REM QB64SHELL - command line prompt shell windows program for QB64

' define all default variables
DEFINT A-Z

' declare redim variables
REM $DYNAMIC

' set all shared variables
REM $INCLUDE:'qb64shell.inc'

' set title icon
$EXEICON:'qb64shell.ico'
_ICON

' declare standard error trap
ON ERROR GOTO Error.Routine

' store command arrays
CALL InitData

' reset multi-dimensional command history arrays
FOR VarQ1 = 1 TO 3
    HistoryCount2(VarQ1) = 0
    FOR VarQ2 = 1 TO MaxHistory
        History2(VarQ1, VarQ2) = ""
    NEXT
NEXT

' initialize some variables
Token.List = " -+*/\^()[]{}<>=|&!%~?:#@`;," + CHR$(34) ' cannot be changed
White.Space = CHR$(32) + CHR$(9) ' can be changed

' force default path
x$ = _STARTDIR$
StartDir$ = x$
IF RIGHT$(StartDir$, 1) <> "\" THEN
    StartDir$ = StartDir$ + "\"
END IF
f$ = x$ + CHR$(0)
x = SetCurrentDirectoryA(f$)

' store default path in table
REDIM DriveTable(26) AS STRING
REDIM NetPathTable(1) AS STRING
REDIM PathTable(1) AS STRING
NetPathCount = 0
PathCount = 1
PathMax = 1
PathTable(1) = _STARTDIR$
IF LEFT$(x$, 2) = "\\" THEN
    NetPathCount = 1
    NetPathTable(1) = _CWD$
ELSE
    V = ASC(LEFT$(x$, 1)) - 64
    DriveTable(V) = _CWD$
END IF

' parse config file
CALL ReadKeys
CALL ReadConfig

' init user profiles
Username = ""
Password = ""
UsernameTemp = ""
CALL InitProfiles

' play intro
IF SoundEffect = 0 THEN
    IF _FILEEXISTS(SoundFile) THEN
        _SNDPLAYFILE (SoundFile)
    END IF
END IF

REM read file menu box variables
' intialize some variables.
CALL InitVars

' initialize server name.
CALL StoreServerName

' read config file.
CALL ReadConfigFile

' get current drive/directory.
CALL GetDrive(Drive.Number, OldDirectory$)

' get file from url
IF SoundEffect = 0 THEN
    CALL GrabURL
END IF

' declare startup settings
_TITLE Program
ON TIMER(1) GOSUB CPUtrap

' start timer trap.
t1 = _FREETIMER
ON TIMER(t1, 1) GOSUB Timer.Trap
TIMER(t1) ON
VarQ = _EXIT

' initialize screen saver
ScreenSaverActive = 0
ScreenSaverTimer = 0!
ScreenSaverResolution = 0!
KeyPressed = 0
MousePressed = 0
 
' start screen saver timer trap.
t2 = _FREETIMER
ON TIMER(t2, 1) GOSUB Saver.Trap
TIMER(t2) ON

' restore screen coordinates
_SCREENMOVE WindowX, WindowY

' get command line
CALL ParseCommandLine
IF CPUOFF = 0 THEN
    TIMER ON
END IF

' start memory monitor timer trap.
MonitorCount = 0
MonitorTimer = 60 ' seconds
MonitorLimit = 10 ' megabytes
t3 = _FREETIMER
ON TIMER(t3, 1) GOSUB Monitor.Trap
TIMER(t3) ON

' define and turn on keys
COLOR 7
CLS
CALL SetKeys
IF KEYOFF = 0 THEN
    KEY ON
END IF

' display program name and version
Status = 0
IF SOFF = 0 THEN
    COLOR 14
    LOCATE 24, 1, 1
    CALL DisplayVersion
    Status = -1
END IF

' display toolbar
CALL Toolbar

' define viewport of text screen
IF SOFF = 0 THEN
    VIEW PRINT 2 TO 23
ELSE
    VIEW PRINT 2 TO 24
END IF

' start prompt input loop
CLS
COLOR 15
IF Status THEN
    LOCATE 23, 1, 1
ELSE
    LOCATE 24, 1, 1
END IF

' get username profile
IF ProfilesActive THEN
    CALL GetUserProfile(V, UsernameTemp)
END IF
StartLoop:
DO
    DO
        ' declare standard error trap
        ON ERROR GOTO Error.Routine
        Error.Resume0:
        Control.Break = False
        Display.Errors = False
        Strip.Blanks = False
        IF StdinHandle THEN
            CLOSE #StdinHandle
        END IF
        DO
            COLOR 15
            LOCATE , 1, 1
            CALL DisplayPrompt
            COLOR 7, 0
            LOCATE , , 1
            Temp0 = 1
            x$ = LTRIM$(KeyboardLine$(1)): PRINT
            ' parse drop down menu selection
            M = CurrentMenu
            N = CurrentMenuSelection
            CurrentMenu = 0
            SELECT CASE M
                CASE 3, 6 ' hex screen/file dump/print
                    COLOR 7, 0
                    PRINT "Error: Only works in screen editor."

                    ' store screen area.
                    CurrentX = CSRLIN
                    CurrentY = POS(0)
                    CALL Toolbar
                    COLOR 7, 0
                    LOCATE CurrentX, CurrentY, 1
                CASE 5 ' jump byte/line
                    COLOR 7, 0
                    PRINT "Error: Jump only works in screen editor."

                    ' store screen area.
                    CurrentX = CSRLIN
                    CurrentY = POS(0)
                    CALL Toolbar
                    COLOR 7, 0
                    LOCATE CurrentX, CurrentY, 1
                CASE 7 ' search byte/line
                    COLOR 7, 0
                    PRINT "Error: Search only works in screen editor."

                    ' store screen area.
                    CurrentX = CSRLIN
                    CurrentY = POS(0)
                    CALL Toolbar
                    COLOR 7, 0
                    LOCATE CurrentX, CurrentY, 1
                CASE 1
                    SELECT CASE N
                        CASE 1 ' open file
                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL SaveScreen

                            ' read information from file menu box.
                            CALL ClearKeys
                            CALL DisplayToolBar
                            CALL Menu3(NetPath2$, OldDrive$, OldDirectory$, FileSpec1$, FileSpec2$)
                            CALL SetKeys

                            ' check filename.
                            Filename = Nul
                            IF FileSpec2$ <> Nul THEN
                                ' construct filename from menu box.
                                IF NetPath2$ = Nul THEN
                                    IF OldDirectory$ = Nul THEN
                                        Filename = OldDrive$ + ":\" + FileSpec2$
                                    ELSE
                                        Filename = OldDrive$ + ":\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                ELSE
                                    IF OldDirectory$ = Nul THEN
                                        Filename = NetPath2$ + "\" + FileSpec2$
                                    ELSE
                                        Filename = NetPath2$ + "\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                END IF
                            END IF

                            ' restore screen lower area.
                            IF Status = 0 THEN
                                VIEW PRINT 2 TO 24
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                                LOCATE 23, 1, 1
                                COLOR 15
                                CALL DisplayPrompt
                                PRINT
                                LOCATE 24, 1, 1
                            ELSE
                                LOCATE 24, 1, 1
                                COLOR 14
                                CALL DisplayVersion
                                VIEW PRINT 2 TO 23
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                            END IF
                            IF KEYOFF = 0 THEN
                                KEY ON
                            END IF

                            ' restore screen area.
                            CALL RestoreScreen
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                            IF LEN(Filename) THEN
                                VIEW PRINT 1 TO 25
                                KEY OFF
                                CLS
                                COLOR 15, 0
                                LOCATE 25, 1
                                VarC = 0
                                Var$ = Filename
                                CALL ClearKeys
                                CALL ScrnEdit(Var$, VarZ, VarC)
                                CALL SetKeys
                                Var1$ = Nul
                                Var2$ = Nul
                                IF VarZ THEN
                                    IF VarZ = -1 THEN
                                        Var1$ = " Saved "
                                        Var2$ = "File stored to disk."
                                    ELSE
                                        IF VarZ = -2 THEN
                                            Var1$ = " Error "
                                            Var2$ = "Error creating file."
                                        ELSE
                                            IF VarZ = -3 THEN
                                                Var1$ = " Error "
                                                Var2$ = "Invalid filename."
                                            ELSE
                                                IF VarZ = -4 THEN
                                                    Var1$ = " Error "
                                                    Var2$ = "Interrupt."
                                                ELSE
                                                    IF VarZ = -5 THEN
                                                        Var1$ = " Error "
                                                        Var2$ = "Overflow."
                                                    END IF
                                                END IF
                                            END IF
                                        END IF
                                    END IF
                                END IF
                                CALL Toolbar
                                IF Status = 0 THEN
                                    VIEW PRINT 2 TO 24
                                    COLOR 7
                                    CLS
                                    CALL RestoreScreen
                                    LOCATE 23, 1, 1
                                    COLOR 15
                                    CALL DisplayPrompt
                                    PRINT
                                    LOCATE 24, 1, 1
                                ELSE
                                    LOCATE 24, 1, 1
                                    COLOR 14
                                    CALL DisplayVersion
                                    VIEW PRINT 2 TO 23
                                    COLOR 7
                                    CLS
                                    CALL RestoreScreen
                                END IF
                                IF KEYOFF = 0 THEN
                                    KEY ON
                                END IF
                                IF LEN(Var1$) THEN
                                    VarX = MessageBox(Var1$, Var2$)
                                END IF
                            END IF
                        CASE 2 ' close file
                            File.Loaded = 0
                            COLOR 7, 0
                            PRINT "File handles closed."

                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL Toolbar
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                        CASE 3 ' print file
                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL SaveScreen

                            ' read information from file menu box.
                            CALL ClearKeys
                            CALL DisplayToolBar
                            CALL Menu3(NetPath2$, OldDrive$, OldDirectory$, FileSpec1$, FileSpec2$)
                            CALL SetKeys

                            ' check filename.
                            Filename = Nul
                            IF FileSpec2$ <> Nul THEN
                                ' construct filename from menu box.
                                IF NetPath2$ = Nul THEN
                                    IF OldDirectory$ = Nul THEN
                                        Filename = OldDrive$ + ":\" + FileSpec2$
                                    ELSE
                                        Filename = OldDrive$ + ":\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                ELSE
                                    IF OldDirectory$ = Nul THEN
                                        Filename = NetPath2$ + "\" + FileSpec2$
                                    ELSE
                                        Filename = NetPath2$ + "\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                END IF
                            END IF

                            ' restore screen lower area.
                            IF Status = 0 THEN
                                VIEW PRINT 2 TO 24
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                                LOCATE 23, 1, 1
                                COLOR 15
                                CALL DisplayPrompt
                                PRINT
                                LOCATE 24, 1, 1
                            ELSE
                                LOCATE 24, 1, 1
                                COLOR 14
                                CALL DisplayVersion
                                VIEW PRINT 2 TO 23
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                            END IF
                            CALL DisplayToolBar
                            IF KEYOFF = 0 THEN
                                KEY ON
                            END IF

                            ' restore screen area.
                            CALL RestoreScreen
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                            ' print file
                            IF LEN(Filename) THEN
                                IF INSTR(Filename, "?") = 0 THEN
                                    x = FREEFILE
                                    OPEN Filename FOR INPUT AS #x
                                    DO UNTIL EOF(x)
                                        LINE INPUT #x, Strng
                                        ' replace unprintable characters
                                        FOR V = 1 TO LEN(Strng)
                                            IF ASC(MID$(Strng, V, 1)) < 32 THEN
                                                MID$(Strng, V, 1) = "."
                                            END IF
                                        NEXT
                                        LPRINT Strng
                                    LOOP
                                    LPRINT CHR$(12);
                                    CLOSE #x
                                END IF
                            END IF
                        CASE 4 ' view file info
                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL SaveScreen

                            ' read information from file menu box.
                            CALL ClearKeys
                            CALL DisplayToolBar
                            CALL Menu3(NetPath2$, OldDrive$, OldDirectory$, FileSpec1$, FileSpec2$)
                            CALL SetKeys

                            ' check filename.
                            Filename = Nul
                            IF FileSpec2$ <> Nul THEN
                                ' construct filename from menu box.
                                IF NetPath2$ = Nul THEN
                                    IF OldDirectory$ = Nul THEN
                                        Filename = OldDrive$ + ":\" + FileSpec2$
                                    ELSE
                                        Filename = OldDrive$ + ":\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                ELSE
                                    IF OldDirectory$ = Nul THEN
                                        Filename = NetPath2$ + "\" + FileSpec2$
                                    ELSE
                                        Filename = NetPath2$ + "\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                END IF
                            END IF

                            ' restore screen lower area.
                            IF Status = 0 THEN
                                VIEW PRINT 2 TO 24
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                                LOCATE 23, 1, 1
                                COLOR 15
                                CALL DisplayPrompt
                                PRINT
                                LOCATE 24, 1, 1
                            ELSE
                                LOCATE 24, 1, 1
                                COLOR 14
                                CALL DisplayVersion
                                VIEW PRINT 2 TO 23
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                            END IF
                            CALL DisplayToolBar
                            IF KEYOFF = 0 THEN
                                KEY ON
                            END IF

                            ' restore screen area.
                            CALL RestoreScreen
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                            IF LEN(Filename) THEN
                                CurrentCommand = 0
                                CommandCount = 0
                                StdoutFile$ = "scrn:"
                                COMMAND2$ = Quote + Filename + Quote + " /C"
                                CALL ViewFileInfo(Filename)
                            END IF
                        CASE 5 ' type file
                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL SaveScreen

                            ' read information from file menu box.
                            CALL ClearKeys
                            CALL DisplayToolBar
                            CALL Menu3(NetPath2$, OldDrive$, OldDirectory$, FileSpec1$, FileSpec2$)
                            CALL SetKeys

                            ' check filename.
                            Filename = Nul
                            IF FileSpec2$ <> Nul THEN
                                ' construct filename from menu box.
                                IF NetPath2$ = Nul THEN
                                    IF OldDirectory$ = Nul THEN
                                        Filename = OldDrive$ + ":\" + FileSpec2$
                                    ELSE
                                        Filename = OldDrive$ + ":\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                ELSE
                                    IF OldDirectory$ = Nul THEN
                                        Filename = NetPath2$ + "\" + FileSpec2$
                                    ELSE
                                        Filename = NetPath2$ + "\" + OldDirectory$ + "\" + FileSpec2$
                                    END IF
                                END IF
                            END IF

                            ' restore screen lower area.
                            IF Status = 0 THEN
                                VIEW PRINT 2 TO 24
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                                LOCATE 23, 1, 1
                                COLOR 15
                                CALL DisplayPrompt
                                PRINT
                                LOCATE 24, 1, 1
                            ELSE
                                LOCATE 24, 1, 1
                                COLOR 14
                                CALL DisplayVersion
                                VIEW PRINT 2 TO 23
                                COLOR 7
                                CLS
                                CALL RestoreScreen
                            END IF
                            CALL DisplayToolBar
                            IF KEYOFF = 0 THEN
                                KEY ON
                            END IF

                            ' restore screen area.
                            CALL RestoreScreen
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                            IF LEN(Filename) THEN
                                IF INSTR(Filename, "?") = 0 THEN
                                    CurrentCommand = 0
                                    CommandCount = 0
                                    StdoutFile$ = "scrn:"
                                    COMMAND2$ = Quote + Filename + Quote + " /C"
                                    CALL TypeY(Var)
                                END IF
                            END IF
                        CASE 6 ' exit program
                            IF QuitBox THEN
                                CALL WriteConfig
                                CALL WriteProfile
                                COLOR 7, 0
                                SYSTEM
                            END IF
                    END SELECT
                CASE 2
                    SELECT CASE N
                        CASE 1 ' ascii chart
                            CALL AsciiChart
                            CALL DisplayToolBar
                        CASE 2 ' hex chart
                            CALL HexChart
                            CALL DisplayToolBar
                    END SELECT
                CASE 4
                    SELECT CASE N
                        CASE 1, 2 ' append byte/string
                            COLOR 7, 0
                            PRINT "Error: Append only works in screen editor."

                            ' store screen area.
                            CurrentX = CSRLIN
                            CurrentY = POS(0)
                            CALL Toolbar
                            COLOR 7, 0
                            LOCATE CurrentX, CurrentY, 1
                        CASE 3 ' redraw
                            VIEW PRINT 1 TO 25
                            KEY OFF
                            CLS
                            COLOR 15, 0
                            LOCATE 25, 1
                            CALL Toolbar
                            IF Status = 0 THEN
                                VIEW PRINT 2 TO 24
                                COLOR 7
                                LOCATE 24, 1, 1
                            ELSE
                                LOCATE 24, 1, 1
                                COLOR 14
                                CALL DisplayVersion
                                VIEW PRINT 2 TO 23
                                COLOR 7
                                LOCATE 23, 1, 1
                            END IF
                            IF KEYOFF = 0 THEN
                                KEY ON
                            END IF
                    END SELECT
            END SELECT
            IF LEN(x$) THEN
                EXIT DO
            END IF
        LOOP
        Display.Errors = False
        Strip.Blanks = False
        IF LEFT$(UCASE$(x$), 5) = "SHELL" THEN
            CALL ShellSub(x$)
            GOTO StartLoop
        END IF
        IF LEN(x$) = 2 THEN
            IF RIGHT$(x$, 1) = ":" THEN
                CALL SwitchDrive(LEFT$(x$, 1))
                GOTO StartLoop
            END IF
        END IF
        IF LEFT$(x$, 2) = "\\" THEN
            CALL SwitchNetpath(x$)
            GOTO StartLoop
        END IF

        ' read all redirect symbols
        CALL PreParse(x$, V, z)
        IF V = -1 THEN
            x = -1
            PRINT "Error: ";
            SELECT CASE z
                CASE 1
                    PRINT "Multiple redirect symbols."
                CASE 2
                    PRINT "Multiple redirect error symbols."
                CASE 3
                    PRINT "Multiple redirect output symbols."
                CASE 4
                    PRINT "Invalid redirect filename characters."
                CASE 5
                    PRINT "Cannot open stdin filename."
                CASE 6
                    PRINT "Cannot open stdout filename."
                CASE 7
                    PRINT "Cannot open stderr filename."
            END SELECT
        ELSE
            ' close output handle
            IF StdoutHandle THEN
                CLOSE StdoutHandle
            END IF

            ' open output handle
            StdoutHandle = FREEFILE
            IF Stdout = 0 THEN
                StdoutFile$ = "scrn:"
                OPEN "scrn:" FOR OUTPUT AS StdoutHandle
            ELSE
                IF Stdout = 1 THEN
                    OPEN StdoutFile$ FOR OUTPUT AS StdoutHandle
                ELSE
                    IF Stdout = 2 THEN
                        OPEN StdoutFile$ FOR APPEND AS StdoutHandle
                    END IF
                END IF
            END IF

            ' close error handle
            IF StderrHandle THEN
                CLOSE StderrHandle
            END IF

            ' open error handle
            StderrHandle = FREEFILE
            IF Stderr = 0 THEN
                StderrFile$ = "scrn:"
                OPEN "scrn:" FOR OUTPUT AS StderrHandle
            ELSE
                IF Stderr = -1 THEN
                    OPEN StderrFile$ FOR APPEND AS StderrHandle
                END IF
            END IF

            ' read stdin
            CALL PreFetch(V)

            ' reset timers
            TIMER OFF
            _TITLE Program
            Display.Errors = False
            Strip.Blanks = False
            CALL SimpleParse(x$, x)
            IF CPUOFF = 0 THEN
                TIMER ON
            ELSE
                _TITLE Program
            END IF
        END IF
        IF x = -2 THEN
            ' declare error routine
            ON ERROR GOTO Whatis.Error.Routine
            Error.Resume1:

            ' whatis input loop
            Control.Break = False
            DO
                IF LEN(x$) THEN
                    Error.Trap.Type = 0
                    Out2 = x$
                    x$ = ""
                    Parse.Type = 1
                    Recurse = 0
                    CALL Enter.Equate
                    EXIT DO
                END IF
                Error.Trap.Type = -1
                COLOR Yellow, Black
                LOCATE , 1, 1
                PRINT "Whatis>";
                LOCATE , , 1
                ' get input line
                COLOR 7, 0
                Temp0 = 2
                Out2 = KeyboardLine$(2): PRINT
                ' parse input line
                CALL Remove.Spaces
                IF UCASE$(Out2) = "QUIT" THEN
                    EXIT DO
                END IF
                ' calculate line
                Parse.Type = 2
                Recurse = 0
                CALL Enter.Equate
            LOOP
        END IF
        IF x = 0 THEN
            TIMER OFF
            _TITLE Program
            CALL ComplexInit(x$, x, z)
            IF x = 0 THEN
                x = -1
                IF z = 1 THEN
                    PRINT "Error: Missing command. Type HELP for help."
                END IF
                IF z = 2 THEN
                    PRINT "Error: Mismatched brackets. Type HELP for help."
                END IF
                IF z = 3 THEN
                    PRINT "Error: Unknown command. Type HELP for help."
                END IF
            ELSE
                FOR CurrentCommand = 1 TO CommandCount
                    ' copy stdout to stdin
                    IF CurrentCommand > 1 THEN
                        IF StdoutLines > StdinMax THEN
                            StdinMax = StdoutLines
                            REDIM _PRESERVE StdinArray(StdinMax) AS STRING
                        END IF
                        FOR StdoutCount = 1 TO StdoutLines
                            StdinArray(StdoutCount) = StdoutArray(StdoutCount)
                        NEXT
                        StdinLines = StdoutLines
                    END IF
                    Display.Errors = False
                    Strip.Blanks = False
                    ' process next command in stream
                    x$ = CommandList(CurrentCommand)
                    CALL ComplexParse(x$, x)
                    IF x = 0 THEN ' command not found
                        EXIT FOR
                    END IF
                    IF x = -2 THEN ' command line switch error
                        EXIT FOR
                    END IF
                NEXT
            END IF
            IF CPUOFF = 0 THEN
                TIMER ON
            ELSE
                _TITLE Program
            END IF
        END IF
        IF x = 0 THEN
            EXIT DO
        END IF
    LOOP
    PRINT "Type HELP for help."
LOOP
SLEEP 5
END

CPUtrap:
S$ = STR$(INT(GetCPULoad * 10000) / 100) ' 100.00
IF INSTR(S$, ".") THEN
    Q$ = MID$(S$, INSTR(S$, ".") + 1)
    S$ = LEFT$(S$, INSTR(S$, ".") - 1)
    Q$ = LEFT$(Q$, 3)
    S$ = S$ + "." + Q$
ELSE
    S$ = S$ + ".0"
END IF

' CPU 100.00%
D$ = "CPU " + S$ + "%" ' 100.00
D$ = LEFT$(D$, 12): D$ = D$ + SPACE$(12 - LEN(D$))
IF CPUMON THEN
    Var# = CDBL(FreePhysicalMem)
    SELECT CASE CPUMON
        CASE 1 ' Megabyte
            Var# = Var# / 1048576#
            Var$ = FormatStringX$(Var#, 2) + " MB. "
        CASE 2 ' Gigabyte
            Var# = Var# / 1073741824#
            Var$ = FormatStringX$(Var#, 3) + " GB. "
    END SELECT
    D$ = D$ + " MEM " + Var$
END IF
_TITLE Program + " - " + D$
RETURN

' whatis error routine
Whatis.Error.Routine:
Error.Mode = ERR
SELECT CASE Error.Mode
    ' standard errors.
    CASE 7
        PRINT #StderrHandle, "Out of memory."
    CASE 9
        PRINT #StderrHandle, "Subscript out of range."
    CASE 14
        PRINT #StderrHandle, "Out of string space."
    CASE 53
        PRINT #StderrHandle, "File not found."
    CASE 57
        PRINT #StderrHandle, "Media error."
    CASE 61
        PRINT #StderrHandle, "Disk full."
    CASE 70
        PRINT #StderrHandle, "Permission denied."
    CASE 71
        PRINT #StderrHandle, "Disk not ready."
        ' extended errors.
    CASE 5
        PRINT #StderrHandle, "Error #001: Illegal function call."
    CASE 6
        PRINT #StderrHandle, "Error #002: Overflow."
    CASE 9
        PRINT #StderrHandle, "Error #003: Subscript out of range."
    CASE 11
        PRINT #StderrHandle, "Error #004: Division by zero."
    CASE 92
        PRINT #StderrHandle, "Error #005: Whatis: "; Strng
    CASE 130
        PRINT #StderrHandle, "Error #006: Bad numeric specifier."
    CASE 140
        PRINT #StderrHandle, "Error #007: Bad exponent specifier."
    CASE 145
        PRINT #StderrHandle, "Error #008: Bad DEFFN declaration."
    CASE ELSE
        PRINT #StderrHandle, "Critical error: '" + STR$(Error.Mode) + "' in Whatis."
        '    PRINT "Error #000: Error"; Error.Mode
END SELECT
IF Error.Trap.Type THEN
    RESUME Error.Resume1
ELSE
    RESUME Error.Resume0
END IF

' critical error trap
Error.Routine:
' process error
DataError = ERR
IF Error.Flag THEN
    RESUME NEXT
END IF
IF Display.Errors THEN
    Error.Level = True
    RESUME NEXT
END IF
' check share violations.
IF ERR = 70 THEN
    Share.Count = Share.Count + 1
    IF Share.Count < MaxViolations THEN
        RESUME
    END IF
    Share.Count = False
END IF

IF Wide.List THEN
    Wide.List = False
    PRINT
END IF

' display error
COLOR Green, Black
PRINT "Critical error:"; DataError; " IDE line:"; _ERRORLINE
Prompt2$ = "Press C to continue, R to retry, A to abort, Q to quit:"
CALL MorePrompt(Prompt2$, "craq", Outpt$)
SELECT CASE Outpt$
    CASE "a"
        END
    CASE "q"
        Error.Level = True
        RESUME Error.Resume0
    CASE "r"
        RESUME
    CASE "c"
        RESUME NEXT
END SELECT
COLOR Plain, Black
END 0

' screen saver timer trap.
Saver.Trap:
IF ScreenSaverActive = 0 THEN
    RETURN
END IF
IF KeyPressed OR MousePressed THEN
    KeyPressed = 0
    MousePressed = 0
    ScreenSaverTimer = TIMER
    RETURN
END IF
IF ScreenSaverResolution = 0! THEN
    RETURN
END IF
IF ScreenSaverTimer = 0! THEN
    RETURN
END IF

' start screen saver
Elapsed! = TIMER - ScreenSaverTimer
IF Elapsed! < 0! THEN Elapsed! = Elapsed! + 86400!
IF Elapsed! >= ScreenSaverResolution THEN
    ScreenSaverTimer = 0! ' disable timer until next keypress.
    s%& = SendMessageA(&HFFFF&, wm_syscommand, sc_screensave, 0&)
END IF
RETURN

' timer ctrl-break trap.
Timer.Trap:
VarQ = _EXIT
IF VarQ THEN ' Control-Break
    Control.Break = True
    IF BreakBox = 0 THEN
        SYSTEM
    END IF
END IF
RETURN

' memory monitor trap
Monitor.Trap:
IF MONITOROFF THEN
    RETURN
END IF
IF MonitorLimit = 0# THEN
    RETURN
END IF
IF MonitorTimer = 0 THEN
    RETURN
END IF
MonitorCount = MonitorCount + 1
IF MonitorCount <= MonitorTimer THEN
    RETURN
END IF
MonitorCount = 0
X# = MonitorLimit * 1048576#
IF FreePhysicalMem > X# THEN RETURN
s1$ = LTRIM$(STR$(MonitorTimer))
s2$ = LTRIM$(STR$(MonitorLimit))
r = MonitorBox(s1$, s2$)
IF DisplaySelect THEN
    MONITOROFF = True
    RETURN
END IF
IF r = 0 THEN RETURN ' exit
IF r = 2 THEN SYSTEM ' system
IF r = 1 THEN ' continue
    s1$ = LTRIM$(RTRIM$(s1$))
    s2$ = LTRIM$(RTRIM$(s2$))
    IF IsNumeric(s1$) THEN
        IF IsSingle(s2$) THEN
            X# = VAL(s1$)
            Y# = VAL(s2$)
            IF X# = 0# AND Y# = 0# THEN
                MONITOROFF = True
            END IF
            IF X# > 0# AND Y# > 0# THEN
                IF X# <= 86400# THEN
                    X! = CSNG(X#)
                    MONITOROFF = False
                    MonitorTimer = X!
                    MonitorLimit = Y#
                END IF
            END IF
        END IF
    END IF
END IF
RETURN

' display prompt text with macros
SUB DisplayPrompt
DIM StoreX AS INTEGER, StoreY AS INTEGER
IF PromptForeground OR PromptBackground THEN
    COLOR PromptForeground, PromptBackground
END IF
Length = 0
Var$ = Prompt$
DO
    IF LEFT$(Var$, 1) = "$" THEN
        Param$ = LEFT$(Var$, 2)
        Param$ = MID$(Param$, 2)
        Var$ = MID$(Var$, 3)
        IF UCASE$(Param$) = "F" THEN
            IF LEFT$(Var$, 1) = "[" THEN
                V = INSTR(2, Var$, "]")
                IF V THEN
                    Param$ = LEFT$(Var$, V - 1)
                    Param$ = MID$(Param$, 2)
                    IF Param$ <> "" THEN ' x;y
                        T = INSTR(Param$, ";")
                        IF T THEN
                            IF Status THEN
                                Z = 23
                            ELSE
                                Z = 24
                            END IF
                            X1 = VAL(LEFT$(Param$, T - 1))
                            Y1 = VAL(MID$(Param$, T + 1))
                            IF X1 >= 2 AND X1 <= Z THEN
                                IF Y1 >= 1 AND Y1 <= 80 THEN
                                    LOCATE X1, Y1, 1
                                END IF
                            END IF
                        END IF
                    END IF
                    Var$ = MID$(Var$, V + 1)
                END IF
            END IF
        END IF
        IF UCASE$(Param$) = "C" THEN
            IF LEFT$(Var$, 1) = "[" THEN
                V = INSTR(2, Var$, "]")
                IF V THEN
                    Param$ = LEFT$(Var$, V - 1)
                    Param$ = MID$(Param$, 2)
                    IF Param$ <> "" THEN
                        T = INT(VAL(Param$))
                        IF T >= 0 AND T <= 15 THEN
                            COLOR T
                        END IF
                    END IF
                    Var$ = MID$(Var$, V + 1)
                END IF
            END IF
        END IF
        IF UCASE$(Param$) = "E" THEN
            IF LEFT$(Var$, 1) = "[" THEN
                V = INSTR(2, Var$, "]")
                IF V THEN
                    Param$ = LEFT$(Var$, V - 1)
                    Param$ = MID$(Param$, 2)
                    IF Param$ <> "" THEN
                        T = INT(VAL(Param$))
                        IF T >= 0 AND T <= 7 THEN
                            COLOR , T
                        END IF
                    END IF
                    Var$ = MID$(Var$, V + 1)
                END IF
            END IF
        END IF
        IF UCASE$(Param$) = "J" THEN
            Length = 0
            CLS
        END IF
        IF UCASE$(Param$) = "P" THEN
            Length = Length + LEN(_CWD$)
            PRINT _CWD$;
        END IF
        IF UCASE$(Param$) = "N" THEN
            Length = Length + 1
            PRINT LEFT$(_CWD$, 1);
        END IF
        IF UCASE$(Param$) = "D" THEN
            Length = Length + 10
            PRINT DATE$;
        END IF
        IF UCASE$(Param$) = "T" THEN
            Length = Length + 8
            PRINT TIME$;
        END IF
        IF UCASE$(Param$) = "V" THEN
            Length = Length + LEN(Version$)
            PRINT Version$;
        END IF
        IF UCASE$(Param$) = "Q" THEN
            GOSUB DisplayVolume
        END IF
        IF UCASE$(Param$) = "S" THEN
            StoreX = CSRLIN
            StoreY = POS(0)
        END IF
        IF UCASE$(Param$) = "U" THEN
            IF StoreX > 0 AND StoreY > 0 THEN
                LOCATE StoreX, StoreY, 1
            END IF
        END IF
        IF UCASE$(Param$) = "_" THEN
            Length = 0
            PRINT
        END IF
        IF UCASE$(Param$) = "$" THEN
            Length = Length + 1
            PRINT "$";
        END IF
        IF UCASE$(Param$) = "G" THEN
            Length = Length + 1
            PRINT ">";
        END IF
        IF UCASE$(Param$) = "L" THEN
            Length = Length + 1
            PRINT "<";
        END IF
        IF UCASE$(Param$) = "H" THEN ' backspace
            IF Length > 0 THEN
                IF POS(0) > 1 THEN ' check cursor location
                    Length = Length - 1
                    LOCATE CSRLIN, POS(0) - 1, 0
                    PRINT " ";
                    LOCATE CSRLIN, POS(0) - 1, 1
                END IF
            END IF
        END IF
    ELSE
        Length = Length + 1
        _CONTROLCHR OFF
        PRINT LEFT$(Var$, 1);
        _CONTROLCHR ON
        LOCATE , , 1
        Var$ = MID$(Var$, 2)
    END IF
    IF Var$ = "" THEN EXIT DO
LOOP
EXIT SUB

' display drive volume for prompt
DisplayVolume:
IF LEFT$(_CWD$, 2) = "\\" THEN
    V = INSTR(3, _CWD$, "\")
    IF V THEN
        Out3$ = LEFT$(_CWD$, V)
    ELSE
        Out3$ = "\\"
    END IF
    Length = Length + LEN(Out3)
    PRINT Out3$;
ELSE
    f$ = UCASE$(LEFT$(_CWD$, 1))
    IF f$ >= "A" AND f$ <= "Z" THEN
        V = GetDriveExists(f$)
        IF V THEN
            CALL Vlabel(f$)
            IF RTRIM$(f$) = Nul THEN
                Out3$ = DriveType
            ELSE
                Out3$ = RTRIM$(f$)
            END IF
            Length = Length + LEN(Out3)
            PRINT Out3$;
        END IF
    END IF
END IF
RETURN
END SUB

SUB ParseCommandLine
Var$ = UCASE$(COMMAND$)
IF Var$ = "/?" THEN
    COLOR 14, 0
    PRINT "Program "; Program; " command line."
    COLOR 15, 0
    PRINT "/CPU[ON|OFF]"
    PRINT "   toggles CPU monitor"
    PRINT "/KEYS[ON|OFF]"
    PRINT "   toggles 25th row keys list"
    PRINT "/STATUS[ON|OFF]"
    PRINT "   toggles 24th row status line"
    PRINT "/MONITOR[ON|OFF]"
    PRINT "   toggles low memory usage monitor"
    PRINT "/CPUMON[0|1|2]"
    PRINT "   toggles memory display monitor"
    PRINT "   0=off/1=MB/2=GB"
    PRINT "/NODE:[n]"
    PRINT "   n is 1 to 9."
    PRINT "/USERNAME:[USER]"
    PRINT "   is profile username"
    COLOR 7, 0
    END
END IF
V = INSTR(Var$, "/STATUS:OFF")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 12)
    SOFF = -1
END IF
V = INSTR(Var$, "/STATUS:ON")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 11)
    SOFF = 0
END IF
V = INSTR(Var$, "/KEYS:OFF")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 9)
    KEYOFF = -1
END IF
V = INSTR(Var$, "/KEYS:ON")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 8)
    KEYOFF = 0
END IF
V = INSTR(Var$, "/CPU:OFF")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 8)
    CPUOFF = -1
END IF
V = INSTR(Var$, "/CPU:ON")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 7)
    CPUOFF = 0
END IF

V = INSTR(Var$, "/CPUMON:0")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 9)
    CPUMON = 0
END IF
V = INSTR(Var$, "/CPUMON:1")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 9)
    CPUMON = 1
END IF
V = INSTR(Var$, "/CPUMON:2")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 9)
    CPUMON = 2
END IF

V = INSTR(Var$, "/MONITOR:OFF")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 12)
    MONITOROFF = -1
END IF
V = INSTR(Var$, "/MONITOR:ON")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 11)
    MONITOROFF = 0
END IF
V = INSTR(Var$, "/DEBUG:ON")
IF V THEN
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 9)
    Debug = -1
END IF

V = INSTR(Var$, "/NODE:")
IF V THEN
    V$ = MID$(Var$, V + 7, 1)
    Var$ = LEFT$(Var$, V - 1) + MID$(Var$, V + 8)
    IF V$ >= "1" AND Var$ <= "9" THEN
        Node = VAL(V$)
    END IF
END IF

V = INSTR(Var$, "/USERNAME:")
IF V THEN
    UsernameTemp = MID$(Var$, V + 10)
    UsernameTemp = LTRIM$(RTRIM$(UsernameTemp))
    Var$ = LEFT$(Var$, V - 1)
END IF
END SUB

SUB InitProfiles
DIM hfind AS _OFFSET

' see if profiles datafile exists.
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) THEN
    X = FREEFILE
    OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
    IF LOF(X) > 0 THEN
        CLOSE X
        EXIT SUB
    END IF
    CLOSE X
END IF

' force file creation
f$ = GetProfileFilename$ + CHR$(0)
hfind = CreateFile(f$, &H180, 0, 0, 2, 0, 0)
IF hfind = 0 THEN
    EXIT SUB
END IF
r = CloseHandle(hfind)

' open profiles datafile
f$ = GetProfileFilename$
X = FREEFILE
DataError = 0
Error.Flag = -1
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
Error.Flag = 0
IF DataError THEN
    EXIT SUB
END IF

' write first record
UserProfileRecord.Username = "SYSOP"
UserProfileRecord.Password = ""
UserProfileRecord.ScreenSaver = 0
UserProfileRecord.ScreenTimer = 0
UserProfileRecord.LastDateOn = DATE$ + " " + TIME$ ' 01/01/1980 HH:MM:SS
UserProfileRecord.CreationDate = DATE$ + " " + TIME$
UserProfileRecord.TimesLoggedOn = 0#

UserProfileRecord.ASCIIZreserved1 = ""
UserProfileRecord.ASCIIZreserved2 = ""
UserProfileRecord.DOSreserved1 = ""
UserProfileRecord.DOSreserved2 = ""
UserProfileRecord.TimesDeleted = 0
UserProfileRecord.TimesUndeleted = 0

UserProfileRecord.MonitorOff = 0
UserProfileRecord.MonitorTimer = 60 ' seconds
UserProfileRecord.MonitorLimit = 10 ' megabytes

UserProfileRecord.Reserved1 = 0%
UserProfileRecord.Reserved2 = 0%
UserProfileRecord.Reserved3 = 0!
UserProfileRecord.Reserved4 = 0!
UserProfileRecord.Reserved5 = 0&
UserProfileRecord.Reserved6 = 0&
UserProfileRecord.Reserved7 = 0#
UserProfileRecord.Reserved8 = 0#

UserProfileRecord.Deleted = 0

PUT #X, 1, UserProfileRecord
CLOSE X
END SUB

' returns Var = True if valid logon
'   Var$ is username override.
SUB GetUserProfile (Var, Var$)
Var = 0
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
GET #X, 1, UserProfileRecord
IF RTRIM$(UserProfileRecord.Password) = "" THEN
    UserProfileRecord.Password = EnCrypt$("PASSWORD")
    PUT X, 1, UserProfileRecord
    COLOR 15
    PRINT "Initializing User Profiles";
    FOR V = 1 TO 3
        _DELAY 1: PRINT ". ";
    NEXT
    PRINT "UN: SYSOP PW: PASSWORD"
END IF
IF LEN(Var$) THEN
    X$ = Var$
    FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
        GET X, Z!, UserProfileRecord
        IF X$ = RTRIM$(UserProfileRecord.Username) THEN
            IF UserProfileRecord.Deleted = 0 THEN
                Username = X$
                ScreenSaverActive = UserProfileRecord.ScreenSaver
                ScreenSaverTimer = 0!
                ScreenSaverResolution = UserProfileRecord.ScreenTimer
                MONITOROFF = UserProfileRecord.MonitorOff
                MonitorTimer = UserProfileRecord.MonitorTimer ' seconds
                MonitorLimit = UserProfileRecord.MonitorLimit ' megabytes
                CLOSE X
                Var = -1
                EXIT SUB
            END IF
        END IF
    NEXT
    PRINT "Unknown username/password."
    CLOSE X
    EXIT SUB
END IF
V = 0
DO
    COLOR 15
    PRINT "Username?";
    X$ = LineInput$(CSRLIN, POS(0), 20): PRINT
    COLOR 15
    PRINT "Password?";
    MaskInput = -1
    Y$ = LineInput$(CSRLIN, POS(0), 8): PRINT
    MaskInput = 0
    _KEYCLEAR
    V = V + 1
    IF X$ = "" OR Y$ = "" OR V > MaxTries THEN
        V = 0
        PRINT "Username entry failed."
        CALL MorePrompt("Exit logon(y/n)?", "yn", Outpt$)
        IF Outpt$ = "y" THEN
            CLOSE X
            EXIT SUB
        END IF
    END IF
    X$ = UCASE$(X$)
    Y$ = UCASE$(Y$)
    FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
        GET X, Z!, UserProfileRecord
        IF X$ = RTRIM$(UserProfileRecord.Username) THEN
            IF EnCrypt$(Y$) = RTRIM$(UserProfileRecord.Password) THEN
                IF UserProfileRecord.Deleted = 0 THEN
                    Username = X$
                    Password = Y$
                    ScreenSaverActive = UserProfileRecord.ScreenSaver
                    ScreenSaverTimer = 0!
                    ScreenSaverResolution = UserProfileRecord.ScreenTimer
                    MONITOROFF = UserProfileRecord.MonitorOff
                    MonitorTimer = UserProfileRecord.MonitorTimer ' seconds
                    MonitorLimit = UserProfileRecord.MonitorLimit ' megabytes
                    CLOSE X
                    Var = -1
                    EXIT SUB
                END IF
            END IF
        END IF
    NEXT
    PRINT "Unknown username/password."
LOOP
END SUB

SUB CountProfiles (Var!)
Var! = 0
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
    GET X, Z!, UserProfileRecord
    IF RTRIM$(UserProfileRecord.Username) > "" THEN
        IF UserProfileRecord.Deleted = 0 THEN
            Var! = Var! + 1!
        END IF
    END IF
NEXT
CLOSE X
END SUB

SUB CountDeletedProfiles (Var!)
Var! = 0
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
    GET X, Z!, UserProfileRecord
    IF RTRIM$(UserProfileRecord.Username) > "" THEN
        IF UserProfileRecord.Deleted THEN
            Var! = Var! + 1!
        END IF
    END IF
NEXT
CLOSE X
END SUB

SUB FindProfile
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
LineCount = 2
IF Status THEN
    LineBreak = 23
ELSE
    LineBreak = 24
END IF
PRINT "Enter username(wildcard ? and * allowed):"
PRINT "Enter: ";
Var$ = LineInput$(CSRLIN, POS(0), 40): PRINT
IF Var$ = "" THEN
    Var$ = "*"
END IF
Var$ = UCASE$(Var$)
COLOR 15
PRINT "Username               Lastdate              Timesloggedin"
PRINT "--------------------   -------------------   -----------------"
Count = 0
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
    GET X, Z!, UserProfileRecord
    IF RTRIM$(UserProfileRecord.Username) > "" THEN
        X$ = RTRIM$(UserProfileRecord.Username)
        CALL InstrSUB1(Flag, Var$, X$, 0)
        IF Flag THEN
            Count = -1
            Outpt$ = UserProfileRecord.Username + "   "
            Outpt$ = Outpt$ + UserProfileRecord.LastDateOn + "   "
            IF UserProfileRecord.Deleted THEN
                Outpt$ = Outpt$ + "<deleted>"
            ELSE
                Outpt$ = Outpt$ + FormatString$(UserProfileRecord.TimesLoggedOn)
            END IF
            PRINT Outpt$
            LineCount = LineCount + 1
            IF LineCount >= LineBreak THEN
                LineCount = 0
                CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt2$)
                IF Outpt2$ = "n" THEN
                    EXIT FOR
                END IF
            END IF
        END IF
    END IF
NEXT
IF Count = 0 THEN
    PRINT "No users found."
END IF
IF LineCount THEN
    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END IF
CLOSE X
END SUB

SUB ListProfiles
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
LineCount = 2
IF Status THEN
    LineBreak = 23
ELSE
    LineBreak = 24
END IF
COLOR 15
PRINT "Username               Lastdate              Timesloggedin"
PRINT "--------------------   -------------------   -----------------"
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
    GET X, Z!, UserProfileRecord
    IF RTRIM$(UserProfileRecord.Username) > "" THEN
        Outpt$ = UserProfileRecord.Username + "   "
        Outpt$ = Outpt$ + UserProfileRecord.LastDateOn + "   "
        IF UserProfileRecord.Deleted THEN
            Outpt$ = Outpt$ + "<deleted>"
        ELSE
            Outpt$ = Outpt$ + FormatString$(UserProfileRecord.TimesLoggedOn)
        END IF
        PRINT Outpt$
        LineCount = LineCount + 1
        IF LineCount >= LineBreak THEN
            LineCount = 0
            CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt2$)
            IF Outpt2$ = "n" THEN
                EXIT FOR
            END IF
        END IF
    END IF
NEXT
IF LineCount THEN
    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END IF
CLOSE X
END SUB

SUB WriteProfile
IF Username = "" THEN EXIT SUB

f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF

X$ = UCASE$(Username$)
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord)
    GET X, Z!, UserProfileRecord
    IF X$ = RTRIM$(UserProfileRecord.Username) THEN
        UserProfileRecord.ScreenSaver = ScreenSaverActive
        UserProfileRecord.ScreenTimer = ScreenSaverResolution
        UserProfileRecord.LastDateOn = DATE$ + " " + TIME$ ' 01/01/1980 HH:MM:SS
        UserProfileRecord.TimesLoggedOn = UserProfileRecord.TimesLoggedOn + 1#
        UserProfileRecord.MonitorOff = MONITOROFF
        UserProfileRecord.MonitorTimer = MonitorTimer ' seconds
        UserProfileRecord.MonitorLimit = MonitorLimit ' megabytes

        PUT X, Z!, UserProfileRecord
        CLOSE X
        EXIT SUB
    END IF
NEXT
CLOSE X
END SUB

SUB ReadConfig
ProfilesActive = -1
Prompt$ = "$P>"
WindowX = _SCREENX: WindowY = _SCREENY
f$ = GetConfigFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
DataError = 0
Error.Flag = -1
OPEN f$ FOR INPUT AS #X
Error.Flag = 0
IF DataError THEN
    EXIT SUB
END IF
IF EOF(X) = 0 THEN
    INPUT #X, WindowX
END IF
IF EOF(X) = 0 THEN
    INPUT #X, WindowY
END IF
DO UNTIL EOF(X)
    LINE INPUT #X, v$
    z$ = v$
    v$ = UCASE$(v$)
    IF v$ = "/STATUS:OFF" THEN
        SOFF = -1
    END IF
    IF v$ = "/STATUS:ON" THEN
        SOFF = 0
    END IF
    IF v$ = "/KEYS:OFF" THEN
        KEYOFF = -1
    END IF
    IF v$ = "/KEYS:ON" THEN
        KEYOFF = 0
    END IF
    IF v$ = "/CPU:OFF" THEN
        CPUOFF = -1
    END IF
    IF v$ = "/CPU:ON" THEN
        CPUOFF = 0
    END IF

    IF v$ = "/CPUMON:0" THEN
        CPUMON = 0
    END IF
    IF v$ = "/CPUMON:1" THEN
        CPUMON = 1
    END IF
    IF v$ = "/CPUMON:2" THEN
        CPUMON = 2
    END IF

    IF v$ = "/MONITOR:OFF" THEN
        MONITOROFF = -1
    END IF
    IF v$ = "/MONITOR:ON" THEN
        MONITOROFF = 0
    END IF
    IF v$ = "SOUNDEFFECT=-1" THEN
        SoundEffect = -1
    END IF
    IF LEFT$(v$, 7) = "PROMPT=" THEN
        Prompt$ = MID$(z$, 8)
    END IF
    IF LEFT$(v$, 14) = "PROFILESACTIVE" THEN
        q$ = MID$(v$, 15)
        q$ = LTRIM$(q$)
        IF LEFT$(q$, 1) = "=" THEN
            q$ = MID$(q$, 2)
            q$ = LTRIM$(q$)
            q$ = RTRIM$(q$)
            IF q$ = "0" THEN
                ProfilesActive = 0
            END IF
            IF q$ = "OFF" THEN
                ProfilesActive = 0
            END IF
            IF q$ = "FALSE" THEN
                ProfilesActive = 0
            END IF
        END IF
    END IF
    v = 0
    q$ = ""
    IF LEFT$(v$, 4) = "KEY " THEN ' key 1=EDIT
        IF MID$(v$, 6, 1) = "=" THEN
            v = VAL(MID$(v$, 5, 1))
            q$ = MID$(z$, 7)
        ELSE
            IF MID$(v$, 7, 1) = "=" THEN
                v = VAL(MID$(v$, 5, 2))
                q$ = MID$(z$, 8)
            END IF
        END IF
        IF v >= 1 AND v <= 12 THEN
            IF RIGHT$(q$, 3) = "+cr" THEN
                FunctionKeys(v) = LEFT$(q$, LEN(q$) - 3) + CHR$(13)
            ELSE
                FunctionKeys(v) = q$
            END IF
        END IF
    END IF
LOOP
CLOSE #X
END SUB

SUB WriteConfig
DIM hfind AS _OFFSET

' force file creation
f$ = GetConfigFilename$ + CHR$(0)
hfind = CreateFile(f$, &H180, 0, 0, 2, 0, 0)
IF hfind = 0 THEN
    EXIT SUB
END IF
r = CloseHandle(hfind)

' open file for output
X = FREEFILE
DataError = 0
Error.Flag = -1
f$ = GetConfigFilename$
OPEN f$ FOR OUTPUT AS #X
Error.Flag = 0
IF DataError THEN
    EXIT SUB
END IF

' write values to file
PRINT #X, _SCREENX
PRINT #X, _SCREENY
IF SOFF = -1 THEN
    PRINT #X, "/STATUS:OFF"
ELSE
    PRINT #X, "/STATUS:ON"
END IF
IF KEYOFF = -1 THEN
    PRINT #X, "/KEYS:OFF"
ELSE
    PRINT #X, "/KEYS:ON"
END IF
IF CPUOFF = -1 THEN
    PRINT #X, "/CPU:OFF"
ELSE
    PRINT #X, "/CPU:ON"
END IF

IF CPUMON = 0 THEN
    PRINT #X, "/CPUMON:0"
ELSE
    IF CPUMON = 1 THEN
        PRINT #X, "/CPUMON:1"
    ELSE
        IF CPUMON = 2 THEN
            PRINT #X, "/CPUMON:2"
        END IF
    END IF
END IF

IF MONITOROFF = -1 THEN
    PRINT #X, "/MONITOR:OFF"
ELSE
    PRINT #X, "/MONITOR:ON"
END IF
PRINT #X, "PROMPT=" + Prompt$
FOR v = 1 TO 12
    IF RIGHT$(FunctionKeys(v), 1) = CHR$(13) THEN
        v$ = LEFT$(FunctionKeys(v), LEN(FunctionKeys(v)) - 1) + "+cr"
    ELSE
        v$ = FunctionKeys(v)
    END IF
    v$ = "KEY" + STR$(v) + "=" + v$
    PRINT #X, v$
NEXT
IF ProfilesActive THEN
    PRINT #X, "PROFILESACTIVE=-1"
ELSE
    PRINT #X, "PROFILESACTIVE=0"
END IF
PRINT #X, "SoundEffect=-1"
CLOSE #X
END SUB

'Common HRESULT Values
'Name            Description                    Value
'S_OK            Operation successful           0x00000000
'E_ABORT         Operation aborted              0x80004004
'E_ACCESSDENIED  General access denied error    0x80070005
'E_FAIL          Unspecified failure            0x80004005

' expiremental url grab..
SUB GrabURL
DIM ASCIIZ AS STRING * 260
DIM Attribute AS _UNSIGNED LONG

' URL to grab (page or a file)
URL$ = "http://www.filegate.net/oredson/emc3.zip"

' File to save URL as
Var$ = GetURLFilename$
URLfile$ = Var$
IF Debug THEN
    ' display files
    PRINT "URL: "; URL$
    PRINT "File: "; URLfile$
END IF
IF _FILEEXISTS(Var$) = 0 THEN
    ' Download it.  Returns 0 if started.
    a% = URLDownloadToFileA%(0, URL$ + CHR$(0), URLfile$ + CHR$(0), 0, 0)
    PRINT "Started. Return: "; HEX$(a%)
END IF

' hide file
Attribute = &H27
ASCIIZ = Var$ + CHR$(0)
x = SetFileAttributes&(ASCIIZ, Attribute)
IF Debug THEN
    CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt2$)
END IF
END SUB

FUNCTION GetConfigFilename$
Var$ = ENVIRON$("PROGRAMFILES")
IF LEN(Var$) = 0 THEN
    Var$ = "C:\Program Files"
END IF
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
IF RIGHT$(Var$, 1) <> "\" THEN
    Var$ = Var$ + "\"
END IF
Var$ = Var$ + "QB64Shell"
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
Var$ = Var$ + "\" + ConfigFilename$
GetConfigFilename$ = Var$
END FUNCTION

FUNCTION GetMenuConfigFilename$
Var$ = ENVIRON$("PROGRAMFILES")
IF LEN(Var$) = 0 THEN
    Var$ = "C:\Program Files"
END IF
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
IF RIGHT$(Var$, 1) <> "\" THEN
    Var$ = Var$ + "\"
END IF
Var$ = Var$ + "QB64Shell"
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
Var$ = Var$ + "\" + MenuConfigFilename$
GetMenuConfigFilename$ = Var$
END FUNCTION

FUNCTION GetProfileFilename$
Var$ = ENVIRON$("PROGRAMFILES")
IF LEN(Var$) = 0 THEN
    Var$ = "C:\Program Files"
END IF
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
IF RIGHT$(Var$, 1) <> "\" THEN
    Var$ = Var$ + "\"
END IF
Var$ = Var$ + "QB64Shell"
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
Var$ = Var$ + "\" + ProfileFilename$
GetProfileFilename$ = Var$
END FUNCTION

FUNCTION GetProfileBackupFilename$
Var$ = ENVIRON$("PROGRAMFILES")
IF LEN(Var$) = 0 THEN
    Var$ = "C:\Program Files"
END IF
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
IF RIGHT$(Var$, 1) <> "\" THEN
    Var$ = Var$ + "\"
END IF
Var$ = Var$ + "QB64Shell"
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
Var$ = Var$ + "\" + ProfileBackupFilename$
GetProfileBackupFilename$ = Var$
END FUNCTION

FUNCTION GetURLFilename$
Var$ = ENVIRON$("PROGRAMFILES")
IF LEN(Var$) = 0 THEN
    Var$ = "C:\Program Files"
END IF
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
IF RIGHT$(Var$, 1) <> "\" THEN
    Var$ = Var$ + "\"
END IF
Var$ = Var$ + "QB64Shell"
IF _DIREXISTS(Var$) = 0 THEN
    f$ = Var$ + CHR$(0)
    x = CreateDirectoryA(f$, 0)
END IF
Var$ = Var$ + "\emc3.zip"
GetURLFilename$ = Var$
END FUNCTION

SUB ReadKeys
FunctionKeys(1) = "EDIT"
FunctionKeys(2) = "PROMPT"
FunctionKeys(3) = "DIR"
FunctionKeys(4) = "TREEDIR"
FunctionKeys(5) = "COPY"
FunctionKeys(6) = "QUIT"
FunctionKeys(7) = "MEM"
FunctionKeys(8) = "CHDIR"
FunctionKeys(9) = "MKDIR"
FunctionKeys(10) = "RMDIR"
FunctionKeys(11) = "SYSTEM"
FunctionKeys(12) = "STATUS"
END SUB

SUB ClearKeys
FOR V = 1 TO 10
    KEY V, ""
NEXT
KEY 30, ""
KEY 31, ""
END SUB

SUB SetKeys
FOR V = 1 TO 10
    KEY V, FunctionKeys(V)
NEXT
KEY 30, FunctionKeys(11)
KEY 31, FunctionKeys(12)
END SUB

SUB DisplayToolBar
L = CSRLIN
P = POS(0)
IF Status = 0 THEN
    VIEW PRINT 1 TO 24
ELSE
    VIEW PRINT 1 TO 23
END IF
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
ELSE
    VIEW PRINT 2 TO 23
END IF
LOCATE L, P, 1
END SUB

SUB DisplayToolBar2
L = CSRLIN
P = POS(0)
LOCATE 1, 1, 1
IF CPUOFF = 0 THEN
    COLOR Green, 1
ELSE
    COLOR Red, 1
END IF
PRINT "CPU";
COLOR 7, 1
LOCATE L, P, 0
END SUB

SUB Toolbar
COLOR 14, 1
LOCATE 1, 1, 1
PRINT SPACE$(80);
LOCATE 1, 1, 1
IF CPUOFF = 0 THEN
    COLOR Green, 1
ELSE
    COLOR Red, 1
END IF
PRINT "CPU";
COLOR 14, 1
FOR Var = 1 TO 7
    COLOR White, 1
    CALL DisplayTab(Var)
NEXT
COLOR 15, 1
LOCATE 1, 74, 1
PRINT "About";
COLOR 7, 0
END SUB

SUB PreFetch (VarZ)
StdinLines = 0
StdinMax = 1
StdoutLines = 0
StdoutMax = 1
REDIM StdinArray(1) AS STRING
REDIM StdoutArray(1) AS STRING
IF Stdin THEN
    IF StdinHandle THEN
        CLOSE #StdinHandle
    END IF
    StdinHandle = FREEFILE
    OPEN StdinFile FOR BINARY AS #StdinHandle
    Count# = 0#
    IF LOF(StdinHandle) > 0 THEN
        DO
            IF EOF(StdinHandle) THEN
                EXIT DO
            END IF
            LINE INPUT #StdinHandle, LineX$
            Count# = Count# + LEN(LineX$) + 2#
            StdinLines = StdinLines + 1
            IF StdinLines > StdinMax THEN
                StdinMax = StdinMax + 1
                REDIM _PRESERVE StdinArray(StdinMax) AS STRING
            END IF
            StdinArray(StdinLines) = LineX$
        LOOP
        IF Count# < LOF(StdinHandle) THEN
            StdinLines = StdinLines + 1
            IF StdinLines > StdinMax THEN
                StdinMax = StdinMax + 1
                REDIM _PRESERVE StdinArray(StdinMax) AS STRING
            END IF
            StdinArray(StdinLines) = ""
        END IF
    END IF
END IF
END SUB

' returns:
'  Var =
'   0  -  no match
'  -1  -  error  (error number in VarZ)
'   1  -  match
'
' StdinFile$  -  standard input filename
'  Stdin%    -1 specified
' StdoutFile$ -  standard output filename
'  Stdout%    1 = output, 2 = append
' StderrFile$ -  standard error filename
'  Stderr%   -1 specified
SUB PreParse (Var$, Var, VarZ)
DIM hfind AS _OFFSET

Stdin% = 0
Stdout% = 0
Stderr% = 0
StdinFile$ = ""
StdoutFile$ = ""
StderrFile$ = ""

' reset return variables
Var = 0

' reset work variables
VarS$ = Var$
VarX$ = Var$

' remove special switchs
GOSUB ParseSymbols

' parse multi token error
IF INSTR(VarX$, ">>>>") THEN
    VarZ = 1
    GOTO ParseError
END IF

' parse sdterr
V = INSTR(VarX$, ">>>")
IF V THEN
    VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 3)
    Var = 1
END IF
V = INSTR(VarX$, ">>>")
IF V THEN
    VarZ = 2
    GOTO ParseError
END IF

' parse stdout append
V = INSTR(VarX$, ">>")
IF V THEN
    VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 2)
    Var = 1
    V = INSTR(VarX$, ">")
    IF V THEN
        VarZ = 3
        GOTO ParseError
    END IF
END IF

' parse stdout
V = INSTR(VarX$, ">")
IF V THEN
    VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 1)
    Var = 1
END IF
V = INSTR(VarX$, ">")
IF V THEN
    VarZ = 3
    GOTO ParseError
END IF

' parse stdin
V = INSTR(VarX$, "<")
IF V THEN
    VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 1)
    Var = 1
END IF
V = INSTR(VarX$, "<")
IF V THEN
    VarZ = 3
    GOTO ParseError
END IF
IF Var = 0 THEN
    EXIT SUB
END IF

' parse filenames
VarX$ = Var$
GOSUB ParseSymbols
L = LEN(VarX$)
DO
    IF MID$(VarX$, L, 1) = "<" THEN
        Stdin% = -1
        StdinFile$ = MID$(VarX$, L + 1)
        VarX$ = LEFT$(VarX$, L - 1)
        L = LEN(VarX$)
        GOTO NextChar
    END IF

    IF INSTR(VarX$, ">>>") THEN
        IF MID$(VarX$, L - 2, 3) = ">>>" THEN
            Stderr% = -1
            StderrFile$ = MID$(VarX$, L + 1)
            VarX$ = LEFT$(VarX$, L - 3)
            L = LEN(VarX$)
            GOTO NextChar
        END IF
    END IF

    IF INSTR(VarX$, ">>") THEN
        IF MID$(VarX$, L - 1, 2) = ">>" THEN
            Stdout% = 2
            StdoutFile$ = MID$(VarX$, L + 1)
            VarX$ = LEFT$(VarX$, L - 2)
            L = LEN(VarX$)
            GOTO NextChar
        END IF
    END IF

    IF INSTR(VarX$, ">") THEN
        IF MID$(VarX$, L, 1) = ">" THEN
            Stdout% = 1
            StdoutFile$ = MID$(VarX$, L + 1)
            VarX$ = LEFT$(VarX$, L - 1)
            L = LEN(VarX$)
            GOTO NextChar
        END IF
    END IF
    L = L - 1
    NextChar:
    IF L <= 0 THEN
        EXIT DO
    END IF
LOOP

' parse filenames
IF LEN(StdinFile$) THEN
    StdinFile$ = RTRIM$(LTRIM$(StdinFile$))
    IF LEFT$(StdinFile$, 1) = CHR$(34) AND RIGHT$(StdinFile$, 1) = CHR$(34) THEN
        StdinFile$ = LEFT$(StdinFile$, LEN(StdinFile$) - 1)
        StdinFile$ = MID$(StdinFile$, 2)
        IF LEN(StdinFile$) = 0 THEN
            VarZ = 4
            GOTO ParseError
        END IF
    ELSE
        IF INSTR(StdinFile$, " ") THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

IF LEN(StdoutFile$) THEN
    StdoutFile$ = RTRIM$(LTRIM$(StdoutFile$))
    IF LEFT$(StdoutFile$, 1) = CHR$(34) AND RIGHT$(StdoutFile$, 1) = CHR$(34) THEN
        StdoutFile$ = LEFT$(StdoutFile$, LEN(StdoutFile$) - 1)
        StdoutFile$ = MID$(StdoutFile$, 2)
        IF LEN(StdoutFile$) = 0 THEN
            VarZ = 4
            GOTO ParseError
        END IF
    ELSE
        IF INSTR(StdoutFile$, " ") THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

IF LEN(StderrFile$) THEN
    StderrFile$ = RTRIM$(LTRIM$(StderrFile$))
    IF LEFT$(StderrFile$, 1) = CHR$(34) AND RIGHT$(StderrFile$, 1) = CHR$(34) THEN
        StderrFile$ = LEFT$(StderrFile$, LEN(StderrFile$) - 1)
        StderrFile$ = MID$(StderrFile$, 2)
        IF LEN(StderrFile$) = 0 THEN
            VarZ = 4
            GOTO ParseError
        END IF
    ELSE
        IF INSTR(StderrFile$, " ") THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

IF LEN(StdinFile$) THEN
    VarQ$ = StdinFile$
    FOR L = LEN(VarQ$) TO 1 STEP -1
        IF MID$(VarQ$, L, 1) = "\" THEN
            VarQ$ = MID$(VarQ$, L + 1)
            EXIT FOR
        END IF
    NEXT
    IF MID$(VarQ$, 2, 1) = ":" THEN
        VarQ$ = MID$(VarQ$, 3)
    END IF
    IF LEN(VarQ$) = 0 THEN
        VarZ = 4
        GOTO ParseError
    END IF
    IF LCASE$(VarQ$) = "nul:" THEN
        Eat$ = ""
    ELSE
        IF ValidFileChar(VarQ$) THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

IF LEN(StdoutFile$) THEN
    VarQ$ = StdoutFile$
    FOR L = LEN(VarQ$) TO 1 STEP -1
        IF MID$(VarQ$, L, 1) = "\" THEN
            VarQ$ = MID$(VarQ$, L + 1)
            EXIT FOR
        END IF
    NEXT
    IF MID$(VarQ$, 2, 1) = ":" THEN
        VarQ$ = MID$(VarQ$, 3)
    END IF
    IF LEN(VarQ$) = 0 THEN
        VarZ = 4
        GOTO ParseError
    END IF
    IF LCASE$(VarQ$) = "nul:" OR LCASE$(VarQ$) = "scrn:" THEN
        Eat$ = ""
    ELSE
        IF ValidFileChar(VarQ$) THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

IF LEN(StderrFile$) THEN
    VarQ$ = StderrFile$
    FOR L = LEN(VarQ$) TO 1 STEP -1
        IF MID$(VarQ$, L, 1) = "\" THEN
            VarQ$ = MID$(VarQ$, L + 1)
            EXIT FOR
        END IF
    NEXT
    IF MID$(VarQ$, 2, 1) = ":" THEN
        VarQ$ = MID$(VarQ$, 3)
    END IF
    IF LEN(VarQ$) = 0 THEN
        VarZ = 4
        GOTO ParseError
    END IF
    IF LCASE$(VarQ$) = "nul:" OR LCASE$(VarQ$) = "scrn:" THEN
        Eat$ = ""
    ELSE
        IF ValidFileChar(VarQ$) THEN
            VarZ = 4
            GOTO ParseError
        END IF
    END IF
END IF

' search for special switches
Var$ = VarS$
V1 = 0
V2 = 0
IF LEN(Var$) THEN
    DO
        V1 = INSTR(V1 + 1, Var$, "/[")
        IF V1 = 0 THEN
            EXIT DO
        END IF
        V2 = INSTR(V1 + 1, Var$, "]")
        IF V2 THEN
            V1 = V2
        ELSE
            V1 = V1 + 1
        END IF
    LOOP
END IF

' trim command line
V = INSTR(V2 + 1, Var$, ">")
IF V THEN
    Var$ = LEFT$(Var$, V - 1)
END IF
V = INSTR(V2 + 1, Var$, "<")
IF V THEN
    Var$ = LEFT$(Var$, V - 1)
END IF

' verify files can be opened
IF Stdin% = -1 THEN
    IF RTRIM$(StdinFile$) = "" THEN
        VarZ = 5
        GOTO ParseError
    END IF
    f$ = StdinFile$ + CHR$(0)
    hfind = CreateFile(f$, &H180, 0, 0, 3, 0, 0) ' open only if exist
    IF hfind = 0 THEN
        VarZ = 5
        GOTO ParseError
    END IF
    x = CloseHandle(hfind)
    IF LCASE$(StdinFile$) = "nul:" THEN
        Eat$ = ""
    ELSE
        IF TestFile2(StdinFile$) = 0 THEN
            VarZ = 5
            GOTO ParseError
        END IF
    END IF
END IF

IF Stdout% > 0 THEN
    IF RTRIM$(StdoutFile$) = "" THEN
        VarZ = 6
        GOTO ParseError
    END IF
    f$ = StdoutFile$ + CHR$(0)
    hfind = CreateFile(f$, &H180, 0, 0, 4, 0, 0) ' open always
    IF hfind = 0 THEN
        VarZ = 6
        GOTO ParseError
    END IF
    x = CloseHandle(hfind)
    IF LCASE$(StdoutFile$) = "nul:" OR LCASE$(StdoutFile$) = "scrn:" THEN
        Eat$ = ""
    ELSE
        IF TestFile(StdoutFile$) = 0 THEN
            VarZ = 6
            GOTO ParseError
        END IF
    END IF
END IF

IF Stderr% = -1 THEN
    IF RTRIM$(StderrFile$) = "" THEN
        VarZ = 7
        GOTO ParseError
    END IF
    f$ = StderrFile$ + CHR$(0)
    hfind = CreateFile(f$, &H180, 0, 0, 4, 0, 0) ' open always
    IF hfind = 0 THEN
        VarZ = 7
        GOTO ParseError
    END IF
    x = CloseHandle(hfind)
    IF LCASE$(StderrFile$) = "nul:" OR LCASE$(StderrFile$) = "scrn:" THEN
        Eat$ = ""
    ELSE
        IF TestFile(StderrFile$) = 0 THEN
            VarZ = 7
            GOTO ParseError
        END IF
    END IF
END IF
EXIT SUB

' remove special switches
ParseSymbols:
IF LEN(VarX$) THEN
    DO
        V1 = INSTR(VarX$, "/[")
        IF V1 = 0 THEN
            EXIT DO
        END IF
        V2 = INSTR(V1 + 1, VarX$, "]")
        IF V2 THEN
            VarX$ = LEFT$(VarX$, V1 - 1) + MID$(VarX$, V2 + 1)
        ELSE
            VarX$ = LEFT$(VarX$, V1 - 1) + MID$(VarX$, V1 + 2)
        END IF
    LOOP
END IF
RETURN

ParseError:
Var = -1
Stdin% = 0
Stdout% = 0
Stderr% = 0
StdinFile$ = ""
StdoutFile$ = ""
StderrFile$ = ""
END SUB

SUB ComplexInit (x$, x, z)
CommandCount = 0
z$ = x$
x = -1
z = 0
' parse /[ and | symbols.
V = 0
DO
    IF V > LEN(z$) THEN
        c$ = RTRIM$(LTRIM$(z$))
        IF LEN(c$) THEN
            CommandCount = CommandCount + 1
            REDIM _PRESERVE CommandList(CommandCount) AS STRING
            CommandList(CommandCount) = c$
        ELSE
            x = 0
            z = 1
        END IF
        EXIT DO
    END IF
    V = V + 1
    IF MID$(z$, V, 2) = "/[" THEN
        V = INSTR(V + 2, z$, "]")
        IF V = 0 THEN
            x = 0
            z = 2
            EXIT DO
        END IF
    END IF
    IF MID$(z$, V, 1) = "|" THEN
        c$ = LEFT$(z$, V - 1)
        c$ = RTRIM$(LTRIM$(c$))
        IF LEN(c$) THEN
            CommandCount = CommandCount + 1
            REDIM _PRESERVE CommandList(CommandCount) AS STRING
            CommandList(CommandCount) = c$
            z$ = MID$(z$, V + 1)
            V = 0
        ELSE
            x = 0
            z = 1
            EXIT DO
        END IF
    END IF
LOOP
IF x = 0 THEN
    EXIT SUB
END IF

' verify all commands in list
FOR V = 1 TO CommandCount
    x1$ = LCASE$(CommandList(V))
    q = 0
    FOR V2 = 1 TO MaxComplex
        V$ = ComplexCommands(V2)
        IF LEFT$(x1$, LEN(V$)) = V$ THEN
            z$ = MID$(x1$, LEN(V$) + 1, 1)
            IF LEN(z$) THEN
                IF IsAlphaNum(z$) THEN
                    x = 0
                    z = 3
                    EXIT SUB
                END IF
            END IF
            q = -1
            EXIT FOR
        END IF
    NEXT
    IF q = 0 THEN
        x = 0
        z = 3
        EXIT FOR
    END IF
NEXT
END SUB

SUB ComplexParse (x$, x)
x = 0
x1$ = LCASE$(x$)
FOR V = 1 TO MaxComplex
    V$ = ComplexCommands(V)
    IF LEFT$(x1$, LEN(V$)) = V$ THEN
        x = -1

        ' display current command line title
        _TITLE Program + " - " + UCASE$(V$)

        ' get next command in list
        x1$ = RTRIM$(x$)
        COMMAND2$ = MID$(x1$, LEN(V$) + 1)

        ' force stdin for all streamed commands
        Stdin = -1

        ' call the actual command in stream
        Var = 0
        SELECT CASE V
            CASE 1
                CALL Xcount(Var)
            CASE 2
                CALL TouchDir(Var)
            CASE 3
                CALL Touch(Var)
            CASE 4
                CALL Dirattr(Var)
            CASE 5
                CALL Setattr(Var)
            CASE 6
                Encrypt.Call = 0
                CALL Encrypt(Var)
            CASE 7
                Encrypt.Call = -1
                CALL Encrypt(Var)
            CASE 8
                CALL Treedel(Var)
            CASE 9
                CALL Stree(Var)
            CASE 10
                CALL Delete(Var)
            CASE 11
                CALL Nameit(Var)
            CASE 12
                CALL Drives(Var)
            CASE 13
                CALL Label(Var)
            CASE 14
                CALL ListVols(Var)
            CASE 15
                CALL Whereis(Var)
            CASE 16
                CALL Copyit(Var)
            CASE 17
                CALL ParseTest
            CASE 18
                CALL Zsort(Var)
            CASE 19
                CALL TypeY(Var)
            CASE 20
                CALL MoreX(Var)
            CASE 21
                CALL FindY(Var)
            CASE 22, 23
                CALL NewDir(Var)
            CASE 24, 25
                CALL MakDir(Var)
            CASE 26, 27
                CALL DelDir(Var)
            CASE 28
                CALL MakeFile(Var)
        END SELECT
        ' command switch error
        IF Var THEN
            x = -2
            EXIT FOR
        END IF
    END IF
NEXT
END SUB

SUB SimpleParse (x$, x)
x = 0
x1$ = LCASE$(x$)
FOR V = 1 TO MaxSimple
    V$ = SimpleCommands(V)
    IF LEFT$(x1$, LEN(V$)) = V$ THEN
        x = -1
        z$ = MID$(x1$, LEN(V$) + 1, 1)
        IF LEN(z$) THEN
            IF IsAlphaNum(z$) THEN
                PRINT "Type HELP for help."
                EXIT SUB
            END IF
        END IF
        IF Stdin THEN
            PRINT "Error: Command does not support redirected input."
            EXIT SUB
        END IF
        IF LTRIM$(RTRIM$(MID$(x1$, LEN(V$) + 1))) = "/?" THEN
            CALL DisplayHelp(V$)
            EXIT SUB
        END IF
        _TITLE Program + " - " + UCASE$(V$)
        IF V = 1 THEN GOTO PromptFunc
        x1$ = RTRIM$(x1$)
        IF SimpleCommandsN(V) = 0 THEN
            IF x1$ <> V$ THEN
                PRINT "Type " + UCASE$(V$) + " /? for help."
                EXIT SUB
            END IF
        END IF
        IF V = 2 THEN GOTO RedrawFunc
        IF V = 3 THEN GOTO CpuFunc
        IF V = 4 THEN GOTO MemFunc
        IF V = 5 THEN GOTO ClsFunc
        IF V = 6 THEN GOTO EditFunc
        IF V = 7 THEN GOTO DebugFunc
        IF V = 8 THEN GOTO VerFunc
        IF V = 9 THEN GOTO HelpFunc
        IF V = 10 THEN GOTO ColorsFunc
        IF V = 11 THEN GOTO QuitProgram
        IF V = 12 THEN GOTO AbortProgram
        IF V = 13 THEN GOTO KeyFunc
        IF V = 14 THEN GOTO StatusFunc
        IF V = 15 THEN GOTO MenuFunc
        IF V = 16 THEN GOTO DateFunc
        IF V = 17 THEN GOTO TimeFunc
        IF V = 18 THEN GOTO ClockFunc
        IF V = 19 THEN GOTO HexCalcFunc
        IF V = 20 THEN GOTO AsciiChartFunc
        IF V = 21 THEN GOTO HexChartFunc
        IF V = 22 THEN GOTO WhatisFunc
        IF V = 23 THEN GOTO ToggleFunc
        IF V = 24 THEN GOTO AboutFunc
        IF V = 25 THEN GOTO ToggleFunc2
        IF V = 26 THEN GOTO ToggleFunc3
        IF V = 27 THEN GOTO BlankFunc
        IF V = 28 THEN GOTO SaverFunc
        IF V = 29 THEN GOTO ProfilesFunc
        IF V = 30 THEN GOTO MonitorFunc
        IF V = 31 THEN GOTO PushdFunc
        IF V = 32 THEN GOTO PopdFunc
        IF V = 33 THEN GOTO CompareFunc
        IF V = 34 THEN GOTO CompareFunc2
        IF V = 35 THEN GOTO FileListFunc
        IF V = 36 THEN GOTO DirListFunc
        IF V = 37 THEN GOTO DriveListFunc
    END IF
NEXT
EXIT SUB

DateFunc:
PRINT "Current date: "; LEFT$(clock$, 15)
PRINT "Enter date(mm-dd-yyyy)";
INPUT z$
IF LEN(z$) THEN
    z = ValidDate(z$)
    IF z THEN
        DATE$ = z$
    ELSE
        PRINT #StdoutHandle, "Invalid date."
    END IF
END IF
EXIT SUB

TimeFunc:
PRINT "Current time: "; TIME$
PRINT "Enter time(hh:mm:ss)";
INPUT z$
IF LEN(z$) THEN
    z = ValidTime(z$)
    IF z THEN
        TIME$ = z$
    ELSE
        PRINT #StdoutHandle, "Invalid time."
    END IF
END IF
EXIT SUB

DebugFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "debug" THEN
    IF Debug THEN
        PRINT #StdoutHandle, "DEBUG is on."
    ELSE
        PRINT #StdoutHandle, "DEBUG is off."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 6)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "off" THEN
        Debug = 0
        PRINT "Debug off."
    ELSE
        IF x1$ = "on" THEN
            Debug = -1
            PRINT "Debug on."
        ELSE
            IF x1$ = "test" THEN
                r = SQR(-1)
            ELSE
                IF x1$ = "error" THEN
                    ' display specific error
                    PRINT "Test value? ";
                    Var$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                    x& = INT(VAL(Var$))
                    y& = FormatMessageA&(&H1200, "", x&, 0, ErrorBuffer$, 260, "")
                    IF y& = 0 THEN
                        ' message does not exist in message table
                        y& = FormatMessageA&(&H1200, "", GetLastError, 0, ErrorBuffer$, 260, "")
                        PRINT "Error 0x"; HEX$(GetLastError); " ("; GetLastError; ")"
                        IF y& > 2 THEN
                            PRINT LEFT$(ErrorBuffer$, y& - 2)
                        END IF
                    ELSE
                        PRINT "Error 0x"; HEX$(x&); " ("; x&; ")"
                        PRINT LEFT$(ErrorBuffer$, y& - 2)
                    END IF
                ELSE
                    IF x1$ = "errorbox" THEN
                        IF ErrorBox(" Debug ", "Sample error test.") THEN
                            PRINT "Entered OK"
                        ELSE
                            PRINT "Entered Cancel"
                        END IF
                    ELSE
                        IF x1$ = "messagebox" THEN
                            r = MessageBox(" Debug ", "Sample message test.")
                        ELSE
                            IF x1$ = "monitorbox" THEN
                                m = DisplaySelect ' store monitor state
                                s1$ = LTRIM$(STR$(MonitorTimer))
                                s2$ = LTRIM$(STR$(MonitorLimit))
                                r = MonitorBox(s1$, s2$) ' 0=exit
                                DisplaySelect = m ' restore monitor state
                                IF r = 2 THEN ' system
                                    PRINT "System"
                                END IF
                                IF r = 1 THEN ' continue
                                    PRINT "Timer: "; s1$
                                    PRINT "Limit: "; s2$
                                END IF
                            ELSE
                                IF x1$ = "timetype" THEN
                                    DateTimeType = DateTimeType + 1
                                    IF DateTimeType = 3 THEN
                                        DateTimeType = 0
                                    END IF
                                    SELECT CASE DateTimeType
                                        CASE 0
                                            PRINT "Create"
                                        CASE 1
                                            PRINT "Access"
                                        CASE 2
                                            PRINT "Modify"
                                    END SELECT
                                ELSE
                                    IF x1$ = "graburl" THEN
                                        CALL GrabURL
                                    ELSE
                                        PRINT #StdoutHandle, "Type DEBUG /? for help."
                                    END IF
                                END IF
                            END IF
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

ToggleFunc:
Fullscreen = NOT Fullscreen
IF Fullscreen THEN
    _FULLSCREEN
    IF _FULLSCREEN = 0 THEN
        _FULLSCREEN _OFF
        Fullscreen = 0
    END IF
ELSE
    _FULLSCREEN _OFF
END IF
VIEW PRINT 1 TO 25
KEY OFF
CLS
COLOR 15, 0
LOCATE 25, 1
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
    COLOR 7
    LOCATE 24, 1, 1
ELSE
    LOCATE 24, 1, 1
    COLOR 14
    CALL DisplayVersion
    VIEW PRINT 2 TO 23
    COLOR 7
    LOCATE 23, 1, 1
END IF
IF KEYOFF = 0 THEN
    KEY ON
END IF
EXIT SUB

' screen
ToggleFunc2:
Fullscreen = -1
_FULLSCREEN
IF _FULLSCREEN = 0 THEN
    _FULLSCREEN _OFF
    Fullscreen = 0
END IF
VIEW PRINT 1 TO 25
KEY OFF
CLS
COLOR 15, 0
LOCATE 25, 1
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
    COLOR 7
    LOCATE 24, 1, 1
ELSE
    LOCATE 24, 1, 1
    COLOR 14
    CALL DisplayVersion
    VIEW PRINT 2 TO 23
    COLOR 7
    LOCATE 23, 1, 1
END IF
IF KEYOFF = 0 THEN
    KEY ON
END IF
EXIT SUB

' window
ToggleFunc3:
Fullscreen = 0
_FULLSCREEN _OFF
VIEW PRINT 1 TO 25
KEY OFF
CLS
COLOR 15, 0
LOCATE 25, 1
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
    COLOR 7
    LOCATE 24, 1, 1
ELSE
    LOCATE 24, 1, 1
    COLOR 14
    CALL DisplayVersion
    VIEW PRINT 2 TO 23
    COLOR 7
    LOCATE 23, 1, 1
END IF
IF KEYOFF = 0 THEN
    KEY ON
END IF
EXIT SUB

AboutFunc:
V$ = Program + " " + FullVersion$ + " OS " + _OS$
PRINT #StdoutHandle, V$
PRINT #StdoutHandle, "Author: " + Author
PRINT #StdoutHandle, "Email: " + Email
EXIT SUB

SaverFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "screensaver" THEN
    IF ScreenSaverActive THEN
        PRINT #StdoutHandle, "Screensaver active."
        IF ScreenSaverResolution = 0! THEN
            PRINT #StdoutHandle, "Screensaver timer not set."
        ELSE
            V! = ScreenSaverResolution
            Var$ = "Screensaver timer:"
            Var$ = Var$ + STR$(V!) + " seconds"
            IF V! > 60! THEN
                V! = V! / 60!
                Var2$ = STR$(V!)
                IF INSTR(Var2$, ".") = 0 THEN
                    Var2$ = Var2$ + ".0"
                ELSE
                    Var2$ = LEFT$(Var2$, INSTR(Var2$, ".") + 1)
                END IF
                Var$ = Var$ + " (" + LTRIM$(Var2$) + " minutes)"
            END IF
            PRINT #StdoutHandle, Var$
        END IF
    ELSE
        PRINT #StdoutHandle, "Screensaver inactive."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 12)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "off" THEN
        ScreenSaverActive = 0
        PRINT #StdoutHandle, "Screensaver off."
        EXIT SUB
    END IF
    IF LEFT$(x1$, 2) = "on" THEN
        IF x1$ = "on" THEN
            ScreenSaverActive = -1
            PRINT #StdoutHandle, "Screensaver on."
            EXIT SUB
        END IF
        x1$ = MID$(x1$, 3)
        x1$ = RTRIM$(LTRIM$(x1$))
        z$ = LEFT$(x1$, 1)
        IF z$ = "," OR z$ = "=" THEN
            x1$ = MID$(x1$, 2)
        ELSE
            IF z$ >= "0" AND z$ <= "9" THEN
                Eat$ = Nul
            ELSE
                PRINT #StdoutHandle, "Type HELP for help."
                EXIT SUB
            END IF
        END IF
    END IF
    x1$ = RTRIM$(LTRIM$(x1$))
    z = ValidTime(x1$)
    IF z THEN ' HH:MM:SS
        V! = INT(VAL(RIGHT$(x1$, 2)))
        V! = V! + INT(VAL(MID$(x1$, 4, 2))) * 60!
        V! = V! + INT(VAL(LEFT$(x1$, 2))) * 3600!
        x1$ = LTRIM$(STR$(V!))
    END IF
    m = 0
    IF RIGHT$(x1$, 1) = "s" THEN
        x1$ = LEFT$(x1$, LEN(x1$) - 1)
    ELSE
        IF RIGHT$(x1$, 1) = "m" THEN
            m = -1
            x1$ = LEFT$(x1$, LEN(x1$) - 1)
        ELSE
            IF RIGHT$(x1$, 1) = "h" THEN
                m = -2
                x1$ = LEFT$(x1$, LEN(x1$) - 1)
            END IF
        END IF
    END IF
    x1$ = RTRIM$(LTRIM$(x1$))
    IF IsNumeric(x1$) THEN
        V! = INT(VAL(x1$))
        IF m = -1 THEN V! = V! * 60! ' minutes
        IF m = -2 THEN V! = V! * 3600! ' hours
        IF V! >= 1! AND V! <= 86400! THEN
            ScreenSaverActive = -1
            ScreenSaverResolution = V!
            PRINT #StdoutHandle, "Screensaver timer set."
        ELSE
            PRINT #StdoutHandle, "Invalid screensaver timer value."
        END IF
    ELSE
        PRINT #StdoutHandle, "Invalid screensaver timer value."
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

ProfilesFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "profiles" THEN
    IF ProfilesActive THEN
        PRINT #StdoutHandle, "User profiles active."
        CALL CountProfiles(V!)
        PRINT #StdoutHandle, "Total of" + STR$(V!) + " users."
        CALL CountDeletedProfiles(V!)
        PRINT #StdoutHandle, "Total of" + STR$(V!) + " deleted users."
        IF Username > "" THEN
            PRINT #StdoutHandle, "You are currently logged on as: " + Username
            CALL GetUserStats(V#)
            PRINT #StdoutHandle, "You have logged on: " + FormatString$(V#) + " times."
        END IF
    ELSE
        PRINT #StdoutHandle, "User profiles not active."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 9)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "on" THEN
        IF ProfilesActive = 0 THEN
            CALL InitProfiles
        ELSE
            CALL WriteProfile
        END IF
        CALL GetUserProfile(V, "")
        IF V THEN
            ProfilesActive = -1
            PRINT #StdoutHandle, "User Profiles on."
        ELSE
            ProfilesActive = 0
            PRINT #StdoutHandle, "User Profiles logon fail."
        END IF
        EXIT SUB
    END IF
    IF x1$ = "off" THEN
        IF ProfilesActive THEN
            CALL WriteProfile
        END IF
        ProfilesActive = 0
        PRINT #StdoutHandle, "User Profiles off."
        EXIT SUB
    END IF
    IF x1$ = "list" THEN
        CALL ListProfiles
        EXIT SUB
    END IF
    IF x1$ = "edit" OR x1$ = "editor" THEN
        CALL EditProfiles
        EXIT SUB
    END IF
    IF x1$ = "logon" OR x1$ = "relogon" THEN
        IF ProfilesActive = 0 THEN
            PRINT #StdoutHandle, "User profiles not active."
            EXIT SUB
        END IF
        CALL WriteProfile
        Username = ""
        Password = ""
        CALL GetUserProfile(V, "")
        EXIT SUB
    END IF
    PRINT #StdoutHandle, "Type PROFILES /? for help."
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

CpuFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "cpu" THEN
    IF CPUOFF THEN
        PRINT #StdoutHandle, "CPU monitor is off."
    ELSE
        PRINT #StdoutHandle, "CPU monitor is on."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 4)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "off" THEN
        TIMER OFF
        CPUOFF = -1
        _TITLE Program
        CALL DisplayToolBar
    ELSE
        IF x1$ = "on" THEN
            CPUOFF = 0
            TIMER ON
            CALL DisplayToolBar
        ELSE
            IF x1$ = "memory" THEN
                SELECT CASE CPUMON
                    CASE 0
                        CPUMON = 1 ' MB
                        PRINT #StdoutHandle, "CPU MB memory monitor on."
                    CASE 1
                        CPUMON = 2 ' GB
                        PRINT #StdoutHandle, "CPU GB memory monitor on."
                    CASE 2
                        CPUMON = 0
                        PRINT #StdoutHandle, "CPU memory monitor off."
                END SELECT
            ELSE
                PRINT #StdoutHandle, "Type CPU /? for help."
            END IF
        END IF
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

MonitorFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "monitor" THEN
    IF MONITOROFF THEN
        PRINT #StdoutHandle, "Memory monitor not active."
    ELSE
        PRINT #StdoutHandle, "Memory monitor active."
        PRINT #StdoutHandle, "Memory counter set to" + STR$(MonitorTimer) + " seconds."
        PRINT #StdoutHandle, "Memory monitor set to" + STR$(MonitorLimit) + " MB."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 8) ' monitor
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "off" THEN
        MONITOROFF = True
        PRINT #StdoutHandle, "Memory monitor off."
    ELSE
        IF x1$ = "on" THEN
            MONITOROFF = False
            PRINT #StdoutHandle, "Memory monitor on."
            MonitorCount = 0
        ELSE
            IF LEFT$(x1$, 5) = "limit" THEN
                x1$ = LTRIM$(MID$(x1$, 6))
                IF LEFT$(x1$, 1) = "=" THEN
                    x1$ = MID$(x1$, 2)
                END IF
                x1$ = LTRIM$(RTRIM$(x1$))
                IF IsSingle(x1$) THEN
                    X# = VAL(x1$)
                    IF X# = 0# THEN
                        MONITOROFF = True
                        PRINT #StdoutHandle, "Memory monitor off."
                    ELSE
                        IF X# > 0# THEN
                            MONITOROFF = False
                            MonitorCount = 0
                            MonitorLimit = X#
                            PRINT #StdoutHandle, "Memory monitor set to" + STR$(MonitorLimit) + " MB."
                        ELSE
                            PRINT #StdoutHandle, "Type MONITOR /? for help."
                        END IF
                    END IF
                ELSE
                    PRINT #StdoutHandle, "Type MONITOR /? for help."
                END IF
            ELSE
                IF LEFT$(x1$, 5) = "timer" THEN
                    x1$ = LTRIM$(MID$(x1$, 6))
                    IF LEFT$(x1$, 1) = "=" THEN
                        x1$ = MID$(x1$, 2)
                    END IF
                    x1$ = LTRIM$(RTRIM$(x1$))
                    IF IsNumeric(x1$) THEN
                        X# = VAL(x1$)
                        IF X# = 0# THEN
                            MONITOROFF = True
                            PRINT #StdoutHandle, "Memory monitor off."
                        ELSE
                            IF X# > 0# THEN
                                IF X# <= 86400# THEN
                                    X! = CSNG(X#)
                                    MONITOROFF = False
                                    MonitorCount = 0
                                    MonitorTimer = X!
                                    PRINT #StdoutHandle, "Memory counter set to" + STR$(MonitorTimer) + " seconds."
                                END IF
                            ELSE
                                PRINT #StdoutHandle, "Type MONITOR /? for help."
                            END IF
                        END IF
                    ELSE
                        PRINT #StdoutHandle, "Type MONITOR /? for help."
                    END IF
                ELSE
                    PRINT #StdoutHandle, "Type MONITOR /? for help."
                END IF
            END IF
        END IF
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

PushdFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "pushd" THEN
    IF PathCount = 1 THEN
        PRINT #StdoutHandle, "No paths in table."
    ELSE
        PRINT #StdoutHandle, STR$(PathCount - 1) + " paths in table."
    END IF
    EXIT SUB
END IF
x1$ = MID$(x1$, 6) ' pushd
x1$ = RTRIM$(LTRIM$(x1$))
IF LEFT$(x1$, 1) = CHR$(34) THEN
    x1$ = MID$(x1$, 2)
END IF
IF RIGHT$(x1$, 1) = CHR$(34) THEN
    x1$ = LEFT$(x1$, LEN(x1$) - 1)
END IF
f$ = x1$ + CHR$(0)
r = SetCurrentDirectoryA(f$)
IF r = 0 THEN
    CALL DisplayWinError(r)
    EXIT SUB
END IF
PathCount = PathCount + 1
IF PathCount > PathMax THEN
    PathMax = PathMax + 1
    REDIM _PRESERVE PathTable(PathMax) AS STRING
END IF
PathTable(PathCount) = _CWD$
EXIT SUB

PopdFunc:
IF PathCount = 1 THEN
    PRINT #StdoutHandle, "No paths in table."
    EXIT SUB
END IF
PathCount = PathCount - 1
f$ = PathTable(PathCount) + CHR$(0)
r = SetCurrentDirectoryA(f$)
IF r = 0 THEN
    CALL DisplayWinError(r)
END IF
EXIT SUB

CompareFunc:
COLOR 15
PRINT "Enter first path? ";
f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
IF f$ <> "" THEN
    COLOR 15
    PRINT "Enter second path? ";
    g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
    IF g$ <> "" THEN
        IF _DIREXISTS(f$) AND _DIREXISTS(g$) THEN
            CALL VerifyFiles(f$, g$)
        ELSE
            COLOR 15
            PRINT #StdoutHandle, "Path not found."
            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
        END IF
    END IF
END IF
EXIT SUB

CompareFunc2:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "compfile" THEN
    ByteList = 0
ELSE
    x1$ = MID$(x1$, 9) ' compfile
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "list" THEN
        ByteList = -1
    ELSE
        PRINT #StdoutHandle, "Type COMPFILE /? for help."
        EXIT SUB
    END IF
END IF
COLOR 15
PRINT "Enter first file? ";
f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
IF f$ <> "" THEN
    COLOR 15
    PRINT "Enter second file? ";
    g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
    IF g$ <> "" THEN
        IF _FILEEXISTS(f$) AND _FILEEXISTS(g$) THEN
            CALL VerifyFiles2(f$, g$, ByteList)
        ELSE
            COLOR 15
            PRINT #StdoutHandle, "File not found."
            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
        END IF
    END IF
END IF
EXIT SUB

FileListFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "listfile" THEN
    Var1$ = "*.*"
ELSE
    x1$ = MID$(x1$, 9) ' filelist
    Var1$ = RTRIM$(LTRIM$(x1$))
END IF
CALL ListFiles(Var1$, -1)
EXIT SUB

DirListFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "listdir" THEN
    Var1$ = "*.*"
ELSE
    x1$ = MID$(x1$, 8) ' listdir
    Var1$ = RTRIM$(LTRIM$(x1$))
END IF
CALL ListFiles(Var1$, 0)
EXIT SUB

DriveListFunc:
CALL ListDrives
EXIT SUB

StatusFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "status" THEN
    PRINT #StdoutHandle, "Type STATUS /? for help."
    EXIT SUB
END IF
x1$ = MID$(x1$, 7)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "off" THEN
        IF Status THEN
            CALL SaveScreen
            SOFF = -1
            Status = 0
            VIEW PRINT 2 TO 24
            COLOR 7
            CLS
            CALL RestoreScreen
            LOCATE 23, 1, 1
            COLOR 15
            CALL DisplayPrompt
            PRINT
            LOCATE 24, 1, 1
        END IF
    ELSE
        IF x1$ = "on" THEN
            IF Status = 0 THEN
                CALL SaveScreen
                SOFF = 0
                Status = -1
                LOCATE 24, 1, 1
                COLOR 14
                CALL DisplayVersion
                VIEW PRINT 2 TO 23
                COLOR 7
                CLS
                CALL RestoreScreen
                PRINT
                LOCATE 23, 1, 1
            END IF
        ELSE
            PRINT #StdoutHandle, "Type STATUS /? for help."
        END IF
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

KeyFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "key" THEN
    PRINT #StdoutHandle, "Type KEY /? for help."
    EXIT SUB
END IF
x1$ = MID$(x1$, 4)
IF LEFT$(x1$, 1) = " " THEN
    x1$ = MID$(x1$, 2)
    x1$ = RTRIM$(LTRIM$(x1$))
    IF x1$ = "on" THEN
        KEYOFF = 0
        KEY ON
    ELSE
        IF x1$ = "off" THEN
            KEYOFF = -1
            KEY OFF
        ELSE
            IF x1$ = "list" THEN
                KEY LIST
            ELSE
                IF x1$ = "reset" THEN
                    CALL ReadKeys
                    CALL SetKeys
                    PRINT #StdoutHandle, "Keys reset."
                ELSE
                    y$ = MID$(x$, 4)
                    y = INSTR(y$, ",")
                    IF y THEN
                        z = VAL(LEFT$(y$, y - 1))
                        z$ = MID$(y$, y + 1)
                        IF z >= 1 AND z <= 12 THEN
                            q = z
                            IF z = 11 THEN z = 30
                            IF z = 12 THEN z = 31
                            IF LCASE$(RIGHT$(z$, 3)) = "+cr" THEN
                                z$ = LEFT$(z$, LEN(z$) - 3) + CHR$(13)
                            END IF
                            KEY z, z$
                            FunctionKeys(q) = z$
                        ELSE
                            PRINT #StdoutHandle, "Type KEY /? for help."
                        END IF
                    ELSE
                        PRINT #StdoutHandle, "Type KEY /? for help."
                    END IF
                END IF
            END IF
        END IF
    END IF
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
EXIT SUB

AbortProgram:
CALL WriteConfig
CALL WriteProfile
COLOR 7, 0
SYSTEM
EXIT SUB

QuitProgram:
IF QuitBox THEN
    CALL WriteConfig
    CALL WriteProfile
    COLOR 7, 0
    SYSTEM
END IF
EXIT SUB

MenuFunc:
CALL Menu
EXIT SUB

MemFunc:
IF x1$ = "mem" THEN
    PRINT #StdoutHandle, "Total Physical Memory: "; FormatString$(CDBL(TotalPhysicalMem))
    PRINT #StdoutHandle, "Free Physical Memory: "; FormatString$(CDBL(FreePhysicalMem))
    Used# = CDBL(TotalPhysicalMem) - CDBL(FreePhysicalMem)
    PRINT #StdoutHandle, "Used Physical Memory: "; FormatString$(Used#)
ELSE
    x1$ = MID$(x1$, 4)
    IF LEFT$(x1$, 1) = " " THEN
        IF LTRIM$(RTRIM$(MID$(x1$, 2))) = "paging" THEN
            PRINT #StdoutHandle, "Total Paging File: "; FormatString$(CDBL(TotalPagingFile))
            PRINT #StdoutHandle, "Free Paging File: "; FormatString$(CDBL(FreePagingFile))
            Used# = CDBL(TotalPagingFile) - CDBL(FreePagingFile)
            PRINT #StdoutHandle, "Used Paging File: "; FormatString$(Used#)
        ELSE
            IF LTRIM$(RTRIM$(MID$(x1$, 2))) = "virtual" THEN
                PRINT #StdoutHandle, "Total Virtual Memory: "; FormatString$(CDBL(TotalVirtualMem))
                PRINT #StdoutHandle, "Free Virtual Memory: "; FormatString$(CDBL(FreeVirtualMem))
                Used# = CDBL(TotalVirtualMem) - CDBL(FreeVirtualMem)
                PRINT #StdoutHandle, "Used Virtual Memory: "; FormatString$(Used#)
            ELSE
                PRINT #StdoutHandle, "Type MEM /? for help."
            END IF
        END IF
    ELSE
        PRINT #StdoutHandle, "Type HELP for help."
    END IF
END IF
EXIT SUB

ClockFunc:
PRINT #StdoutHandle, clock$
EXIT SUB

VerFunc:
PRINT #StdoutHandle, Program + " " + FullVersion$
EXIT SUB

ClsFunc:
CLS
EXIT SUB

BlankFunc:
s%& = SendMessageA(&HFFFF&, wm_syscommand, sc_screensave, 0&)
EXIT SUB
IF _FILEEXISTS("scrnsave.exe") THEN
    SHELL _HIDE "scrnsave.exe"
END IF
EXIT SUB

EditFunc:
IF x1$ = "edit" THEN
    CALL SaveScreen
    IF File.Loaded THEN
        IF ErrorBox(" EditFile ", "Continue editing?") THEN
            VarC = -1
            GOTO EditFunc1
        END IF
    END IF
    Var$ = FilenameBox$(3)
    IF Var$ = "" THEN
        EXIT SUB
    END IF
    VarC = 0
    GOTO EditFunc1
ELSE
    x1$ = MID$(x1$, 5)
    IF LEFT$(x1$, 1) = " " THEN
        Var$ = MID$(x1$, 2)
        Var$ = RTRIM$(LTRIM$(Var$))
        IF LEN(Var$) THEN
            CALL SaveScreen
            VarC = 0
            GOTO EditFunc1
        END IF
        PRINT #StdoutHandle, "Type EDIT /? for help."
    ELSE
        PRINT #StdoutHandle, "Type HELP for help."
    END IF
END IF
EXIT SUB

EditFunc1:
VIEW PRINT 1 TO 25
KEY OFF
CLS
COLOR 15, 0
LOCATE 25, 1
IF LEFT$(Var$, 1) = CHR$(34) THEN
    IF RIGHT$(Var$, 1) = CHR$(34) THEN
        Var$ = LEFT$(Var$, LEN(Var$) - 1)
        Var$ = MID$(Var$, 2)
    END IF
END IF
CALL ClearKeys
CALL ScrnEdit(Var$, VarZ, VarC)
CALL SetKeys
Var1$ = Nul
Var2$ = Nul
IF VarZ THEN
    IF VarZ = -1 THEN
        Var1$ = " Saved "
        Var2$ = "File stored to disk."
    ELSE
        IF VarZ = -2 THEN
            Var1$ = " Error "
            Var2$ = "Error creating file."
        ELSE
            IF VarZ = -3 THEN
                Var1$ = " Error "
                Var2$ = "Invalid filename."
            ELSE
                IF VarZ = -4 THEN
                    Var1$ = " Error "
                    Var2$ = "Interrupt."
                ELSE
                    IF VarZ = -5 THEN
                        Var1$ = " Error "
                        Var2$ = "Overflow."
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
    COLOR 7
    CLS
    CALL RestoreScreen
    LOCATE 23, 1, 1
    COLOR 15
    CALL DisplayPrompt
    PRINT
    LOCATE 24, 1, 1
ELSE
    LOCATE 24, 1, 1
    COLOR 14
    CALL DisplayVersion
    VIEW PRINT 2 TO 23
    COLOR 7
    CLS
    CALL RestoreScreen
END IF
IF KEYOFF = 0 THEN
    KEY ON
END IF
IF LEN(Var1$) THEN
    VarX = MessageBox(Var1$, Var2$)
END IF
EXIT SUB

RedrawFunc:
VIEW PRINT 1 TO 25
KEY OFF
CLS
COLOR 15, 0
LOCATE 25, 1
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
    COLOR 7
    LOCATE 24, 1, 1
ELSE
    LOCATE 24, 1, 1
    COLOR 14
    CALL DisplayVersion
    VIEW PRINT 2 TO 23
    COLOR 7
    LOCATE 23, 1, 1
END IF
IF KEYOFF = 0 THEN
    KEY ON
END IF
EXIT SUB

HexCalcFunc:
CALL HexCalc
EXIT SUB

AsciiChartFunc:
CALL AsciiChart
EXIT SUB

HexChartFunc:
CALL HexChart
EXIT SUB

WhatisFunc:
x1$ = MID$(x1$, 7)
x$ = LTRIM$(x1$)
x = -2
EXIT SUB

HelpFunc:
IF x1$ = "help" THEN
    CALL DisplayHelp("commands")
ELSE
    x1$ = MID$(x1$, 5)
    IF LEFT$(x1$, 1) = " " THEN
        x1$ = MID$(x1$, 2)
        x1$ = RTRIM$(x1$)
        CALL DisplayHelp(x1$)
    ELSE
        PRINT #StdoutHandle, "Type HELP for help."
    END IF
END IF
EXIT SUB

PromptFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "prompt" THEN
    PRINT "Prompt is "; Prompt$
    EXIT SUB
END IF
x1$ = MID$(x1$, 7)
IF LEFT$(x1$, 1) = " " THEN
    Prompt$ = MID$(x$, 8)
ELSE
    PRINT #StdoutHandle, "Type PROMPT /? for help."
END IF
EXIT SUB

ColorsFunc:
x1$ = RTRIM$(LTRIM$(x1$))
IF x1$ = "color" THEN
    PRINT #StdoutHandle, "Type COLOR /? for help."
    EXIT SUB
END IF
x1$ = MID$(x1$, 6)
IF LEFT$(x1$, 1) = " " THEN
    y$ = MID$(x1$, 2)
    V = INSTR(y$, ",")
    IF V THEN
        f$ = LEFT$(y$, V - 1)
        b$ = MID$(y$, V + 1)
        f$ = LTRIM$(RTRIM$(f$))
        b$ = LTRIM$(RTRIM$(b$))
        IF f$ = "" AND b$ = "" THEN
            PRINT #StdoutHandle, "Type COLOR /? for help."
            EXIT SUB
        END IF
        IF LEN(f$) AND LEN(b$) THEN
            fore = INT(VAL(f$))
            back = INT(VAL(b$))
            IF fore >= 0 AND fore <= 15 AND back >= 0 AND back <= 7 THEN
                PromptForeground = fore
                PromptBackground = back
                EXIT SUB
            END IF
        ELSE
            IF LEN(f$) THEN
                fore = INT(VAL(f$))
                IF fore >= 0 AND fore <= 15 THEN
                    PromptForeground = fore
                    EXIT SUB
                END IF
            ELSE
                IF LEN(b$) THEN
                    back = INT(VAL(b$))
                    IF back >= 0 AND back <= 7 THEN
                        PromptBackground = back
                        EXIT SUB
                    END IF
                END IF
            END IF
        END IF
    ELSE
        fore = INT(VAL(y$))
        IF fore >= 0 AND fore <= 15 THEN
            PromptForeground = fore
            EXIT SUB
        END IF
    END IF
    PRINT #StdoutHandle, "Type COLOR /? for help."
ELSE
    PRINT #StdoutHandle, "Type HELP for help."
END IF
END SUB

SUB DisplayVersion
V$ = Program + " " + FullVersion$ + " OS " + _OS$
V$ = LEFT$(V$, 79)
PRINT V$;
END SUB

SUB DisplayHelp (V$)
SELECT CASE V$
    CASE "shell"
        PRINT "SHELL"
        PRINT
        PRINT "  Shells to DOS"
        PRINT
        PRINT "SHELL <command>"
        PRINT
        PRINT "  Shells to DOS and performs command"
    CASE "redraw"
        PRINT #StdoutHandle, "REDRAW"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Redraws the main shell screen."
    CASE "menu"
        PRINT #StdoutHandle, "MENU"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Starts simple menu function for access to files/directories."
    CASE "date"
        PRINT #StdoutHandle, "DATE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Displays and sets date."
    CASE "time"
        PRINT #StdoutHandle, "TIME"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Displays and sets time."
    CASE "toggle"
        PRINT #StdoutHandle, "TOGGLE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Toggles full screen mode."
    CASE "cpu"
        PRINT #StdoutHandle, "CPU [ON][OFF][MEMORY]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns cpu status on or off."
        PRINT #StdoutHandle, "  'memory' toggles memory usage on or off."
    CASE "screensaver"
        PRINT #StdoutHandle, "SCREENSAVER [ON][OFF][<n>]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns screensaver on, off, or sets to <n> seconds."
    CASE "debug"
        PRINT #StdoutHandle, "DEBUG [ON][OFF]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns debug on or off."
    CASE "status"
        PRINT #StdoutHandle, "STATUS [ON][OFF]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns status line on or off."
    CASE "key"
        PRINT #StdoutHandle, "KEY [ON][OFF][LIST][RESET][<n>,<text>]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns key statusline on, off, lists current keys, resets keys,"
        PRINT #StdoutHandle, "  assigns <text> to a key number to <n> from 1 to 12."
    CASE "system"
        PRINT #StdoutHandle, "SYSTEM"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Abort program and exit to system."
    CASE "quit"
        PRINT #StdoutHandle, "QUIT"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Quit program and exit to system."
    CASE "mem"
        PRINT #StdoutHandle, "MEM [VIRTUAL][PAGING]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Display total/free/used physical system RAM."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "    'virtual' displays virtual RAM."
        PRINT #StdoutHandle, "    'paging' displays paging file RAM."
    CASE "clock"
        PRINT #StdoutHandle, "CLOCK"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Display current date and time."
    CASE "ver"
        PRINT #StdoutHandle, "VER"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Display program version."
    CASE "cls"
        PRINT #StdoutHandle, "CLS"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Clears text screen viewport."
    CASE "edit"
        PRINT #StdoutHandle, "EDIT"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Starts the full screen editor."
    CASE "hexcalc"
        PRINT #StdoutHandle, "HEXCALC"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Starts the Hexidecimal-Decimal calculator."
    CASE "ascii"
        PRINT #StdoutHandle, "ASCII"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Displays charts of ASCII values and characters."
    CASE "hexchart"
        PRINT #StdoutHandle, "HEXCHART"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Displays charts of hexidecimal values and characters."
    CASE "whatis"
        PRINT #StdoutHandle, "WHATIS"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Starts the recursive descent parser."
    CASE "help"
        PRINT #StdoutHandle, "HELP [COMMAND]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Displays commands available or help for specific command."
    CASE "prompt"
        PRINT #StdoutHandle, "PROMPT [text]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Changes the QB64shell command prompt."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "    text    Specifies a new command prompt."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Prompt can contain the following special codes:"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "    $D   Current date      $T   Current time"
        PRINT #StdoutHandle, "    $N   Current drive     $P   Current drive and path"
        PRINT #StdoutHandle, "    $Q   Current volume    $V   QB64shell version number"
        PRINT #StdoutHandle, "    $H   Backspace         $_   Carriage return and linefeed"
        PRINT #StdoutHandle, "    $J   Clear screen      $$   Display a $ character."
        PRINT #StdoutHandle, "    $G   Display > symbol  $L   Display a < symbol"
        PRINT #StdoutHandle, "    $S   Store current cursor position"
        PRINT #StdoutHandle, "    $U   Restore cursor position"
        PRINT #StdoutHandle, "    $F[x;y]  Set cursor on screen"
        PRINT #StdoutHandle, "    $C[n] Change foreground color"
        PRINT #StdoutHandle, "    $E[n] Change background color"
    CASE "color"
        PRINT #StdoutHandle, "COLOR <foreground>,<background>"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Changes the shell prompt color."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Color code values:"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Fore/Background             Foreground"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "   0               Black        8               Gray"
        PRINT #StdoutHandle, "   1               Blue         9               Light Blue"
        PRINT #StdoutHandle, "   2               Green       10               Light green"
        PRINT #StdoutHandle, "   3               Cyan        11               Light cyan"
        PRINT #StdoutHandle, "   4               Red         12               Light red"
        PRINT #StdoutHandle, "   5               Magenta     13               Light magenta"
        PRINT #StdoutHandle, "   6               Brown       14               Yellow"
        PRINT #StdoutHandle, "   7               White       15               High-intensity white"
    CASE "commands"
        PRINT #StdoutHandle, "QB64shell commands:"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Basic commands:"
        PRINT #StdoutHandle, "  CLS     CPU       KEY       MEM       VER      TOGGLE   MONITOR"
        PRINT #StdoutHandle, "  CLOCK   DATE      TIME      MENU      HELP     EDIT     PROFILES"
        PRINT #StdoutHandle, "  DEBUG   COLOR     PROMPT    STATUS    SYSTEM   QUIT     SCREENSAVER"
        PRINT #StdoutHandle, "  ASCII   HEXCHART  HEXCALC   REDRAW    WHATIS   COUNT"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Filters:  FIND      MORE      SORT      TYPE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Filename commands:"
        PRINT #StdoutHandle, "  COMPFILE COPY     DELETE    DIR       MKFILE    RENAME"
        PRINT #StdoutHandle, "  ENCRYPT  DECRYPT  GETATTR   SETATTR   LISTFILE  TOUCHFILE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Directory commands: COMPDIR   LISTDIR   PUSHD     POPD"
        PRINT #StdoutHandle, "  TREEDIR  TREEDEL  TOUCHDIR  CHDIR     MKDIR     RMDIR"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Volume commands:"
        PRINT #StdoutHandle, "  DRIVES   LABEL    VOL       LISTDRIVE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "See also " + Quote + "HELP STREAM" + Quote + " for additional command parsing."
    CASE "stream"
        PRINT #StdoutHandle, "Streaming is done using the following symbols on the command line:"
        PRINT #StdoutHandle, "  <  redirect input from file."
        PRINT #StdoutHandle, "  >  redirect output to file."
        PRINT #StdoutHandle, "  >>  append redirected output to file."
        PRINT #StdoutHandle, "  >>>  append errors to file."
        PRINT #StdoutHandle, "  |   redirect i/o from command to next."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Example usage:"
        PRINT #StdoutHandle, "   Type Filelist.txt /c | Sort /c > output.lst"
        PRINT #StdoutHandle, "   Treedir temp*.* /r/b/c | Dir *.lst /c"
    CASE "profiles"
        PRINT #StdoutHandle, "PROFILES [ON][OFF][LIST][EDIT]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns profiles on or off, lists profiles, edits profiles."
    CASE "monitor"
        PRINT #StdoutHandle, "MONITOR [ON][OFF][LIMIT][TIMER]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Turns memory monitor on or off, sets limit and timer."
        PRINT #StdoutHandle, "     Limit is free RAM in MB. For example, MONITOR LIMIT=20"
        PRINT #StdoutHandle, "     Timer is checking in seconds. For example, MONITOR TIMER=60"
    CASE "pushd"
        PRINT #StdoutHandle, "PUSHD <path/netpath>"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Pushs a fully qualified pathname or a network pathname"
        PRINT #StdoutHandle, "  on the stack and moves default prompt to the path."
    CASE "popd"
        PRINT #StdoutHandle, "POPD"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Pops the current path off the stack."
    CASE "compdir"
        PRINT #StdoutHandle, "COMPDIR"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Compares the files in two directories for equivalent"
        PRINT #StdoutHandle, "  files of same name, size, and content."
    CASE "compfile"
        PRINT #StdoutHandle, "COMPFILE [LIST]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Compares two files for equivalent size and contents."
        PRINT #StdoutHandle, "  Optional LIST displays table of offsets and hex/ascii values."
    CASE "listfile"
        PRINT #StdoutHandle, "LISTFILE [<spec>]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Lists files and files in a specification."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Where the 7 attributes from left to right are:"
        PRINT #StdoutHandle, "  A - archive"
        PRINT #StdoutHandle, "  D - directory"
        PRINT #StdoutHandle, "  S - system"
        PRINT #StdoutHandle, "  R - read-only"
        PRINT #StdoutHandle, "  H - hidden"
        PRINT #StdoutHandle, "  C - compressed"
        PRINT #StdoutHandle, "  E - encrypted"
    CASE "listdir"
        PRINT #StdoutHandle, "LISTDIR [<spec>]"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Lists directories and directories in a specification."
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "Where the 7 attributes from left to right are:"
        PRINT #StdoutHandle, "  A - archive"
        PRINT #StdoutHandle, "  D - directory"
        PRINT #StdoutHandle, "  S - system"
        PRINT #StdoutHandle, "  R - read-only"
        PRINT #StdoutHandle, "  H - hidden"
        PRINT #StdoutHandle, "  C - compressed"
        PRINT #StdoutHandle, "  E - encrypted"
    CASE "listdrive"
        PRINT #StdoutHandle, "LISTDRIVE"
        PRINT #StdoutHandle, ""
        PRINT #StdoutHandle, "  Lists current drives, and their labels, volume types, and"
        PRINT #StdoutHandle, "  free/total/used space."
    CASE ELSE
        PRINT #StdoutHandle, "Type HELP /? for help."
END SELECT
END SUB

SUB HelpAbout
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

' get box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF AboutXcoor THEN Xcoor3 = AboutXcoor
IF AboutYcoor THEN Ycoor3 = AboutYcoor

' adjust xcoor
IF Status THEN
    Z = 23
    IF Xcoor3 + 7 > Z THEN
        Xcoor3 = Xcoor3 - 1
    END IF
END IF

' draw box
GOSUB DrawAboutBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X& = _KEYHIT
    IF X& > 0 THEN
        IF X& = KEY_RALT& THEN
            EXIT DO
        END IF
        IF X& = KEY_LALT& THEN
            EXIT DO
        END IF
        'IF X& = KEY_RCTRL& THEN
        '    EXIT DO
        'END IF
        'IF X& = KEY_LCTRL& THEN
        '    EXIT DO
        'END IF
        IF X& = KEY_RSHIFT& THEN
            EXIT DO
        END IF
        IF X& = KEY_LSHIFT& THEN
            EXIT DO
        END IF
    END IF
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE CHR$(1) ' ctrl-a
                        X$ = ""
                        BackGround4 = BackGround4 + 1
                        IF BackGround4 = 8 THEN
                            BackGround4 = 0
                        END IF
                        GOSUB DrawAboutBox
                    CASE CHR$(2) ' ctrl-b
                        X$ = ""
                        BackGround4 = BackGround4 - 1
                        IF BackGround4 < 0 THEN
                            BackGround4 = 7
                        END IF
                        GOSUB DrawAboutBox
                    CASE CHR$(4) ' ctrl-d
                        X$ = ""
                        AboutBoxBorderColor = AboutBoxBorderColor + 1
                        IF AboutBoxBorderColor = 16 THEN
                            AboutBoxBorderColor = 0
                        END IF
                        GOSUB DrawAboutBox
                    CASE CHR$(5) ' ctrl-e
                        X$ = ""
                        AboutBoxTitleColor = AboutBoxTitleColor + 1
                        IF AboutBoxTitleColor = 16 THEN
                            AboutBoxTitleColor = 0
                        END IF
                        GOSUB DrawAboutBox
                    CASE CHR$(6) ' ctrl-f
                        X$ = ""
                        AboutBoxTextColor = AboutBoxTextColor + 1
                        IF AboutBoxTextColor = 16 THEN
                            AboutBoxTextColor = 0
                        END IF
                        GOSUB DrawAboutBox
                    CASE ELSE
                        EXIT DO
                END SELECT
            CASE 2
                V = ASC(RIGHT$(X$, 1))
                SELECT CASE V
                    CASE 72 ' up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawAboutBox
                        END IF
                        _KEYCLEAR
                    CASE 80 ' down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawAboutBox
                        END IF
                        _KEYCLEAR
                    CASE 75 ' left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawAboutBox
                        END IF
                        _KEYCLEAR
                    CASE 77 ' right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawAboutBox
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF LEN(X$) THEN
        EXIT DO
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen2
                            GOSUB DrawAboutBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX >= Xcoor3 + 1 AND MouseX <= Xcoor3 + 17 THEN
                IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
    IF MouseButton2 THEN
        IF MouseX >= Xcoor3 AND MouseX <= Xcoor3 + 17 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                EXIT DO
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates.
AboutXcoor = Xcoor3
AboutYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen2
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
EXIT SUB

' draw box
DrawAboutBox:
COLOR AboutBoxBorderColor, BackGround4
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR AboutBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 12, 0
PRINT " About ";

' display about text
COLOR AboutBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
PRINT "Program: " + Program
LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
PRINT "Version: " + Version + " " + Release
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
PRINT "Publish: " + Publish
LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
PRINT "Author: " + Author
LOCATE Xcoor3 + 5, Ycoor3 + 1, 0
PRINT "Email: " + Email
LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
PRINT "Url: " + Urls
RETURN
END SUB

REM MessageBox with OK/Cancel
FUNCTION QuitBox
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

' declare quit box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF QuitXcoor THEN Xcoor3 = QuitXcoor
IF QuitYcoor THEN Ycoor3 = QuitYcoor

' adjust xcoor
IF Status THEN
    Z = 23
    IF Xcoor3 + 7 > Z THEN
        Xcoor3 = Xcoor3 - 1
    END IF
END IF

' draw box
BoxButton = 1
GOSUB DrawQuitBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE "O"
                        BoxButton = 1
                        EXIT DO
                    CASE "C"
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE CHR$(1) ' ctrl-a
                        BackGround6 = BackGround6 + 1
                        IF BackGround6 = 8 THEN
                            BackGround6 = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(2) ' ctrl-b
                        BackGround7 = BackGround7 + 1
                        IF BackGround7 = 8 THEN
                            BackGround7 = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(4) ' ctrl-d
                        QuitBoxBorderColor = QuitBoxBorderColor + 1
                        IF QuitBoxBorderColor = 16 THEN
                            QuitBoxBorderColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(5) ' ctrl-e
                        QuitBoxTitleColor = QuitBoxTitleColor + 1
                        IF QuitBoxTitleColor = 16 THEN
                            QuitBoxTitleColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(6) ' ctrl-f
                        QuitBoxTextColor = QuitBoxTextColor + 1
                        IF QuitBoxTextColor = 16 THEN
                            QuitBoxTextColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(7) ' ctrl-g
                        QuitBoxButton1Color = QuitBoxButton1Color + 1
                        IF QuitBoxButton1Color = 16 THEN
                            QuitBoxButton1Color = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(8) ' ctrl-h
                        QuitBoxButton2Color = QuitBoxButton2Color + 1
                        IF QuitBoxButton2Color = 16 THEN
                            QuitBoxButton2Color = 0
                        END IF
                        GOSUB DrawQuitBox
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 75, 15 ' left/shift-tab
                        IF BoxButton = 2 THEN
                            BoxButton = 1
                        ELSE
                            BoxButton = 2
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE 77 ' right
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen2
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                CALL RestoreScreen2
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen2
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen2
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen2
                            GOSUB DrawQuitBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 4 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 4 THEN
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                IF BoxButton = 2 THEN
                    BoxButton = 1
                    GOSUB DrawQuitBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                IF BoxButton = 1 THEN
                    BoxButton = 2
                    GOSUB DrawQuitBoxButtons
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
QuitXcoor = Xcoor3
QuitYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen2
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN
    QuitBox = -1
ELSE
    QuitBox = 0
END IF
EXIT FUNCTION

' draw box
DrawQuitBox:
COLOR QuitBoxBorderColor, BackGround6
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR QuitBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 12, 0
PRINT " Quit ";

' display quit text
COLOR QuitBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT QuitBoxText$
GOSUB DrawQuitBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawQuitBoxButtons:
IF BoxButton = 1 THEN
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR QuitBoxButton1Color, BackGround7
    PRINT "<OK>";
    LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 4, Ycoor3 + 8, 0
    COLOR QuitBoxButton2Color, BackGround7
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
ELSE
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR QuitBoxButton2Color, BackGround7
    PRINT "<OK>";
    LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 4, Ycoor3 + 8, 0
    COLOR QuitBoxButton1Color, BackGround7
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
END IF
COLOR 7, 0
RETURN
END FUNCTION

REM MessageBox with YES/NO/Cancel
REM    only with full screen editor.
FUNCTION QuitBox2 (Var1$)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare quit box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF QuitXcoor2 THEN Xcoor3 = QuitXcoor2
IF QuitYcoor2 THEN Ycoor3 = QuitYcoor2

' draw box
BoxButton = 1
GOSUB DrawQuitBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 3
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE CHR$(1) ' ctrl-a
                        BackGround6 = BackGround6 + 1
                        IF BackGround6 = 8 THEN
                            BackGround6 = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(2) ' ctrl-b
                        BackGround7 = BackGround7 + 1
                        IF BackGround7 = 8 THEN
                            BackGround7 = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(4) ' ctrl-d
                        QuitBoxBorderColor = QuitBoxBorderColor + 1
                        IF QuitBoxBorderColor = 16 THEN
                            QuitBoxBorderColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(5) ' ctrl-e
                        QuitBoxTitleColor = QuitBoxTitleColor + 1
                        IF QuitBoxTitleColor = 16 THEN
                            QuitBoxTitleColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(6) ' ctrl-f
                        QuitBoxTextColor = QuitBoxTextColor + 1
                        IF QuitBoxTextColor = 16 THEN
                            QuitBoxTextColor = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(7) ' ctrl-g
                        QuitBoxButton1Color = QuitBoxButton1Color + 1
                        IF QuitBoxButton1Color = 16 THEN
                            QuitBoxButton1Color = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE CHR$(8) ' ctrl-h
                        QuitBoxButton2Color = QuitBoxButton2Color + 1
                        IF QuitBoxButton2Color = 16 THEN
                            QuitBoxButton2Color = 0
                        END IF
                        GOSUB DrawQuitBox
                    CASE ELSE
                        IF UCASE$(X$) = "Y" THEN
                            BoxButton = 1
                            EXIT DO
                        END IF
                        IF UCASE$(X$) = "N" THEN
                            BoxButton = 2
                            EXIT DO
                        END IF
                        IF UCASE$(X$) = "C" THEN
                            BoxButton = 3
                            EXIT DO
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 75, 15 ' left/shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 3
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE 77 ' right
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        GOSUB DrawQuitBoxButtons
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        X$ = ""
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        X$ = ""
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawQuitBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawQuitBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 4 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 6 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 9 AND MouseY <= Ycoor3 + 12 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 15 AND MouseY <= Ycoor3 + 22 THEN
                    BoxButton = 3
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 4 THEN
            '<YES>  <NO>  <Cancel>
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 6 THEN
                IF BoxButton <> 1 THEN
                    BoxButton = 1
                    GOSUB DrawQuitBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 9 AND MouseY <= Ycoor3 + 12 THEN
                IF BoxButton <> 2 THEN
                    BoxButton = 2
                    GOSUB DrawQuitBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 15 AND MouseY <= Ycoor3 + 22 THEN
                IF BoxButton <> 3 THEN
                    BoxButton = 3
                    GOSUB DrawQuitBoxButtons
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
QuitXcoor2 = Xcoor3
QuitYcoor2 = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN ' YES
    QuitBox2 = -1
ELSE
    IF BoxButton = 2 THEN ' NO
        QuitBox2 = 1
    ELSE
        QuitBox2 = 0 ' Cancel
    END IF
END IF
EXIT FUNCTION

' draw box
DrawQuitBox:
COLOR QuitBoxBorderColor, BackGround6
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR QuitBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 12, 0
PRINT " Quit ";

' display quit text
COLOR QuitBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 2, 0
PRINT Var1$;
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT QuitBoxText2$
GOSUB DrawQuitBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawQuitBoxButtons:
SELECT CASE BoxButton
    '<YES>  <NO>  <Cancel>
    CASE 1
        LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
        COLOR QuitBoxButton1Color, QuitBoxButtonBackGround
        PRINT "<YES>";
        LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "Y";

        LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<NO>";
        LOCATE Xcoor3 + 4, Ycoor3 + 10, 0
        COLOR Magenta
        PRINT "N";

        LOCATE Xcoor3 + 4, Ycoor3 + 15, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 4, Ycoor3 + 16, 0
        COLOR Magenta
        PRINT "C";
    CASE 2
        LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<YES>";
        LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "Y";

        LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
        COLOR QuitBoxButton1Color, QuitBoxButtonBackGround
        PRINT "<NO>";
        LOCATE Xcoor3 + 4, Ycoor3 + 10, 0
        COLOR Magenta
        PRINT "N";

        LOCATE Xcoor3 + 4, Ycoor3 + 15, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 4, Ycoor3 + 16, 0
        COLOR Magenta
        PRINT "C";
    CASE 3
        LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<YES>";
        LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "Y";

        LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
        COLOR QuitBoxButton2Color, QuitBoxButtonBackGround
        PRINT "<NO>";
        LOCATE Xcoor3 + 4, Ycoor3 + 10, 0
        COLOR Magenta
        PRINT "N";

        LOCATE Xcoor3 + 4, Ycoor3 + 15, 0
        COLOR QuitBoxButton1Color, QuitBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 4, Ycoor3 + 16, 0
        COLOR Magenta
        PRINT "C";
END SELECT
COLOR 7, 0
RETURN
END FUNCTION

REM MessageBox with OK/Cancel
FUNCTION ErrorBox (Var0$, Var1$)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

' declare error box coordinates.
Xcoor3 = 10
Ycoor3 = 10

' set error box colors
ErrorBoxBorderColor = Yellow
ErrorBoxTitleColor = White
ErrorBoxTextColor = White
ErrorBoxButton1Color = White
ErrorBoxButton2Color = Plain
ErrorBackGround1 = 1
ErrorBackGround2 = 0

' draw box
BoxButton = 1
GOSUB DrawErrorBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE "O"
                        BoxButton = 1
                        EXIT DO
                    CASE "C"
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawErrorBoxButtons
                    CASE CHR$(1) ' ctrl-a
                        ErrorBackGround1 = ErrorBackGround1 + 1
                        IF ErrorBackGround1 = 8 THEN
                            ErrorBackGround1 = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(2) ' ctrl-b
                        ErrorBackGround2 = ErrorBackGround2 + 1
                        IF ErrorBackGround2 = 8 THEN
                            ErrorBackGround2 = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(4) ' ctrl-d
                        ErrorBoxBorderColor = ErrorBoxBorderColor + 1
                        IF ErrorBoxBorderColor = 16 THEN
                            ErrorBoxBorderColor = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(5) ' ctrl-e
                        ErrorBoxTitleColor = ErrorBoxTitleColor + 1
                        IF ErrorBoxTitleColor = 16 THEN
                            ErrorBoxTitleColor = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(6) ' ctrl-f
                        ErrorBoxTextColor = ErrorBoxTextColor + 1
                        IF ErrorBoxTextColor = 16 THEN
                            ErrorBoxTextColor = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(7) ' ctrl-g
                        ErrorBoxButton1Color = ErrorBoxButton1Color + 1
                        IF ErrorBoxButton1Color = 16 THEN
                            ErrorBoxButton1Color = 0
                        END IF
                        GOSUB DrawErrorBox
                    CASE CHR$(8) ' ctrl-h
                        ErrorBoxButton2Color = ErrorBoxButton2Color + 1
                        IF ErrorBoxButton2Color = 16 THEN
                            ErrorBoxButton2Color = 0
                        END IF
                        GOSUB DrawErrorBox
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 75, 15 ' left/shift-tab
                        IF BoxButton = 2 THEN
                            BoxButton = 1
                        ELSE
                            BoxButton = 2
                        END IF
                        GOSUB DrawErrorBoxButtons
                    CASE 77 ' right
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawErrorBoxButtons
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen2
                                GOSUB DrawErrorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                CALL RestoreScreen2
                                GOSUB DrawErrorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen2
                                GOSUB DrawErrorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen2
                                GOSUB DrawErrorBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen2
                            GOSUB DrawErrorBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 4 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 4 THEN
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                IF BoxButton = 2 THEN
                    BoxButton = 1
                    GOSUB DrawErrorBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                IF BoxButton = 1 THEN
                    BoxButton = 2
                    GOSUB DrawErrorBoxButtons
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' restore screen area.
CALL RestoreScreen2
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN
    ErrorBox = -1
ELSE
    ErrorBox = 0
END IF
EXIT FUNCTION

' draw box
DrawErrorBox:
COLOR ErrorBoxBorderColor, ErrorBackGround1
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
TempX = 16 - LEN(Var0$) \ 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR ErrorBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT Var0$;

' display error text
COLOR ErrorBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT Var1$
GOSUB DrawErrorBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawErrorBoxButtons:
IF BoxButton = 1 THEN
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR ErrorBoxButton1Color, ErrorBackGround2
    PRINT "<OK>";
    LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 4, Ycoor3 + 8, 0
    COLOR ErrorBoxButton2Color, ErrorBackGround2
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
ELSE
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR ErrorBoxButton2Color, ErrorBackGround2
    PRINT "<OK>";
    LOCATE Xcoor3 + 4, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 4, Ycoor3 + 8, 0
    COLOR ErrorBoxButton1Color, ErrorBackGround2
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
END IF
COLOR 7, 0
RETURN
END FUNCTION

REM MessageBox with OK
FUNCTION MessageBox (Var0$, Var1$)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

' declare error box coordinates.
Xcoor3 = 10
Ycoor3 = 10

' set error box colors
MessageBoxBorderColor = Yellow
MessageBoxTitleColor = White
MessageBoxTextColor = White
MessageBoxButton1Color = White
MessageBoxButton2Color = Plain
MessageBackGround1 = 1
MessageBackGround2 = 0

' draw box
GOSUB DrawMessageBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE "O"
                        EXIT DO
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        EXIT DO
                    CASE CHR$(1) ' ctrl-a
                        MessageBackGround1 = MessageBackGround1 + 1
                        IF MessageBackGround1 = 8 THEN
                            MessageBackGround1 = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(2) ' ctrl-b
                        MessageBackGround2 = MessageBackGround2 + 1
                        IF MessageBackGround2 = 8 THEN
                            MessageBackGround2 = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(4) ' ctrl-d
                        MessageBoxBorderColor = MessageBoxBorderColor + 1
                        IF MessageBoxBorderColor = 16 THEN
                            MessageBoxBorderColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(5) ' ctrl-e
                        MessageBoxTitleColor = MessageBoxTitleColor + 1
                        IF MessageBoxTitleColor = 16 THEN
                            MessageBoxTitleColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(6) ' ctrl-f
                        MessageBoxTextColor = MessageBoxTextColor + 1
                        IF MessageBoxTextColor = 16 THEN
                            MessageBoxTextColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(7) ' ctrl-g
                        MessageBoxButton1Color = MessageBoxButton1Color + 1
                        IF MessageBoxButton1Color = 16 THEN
                            MessageBoxButton1Color = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(8) ' ctrl-h
                        MessageBoxButton2Color = MessageBoxButton2Color + 1
                        IF MessageBoxButton2Color = 16 THEN
                            MessageBoxButton2Color = 0
                        END IF
                        GOSUB DrawMessageBox
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen2
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                CALL RestoreScreen2
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen2
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen2
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen2
                            GOSUB DrawMessageBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 4 THEN
                IF MouseY >= Ycoor3 + 14 AND MouseY <= Ycoor3 + 17 THEN
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' restore screen area.
CALL RestoreScreen2
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
MessageBox = -1
EXIT FUNCTION

' draw box
DrawMessageBox:
COLOR MessageBoxBorderColor, MessageBackGround1
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
TempX = 16 - LEN(Var0$) \ 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR MessageBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT Var0$;

' display message text
COLOR MessageBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT Var1$
GOSUB DrawMessageBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawMessageBoxButtons:
LOCATE Xcoor3 + 4, Ycoor3 + 14, 0
COLOR MessageBoxButton1Color, MessageBackGround2
PRINT "<OK>";

LOCATE Xcoor3 + 4, Ycoor3 + 15, 0
COLOR Magenta
PRINT "O";

COLOR 7, 0
RETURN
END FUNCTION

REM HelpBox with OK
FUNCTION HelpBox
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare error box coordinates.
Xcoor3 = 10
Ycoor3 = 10

' set error box colors
MessageBoxBorderColor = Yellow
MessageBoxTitleColor = White
MessageBoxTextColor = White
MessageBoxButton1Color = White
MessageBoxButton2Color = Plain
MessageBackGround1 = 1
MessageBackGround2 = 0

' draw box
GOSUB DrawMessageBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE "O"
                        EXIT DO
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        EXIT DO
                    CASE CHR$(1) ' ctrl-a
                        MessageBackGround1 = MessageBackGround1 + 1
                        IF MessageBackGround1 = 8 THEN
                            MessageBackGround1 = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(2) ' ctrl-b
                        MessageBackGround2 = MessageBackGround2 + 1
                        IF MessageBackGround2 = 8 THEN
                            MessageBackGround2 = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(4) ' ctrl-d
                        MessageBoxBorderColor = MessageBoxBorderColor + 1
                        IF MessageBoxBorderColor = 16 THEN
                            MessageBoxBorderColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(5) ' ctrl-e
                        MessageBoxTitleColor = MessageBoxTitleColor + 1
                        IF MessageBoxTitleColor = 16 THEN
                            MessageBoxTitleColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(6) ' ctrl-f
                        MessageBoxTextColor = MessageBoxTextColor + 1
                        IF MessageBoxTextColor = 16 THEN
                            MessageBoxTextColor = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(7) ' ctrl-g
                        MessageBoxButton1Color = MessageBoxButton1Color + 1
                        IF MessageBoxButton1Color = 16 THEN
                            MessageBoxButton1Color = 0
                        END IF
                        GOSUB DrawMessageBox
                    CASE CHR$(8) ' ctrl-h
                        MessageBoxButton2Color = MessageBoxButton2Color + 1
                        IF MessageBoxButton2Color = 16 THEN
                            MessageBoxButton2Color = 0
                        END IF
                        GOSUB DrawMessageBox
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        X$ = ""
                        IF Xcoor3 < 16 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        X$ = ""
                        IF Xcoor3 < 12 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        ELSE
                            IF Xcoor3 < 16 THEN
                                Xcoor3 = 16
                                CALL RestoreScreen3
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawMessageBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 16 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawMessageBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 7 THEN
                IF MouseY >= Ycoor3 + 14 AND MouseY <= Ycoor3 + 17 THEN
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
HelpBox = -1
EXIT FUNCTION

' draw box
DrawMessageBox:
COLOR MessageBoxBorderColor, MessageBackGround1
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 7
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 8, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
Var0$ = " Help "
TempX = 16 - LEN(Var0$) \ 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR MessageBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT Var0$;

' display message text
COLOR MessageBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 2, 0
PRINT " F1 - Help      Alt-D Dump"
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT " F2 - Savefile  Alt-J Jump"
LOCATE Xcoor3 + 3, Ycoor3 + 2, 0
PRINT " F3 - Search    Alt-E Edit"
LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
PRINT " Ctrl-E Repeat   (append)"
LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
PRINT " Ctrl-F Search  Alt-P Print"
LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
PRINT " Ctrl-G Goto    Alt-S Search"
GOSUB DrawMessageBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawMessageBoxButtons:
LOCATE Xcoor3 + 7, Ycoor3 + 14, 0
COLOR MessageBoxButton1Color, MessageBackGround2
PRINT "<OK>";

LOCATE Xcoor3 + 7, Ycoor3 + 15, 0
COLOR Magenta
PRINT "O";

COLOR 7, 0
RETURN
END FUNCTION

FUNCTION SearchBox (X$, Var1) ' Var1=1 Hex, =2 Ascii.
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare input box coordinates.
XC = Xcoor3 + 3
YC = Ycoor3 + 8
LC = 22
IF SearchXcoor THEN Xcoor3 = SearchXcoor
IF SearchYcoor THEN Ycoor3 = SearchYcoor

' draw box
BoxButton = 1
WildcardSelect = 0
SensitiveSelect = 0
GOSUB DrawSearchBox

' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
LineX = 0
IF LEN(X$) THEN
    LineX = -1
END IF
Insert = -1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE I$
                    CASE CHR$(8) ' backspace
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF LineX THEN
                                    LineX = 0
                                    X$ = ""
                                    Xposition1 = 1
                                    Xposition2 = 0
                                    GOSUB DisplayFileLine
                                END IF
                                IF Xposition2 > 0 THEN
                                    IF Insert THEN ' destructive backspace
                                        X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                                    END IF
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE CHR$(13)
                        IF BoxButton = 5 THEN
                            BoxButton = 0
                        ELSE
                            BoxButton = -1
                        END IF
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 0
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 6 THEN
                            BoxButton = 1
                        END IF
                        SELECT CASE BoxButton
                            CASE 1 ' entry area
                                GOSUB DisplayFileLine
                            CASE 2 ' checkbox1
                                LOCATE Xcoor3 + 4, Ycoor3 + 2, 1, 8, 8
                            CASE 3 ' checkbox2
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' ok
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 5, 1, 8, 8
                            CASE 5 'cancel
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                        END SELECT
                    CASE CHR$(1) ' ctrl-a
                        SearchBoxBackGround = SearchBoxBackGround + 1
                        IF SearchBoxBackGround = 8 THEN
                            SearchBoxBackGround = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(2) ' ctrl-b
                        SearchBoxButtonBackGround = SearchBoxButtonBackGround + 1
                        IF SearchBoxButtonBackGround = 8 THEN
                            SearchBoxButtonBackGround = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(4) ' ctrl-d
                        SearchBoxBorderColor = SearchBoxBorderColor + 1
                        IF SearchBoxBorderColor = 16 THEN
                            SearchBoxBorderColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(5) ' ctrl-e
                        SearchBoxTitleColor = SearchBoxTitleColor + 1
                        IF SearchBoxTitleColor = 16 THEN
                            SearchBoxTitleColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(6) ' ctrl-f
                        SearchBoxTextColor = SearchBoxTextColor + 1
                        IF SearchBoxTextColor = 16 THEN
                            SearchBoxTextColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(7) ' ctrl-g
                        SearchBoxButton1Color = SearchBoxButton1Color + 1
                        IF SearchBoxButton1Color = 16 THEN
                            SearchBoxButton1Color = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(11) ' ctrl-k
                        SearchBoxButton2Color = SearchBoxButton2Color + 1
                        IF SearchBoxButton2Color = 16 THEN
                            SearchBoxButton2Color = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE ELSE
                        IF BoxButton >= 2 AND BoxButton <= 5 THEN
                            IF UCASE$(I$) = "W" THEN
                                WildcardSelect = NOT WildcardSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                            IF UCASE$(I$) = "S" THEN
                                SensitiveSelect = NOT SensitiveSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                            IF UCASE$(I$) = "O" THEN
                                BoxButton = -1
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "C" THEN
                                BoxButton = 0
                                EXIT DO
                            END IF
                        END IF
                        IF BoxButton = 2 THEN ' checkbox1
                            IF I$ = " " THEN
                                WildcardSelect = NOT WildcardSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                        END IF
                        IF BoxButton = 3 THEN ' checkbox2
                            IF I$ = " " THEN
                                SensitiveSelect = NOT SensitiveSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                        END IF
                        IF BoxButton = 1 THEN
                            ' append/insert character into edit area.
                            VarI = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                VarI = 0 ' disallow control charcters
                            END IF
                            IF VarI THEN
                                IF LineX THEN
                                    LineX = 0
                                    X$ = ""
                                    Xposition1 = 1
                                    Xposition2 = 0
                                END IF
                                IF Insert THEN
                                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > LC THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Xposition2 + 1 > LEN(X$) THEN
                                        X$ = X$ + I$ ' append
                                        Xposition2 = Xposition2 + 1
                                        IF Xposition2 > LC THEN
                                            Xposition1 = Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X$, Xposition2 + 1, 1) = I$
                                        Xposition2 = Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        IF BoxButton = 1 THEN
                            Insert = NOT Insert
                            IF Insert THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                    CASE 83 ' Delete
                        IF BoxButton = 1 THEN
                            LineX = 0
                            GOSUB DisplayFileLine
                            IF LEN(X$) > 0 THEN
                                X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 71 ' Home
                        IF BoxButton = 1 THEN
                            LineX = 0
                            IF Xposition2 > 0 THEN
                                Xposition1 = 1
                                Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 79 ' End
                        IF BoxButton = 1 THEN
                            LineX = 0
                            IF Xposition2 < LEN(X$) THEN
                                Xposition2 = LEN(X$)
                                Xposition1 = Xposition2 - LC + 1
                                IF Xposition1 < 1 THEN
                                    Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 75 ' Left
                        IF BoxButton = 1 THEN
                            LineX = 0
                            IF LEN(X$) THEN
                                IF Xposition2 > 0 THEN
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF BoxButton = 1 THEN
                            LineX = 0
                            IF LEN(X$) THEN
                                IF Xposition2 < LEN(X$) THEN
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > Xposition1 + LC - 1 THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 72 ' up
                        IF BoxButton = 2 THEN ' checkbox1
                            WildcardSelect = -1
                            GOSUB DrawSearchCheckBoxs
                        END IF
                        IF BoxButton = 3 THEN ' checkbox2
                            SensitiveSelect = -1
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 80 ' down
                        IF BoxButton = 2 THEN ' checkbox1
                            WildcardSelect = 0
                            GOSUB DrawSearchCheckBoxs
                        END IF
                        IF BoxButton = 3 THEN ' checkbox2
                            SensitiveSelect = 0
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 15 ' shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 5
                        END IF
                        SELECT CASE BoxButton
                            CASE 1 ' entry area
                                GOSUB DisplayFileLine
                            CASE 2 ' checkbox1
                                LOCATE Xcoor3 + 4, Ycoor3 + 2, 1, 8, 8
                            CASE 3 ' checkbox2
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' ok
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 5, 1, 8, 8
                            CASE 5 'cancel
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                        END SELECT
                    CASE 141 ' ctrl-up
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE ' click on button
            IF MouseX = XC THEN
                ' entry area
                IF MouseY >= YC AND MouseY <= YC + LC THEN
                    BoxButton = 1
                    GOSUB DrawSearchBox
                END IF
            END IF
            IF MouseX = Xcoor3 + 4 THEN
                ' click checkbox 1
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 3 THEN
                    WildcardSelect = NOT WildcardSelect
                    GOSUB DrawSearchCheckBoxs
                END IF
            END IF
            IF MouseX = Xcoor3 + 5 THEN
                ' click checkbox 2
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 3 THEN
                    SensitiveSelect = NOT SensitiveSelect
                    GOSUB DrawSearchCheckBoxs
                END IF
            END IF
            IF MouseX = Xcoor3 + 6 THEN
                ' click OK
                IF MouseY >= Ycoor3 + 5 AND MouseY <= Ycoor3 + 8 THEN
                    BoxButton = -1
                    EXIT DO
                END IF
                ' click cancel
                IF MouseY >= Ycoor3 + 11 AND MouseY <= Ycoor3 + 18 THEN
                    BoxButton = 0
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
SearchXcoor = Xcoor3
SearchYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton THEN
    SearchBox = -1
ELSE
    SearchBox = 0
END IF
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawSearchBox:
COLOR SearchBoxBorderColor, SearchBoxBackGround
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
SearchBoxTitle$ = " Search "
TempX = 16 - LEN(SearchBoxTitle$) / 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR SearchBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT SearchBoxTitle$;

' display search text
IF Var1 = 1 THEN
    COLOR SearchBoxTextColor
    LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
    PRINT "Enter HEX bytes"
    LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
    PRINT "  (space separated)"
END IF
IF Var1 = 2 THEN
    COLOR SearchBoxTextColor
    LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
    PRINT "Enter ASCII string:"
END IF

LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
PRINT "Entry: ";

LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
COLOR SearchBoxButton1Color, SearchBoxBackGround
PRINT "[ ] Use Wildcards";
LOCATE Xcoor3 + 4, Ycoor3 + 9, 0
COLOR Magenta
PRINT "W";

LOCATE Xcoor3 + 5, Ycoor3 + 1, 0
COLOR SearchBoxButton1Color, SearchBoxBackGround
PRINT "[ ] Case Sensitive";
LOCATE Xcoor3 + 5, Ycoor3 + 10, 0
COLOR Magenta
PRINT "S";

GOSUB DrawSearchCheckBoxs
GOSUB DrawSearchBoxButtons
XC = Xcoor3 + 3
YC = Ycoor3 + 8

COLOR White, Black
LOCATE XC, YC, 1
PRINT SPACE$(LC);

SELECT CASE BoxButton
    CASE 1 ' entry area
        GOSUB DisplayFileLine
    CASE 2 ' checkbox1
        LOCATE Xcoor3 + 4, Ycoor3 + 2, 1
    CASE 3 ' checkbox2
        LOCATE Xcoor3 + 5, Ycoor3 + 2, 1
    CASE 4 ' ok
        GOSUB DrawSearchBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 5, 1
    CASE 5 'cancel
        GOSUB DrawSearchBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 11, 1
END SELECT
LOCATE , , , 8, 8
RETURN

' display checkboxs
DrawSearchCheckBoxs:
'-[ ] Use wildcards
'-[ ] Case-sensitive

IF WildcardSelect THEN
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT "X";
ELSE
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT " ";
END IF
IF SensitiveSelect THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT "X";
ELSE
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT " ";
END IF
IF BoxButton = 2 THEN
    LOCATE Xcoor3 + 4, Ycoor3 + 2, 1, 8, 8
ELSE
    IF BoxButton = 3 THEN
        LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
    END IF
END IF
RETURN

' display buttuns
DrawSearchBoxButtons:
SELECT CASE BoxButton
    CASE 4 ' OK
        LOCATE Xcoor3 + 6, Ycoor3 + 5, 0
        COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 6, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";
    CASE 5 ' cancel
        LOCATE Xcoor3 + 6, Ycoor3 + 5, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 6, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";
    CASE ELSE
        LOCATE Xcoor3 + 6, Ycoor3 + 5, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 6, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";
END SELECT
COLOR White, Black
RETURN

' display current entry area.
DisplayFileLine:
LOCATE XC, YC, 1
PRINT SPACE$(LC);

' display edit line and position cursor.
LOCATE XC, YC, 1
IF LEN(X$) THEN
    IF LineX THEN
        COLOR 0, 7
    ELSE
        COLOR White
    END IF
    PRINT MID$(X$, Xposition1, LC);
END IF
LOCATE XC, YC + Xposition2 - Xposition1 + 1, 1
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
COLOR White, Black
RETURN

ToggleInsert:
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN
END FUNCTION

REM SearchReplaceBox
FUNCTION SearchBox2 (X1$, X2$)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare input box coordinates.
XC1 = Xcoor3 + 2
YC1 = Ycoor3 + 1
XC2 = Xcoor3 + 4
YC2 = Ycoor3 + 1
LC = 30
IF SearchXcoor THEN Xcoor3 = SearchXcoor
IF SearchYcoor THEN Ycoor3 = SearchYcoor

' draw box
BoxButton = 1
SensitiveSelect = 0
GOSUB DrawSearchBox

' init entry area position.
Box1Xposition1 = 1
Box1Xposition2 = 0
Box2Xposition1 = 1
Box2Xposition2 = 0

' display edit entry area.
Insert1 = -1
Insert2 = -1
LineX2 = 0
IF LEN(X2$) THEN
    LineX2 = -1
END IF
BoxButton = 2
GOSUB DisplayFileLine

LineX1 = 0
IF LEN(X1$) THEN
    LineX1 = -1
END IF
BoxButton = 1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE I$
                    CASE CHR$(8) ' backspace
                        IF BoxButton = 1 THEN
                            IF LEN(X1$) THEN
                                IF LineX1 THEN
                                    LineX1 = 0
                                    X1$ = ""
                                    Box1Xposition1 = 1
                                    Box1Xposition2 = 0
                                    GOSUB DisplayFileLine
                                END IF
                                IF Box1Xposition2 > 0 THEN
                                    IF Insert1 THEN ' destructive backspace
                                        X1$ = LEFT$(X1$, Box1Xposition2 - 1) + MID$(X1$, Box1Xposition2 + 1)
                                    END IF
                                    Box1Xposition2 = Box1Xposition2 - 1
                                    IF Box1Xposition2 - LC <= Box1Xposition1 THEN
                                        Box1Xposition1 = Box1Xposition1 - 1
                                        IF Box1Xposition1 < 1 THEN
                                            Box1Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            IF LEN(X2$) THEN
                                IF LineX2 THEN
                                    LineX2 = 0
                                    X2$ = ""
                                    Box2Xposition1 = 1
                                    Box2Xposition2 = 0
                                    GOSUB DisplayFileLine
                                END IF
                                IF Box2Xposition2 > 0 THEN
                                    IF Insert2 THEN ' destructive backspace
                                        X2$ = LEFT$(X2$, Box2Xposition2 - 1) + MID$(X2$, Box2Xposition2 + 1)
                                    END IF
                                    Box2Xposition2 = Box2Xposition2 - 1
                                    IF Box2Xposition2 - LC <= Box2Xposition1 THEN
                                        Box2Xposition1 = Box2Xposition1 - 1
                                        IF Box2Xposition1 < 1 THEN
                                            Box2Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27) ' exit
                        BoxButton = 6
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 7 THEN
                            BoxButton = 1
                        END IF
                        SELECT CASE BoxButton
                            CASE 1, 2 ' entry area
                                GOSUB DrawSearchBoxButtons
                                GOSUB DisplayFileLine
                            CASE 3 ' checkbox
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' replace
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 1, 1, 8, 8
                            CASE 5 ' replaceall
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                            CASE 6 ' exit
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 25, 1, 8, 8
                        END SELECT
                    CASE CHR$(1) ' ctrl-a
                        SearchBoxBackGround = SearchBoxBackGround + 1
                        IF SearchBoxBackGround = 8 THEN
                            SearchBoxBackGround = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(2) ' ctrl-b
                        SearchBoxButtonBackGround = SearchBoxButtonBackGround + 1
                        IF SearchBoxButtonBackGround = 8 THEN
                            SearchBoxButtonBackGround = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(4) ' ctrl-d
                        SearchBoxBorderColor = SearchBoxBorderColor + 1
                        IF SearchBoxBorderColor = 16 THEN
                            SearchBoxBorderColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(5) ' ctrl-e
                        SearchBoxTitleColor = SearchBoxTitleColor + 1
                        IF SearchBoxTitleColor = 16 THEN
                            SearchBoxTitleColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(6) ' ctrl-f
                        SearchBoxTextColor = SearchBoxTextColor + 1
                        IF SearchBoxTextColor = 16 THEN
                            SearchBoxTextColor = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(7) ' ctrl-g
                        SearchBoxButton1Color = SearchBoxButton1Color + 1
                        IF SearchBoxButton1Color = 16 THEN
                            SearchBoxButton1Color = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE CHR$(11) ' ctrl-k
                        SearchBoxButton2Color = SearchBoxButton2Color + 1
                        IF SearchBoxButton2Color = 16 THEN
                            SearchBoxButton2Color = 0
                        END IF
                        GOSUB DrawSearchBox
                    CASE ELSE
                        IF BoxButton >= 3 AND BoxButton <= 6 THEN
                            IF UCASE$(I$) = "S" THEN
                                SensitiveSelect = NOT SensitiveSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                            IF UCASE$(I$) = "R" THEN
                                BoxButton = 4
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "A" THEN
                                BoxButton = 5
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "E" THEN
                                BoxButton = 6
                                EXIT DO
                            END IF
                        END IF
                        IF BoxButton = 3 THEN ' checkbox
                            IF I$ = " " THEN
                                SensitiveSelect = NOT SensitiveSelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                        END IF
                        IF BoxButton = 1 THEN
                            ' append/insert character into edit area.
                            VarI = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                VarI = 0 ' disallow control charcters
                            END IF
                            IF VarI THEN
                                IF LineX1 THEN
                                    LineX1 = 0
                                    X1$ = ""
                                    Box1Xposition1 = 1
                                    Box1Xposition2 = 0
                                END IF
                                IF Insert1 THEN
                                    X1$ = LEFT$(X1$, Box1Xposition2) + I$ + MID$(X1$, Box1Xposition2 + 1)
                                    Box1Xposition2 = Box1Xposition2 + 1
                                    IF Box1Xposition2 > LC THEN
                                        Box1Xposition1 = Box1Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Box1Xposition2 + 1 > LEN(X1$) THEN
                                        X1$ = X1$ + I$ ' append
                                        Box1Xposition2 = Box1Xposition2 + 1
                                        IF Box1Xposition2 > LC THEN
                                            Box1Xposition1 = Box1Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X1$, Box1Xposition2 + 1, 1) = I$
                                        Box1Xposition2 = Box1Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            ' append/insert character into edit area.
                            VarI = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                VarI = 0 ' disallow control charcters
                            END IF
                            IF VarI THEN
                                IF LineX2 THEN
                                    LineX2 = 0
                                    X2$ = ""
                                    Box2Xposition1 = 1
                                    Box2Xposition2 = 0
                                END IF
                                IF Insert2 THEN
                                    X2$ = LEFT$(X2$, Box2Xposition2) + I$ + MID$(X2$, Box2Xposition2 + 1)
                                    Box2Xposition2 = Box2Xposition2 + 1
                                    IF Box2Xposition2 > LC THEN
                                        Box2Xposition1 = Box2Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Box2Xposition2 + 1 > LEN(X2$) THEN
                                        X2$ = X2$ + I$ ' append
                                        Box2Xposition2 = Box2Xposition2 + 1
                                        IF Box2Xposition2 > LC THEN
                                            Box2Xposition1 = Box2Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X2$, Box2Xposition2 + 1, 1) = I$
                                        Box2Xposition2 = Box2Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        IF BoxButton = 1 THEN
                            Insert1 = NOT Insert1
                            IF Insert1 THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            Insert2 = NOT Insert2
                            IF Insert2 THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                    CASE 83 ' Delete
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            GOSUB DisplayFileLine
                            IF LEN(X1$) > 0 THEN
                                X1$ = LEFT$(X1$, Box1Xposition2) + MID$(X1$, Box1Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            GOSUB DisplayFileLine
                            IF LEN(X2$) > 0 THEN
                                X2$ = LEFT$(X2$, Box2Xposition2) + MID$(X2$, Box2Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 71 ' Home
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF Box1Xposition2 > 0 THEN
                                Box1Xposition1 = 1
                                Box1Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF Box2Xposition2 > 0 THEN
                                Box2Xposition1 = 1
                                Box2Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 79 ' End
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF Box1Xposition2 < LEN(X1$) THEN
                                Box1Xposition2 = LEN(X1$)
                                Box1Xposition1 = Box1Xposition2 - LC + 1
                                IF Box1Xposition1 < 1 THEN
                                    Box1Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF Box2Xposition2 < LEN(X2$) THEN
                                Box2Xposition2 = LEN(X2$)
                                Box2Xposition1 = Box2Xposition2 - LC + 1
                                IF Box2Xposition1 < 1 THEN
                                    Box2Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 75 ' Left
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF LEN(X1$) THEN
                                IF Box1Xposition2 > 0 THEN
                                    Box1Xposition2 = Box1Xposition2 - 1
                                    IF Box1Xposition2 - LC <= Box1Xposition1 THEN
                                        Box1Xposition1 = Box1Xposition1 - 1
                                        IF Box1Xposition1 < 1 THEN
                                            Box1Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF LEN(X2$) THEN
                                IF Box2Xposition2 > 0 THEN
                                    Box2Xposition2 = Box2Xposition2 - 1
                                    IF Box2Xposition2 - LC <= Box2Xposition1 THEN
                                        Box2Xposition1 = Box2Xposition1 - 1
                                        IF Box2Xposition1 < 1 THEN
                                            Box2Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF LEN(X1$) THEN
                                IF Box1Xposition2 < LEN(X1$) THEN
                                    Box1Xposition2 = Box1Xposition2 + 1
                                    IF Box1Xposition2 > Box1Xposition1 + LC - 1 THEN
                                        Box1Xposition1 = Box1Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF LEN(X2$) THEN
                                IF Box2Xposition2 < LEN(X2$) THEN
                                    Box2Xposition2 = Box2Xposition2 + 1
                                    IF Box2Xposition2 > Box2Xposition1 + LC - 1 THEN
                                        Box2Xposition1 = Box2Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 72 ' up
                        IF BoxButton = 3 THEN ' checkbox
                            SensitiveSelect = -1
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 80 ' down
                        IF BoxButton = 3 THEN ' checkbox
                            SensitiveSelect = 0
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 15 ' shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 6
                        END IF
                        SELECT CASE BoxButton
                            CASE 1, 2 ' entry area
                                GOSUB DrawSearchBoxButtons
                                GOSUB DisplayFileLine
                            CASE 3 ' checkbox
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' replace
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 1, 1, 8, 8
                            CASE 5 ' replace all
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                            CASE 6 ' exit
                                GOSUB DrawSearchBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 25, 1, 8, 8
                        END SELECT
                    CASE 141 ' ctrl-up
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawSearchBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawSearchBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE ' click on button
            IF MouseX = XC1 THEN
                ' entry area 1
                IF MouseY >= YC1 AND MouseY <= YC1 + LC THEN
                    BoxButton = 1
                    GOSUB DrawSearchBox
                END IF
            END IF
            IF MouseX = XC2 THEN
                ' entry area 2
                IF MouseY >= YC2 AND MouseY <= YC2 + LC THEN
                    BoxButton = 2
                    GOSUB DrawSearchBox
                END IF
            END IF
            IF MouseX = Xcoor3 + 5 THEN
                ' click checkbox
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 3 THEN
                    SensitiveSelect = NOT SensitiveSelect
                    GOSUB DrawSearchCheckBoxs
                END IF
            END IF
            IF MouseX = Xcoor3 + 6 THEN
                ' click <replace>
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 9 THEN
                    BoxButton = 4
                    EXIT DO
                END IF
                ' click <replace all>
                IF MouseY >= Ycoor3 + 11 AND MouseY <= Ycoor3 + 23 THEN
                    BoxButton = 5
                    EXIT DO
                END IF
                ' click <exit>
                IF MouseY >= Ycoor3 + 25 AND MouseY <= Ycoor3 + 30 THEN
                    BoxButton = 6
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
SearchXcoor = Xcoor3
SearchYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
SELECT CASE BoxButton
    CASE 1 ' search
        SearchBox2 = 1
    CASE 2 ' replace
        SearchBox2 = 2
    CASE 3 ' sensitive
        SearchBox2 = 1
    CASE 4 ' replace
        SearchBox2 = 1
    CASE 5 ' replaceall
        SearchBox2 = 2
    CASE 6 ' cancel
        SearchBox2 = 0
END SELECT
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawSearchBox:
COLOR SearchBoxBorderColor, SearchBoxBackGround
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
SearchBoxTitle$ = " Search "
TempX = 16 - LEN(SearchBoxTitle$) / 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR SearchBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT SearchBoxTitle$;

' display search text
COLOR SearchBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
PRINT "Search string:"
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
PRINT "Replacement string:"

LOCATE Xcoor3 + 5, Ycoor3 + 1, 0
COLOR SearchBoxButton1Color, SearchBoxBackGround
PRINT "[ ] Case Sensitive";
LOCATE Xcoor3 + 5, Ycoor3 + 10, 0
COLOR Magenta
PRINT "S";

GOSUB DrawSearchCheckBoxs
GOSUB DrawSearchBoxButtons
LOCATE , , , 8, 8
XC1 = Xcoor3 + 2
YC1 = Ycoor3 + 1
XC2 = Xcoor3 + 4
YC2 = Ycoor3 + 1

COLOR White, Black
LOCATE XC1, YC1, 1
PRINT SPACE$(LC);
LOCATE XC2, YC2, 1
PRINT SPACE$(LC);
IF LEN(X1$) THEN
    LOCATE XC1, YC1, 1
    COLOR White
    PRINT MID$(X1$, Box1Xposition1, LC);
END IF
IF LEN(X2$) THEN
    LOCATE XC2, YC2, 1
    COLOR White
    PRINT MID$(X2$, Box2Xposition1, LC);
END IF

SELECT CASE BoxButton
    CASE 1, 2 ' entry area
        GOSUB DisplayFileLine
    CASE 3 ' checkbox
        LOCATE Xcoor3 + 5, Ycoor3 + 2, 1
    CASE 4 ' replace
        GOSUB DrawSearchBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 1
    CASE 5 ' replaceall
        GOSUB DrawSearchBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 11, 1
    CASE 6 ' cancel
        GOSUB DrawSearchBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 25, 1
END SELECT
RETURN

' display checkboxs
DrawSearchCheckBoxs:
'-[ ] Case-sensitive

IF SensitiveSelect THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT "X";
ELSE
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
    PRINT " ";
END IF
IF BoxButton = 3 THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
END IF
RETURN

' display buttuns
DrawSearchBoxButtons:
'-(Replace> <Replace All> <Exit>
SELECT CASE BoxButton
    CASE 4 ' replace
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace All>";
        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR Magenta
        PRINT "A";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE 5 ' replaceall
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
        PRINT "<Replace All>";
        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR Magenta
        PRINT "A";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE 6 ' cancel
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace All>";
        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR Magenta
        PRINT "A";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR SearchBoxButton1Color, SearchBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE ELSE
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Replace All>";
        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR Magenta
        PRINT "A";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR SearchBoxButton2Color, SearchBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
END SELECT
COLOR White, Black
RETURN

' display current entry area.
DisplayFileLine:
IF BoxButton = 1 THEN
    LOCATE XC1, YC1, 1
    PRINT SPACE$(LC);
END IF
IF BoxButton = 2 THEN
    LOCATE XC2, YC2, 1
    PRINT SPACE$(LC);
END IF

' display edit line and position cursor.
IF BoxButton = 1 THEN
    LOCATE XC1, YC1, 1
    IF LEN(X1$) THEN
        IF LineX1 THEN
            COLOR 0, 7
        ELSE
            COLOR White
        END IF
        PRINT MID$(X1$, Box1Xposition1, LC);
    END IF
    LOCATE XC1, YC1 + Box1Xposition2 - Box1Xposition1 + 1, 1
    IF Insert1 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
    COLOR White, Black
END IF

IF BoxButton = 2 THEN
    LOCATE XC2, YC2, 1
    IF LEN(X2$) THEN
        IF LineX2 THEN
            COLOR 0, 7
        ELSE
            COLOR White
        END IF
        PRINT MID$(X2$, Box2Xposition1, LC);
    END IF
    LOCATE XC2, YC2 + Box2Xposition2 - Box2Xposition1 + 1, 1
    IF Insert2 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
    COLOR White, Black
END IF
RETURN

' reset cursor
ToggleInsert:
IF BoxButton = 1 THEN
    IF Insert1 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
END IF
IF BoxButton = 2 THEN
    IF Insert2 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
END IF
RETURN

END FUNCTION

REM ReplaceBox w/ Replace/Skip/Cancel
FUNCTION ReplaceBox (X, X!)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' get replace box colors
ReplaceBoxBorderColor = Yellow
ReplaceBoxTitleColor = White
ReplaceBoxTextColor = White
ReplaceBoxButton1Color = White
ReplaceBoxButton2Color = Plain
ReplaceBackGround = 1
ReplaceButtonBackGround = 0

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare replace box coordinates.
Xcoor3 = X
Ycoor3 = 10

' draw box
BoxButton = 1
GOSUB DrawReplaceBox

' wait for keypress or mouse
_KEYCLEAR
DO
    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 3
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        GOSUB DrawReplaceBoxButtons
                    CASE CHR$(1) ' ctrl-a
                        ReplaceBackGround = ReplaceBackGround + 1
                        IF ReplaceBackGround = 8 THEN
                            ReplaceBackGround = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(2) ' ctrl-b
                        ReplaceButtonBackGround = ReplaceButtonBackGround + 1
                        IF ReplaceButtonBackGround = 8 THEN
                            ReplaceButtonBackGround = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(4) ' ctrl-d
                        ReplaceBoxBorderColor = ReplaceBoxBorderColor + 1
                        IF ReplaceBoxBorderColor = 16 THEN
                            ReplaceBoxBorderColor = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(5) ' ctrl-e
                        ReplaceBoxTitleColor = ReplaceBoxTitleColor + 1
                        IF ReplaceBoxTitleColor = 16 THEN
                            ReplaceBoxTitleColor = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(6) ' ctrl-f
                        ReplaceBoxTextColor = ReplaceBoxTextColor + 1
                        IF ReplaceBoxTextColor = 16 THEN
                            ReplaceBoxTextColor = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(7) ' ctrl-g
                        ReplaceBoxButton1Color = ReplaceBoxButton1Color + 1
                        IF ReplaceBoxButton1Color = 16 THEN
                            ReplaceBoxButton1Color = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE CHR$(8) ' ctrl-h
                        ReplaceBoxButton2Color = ReplaceBoxButton2Color + 1
                        IF ReplaceBoxButton2Color = 16 THEN
                            ReplaceBoxButton2Color = 0
                        END IF
                        GOSUB DrawReplaceBox
                    CASE ELSE
                        IF UCASE$(X$) = "R" THEN
                            BoxButton = 1
                            EXIT DO
                        END IF
                        IF UCASE$(X$) = "S" THEN
                            BoxButton = 2
                            EXIT DO
                        END IF
                        IF UCASE$(X$) = "C" THEN
                            BoxButton = 3
                            EXIT DO
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 75, 15 ' left/shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 3
                        END IF
                        GOSUB DrawReplaceBoxButtons
                    CASE 77 ' right
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        GOSUB DrawReplaceBoxButtons
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        X$ = ""
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawReplaceBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        X$ = ""
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawReplaceBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawReplaceBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawReplaceBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawReplaceBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 6 THEN
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 9 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 12 AND MouseY <= Ycoor3 + 17 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 20 AND MouseY <= Ycoor3 + 27 THEN
                    BoxButton = 3
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 6 THEN
            '<Replace>  <Skip>  <Cancel>
            IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 9 THEN
                IF BoxButton <> 1 THEN
                    BoxButton = 1
                    GOSUB DrawReplaceBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 12 AND MouseY <= Ycoor3 + 17 THEN
                IF BoxButton <> 2 THEN
                    BoxButton = 2
                    GOSUB DrawReplaceBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 20 AND MouseY <= Ycoor3 + 27 THEN
                IF BoxButton <> 3 THEN
                    BoxButton = 3
                    GOSUB DrawReplaceBoxButtons
                END IF
            END IF
        END IF
    END IF
LOOP
_KEYCLEAR
_DELAY .1

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN ' Replace
    ReplaceBox = -1
ELSE
    IF BoxButton = 2 THEN ' Skip
        ReplaceBox = 1
    ELSE
        ReplaceBox = 0 ' Cancel
    END IF
END IF
EXIT FUNCTION

' draw box
DrawReplaceBox:
COLOR ReplaceBoxBorderColor, ReplaceBackGround
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR ReplaceBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 10, 0
PRINT " Replace ";

' display replace text
COLOR ReplaceBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
Var1$ = "Text in line:" + STR$(X!)
PRINT Var1$;
LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
Var1$ = "Replace text?"
PRINT Var1$;
GOSUB DrawReplaceBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawReplaceBoxButtons:
SELECT CASE BoxButton
    '<Replace>  <Skip>  <Cancel>
    CASE 1
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR ReplaceBoxButton1Color, ReplaceButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Skip>";
        LOCATE Xcoor3 + 6, Ycoor3 + 13, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 21, 0
        COLOR Magenta
        PRINT "C";
    CASE 2
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR ReplaceBoxButton1Color, ReplaceButtonBackGround
        PRINT "<Skip>";
        LOCATE Xcoor3 + 6, Ycoor3 + 13, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 20, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 21, 0
        COLOR Magenta
        PRINT "C";
    CASE 3
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Replace>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "R";

        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR ReplaceBoxButton2Color, ReplaceButtonBackGround
        PRINT "<Skip>";
        LOCATE Xcoor3 + 6, Ycoor3 + 13, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 4, Ycoor3 + 20, 0
        COLOR ReplaceBoxButton1Color, ReplaceButtonBackGround
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 21, 0
        COLOR Magenta
        PRINT "C";
END SELECT
COLOR 7, 0
RETURN
END FUNCTION

REM MemoryMonitorBox
FUNCTION MonitorBox (X1$, X2$)
DIM TempArrayY3(1 TO 2000) AS INTEGER
DIM TempArrayZ3(1 TO 2000) AS INTEGER

' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
GOSUB SaveScreenX

' get search box colors
MonitorBoxBorderColor = Yellow
MonitorBoxTitleColor = White
MonitorBoxTextColor = White
MonitorBoxButton1Color = White
MonitorBoxButton2Color = Plain
MonitorBoxBackGround = 1
MonitorBoxButtonBackGround = 0

' declare input box coordinates.
Xcoor3 = 10
Ycoor3 = 10
XC1 = Xcoor3 + 2
YC1 = Ycoor3 + 1
XC2 = Xcoor3 + 4
YC2 = Ycoor3 + 1
LC = 30

' draw box
BoxButton = 1
DisplaySelect = 0
GOSUB DrawMonitorBox

' init entry area position.
Box1Xposition1 = 1
Box1Xposition2 = 0
Box2Xposition1 = 1
Box2Xposition2 = 0

' display edit entry area.
Insert1 = -1
Insert2 = -1
LineX2 = 0
IF LEN(X2$) THEN
    LineX2 = -1
END IF
BoxButton = 2
GOSUB DisplayFileLine

LineX1 = 0
IF LEN(X1$) THEN
    LineX1 = -1
END IF
BoxButton = 1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE I$
                    CASE CHR$(8) ' backspace
                        IF BoxButton = 1 THEN
                            IF LEN(X1$) THEN
                                IF LineX1 THEN
                                    LineX1 = 0
                                    X1$ = ""
                                    Box1Xposition1 = 1
                                    Box1Xposition2 = 0
                                    GOSUB DisplayFileLine
                                END IF
                                IF Box1Xposition2 > 0 THEN
                                    IF Insert1 THEN ' destructive backspace
                                        X1$ = LEFT$(X1$, Box1Xposition2 - 1) + MID$(X1$, Box1Xposition2 + 1)
                                    END IF
                                    Box1Xposition2 = Box1Xposition2 - 1
                                    IF Box1Xposition2 - LC <= Box1Xposition1 THEN
                                        Box1Xposition1 = Box1Xposition1 - 1
                                        IF Box1Xposition1 < 1 THEN
                                            Box1Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            IF LEN(X2$) THEN
                                IF LineX2 THEN
                                    LineX2 = 0
                                    X2$ = ""
                                    Box2Xposition1 = 1
                                    Box2Xposition2 = 0
                                    GOSUB DisplayFileLine
                                END IF
                                IF Box2Xposition2 > 0 THEN
                                    IF Insert2 THEN ' destructive backspace
                                        X2$ = LEFT$(X2$, Box2Xposition2 - 1) + MID$(X2$, Box2Xposition2 + 1)
                                    END IF
                                    Box2Xposition2 = Box2Xposition2 - 1
                                    IF Box2Xposition2 - LC <= Box2Xposition1 THEN
                                        Box2Xposition1 = Box2Xposition1 - 1
                                        IF Box2Xposition1 < 1 THEN
                                            Box2Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27) ' exit
                        BoxButton = 6
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 7 THEN
                            BoxButton = 1
                        END IF
                        SELECT CASE BoxButton
                            CASE 1, 2 ' entry area
                                GOSUB DrawMonitorBoxButtons
                                GOSUB DisplayFileLine
                            CASE 3 ' checkbox
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' replace
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 1, 1, 8, 8
                            CASE 5 ' replaceall
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                            CASE 6 ' exit
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 25, 1, 8, 8
                        END SELECT
                    CASE CHR$(1) ' ctrl-a
                        MonitorBoxBackGround = MonitorBoxBackGround + 1
                        IF MonitorBoxBackGround = 8 THEN
                            MonitorBoxBackGround = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(2) ' ctrl-b
                        MonitorBoxButtonBackGround = MonitorBoxButtonBackGround + 1
                        IF MonitorBoxButtonBackGround = 8 THEN
                            MonitorBoxButtonBackGround = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(4) ' ctrl-d
                        MonitorBoxBorderColor = MonitorBoxBorderColor + 1
                        IF MonitorBoxBorderColor = 16 THEN
                            MonitorBoxBorderColor = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(5) ' ctrl-e
                        MonitorBoxTitleColor = MonitorBoxTitleColor + 1
                        IF MonitorBoxTitleColor = 16 THEN
                            MonitorBoxTitleColor = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(6) ' ctrl-f
                        MonitorBoxTextColor = MonitorBoxTextColor + 1
                        IF MonitorBoxTextColor = 16 THEN
                            MonitorBoxTextColor = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(7) ' ctrl-g
                        MonitorBoxButton1Color = MonitorBoxButton1Color + 1
                        IF MonitorBoxButton1Color = 16 THEN
                            MonitorBoxButton1Color = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE CHR$(11) ' ctrl-k
                        MonitorBoxButton2Color = MonitorBoxButton2Color + 1
                        IF MonitorBoxButton2Color = 16 THEN
                            MonitorBoxButton2Color = 0
                        END IF
                        GOSUB DrawMonitorBox
                    CASE ELSE
                        IF BoxButton >= 3 AND BoxButton <= 6 THEN
                            IF UCASE$(I$) = "D" THEN
                                DisplaySelect = NOT DisplaySelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                            IF UCASE$(I$) = "S" THEN
                                BoxButton = 4
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "C" THEN
                                BoxButton = 5
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "E" THEN
                                BoxButton = 6
                                EXIT DO
                            END IF
                        END IF
                        IF BoxButton = 3 THEN ' checkbox
                            IF I$ = " " THEN
                                DisplaySelect = NOT DisplaySelect
                                GOSUB DrawSearchCheckBoxs
                            END IF
                        END IF
                        IF BoxButton = 1 THEN
                            ' append/insert character into edit area.
                            VarI = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                VarI = 0 ' disallow control charcters
                            END IF
                            IF VarI THEN
                                IF LineX1 THEN
                                    LineX1 = 0
                                    X1$ = ""
                                    Box1Xposition1 = 1
                                    Box1Xposition2 = 0
                                END IF
                                IF Insert1 THEN
                                    X1$ = LEFT$(X1$, Box1Xposition2) + I$ + MID$(X1$, Box1Xposition2 + 1)
                                    Box1Xposition2 = Box1Xposition2 + 1
                                    IF Box1Xposition2 > LC THEN
                                        Box1Xposition1 = Box1Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Box1Xposition2 + 1 > LEN(X1$) THEN
                                        X1$ = X1$ + I$ ' append
                                        Box1Xposition2 = Box1Xposition2 + 1
                                        IF Box1Xposition2 > LC THEN
                                            Box1Xposition1 = Box1Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X1$, Box1Xposition2 + 1, 1) = I$
                                        Box1Xposition2 = Box1Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            ' append/insert character into edit area.
                            VarI = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                VarI = 0 ' disallow control charcters
                            END IF
                            IF VarI THEN
                                IF LineX2 THEN
                                    LineX2 = 0
                                    X2$ = ""
                                    Box2Xposition1 = 1
                                    Box2Xposition2 = 0
                                END IF
                                IF Insert2 THEN
                                    X2$ = LEFT$(X2$, Box2Xposition2) + I$ + MID$(X2$, Box2Xposition2 + 1)
                                    Box2Xposition2 = Box2Xposition2 + 1
                                    IF Box2Xposition2 > LC THEN
                                        Box2Xposition1 = Box2Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Box2Xposition2 + 1 > LEN(X2$) THEN
                                        X2$ = X2$ + I$ ' append
                                        Box2Xposition2 = Box2Xposition2 + 1
                                        IF Box2Xposition2 > LC THEN
                                            Box2Xposition1 = Box2Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X2$, Box2Xposition2 + 1, 1) = I$
                                        Box2Xposition2 = Box2Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        IF BoxButton = 1 THEN
                            Insert1 = NOT Insert1
                            IF Insert1 THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            Insert2 = NOT Insert2
                            IF Insert2 THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                    CASE 83 ' Delete
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            GOSUB DisplayFileLine
                            IF LEN(X1$) > 0 THEN
                                X1$ = LEFT$(X1$, Box1Xposition2) + MID$(X1$, Box1Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            GOSUB DisplayFileLine
                            IF LEN(X2$) > 0 THEN
                                X2$ = LEFT$(X2$, Box2Xposition2) + MID$(X2$, Box2Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 71 ' Home
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF Box1Xposition2 > 0 THEN
                                Box1Xposition1 = 1
                                Box1Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF Box2Xposition2 > 0 THEN
                                Box2Xposition1 = 1
                                Box2Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 79 ' End
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF Box1Xposition2 < LEN(X1$) THEN
                                Box1Xposition2 = LEN(X1$)
                                Box1Xposition1 = Box1Xposition2 - LC + 1
                                IF Box1Xposition1 < 1 THEN
                                    Box1Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF Box2Xposition2 < LEN(X2$) THEN
                                Box2Xposition2 = LEN(X2$)
                                Box2Xposition1 = Box2Xposition2 - LC + 1
                                IF Box2Xposition1 < 1 THEN
                                    Box2Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 75 ' Left
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF LEN(X1$) THEN
                                IF Box1Xposition2 > 0 THEN
                                    Box1Xposition2 = Box1Xposition2 - 1
                                    IF Box1Xposition2 - LC <= Box1Xposition1 THEN
                                        Box1Xposition1 = Box1Xposition1 - 1
                                        IF Box1Xposition1 < 1 THEN
                                            Box1Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF LEN(X2$) THEN
                                IF Box2Xposition2 > 0 THEN
                                    Box2Xposition2 = Box2Xposition2 - 1
                                    IF Box2Xposition2 - LC <= Box2Xposition1 THEN
                                        Box2Xposition1 = Box2Xposition1 - 1
                                        IF Box2Xposition1 < 1 THEN
                                            Box2Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF BoxButton = 1 THEN
                            LineX1 = 0
                            IF LEN(X1$) THEN
                                IF Box1Xposition2 < LEN(X1$) THEN
                                    Box1Xposition2 = Box1Xposition2 + 1
                                    IF Box1Xposition2 > Box1Xposition1 + LC - 1 THEN
                                        Box1Xposition1 = Box1Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                        IF BoxButton = 2 THEN
                            LineX2 = 0
                            IF LEN(X2$) THEN
                                IF Box2Xposition2 < LEN(X2$) THEN
                                    Box2Xposition2 = Box2Xposition2 + 1
                                    IF Box2Xposition2 > Box2Xposition1 + LC - 1 THEN
                                        Box2Xposition1 = Box2Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 72 ' up
                        IF BoxButton = 3 THEN ' checkbox
                            DisplaySelect = -1
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 80 ' down
                        IF BoxButton = 3 THEN ' checkbox
                            DisplaySelect = 0
                            GOSUB DrawSearchCheckBoxs
                        END IF
                    CASE 15 ' shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 6
                        END IF
                        SELECT CASE BoxButton
                            CASE 1, 2 ' entry area
                                GOSUB DrawMonitorBoxButtons
                                GOSUB DisplayFileLine
                            CASE 3 ' checkbox
                                LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
                            CASE 4 ' replace
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 1, 1, 8, 8
                            CASE 5 ' replace all
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 11, 1, 8, 8
                            CASE 6 ' exit
                                GOSUB DrawMonitorBoxButtons
                                LOCATE Xcoor3 + 6, Ycoor3 + 25, 1, 8, 8
                        END SELECT
                    CASE 141 ' ctrl-up
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                GOSUB RestoreScreenX
                                GOSUB DrawMonitorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                GOSUB RestoreScreenX
                                GOSUB DrawMonitorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                GOSUB RestoreScreenX
                                GOSUB DrawMonitorBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                GOSUB RestoreScreenX
                                GOSUB DrawMonitorBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            GOSUB RestoreScreenX
                            GOSUB DrawMonitorBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE ' click on button
            IF MouseX = XC1 THEN
                ' entry area 1
                IF MouseY >= YC1 AND MouseY <= YC1 + LC THEN
                    BoxButton = 1
                    GOSUB DrawMonitorBox
                END IF
            END IF
            IF MouseX = XC2 THEN
                ' entry area 2
                IF MouseY >= YC2 AND MouseY <= YC2 + LC THEN
                    BoxButton = 2
                    GOSUB DrawMonitorBox
                END IF
            END IF
            IF MouseX = Xcoor3 + 5 THEN
                ' click checkbox
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 3 THEN
                    DisplaySelect = NOT DisplaySelect
                    GOSUB DrawSearchCheckBoxs
                END IF
            END IF
            IF MouseX = Xcoor3 + 6 THEN
                ' click <system>
                IF MouseY >= Ycoor3 + 1 AND MouseY <= Ycoor3 + 8 THEN
                    BoxButton = 4
                    EXIT DO
                END IF
                ' click <continue>
                IF MouseY >= Ycoor3 + 11 AND MouseY <= Ycoor3 + 20 THEN
                    BoxButton = 5
                    EXIT DO
                END IF
                ' click <exit>
                IF MouseY >= Ycoor3 + 25 AND MouseY <= Ycoor3 + 30 THEN
                    BoxButton = 6
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' restore screen area.
GOSUB RestoreScreenX
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
SELECT CASE BoxButton
    CASE 1 ' timer
        MonitorBox = 1
    CASE 2 ' limit
        MonitorBox = 1
    CASE 3 ' don't display
        MonitorBox = 1
    CASE 4 ' restart
        MonitorBox = 2
    CASE 5 ' continue
        MonitorBox = 1
    CASE 6 ' exit
        MonitorBox = 0
END SELECT
LOCATE , , 1, 8, 8
_KEYCLEAR
Monitor.Box = -1
EXIT FUNCTION

' draw box
DrawMonitorBox:
COLOR MonitorBoxBorderColor, MonitorBoxBackGround
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
MonitorBoxTitle$ = " MEMORY LOW WARNING "
TempX = 16 - LEN(MonitorBoxTitle$) / 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR MonitorBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT MonitorBoxTitle$;

' display search text
COLOR MonitorBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
PRINT "Monitor timer(seconds):"
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
PRINT "Monitor limit(MB.):"

LOCATE Xcoor3 + 5, Ycoor3 + 1, 0
COLOR MonitorBoxButton1Color, MonitorBoxBackGround
PRINT "[ ] Don't display again";
LOCATE Xcoor3 + 5, Ycoor3 + 5, 0
COLOR Magenta
PRINT "D";

GOSUB DrawSearchCheckBoxs
GOSUB DrawMonitorBoxButtons
LOCATE , , , 8, 8
XC1 = Xcoor3 + 2
YC1 = Ycoor3 + 1
XC2 = Xcoor3 + 4
YC2 = Ycoor3 + 1

COLOR White, Black
LOCATE XC1, YC1, 1
PRINT SPACE$(LC);
LOCATE XC2, YC2, 1
PRINT SPACE$(LC);
IF LEN(X1$) THEN
    LOCATE XC1, YC1, 1
    COLOR White
    PRINT MID$(X1$, Box1Xposition1, LC);
END IF
IF LEN(X2$) THEN
    LOCATE XC2, YC2, 1
    COLOR White
    PRINT MID$(X2$, Box2Xposition1, LC);
END IF

SELECT CASE BoxButton
    CASE 1, 2 ' entry area
        GOSUB DisplayFileLine
    CASE 3 ' checkbox
        LOCATE Xcoor3 + 5, Ycoor3 + 2, 1
    CASE 4 ' replace
        GOSUB DrawMonitorBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 1
    CASE 5 ' replaceall
        GOSUB DrawMonitorBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 11, 1
    CASE 6 ' cancel
        GOSUB DrawMonitorBoxButtons
        LOCATE Xcoor3 + 6, Ycoor3 + 25, 1
END SELECT
RETURN

' display checkboxs
DrawSearchCheckBoxs:
'-[ ] Don't display again

IF DisplaySelect THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR MonitorBoxButton1Color, MonitorBoxButtonBackGround
    PRINT "X";
ELSE
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR MonitorBoxButton1Color, MonitorBoxButtonBackGround
    PRINT " ";
END IF
IF BoxButton = 3 THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 1, 8, 8
END IF
RETURN

' display buttuns
DrawMonitorBoxButtons:
'-(System>  <Continue>    <Exit>
SELECT CASE BoxButton
    CASE 4 ' replace
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR MonitorBoxButton1Color, MonitorBoxButtonBackGround
        PRINT "<System>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Continue>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE 5 ' replaceall
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<System>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR MonitorBoxButton1Color, MonitorBoxButtonBackGround
        PRINT "<Continue>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE 6 ' cancel
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<System>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Continue>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR MonitorBoxButton1Color, MonitorBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
    CASE ELSE
        LOCATE Xcoor3 + 6, Ycoor3 + 1, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<System>";
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR Magenta
        PRINT "S";

        LOCATE Xcoor3 + 6, Ycoor3 + 11, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Continue>";
        LOCATE Xcoor3 + 6, Ycoor3 + 12, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 25, 0
        COLOR MonitorBoxButton2Color, MonitorBoxButtonBackGround
        PRINT "<Exit>";
        LOCATE Xcoor3 + 6, Ycoor3 + 26, 0
        COLOR Magenta
        PRINT "E";
END SELECT
COLOR White, Black
RETURN

' display current entry area.
DisplayFileLine:
IF BoxButton = 1 THEN
    LOCATE XC1, YC1, 1
    PRINT SPACE$(LC);
END IF
IF BoxButton = 2 THEN
    LOCATE XC2, YC2, 1
    PRINT SPACE$(LC);
END IF

' display edit line and position cursor.
IF BoxButton = 1 THEN
    LOCATE XC1, YC1, 1
    IF LEN(X1$) THEN
        IF LineX1 THEN
            COLOR 0, 7
        ELSE
            COLOR White
        END IF
        PRINT MID$(X1$, Box1Xposition1, LC);
    END IF
    LOCATE XC1, YC1 + Box1Xposition2 - Box1Xposition1 + 1, 1
    IF Insert1 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
    COLOR White, Black
END IF

IF BoxButton = 2 THEN
    LOCATE XC2, YC2, 1
    IF LEN(X2$) THEN
        IF LineX2 THEN
            COLOR 0, 7
        ELSE
            COLOR White
        END IF
        PRINT MID$(X2$, Box2Xposition1, LC);
    END IF
    LOCATE XC2, YC2 + Box2Xposition2 - Box2Xposition1 + 1, 1
    IF Insert2 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
    COLOR White, Black
END IF
RETURN

' reset cursor
ToggleInsert:
IF BoxButton = 1 THEN
    IF Insert1 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
END IF
IF BoxButton = 2 THEN
    IF Insert2 THEN
        LOCATE , , 1, 8, 8 ' line cursor
    ELSE
        LOCATE , , 1, 0, 8 ' block cursor
    END IF
END IF
RETURN

' screen save
SaveScreenX:
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        TempZ1 = SCREEN(Var1, Var2) ' screen char
        TempZ2 = SCREEN(Var1, Var2, 1) ' char color
        TempArrayY3((Var1 - 1) * 80 + Var2) = TempZ1
        TempArrayZ3((Var1 - 1) * 80 + Var2) = TempZ2
    NEXT
NEXT
RETURN

' screen restore
RestoreScreenX:
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        VarB = INT(TempArrayZ3((Var1 - 1) * 80 + Var2) / 16)
        VarF = TempArrayZ3((Var1 - 1) * 80 + Var2) MOD 16
        TempZ1 = TempArrayY3((Var1 - 1) * 80 + Var2)
        LOCATE Var1, Var2, 1
        COLOR VarF, VarB
        _CONTROLCHR OFF
        PRINT CHR$(TempZ1);
        _CONTROLCHR ON
    NEXT
NEXT
RETURN
END FUNCTION

REM BreakBox with OK/Quit
FUNCTION BreakBox
DIM TempArrayY3(1 TO 2000) AS INTEGER
DIM TempArrayZ3(1 TO 2000) AS INTEGER

' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
GOSUB SaveScreenX

' declare break box coordinates.
Xcoor3 = 10
Ycoor3 = 10

' set break box colors
BreakBoxBorderColor = Yellow
BreakBoxTitleColor = White
BreakBoxTextColor = White
BreakBoxButton1Color = White
BreakBoxButton2Color = Plain
BreakBackGround1 = 1
BreakBackGround2 = 0

' draw box
BoxButton = 1
GOSUB DrawBreakBox

' wait for keypress or mouse
_KEYCLEAR
DO
    VarQ = _EXIT
    IF VarQ THEN ' Control-Break
        SYSTEM
    END IF

    _LIMIT 30
    X$ = INKEYx$
    IF LEN(X$) THEN
        SELECT CASE LEN(X$)
            CASE 1
                SELECT CASE UCASE$(X$)
                    CASE "O"
                        BoxButton = 1
                        EXIT DO
                    CASE "Q"
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 1
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawBreakBoxButtons
                    CASE CHR$(1) ' ctrl-a
                        BreakBackGround1 = BreakBackGround1 + 1
                        IF BreakBackGround1 = 8 THEN
                            BreakBackGround1 = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(2) ' ctrl-b
                        BreakBackGround2 = BreakBackGround2 + 1
                        IF BreakBackGround2 = 8 THEN
                            BreakBackGround2 = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(4) ' ctrl-d
                        BreakBoxBorderColor = BreakBoxBorderColor + 1
                        IF BreakBoxBorderColor = 16 THEN
                            BreakBoxBorderColor = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(5) ' ctrl-e
                        BreakBoxTitleColor = BreakBoxTitleColor + 1
                        IF BreakBoxTitleColor = 16 THEN
                            BreakBoxTitleColor = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(6) ' ctrl-f
                        BreakBoxTextColor = BreakBoxTextColor + 1
                        IF BreakBoxTextColor = 16 THEN
                            BreakBoxTextColor = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(7) ' ctrl-g
                        BreakBoxButton1Color = BreakBoxButton1Color + 1
                        IF BreakBoxButton1Color = 16 THEN
                            BreakBoxButton1Color = 0
                        END IF
                        GOSUB DrawBreakBox
                    CASE CHR$(8) ' ctrl-h
                        BreakBoxButton2Color = BreakBoxButton2Color + 1
                        IF BreakBoxButton2Color = 16 THEN
                            BreakBoxButton2Color = 0
                        END IF
                        GOSUB DrawBreakBox
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(X$, 1))
                    CASE 75, 15 ' left/shift-tab
                        IF BoxButton = 2 THEN
                            BoxButton = 1
                        ELSE
                            BoxButton = 2
                        END IF
                        GOSUB DrawBreakBoxButtons
                    CASE 77 ' right
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawBreakBoxButtons
                    CASE 141 ' ctrl-up
                        X$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        X$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        X$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        X$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                GOSUB RestoreScreenX
                                GOSUB DrawBreakBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        X$ = ""
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                GOSUB RestoreScreenX
                                GOSUB DrawBreakBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        X$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                GOSUB RestoreScreenX
                                GOSUB DrawBreakBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        X$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                GOSUB RestoreScreenX
                                GOSUB DrawBreakBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            GOSUB RestoreScreenX
                            GOSUB DrawBreakBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 5 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 13 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 5 THEN
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                IF BoxButton = 2 THEN
                    BoxButton = 1
                    GOSUB DrawBreakBoxButtons
                END IF
            END IF
            IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                IF BoxButton = 1 THEN
                    BoxButton = 2
                    GOSUB DrawBreakBoxButtons
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' restore screen area.
GOSUB RestoreScreenX
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN
    BreakBox = -1
ELSE
    BreakBox = 0
END IF
EXIT FUNCTION

' draw box
DrawBreakBox:
COLOR BreakBoxBorderColor, BreakBackGround1
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);

' display box title
Var0$ = " Break "
TempX = 16 - LEN(Var0$) \ 2 ' center of titlebar
TempX = Ycoor3 + TempX
IF TempX < 1 THEN TempX = 1
COLOR BreakBoxTitleColor
LOCATE Xcoor3, TempX, 0
PRINT Var0$;

' display error text
COLOR BreakBoxTextColor

Var1$ = "Control-Break trap."
LOCATE Xcoor3 + 1, Ycoor3 + 2, 0
PRINT Var1$

Var2$ = "Press <OK> to continue."
LOCATE Xcoor3 + 2, Ycoor3 + 2, 0
PRINT Var2$

Var3$ = "Press <Quit> to end program."
LOCATE Xcoor3 + 3, Ycoor3 + 2, 0
PRINT Var3$

GOSUB DrawBreakBoxButtons
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawBreakBoxButtons:
IF BoxButton = 1 THEN
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR BreakBoxButton1Color, BreakBackGround2
    PRINT "<OK>";
    LOCATE Xcoor3 + 5, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 5, Ycoor3 + 8, 0
    COLOR BreakBoxButton2Color, BreakBackGround2
    PRINT "<Quit>";
    LOCATE Xcoor3 + 5, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "Q";
ELSE
    LOCATE Xcoor3 + 5, Ycoor3 + 2, 0
    COLOR BreakBoxButton2Color, BreakBackGround2
    PRINT "<OK>";
    LOCATE Xcoor3 + 5, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 5, Ycoor3 + 8, 0
    COLOR BreakBoxButton1Color, BreakBackGround2
    PRINT "<Quit>";
    LOCATE Xcoor3 + 5, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "Q";
END IF
COLOR 7, 0
RETURN

' screen save
SaveScreenX:
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        TempZ1 = SCREEN(Var1, Var2) ' screen char
        TempZ2 = SCREEN(Var1, Var2, 1) ' char color
        TempArrayY3((Var1 - 1) * 80 + Var2) = TempZ1
        TempArrayZ3((Var1 - 1) * 80 + Var2) = TempZ2
    NEXT
NEXT
RETURN

' screen restore
RestoreScreenX:
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        VarB = INT(TempArrayZ3((Var1 - 1) * 80 + Var2) / 16)
        VarF = TempArrayZ3((Var1 - 1) * 80 + Var2) MOD 16
        TempZ1 = TempArrayY3((Var1 - 1) * 80 + Var2)
        LOCATE Var1, Var2, 1
        COLOR VarF, VarB
        _CONTROLCHR OFF
        PRINT CHR$(TempZ1);
        _CONTROLCHR ON
    NEXT
NEXT
RETURN
END FUNCTION

' formats a double numeric string
FUNCTION FormatString$ (s#)
x$ = ""
s$ = STR$(s#)
IF INSTR(s$, "D") THEN ' return string
    FormatString$ = s$
    EXIT FUNCTION
END IF
IF LEFT$(s$, 1) = "-" THEN ' store sign
    e$ = "-"
    s$ = MID$(s$, 2)
END IF
s$ = LTRIM$(s$) ' format string
IF INSTR(s$, ".") THEN
    q$ = MID$(s$, INSTR(s$, "."))
    s$ = LEFT$(s$, INSTR(s$, ".") - 1)
END IF
FOR l = LEN(s$) TO 3 STEP -3
    x$ = MID$(s$, l - 2, 3) + "," + x$
NEXT
IF l > 0 THEN
    x$ = MID$(s$, 1, l) + "," + x$
END IF
IF LEN(s$) < 3 THEN
    x$ = s$
END IF
IF RIGHT$(x$, 1) = "," THEN
    x$ = LEFT$(x$, LEN(x$) - 1)
END IF
x$ = e$ + x$ + q$ ' construct string
FormatString$ = x$
END FUNCTION

SUB ParseTest
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' display parsed command line
        PRINT Command.Work

        ' check search filename
        IF Command.Line = "" THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP
END SUB

' uses direct keyboard input to get an input line
'   and edit multi-dimensional command history arrays with line-wrapping,
'     where Temp0 (history array type) equals:
'       1=command prompt
'       2=whatis prompt
'       3=debug prompt

FUNCTION KeyboardLine$ (VarP)
Insert = -1
HistoryCount2(Temp0) = 0 ' reset
NetPathHistory = 0

' position of prompt
L = CSRLIN
P = POS(0)
LOCATE L, P, 1

' position in entry string
X = 1
Var$ = ""
MaxLength = 128
PRINT "";
LOCATE , , 1, 8, 8
CurrentMenu = 0
Control.Break = False
Monitor.Box = False
DO
    ' check alt key press
    Q& = _KEYHIT
    IF Q& = KEY_LALT& THEN
        CurrentMenu = 1
        CALL Menu2
        GOSUB DisplayLine2
        Var$ = ""
        EXIT DO
    END IF
    IF Q& = KEY_RALT& THEN
        CurrentMenu = 1
        CALL Menu2
        GOSUB DisplayLine2
        Var$ = ""
        EXIT DO
    END IF

    ' check mouse activity
    Z = MouseDriver
    IF MouseButton1 THEN
        ' select drop down menu.
        IF MouseX = 1 THEN
            NewMenuSelection = 0
            SELECT CASE MouseY
                CASE 1 TO 3 ' cpu
                    CPUOFF = NOT CPUOFF
                    IF CPUOFF THEN
                        TIMER OFF
                        _TITLE Program
                        CALL DisplayToolBar
                    ELSE
                        TIMER ON
                        CALL DisplayToolBar
                    END IF
                CASE 74 TO 78 ' about
                    CALL HelpAbout
                CASE 6 TO 9 ' File
                    NewMenuSelection = 1
                CASE 16 TO 21 ' Charts
                    NewMenuSelection = 2
                CASE 26 TO 29 ' Dump
                    NewMenuSelection = 3
                CASE 36 TO 39 ' Edit
                    NewMenuSelection = 4
                CASE 46 TO 49 ' Jump
                    NewMenuSelection = 5
                CASE 56 TO 60 ' Print
                    NewMenuSelection = 6
                CASE 64 TO 69 ' Search
                    NewMenuSelection = 7
            END SELECT
            IF NewMenuSelection THEN
                CurrentMenu = NewMenuSelection
                NewMenuSelection = 0
                CurrentMenuSelection = 1
                CALL DropDownMenu(Var)
                GOSUB DisplayLine2
                Var$ = ""
                EXIT DO
            END IF
        END IF
    END IF

    ' get key input
    _LIMIT 50
    LOCATE , , 1
    Var2$ = INKEYx$
    IF Control.Break THEN
        GOSUB DisplayLine2
        Var$ = ""
        EXIT DO
    END IF
    IF LEN(Var2$) THEN
        SELECT CASE LEN(Var2$)
            CASE 1
                SELECT CASE Var2$
                    CASE CHR$(3) ' Control-C
                        GOSUB DisplayLine2
                        Var$ = ""
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF VarP = 1 THEN
                            IF NetPathHistory = 0 THEN
                                IF NetPathCount >= 1 THEN
                                    NetPathHistory = 1
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(NetPathTable(1))
                                    GOSUB DisplayLine
                                END IF
                            ELSE
                                IF NetPathHistory < NetPathCount THEN
                                    NetPathHistory = NetPathHistory + 1
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(NetPathTable(NetPathHistory))
                                    GOSUB DisplayLine
                                END IF
                            END IF
                        ELSE
                            IF X + 8 <= LEN(Var$) THEN
                                X = X + 8
                            ELSE
                                X = LEN(Var$) + 1
                            END IF
                            GOSUB DisplayLine2
                        END IF
                    CASE CHR$(13) ' return
                        GOSUB EnterHistory
                        EXIT DO
                    CASE CHR$(8) ' backspace
                        IF X > 1 THEN
                            IF Insert THEN
                                Var$ = LEFT$(Var$, X - 2) + MID$(Var$, X)
                                X = X - 1
                                GOSUB DisplayLine
                            ELSE
                                X = X - 1
                                GOSUB DisplayLine2
                            END IF
                        END IF
                    CASE CHR$(27) ' escape
                        NetPathHistory = 0
                        GOSUB EraseLine
                    CASE ELSE ' store input
                        IF Insert THEN
                            ' always insert char
                            Var$ = LEFT$(Var$, X - 1) + Var2$ + MID$(Var$, X)
                            ' truncate end of line
                            Var$ = LEFT$(Var$, MaxLength)
                            IF X <= MaxLength THEN
                                X = X + 1
                            END IF
                            GOSUB DisplayLine
                        ELSE
                            IF X <= MaxLength THEN
                                Var$ = LEFT$(Var$, X - 1) + Var2$ + MID$(Var$, X + 1)
                                X = X + 1
                                GOSUB DisplayLine2
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(Var2$, 1))
                    CASE 141 ' Ctrl-Up = insert ascii sequence
                        Key1 = 0
                        Key1$ = ""
                        DO
                            _LIMIT 50
                            V$ = INKEYx$
                            IF LEN(V$) THEN
                                IF V$ >= "0" AND V$ <= "9" THEN
                                    Key1 = Key1 + 1
                                    Key1$ = Key1$ + V$
                                    IF Key1 = 3 THEN
                                        EXIT DO
                                    END IF
                                END IF
                            END IF
                        LOOP
                        Var1 = VAL(Key1$)
                        IF Var1 >= 0 AND Var1 <= 255 THEN
                            Var2$ = CHR$(Var1)
                            IF Insert THEN
                                ' always insert char
                                Var$ = LEFT$(Var$, X - 1) + Var2$ + MID$(Var$, X)
                                ' truncate end of line
                                Var$ = LEFT$(Var$, MaxLength)
                                IF X <= MaxLength THEN
                                    X = X + 1
                                END IF
                                GOSUB DisplayLine
                            ELSE
                                IF X <= MaxLength THEN
                                    Var$ = LEFT$(Var$, X - 1) + Var2$ + MID$(Var$, X + 1)
                                    X = X + 1
                                    GOSUB DisplayLine2
                                END IF
                            END IF
                        END IF
                    CASE 71 ' Home
                        X = 1
                        GOSUB DisplayLine2
                    CASE 79 ' End
                        X = LEN(Var$) + 1
                        GOSUB DisplayLine2
                    CASE 115 ' Ctrl-Left
                        DO
                            IF X = 1 THEN
                                EXIT DO
                            END IF
                            X = X - 1
                            IF MID$(Var$, X - 1, 1) = " " THEN
                                IF MID$(Var$, X, 1) <> " " THEN
                                    EXIT DO
                                END IF
                            END IF
                        LOOP
                        GOSUB DisplayLine2
                    CASE 116 ' Ctrl-Right
                        DO
                            IF X > LEN(Var$) THEN
                                EXIT DO
                            END IF
                            IF MID$(Var$, X, 1) = " " THEN
                                IF MID$(Var$, X + 1, 1) <> " " THEN
                                    X = X + 1
                                    EXIT DO
                                END IF
                            END IF
                            X = X + 1
                        LOOP
                        GOSUB DisplayLine2
                    CASE 15 ' Shift-Tab
                        IF VarP = 1 THEN
                            IF NetPathHistory = 0 THEN
                                IF NetPathCount >= 1 THEN
                                    NetPathHistory = NetPathCount
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(NetPathTable(NetPathHistory))
                                    GOSUB DisplayLine
                                END IF
                            ELSE
                                IF NetPathHistory > 1 THEN
                                    NetPathHistory = NetPathHistory - 1
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(NetPathTable(NetPathHistory))
                                    GOSUB DisplayLine
                                END IF
                            END IF
                        ELSE
                            IF X - 8 >= 1 THEN
                                X = X - 8
                            ELSE
                                X = 1
                            END IF
                            GOSUB DisplayLine2
                        END IF
                    CASE 75 ' left
                        IF X > 1 THEN
                            X = X - 1
                            GOSUB DisplayLine2
                        END IF
                    CASE 77 ' right
                        IF X <= LEN(Var$) THEN
                            X = X + 1
                            GOSUB DisplayLine2
                        END IF
                    CASE 82 ' Insert
                        Insert = NOT Insert
                        IF Insert THEN
                            LOCATE , , 1, 8, 8 ' line cursor
                        ELSE
                            LOCATE , , 1, 0, 8 ' blobk cursor
                        END IF
                    CASE 83 ' Delete
                        IF X > 0 THEN
                            Var$ = LEFT$(Var$, X - 1) + MID$(Var$, X + 1)
                            GOSUB DisplayLine
                        END IF
                    CASE 72 ' up - back history
                        GOSUB GetCount
                        IF Var1 > 0 THEN
                            SELECT CASE HistoryCount2(Temp0)
                                CASE 0
                                    GOSUB EraseLine
                                    HistoryCount2(Temp0) = Var1
                                    Var$ = RTRIM$(History2(Temp0, Var1))
                                    X = LEN(Var$) + 1
                                    GOSUB DisplayLine
                                CASE 1 TO MaxHistory - 1
                                    IF HistoryCount2(Temp0) + 1 <= Var1 THEN
                                        GOSUB EraseLine
                                        HistoryCount2(Temp0) = HistoryCount2(Temp0) + 1
                                        Var$ = RTRIM$(History2(Temp0, HistoryCount2(Temp0)))
                                        X = LEN(Var$) + 1
                                        GOSUB DisplayLine
                                    ELSE
                                        IF HistoryCount2(Temp0) = Var1 THEN
                                            GOSUB EraseLine
                                            Var$ = RTRIM$(History2(Temp0, Var1))
                                            X = LEN(Var$) + 1
                                            GOSUB DisplayLine
                                        END IF
                                    END IF
                                CASE MaxHistory
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(History2(Temp0, MaxHistory))
                                    X = LEN(Var$) + 1
                                    GOSUB DisplayLine
                            END SELECT
                        END IF
                    CASE 80 ' down - forward history
                        GOSUB GetCount
                        IF Var1 > 0 THEN
                            SELECT CASE HistoryCount2(Temp0)
                                CASE 0
                                    GOSUB EraseLine
                                    HistoryCount2(Temp0) = Var1
                                    Var$ = RTRIM$(History2(Temp0, Var1))
                                    X = LEN(Var$) + 1
                                    GOSUB DisplayLine
                                CASE 1
                                    GOSUB EraseLine
                                    Var$ = RTRIM$(History2(Temp0, 1))
                                    X = LEN(Var$) + 1
                                    GOSUB DisplayLine
                                CASE ELSE
                                    IF HistoryCount2(Temp0) - 1 > 0 THEN
                                        GOSUB EraseLine
                                        HistoryCount2(Temp0) = HistoryCount2(Temp0) - 1
                                        Var$ = RTRIM$(History2(Temp0, HistoryCount2(Temp0)))
                                        X = LEN(Var$) + 1
                                        GOSUB DisplayLine
                                    END IF
                            END SELECT
                        END IF
                    CASE 119 ' ctrl-home
                        GOSUB GetCount
                        IF Var1 > 0 THEN
                            GOSUB EraseLine
                            HistoryCount2(Temp0) = 1
                            Var$ = RTRIM$(History2(Temp0, 1))
                            X = LEN(Var$) + 1
                            GOSUB DisplayLine
                        END IF
                    CASE 117 ' ctrl-end
                        GOSUB GetCount
                        IF Var1 > 0 THEN
                            GOSUB EraseLine
                            HistoryCount2(Temp0) = Var1
                            Var$ = RTRIM$(History2(Temp0, Var1))
                            X = LEN(Var$) + 1
                            GOSUB DisplayLine
                        END IF
                END SELECT
        END SELECT
    END IF
LOOP

' return input line
LOCATE , , 1, 8, 8
KeyboardLine$ = Var$
EXIT FUNCTION

EnterHistory:
IF LEN(Var$) THEN
    GOSUB GetCount
    ' init
    IF Var1 = 0 THEN
        HistoryCount2(Temp0) = 1
        History2(Temp0, 1) = Var$
    ELSE
        ' add history
        IF HistoryCount2(Temp0) = 0 THEN
            IF Var1 < MaxHistory THEN
                HistoryCount2(Temp0) = Var1 + 1
                History2(Temp0, Var1 + 1) = Var$
            ELSE
                FOR Var = 1 TO MaxHistory - 1
                    History2(Temp0, Var) = History2(Temp0, Var + 1)
                NEXT
                HistoryCount2(Temp0) = MaxHistory
                History2(Temp0, MaxHistory) = Var$
            END IF
        ELSE
            ' pack history
            IF Var1 = MaxHistory THEN
                FOR Var = HistoryCount2(Temp0) TO MaxHistory - 1
                    History2(Temp0, Var) = History2(Temp0, Var + 1)
                NEXT
                HistoryCount2(Temp0) = MaxHistory
                History2(Temp0, MaxHistory) = Var$
            ELSE
                FOR Var = HistoryCount2(Temp0) TO Var1 - 1
                    History2(Temp0, Var) = History2(Temp0, Var + 1)
                NEXT
                History2(Temp0, Var1) = Var$
            END IF
        END IF
    END IF
END IF
RETURN

EraseLine:
LOCATE L, P, 1
FOR V = 1 TO LEN(Var$)
    PRINT " ";
NEXT
LOCATE L, P, 1
Var$ = ""
X = 1
RETURN

' display current entry area.
DisplayLine:

' display edit line.
LOCATE L, P, 1
Var2$ = Var$ + " "
FOR V = 1 TO LEN(Var2$)
    _CONTROLCHR OFF
    PRINT MID$(Var2$, V, 1);
    _CONTROLCHR ON
    ' bump prompt up if at corner
    IF Status THEN
        IF CSRLIN = 23 AND POS(0) = 80 THEN L = L - 1: PRINT
    ELSE
        IF CSRLIN = 24 AND POS(0) = 80 THEN L = L - 1: PRINT
    END IF
NEXT

DisplayLine2:

' position cursor.
LOCATE L, P, 1
FOR V = 1 TO X - 1
    _CONTROLCHR OFF
    PRINT MID$(Var$, V, 1);
    _CONTROLCHR ON
NEXT
COLOR 7, 0
RETURN

GetCount:
Var1 = 0
FOR Var2 = MaxHistory TO 1 STEP -1
    IF RTRIM$(History2(Temp0, Var2)) <> "" THEN
        Var1 = Var2
        EXIT FOR
    END IF
NEXT
RETURN
END FUNCTION

FUNCTION ValidDate (Var$)
' mm-dd-yyyy
Var$ = RTRIM$(Var$)
IF LEN(Var$) <> 10 THEN
    ValidDate = False
    EXIT FUNCTION
END IF
FOR Var = 1 TO 10
    V$ = MID$(Var$, Var, 1)
    SELECT CASE Var
        CASE 1, 2, 4, 5, 7, 8, 9, 10
            IF V$ >= "0" AND V$ <= "9" THEN
                Eat$ = ""
            ELSE
                ValidDate = False
                EXIT FUNCTION
            END IF
        CASE ELSE
            IF V$ <> "-" THEN
                ValidDate = False
                EXIT FUNCTION
            END IF
    END SELECT
NEXT
M = INT(VAL(MID$(Var$, 1, 2)))
D = INT(VAL(MID$(Var$, 4, 2)))
Y = INT(VAL(MID$(Var$, 7, 4)))
IF M >= 1 AND M <= 12 THEN
    IF D >= 1 AND D <= 31 THEN
        IF Y >= 1980 AND Y <= 2079 THEN
            L = 0
            IF Y / 4 = Y \ 4 THEN
                L = -1
            END IF
            IF Y / 100 = Y \ 100 THEN
                L = 0
            END IF
            IF Y / 400 = Y \ 400 THEN
                L = -1
            END IF
            SELECT CASE M
                CASE 1, 3, 5, 7, 8, 10, 12
                    IF D <= 31 THEN
                        ValidDate = -1
                        EXIT FUNCTION
                    END IF
                CASE 4, 6, 9, 11
                    IF D <= 30 THEN
                        ValidDate = -1
                        EXIT FUNCTION
                    END IF
                CASE 2
                    IF L THEN
                        IF D <= 29 THEN
                            ValidDate = -1
                            EXIT FUNCTION
                        END IF
                    END IF
                    IF D <= 28 THEN
                        ValidDate = -1
                        EXIT FUNCTION
                    END IF
            END SELECT
        END IF
    END IF
END IF
ValidDate = 0
END FUNCTION

FUNCTION ValidTime (Var$)
' hh:mm:ss
Var$ = RTRIM$(Var$)
IF LEN(Var$) <> 8 THEN
    ValidTime = False
    EXIT FUNCTION
END IF
FOR Var = 1 TO 8
    V$ = MID$(Var$, Var, 1)
    SELECT CASE Var
        CASE 1, 2, 4, 5, 7, 8
            IF V$ >= "0" AND V$ <= "9" THEN
                Eat$ = ""
            ELSE
                ValidTime = False
                EXIT FUNCTION
            END IF
        CASE ELSE
            IF V$ <> ":" THEN
                ValidTime = False
                EXIT FUNCTION
            END IF
    END SELECT
NEXT
H = INT(VAL(MID$(Var$, 1, 2)))
M = INT(VAL(MID$(Var$, 4, 2)))
S = INT(VAL(MID$(Var$, 7, 2)))
IF H >= 0 AND H <= 23 THEN
    IF M >= 0 AND M <= 59 THEN
        IF S >= 0 AND S <= 59 THEN
            ValidTime = -1
            EXIT FUNCTION
        END IF
    END IF
END IF
ValidTime = 0
END FUNCTION

' Def FNclock$=Format$(Now,"ddd dd mmm yyyy\, hh:mma/p")
FUNCTION clock$
' mm-dd-yyyy hh:mm:ss

M = VAL(LEFT$(DATE$, 2))
D = VAL(MID$(DATE$, 4, 2))
Y = VAL(RIGHT$(DATE$, 4))

IF M < 3 THEN M = M + 12: Y = Y - 1
W = ((13 * M + 3) \ 5 + D + Y + Y \ 4 - Y \ 100 + Y \ 400 + 1) MOD 7
clock$ = MID$("SunMonTueWedThuFriSat", W * 3 + 1, 3)
clock$ = clock$ + " " + MID$(DATE$, 4, 2) ' dd

T = (VAL(MID$(DATE$, 1, 2)) - 1) * 3 + 1
clock$ = clock$ + " " + MID$("JanFebMarAprMayJunJulAugSepOctNovDec", T, 3)

clock$ = clock$ + " " + MID$(DATE$, 7, 4) ' yyyy

T = VAL(LEFT$(TIME$, 2))
IF T > 11 THEN T = T - 12
IF T = 0 THEN T = 12
clock$ = clock$ + ", " + RIGHT$("00" + LTRIM$(STR$(T)), 2) ' hh
clock$ = clock$ + ":" + MID$(TIME$, 4, 2) ' mm

T = VAL(LEFT$(TIME$, 2))
IF T < 12 THEN
    clock$ = clock$ + "a"
ELSE
    clock$ = clock$ + "p"
END IF
END FUNCTION

SUB ViewFileInfo (Var$)
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _UNSIGNED _OFFSET

IF INSTR(Var$, "?") THEN
    Var1$ = " Error "
    Var2$ = "Invalid filename."
    VarX = MessageBox(Var1$, Var2$)
    EXIT SUB
END IF

VarX$ = Var$
IF MID$(VarX$, 2, 1) = ":" THEN
    PRINT "Drive "; LEFT$(VarX$, 3)
    VarX$ = MID$(VarX$, 3)
END IF
IF LEFT$(VarX$, 2) = "\\" THEN
    V = INSTR(3, VarX$, "\")
    V = INSTR(V + 1, VarX$, "\")
    PRINT "Netpath: "; LEFT$(VarX$, V)
    VarX$ = MID$(VarX$, V)
END IF
FOR V = LEN(VarX$) TO 1 STEP -1
    IF MID$(VarX$, V, 1) = "\" THEN
        Path$ = LEFT$(VarX$, V)
        File$ = MID$(VarX$, V + 1)
        EXIT FOR
    END IF
NEXT
PRINT "Pathname: "; Path$

File$ = UnAmbiguateFile$(Var$)
PRINT "Filename: "; File$

VarX$ = UCASE$(AmbiguateFile$(Var$))
PRINT "Short filename: "; VarX$

' display file attributes/date/time.
FileSize = False
TempASCIIZ = Var$ + CHR$(0)
hfind = FindFirstFileA(_OFFSET(TempASCIIZ), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    Attribute = finddata.dwFileAttributes
    PRINT "Attributes: ";
    IF (Attribute AND &H1) = &H1 THEN ' read-only
        PRINT "Read-Only ";
    END IF
    IF (Attribute AND &H2) = &H2 THEN ' hidden
        PRINT "Hidden ";
    END IF
    IF (Attribute AND &H4) = &H4 THEN ' system
        PRINT "System ";
    END IF
    IF (Attribute AND &H20) = &H20 THEN ' archive
        PRINT "Archive";
    END IF
    PRINT
    FileSize = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow
    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
    PRINT "Create: "; GetDateTime(z)
    x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
    PRINT "Access: "; GetDateTime(z)
    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    PRINT "Modify: "; GetDateTime(z)
    x = FindClose(hfind)
END IF
PRINT "Filesize: "; FormatString$(FileSize)

x = FileEncryptionStatusA(TempASCIIZ, f&)
IF x THEN
    PRINT "Encryption: ";
    SELECT CASE f&
        CASE 0
            PRINT "File can be encrypted."
        CASE 1
            PRINT "File is encrypted."
        CASE 2
            PRINT "File is system."
        CASE 3
            PRINT "File is root."
        CASE 4
            PRINT "File is system directory."
        CASE 5
            PRINT "Encryption status unknown."
        CASE 6
            PRINT "File system does not support encryption."
        CASE 7 ' reserved
        CASE 8
            PRINT "File is read-only."
        CASE ELSE
            PRINT "Unknown."
    END SELECT
END IF
END SUB

' encrypts string
FUNCTION EnCrypt$ (Var2$)
Var$ = Var2$
Var1 = LEN(Var$)
IF Var1 = 0 THEN
    EnCrypt = ""
    EXIT FUNCTION
END IF
FOR Var2 = 0 TO Var1 - 1
    VarZ = ASC(MID$(Var$, Var2 + 1, 1))
    VarZ = VarZ XOR Var1
    IF Var2 = False THEN
        VarZ = VarZ XOR Primekey
    ELSE
        VarZ = VarZ XOR Var2
        VarZ = VarZ XOR ASC(MID$(Var$, Var2, 1))
    END IF
    MID$(Var$, Var2 + 1, 1) = CHR$(VarZ)
NEXT
IF Var1 = 1 THEN
    EnCrypt = Var$
    EXIT FUNCTION
END IF
FOR Var2 = Var1 - 1 TO 0 STEP -1
    VarZ = ASC(MID$(Var$, Var2 + 1, 1))
    VarZ = VarZ XOR Var1
    IF Var2 = Var1 - 1 THEN
        VarZ = VarZ XOR Primekey
    ELSE
        VarZ = VarZ XOR Var2
        VarZ = VarZ XOR ASC(MID$(Var$, Var2 + 2, 1))
    END IF
    MID$(Var$, Var2 + 1, 1) = CHR$(VarZ)
NEXT
EnCrypt = Var$
END FUNCTION

' user profiles editor menu
SUB EditProfiles
' check locked profiles
IF LockProfiles THEN
    IF ProfilesActive = 0 THEN
        PRINT "User profiles not active."
        EXIT SUB
    END IF
    IF Username <> "SYSOP" THEN
        PRINT "User profiles editor locked."
        EXIT SUB
    END IF
END IF
DO
    COLOR 14
    PRINT "User Profiles Editor Options"
    COLOR 15
    PRINT "  (A)dd"
    PRINT "  (C)hange"
    PRINT "  (D)elete"
    PRINT "  (F)ind"
    PRINT "  (L)ist"
    PRINT "  (P)ack"
    PRINT "  (S)ort"
    PRINT "  (U)ndelete"
    COLOR 14
    PRINT "Enter option(Q to quit)?";
    COLOR 15
    DO
        _LIMIT 50
        LOCATE , , 1
        x$ = INKEYx$
        IF LEN(x$) THEN
            PRINT
            EXIT DO
        END IF
    LOOP
    IF UCASE$(x$) = "Q" THEN
        COLOR 7
        EXIT SUB
    END IF
    SELECT CASE UCASE$(x$)
        CASE "A"
            CALL AddProfile
        CASE "C"
            CALL ChangeProfile
        CASE "D"
            CALL DeleteProfile
        CASE "F"
            CALL FindProfile
        CASE "L"
            CALL ListProfiles
        CASE "P"
            CALL PackProfiles
        CASE "S"
            CALL SortProfiles
        CASE "U"
            CALL UndeleteProfile
    END SELECT
LOOP
END SUB

SUB AddProfile
f$ = GetProfileFilename$
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
DO
    COLOR 15
    PRINT "Username?";
    X$ = LineInput$(CSRLIN, POS(0), 20): PRINT
    COLOR 15
    PRINT "Password?";
    MaskInput = -1
    Y$ = LineInput$(CSRLIN, POS(0), 8): PRINT
    MaskInput = 0
    IF X$ = "" OR Y$ = "" THEN
        PRINT "Username entry failed."
        EXIT DO
    END IF
    X$ = UCASE$(X$)
    Y$ = UCASE$(Y$)
    IF LEN(X$) < 3 THEN
        PRINT "Username must be at least 3 characters."
    ELSE
        IF LEN(Y$) < 3 THEN
            PRINT "Password must be at least 3 characters."
        ELSE
            IF LEN(X$) > 20 THEN
                PRINT "Username can be at most 20 characters."
            ELSE
                IF LEN(Y$) > 8 THEN
                    PRINT "Password can be at most 8 characters."
                ELSE
                    IF INSTR(X$, " ") THEN
                        PRINT "Invalid characters in username."
                    ELSE
                        V = 0
                        FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord2)
                            GET X, Z!, UserProfileRecord2
                            IF X$ = RTRIM$(UserProfileRecord2.Username) THEN
                                PRINT "Username already exists."
                                V = -1
                                EXIT FOR
                            END IF
                        NEXT
                        IF V = 0 THEN
                            UserProfileRecord2.Username = X$
                            UserProfileRecord2.Password = EnCrypt$(Y$)
                            UserProfileRecord2.ScreenSaver = 0
                            UserProfileRecord2.ScreenTimer = 0
                            UserProfileRecord2.LastDateOn = DATE$ + " " + TIME$ ' 01/01/1980 HH:MM:SS
                            UserProfileRecord2.CreationDate = DATE$ + " " + TIME$
                            UserProfileRecord2.TimesLoggedOn = 0#

                            UserProfileRecord2.ASCIIZreserved1 = ""
                            UserProfileRecord2.ASCIIZreserved2 = ""
                            UserProfileRecord2.DOSreserved1 = ""
                            UserProfileRecord2.DOSreserved2 = ""
                            UserProfileRecord2.TimesDeleted = 0
                            UserProfileRecord2.TimesUndeleted = 0

                            UserProfileRecord2.MonitorOff = 0
                            UserProfileRecord2.MonitorTimer = 60 ' seconds
                            UserProfileRecord2.MonitorLimit = 10 ' megabytes

                            UserProfileRecord2.Reserved1 = 0%
                            UserProfileRecord2.Reserved2 = 0%
                            UserProfileRecord2.Reserved3 = 0!
                            UserProfileRecord2.Reserved4 = 0!
                            UserProfileRecord2.Reserved5 = 0&
                            UserProfileRecord2.Reserved6 = 0&
                            UserProfileRecord2.Reserved7 = 0#
                            UserProfileRecord2.Reserved8 = 0#

                            UserProfileRecord2.Deleted = 0
                            Z! = LOF(X) / LEN(UserProfileRecord2) + 1!
                            PUT #X, Z!, UserProfileRecord2
                            PRINT "Username created."
                            EXIT DO
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
LOOP
CLOSE X
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB ChangeProfile
f$ = GetProfileFilename$
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
DO
    COLOR 15
    PRINT "Username?";
    X$ = LineInput$(CSRLIN, POS(0), 20): PRINT
    COLOR 15
    PRINT "Password?";
    MaskInput = -1
    Y$ = LineInput$(CSRLIN, POS(0), 8): PRINT
    MaskInput = 0
    IF X$ = "" OR Y$ = "" THEN
        PRINT "Username entry failed."
        EXIT DO
    END IF
    X$ = UCASE$(X$)
    Y$ = UCASE$(Y$)
    IF LEN(X$) < 3 THEN
        PRINT "Username must be at least 3 characters."
    ELSE
        IF LEN(Y$) < 3 THEN
            PRINT "Password must be at least 3 characters."
        ELSE
            IF LEN(X$) > 20 THEN
                PRINT "Username can be at most 20 characters."
            ELSE
                IF LEN(Y$) > 8 THEN
                    PRINT "Password can be at most 8 characters."
                ELSE
                    FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord2)
                        GET X, Z!, UserProfileRecord2
                        IF X$ = RTRIM$(UserProfileRecord2.Username) THEN
                            UserProfileRecord2.Password = EnCrypt$(Y$)
                            PUT #X, Z!, UserProfileRecord2
                            PRINT "Username changed."
                            EXIT DO
                        END IF
                    NEXT
                    PRINT "Username not found."
                END IF
            END IF
        END IF
    END IF
LOOP
CLOSE X
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB DeleteProfile
f$ = GetProfileFilename$
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
DO
    COLOR 15
    PRINT "Username?";
    X$ = LineInput$(CSRLIN, POS(0), 20): PRINT
    COLOR 15
    IF X$ = "" THEN
        PRINT "Username entry failed."
        EXIT DO
    END IF
    X$ = UCASE$(X$)
    IF X$ = "SYSOP" THEN
        PRINT "Cannot delete Sysop."
        EXIT DO
    END IF
    IF Username > "" THEN
        IF X$ = Username THEN
            PRINT "Cannot delete yourself."
            EXIT DO
        END IF
    END IF
    IF LEN(X$) < 3 THEN
        PRINT "Username must be at least 3 characters."
    ELSE
        IF LEN(X$) > 20 THEN
            PRINT "Username can be at most 20 characters."
        ELSE
            FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord2)
                GET X, Z!, UserProfileRecord2
                IF X$ = RTRIM$(UserProfileRecord2.Username) THEN
                    IF UserProfileRecord2.Deleted THEN
                        PRINT "Username already deleted."
                    ELSE
                        UserProfileRecord2.Deleted = -1
                        UserProfileRecord2.TimesDeleted = UserProfileRecord2.TimesDeleted + 1
                        PUT #X, Z!, UserProfileRecord2
                        PRINT "Username deleted."
                    END IF
                    EXIT DO
                END IF
            NEXT
            PRINT "Username not found."
        END IF
    END IF
LOOP
CLOSE X
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB PackProfiles
f$ = GetProfileFilename$
g$ = GetProfileBackupFilename$
IF _FILEEXISTS(g$) THEN
    x$ = g$ + CHR$(0)
    r = DeleteFileA(x$)
END IF

X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
Y = FREEFILE
OPEN g$ FOR RANDOM AS #Y LEN = LEN(UserProfileRecord2)

' copy datafile to backup
FOR Z1! = 1! TO LOF(X) / LEN(UserProfileRecord2)
    GET X, Z1!, UserProfileRecord2
    PUT Y, Z1!, UserProfileRecord2
NEXT
CLOSE X, Y

' delete profiles datafile
x$ = f$ + CHR$(0)
X = DeleteFileA(x$)

' copy backup to datafile packing deleted usernames
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
Y = FREEFILE
OPEN g$ FOR RANDOM AS #Y LEN = LEN(UserProfileRecord2)

FOR Z1! = 1! TO LOF(Y) / LEN(UserProfileRecord2)
    GET Y, Z1!, UserProfileRecord2
    IF UserProfileRecord2.Deleted = 0 THEN
        Z2! = LOF(X) / LEN(UserProfileRecord2) + 1!
        PUT #X, Z2!, UserProfileRecord2
    END IF
NEXT
CLOSE X, Y
PRINT "Profiles packed."
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB SortProfiles
f$ = GetProfileFilename$
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
CALL MorePrompt("Sort (a)scending/(d)escending?", "ad", Outpt2$)
V = 0
IF Outpt2$ = "a" THEN
    V = -1
END IF
FOR Z1! = 2! TO LOF(X) / LEN(UserProfileRecord)
    FOR Z2! = Z1! + 1! TO LOF(X) / LEN(UserProfileRecord)
        GET X, Z1!, UserProfileRecord
        GET X, Z2!, UserProfileRecord2
        IF V THEN
            IF UserProfileRecord.Username > UserProfileRecord2.Username THEN
                PUT X, Z1!, UserProfileRecord2
                PUT X, Z2!, UserProfileRecord
            END IF
        ELSE
            IF UserProfileRecord.Username < UserProfileRecord2.Username THEN
                PUT X, Z1!, UserProfileRecord2
                PUT X, Z2!, UserProfileRecord
            END IF
        END IF
    NEXT
NEXT
CLOSE X
PRINT "User profiles sorted."
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB UndeleteProfile
f$ = GetProfileFilename$
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord2)
DO
    COLOR 15
    PRINT "Username?";
    X$ = LineInput$(CSRLIN, POS(0), 20): PRINT
    COLOR 15
    IF X$ = "" THEN
        PRINT "Username entry failed."
        EXIT DO
    END IF
    X$ = UCASE$(X$)
    IF LEN(X$) < 3 THEN
        PRINT "Username must be at least 3 characters."
    ELSE
        IF LEN(X$) > 20 THEN
            PRINT "Username can be at most 20 characters."
        ELSE
            FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord2)
                GET X, Z!, UserProfileRecord2
                IF X$ = RTRIM$(UserProfileRecord2.Username) THEN
                    IF UserProfileRecord2.Deleted = 0 THEN
                        PRINT "Username not deleted."
                    ELSE
                        UserProfileRecord2.Deleted = 0
                        UserProfileRecord2.TimesUndeleted = UserProfileRecord2.TimesUndeleted + 1
                        PUT #X, Z!, UserProfileRecord2
                        PRINT "Username undeleted."
                    END IF
                    EXIT DO
                END IF
            NEXT
            PRINT "Username not found."
        END IF
    END IF
LOOP
CLOSE X
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END SUB

SUB GetUserStats (V#)
V# = 0#
f$ = GetProfileFilename$
IF _FILEEXISTS(f$) = 0 THEN EXIT SUB
X = FREEFILE
OPEN f$ FOR RANDOM AS #X LEN = LEN(UserProfileRecord)
IF LOF(X) = 0 THEN
    CLOSE X
    EXIT SUB
END IF
FOR Z! = 1! TO LOF(X) / LEN(UserProfileRecord2)
    GET X, Z!, UserProfileRecord2
    IF RTRIM$(UserProfileRecord2.Username) = RTRIM$(Username) THEN
        V# = UserProfileRecord2.TimesLoggedOn
        EXIT FOR
    END IF
NEXT
CLOSE #X
END SUB

' simple file edit menu
SUB Menu
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM hfind AS _OFFSET

c$ = _CWD$
DO
    COLOR 14
    PRINT "Main Options for ";
    COLOR 12
    PRINT _CWD$
    COLOR 15
    PRINT "  (D)irectory"
    PRINT "  (F)iles"
    PRINT "  (V)olume"
    PRINT "  (S)ystem"
    PRINT "  (Q)uit"
    COLOR 14
    PRINT "Enter option?";
    COLOR 15
    DO
        _LIMIT 50
        LOCATE , , 1
        x$ = INKEYx$
        IF LEN(x$) THEN
            PRINT
            EXIT DO
        END IF
    LOOP
    IF UCASE$(x$) = "Q" THEN
        COLOR 7
        f$ = c$ + CHR$(0)
        x = SetCurrentDirectoryA(f$)
        EXIT SUB
    END IF
    SELECT CASE UCASE$(x$)
        CASE "S"
            DO
                COLOR 14
                PRINT "System Options:"
                COLOR 15
                PRINT "  (S)ystem RAM"
                PRINT "  (P)aging RAM"
                PRINT "  (V)irtual RAM"
                PRINT "  (Q)uit"
                COLOR 14
                PRINT "Enter option?";
                COLOR 15
                DO
                    _LIMIT 50
                    LOCATE , , 1
                    x$ = INKEYx$
                    IF LEN(x$) THEN
                        PRINT
                        EXIT DO
                    END IF
                LOOP
                IF UCASE$(x$) = "Q" THEN EXIT DO
                SELECT CASE UCASE$(x$)
                    CASE "S"
                        COLOR 14
                        PRINT "Total Physical Memory: "; FormatString$(CDBL(TotalPhysicalMem))
                        PRINT "Free Physical Memory: "; FormatString$(CDBL(FreePhysicalMem))
                        Used# = CDBL(TotalPhysicalMem) - CDBL(FreePhysicalMem)
                        PRINT "Used Physical Memory: "; FormatString$(Used#)
                        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                    CASE "P"
                        COLOR 14
                        PRINT "Total Paging File: "; FormatString$(CDBL(TotalPagingFile))
                        PRINT "Free Paging File: "; FormatString$(CDBL(FreePagingFile))
                        Used# = CDBL(TotalPagingFile) - CDBL(FreePagingFile)
                        PRINT "Used Paging File: "; FormatString$(Used#)
                        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                    CASE "V"
                        COLOR 14
                        PRINT "Total Virtual Memory: "; FormatString$(CDBL(TotalVirtualMem))
                        PRINT "Free Virtual Memory: "; FormatString$(CDBL(FreeVirtualMem))
                        Used# = CDBL(TotalVirtualMem) - CDBL(FreeVirtualMem)
                        PRINT "Used Virtual Memory: "; FormatString$(Used#)
                        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                END SELECT
            LOOP
        CASE "D"
            DO
                COLOR 14
                PRINT "Directory Options:"
                COLOR 15
                PRINT "  (A)ttribute override"
                PRINT "  (C)reate directory"
                PRINT "  (L)ist directories"
                PRINT "  (R)emove directory"
                PRINT "  (S)et current directory"
                PRINT "  (V)erify duplicates"
                PRINT "  (Q)uit"
                COLOR 14
                PRINT "Enter option?";
                COLOR 15
                DO
                    _LIMIT 50
                    LOCATE , , 1
                    x$ = INKEYx$
                    IF LEN(x$) THEN
                        PRINT
                        EXIT DO
                    END IF
                LOOP
                IF UCASE$(x$) = "Q" THEN EXIT DO
                SELECT CASE UCASE$(x$)
                    CASE "A"
                        PRINT "Enter pathname? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            ASCIIZ = f$ + CHR$(0)
                            Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
                            IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
                                Attribute = finddata.dwFileAttributes
                                x = FindClose(Wfile.Handle)
                                ' check for directory
                                IF (Attribute AND &H10) = &H10 THEN
                                    DO
                                        COLOR 14
                                        PRINT "Attributes:"
                                        COLOR 15
                                        PRINT "  (1)Read-Only ";
                                        IF (Attribute AND &H1) = &H1 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (2)Hidden ";
                                        IF (Attribute AND &H2) = &H2 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (3)System ";
                                        IF (Attribute AND &H4) = &H4 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (4)Archive ";
                                        IF (Attribute AND &H20) = &H20 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (Q)uit"
                                        COLOR 14
                                        PRINT "Enter option?";
                                        COLOR 15
                                        DO
                                            _LIMIT 50
                                            LOCATE , , 1
                                            x$ = INKEYx$
                                            IF LEN(x$) THEN
                                                PRINT
                                                EXIT DO
                                            END IF
                                        LOOP
                                        IF UCASE$(x$) = "Q" THEN EXIT DO
                                        SELECT CASE x$
                                            CASE "1"
                                                Attribute = Attribute XOR &H1
                                            CASE "2"
                                                Attribute = Attribute XOR &H2
                                            CASE "3"
                                                Attribute = Attribute XOR &H4
                                            CASE "4"
                                                Attribute = Attribute XOR &H20
                                        END SELECT
                                    LOOP
                                    ' change filename attribute
                                    x = SetFileAttributes&(ASCIIZ, Attribute)
                                ELSE
                                    COLOR 15
                                    PRINT "Specified filename."
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            ELSE
                                COLOR 15
                                PRINT "Path not found."
                                CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                            END IF
                        END IF
                    CASE "V"
                        COLOR 15
                        PRINT "Enter first path? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            COLOR 15
                            PRINT "Enter second path? ";
                            g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                            IF g$ <> "" THEN
                                IF _DIREXISTS(f$) AND _DIREXISTS(g$) THEN
                                    CALL VerifyFiles(f$, g$)
                                ELSE
                                    COLOR 15
                                    PRINT "Path not found."
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            END IF
                        END IF
                    CASE "L"
                        PRINT "Enter directory spec? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            LineCount = 0
                            IF Status THEN
                                LineBreak = 23
                            ELSE
                                LineBreak = 24
                            END IF
                            Quit = 0
                            ' start directory search
                            ASCIIZ = f$ + CHR$(0)
                            Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
                            IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
                                DO
                                    Attribute = finddata.dwFileAttributes
                                    ' check for directory
                                    IF (Attribute AND &H10) = &H10 THEN
                                        ' store directory name
                                        Directory$ = finddata.cFileName
                                        Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)
                                        ' check directory name
                                        IF Directory$ <> "." AND Directory$ <> ".." THEN
                                            PRINT Directory$
                                            LineCount = LineCount + 1
                                            IF LineCount >= LineBreak THEN
                                                LineCount = 0
                                                CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt$)
                                                IF Outpt$ = "n" THEN
                                                    EXIT DO
                                                END IF
                                            END IF
                                        END IF
                                    END IF
                                LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
                                x = FindClose(Wfile.Handle)
                                IF LineCount THEN
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            ELSE
                                COLOR 15
                                PRINT "File not found."
                            END IF
                        END IF
                    CASE "C"
                        PRINT "Enter directory? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            x = CreateDirectoryStructure(f$)
                            COLOR 14
                            IF x = 0 THEN
                                IF NetPathError THEN
                                    PRINT "Netpath error."
                                ELSE
                                    PRINT "Error 0x" + HEX$(GetLastError) + " creating directory."
                                    PRINT DisplayWinError$(x)
                                END IF
                            ELSE
                                PRINT "Directory created."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "R"
                        PRINT "Enter directory? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            IF RIGHT$(f$, 1) <> "\" THEN f$ = f$ + "\"
                            COLOR 14
                            IF _DIREXISTS(f$) = 0 THEN
                                PRINT "Directory not found."
                            ELSE
                                CALL Silentdel_Directories(f$)
                                IF _DIREXISTS(f$) THEN
                                    PRINT "Error deleting directory."
                                ELSE
                                    PRINT "Directory removed."
                                END IF
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "S"
                        PRINT "Enter directory? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            x = SetCurrentDirectoryA(f$)
                            COLOR 14
                            IF x = 0 THEN
                                PRINT "Error 0x" + HEX$(GetLastError) + " changing directory."
                                PRINT DisplayWinError$(x)
                            ELSE
                                PRINT "Directory changed."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                END SELECT
            LOOP
        CASE "F"
            DO
                COLOR 14
                PRINT "File Options:"
                COLOR 15
                PRINT "  (A)ppend file"
                PRINT "  (C)opy file"
                PRINT "  (D)elete file"
                PRINT "  (E)ncrypt file"
                PRINT "  (G)et filesize"
                PRINT "  (L)ist files"
                PRINT "  (M)ake file"
                PRINT "  (R)ename file"
                PRINT "  (S)et attributes"
                PRINT "  (T)est file"
                PRINT "  (U)nencrypt file"
                PRINT "  (V)erify files"
                PRINT "  (Q)uit"
                COLOR 14
                PRINT "Enter option?";
                COLOR 15
                DO
                    _LIMIT 50
                    LOCATE , , 1
                    x$ = INKEYx$
                    IF LEN(x$) THEN
                        PRINT
                        EXIT DO
                    END IF
                LOOP
                IF UCASE$(x$) = "Q" THEN EXIT DO
                SELECT CASE UCASE$(x$)
                    CASE "V"
                        COLOR 15
                        PRINT "Enter first file? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            COLOR 15
                            PRINT "Enter second file? ";
                            g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                            IF g$ <> "" THEN
                                IF _FILEEXISTS(f$) AND _FILEEXISTS(g$) THEN
                                    CALL VerifyFiles2(f$, g$, 0)
                                ELSE
                                    COLOR 15
                                    PRINT "File not found."
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            END IF
                        END IF
                    CASE "G"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            ASCIIZ = f$ + CHR$(0)
                            Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
                            IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
                                File.Size# = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow
                                PRINT FormatString$(File.Size#)
                                x = FindClose(Wfile.Handle)
                            ELSE
                                COLOR 15
                                PRINT "File not found."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "S"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            ASCIIZ = f$ + CHR$(0)
                            Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
                            IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
                                Attribute = finddata.dwFileAttributes
                                x = FindClose(Wfile.Handle)
                                ' check for directory
                                IF (Attribute AND &H10) <> &H10 THEN
                                    DO
                                        COLOR 14
                                        PRINT "Attributes:"
                                        COLOR 15
                                        PRINT "  (1)Read-Only ";
                                        IF (Attribute AND &H1) = &H1 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (2)Hidden ";
                                        IF (Attribute AND &H2) = &H2 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (3)System ";
                                        IF (Attribute AND &H4) = &H4 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (4)Archive ";
                                        IF (Attribute AND &H20) = &H20 THEN
                                            PRINT "<set>"
                                        ELSE
                                            PRINT "<off>"
                                        END IF
                                        PRINT "  (Q)uit"
                                        COLOR 14
                                        PRINT "Enter option?";
                                        COLOR 15
                                        DO
                                            _LIMIT 50
                                            LOCATE , , 1
                                            x$ = INKEYx$
                                            IF LEN(x$) THEN
                                                PRINT
                                                EXIT DO
                                            END IF
                                        LOOP
                                        IF UCASE$(x$) = "Q" THEN EXIT DO
                                        SELECT CASE x$
                                            CASE "1"
                                                Attribute = Attribute XOR &H1
                                            CASE "2"
                                                Attribute = Attribute XOR &H2
                                            CASE "3"
                                                Attribute = Attribute XOR &H4
                                            CASE "4"
                                                Attribute = Attribute XOR &H20
                                        END SELECT
                                    LOOP
                                    ' change filename attribute
                                    x = SetFileAttributes&(ASCIIZ, Attribute)
                                ELSE
                                    COLOR 15
                                    PRINT "Specified directory."
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            ELSE
                                COLOR 15
                                PRINT "File not found."
                                CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                            END IF
                        END IF
                    CASE "L"
                        PRINT "Enter file spec? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            LineCount = 0
                            IF Status THEN
                                LineBreak = 23
                            ELSE
                                LineBreak = 24
                            END IF
                            Quit = 0
                            ' start directory search
                            ASCIIZ = f$ + CHR$(0)
                            Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
                            IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
                                DO
                                    Attribute = finddata.dwFileAttributes
                                    ' check for directory
                                    IF (Attribute AND &H10) <> &H10 THEN
                                        ' store directory name
                                        Directory$ = finddata.cFileName
                                        Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)
                                        ' check directory name
                                        IF Directory$ <> "." AND Directory$ <> ".." THEN
                                            PRINT Directory$
                                            LineCount = LineCount + 1
                                            IF LineCount >= LineBreak THEN
                                                LineCount = 0
                                                CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt$)
                                                IF Outpt$ = "n" THEN
                                                    EXIT DO
                                                END IF
                                            END IF
                                        END IF
                                    END IF
                                LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
                                x = FindClose(Wfile.Handle)
                                IF LineCount THEN
                                    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                                END IF
                            ELSE
                                COLOR 15
                                PRINT "File not found."
                            END IF
                        END IF
                    CASE "T"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            x = FileEncryptionStatusA(f$, f&)

                            'FileEncryptionStatus return variable:

                            '0 = can be encrypted
                            '1 = file is encrypted
                            '2 = file is a system file (cannot be encrypted)
                            '3 = file is root (cannot be encryped)
                            '4 = file is system directory (cannot be encrypted)
                            '5 = encryption status unknown. file may be encrypted
                            '6 = file system does not support file encryption
                            '7 = reserved
                            '8 = file is read-only

                            IF x = 0 THEN
                                PRINT "Error 0x" + HEX$(GetLastError) + " accessing file."
                                PRINT DisplayWinError$(x)
                            END IF
                            IF x THEN
                                SELECT CASE f&
                                    CASE 0
                                        PRINT "File can be encrypted."
                                    CASE 1
                                        PRINT "File is encrypted."
                                    CASE 2
                                        PRINT "File is system."
                                    CASE 3
                                        PRINT "File is root."
                                    CASE 4
                                        PRINT "File is system directory."
                                    CASE 5
                                        PRINT "Encryption status unknown."
                                    CASE 6
                                        PRINT "File system does not support encryption."
                                    CASE 7 ' reserved
                                    CASE 8
                                        PRINT "File is read-only."
                                END SELECT
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "E"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            x = EncryptFile(f$)
                            COLOR 14
                            IF x = 0 THEN
                                PRINT "Error 0x" + HEX$(GetLastError) + " encrypting file."
                                PRINT DisplayWinError$(x)
                            ELSE
                                PRINT "File encrypted."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "U"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            x = DecryptFile(f$, 0) ' 0=reserved
                            COLOR 14
                            IF x = 0 THEN
                                PRINT "Error 0x" + HEX$(GetLastError) + " decrypting file."
                                PRINT DisplayWinError$(x)
                            ELSE
                                PRINT "File decrypted."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "D"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            x = DeleteFileA(f$)
                            COLOR 14
                            IF x = 0 THEN
                                PRINT "Error 0x" + HEX$(GetLastError) + " deleting file."
                                PRINT DisplayWinError$(x)
                            ELSE
                                PRINT "File deleted."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "A"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            hfind = CreateFile(f$, &H180, 0, 0, 4, 0, 0)
                            COLOR 14
                            IF hfind <> INVALID_HANDLE_VALUE THEN
                                PRINT "File accessed for append."
                                x = CloseHandle(hfind)
                            ELSE
                                PRINT "Error 0x" + HEX$(GetLastError) + " appending file."
                                PRINT DisplayWinError$(x)
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "M"
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = f$ + CHR$(0)
                            hfind = CreateFile(f$, &H180, 0, 0, 1, 0, 0)
                            COLOR 14
                            IF hfind <> INVALID_HANDLE_VALUE THEN
                                PRINT "File created."
                                x = CloseHandle(hfind)
                            ELSE
                                PRINT "Error 0x" + HEX$(GetLastError) + " creating file."
                                PRINT DisplayWinError$(x)
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "R"
                        COLOR 15
                        PRINT "Enter filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            COLOR 15
                            PRINT "Enter new filename? ";
                            g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                            IF g$ <> "" THEN
                                f$ = f$ + CHR$(0)
                                g$ = g$ + CHR$(0)
                                x = MoveFile(f$, g$)
                                COLOR 14
                                IF x = 0 THEN
                                    PRINT "Error 0x" + HEX$(GetLastError) + " renaming filename."
                                    PRINT DisplayWinError$(x)
                                ELSE
                                    PRINT "Filename renamed."
                                END IF
                                CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                            END IF
                        END IF
                    CASE "C"
                        COLOR 15
                        PRINT "Enter source filename? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            COLOR 15
                            PRINT "Enter dest filename? ";
                            g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                            IF g$ <> "" THEN
                                f$ = f$ + CHR$(0)
                                g$ = g$ + CHR$(0)
                                x = CopyFile(f$, g$, 0) ' 0=overwrite, -1=don't
                                COLOR 14
                                IF x = 0 THEN
                                    PRINT "Error 0x" + HEX$(GetLastError) + " copying filename."
                                    PRINT DisplayWinError$(x)
                                ELSE
                                    PRINT "Filename copied."
                                END IF
                                CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                            END IF
                        END IF
                END SELECT
            LOOP
        CASE "V"
            DO
                COLOR 14
                PRINT "Volume Options:"
                COLOR 15
                PRINT "  (C)hange volume"
                PRINT "  (D)isplay serial"
                PRINT "  (G)et system type"
                PRINT "  (L)ist volume"
                PRINT "  (R)ead disk space"
                PRINT "  (Q)uit"
                COLOR 14
                PRINT "Enter option?";
                COLOR 15
                DO
                    _LIMIT 50
                    LOCATE , , 1
                    x$ = INKEYx$
                    IF LEN(x$) THEN
                        PRINT
                        EXIT DO
                    END IF
                LOOP
                IF UCASE$(x$) = "Q" THEN EXIT DO
                COLOR 14
                SELECT CASE UCASE$(x$)
                    CASE "D"
                        PRINT "Enter drive letter? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = UCASE$(f$)
                            IF f$ >= "A" AND f$ <= "Z" THEN
                                v = GetDriveExists(f$)
                                IF v THEN
                                    ' get drive info.                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Out3$ = ""
                                    Vname$ = SPACE$(MAX_PATH)
                                    Fname$ = SPACE$(MAX_PATH)
                                    VarX$ = f$ + ":\" + CHR$(0)
                                    R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
                                    IF R THEN
                                        ' get serial number.
                                        Out3$ = LEFT$(HEX$(serial~&), 4) + "-" + RIGHT$(HEX$(serial~&), 4)
                                        PRINT f$; ":"; Out3$
                                    ELSE
                                        PRINT "Media not inserted."
                                    END IF
                                ELSE
                                    PRINT "Drive not found."
                                END IF
                            ELSE
                                PRINT "Invalid drive eltter."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "G"
                        PRINT "Enter drive letter? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = UCASE$(f$)
                            IF f$ >= "A" AND f$ <= "Z" THEN
                                v = GetDriveExists(f$)
                                IF v THEN
                                    ' get drive info.                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Out3$ = ""
                                    Vname$ = SPACE$(MAX_PATH)
                                    Fname$ = SPACE$(MAX_PATH)
                                    VarX$ = f$ + ":\" + CHR$(0)
                                    R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
                                    IF R THEN
                                        ' get system type.
                                        Out3$ = RTRIM$(Fname$)
                                        PRINT f$; ":"; Out3$
                                    ELSE
                                        PRINT "Media not inserted."
                                    END IF
                                ELSE
                                    PRINT "Drive not found."
                                END IF
                            ELSE
                                PRINT "Invalid drive eltter."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "R"
                        PRINT "Enter drive letter? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = UCASE$(f$)
                            IF f$ >= "A" AND f$ <= "Z" THEN
                                v = GetDriveExists(f$)
                                IF v THEN
                                    ' get drive info.                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Out3$ = ""
                                    Vname$ = SPACE$(MAX_PATH)
                                    Fname$ = SPACE$(MAX_PATH)
                                    VarX$ = f$ + ":\" + CHR$(0)
                                    R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
                                    IF R THEN
                                        ' get drive space.
                                        R = GetDiskFreeSpaceExA(VarX$, free~&&, total~&&, free2~&&)
                                        IF R THEN
                                            PRINT "Total: "; FormatString$(CDBL(total~&&))
                                            PRINT "Free: "; FormatString$(CDBL(free~&&))
                                        ELSE
                                            PRINT "Disk space unavailable."
                                        END IF
                                    ELSE
                                        PRINT "Media not inserted."
                                    END IF
                                ELSE
                                    PRINT "Drive not found."
                                END IF
                            ELSE
                                PRINT "Invalid drive eltter."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "L"
                        PRINT "Enter drive letter? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = UCASE$(f$)
                            IF f$ >= "A" AND f$ <= "Z" THEN
                                v = GetDriveExists(f$)
                                IF v THEN
                                    PRINT f$; ":";
                                    ' get drive info.                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Out3$ = ""
                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Vname$ = SPACE$(MAX_PATH)
                                    Fname$ = SPACE$(MAX_PATH)
                                    R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
                                    IF R THEN
                                        ' get volume label.
                                        Out3$ = RTRIM$(Vname$)
                                        v = INSTR(Out3$, CHR$(0))
                                        IF v THEN Out3$ = LEFT$(Out3$, v - 1)
                                    END IF
                                    IF RTRIM$(Out3$) = "" THEN
                                        PRINT DriveType
                                    ELSE
                                        PRINT Out3$
                                    END IF
                                ELSE
                                    PRINT "Drive not found."
                                END IF
                            ELSE
                                PRINT "Invalid drive eltter."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                    CASE "C"
                        PRINT "Enter drive letter? ";
                        f$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                        IF f$ <> "" THEN
                            f$ = UCASE$(f$)
                            IF f$ >= "A" AND f$ <= "Z" THEN
                                v = GetDriveExists(f$)
                                IF v THEN
                                    ' get drive info.
                                    VarX$ = f$ + ":\" + CHR$(0)
                                    Vname$ = SPACE$(MAX_PATH)
                                    Fname$ = SPACE$(MAX_PATH)
                                    R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
                                    IF R THEN
                                        PRINT "Enter drive volume? ";
                                        g$ = LineInput$(CSRLIN, POS(0), 40): PRINT
                                        ' change volume label
                                        f$ = f$ + ":\" + CHR$(0)
                                        g$ = g$ + CHR$(0)
                                        x = SetVolumeLabelA%(f$, g$)
                                        ' check volume flag
                                        IF x = 0 THEN
                                            PRINT "Error 0x" + HEX$(GetLastError) + " naming volume."
                                            PRINT DisplayWinError(x)
                                        ELSE
                                            PRINT "Volume changed."
                                        END IF
                                    ELSE
                                        PRINT "Media not inserted."
                                    END IF
                                ELSE
                                    PRINT "Drive not found."
                                END IF
                            ELSE
                                PRINT "Invalid drive letter."
                            END IF
                            CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
                        END IF
                END SELECT
            LOOP
    END SELECT
LOOP
END SUB

SUB VerifyFiles (Var1$, Var2$)
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM hfind AS _OFFSET

DIM Verify1(1) AS STRING
DIM Verify2(1) AS STRING
DIM VerifyCount1 AS SINGLE
DIM VerifyCount2 AS SINGLE
VerifyCount1 = 0!
VerifyCount2 = 0!

DIM SizeArray1(1) AS DOUBLE
DIM SizeArray2(1) AS DOUBLE

' make search path
Var3$ = Var1$
IF RIGHT$(Var3$, 1) <> "\" THEN
    Var3$ = Var3$ + "\"
END IF
TempPath1$ = Var3$
Var3$ = Var3$ + "*.*"

Var4$ = Var2$
IF RIGHT$(Var4$, 1) <> "\" THEN
    Var4$ = Var4$ + "\"
END IF
TempPath2$ = Var4$
Var4$ = Var4$ + "*.*"

' store first path files
ASCIIZ = Var3$ + CHR$(0)
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    DO
        Attribute = finddata.dwFileAttributes
        ' check for directory
        IF (Attribute AND &H10) <> &H10 THEN
            ' store directory name
            Directory$ = finddata.cFileName
            Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)
            ' check directory name
            IF Directory$ <> "." AND Directory$ <> ".." THEN
                VerifyCount1 = VerifyCount1 + 1!
                REDIM _PRESERVE Verify1(VerifyCount1) AS STRING
                Verify1(VerifyCount1) = Directory$

                REDIM _PRESERVE SizeArray1(VerifyCount1) AS DOUBLE
                SizeArray1(VerifyCount1) = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF

' store second path files
ASCIIZ = Var4$ + CHR$(0)
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    DO
        Attribute = finddata.dwFileAttributes
        ' check for directory
        IF (Attribute AND &H10) <> &H10 THEN
            ' store directory name
            Directory$ = finddata.cFileName
            Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)
            ' check directory name
            IF Directory$ <> "." AND Directory$ <> ".." THEN
                VerifyCount2 = VerifyCount2 + 1!
                REDIM _PRESERVE Verify2(VerifyCount2) AS STRING
                Verify2(VerifyCount2) = Directory$

                REDIM _PRESERVE SizeArray2(VerifyCount2) AS DOUBLE
                SizeArray2(VerifyCount2) = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF

' check files found
IF VerifyCount1 = 0! OR VerifyCount2 = 0! THEN
    COLOR 15
    PRINT "No matching files found."
    EXIT SUB
END IF

' compare both arrays
LineCount = 0
IF Status THEN
    LineBreak = 23
ELSE
    LineBreak = 24
END IF
Found = 0
FOR Var1! = 1! TO VerifyCount1
    FOR Var2! = 1! TO VerifyCount2
        IF Verify1(Var1!) = Verify2(Var2!) THEN
            IF SizeArray1(Var1!) = SizeArray2(Var2!) THEN
                IF FileCompare(Var1$, Var2$, Verify1(Var1!), Verify2(Var2!)) = -1 THEN
                    IF Found = 0 THEN
                        COLOR 14
                        PRINT "Filename1", "Filename2", "Filesize"
                        PRINT "---------", "---------", "--------"
                    END IF
                    Found = -1
                    COLOR 15
                    File1$ = AmbiguateFile$(TempPath1$ + Verify1(Var1!))
                    File2$ = AmbiguateFile$(TempPath2$ + Verify2(Var2!))
                    PRINT UCASE$(File1$), UCASE$(File2$), FormatString$(SizeArray1(Var1!))
                    LineCount = LineCount + 1
                    IF LineCount >= LineBreak THEN
                        LineCount = 0
                        CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt$)
                        IF Outpt$ = "n" THEN
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    NEXT
NEXT
IF Found = 0 THEN
    COLOR 15
    PRINT "No matching files found."
END IF
IF LineCount THEN
    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
END IF
END SUB

SUB VerifyFiles2 (Var1$, Var2$, VarX)
DIM Byte1 AS STRING * 1
DIM Byte2 AS STRING * 1
X = FREEFILE
OPEN Var1$ FOR BINARY SHARED AS #X LEN = xbuflen
Y = FREEFILE
OPEN Var2$ FOR BINARY SHARED AS #Y LEN = xbuflen
IF LOF(X) <> LOF(Y) THEN
    PRINT "Files not same size."
    CLOSE X, Y
    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
    EXIT SUB
END IF
IF LOF(X) = 0 AND LOF(Y) = 0 THEN
    PRINT "Files both zero."
    CLOSE X, Y
    CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
    EXIT SUB
END IF
Count# = 0#
Percent = 0
Last.Percent = 0
File.Size = LOF(X)

Block# = 0#
DO UNTIL EOF(X)
    GOSUB Display.Percent
    GET #X, , xbuffer1
    GET #Y, , xbuffer2
    Block# = Block# + 1#
    IF xbuffer1 <> xbuffer2 THEN
        IF VarX THEN
            GOTO DisplayX
        END IF
        PRINT "Files do not match."
        CLOSE #X, #Y
        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
        EXIT SUB
    END IF
    IF INKEYx$ = CHR$(27) THEN
        PRINT "Files matching interrupted."
        CLOSE #X, #Y
        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
        EXIT SUB
    END IF
LOOP
CLOSE #X, #Y
PRINT "Files match."
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
EXIT SUB

DisplayX:
CLOSE #X, #Y
X = FREEFILE
OPEN Var1$ FOR BINARY SHARED AS #X LEN = xbuflen
Y = FREEFILE
OPEN Var2$ FOR BINARY SHARED AS #Y LEN = xbuflen

PRINT #StdoutHandle, "Offset         Hex    ASCII"
PRINT #StdoutHandle, "-------------  -----  -------"

LineCount = 0
IF Status THEN
    LineBreak = 23
ELSE
    LineBreak = 24
END IF
Z# = 0#
FOR V# = 1# TO Block# - 1#
    GET #X, , xbuffer1
    GET #Y, , xbuffer2
    Z# = Z# + xbuflen
NEXT
DO UNTIL EOF(X)
    GET #X, , xbuffer1
    GET #Y, , xbuffer2
    FOR C# = 1# TO xbuflen
        Z# = Z# + 1#
        Byte1 = MID$(xbuffer1, C#, 1#)
        Byte2 = MID$(xbuffer2, C#, 1#)
        IF Byte1 <> Byte2 THEN
            V1$ = LTRIM$(STR$(Z#)): V1$ = V1$ + SPACE$(15 - LEN(V1$))
            V2$ = RIGHT$("00" + HEX$(ASC(Byte1)), 2) + "\" + RIGHT$("00" + HEX$(ASC(Byte2)), 2)
            V3$ = RIGHT$("000" + LTRIM$(STR$(ASC(Byte1))), 3) + "\" + RIGHT$("000" + LTRIM$(STR$(ASC(Byte2))), 3)
            VX$ = V1$ + V2$ + "  " + V3$
            PRINT #StdoutHandle, VX$
            LineCount = LineCount + 1
            IF LineCount >= LineBreak THEN
                LineCount = 0
                IF LCASE$(StderrFile$) = "scrn:" THEN
                    CALL MorePrompt("-more-", "yn " + CHR$(13), Outpt$)
                    IF Outpt$ = "n" THEN
                        EXIT DO
                    END IF
                END IF
            END IF
        END IF
    NEXT
LOOP
CLOSE X, Y
IF LineCount THEN
    IF LCASE$(StderrFile$) = "scrn:" THEN
        CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
    END IF
END IF
EXIT SUB

' display percentage scanned in title
Display.Percent:
IF File.Size > 0 THEN
    Count# = Count# + xbuflen
    IF Count# > File.Size THEN
        Count# = File.Size
    END IF
    Percent = (Count# / File.Size) * 100!
    Percent = INT(Percent)
    IF Percent <> Last.Percent THEN
        Last.Percent = Percent
        P$ = RIGHT$(STR$(Percent + 1000%), 3) + "%"
        V$ = Program + " - COMPARE - " + P$
        _TITLE V$
    END IF
END IF
RETURN
END SUB

FUNCTION FileCompare (Path1$, Path2$, Filename1$, Filename2$)
DIM Byte1 AS STRING * 1
DIM Byte2 AS STRING * 1
FileComp1$ = Path1$
IF RIGHT$(FileComp1$, 1) <> "\" THEN
    FileComp1$ = FileComp1$ + "\"
END IF
FileComp1$ = FileComp1$ + Filename1$

FileComp2$ = Path2$
IF RIGHT$(FileComp2$, 1) <> "\" THEN
    FileComp2$ = FileComp2$ + "\"
END IF
FileComp2$ = FileComp2$ + Filename2$

IF _FILEEXISTS(FileComp1$) = 0 THEN
    FileCompare = 0
    EXIT FUNCTION
END IF

IF _FILEEXISTS(FileComp2$) = 0 THEN
    FileCompare = 0
    EXIT FUNCTION
END IF

Count# = 0#
Percent = 0
Last.Percent = 0

X = FREEFILE
OPEN FileComp1$ FOR BINARY SHARED AS #X LEN = xbuflen
Y = FREEFILE
OPEN FileComp2$ FOR BINARY SHARED AS #Y LEN = xbuflen
IF LOF(X) = 0 AND LOF(Y) = 0 THEN
    CLOSE #X, #Y
    FileCompare = -1
    EXIT FUNCTION
END IF
File.Size = LOF(X)
DO UNTIL EOF(X)
    GOSUB Display.Percent
    GET #X, , xbuffer1
    GET #Y, , xbuffer2
    IF xbuffer1 <> xbuffer2 THEN
        CLOSE #X, #Y
        FileCompare = 0
        EXIT FUNCTION
    END IF
    IF INKEYx$ = CHR$(27) THEN
        CLOSE #X, #Y
        FileCompare = -2
        EXIT FUNCTION
    END IF
LOOP
CLOSE #X, #Y
FileCompare = -1
EXIT FUNCTION

' display percentage scanned in title
Display.Percent:
IF File.Size > 0 THEN
    Count# = Count# + xbuflen
    IF Count# > File.Size THEN
        Count# = File.Size
    END IF
    Percent = (Count# / File.Size) * 100!
    Percent = INT(Percent)
    IF Percent <> Last.Percent THEN
        Last.Percent = Percent
        P$ = RIGHT$(STR$(Percent + 1000%), 3) + "%"
        V$ = Program + " - COMPARE - " + P$
        _TITLE V$
    END IF
END IF
RETURN
END FUNCTION

' return -1 if drive exists
FUNCTION GetDriveExists (V$)
VarX$ = V$ + ":\" + CHR$(0)
VarX = GetDriveType(VarX$)
DriveType = ""
SELECT CASE VarX
    CASE 0
        DriveType = "[UNKNOWN]"
    CASE 1
        DriveType = "[BADROOT]"
    CASE 2
        DriveType = "[REMOVABLE]"
    CASE 3
        DriveType = "[FIXED]"
    CASE 4
        DriveType = "[REMOTE]"
    CASE 5
        DriveType = "[CDROM]"
    CASE 6
        DriveType = "[RAMDISK]"
END SELECT
IF VarX > 1 THEN
    GetDriveExists = -1
ELSE
    GetDriveExists = 0
END IF
END FUNCTION

' display windows error message
FUNCTION DisplayWinError$ (x)
' define error message value
v& = GetLastError
' call windows error message routine
x& = FormatMessageA&(&H1200, "", v&, 0, ErrorBuffer$, 260, "")
IF x& THEN
    DisplayWinError$ = LEFT$(ErrorBuffer$, x& - 2)
ELSE
    DisplayWinError$ = "Unknown error 0x" + HEX$(v&) + "."
END IF
END FUNCTION

REM hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, 0, 0)
' parameters:
'   (1) pointer to filename
'   (2) access:
'     x80(128)  - read
'     x100(256) - write
'   (3) sharing
'   (4) security attributes
'   (5) create:
'     1 CREATE_NEW    = create only if not exist
'     2 CREATE_ALWAYS = always create new file
'     3 OPEN_EXISTING = open only if exists
'     4 OPEN_ALWAYS   = open file always
'     5 TRUNCATE_EXISTING = open/truncate to 0 only if exists
'   (6) - flags (standard OSHA)
'open direcctory flag for parameter 6:
'FILE_FLAG_BACKUP_SEMANTICS flag 0x02000000

SUB HexCalc

' Note: In HexCalc the following keys are active:
'   Ctrl-A - move up, Ctrl-B - move down
'   Ctrl-C - move left, Ctrl-D - move right
'   Ctrl-E, Ctrl-F - cycle foreground
'   Ctrl-J, Ctrl-K - cycle background

' coordinate of hexcalc box
Xcoor = 4
Ycoor = 6
IF HexCalcXcoor THEN Xcoor = HexCalcXcoor
IF HexCalcYcoor THEN Ycoor = HexCalcYcoor

' adjust xcoor
IF Status THEN
    Z = 23
    IF Xcoor + 3 > Z THEN
        Xcoor = Xcoor - 1
    END IF
END IF

' coordinate of hexcalc help box
Xcoor2 = 4
Ycoor2 = 6
IF HexCalcHelpXcoor THEN Xcoor2 = HexCalcHelpXcoor
IF HexCalcHelpYcoor THEN Ycoor2 = HexCalcHelpYcoor

' adjust xcoor
IF Status THEN
    Z = 23
    IF Xcoor2 + 4 > Z THEN
        Xcoor2 = Xcoor2 - 1
    END IF
END IF

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen

' reset insert mode.
Ins = -1

GOSUB MakeBox ' create edit box.

CalcBox = 1 ' reset to left box.

Box$ = "0" ' current box
Box1$ = "0" ' left box string
Box2$ = "0" ' right box string

GOSUB SelectBox2 ' display box 2.
GOSUB SelectBox1 ' display box 1.

DO ' calculator loop
    _LIMIT 100
    Var$ = INKEYx$ ' get keystroke
    IF LEN(Var$) THEN ' check keystroke
        SELECT CASE LEN(Var$)
            CASE 1 ' single ascii key
                SELECT CASE ASC(Var$)
                    CASE 1 ' control-a (up)
                        IF Xcoor > 2 THEN
                            Xcoor = Xcoor - 1
                            GOSUB ResetBox
                        END IF
                    CASE 2 ' control-b (down)
                        IF Status THEN
                            Z = 21
                        ELSE
                            Z = 22
                        END IF
                        IF Xcoor < Z THEN
                            Xcoor = Xcoor + 1
                            GOSUB ResetBox
                        END IF
                    CASE 3 ' control-c (left)
                        IF Ycoor > 1 THEN
                            Ycoor = Ycoor - 1
                            GOSUB ResetBox
                        END IF
                    CASE 4 ' control-d (right)
                        IF Ycoor < 50 THEN
                            Ycoor = Ycoor + 1
                            GOSUB ResetBox
                        END IF
                    CASE 5 ' control-e
                        IF HexCalcForeground1 = 15 THEN
                            HexCalcForeground1 = 0
                        ELSE
                            HexCalcForeground1 = HexCalcForeground1 + 1
                        END IF
                        CalcBoxBorderColor = HexCalcForeground1
                        CalcBoxTitleColor = HexCalcForeground1
                        CalcBoxTextColor = HexCalcForeground1
                        GOSUB ResetBox
                    CASE 6 ' control-f
                        IF HexCalcForeground2 = 15 THEN
                            HexCalcForeground2 = 0
                        ELSE
                            HexCalcForeground2 = HexCalcForeground2 + 1
                        END IF
                        GOSUB ResetBox
                    CASE 10 ' control-j
                        IF HexCalcBackground1 = 7 THEN
                            HexCalcBackground1 = 0
                        ELSE
                            HexCalcBackground1 = HexCalcBackground1 + 1
                        END IF
                        GOSUB ResetBox
                    CASE 11 ' control-k
                        IF HexCalcBackground2 = 7 THEN
                            HexCalcBackground2 = 0
                        ELSE
                            HexCalcBackground2 = HexCalcBackground2 + 1
                        END IF
                        GOSUB ResetBox
                    CASE 8 ' backspace
                        IF Ycoor1 - Column3 > 0 THEN
                            IF Ins THEN
                                Box$ = LEFT$(Box$, Ycoor1 - Column3 - 1) + MID$(Box$, Ycoor1 - Column3 + 1)
                                Ycoor1 = Ycoor1 - 1
                                LOCATE Xcoor1, Ycoor1, 1
                                Var$ = MID$(Box$, Ycoor1 - Column3 + 1) + " "
                                COLOR HexCalcForeground2, HexCalcBackground1
                                PRINT Var$;
                                LOCATE Xcoor1, Ycoor1, 1
                                GOSUB PrintBoxes
                            ELSE ' move left
                                Ycoor1 = Ycoor1 - 1
                                LOCATE Xcoor1, Ycoor1, 1
                            END IF
                        END IF
                    CASE 9 ' tab
                        IF CalcBox = 1 THEN
                            CalcBox = 2
                            GOSUB SelectBox2
                            GOSUB PrintBoxes
                        END IF
                    CASE 27 ' escape
                        EXIT DO
                    CASE 63 ' ?
                        GOSUB HelpBox
                        GOSUB ResetBox
                    CASE ELSE ' keystroke
                        VarX = 0 ' valid char flag
                        TempX$ = UCASE$(Var$)
                        SELECT CASE TempX$ ' validate char
                            CASE "0" TO "9"
                                VarX = -1
                            CASE "A" TO "F"
                                IF CalcBox = 1 THEN
                                    VarX = -1
                                END IF
                        END SELECT
                        IF VarX THEN
                            IF Ycoor1 - Column3 + 1 <= LineLength THEN
                                ' insert digit
                                IF Ins THEN
                                    Box$ = LEFT$(Box$, Ycoor1 - Column3) + TempX$ + MID$(Box$, Ycoor1 - Column3 + 1)
                                    Box$ = LEFT$(Box$, LineLength)
                                    LOCATE Xcoor1, Ycoor1, 1
                                    Var$ = MID$(Box$, Ycoor1 - Column3 + 1)
                                    COLOR HexCalcForeground2, HexCalcBackground1
                                    PRINT Var$;
                                    Ycoor1 = Ycoor1 + 1
                                    LOCATE Xcoor1, Ycoor1, 1
                                    GOSUB PrintBoxes
                                ELSE ' overstrike char
                                    IF Ycoor1 - Column3 + 1 > LEN(Box$) THEN
                                        Box$ = Box$ + TempX$ ' append box
                                    ELSE
                                        MID$(Box$, Ycoor1 - Column3 + 1, 1) = TempX$ ' replace
                                    END IF
                                    LOCATE Xcoor1, Ycoor1, 1
                                    COLOR HexCalcForeground2, HexCalcBackground1
                                    PRINT TempX$;
                                    Ycoor1 = Ycoor1 + 1
                                    LOCATE Xcoor1, Ycoor1, 1
                                    GOSUB PrintBoxes
                                END IF
                            END IF
                        END IF
                END SELECT
            CASE 2 ' extended key
                SELECT CASE ASC(RIGHT$(Var$, 1))
                    CASE 15 ' shift-tab
                        IF CalcBox = 2 THEN
                            CalcBox = 1
                            GOSUB SelectBox1
                            GOSUB PrintBoxes
                        END IF
                    CASE 71 ' home
                        Ycoor1 = Column3
                    CASE 79 ' end
                        Ycoor1 = LEN(Box$) + Column3
                    CASE 77 ' right
                        IF Ycoor1 - Column3 + 1 <= LEN(Box$) THEN
                            Ycoor1 = Ycoor1 + 1
                        END IF
                    CASE 75 ' left
                        IF Ycoor1 - Column3 > 0 THEN
                            Ycoor1 = Ycoor1 - 1
                        END IF
                    CASE 82 ' insert
                        Ins = NOT Ins
                        COLOR CalcBoxTitleColor, HexCalcBackground2
                        IF Ins THEN
                            LOCATE Xcoor, Ycoor + 24, 1
                            PRINT "<ins>";
                        ELSE
                            LOCATE Xcoor, Ycoor + 24, 1
                            PRINT STRING$(5, Hline);
                        END IF
                        COLOR HexCalcForeground2, HexCalcBackground1
                    CASE 83 ' delete
                        IF Ycoor1 - Column3 + 1 <= LEN(Box$) THEN
                            Box$ = LEFT$(Box$, Ycoor1 - Column3) + MID$(Box$, Ycoor1 - Column3 + 2)
                            LOCATE Xcoor1, Ycoor1, 1
                            Var$ = MID$(Box$, Ycoor1 - Column3 + 1) + " "
                            COLOR HexCalcForeground2, HexCalcBackground1
                            PRINT Var$;
                            LOCATE Xcoor1, Ycoor1, 1
                            GOSUB PrintBoxes
                        END IF
                    CASE 119 ' control-home
                        Ycoor1 = Column3
                    CASE 117 ' control-end
                        Ycoor1 = LEN(Box$) + Column3
                END SELECT
                LOCATE Xcoor1, Ycoor1, 1
        END SELECT
    END IF
    IF Status THEN
        Z = 21
    ELSE
        Z = 22
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor THEN
            IF MouseY >= Ycoor AND MouseY <= Ycoor + 30 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z THEN
                                Xcoor = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 50 THEN
                                Ycoor = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            GOSUB ResetBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        END IF
    END IF
LOOP

' store coordinates
HexCalcXcoor = Xcoor
HexCalcYcoor = Ycoor
HexCalcHelpXcoor = Xcoor2
HexCalcHelpYcoor = Ycoor2

CALL RestoreScreen ' restore screen area.
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
EXIT SUB

' display edit box
ResetBox:

' restore screen area.
CALL RestoreScreen

' create edit box.
GOSUB MakeBox
Xcoor1 = Xcoor + 1
Ymove = Ycoor1 - Column3

' display box area 1
COLOR HexCalcForeground2, HexCalcBackground1
LOCATE Xcoor1, Ycoor + 3, 1
PRINT SPACE$(8);
LOCATE Xcoor1, Ycoor + 3, 1
IF CalcBox = 1 THEN
    PRINT Box$;
ELSE
    PRINT Box1$;
END IF

' display box area 2
COLOR HexCalcForeground2, HexCalcBackground1
LOCATE Xcoor1, Ycoor + 15, 1
PRINT SPACE$(10);
LOCATE Xcoor1, Ycoor + 15, 1
IF CalcBox = 2 THEN
    PRINT Box$;
ELSE
    PRINT Box2$;
END IF

' reset editing area
IF CalcBox = 1 THEN
    Ycoor1 = Ycoor + 3
    Column3 = Ycoor1
    Ycoor1 = Ycoor1 + Ymove
    LineLength = 8
ELSE
    Ycoor1 = Ycoor + 15
    Column3 = Ycoor1
    Ycoor1 = Ycoor1 + Ymove
    LineLength = 10
END IF
LOCATE Xcoor1, Ycoor1, 1
RETURN

' create edit box.
MakeBox:
' Y-coordinate of editing boxes.
Box1 = Ycoor + 3
Box2 = Ycoor + 15

' display editing box.
COLOR , HexCalcBackground2

' display line 1
LOCATE Xcoor, Ycoor, 1
COLOR CalcBoxBorderColor
PRINT " " + CHR$(ULcorner) + CHR$(Hline);
COLOR CalcBoxTitleColor
PRINT "<esc>=Quit,?=Help";
COLOR CalcBoxBorderColor
PRINT STRING$(4, Hline);
COLOR CalcBoxTitleColor
PRINT "<ins>";
COLOR CalcBoxBorderColor
PRINT CHR$(URcorner) + " ";

' display line 2
LOCATE Xcoor + 1, Ycoor, 1
COLOR CalcBoxBorderColor
PRINT " " + CHR$(Vline);
COLOR CalcBoxTextColor
PRINT SPACE$(9);
PRINT "Hex";
PRINT SPACE$(11);
PRINT "Dec ";
COLOR CalcBoxBorderColor
PRINT CHR$(Vline) + " ";

' display line 3
LOCATE Xcoor + 2, Ycoor, 1
COLOR CalcBoxBorderColor
PRINT " " + CHR$(LLcorner) + CHR$(Hline);
COLOR CalcBoxTitleColor
PRINT "<tab>/<shift-tab>=switch";
COLOR CalcBoxBorderColor
PRINT CHR$(Hline) + CHR$(Hline) + CHR$(LRcorner) + " ";
RETURN

' move to box 1 (left box).
SelectBox1:
Box2$ = Box$ ' store right box.
Box$ = Box1$ ' reset to left box.

' reset editing parameters.
Xcoor1 = Xcoor + 1
Ycoor1 = Ycoor + 3
Column3 = Ycoor1
LineLength = 8

' display left box.
COLOR HexCalcForeground2, HexCalcBackground1
LOCATE Xcoor1, Ycoor1, 1
PRINT SPACE$(8);
LOCATE Xcoor1, Ycoor1, 1
PRINT Box$;
Ycoor1 = Ycoor1 + LEN(Box$)
LOCATE Xcoor1, Ycoor1, 1
RETURN

' move to box 2 (right box).
SelectBox2:
Box1$ = Box$ ' store left box.
Box$ = Box2$ ' reset to right box.

' reset editing parameters.
Xcoor1 = Xcoor + 1
Ycoor1 = Ycoor + 15
Column3 = Ycoor1
LineLength = 10

' display right box.
COLOR HexCalcForeground2, HexCalcBackground1
LOCATE Xcoor1, Ycoor1, 1
PRINT SPACE$(10);
LOCATE Xcoor1, Ycoor1, 1
PRINT Box$;
Ycoor1 = Ycoor1 + LEN(Box$)
LOCATE Xcoor1, Ycoor1, 1
RETURN

' calculates value of current box,
'  displays result in calculated form in opposite box.
PrintBoxes:
IF CalcBox = 1 THEN ' is in hex box.
    ' strip leading zeroes for signed bit conversion.
    TempBox$ = Box$
    DO
        IF LEFT$(TempBox$, 1) = "0" THEN
            TempBox$ = MID$(TempBox$, 2)
        ELSE
            EXIT DO
        END IF
    LOOP
    Value# = VAL("&H" + TempBox$) ' retreive decimal value.
    IF Value# < 0# THEN ' check twos-complement from signed value.
        IF LEN(TempBox$) = 8 THEN ' 8000 0000 to FFFF FFFF
            Value# = Value# + 4294967296#
        ELSE
            IF LEN(TempBox$) = 4 THEN ' 8000 to FFFF
                Value# = Value# + 65536#
            END IF
        END IF
    END IF
    ' display decimal value.
    Box2$ = MID$(STR$(Value#), 2)
    COLOR HexCalcForeground2, HexCalcBackground1
    LOCATE Xcoor1, Box2, 1
    PRINT SPACE$(10);
    LOCATE Xcoor1, Box2, 1
    PRINT Box2$;
END IF
IF CalcBox = 2 THEN ' is in decimal box.
    Value# = VAL(Box$) ' retreive hex value.
    IF Value# >= 4294967296# THEN ' check overflow.
        COLOR 15, 0
        LOCATE Xcoor1, Box1, 1
        Var$ = "overflow"
        PRINT Var$;
        Box1$ = "0"
    ELSE
        IF Value# > 2147483647# THEN ' force to twos-complement.
            Value# = Value# - 4294967296# ' 8000 0000 to FFFF FFFF
        END IF
        ' display hex value.
        Box1$ = HEX$(Value#)
        COLOR HexCalcForeground2, HexCalcBackground1
        LOCATE Xcoor1, Box1, 1
        PRINT SPACE$(8);
        LOCATE Xcoor1, Box1, 1
        PRINT Box1$;
    END IF
END IF
LOCATE Xcoor1, Ycoor1, 1
RETURN

' create help box.
HelpBox:
DO
    ' restore screen area.
    CALL RestoreScreen

    ' make help box
    COLOR HexCalcHelpForeground1, HexCalcHelpBackground2 ' yellow, blue
    LOCATE Xcoor2, Ycoor2, 1
    PRINT CHR$(ULcorner) + STRING$(39, Hline) + CHR$(URcorner);
    LOCATE Xcoor2 + 1, Ycoor2, 1
    PRINT CHR$(Vline);
    LOCATE Xcoor2 + 1, Ycoor2 + 40, 1
    PRINT CHR$(Vline);
    LOCATE Xcoor2 + 2, Ycoor2, 1
    PRINT CHR$(Vline);
    LOCATE Xcoor2 + 2, Ycoor2 + 40, 1
    PRINT CHR$(Vline);
    LOCATE Xcoor2 + 3, Ycoor2, 1
    PRINT CHR$(LLcorner) + STRING$(39, Hline) + CHR$(LRcorner);

    ' make help box title
    COLOR HexCalcHelpForeground1, HexCalcHelpBackground2 ' yellow, blue
    LOCATE Xcoor2, Ycoor2 + 10, 1
    PRINT " Additional Keys ";

    ' make help box contents
    COLOR HexCalcHelpForeground2, HexCalcHelpBackground1 ' white, black
    LOCATE Xcoor2 + 1, Ycoor2 + 1, 1
    PRINT "Ctrl-A = Move Up, Ctrl-B = Move Down   ";
    LOCATE Xcoor2 + 2, Ycoor2 + 1, 1
    PRINT "Ctrl-C = Move Left, Ctrl-D = Move Right";

    ' make help box title
    COLOR HexCalcHelpForeground1, HexCalcHelpBackground2 ' yellow, blue
    LOCATE Xcoor2 + 3, Ycoor2 + 9, 1
    PRINT "Press <esc> to Exit:";

    ' help box loop
    DO
        _LIMIT 100
        X$ = INKEYx$
        IF LEN(X$) = 2 THEN ' extended key
            SELECT CASE ASC(RIGHT$(X$, 1))
                CASE 76, 143 ' center
                    Xcoor2 = 4
                    Ycoor2 = 6
                    EXIT DO
                CASE 75, 115, 155 ' left/ctrl-left/alt-left
                    IF Ycoor2 > 1 THEN
                        Ycoor2 = Ycoor2 - 1
                        EXIT DO
                    END IF
                CASE 77, 116, 157 ' right/ctrl-right/alt-right
                    IF Ycoor2 < 40 THEN
                        Ycoor2 = Ycoor2 + 1
                        EXIT DO
                    END IF
                CASE 72, 141, 152 ' up/ctrl-up/alt-up
                    IF Xcoor2 > 2 THEN
                        Xcoor2 = Xcoor2 - 1
                        EXIT DO
                    END IF
                CASE 80, 145, 160 ' down/ctrl-dn/alt-dn
                    IF Status THEN
                        Z = 20
                    ELSE
                        Z = 21
                    END IF
                    IF Xcoor2 < Z THEN
                        Xcoor2 = Xcoor2 + 1
                        EXIT DO
                    END IF
            END SELECT
        END IF
        IF LEN(X$) = 1 THEN
            SELECT CASE ASC(X$)
                CASE 1 ' control-a (up)
                    IF Xcoor2 > 2 THEN
                        Xcoor2 = Xcoor2 - 1
                        EXIT DO
                    END IF
                CASE 2 ' control-b (down)
                    IF Status THEN
                        Z = 20
                    ELSE
                        Z = 21
                    END IF
                    IF Xcoor2 < Z THEN
                        Xcoor2 = Xcoor2 + 1
                        EXIT DO
                    END IF
                CASE 3 ' control-c (left)
                    IF Ycoor2 > 1 THEN
                        Ycoor2 = Ycoor2 - 1
                        EXIT DO
                    END IF
                CASE 4 ' control-d (right)
                    IF Ycoor2 < 40 THEN
                        Ycoor2 = Ycoor2 + 1
                        EXIT DO
                    END IF
                CASE 5 ' control-e
                    IF HexCalcHelpForeground1 = 15 THEN
                        HexCalcHelpForeground1 = 0
                    ELSE
                        HexCalcHelpForeground1 = HexCalcHelpForeground1 + 1
                    END IF
                    EXIT DO
                CASE 6 ' control-f
                    IF HexCalcHelpForeground2 = 15 THEN
                        HexCalcHelpForeground2 = 0
                    ELSE
                        HexCalcHelpForeground2 = HexCalcHelpForeground2 + 1
                    END IF
                    EXIT DO
                CASE 10 ' control-j
                    IF HexCalcHelpBackground1 = 7 THEN
                        HexCalcHelpBackground1 = 0
                    ELSE
                        HexCalcHelpBackground1 = HexCalcHelpBackground1 + 1
                    END IF
                    EXIT DO
                CASE 11 ' control-k
                    IF HexCalcHelpBackground2 = 7 THEN
                        HexCalcHelpBackground2 = 0
                    ELSE
                        HexCalcHelpBackground2 = HexCalcHelpBackground2 + 1
                    END IF
                    EXIT DO
                CASE 27 ' escape
                    RETURN
            END SELECT
        END IF
        IF Status THEN
            Z = 20
        ELSE
            Z = 21
        END IF
        MoveBox = 0
        X = MouseDriver
        IF MouseButton1 THEN
            ' hover over titlebar
            IF MouseX = Xcoor2 THEN
                IF MouseY >= Ycoor2 AND MouseY <= Ycoor2 + 40 THEN
                    ' store mouse XY during click
                    MouseTempX = MouseX
                    MouseTempY = MouseY
                    DO
                        X = MouseDriver
                        IF MouseX OR MouseY THEN ' drag
                            ' difference in mouse X
                            IF MouseX <> MouseTempX THEN
                                IF MouseX >= 2 AND MouseX <= Z THEN
                                    Xcoor2 = MouseX
                                    MouseTempX = MouseX
                                    MoveBox = -1
                                END IF
                            END IF
                            ' difference in mouse Y
                            IF MouseY <> MouseTempY THEN
                                MoveY = Ycoor2 + (MouseY - MouseTempY)
                                IF MoveY >= 1 AND MoveY <= 40 THEN
                                    Ycoor2 = MoveY
                                    MouseTempY = MouseY
                                    MoveBox = -1
                                END IF
                            END IF
                            ' move box
                            IF MoveBox THEN
                                EXIT DO
                            END IF
                        END IF
                    LOOP UNTIL MouseButton1 = 0
                    IF MoveBox THEN
                        MoveBox = 0
                        EXIT DO
                    END IF
                END IF
            END IF
        END IF
    LOOP
LOOP
RETURN
END SUB

' Displays ASCII character chart. v1.1a. PD 2006.
'  v1.1a modify unprintable char output for QB64.
SUB AsciiChart
' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

COLOR 7, 0
CLS
COLOR 15, 0
PRINT "ASCII Chart for characters 1 to 127:"
FOR Char = 1 TO 127
    DisplayChar$ = CHR$(Char)
    Var$ = MID$(STR$(Char), 2)
    Var1$ = RIGHT$("00" + Var$, 3) + " "
    COLOR 10, 0
    PRINT Var1$;
    COLOR 14, 0
    _CONTROLCHR OFF
    PRINT DisplayChar$;
    _CONTROLCHR ON
    PRINT " ";
    ' check full screen line
    IF (Char MOD 10) = 0 THEN PRINT
NEXT
PRINT
COLOR 15, 0
PRINT "Press a key:";
WHILE INKEYx$ = "": WEND
x = ClearKeyboard
COLOR 7, 0
PRINT
CLS
COLOR 15, 0
PRINT "ASCII Chart for characters 128 to 255:"
FOR Char = 128 TO 255
    DisplayChar$ = CHR$(Char)
    Var$ = MID$(STR$(Char), 2)
    Var1$ = RIGHT$("00" + Var$, 3) + " "
    COLOR 10, 0
    PRINT Var1$;
    COLOR 14, 0
    _CONTROLCHR OFF
    PRINT DisplayChar$;
    _CONTROLCHR ON
    PRINT " ";
    ' check full screen line
    IF ((Char - 7) MOD 10) = 0 THEN PRINT
NEXT
PRINT
COLOR 15, 0
PRINT "Press a key:";
WHILE INKEYx$ = "": WEND
x = ClearKeyboard
CALL RestoreScreen2 ' restore screen area.
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
EXIT SUB
END SUB

' Displays HEX character chart. v1.1a. PD 2006.
'  v1.1a modify unprintable char output for QB64.
SUB HexChart
' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

COLOR 7, 0
CLS
COLOR 15, 0
PRINT "HEX Chart for characters 1 to 127:"
FOR Char = 1 TO 127
    DisplayChar$ = CHR$(Char)
    Var$ = HEX$(Char)
    Var1$ = RIGHT$("0" + Var$, 2) + " "
    COLOR 10, 0
    PRINT Var1$;
    COLOR 14, 0
    _CONTROLCHR OFF
    PRINT DisplayChar$;
    _CONTROLCHR ON
    PRINT " ";
    ' check full screen line
    IF (Char MOD 10) = 0 THEN PRINT
NEXT
PRINT
COLOR 15, 0
PRINT "Press a key:";
WHILE INKEYx$ = "": WEND
x = ClearKeyboard
COLOR 7, 0
PRINT
CLS
COLOR 15, 0
PRINT "HEX Chart for characters 128 to 255:"
FOR Char = 128 TO 255
    DisplayChar$ = CHR$(Char)
    Var$ = HEX$(Char)
    Var1$ = RIGHT$("0" + Var$, 2) + " "
    COLOR 10, 0
    PRINT Var1$;
    COLOR 14, 0
    PRINT DisplayChar$; " ";
    ' check full screen line
    IF ((Char - 7) MOD 10) = 0 THEN PRINT
NEXT
PRINT
COLOR 15, 0
PRINT "Press a key:";
WHILE INKEYx$ = "": WEND
x = ClearKeyboard
CALL RestoreScreen2 ' restore screen area.
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
EXIT SUB
END SUB

' routine accepts an operation and performs on two values, string or numeric
SUB Arith (Token.Parsed$, Temp#, Temp2#)
IF Last.Token THEN
    SELECT CASE Token.Parsed$
        CASE "-"
            Temp# = Temp# - Temp2#
        CASE "+"
            Temp# = Temp# + Temp2#
        CASE "/"
            Temp# = Temp# / Temp2#
        CASE "\"
            Temp# = Temp# \ Temp2#
        CASE "*"
            Temp# = Temp# * Temp2#
        CASE "^"
            Temp# = Temp# ^ Temp2#
        CASE "<"
            Temp# = (Temp# < Temp2#)
        CASE ">"
            Temp# = (Temp# > Temp2#)
        CASE "<<"
            Temp# = Temp# * 2#
        CASE ">>"
            Temp# = Temp# \ 2#
        CASE "++"
            Temp# = Temp# + 1#
        CASE "--"
            Temp# = Temp# - 1#
        CASE "**"
            Temp# = Temp# ^ 2#
        CASE "//"
            Temp# = SQR(Temp#)
        CASE "="
            Temp# = (Temp# = Temp2#)
        CASE "<="
            Temp# = (Temp# <= Temp2#)
        CASE ">="
            Temp# = (Temp# >= Temp2#)
        CASE "<>"
            Temp# = (Temp# <> Temp2#)
        CASE "|"
            Temp# = Temp# OR Temp2#
        CASE "&"
            Temp# = Temp# AND Temp2#
        CASE "%"
            Temp# = Temp# MOD Temp2#
        CASE "~"
            Temp# = Temp# XOR Temp2#
        CASE "?"
            Temp# = Temp# IMP Temp2#
        CASE ":"
            Temp# = Temp# EQV Temp2#
        CASE "#"
            Temp# = NOT (Temp# OR Temp2#)
        CASE "@"
            Temp# = NOT (Temp# IMP Temp2#)
        CASE "`"
            Temp# = NOT (Temp# AND Temp2#)
    END SELECT
    EXIT SUB
END IF
SELECT CASE Token.Parsed$
    CASE "-"
        IF RIGHT$(Out3, LEN(Out4)) = Out4 THEN
            Out3 = LEFT$(Out3, LEN(Out3) - LEN(Out4))
        END IF
    CASE "+"
        Out3 = Out3 + Out4
    CASE "/", "\"
        IF LEN(Out3) > False AND LEN(Out4) > False THEN
            Imbedded = INSTR(Out3, Out4)
            WHILE Imbedded
                Out3 = LEFT$(Out3, Imbedded - 1) + MID$(Out3, Imbedded + LEN(Out4))
                Imbedded = INSTR(Out3, Out4)
            WEND
        END IF
    CASE "<"
        Last.Token = True
        Temp# = Out3 < Out4
    CASE ">"
        Last.Token = True
        Temp# = Out3 > Out4
    CASE "="
        Last.Token = True
        Temp# = Out3 = Out4
    CASE "<="
        Last.Token = True
        Temp# = Out3 <= Out4
    CASE ">="
        Last.Token = True
        Temp# = Out3 >= Out4
    CASE "<>"
        Last.Token = True
        Temp# = Out3 <> Out4
END SELECT
END SUB

SUB Assignment1 (S$, S%)
IF LEFT$(S$, 2) = "<<" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = Variables(S%) * 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ">>" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = Variables(S%) \ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "--" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = Variables(S%) - 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "++" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = Variables(S%) + 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "**" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = Variables(S%) ^ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "//" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Variables(S%) = SQR(Variables(S%))
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "-=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) - T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "+=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) + T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "/=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) / T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "\=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) \ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "*=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) * T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "^=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) ^ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 3) = "^-=" THEN
    Out2 = MID$(S$, 4)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) ^ (-T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "%=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) MOD T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "|=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) OR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "&=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) AND T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "~=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) XOR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "?=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) IMP T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ":=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = Variables(S%) EQV T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "#=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = NOT (Variables(S%) OR T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "@=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = NOT (Variables(S%) IMP T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "`=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = NOT (Variables(S%) AND T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 1) = "=" THEN
    Out2 = MID$(S$, 2)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
        Assign = True
        Variables(S%) = T#
    END IF
    EXIT SUB
END IF
END SUB

SUB Assignment2 (S$, S1%, S2%)
IF LEFT$(S$, 2) = "<<" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = Arrays(S1%, S2%) * 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ">>" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = Arrays(S1%, S2%) \ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "--" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = Arrays(S1%, S2%) - 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "++" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = Arrays(S1%, S2%) + 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "**" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = Arrays(S1%, S2%) ^ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "//" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
        Arrays(S1%, S2%) = SQR(Arrays(S1%, S2%))
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "-=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) - T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "+=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) + T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "/=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) / T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "\=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) \ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "*=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) * T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "^=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) ^ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 3) = "^-=" THEN
    Out2 = MID$(S$, 4)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) ^ (-T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "%=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) MOD T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "|=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) OR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "&=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) AND T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "~=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) XOR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "?=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) IMP T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ":=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = Arrays(S1%, S2%) EQV T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "#=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) OR T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "@=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) IMP T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "`=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) AND T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 1) = "=" THEN
    Out2 = MID$(S$, 2)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
        Assign = True
        Arrays(S1%, S2%) = T#
    END IF
END IF
END SUB

' parses single letter tokens
SUB Parse.Alphabetic1 (Temp#)
SELECT CASE UCASE$(Strng) ' test variable symbol
    CASE "A" TO "Z"
        Element = ASC(UCASE$(Strng)) - 64
        Element.Type$ = MID$(Out2, Token.Index, 1)
        SELECT CASE Element.Type$
            CASE "("
                CALL Get.Token
                CALL Get.Token
                CALL Parse1(Temp#)
                IF Strng = ")" THEN
                    Temp# = Arrays(Element, CINT(Temp#))
                    Last.Token = True
                ELSE
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
            CASE "["
                CALL Get.Token
                CALL Get.Token
                CALL Parse1(Temp#)
                IF Strng = "]" THEN
                    Temp# = Arrays(Element, CINT(Temp#))
                    Last.Token = True
                ELSE
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
            CASE "{"
                CALL Get.Token
                CALL Get.Token
                CALL Parse1(Temp#)
                IF Strng = "}" THEN
                    Temp# = Arrays(Element, CINT(Temp#))
                    Last.Token = True
                ELSE
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
            CASE ELSE
                Temp# = Variables(Element)
                Last.Token = True
        END SELECT
    CASE ELSE
        ERROR 92
END SELECT
END SUB

' parses two-letter tokens
SUB Parse.Alphabetic2 (Temp#)
SELECT CASE UCASE$(Strng)
    ' user defined functions
    CASE "FN"
        CALL ParseFN(Temp#)
    CASE "FZ"
        CALL ParseFZ(Temp#)
        ' functions w/o parameters
    CASE "PI" ' calculate PI
        Temp# = ATN(1#) * 4#
        Last.Token = True
    CASE "EX" ' calculate E
        Temp# = EXP(1#)
        Last.Token = True
        ' special case functions
    CASE "OR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# OR Temp#
        CALL Get.Token4
        Last.Token = True
    CASE ELSE
        IF MID$(Strng, 2, 1) = "$" THEN
            SELECT CASE MID$(Out2, Token.Index, 1)
                CASE "(", "[", "{"
                    Strng = "<bad string token>"
                    ERROR 92
            END SELECT
            SELECT CASE UCASE$(LEFT$(Strng, 1))
                CASE "A" TO "Z"
                    Element = ASC(UCASE$(LEFT$(Strng, 1))) - 64
                    Out3 = Strngs(Element)
                    Last.Token = False
            END SELECT
        ELSE
            ERROR 92
        END IF
END SELECT
END SUB

' parses FN(<x>)
SUB ParseFN (Temp#)
Var$ = LEFT$(Out2, Token.Index - 3)
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Var2$ = MID$(Out2, Token.Index)
    Out2 = Definitions(Func%)
    Out2 = LTRIM$(Out2)
    ' test for nul token
    IF Out2 = Nul THEN
        Strng = "<bad FN token>"
        ERROR 92
    END IF
    ' count recursing calls
    Recurse = Recurse + 1
    IF Recurse > Max.Recurse THEN
        Strng = "<bad FN recurse>"
        ERROR 92
    END IF
    ' recursively call functions
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(Temp#)
    ' test huge recursed FN
    IF LEN(Var$) + LEN(Var2$) >= 1024 THEN
        Var$ = Nul
        Var2$ = Nul
        Strng = "<bad FN string>"
        ERROR 92
    END IF
    Out2 = Var$ + Var2$
    Token.Index = LEN(Var$) + 1
    EXIT SUB
END IF
Strng = "<bad FN range>"
ERROR 92
END SUB

' parses FZ(<x>)
SUB ParseFZ (Temp#)
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Out3 = Definitions(Func%)
    Out3 = LTRIM$(Out3)
    Out3 = RTRIM$(Out3)
    IF Out3 = Nul THEN
        Strng = "<bad FZ token>"
        ERROR 92
    END IF
    Last.Token = False
    EXIT SUB
END IF
Strng = "<bad FZ range>"
ERROR 92
END SUB

' parses three and greater letter tokens
SUB Parse.Alphabetic3 (Temp#)
SELECT CASE UCASE$(Strng)
    ' constant functions
    CASE "AUTHOR$"
        Out3 = "Erik Jon Oredson AS. CSci"
        Last.Token = False
    CASE "EMAIL$"
        Out3 = "eoredson@gmail.com"
        Last.Token = False
    CASE "URL$"
        Out3 = "www.filegate.net"
        Last.Token = False
    CASE "VERSION$"
        Out3 = "1.0a"
        Last.Token = False
    CASE "RELEASE$"
        Out3 = "1.0"
        Last.Token = False
    CASE "PROGRAM$"
        Out3 = "what64"
        Last.Token = False
    CASE "PUBLISH$"
        Out3 = "10/15/2014"
        Last.Token = False
    CASE "COOKIE$"
        Out3 = "Thank you for all the cookies!"
        Last.Token = False
    CASE "ERROR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        ERROR CINT(Temp#)
        CALL Get.Token4
        Last.Token = True
        ' functions w/o parameters
    CASE "FALSE"
        Temp# = FalseD
        Last.Token = True
    CASE "TRUE"
        Temp# = TrueD
        Last.Token = True
    CASE "TIMER"
        Temp# = TIMER
        Last.Token = True
    CASE "RND"
        Temp# = RND
        Last.Token = True
        ' boolean functions
    CASE "AND"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# AND Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "MOD"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# MOD Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "NOR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = NOT (Number# OR Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "NON"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = NOT (Number# IMP Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "XAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = NOT (Number# AND Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "XOR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# XOR Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "IMP"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# IMP Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "EQV"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# EQV Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "NOT"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = NOT Temp#
        CALL Get.Token4
        Last.Token = True
        ' third-order boolean
    CASE "TAND"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number1# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Number2# = Temp#
        CALL Get.Token4
        SELECT CASE Number1#
            CASE -1, 0, 1
                SELECT CASE Number2#
                    CASE -1, 0, 1
                        Temp# = Number1# AND Number2#
                        Last.Token = True
                        EXIT SUB
                END SELECT
        END SELECT
        ERROR 5
    CASE "TOR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number1# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Number2# = Temp#
        CALL Get.Token4
        SELECT CASE Number1#
            CASE -1, 0, 1
                SELECT CASE Number2#
                    CASE -1, 0, 1
                        Temp# = Number1# OR Number2#
                        Last.Token = True
                        EXIT SUB
                END SELECT
        END SELECT
        ERROR 5
    CASE "TXOR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number1# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Number2# = Temp#
        CALL Get.Token4
        SELECT CASE Number1#
            CASE -1, 0, 1
                SELECT CASE Number2#
                    CASE -1, 0, 1
                        Temp# = Number1# XOR Number2#
                        Last.Token = True
                        EXIT SUB
                END SELECT
        END SELECT
        ERROR 5
        ' date\time functions
    CASE "CLOCK"
        Temp# = TIMER
        Last.Token = True
    CASE "DATE$"
        Out3 = DATE$
        Last.Token = False
    CASE "TIME$"
        Out3 = TIME$
        Last.Token = False
        ' conversion/calculation functions
    CASE "BIN$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = Nul
        IF Temp# >= FalseD THEN
            Digits = False
            DO
                IF 2 ^ (Digits + 1) > Temp# THEN
                    EXIT DO
                END IF
                Digits = Digits + 1
            LOOP
            FOR Power = Digits TO 0 STEP -1
                IF Temp# - 2 ^ Power >= 0 THEN
                    Temp# = Temp# - 2 ^ Power
                    Out3 = Out3 + "1"
                ELSE
                    Out3 = Out3 + "0"
                END IF
            NEXT
        END IF
        CALL Get.Token4
        Last.Token = False
    CASE "CHR$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = CHR$(CINT(Temp#))
        CALL Get.Token4
        Last.Token = False
    CASE "HEX$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = HEX$(Temp#)
        CALL Get.Token4
        Last.Token = False
    CASE "OCT$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = OCT$(Temp#)
        CALL Get.Token4
        Last.Token = False
    CASE "STR$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = STR$(Temp#)
        CALL Get.Token4
        Last.Token = False
    CASE "LCASE$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = LCASE$(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "UCASE$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = UCASE$(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "RTRIM$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = RTRIM$(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "LTRIM$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = LTRIM$(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "TRIM$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = LTRIM$(RTRIM$(Out3))
        CALL Get.Token4
        Last.Token = False
    CASE "SPACE$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = STRING$(Temp#, 32)
        CALL Get.Token4
        Last.Token = False
    CASE "MID$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp4# = Temp#
        IF Strng = "," THEN
            CALL Get.Token
            CALL Parse1(Temp#)
            Length# = Temp#
            Out3 = MID$(Out3, Temp4#, Length#)
        ELSE
            Out3 = MID$(Out3, Temp4#)
        END IF
        CALL Get.Token4
        Last.Token = False
    CASE "REMID$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL Get.Token
        CALL Parse1(Temp#)
        Start# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Length# = Temp#
        Out3 = LEFT$(Out3, Start# - 1) + MID$(Out3, Start# + Length#)
        CALL Get.Token4
        Last.Token = False
    CASE "DEMID$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        String1$ = Out3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF LEN(String1$) AND LEN(Out3) THEN
            Imbedded = INSTR(String1$, Out3)
            WHILE Imbedded
                String1$ = LEFT$(String1$, Imbedded - 1) + MID$(String1$, Imbedded + LEN(Out3))
                Imbedded = INSTR(String1$, Out3)
            WEND
        END IF
        Out3 = String1$
        CALL Get.Token4
        Last.Token = False
    CASE "REPLACE$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        String1$ = Out3
        CALL Get.Token
        CALL Parse1(Temp#)
        String2$ = Out3
        CALL Get.Token
        CALL Parse1(Temp#)
        String3$ = Out3
        IF String1$ = Nul OR String2$ = Nul OR String2$ = String3$ THEN
            Out3 = String1$
            EXIT SUB
        END IF
        Temp1 = 1
        DO
            IF String1$ = Nul THEN
                EXIT DO
            END IF
            Var1 = INSTR(Temp1, String1$, String2$)
            IF Var1 = False THEN
                EXIT DO
            END IF
            String1$ = LEFT$(String1$, Var1 - 1) + String3$ + MID$(String1$, Var1 + LEN(String2$))
            Temp1 = Temp1 + LEN(String3$)
        LOOP
        Out3 = String1$
        CALL Get.Token4
        Last.Token = False
    CASE "LEFT$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = LEFT$(Out3, Temp#)
        CALL Get.Token4
        Last.Token = False
    CASE "RIGHT$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = RIGHT$(Out3, Temp#)
        CALL Get.Token4
        Last.Token = False
    CASE "STRING$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Out3 = STRING$(Number#, Temp#)
        ELSE
            Out3 = STRING$(Number#, Out3)
        END IF
        CALL Get.Token4
        Last.Token = False
    CASE "ABS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ABS(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "ASC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ASC(Out3)
        CALL Get.Token4
        Last.Token = True
    CASE "EXP"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = EXP(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "FBN" ' calculate xth fibonachi value
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Fibo.A# = 1
        Fibo.B# = 1
        IF Temp# >= 1 THEN
            FOR Fibo.Number# = 1 TO Temp# - 1
                Fibo.Temp# = Fibo.B#
                Fibo.B# = Fibo.A# + Fibo.B#
                Fibo.A# = Fibo.Temp#
            NEXT
        END IF
        Temp# = Fibo.B#
        CALL Get.Token4
        Last.Token = True
    CASE "FCT"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Factorial# = 1
        Factorial.Count# = INT(Temp#)
        IF Factorial.Count# >= 1 THEN
            FOR Factorial.Count# = 1 TO Temp#
                Factorial# = Factorial# * Factorial.Count#
            NEXT
        END IF
        Temp# = Factorial#
        CALL Get.Token4
        Last.Token = True
    CASE "FIX"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = FIX(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "INSTR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Start# = Temp#
            CALL Get.Token
            CALL Parse1(Temp#)
            Stored.String$ = Out3
            CALL Get.Token
            CALL Parse1(Temp#)
        ELSE
            Start# = 1#
            Stored.String$ = Out3
            CALL Get.Token
            CALL Parse1(Temp#)
        END IF
        Temp# = INSTR(Start#, Stored.String$, Out3)
        CALL Get.Token4
        Last.Token = True
    CASE "INT"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = INT(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "LEN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LEN(Out3)
        CALL Get.Token4
        Last.Token = True
    CASE "LOG"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "ISEVEN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (Temp# MOD 2#) = 0
        CALL Get.Token4
        Last.Token = True
    CASE "ISODD"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (Temp# MOD 2#) = 1
        CALL Get.Token4
        Last.Token = True
    CASE "ISPRM" ' calculate x is prime
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Prime# = INT(Temp#)
        Prime.Flag = True
        IF Prime# > 2# THEN
            FOR Loop.Count# = 2# TO INT(SQR(Prime#))
                IF Prime# / Loop.Count# = INT(Prime# / Loop.Count#) THEN
                    Prime.Flag = False
                    EXIT FOR
                END IF
            NEXT
        END IF
        IF Prime.Flag = False THEN
            Temp# = FalseD
        ELSE
            Temp# = TrueD
        END IF
        CALL Get.Token4
        Last.Token = True
    CASE "PRF" ' calculate xth perfect number
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL Get.Token4
        Last.Token = True
        Perfect# = INT(Temp#)
        SELECT CASE Perfect#
            CASE 1#
                X# = 1#
            CASE 2#
                X# = 2#
            CASE 3#
                X# = 4#
            CASE 4#
                X# = 6#
            CASE 5#
                X# = 12#
            CASE 6#
                X# = 16#
            CASE 7#
                X# = 18#
                'CASE 8# ' overflows beyond double precision
                '  X# = 30#
            CASE ELSE
                ERROR 130
        END SELECT
        ' equation for perfect number
        Temp# = 2 ^ X# * (2 ^ (X# + 1) - 1)
        Last.Token = True
    CASE "PRM" ' calculate xth prime
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Prime.Number# = INT(Temp#)
        Prime# = 1
        IF Prime.Number# > FalseD THEN
            Prime.Counter# = False
            DO
                Prime# = Prime# + 1
                Prime.Flag = False
                FOR Factor# = 2# TO INT(SQR(Prime#))
                    IF Prime# / Factor# = INT(Prime# / Factor#) THEN
                        Prime.Flag = True
                        EXIT FOR
                    END IF
                NEXT
                IF Prime.Flag = False THEN
                    Prime.Counter# = Prime.Counter# + 1
                END IF
                IF Prime.Counter# = Prime.Number# THEN
                    EXIT DO
                END IF
            LOOP
        END IF
        Temp# = Prime#
        CALL Get.Token4
        Last.Token = True
    CASE "SGN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = SGN(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "SQR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = SQR(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "CBR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Temp# ^ (1# / 3#)
        CALL Get.Token4
        Last.Token = True
    CASE "GCM"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number1# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Number2# = Temp#
        ' find last factor which divides into both numbers
        Factor1# = 1
        Number3# = Number1#
        FOR Factor2# = 2 TO Number1#
            DO
                IF Number3# / Factor2# = Number3# \ Factor2# THEN
                    IF Number2# / Factor2# = Number2# \ Factor2# THEN
                        IF Factor2# > Factor1# THEN
                            Factor1# = Factor2#
                        END IF
                    END IF
                    Number3# = Number3# / Factor2#
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
        Temp# = Factor1#
        CALL Get.Token4
        Last.Token = True
    CASE "LCD"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number1# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Number2# = Temp#
        ' find first factor which divides into both numbers
        Factor1# = 0
        FOR Factor2# = 2 TO Number1#
            IF Number1# / Factor2# = Number1# \ Factor2# THEN
                IF Number2# / Factor2# = Number2# \ Factor2# THEN
                    Factor1# = Factor2#
                    EXIT FOR
                END IF
            END IF
        NEXT
        IF Factor1# = 0# THEN
            Factor1# = 1#
        END IF
        IF Factor1# = Number2# THEN
            Factor1# = 1#
        END IF
        Temp# = Factor1#
        CALL Get.Token4
        Last.Token = True
    CASE "NTH"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# ^ (1# / Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "PWR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = Number# ^ Temp#
        CALL Get.Token4
        Last.Token = True
    CASE "VAL"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = VAL(Out3)
        CALL Get.Token4
        Last.Token = True

        REM file functions..

    CASE "FILEEXISTS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = _FILEEXISTS(Out3)
        CALL Get.Token4
        Last.Token = True
    CASE "DIREXISTS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = _DIREXISTS(Out3)
        CALL Get.Token4
        Last.Token = True
    CASE "DRIVEEXISTS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Var$ = Nul
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Var$ = CHR$(Var + 64)
            END IF
        ELSE
            Var = ASC(UCASE$(Out3))
            IF Var >= 65 AND Var <= 90 THEN
                Var$ = CHR$(Var)
                Var = Var - 64
            END IF
        END IF
        Temp# = 0#
        IF LEN(Var$) THEN
            Out3 = Var$
            IF DRIVEEXISTS(Var) = 0 THEN
                Temp# = -1#
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        CALL Get.Token4
        Last.Token = True
    CASE "DRIVETYPE"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Var$ = Nul
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Var$ = CHR$(Var + 64)
            END IF
        ELSE
            Var = ASC(UCASE$(Out3))
            IF Var >= 65 AND Var <= 90 THEN
                Var$ = CHR$(Var)
                Var = Var - 64
            END IF
        END IF
        Out3 = Nul
        IF LEN(Var$) THEN
            IF DRIVEEXISTS(Var) = 0 THEN
                Out3 = DriveType
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        CALL Get.Token4
        Last.Token = False
    CASE "AMBIG$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = AmbiguateFile$(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "CURDIR$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Out3 = _CWD$
        CALL Get.Token4
        Last.Token = False
    CASE "ENVIRON$"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            Out3 = ENVIRON$(Var)
        ELSE
            Out3 = ENVIRON$(Out3)
        END IF
        CALL Get.Token4
        Last.Token = False
    CASE "FILES"
        Parse.Type = -1
        CALL ListFiles("*.*", True)
        LastToken = False
    CASE "DIRS"
        Parse.Type = -1
        CALL ListFiles("*.*", False)
        LastToken = False
    CASE "DRIVES"
        CALL ListDrives
        Parse.Type = -1
        Last.Token = True
    CASE "FILEDATE"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL FileDate(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "FILETIME"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL FileTime(Out3)
        CALL Get.Token4
        Last.Token = False
    CASE "FILESIZE"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL GetFileSize(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "FILEATTR"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL FileAttr(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "FILEBITS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        CALL FileAttr(Temp#)
        CALL Get.Token4
        Out3 = SPACE$(5)
        IF (Temp# AND &H20) = &H20 THEN
            MID$(Out3, 1, 1) = "A" ' archive
        END IF
        IF (Temp# AND &H10) = &H10 THEN
            MID$(Out3, 2, 1) = "D" ' directory
        END IF
        IF (Temp# AND &H4) = &H4 THEN
            MID$(Out3, 3, 1) = "S" ' system
        END IF
        IF (Temp# AND &H2) = &H2 THEN
            MID$(Out3, 4, 1) = "H" ' hidden
        END IF
        IF (Temp# AND &H1) = &H1 THEN
            MID$(Out3, 5, 1) = "R" ' read-only
        END IF
        Last.Token = False
    CASE "VLABEL"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Out3 = CHR$(Var + 64)
                Var$ = Out3
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        ELSE
            Var$ = UCASE$(Out3)
            IF Var$ >= "A" AND Var$ <= "Z" THEN
                Out3 = Var$
                Var = ASC(Out3) - 64
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        END IF
        Last.Token = False
        IF LEN(Var$) THEN
            Temp# = 0#
            IF DRIVEEXISTS(Var) = 0 THEN
                CALL Vlabel(Out3)
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        CALL Get.Token4
    CASE "VSERIAL"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Out3 = CHR$(Var + 64)
                Var$ = Out3
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        ELSE
            Var$ = UCASE$(Out3)
            IF Var$ >= "A" AND Var$ <= "Z" THEN
                Out3 = Var$
                Var = ASC(Var$) - 64
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        END IF
        Last.Token = False
        IF LEN(Var$) THEN
            Temp# = 0#
            IF DRIVEEXISTS(Var) = 0 THEN
                CALL Vserial(Out3)
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        CALL Get.Token4
    CASE "VTYPE"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Out3 = CHR$(Var + 64)
                Var$ = Out3
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        ELSE
            Var$ = UCASE$(Out3)
            IF Var$ >= "A" AND Var$ <= "Z" THEN
                Out3 = Var$
                Var = ASC(Out3) - 64
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        END IF
        Last.Token = False
        IF LEN(Var$) THEN
            Temp# = 0#
            IF DRIVEEXISTS(Var) = 0 THEN
                CALL Vtype(Out3)
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        CALL Get.Token4
    CASE "DRV"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Out3 = CHR$(Var + 64)
                Var$ = Out3
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        ELSE
            Var$ = UCASE$(Out3)
            IF Var$ >= "A" AND Var$ <= "Z" THEN
                Out3 = Var$
                Var = ASC(Out3) - 64
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        END IF
        IF LEN(Var$) THEN
            Temp# = 0#
            IF DRIVEEXISTS(Var) = 0 THEN
                CALL FreeSpace(Out3)
                Temp# = INT(VAL(Out3))
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        Last.Token = True
        CALL Get.Token4
    CASE "TDRV"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        IF Last.Token THEN
            Var = CINT(Temp#)
            IF Var >= 1 AND Var <= 26 THEN
                Out3 = CHR$(Var + 64)
                Var$ = Out3
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        ELSE
            Var$ = UCASE$(Out3)
            IF Var$ >= "A" AND Var$ <= "Z" THEN
                Out3 = Var$
                Var = ASC(Out3) - 64
            ELSE
                Out3 = Nul
                Var$ = Nul
            END IF
        END IF
        IF LEN(Out3) THEN
            Temp# = 0#
            IF DRIVEEXISTS(Var) = 0 THEN
                CALL TotalSpace(Out3)
                Temp# = INT(VAL(Out3))
            ELSE
                ERROR 71
            END IF
        ELSE
            ERROR 71
        END IF
        Last.Token = True
        CALL Get.Token4

        REM Basic trigonometric functions..

    CASE "ATN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "COS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = COS(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "COT"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = COS(Temp#) / SIN(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "CSC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = 1# / SIN(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "SEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = 1# / COS(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "SIN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = SIN(Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE "TAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = TAN(Temp#)
        CALL Get.Token4
        Last.Token = True

        REM Remaining nonintrinsic trigonmetric functions..

        'Inverse Cosine
    CASE "ARCCOS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(-Temp# / SQR(-Temp# * Temp# + 1)) + 2 * ATN(1)
        CALL Get.Token4
        Last.Token = True
        'Inverse Cosecant
    CASE "ARCCOSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(Temp# / SQR(Temp# * Temp# - 1)) + (SGN(Temp#) - 1) * (2 * ATN(1))
        CALL Get.Token4
        Last.Token = True
        'Inverse Cotangent
    CASE "ARCCOTAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(Temp#) + 2 * ATN(1)
        CALL Get.Token4
        Last.Token = True
        'Inverse Secant
    CASE "ARCSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(Temp# / SQR(Temp# * Temp# - 1)) + SGN((Temp#) - 1) * (2 * ATN(1))
        CALL Get.Token4
        Last.Token = True
        'Inverse Sine
    CASE "ARCSIN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ATN(Temp# / SQR(-Temp# * Temp# + 1))
        CALL Get.Token4
        Last.Token = True
        'Convert Degrees to Radians
    CASE "DEGTORAD"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = ((ATN(1) * 4) / 180) * Temp#
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Cosine
    CASE "HARCCOS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG(Temp# + SQR(Temp# * Temp# - 1))
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Cosecant
    CASE "HARCCOSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG((SGN(Temp#) * SQR(Temp# * Temp# + 1) + 1) / Temp#)
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Cotangent
    CASE "HARCCOTAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG((Temp# + 1) / (Temp# - 1)) / 2
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Secant
    CASE "HARCSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG((SQR(-Temp# * Temp# + 1) + 1) / Temp#)
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Sine
    CASE "HARCSIN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG(Temp# + SQR(Temp# * Temp# + 1))
        CALL Get.Token4
        Last.Token = True
        'Inverse Hyperbolic Tangent
    CASE "HARCTAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = LOG((1 + Temp#) / (1 - Temp#)) / 2
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Cosine
    CASE "HCOS"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (EXP(Temp#) + EXP(-Temp#)) / 2
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Cosecant
    CASE "HCOSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = 2 / (EXP(Temp#) - EXP(-Temp#))
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Cotangent
    CASE "HCOTAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (EXP(Temp#) + EXP(-Temp#)) / (EXP(Temp#) - EXP(-Temp#))
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Secant
    CASE "HSEC"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = 2 / (EXP(Temp#) + EXP(-Temp#))
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Sine
    CASE "HSIN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (EXP(Temp#) - EXP(-Temp#)) / 2
        CALL Get.Token4
        Last.Token = True
        'Hyperbolic Tangent
    CASE "HTAN"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (EXP(Temp#) - EXP(-Temp#)) / (EXP(Temp#) + EXP(-Temp#))
        CALL Get.Token4
        Last.Token = True
        'Convert Radians to Degrees
    CASE "RADTODEG"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = (180 / (ATN(1) * 4)) * Temp#
        CALL Get.Token4
        Last.Token = True
        ' Hypotenuse
    CASE "HYPOT"
        CALL Get.Token3
        CALL Get.Token
        CALL Parse1(Temp#)
        Number# = Temp#
        CALL Get.Token
        CALL Parse1(Temp#)
        Temp# = SQR(Number# * Number# + Temp# * Temp#)
        CALL Get.Token4
        Last.Token = True
    CASE ELSE
        ERROR 92
END SELECT
END SUB

' beginning of the command/equation parser
SUB Enter.Equate
Strng = Nul
Var2$ = Out2
Out2 = LTRIM$(Out2)
IF Out2 = Nul THEN
    EXIT SUB
END IF
Store.Input$ = Out2
Assign = False
GOSUB Assignment
IF Assign THEN
    EXIT SUB
END IF
Out2 = Store.Input$
D$ = UCASE$(Out2)
IF LEFT$(D$, 3) = "MID" THEN
    GOSUB Assign.Mid.String
    IF Assign THEN
        EXIT SUB
    END IF
END IF
IF LEFT$(D$, 4) = "LEFT" THEN
    GOSUB Assign.Left.String
    IF Assign THEN
        EXIT SUB
    END IF
END IF
IF LEFT$(D$, 5) = "RIGHT" THEN
    GOSUB Assign.Right.String
    IF Assign THEN
        EXIT SUB
    END IF
END IF
IF LEFT$(D$, 5) = "DEFFN" THEN
    GOSUB Define.Function
    IF Assign THEN
        EXIT SUB
    END IF
END IF
Out2 = Store.Input$
Last.Token = False
Out3 = Nul
Out4 = Nul
Temp# = False
Token.Index = 1
CALL Get.Token
CALL Parse1(T#)

' parse leftover tokens
WHILE LEN(Strng)
    CALL Parse1(X#)
    ' toss away any unwanted tokens
    IF Last.Token = 3 THEN
        Strng = "<extra token>"
        ERROR 92
    END IF
WEND

' display output result
IF Parse.Type > 0 THEN
    IF Parse.Type = 1 THEN
        PRINT #StdoutHandle, Var2$ + " Equals: ";
    END IF
    IF Last.Token = False THEN
        PRINT #StdoutHandle, Out3
    ELSE
        PRINT #StdoutHandle, LTRIM$(STR$(T#))
    END IF
END IF
EXIT SUB

Assignment:
Assign = False
Temp1$ = UCASE$(LEFT$(Out2, 1))
IF Temp1$ = "?" THEN
    Temp1$ = MID$(Out2, 2)
    Temp2$ = LEFT$(Temp1$, 1)
    IF Temp2$ = "=" THEN
        Temp1$ = MID$(Temp1$, 2)
        Out2 = Temp1$
        CALL Equate(Temp4#)
        Assign = True
        EXIT SUB
    END IF
END IF
IF Temp1$ >= "A" AND Temp1$ <= "Z" THEN
    Variable = ASC(Temp1$) - 64
    Temp1$ = MID$(Out2, 2)
    Temp1$ = LTRIM$(Temp1$)
    CALL Assignment1(Temp1$, Variable)
    IF Assign THEN
        GOSUB Assign.End
        RETURN
    END IF
    IF LEFT$(Temp1$, 1) = "$" THEN
        Temp1$ = MID$(Temp1$, 2)
        Temp1$ = LTRIM$(Temp1$)
        IF LEFT$(Temp1$, 1) = "(" THEN
            Strng = "<bad string token>"
            ERROR 92
        END IF
        IF LEFT$(Temp1$, 1) = "[" THEN
            Strng = "<bad string token>"
            ERROR 92
        END IF
        IF LEFT$(Temp1$, 1) = "{" THEN
            Strng = "<bad string token>"
            ERROR 92
        END IF
        IF LEFT$(Temp1$, 1) = "=" THEN
            Out2 = MID$(Temp1$, 2)
            Out2 = LTRIM$(Out2)
            Token.Index = 1
            CALL Get.Token
            CALL Parse1(Temp3#)
            IF Last.Token = False THEN
                Strngs(Variable) = Out3
                Assign = True
                GOSUB Assign.End
                RETURN
            END IF
        END IF
    END IF
    IF LEFT$(Temp1$, 1) = "(" THEN
        Out2 = MID$(Temp1$, 2)
        Out2 = LTRIM$(Out2)
        Last.Token = False
        Token.Index = 1
        CALL Get.Token
        CALL Parse1(Temp3#)
        IF Last.Token = True THEN
            Element = CINT(Temp3#)
            Temp1$ = MID$(Out2, Token.Index)
            Temp1$ = LTRIM$(Temp1$)
            CALL Assignment2(Temp1$, Variable, Element)
            IF Assign THEN
                GOSUB Assign.End
                RETURN
            END IF
        END IF
    END IF
END IF
RETURN

Assign.Mid.String:
Out2 = MID$(Out2, 4)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
        Variable = ASC(V$) - 64
        IF MID$(Out2, 2, 1) = "$" THEN
            Out2 = MID$(Out2, 4)
            Last.Token = False
            Token.Index = 1
            CALL Get.Token
            CALL Parse1(T#)
            Start = CINT(T#)
            IF Strng = "," THEN
                CALL Get.Token
                CALL Parse1(T#)
                Length = CINT(T#)
                Out2 = MID$(Out2, Token.Index)
                Out2 = LTRIM$(Out2)
                IF LEFT$(Out2, 1) = "=" THEN
                    Out2 = MID$(Out2, 2)
                    Out2 = LTRIM$(Out2)
                    Token.Index = 1
                    CALL Get.Token
                    CALL Parse1(T#)
                    IF Last.Token = False THEN
                        Assign = True
                        IF Strngs(Variable) = Nul THEN
                            IF Start = 1 THEN
                                IF LEN(Out3) = Length THEN
                                    Strngs(Variable) = Out3
                                    GOSUB Assign.End
                                    RETURN
                                END IF
                            END IF
                            Strng = "<bad string>"
                            ERROR 92
                            RETURN
                        END IF
                        MID$(Strngs(Variable), Start, Length) = Out3
                        GOSUB Assign.End
                        RETURN
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF
RETURN

Assign.Left.String:
Out2 = MID$(Out2, 5)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
        Variable = ASC(V$) - 64
        IF MID$(Out2, 2, 1) = "$" THEN
            Out2 = MID$(Out2, 4)
            Last.Token = False
            Token.Index = 1
            CALL Get.Token
            CALL Parse1(T#)
            Start = CINT(T#)
            Out2 = MID$(Out2, Token.Index)
            Out2 = LTRIM$(Out2)
            IF LEFT$(Out2, 1) = "=" THEN
                Assign = True
                Out2 = MID$(Out2, 2)
                Out2 = LTRIM$(Out2)
                Token.Index = 1
                CALL Get.Token
                CALL Parse1(T#)
                IF Last.Token = False THEN
                    Assign = True
                    IF Strngs(Variable) = Nul THEN
                        IF LEN(Out3) = Start THEN
                            Strngs(Variable) = Out3
                            GOSUB Assign.End
                            RETURN
                        END IF
                        Strng = "<bad string>"
                        ERROR 92
                        RETURN
                    END IF
                    MID$(Strngs(Variable), 1, Start) = Out3
                    GOSUB Assign.End
                    RETURN
                END IF
            END IF
        END IF
    END IF
END IF
RETURN

Assign.Right.String:
Out2 = MID$(Out2, 6)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
        Variable = ASC(V$) - 64
        IF MID$(Out2, 2, 1) = "$" THEN
            Out2 = MID$(Out2, 4)
            Last.Token = False
            Token.Index = 1
            CALL Get.Token
            CALL Parse1(T#)
            Start = CINT(T#)
            Out2 = MID$(Out2, Token.Index)
            Out2 = LTRIM$(Out2)
            IF LEFT$(Out2, 1) = "=" THEN
                Assign = True
                Out2 = MID$(Out2, 2)
                Out2 = LTRIM$(Out2)
                Token.Index = 1
                CALL Get.Token
                CALL Parse1(T#)
                Length = LEN(Strngs(Variable)) - Start + 1
                IF Last.Token = False THEN
                    Assign = True
                    IF Length > False THEN
                        MID$(Strngs(Variable), Length, Start) = Out3
                        GOSUB Assign.End
                        RETURN
                    ELSE
                        IF Strngs(Variable) = Nul THEN
                            IF LEN(Out3) = Start THEN
                                Strngs(Variable) = Out3
                                GOSUB Assign.End
                                RETURN
                            END IF
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF
RETURN

Assign.End:
IF LEN(Strng) THEN
    CALL Parse1(X#)
    Strng = "<extra token>"
    ERROR 92
END IF
RETURN

Define.Function:
Out2 = MID$(Out2, 6)
Out2 = LTRIM$(Out2)
Token.Index = 1
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Out2 = MID$(Out2, Token.Index)
    Out2 = LTRIM$(Out2)
    IF LEFT$(Out2, 1) = "=" THEN
        Out2 = MID$(Out2, 2)
        Definitions(Func%) = Out2
        Assign = True
        RETURN
    END IF
END IF
ERROR 145
END SUB

' returns next token in string form,
' increments pointer to next token,
' determines token type.
SUB Get.Token
Strng = Nul ' reset token
Token = False ' reset token type
' locate symbols
CALL Get.Token2(Token.Exists)
IF Token.Exists THEN
    EXIT SUB
END IF
IF Token.Index > LEN(Out2) THEN
    Strng = Nul
    Token = False
    Token.Index = Token.Index + 1
    EXIT SUB
END IF
' locate expression symbol
Token.Element$ = MID$(Out2, Token.Index, 1)
IF INSTR(Token.List, Token.Element$) THEN
    Token = 1 ' store token type
    Strng = Token.Element$ ' store token
    Token.Index = Token.Index + 1 ' increment pointer
    EXIT SUB
END IF
' locate expression is number
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ >= "0" AND Token.Element$ <= "9" THEN
    ' increment token until token is other than number
    DO
        IF LEN(Token.Element$) = False THEN
            EXIT DO
        END IF
        IF INSTR(Token.List, Token.Element$) THEN
            EXIT DO
        END IF
        Strng = Strng + Token.Element$
        Token.Index = Token.Index + 1
        Token.Element$ = MID$(Out2, Token.Index, 1)
    LOOP
    Token = 2 ' store token type
    EXIT SUB
END IF
' locate expression is number beginning with decimal
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ = "." THEN
    ' increment token until token is other than number
    DO
        IF LEN(Token.Element$) = False THEN
            EXIT DO
        END IF
        IF INSTR(Token.List, Token.Element$) THEN
            EXIT DO
        END IF
        Strng = Strng + Token.Element$
        Token.Index = Token.Index + 1
        Token.Element$ = MID$(Out2, Token.Index, 1)
    LOOP
    Token = 2 ' store token type
    EXIT SUB
END IF
' locate expression is alphabetic
Token.Element$ = UCASE$(MID$(Out2, Token.Index, 1))
IF Token.Element$ >= "A" AND Token.Element$ <= "Z" THEN
    ' increment token until token is other than alphabetic
    DO
        IF LEN(Token.Element$) = False THEN
            EXIT DO
        END IF
        IF INSTR(Token.List, Token.Element$) THEN
            EXIT DO
        END IF
        Strng = Strng + Token.Element$
        Token.Index = Token.Index + 1
        Token.Element$ = UCASE$(MID$(Out2, Token.Index, 1))
    LOOP
    Token = 3 ' store token type
    EXIT SUB
END IF
Token = False
Strng = Token.Element$
Token.Index = Token.Index + 1
END SUB

' gets two-symbol tokens,
' increments pointer to next token.
SUB Get.Token2 (Token.Exists)
Token.Exists = False
Stored.Token$ = Out2
' locate expression symbols
Next.Token$ = MID$(Stored.Token$, Token.Index, 2)
SELECT CASE Next.Token$
    CASE "<<"
        Token = 1 ' store token type
        Strng = "<<" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE ">>"
        Token = 1 ' store token type
        Strng = ">>" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "--"
        Token = 1 ' store token type
        Strng = "--" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "++"
        Token = 1 ' store token type
        Strng = "++" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "**"
        Token = 1 ' store token type
        Strng = "**" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "//"
        Token = 1 ' store token type
        Strng = "//" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE ">=", "=>"
        Token = 1 ' store token type
        Strng = ">=" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "<=", "=<"
        Token = 1 ' store token type
        Strng = "<=" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
    CASE "<>", "><"
        Token = 1 ' store token type
        Strng = "<>" ' store token
        Token.Index = Token.Index + 2 ' increment pointer
        Token.Exists = True
END SELECT
END SUB

' verifies next token is opening parenthesis.
SUB Get.Token3
Strng = Nul
Token = False
IF Token.Index > LEN(Out2) THEN
    Strng = "<missing opening parenthesis>"
    ERROR 92
END IF
' locate parenthesis symbol
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ = "(" THEN
    Token = 1 ' store token type
    Strng = Token.Element$ ' store token
    Token.Index = Token.Index + 1 ' increment pointer
    EXIT SUB
END IF
Strng = "<missing opening parenthesis>"
ERROR 92
END SUB

' verifies last token is closing parenthesis.
SUB Get.Token4
' locate parenthesis symbol
IF Strng = ")" THEN
    EXIT SUB
END IF
Strng = "<missing closing parenthesis>"
ERROR 92
END SUB

' process a number
SUB Parse.Numeric (Temp#)
SELECT CASE RIGHT$(UCASE$(Strng), 1) ' compare character after number
    CASE "Q" ' perfect number suffix
        FOR Digit = 1 TO LEN(Strng) - 1
            OutX$ = UCASE$(MID$(Strng, Digit, 1))
            SELECT CASE OutX$
                CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
                    ' nul
                CASE ELSE
                    ERROR 130
            END SELECT
        NEXT
        Temp# = CDBL(VAL(LEFT$(Strng, LEN(Strng) - 1)))
        Perfect# = 0#
        Perfect.Number# = INT(Temp#)
        IF ABS(Perfect.Number#) > 7# THEN
            ERROR 130
        END IF
        IF Perfect.Number# > FalseD THEN
            SELECT CASE Perfect.Number#
                CASE 1#
                    X# = 1#
                CASE 2#
                    X# = 2#
                CASE 3#
                    X# = 4#
                CASE 4#
                    X# = 6#
                CASE 5#
                    X# = 12#
                CASE 6#
                    X# = 16#
                CASE 7#
                    X# = 18#
                    'CASE 8# ' overflows beyond double precision
                    '  X# = 30#
                CASE ELSE
                    ERROR 130
            END SELECT
            ' equation for perfect number
            Perfect# = 2 ^ X# * (2 ^ (X# + 1) - 1)
        END IF
        Temp# = Perfect#
    CASE "P" ' prime number suffix
        FOR Digit = 1 TO LEN(Strng) - 1
            OutX$ = UCASE$(MID$(Strng, Digit, 1))
            SELECT CASE OutX$
                CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
                    ' nul
                CASE ELSE
                    ERROR 130
            END SELECT
        NEXT
        Temp# = CDBL(VAL(LEFT$(Strng, LEN(Strng) - 1)))
        Prime.Number# = INT(Temp#)
        Prime# = 1
        IF Prime.Number# > False THEN
            Prime.Counter# = False
            DO
                Prime# = Prime# + 1
                Prime.Flag = False
                FOR Factor# = 2# TO INT(SQR(Prime#))
                    IF Prime# / Factor# = INT(Prime# / Factor#) THEN
                        Prime.Flag = True
                        EXIT FOR
                    END IF
                NEXT
                IF Prime.Flag = False THEN
                    Prime.Counter# = Prime.Counter# + 1
                END IF
                IF Prime.Counter# = Prime.Number# THEN
                    EXIT DO
                END IF
            LOOP
        END IF
        Temp# = Prime#
    CASE "F" ' factorial
        Temp2 = False
        Temp3 = LEN(Strng)
        DO
            IF UCASE$(MID$(Strng, Temp3, 1)) <> "F" THEN
                EXIT DO
            END IF
            Temp2 = Temp2 + 1
            Temp3 = Temp3 - 1
        LOOP
        FOR Decimal.Digit = 1 TO Temp3
            OutX$ = UCASE$(MID$(Strng, Decimal.Digit, 1))
            SELECT CASE OutX$
                CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
                    ' nul
                CASE ELSE
                    ERROR 130
            END SELECT
        NEXT
        Temp4# = VAL(LEFT$(Strng, Temp3))
        FOR Number = 1 TO Temp2
            Temp# = 1#
            FOR Factorial# = 2# TO Temp4#
                Temp# = Temp# * Factorial#
            NEXT
            Temp4# = Temp#
        NEXT
    CASE "H" ' hexidecimal suffix
        Temp# = False
        Hex.Power = False
        FOR Hex.Digit = LEN(Strng) - 1 TO 1 STEP -1
            OutX$ = UCASE$(MID$(Strng, Hex.Digit, 1))
            SELECT CASE OutX$
                CASE "0" TO "9", "A" TO "F"
                    Hex.Value = VAL("&H" + OutX$)
                    Temp# = Temp# + Hex.Value * 16 ^ Hex.Power
                    Hex.Power = Hex.Power + 1
                CASE ELSE
                    ERROR 130
            END SELECT
        NEXT
    CASE "O" ' octal suffix
        Temp# = False
        Octal.Power = False
        FOR Octal.Digit = LEN(Strng) - 1 TO 1 STEP -1
            OutX$ = MID$(Strng, Octal.Digit, 1)
            IF OutX$ >= "0" AND OutX$ <= "7" THEN
                Octal.Value = VAL(OutX$)
                Temp# = Temp# + Octal.Value * 8 ^ Octal.Power
                Octal.Power = Octal.Power + 1
            ELSE
                ERROR 130
            END IF
        NEXT
    CASE "B" ' binary suffix
        Temp# = False
        Binary.Power = False
        FOR Binary.Digit = LEN(Strng) - 1 TO 1 STEP -1
            OutX$ = MID$(Strng, Binary.Digit, 1)
            SELECT CASE OutX$
                CASE "0"
                    ' nul
                CASE "1"
                    Temp# = Temp# + 2 ^ Binary.Power
                CASE ELSE
                    ERROR 130
            END SELECT
            Binary.Power = Binary.Power + 1
        NEXT
    CASE "0" TO "9", "D", "E" ' decimal/exponent suffix
        Decimal = False
        Exponent = False
        Number$ = Strng
        Start.Token:
        FOR Decimal.Digit = 1 TO LEN(Number$)
            OutX$ = UCASE$(MID$(Number$, Decimal.Digit, 1))
            SELECT CASE OutX$
                CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "-", "+"
                    ' nul
                CASE "D", "E"
                    IF Exponent THEN
                        ERROR 140
                    END IF
                    Exponent = True
                    Exponent$ = UCASE$(RIGHT$(Number$, 1))
                    IF Exponent$ = "D" OR Exponent$ = "E" THEN
                        ' concatenate pre-parsed tokens into exponent
                        Unary$ = MID$(Out2, Token.Index, 1)
                        IF Unary$ = "+" OR Unary$ = "-" THEN
                            Var$ = LEFT$(Out2, Token.Index - 1)
                            CALL Get.Token
                            Out2 = Var$ + Strng + MID$(Out2, Token.Index)
                            CALL Get.Token
                            IF Strng = Nul THEN
                                ERROR 140
                            END IF
                            Number$ = Number$ + Unary$ + Strng
                            IF Strng = "-" OR Strng = "+" THEN
                                ERROR 140
                            END IF
                            IF INSTR(Strng, ".") THEN
                                ERROR 140
                            END IF
                            DO
                                IF LEFT$(Number$, 1) = "-" THEN
                                    Number$ = MID$(Number$, 2)
                                ELSE
                                    IF LEFT$(Number$, 1) = "+" THEN
                                        Number$ = MID$(Number$, 2)
                                    ELSE
                                        EXIT DO
                                    END IF
                                END IF
                            LOOP
                            Decimal = False
                            Exponent = False
                            GOTO Start.Token
                        ELSE
                            ERROR 140
                        END IF
                    END IF
                CASE "."
                    IF Exponent THEN
                        ERROR 140
                    END IF
                    IF Decimal THEN
                        ERROR 140
                    END IF
                    Decimal = True
                CASE ELSE
                    ERROR 130
            END SELECT
        NEXT
        Temp# = CDBL(VAL(Number$))
    CASE ELSE
        ERROR 130
END SELECT
CALL Get.Token
Last.Token = True
END SUB

' parse string in quotes
SUB Parse.Quoted
Out3 = Nul
DO UNTIL MID$(Out2, Token.Index, 1) = CHR$(34)
    Out3 = Out3 + MID$(Out2, Token.Index, 1)
    Token.Index = Token.Index + 1
    IF Token.Index > LEN(Out2) THEN
        ERROR 92
        EXIT DO
    END IF
LOOP
CALL Get.Token
CALL Get.Token
Last.Token = False
END SUB

REM entry to expression parser, each following parser is higher precedence,
REM this routine called by higher precedence routines recursively.

REM precedence of operators:
REM   !, +, -  Not, Unary plus, Unary minus
REM   <<, >>, --, ++, **, //  Dual unary
REM   ^        Power
REM   *, /     Multiplication, Division
REM   \        Integer Division
REM   %        Modulo
REM   +, -     Addition, Subtraction
REM   =, <, >, <>, >=, <=   Relational
REM   |, &, ~, ?, :, #, @, `  Logical (or, and, xor, imp, eqv, nor, non, xan)

SUB Equate (Temp#)
Last.Token = False
Out3 = Nul ' reset string storage
Out4 = Nul ' reset string concatenate storage
Temp# = False ' reset result
Token.Index = 1 ' reset pointer to expression token
CALL Get.Token ' read next token
CALL Parse1(Temp#) ' entry to parse the expression
' parse leftover tokens
WHILE LEN(Strng)
    CALL Parse1(X#)
    ' toss away any unwanted tokens
    IF Last.Token = 3 THEN
        Strng = "<extra token>"
        ERROR 92
    END IF
WEND
END SUB

' logical parser
SUB Parse1 (Temp#)
CALL Parse2(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "|", "&", "~", "?", ":", "#", "@", "`"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse2(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' relational parser
SUB Parse2 (Temp#)
CALL Parse3(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "<", ">", "=", ">=", "<=", "<>"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse3(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' addition/subtraction parser
SUB Parse3 (Temp#)
CALL Parse4(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "+", "-"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse4(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' modulo parser
SUB Parse4 (Temp#)
CALL Parse5(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "%"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse5(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' integer division parser
SUB Parse5 (Temp#)
CALL Parse6(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "\"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse6(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' multiplication/division parser
SUB Parse6 (Temp#)
CALL Parse7(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "*", "/"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse7(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' power parser
SUB Parse7 (Temp#)
CALL Parse7a(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "^"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse7a(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' dual-unary parser
SUB Parse7a (Temp#)
CALL Parse8(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "<<", ">>", "--", "++", "**", "//"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    '    CALL Parse8(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' not/unary plus/unary negative parser
SUB Parse8 (Temp#)
Token.Negate$ = Nul ' reset not/unary tokens
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
        CASE "!", "-", "+", "--", "++"
            ' Nul
        CASE ELSE
            EXIT DO
    END SELECT
    CALL Get.Token ' read next token
    ' store combined not/unary tokens
    Token.Negate$ = Token.Negate$ + Token.Parsed$
    Token.Parsed$ = Strng
LOOP
CALL Parse9(Temp#) ' get next operator precedence
' process the combined operators in reverse
FOR Token.Type = LEN(Token.Negate$) TO 1 STEP -1
    SELECT CASE MID$(Token.Negate$, Token.Type, 1)
        CASE "+"
            ' nul calculation for unary plus
        CASE "-"
            Temp# = -Temp# ' perform negate
        CASE "!" ' not
            Temp# = NOT Temp# ' perform not calculation
    END SELECT
NEXT
END SUB

' string/numeric expression parser,
' routine returns calculated expression string/value,
' calls Parse1 recursively for values inside parenthesis/functions.
SUB Parse9 (Temp#)
SELECT CASE Token ' determine token type
    CASE 0 ' unknown token
        IF Last.Token = 3 THEN
            Strng = "<extra token>"
            ERROR 92
        END IF
        IF Strng = Nul THEN
            Strng = "<missing token>"
        END IF
        ERROR 92
        Last.Token = False
    CASE 1 ' token is symbol
        Token.Parsed$ = Strng
        SELECT CASE Token.Parsed$ ' determine token
            CASE ",", ";" ' separaters
                Last.Token = 3
            CASE CHR$(34) ' parse string
                CALL Parse.Quoted
                Quotes = True
            CASE "(" ' calculate opening parenthesis
                CALL Get.Token ' read next token value inside parenthesis
                IF Strng = ")" THEN
                    Strng = "<empty closing token>"
                    ERROR 92
                END IF
                IF Strng <> ")" THEN
                    DO ' calculate value
                        CALL Parse1(Temp#) ' call parse entry
                    LOOP UNTIL Strng = ")" OR Token = False ' check closing parenthesis
                END IF
                IF Token = False THEN
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
                CALL Get.Token ' read next token after parenthesis
            CASE "["
                CALL Get.Token
                IF Strng = "]" THEN
                    Strng = "<empty closing token>"
                    ERROR 92
                END IF
                IF Strng <> "]" THEN
                    DO
                        CALL Parse1(Temp#)
                    LOOP UNTIL Strng = "]" OR Token = False
                END IF
                IF Token = False THEN
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
                CALL Get.Token
            CASE "{"
                CALL Get.Token
                IF Strng = "}" THEN
                    Strng = "<empty closing token>"
                    ERROR 92
                END IF
                IF Strng <> "}" THEN
                    DO
                        CALL Parse1(Temp#)
                    LOOP UNTIL Strng = "}" OR Token = False
                END IF
                IF Token = False THEN
                    Strng = "<missing closing token>"
                    ERROR 92
                END IF
                CALL Get.Token
            CASE ")", "]", "}" ' check token is closing parenthesis
                Strng = "<extra closing token>"
                ERROR 92
        END SELECT
    CASE 2 ' token is numeric value
        CALL Parse.Numeric(Temp#)
    CASE 3 ' token type is alphabetic
        SELECT CASE LEN(Strng) ' check variable type
            CASE 1
                CALL Parse.Alphabetic1(Temp#)
            CASE 2
                CALL Parse.Alphabetic2(Temp#)
            CASE ELSE
                CALL Parse.Alphabetic3(Temp#)
        END SELECT
        CALL Get.Token
END SELECT
END SUB

' remove all white spaces, skip white spaces in quotes.
SUB Remove.Spaces
Var$ = Out2
Temp = False
DO
    Temp = Temp + 1
    IF Temp > LEN(Var$) THEN
        EXIT DO
    END IF
    IF MID$(Var$, Temp, 1) = CHR$(34) THEN
        DO
            Temp = Temp + 1
            IF Temp > LEN(Var$) THEN
                EXIT DO
            END IF
            IF MID$(Var$, Temp, 1) = CHR$(34) THEN
                EXIT DO
            END IF
        LOOP
    END IF
    FOR Blanks = 1 TO LEN(White.Space)
        IF MID$(Var$, Temp, 1) = MID$(White.Space, Blanks, 1) THEN
            Var$ = LEFT$(Var$, Temp - 1) + MID$(Var$, Temp + 1)
            Temp = Temp - 1
            EXIT FOR
        END IF
    NEXT
LOOP
Out2 = Var$
END SUB

SUB MorePrompt (Input.String$, Input.Mask$, Output.String$)
COLOR White, Black
IF Input.String$ = "-more-" THEN
    PRINT Input.String$;
ELSE
    PRINT Input.String$ + " ";
END IF
Input.Char$ = Nul
LOCATE , , 1
DO
    _LIMIT 100
    Input.Char$ = INKEYx$
    IF LEN(Input.Char$) THEN
        Input.Char$ = LCASE$(Input.Char$)
        IF INSTR(Input.Mask$, Input.Char$) THEN
            IF Input.Char$ = CHR$(13) THEN
                PRINT
            ELSE
                PRINT Input.Char$
            END IF
            Output.String$ = Input.Char$
            EXIT DO
        END IF
    END IF
LOOP
x = ClearKeyboard
END SUB

' input:  OutX$ - file to edit
'         VarC  - resume editing
' output: VarZ  - file save success
SUB ScrnEdit (OutX$, VarZ, VarC)

' array of file contents
STATIC Temp.ArrayS() AS STRING

' file handles
DIM hfind AS _OFFSET
DIM Tempfile1 AS INTEGER

' file declarations
STATIC Filename AS STRING
STATIC LastSearch AS SINGLE
STATIC SearchString AS STRING
STATIC Line.Number AS LONG
STATIC Max.Lines AS LONG
STATIC Max.File AS LONG
STATIC Search AS STRING
STATIC Replace AS STRING
Insert = -1
IF VarC THEN
    OutX$ = Filename
ELSE
    Filename = OutX$
END IF
FOR V = LEN(OutX$) TO 1 STEP -1
    IF MID$(OutX$, V, 1) = "\" THEN
        OutX$ = MID$(OutX$, V + 1)
        EXIT FOR
    END IF
NEXT
IF MID$(OutX$, 2, 1) = ":" THEN
    OutX$ = MID$(OutX$, 3)
END IF
OutX$ = LCASE$(OutX$)
MID$(OutX$, 1, 1) = UCASE$(MID$(OutX$, 1, 1))
IF VarC = 0 THEN
    Max.Lines = False
    Max.File = 1024
    REDIM Temp.ArrayS(1024) AS STRING
    IF OutX$ = "" THEN
        VarZ = -3
        EXIT SUB
    END IF
    IF ValidFileChar(OutX$) THEN
        VarZ = -3
        EXIT SUB
    END IF
    IF _FILEEXISTS(Filename) = 0 THEN
        ' create file
        f$ = Filename + CHR$(0)
        hfind = CreateFile(f$, &H180, 0, 0, 1, 0, 0)
        IF hfind <> INVALID_HANDLE_VALUE THEN
            r = CloseHandle(hfind)
        ELSE
            VarZ = -2
            EXIT SUB
        END IF
    ELSE
        COLOR 7, 0
        PRINT "Loading file.."
        Tempfile1 = FREEFILE
        OPEN Filename FOR BINARY AS #Tempfile1
        Count# = 0#
        Count2# = 0#
        Percent = 0
        Last.Percent = 0
        File.Size = LOF(Tempfile1)
        Control.Break = False
        IF LOF(Tempfile1) > 0 THEN
            DO UNTIL EOF(Tempfile1)
                LINE INPUT #Tempfile1, LineX$
                Count# = Count# + LEN(LineX$) + 2#
                Max.Lines = Max.Lines + 1
                IF Max.Lines >= 2147483646& THEN
                    File.Loaded = False
                    VarZ = -5
                    ERASE Temp.ArrayS
                    _TITLE Program
                    EXIT SUB
                END IF
                IF Max.Lines > Max.File THEN
                    Max.File = Max.File + 1
                    REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
                END IF
                Temp.ArrayS(Max.Lines) = LineX$
                GOSUB Display.Percent
                IF Control.Break THEN
                    Control.Break = False
                    File.Loaded = False
                    VarZ = -4
                    _TITLE Program
                    EXIT SUB
                END IF
            LOOP
            IF Count# < LOF(Tempfile1) THEN
                Max.Lines = Max.Lines + 1
                IF Max.Lines > Max.File THEN
                    Max.File = Max.File + 1
                    REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
                END IF
                Temp.ArrayS(Max.Lines) = ""
            END IF
        END IF
        CLOSE #Tempfile1
    END IF
    VarX$ = UCASE$(AmbiguateFile$(Filename))
    VarX$ = Program + " - EDIT - " + VarX$
    _TITLE VarX$
END IF
Xcoor = 1
Ycoor = 1
Column = 1
Row = 1
IF Max.Lines = 0 THEN
    Line.Number = 0
ELSE
    Line.Number = 1
END IF
File.Loaded = -1
StoreStatus = Status
Status = -1
REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
GOSUB Display.Screen
DO
    Var# = CDBL(Line.Number)
    IF Var# = 0# THEN
        Var$ = " 0"
    ELSE
        Var$ = " " + FormatString$(Var#)
    END IF

    V = Column + Ycoor - 1
    IF V = 0 THEN
        Var2$ = " 0"
    ELSE
        Var2$ = " " + FormatString$(CDBL(V))
    END IF

    Strng = "Row" + Var$ + " Column" + Var2$
    IF V > 0 AND V <= LEN(Temp.ArrayS(Line.Number)) THEN
        V2 = ASC(MID$(Temp.ArrayS(Line.Number), V, 1))
        Strng = Strng + " Ascii" + STR$(V2)

        SELECT CASE V2 ' parse control codes
            CASE 0
                Strng = Strng + " (nul)"
            CASE 1
                Strng = Strng + " (soh)"
            CASE 2
                Strng = Strng + " (stx)"
            CASE 3
                Strng = Strng + " (etx)"
            CASE 4
                Strng = Strng + " (eot)"
            CASE 5
                Strng = Strng + " (enq)"
            CASE 6
                Strng = Strng + " (ack)"
            CASE 7
                Strng = Strng + " (bel)"
            CASE 8
                Strng = Strng + " (bs)"
            CASE 9
                Strng = Strng + " (tab)"
            CASE 10
                Strng = Strng + " (lf)"
            CASE 11
                Strng = Strng + " (vt)"
            CASE 12
                Strng = Strng + " (np)"
            CASE 13
                Strng = Strng + " (cr)"
            CASE 14
                Strng = Strng + " (so)"
            CASE 15
                Strng = Strng + " (si)"
            CASE 16
                Strng = Strng + " (dle)"
            CASE 17
                Strng = Strng + " (dc1)"
            CASE 18
                Strng = Strng + " (dc2)"
            CASE 19
                Strng = Strng + " (dc3)"
            CASE 20
                Strng = Strng + " (dc4)"
            CASE 21
                Strng = Strng + " (nak)"
            CASE 22
                Strng = Strng + " (syn)"
            CASE 23
                Strng = Strng + " (etb)"
            CASE 24
                Strng = Strng + " (can)"
            CASE 25
                Strng = Strng + " (em)"
            CASE 26
                Strng = Strng + " (eof)"
            CASE 27
                Strng = Strng + " (esc)"
            CASE 28
                Strng = Strng + " (fs)"
            CASE 29
                Strng = Strng + " (gs)"
            CASE 30
                Strng = Strng + " (rs)"
            CASE 31
                Strng = Strng + " (us)"
            CASE 32
                Strng = Strng + " (spc)"
            CASE 255
                Eat$ = ""
                ' Strng = Strng + " (" + CHR$(255) + ")"
            CASE ELSE
                Strng = Strng + " (" + CHR$(V2) + ")"
        END SELECT
    END IF
    Strng = LEFT$(Strng, 41)
    Strng = Strng + SPACE$(41 - LEN(Strng))

    LOCATE 25, 1, 1
    COLOR 15, 1
    _CONTROLCHR OFF
    PRINT Strng;
    _CONTROLCHR ON

    COLOR 15, 0
    LOCATE Xcoor + 1, Ycoor, 1
    PRINT "";
    IF Insert THEN
        LOCATE , , , 8, 8
    ELSE
        LOCATE , , , 0, 8
    END IF
    Control.Break = False
    Monitor.Box = False
    DO
        ' check alt key press
        CurrentMenu = False
        Q& = _KEYHIT
        IF Q& = KEY_LALT& THEN
            CurrentMenu = 1
            CALL Menu2
            VIEW PRINT 1 TO 25
            GOSUB Display.Screen
            IF CurrentMenu = 0 THEN
                EXIT DO
            END IF
        END IF
        IF Q& = KEY_RALT& THEN
            CurrentMenu = 1
            CALL Menu2
            VIEW PRINT 1 TO 25
            GOSUB Display.Screen
            IF CurrentMenu = 0 THEN
                EXIT DO
            END IF
        END IF

        ' check mouse activity
        Z = MouseDriver
        IF MouseButton1 THEN
            ' select drop down menu.
            IF MouseX = 1 THEN
                NewMenuSelection = 0
                SELECT CASE MouseY
                    CASE 1 TO 3 ' cpu
                        CPUOFF = NOT CPUOFF
                        IF CPUOFF THEN
                            TIMER OFF
                            _TITLE Program
                            CALL DisplayToolBar
                        ELSE
                            TIMER ON
                            CALL DisplayToolBar
                        END IF
                    CASE 74 TO 78 ' about
                        CALL HelpAbout
                    CASE 6 TO 9 ' File
                        NewMenuSelection = 1
                    CASE 16 TO 21 ' Charts
                        NewMenuSelection = 2
                    CASE 26 TO 29 ' Dump
                        NewMenuSelection = 3
                    CASE 36 TO 39 ' Edit
                        NewMenuSelection = 4
                    CASE 46 TO 49 ' Jump
                        NewMenuSelection = 5
                    CASE 56 TO 60 ' Print
                        NewMenuSelection = 6
                    CASE 64 TO 69 ' Search
                        NewMenuSelection = 7
                END SELECT
                IF NewMenuSelection THEN
                    CurrentMenu = NewMenuSelection
                    NewMenuSelection = 0
                    CurrentMenuSelection = 1
                    CALL DropDownMenu(Var)
                    VIEW PRINT 1 TO 25
                    GOSUB Display.Screen
                    IF CurrentMenu = 0 THEN
                        EXIT DO
                    END IF
                END IF
            END IF
        END IF
        IF CurrentMenu THEN
            SELECT CASE CurrentMenu
                CASE 7 ' search
                    SELECT CASE CurrentMenuSelection
                        CASE 1 ' search string
                            IF Max.Lines > 0 THEN
                                X$ = SearchString$
                                x = SearchBox(X$, 2)
                                IF x THEN
                                    IF LEN(X$) THEN
                                        SearchString$ = X$
                                        Flag = 0
                                        S! = Line.Number
                                        FOR X! = S! TO Max.Lines
                                            IF WildcardSelect THEN
                                                CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                            ELSE
                                                IF SensitiveSelect THEN
                                                    S$ = Temp.ArrayS(X!)
                                                ELSE
                                                    S$ = UCASE$(Temp.ArrayS(X!))
                                                    X$ = UCASE$(X$)
                                                END IF
                                                Flag = INSTR(S$, X$)
                                            END IF
                                            IF Flag THEN
                                                Line.Number = X!
                                                IF Line.Number <= 23 THEN
                                                    Xcoor = Line.Number
                                                ELSE
                                                    Xcoor = 1
                                                END IF
                                                Ycoor = 1
                                                Column = 1
                                                LastSearch = X!
                                                EXIT FOR
                                            END IF
                                        NEXT
                                        IF Flag = 0 THEN
                                            FOR X! = 1! TO S!
                                                IF WildcardSelect THEN
                                                    CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                                ELSE
                                                    IF SensitiveSelect THEN
                                                        S$ = Temp.ArrayS(X!)
                                                    ELSE
                                                        S$ = UCASE$(Temp.ArrayS(X!))
                                                        X$ = UCASE$(X$)
                                                    END IF
                                                    Flag = INSTR(S$, X$)
                                                END IF
                                                IF Flag THEN
                                                    Line.Number = X!
                                                    IF Line.Number <= 23 THEN
                                                        Xcoor = Line.Number
                                                    ELSE
                                                        Xcoor = 1
                                                    END IF
                                                    Ycoor = 1
                                                    Column = 1
                                                    LastSearch = X!
                                                    EXIT FOR
                                                END IF
                                            NEXT
                                        END IF
                                        IF Flag = 0 THEN
                                            x = MessageBox(" Search ", "Text not found.")
                                        END IF
                                    END IF
                                END IF
                                COLOR 15, 0
                                GOSUB Display.Screen
                            END IF
                        CASE 2 ' search bytes
                            IF Max.Lines > 0 THEN
                                X$ = SearchString$
                                x = SearchBox(X$, 1)
                                IF x THEN
                                    IF LEN(X$) THEN
                                        CALL CheckHexBytes2(X$)
                                        IF ValidByteString = 0 THEN
                                            QZ = MessageBox(" Error ", "Invalid HEX byte string.")
                                        ELSE
                                            SearchString$ = X$
                                            Flag = 0
                                            S! = Line.Number
                                            FOR X! = S! TO Max.Lines
                                                IF WildcardSelect THEN
                                                    CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                                ELSE
                                                    IF SensitiveSelect THEN
                                                        S$ = Temp.ArrayS(X!)
                                                    ELSE
                                                        S$ = UCASE$(Temp.ArrayS(X!))
                                                        X$ = UCASE$(X$)
                                                    END IF
                                                    Flag = INSTR(S$, X$)
                                                END IF
                                                IF Flag THEN
                                                    Line.Number = X!
                                                    IF Line.Number <= 23 THEN
                                                        Xcoor = Line.Number
                                                    ELSE
                                                        Xcoor = 1
                                                    END IF
                                                    Ycoor = 1
                                                    Column = 1
                                                    LastSearch = X!
                                                    EXIT FOR
                                                END IF
                                            NEXT
                                            IF Flag = 0 THEN
                                                FOR X! = 1! TO S!
                                                    IF WildcardSelect THEN
                                                        CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                                    ELSE
                                                        IF SensitiveSelect THEN
                                                            S$ = Temp.ArrayS(X!)
                                                        ELSE
                                                            S$ = UCASE$(Temp.ArrayS(X!))
                                                            X$ = UCASE$(X$)
                                                        END IF
                                                        Flag = INSTR(S$, X$)
                                                    END IF
                                                    IF Flag THEN
                                                        Line.Number = X!
                                                        IF Line.Number <= 23 THEN
                                                            Xcoor = Line.Number
                                                        ELSE
                                                            Xcoor = 1
                                                        END IF
                                                        Ycoor = 1
                                                        Column = 1
                                                        LastSearch = X!
                                                        EXIT FOR
                                                    END IF
                                                NEXT
                                            END IF
                                            IF Flag = 0 THEN
                                                x = MessageBox(" Search ", "Text not found.")
                                            END IF
                                        END IF
                                        COLOR 15, 0
                                        GOSUB Display.Screen
                                    END IF
                                END IF
                            END IF
                        CASE 3 ' searchreplace
                            IF Max.Lines > 0 THEN
                                x = SearchBox2(Search$, Replace$)
                                IF x = 1 THEN ' replace
                                    ' store current screen
                                    Store1 = Xcoor
                                    Store2 = Ycoor
                                    Store3 = Column
                                    Store! = Line.Number
                                    ' start replace
                                    Count! = 0!
                                    IF LEN(Search$) THEN
                                        FOR X! = Store! TO Max.Lines
                                            S1 = 1
                                            Text$ = Temp.ArrayS(X!)
                                            DO
                                                IF SensitiveSelect THEN
                                                    P1 = INSTR(S1, Text$, Search$)
                                                ELSE
                                                    P1 = INSTR(S1, UCASE$(Text$), UCASE$(Search$))
                                                END IF
                                                IF P1 THEN
                                                    ' jump to line
                                                    Ycoor = 1
                                                    Column = P1
                                                    Line.Number = X!
                                                    IF Line.Number <= 23 THEN
                                                        Xcoor = Line.Number
                                                    ELSE
                                                        Xcoor = 1
                                                    END IF
                                                    ' display screen
                                                    COLOR 15, 0
                                                    GOSUB Display.Screen
                                                    ' display text to replace
                                                    LOCATE Xcoor + 1, 1
                                                    COLOR 15, 1
                                                    Var1$ = MID$(Text$, P1, LEN(Search$))
                                                    PRINT Var1$;
                                                    COLOR 15, 0
                                                    ' make box bounce
                                                    IF Xcoor < 10 THEN
                                                        q = 12
                                                    ELSE
                                                        q = 2
                                                    END IF
                                                    ' z=-1 replace, z=1 skip
                                                    Z = ReplaceBox(q, X!)
                                                    IF Z = 0 THEN ' cancel
                                                        EXIT FOR
                                                    END IF
                                                    IF Z = -1 THEN
                                                        Text$ = LEFT$(Text$, P1 - 1) + Replace$ + MID$(Text$, P1 + LEN(Search$))
                                                        Temp.ArrayS(X!) = Text$
                                                        Count! = Count! + 1!
                                                    END IF
                                                ELSE
                                                    EXIT DO
                                                END IF
                                                S1 = P1 + LEN(Replace$)
                                            LOOP
                                        NEXT
                                    END IF
                                    ' restore current screen
                                    Xcoor = Store1
                                    Ycoor = Store2
                                    Column = Store3
                                    Line.Number = Store!
                                    COLOR 15, 0
                                    GOSUB Display.Screen
                                    ' display replacements
                                    IF Count! = 0! THEN
                                        Z = MessageBox(" Search ", "Text not found.")
                                    ELSE
                                        Z = MessageBox(" Replaced ", STR$(Count!) + " strings.")
                                    END IF
                                END IF
                                IF x = 2 THEN ' replaceall
                                    Count! = 0!
                                    IF LEN(Search$) THEN
                                        FOR X! = 1! TO Max.Lines
                                            Text$ = Temp.ArrayS(X!)
                                            Flag = 0
                                            S1 = 1
                                            DO
                                                IF SensitiveSelect THEN
                                                    P1 = INSTR(S1, Text$, Search$)
                                                ELSE
                                                    P1 = INSTR(S1, UCASE$(Text$), UCASE$(Search$))
                                                END IF
                                                IF P1 THEN
                                                    Flag = -1
                                                    Text$ = LEFT$(Text$, P1 - 1) + Replace$ + MID$(Text$, P1 + LEN(Search$))
                                                    Count! = Count! + 1!
                                                ELSE
                                                    EXIT DO
                                                END IF
                                                S1 = P1 + LEN(Replace$)
                                            LOOP
                                            IF Flag THEN
                                                Temp.ArrayS(X!) = Text$
                                            END IF
                                        NEXT
                                    END IF
                                    COLOR 15, 0
                                    GOSUB Display.Screen
                                    IF Count! = 0! THEN
                                        Z = MessageBox(" Search ", "Text not found.")
                                    ELSE
                                        Z = MessageBox(" Replaced ", STR$(Count!) + " strings.")
                                    END IF
                                END IF
                            END IF
                    END SELECT
                CASE 4 ' append
                    SELECT CASE CurrentMenuSelection
                        CASE 1 ' append hex bytes
                            VarA$ = AppendBox$(1)
                            IF LEN(VarA$) THEN
                                CALL CheckHexBytes(VarA$)
                                IF ValidByteString = 0 THEN
                                    QZ = MessageBox(" Error ", "Invalid HEX byte string.")
                                ELSE
                                    VarB$ = ""
                                    DO
                                        VarB$ = VarB$ + CHR$(VAL("&H" + LEFT$(VarA$, 2)))
                                        VarA$ = MID$(VarA$, 3)
                                        IF VarA$ = "" THEN
                                            EXIT DO
                                        END IF
                                    LOOP
                                    ' append always extends array
                                    IF Max.Lines + 1 > Max.File THEN
                                        Max.File = Max.File + 1
                                        REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
                                    END IF
                                    ' append line to eof
                                    Max.Lines = Max.Lines + 1
                                    Temp.ArrayS(Max.Lines) = VarB$
                                    ' control-end
                                    Line.Number = Max.Lines
                                    IF Max.Lines < 22 THEN
                                        Xcoor = Max.Lines
                                    ELSE
                                        Xcoor = 22
                                    END IF
                                    Ycoor = 1
                                    Column = 1
                                    GOSUB Display.Screen
                                END IF
                            END IF
                        CASE 2 ' append ascii string
                            VarA$ = AppendBox$(2)
                            IF LEN(VarA$) THEN
                                ' append always extends array
                                IF Max.Lines + 1 > Max.File THEN
                                    Max.File = Max.File + 1
                                    REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
                                END IF
                                ' append line to eof
                                Max.Lines = Max.Lines + 1
                                Temp.ArrayS(Max.Lines) = VarA$
                                ' control-end
                                Line.Number = Max.Lines
                                IF Max.Lines < 22 THEN
                                    Xcoor = Max.Lines
                                ELSE
                                    Xcoor = 22
                                END IF
                                Ycoor = 1
                                Column = 1
                                GOSUB Display.Screen
                            END IF
                    END SELECT
                CASE 5 ' jump
                    IF CurrentMenuSelection = 1 THEN ' jump to byte
                        IF Max.Lines > 0 THEN
                            GOSUB Jump.Byte
                            IF VarX THEN
                                Line.Number = Count!
                                IF Line.Number <= 23 THEN
                                    Xcoor = Line.Number
                                ELSE
                                    Xcoor = 1
                                END IF
                                Ycoor = 1
                                Column = 1
                            END IF
                        END IF
                        COLOR 15, 0
                        GOSUB Display.Screen
                    END IF
                    IF CurrentMenuSelection = 2 THEN ' jump to line
                        IF Max.Lines > 0 THEN
                            GOSUB Jump.Line
                            IF VarX THEN
                                Line.Number = Count!
                                IF Line.Number <= 23 THEN
                                    Xcoor = Line.Number
                                ELSE
                                    Xcoor = 1
                                END IF
                                Ycoor = 1
                                Column = 1
                            END IF
                        END IF
                        COLOR 15, 0
                        GOSUB Display.Screen
                    END IF
                CASE 3
                    IF CurrentMenuSelection = 1 THEN ' hex screen dump
                        HexDump = -1
                        GOSUB Screen.Dump
                    END IF
                    IF CurrentMenuSelection = 2 THEN ' hex file dump
                        HexDump = -1
                        GOSUB File.Dump
                    END IF
                CASE 6
                    IF CurrentMenuSelection = 1 THEN ' hex screen print
                        HexDump = -2
                        GOSUB Screen.Print
                    END IF
                    IF CurrentMenuSelection = 2 THEN ' hex file print
                        HexDump = -2
                        GOSUB File.Print
                    END IF
                CASE 1
                    IF CurrentMenuSelection = 2 THEN ' close file
                        TempX$ = CHR$(27)
                    END IF
                    IF CurrentMenuSelection = 6 THEN ' exit program
                        TempX$ = CHR$(27)
                    END IF
                CASE 2 ' chart
                    IF CurrentMenuSelection = 1 THEN ' ascii chart
                        VIEW PRINT 2 TO 24
                        CALL SetKeys
                        KEY ON
                        CALL AsciiChart
                        KEY OFF
                        CALL ClearKeys
                        VIEW PRINT 1 TO 25
                        GOSUB Display.Screen
                    END IF
                    IF CurrentMenuSelection = 2 THEN ' hex chart
                        VIEW PRINT 2 TO 24
                        CALL SetKeys
                        KEY ON
                        CALL HexChart
                        KEY OFF
                        CALL ClearKeys
                        VIEW PRINT 1 TO 25
                        GOSUB Display.Screen
                    END IF
            END SELECT
            EXIT DO
        END IF

        _LIMIT 30
        TempX$ = INKEYx$
        IF TempX$ = CHR$(3) THEN
            _DELAY 1.2
        END IF
        IF Control.Break THEN
            TempX$ = Nul
            Control.Break = False
            GOSUB Display.Screen
            EXIT DO
        END IF
        IF TempX$ <> "" THEN
            EXIT DO
        END IF
    LOOP
    SELECT CASE LEN(TempX$)
        CASE 1
            SELECT CASE TempX$
                CASE CHR$(1) ' ctrl-a insert char
                    ' insert ascii character
                    Ascii$ = AsciiBox$
                    IF LEN(Ascii$) THEN
                        Ascii.Value = INT(VAL(Ascii$))
                        IF Ascii.Value >= 0 AND Ascii.Value <= 255 THEN
                            TempX$ = CHR$(Ascii.Value)
                            IF Max.Lines = 0 THEN ' init array
                                Line.Number = 1
                                Temp.ArrayS(1) = TempX$
                                Max.Lines = 1
                                Ycoor = 2
                                Column = 1
                                GOSUB Display.Screen
                            ELSE
                                T$ = Temp.ArrayS(Line.Number)
                                IF Insert THEN
                                    T$ = LEFT$(T$, Column + Ycoor - 2) + TempX$ + MID$(T$, Column + Ycoor - 1)
                                ELSE
                                    IF Column + Ycoor - 1 > LEN(T$) THEN
                                        T$ = T$ + TempX$
                                    ELSE
                                        MID$(T$, Column + Ycoor - 1, 1) = TempX$
                                    END IF
                                END IF
                                Temp.ArrayS(Line.Number) = T$
                                IF Ycoor < 80 THEN
                                    Ycoor = Ycoor + 1
                                ELSE
                                    Column = Column + 1
                                END IF
                                GOSUB Display.Screen
                            END IF
                        END IF
                    END IF
                CASE CHR$(5) ' ctrl-e RepeatFind
                    IF Max.Lines > 0 THEN
                        X$ = SearchString$
                        IF LEN(X$) THEN
                            X$ = UCASE$(X$)
                            Start! = Line.Number + 1!
                            Flag = 0
                            FOR X! = Start! TO Max.Lines
                                IF INSTR(UCASE$(Temp.ArrayS(X!)), X$) THEN
                                    Line.Number = X!
                                    IF Line.Number <= 23 THEN
                                        Xcoor = Line.Number
                                    ELSE
                                        Xcoor = 1
                                    END IF
                                    Ycoor = 1
                                    Column = 1
                                    LastSearch = X!
                                    Flag = -1
                                    EXIT FOR
                                END IF
                            NEXT
                            IF Flag = 0 THEN
                                FOR X! = 1! TO Start!
                                    IF INSTR(UCASE$(Temp.ArrayS(X!)), X$) THEN
                                        Line.Number = X!
                                        IF Line.Number <= 23 THEN
                                            Xcoor = Line.Number
                                        ELSE
                                            Xcoor = 1
                                        END IF
                                        Ycoor = 1
                                        Column = 1
                                        LastSearch = X!
                                        Flag = -1
                                        EXIT FOR
                                    END IF
                                NEXT
                            END IF
                        END IF
                        COLOR 15, 0
                        GOSUB Display.Screen
                    END IF
                CASE CHR$(6) ' ctrl-f Find
                    IF Max.Lines > 0 THEN
                        COLOR 15, 1
                        LOCATE 25, 1
                        Strng = SPACE$(41)
                        PRINT Strng;
                        LOCATE 25, 1
                        Strng = "Enter search string: "
                        PRINT Strng;
                        COLOR 15, 0
                        LOCATE 25, 22
                        PRINT SPACE$(50);
                        X$ = LineInput$(25, 22, 50)
                        IF LEN(X$) THEN
                            SearchString = X$
                            X$ = UCASE$(X$)
                            Start! = Line.Number
                            Flag = 0
                            FOR X! = Start! TO Max.Lines
                                IF INSTR(UCASE$(Temp.ArrayS(X!)), X$) THEN
                                    Line.Number = X!
                                    IF Line.Number <= 23 THEN
                                        Xcoor = Line.Number
                                    ELSE
                                        Xcoor = 1
                                    END IF
                                    Ycoor = 1
                                    Column = 1
                                    LastSearch = X!
                                    Flag = -1
                                    EXIT FOR
                                END IF
                            NEXT
                            IF Flag = 0 THEN
                                FOR X! = 1! TO Start!
                                    IF INSTR(UCASE$(Temp.ArrayS(X!)), X$) THEN
                                        Line.Number = X!
                                        IF Line.Number <= 23 THEN
                                            Xcoor = Line.Number
                                        ELSE
                                            Xcoor = 1
                                        END IF
                                        Ycoor = 1
                                        Column = 1
                                        LastSearch = X!
                                        Flag = -1
                                        EXIT FOR
                                    END IF
                                NEXT
                            END IF
                        END IF
                        COLOR 15, 0
                        GOSUB Display.Screen
                    END IF
                CASE CHR$(7) ' ctrl-g Goto
                    IF Max.Lines > 0 THEN
                        GOSUB Jump.Line
                        IF VarX THEN
                            Line.Number = Count!
                            IF Line.Number <= 23 THEN
                                Xcoor = Line.Number
                            ELSE
                                Xcoor = 1
                            END IF
                            Ycoor = 1
                            Column = 1
                        END IF
                    END IF
                    COLOR 15, 0
                    GOSUB Display.Screen
                CASE CHR$(13) ' insert line
                    IF Max.Lines = 0 THEN ' init array
                        Line.Number = 1
                        Temp.ArrayS(1) = ""
                        Max.Lines = 1
                    ELSE
                        ' enter always extends array
                        IF Max.Lines + 1 > Max.File THEN
                            Max.File = Max.File + 1
                            REDIM _PRESERVE Temp.ArrayS(Max.File) AS STRING
                        END IF
                        IF Max.Lines < Max.File THEN
                            ' above
                            IF Column + Ycoor - 1 = 1 THEN
                                FOR Temp1 = Max.Lines TO Line.Number STEP -1
                                    Temp.ArrayS(Temp1 + 1) = Temp.ArrayS(Temp1)
                                NEXT
                                Temp.ArrayS(Line.Number) = ""
                                Max.Lines = Max.Lines + 1
                                Line.Number = Line.Number + 1
                                IF Xcoor < 23 THEN
                                    Xcoor = Xcoor + 1
                                END IF
                            ELSE
                                ' below
                                IF Column + Ycoor - 1 > LEN(Temp.ArrayS(Line.Number)) THEN
                                    Line.Number = Line.Number + 1 ' below
                                    FOR Temp1 = Max.Lines TO Line.Number STEP -1
                                        Temp.ArrayS(Temp1 + 1) = Temp.ArrayS(Temp1)
                                    NEXT
                                    Temp.ArrayS(Line.Number) = ""
                                    Max.Lines = Max.Lines + 1
                                    IF Xcoor < 23 THEN
                                        Xcoor = Xcoor + 1
                                    END IF
                                ELSE
                                    ' split line
                                    FOR Temp1 = Max.Lines TO Line.Number + 1 STEP -1
                                        Temp.ArrayS(Temp1 + 1) = Temp.ArrayS(Temp1)
                                    NEXT
                                    T$ = Temp.ArrayS(Line.Number)
                                    U$ = MID$(T$, Column + Ycoor - 1)
                                    T$ = LEFT$(T$, Column + Ycoor - 2)
                                    Temp.ArrayS(Line.Number) = T$
                                    Line.Number = Line.Number + 1 ' below
                                    Temp.ArrayS(Line.Number) = U$
                                    Max.Lines = Max.Lines + 1
                                    IF Xcoor < 23 THEN
                                        Xcoor = Xcoor + 1
                                    END IF
                                END IF
                            END IF
                        END IF
                    END IF
                    ' enter always moves to 1,1
                    Column = 1
                    Ycoor = 1
                    GOSUB Display.Screen
                CASE CHR$(8) ' backspace
                    q = Column + Ycoor - 1
                    IF Line.Number <= 1 AND q <= 1 THEN ' no backspace at 1,1
                        Eat$ = ""
                    ELSE
                        ' remove current line
                        IF Temp.ArrayS(Line.Number) = "" THEN
                            ' pack array
                            FOR Temp1 = Line.Number TO Max.Lines - 1
                                Temp.ArrayS(Temp1) = Temp.ArrayS(Temp1 + 1)
                            NEXT
                            IF Xcoor > 1 THEN
                                Xcoor = Xcoor - 1
                            END IF
                            Line.Number = Line.Number - 1
                            Max.Lines = Max.Lines - 1
                            ' position at eol
                            End.Line = LEN(Temp.ArrayS(Line.Number))
                            IF End.Line >= 80 THEN
                                Column = End.Line - 78
                                Ycoor = End.Line - Column + 2
                            ELSE
                                Ycoor = End.Line - Column + 2
                            END IF
                            GOSUB Display.Screen
                        ELSE
                            ' concatenate previous line
                            IF Column + Ycoor - 1 <= 1 THEN
                                IF Line.Number > 1 THEN
                                    IF Line.Number <= Max.Lines THEN
                                        Var = 0
                                        IF Temp.ArrayS(Line.Number - 1) = "" THEN
                                            IF Ycoor = 1 THEN
                                                Var = -1
                                            END IF
                                        END IF

                                        ' position at eol
                                        IF Var = 0 THEN
                                            End.Line = LEN(Temp.ArrayS(Line.Number - 1))
                                        END IF

                                        Temp.ArrayS(Line.Number - 1) = Temp.ArrayS(Line.Number - 1) + Temp.ArrayS(Line.Number)
                                        FOR Temp1 = Line.Number TO Max.Lines - 1
                                            Temp.ArrayS(Temp1) = Temp.ArrayS(Temp1 + 1)
                                        NEXT
                                        IF Xcoor > 1 THEN
                                            Xcoor = Xcoor - 1
                                        END IF
                                        Line.Number = Line.Number - 1
                                        Max.Lines = Max.Lines - 1
                                        Column = 1
                                        Ycoor = 1

                                        ' position at eol
                                        IF Var = 0 THEN
                                            IF End.Line >= 80 THEN
                                                Column = End.Line - 78
                                                Ycoor = End.Line - Column + 2
                                            ELSE
                                                Ycoor = End.Line - Column + 2
                                            END IF
                                        END IF

                                        GOSUB Display.Screen
                                    END IF
                                END IF
                            ELSE
                                ' backspace 1 character
                                IF Column + Ycoor - 2 > False THEN
                                    T$ = Temp.ArrayS(Line.Number)
                                    T$ = LEFT$(T$, Column + Ycoor - 3) + MID$(T$, Column + Ycoor - 1)
                                    Temp.ArrayS(Line.Number) = T$
                                    ' move cursor 1 left
                                    IF Ycoor > 1 THEN
                                        Ycoor = Ycoor - 1
                                        GOSUB Display.Screen
                                    ELSE
                                        IF Column > 1 THEN
                                            Column = Column - 1
                                            GOSUB Display.Screen
                                        END IF
                                    END IF
                                END IF
                            END IF
                        END IF
                    END IF
                CASE CHR$(9) ' tab
                    IF Column + Ycoor + 7 <= LEN(Temp.ArrayS(Line.Number)) THEN
                        IF Ycoor + 8 < 80 THEN
                            Ycoor = Ycoor + 8
                        ELSE
                            Column = Column + 8
                            GOSUB Display.Screen
                        END IF
                    ELSE ' end
                        End.Line = LEN(Temp.ArrayS(Line.Number))
                        IF End.Line >= 80 THEN
                            Column = End.Line - 78
                            Ycoor = End.Line - Column + 2
                            GOSUB Display.Screen
                        ELSE
                            Ycoor = End.Line - Column + 2
                        END IF
                    END IF
                CASE CHR$(27) ' escape, exit editor
                    VarX$ = "File: " + UCASE$(AmbiguateFile$(Filename))
                    Quit = QuitBox2(VarX$) ' -1=yes, 1=no, 0=cancel
                    IF Quit = -1 THEN
                        Yes = -1
                        EXIT DO
                    END IF
                    IF Quit = 1 THEN
                        Yes = 0
                        EXIT DO
                    END IF
                CASE ELSE ' insert character
                    IF Max.Lines = 0 THEN ' init array
                        Line.Number = 1
                        Temp.ArrayS(1) = TempX$
                        Max.Lines = 1
                        Ycoor = 2
                        Column = 1
                        GOSUB Display.Screen
                    ELSE
                        T$ = Temp.ArrayS(Line.Number)
                        IF Insert THEN
                            T$ = LEFT$(T$, Column + Ycoor - 2) + TempX$ + MID$(T$, Column + Ycoor - 1)
                        ELSE
                            IF Column + Ycoor - 1 > LEN(T$) THEN
                                T$ = T$ + TempX$
                            ELSE
                                MID$(T$, Column + Ycoor - 1, 1) = TempX$
                            END IF
                        END IF
                        Temp.ArrayS(Line.Number) = T$
                        IF Ycoor < 80 THEN
                            Ycoor = Ycoor + 1
                        ELSE
                            Column = Column + 1
                        END IF
                        GOSUB Display.Screen
                    END IF
            END SELECT
        CASE 2
            Temp = ASC(MID$(TempX$, 2, 1))
            SELECT CASE Temp
                CASE 59 ' F1 - Help
                    x = HelpBox
                CASE 60 ' F2 - savefile
                    IF ErrorBox(" Savefile ", "Save file to disk?") THEN
                        GOSUB SaveFile
                        x = MessageBox(" Savefile ", "File saved to disk.")
                    END IF
                CASE 61 ' F3 - repeatfind
                    IF Max.Lines > 0 THEN
                        X$ = SearchString$
                        x = 0
                        IF LEN(X$) = 0 THEN
                            x = SearchBox(X$, 2)
                            Start! = Line.Number
                        ELSE
                            x = -1
                            Start! = Line.Number + 1!
                        END IF
                        IF x THEN
                            IF LEN(X$) THEN
                                SearchString$ = X$
                                Flag = 0
                                FOR X! = Start! TO Max.Lines
                                    IF WildcardSelect THEN
                                        CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                    ELSE
                                        IF SensitiveSelect THEN
                                            S$ = Temp.ArrayS(X!)
                                        ELSE
                                            S$ = UCASE$(Temp.ArrayS(X!))
                                            X$ = UCASE$(X$)
                                        END IF
                                        Flag = INSTR(S$, X$)
                                    END IF
                                    IF Flag THEN
                                        Line.Number = X!
                                        IF Line.Number <= 23 THEN
                                            Xcoor = Line.Number
                                        ELSE
                                            Xcoor = 1
                                        END IF
                                        Ycoor = 1
                                        Column = 1
                                        LastSearch = X!
                                        EXIT FOR
                                    END IF
                                NEXT
                                IF Flag = 0 THEN
                                    FOR X! = 1! TO Start!
                                        IF WildcardSelect THEN
                                            CALL InstrSUB1(Flag, X$, Temp.ArrayS(X!), SensitiveSelect)
                                        ELSE
                                            IF SensitiveSelect THEN
                                                S$ = Temp.ArrayS(X!)
                                            ELSE
                                                S$ = UCASE$(Temp.ArrayS(X!))
                                                X$ = UCASE$(X$)
                                            END IF
                                            Flag = INSTR(S$, X$)
                                        END IF
                                        IF Flag THEN
                                            Line.Number = X!
                                            IF Line.Number <= 23 THEN
                                                Xcoor = Line.Number
                                            ELSE
                                                Xcoor = 1
                                            END IF
                                            Ycoor = 1
                                            Column = 1
                                            LastSearch = X!
                                            EXIT FOR
                                        END IF
                                    NEXT
                                END IF
                                IF Flag = 0 THEN
                                    x = MessageBox(" Search ", "Text not found.")
                                END IF
                            END IF
                        END IF
                        COLOR 15, 0
                        GOSUB Display.Screen
                    END IF
                CASE 15 ' shift-tab
                    IF Column + Ycoor - 8 > False THEN
                        IF Column + Ycoor - 7 > 1 THEN
                            Ycoor = Ycoor - 7
                        ELSE
                            IF Column > 8 THEN
                                Column = Column - 7
                                GOSUB Display.Screen
                            END IF
                        END IF
                    ELSE ' home
                        IF Column > 1 THEN
                            Column = 1
                            Ycoor = 1
                            GOSUB Display.Screen
                        ELSE
                            Column = 1
                            Ycoor = 1
                        END IF
                    END IF
                CASE 83, 147 ' delete/control-delete
                    IF Line.Number > 0 THEN
                        ' remove line
                        IF Temp.ArrayS(Line.Number) = "" THEN
                            IF Line.Number = 1 AND Max.Lines = 1 THEN
                                Line.Number = 0
                                Max.Lines = 0
                                GOSUB Display.Screen
                            ELSE
                                ' pack array
                                FOR Temp1 = Line.Number TO Max.Lines - 1
                                    Temp.ArrayS(Temp1) = Temp.ArrayS(Temp1 + 1)
                                NEXT
                                IF Line.Number = Max.Lines THEN
                                    Line.Number = Line.Number - 1
                                    IF Xcoor > 1 THEN
                                        Xcoor = Xcoor - 1
                                    END IF
                                END IF
                                Max.Lines = Max.Lines - 1
                                Column = 1
                                Ycoor = 1
                                GOSUB Display.Screen
                            END IF
                        ELSE
                            ' concatenate line
                            IF Column + Ycoor - 1 > LEN(Temp.ArrayS(Line.Number)) THEN
                                IF Line.Number + 1 <= Max.Lines THEN
                                    T$ = Temp.ArrayS(Line.Number)
                                    T$ = T$ + Temp.ArrayS(Line.Number + 1)
                                    Temp.ArrayS(Line.Number) = T$

                                    ' pack array
                                    FOR Temp1 = Line.Number + 1 TO Max.Lines - 1
                                        Temp.ArrayS(Temp1) = Temp.ArrayS(Temp1 + 1)
                                    NEXT
                                    Max.Lines = Max.Lines - 1
                                    GOSUB Display.Screen
                                END IF
                            ELSE
                                ' delete character
                                T$ = Temp.ArrayS(Line.Number)
                                T$ = LEFT$(T$, Column + Ycoor - 2) + MID$(T$, Column + Ycoor)
                                Temp.ArrayS(Line.Number) = T$
                                GOSUB Display.Screen
                            END IF
                        END IF
                    END IF
                CASE 71 ' home
                    IF Column > 1 THEN
                        Column = 1
                        Ycoor = 1
                        GOSUB Display.Screen
                    ELSE
                        Column = 1
                        Ycoor = 1
                    END IF
                CASE 79 ' end
                    End.Line = LEN(Temp.ArrayS(Line.Number))
                    IF End.Line >= 80 THEN
                        Column = End.Line - 78
                        Ycoor = End.Line - Column + 2
                        GOSUB Display.Screen
                    ELSE
                        Ycoor = End.Line - Column + 2
                    END IF
                CASE 75, 155 ' left/alt-left
                    IF Ycoor > 1 THEN
                        Ycoor = Ycoor - 1
                    ELSE
                        IF Column > 1 THEN
                            Column = Column - 1
                            GOSUB Display.Screen
                        END IF
                    END IF
                CASE 77, 157 ' right/alt-right
                    IF Column + Ycoor - 1 <= LEN(Temp.ArrayS(Line.Number)) THEN
                        IF Ycoor < 80 THEN
                            Ycoor = Ycoor + 1
                        ELSE
                            Column = Column + 1
                            GOSUB Display.Screen
                        END IF
                    END IF
                CASE 72, 141, 152 ' up/ctrl-up/alt-up
                    Var = False
                    IF Line.Number > 1 THEN
                        Line.Number = Line.Number - 1
                        IF Xcoor = 1 THEN
                            Var = -1
                        ELSE
                            Xcoor = Xcoor - 1
                        END IF
                        End.Line = LEN(Temp.ArrayS(Line.Number)) + 1
                        IF Column + Ycoor - 1 > End.Line THEN
                            Ycoor = End.Line - Column + 1
                            IF Ycoor <= 0 THEN
                                Ycoor = End.Line
                                IF Column > 1 THEN
                                    Var = -1
                                END IF
                                Column = 1
                            END IF
                        END IF
                        IF Var THEN
                            GOSUB Display.Screen
                        END IF
                    END IF
                CASE 80, 145, 160 ' down/ctrl-dn/alt-dn
                    Var = False
                    IF Line.Number < Max.Lines THEN
                        Line.Number = Line.Number + 1
                        IF Xcoor >= 23 THEN
                            Var = -1
                        ELSE
                            Xcoor = Xcoor + 1
                        END IF
                        End.Line = LEN(Temp.ArrayS(Line.Number)) + 1
                        IF Column + Ycoor - 1 > End.Line THEN
                            Ycoor = End.Line - Column + 1
                            IF Ycoor <= 0 THEN
                                Ycoor = End.Line
                                IF Column > 1 THEN
                                    Var = -1
                                END IF
                                Column = 1
                            END IF
                        END IF
                        IF Var THEN
                            GOSUB Display.Screen
                        END IF
                    END IF
                CASE 73, 132 ' pageup/ctrl-pageup
                    IF Line.Number > 1 THEN
                        IF Line.Number - 22 > 1 THEN
                            Line.Number = Line.Number - 22
                        ELSE
                            Line.Number = 1
                        END IF
                        Xcoor = 1
                        End.Line = LEN(Temp.ArrayS(Line.Number)) + 1
                        IF Column + Ycoor - 1 > End.Line THEN
                            Ycoor = End.Line - Column + 1
                            IF Ycoor <= 0 THEN
                                Ycoor = End.Line
                                IF Column > 1 THEN
                                    Var = -1
                                END IF
                                Column = 1
                            END IF
                        END IF
                        GOSUB Display.Screen
                    END IF
                CASE 81, 118 ' pagedown/ctrl-pagedn
                    IF Line.Number < Max.Lines THEN
                        IF Line.Number + 22 < Max.Lines THEN
                            Line.Number = Line.Number + 22
                            Xcoor = 1
                        ELSE
                            Line.Number = Max.Lines
                            IF Max.Lines < 22 THEN
                                Xcoor = Max.Lines
                            ELSE
                                Xcoor = 22
                            END IF
                        END IF
                        End.Line = LEN(Temp.ArrayS(Line.Number)) + 1
                        IF Column + Ycoor - 1 > End.Line THEN
                            Ycoor = End.Line - Column + 1
                            IF Ycoor <= 0 THEN
                                Ycoor = End.Line
                                IF Column > 1 THEN
                                    Var = -1
                                END IF
                                Column = 1
                            END IF
                        END IF
                        GOSUB Display.Screen
                    END IF
                CASE 82 ' insert
                    Insert = NOT Insert
                    IF Insert THEN
                        LOCATE , , , 8, 8
                    ELSE
                        LOCATE , , , 0, 8
                    END IF
                CASE 116 ' control-right
                    IF Column + Ycoor - 1 > LEN(Temp.ArrayS(Line.Number)) THEN ' eol
                        IF Line.Number + 1 <= Max.Lines THEN
                            Line.Number = Line.Number + 1
                            IF Xcoor = 22 THEN
                                Var = -1
                            ELSE
                                Xcoor = Xcoor + 1
                            END IF
                            Ycoor = 1
                            Column = 1
                            GOSUB Display.Screen
                        END IF
                    ELSE
                        P = 0
                        FOR Temp2 = Column + Ycoor - 1 TO LEN(Temp.ArrayS(Line.Number))
                            P = P + 1
                            IF MID$(Temp.ArrayS(Line.Number), Temp2, 1) = " " THEN
                                IF MID$(Temp.ArrayS(Line.Number), Temp2 + 1, 1) <> " " THEN
                                    EXIT FOR
                                END IF
                            END IF
                        NEXT
                        'shift right
                        FOR Var3 = 1 TO P
                            IF Ycoor < 80 THEN
                                Ycoor = Ycoor + 1
                            ELSE
                                Column = Column + 1
                            END IF
                        NEXT
                        GOSUB Display.Screen
                    END IF
                CASE 115 ' control-left
                    IF Column + Ycoor - 1 <= 1 THEN
                        IF Line.Number > 1 THEN
                            Line.Number = Line.Number - 1
                            IF Xcoor = 1 THEN
                                Var = -1
                            ELSE
                                Xcoor = Xcoor - 1
                            END IF
                            End.Line = LEN(Temp.ArrayS(Line.Number))
                            IF End.Line >= 80 THEN
                                Column = End.Line - 78
                                Ycoor = End.Line - Column + 2
                            ELSE
                                Ycoor = End.Line - Column + 2
                            END IF
                            GOSUB Display.Screen
                        END IF
                    ELSE
                        P = 0
                        V = 0
                        FOR Temp2 = Column + Ycoor - 3 TO 1 STEP -1
                            P = P + 1
                            IF MID$(Temp.ArrayS(Line.Number), Temp2, 1) = " " THEN
                                IF MID$(Temp.ArrayS(Line.Number), Temp2 + 1, 1) <> " " THEN
                                    V = -1
                                    EXIT FOR
                                END IF
                            END IF
                        NEXT
                        IF V = 0 THEN
                            P = P + 1
                        END IF
                        ' shift left
                        FOR Var3 = 1 TO P
                            IF Ycoor > 1 THEN
                                Ycoor = Ycoor - 1
                            ELSE
                                IF Column > 1 THEN
                                    Column = Column - 1
                                END IF
                            END IF
                        NEXT
                        GOSUB Display.Screen
                    END IF
                CASE 119 ' control-home
                    IF Max.Lines > 0 THEN
                        Line.Number = 1
                        Xcoor = 1
                        Ycoor = 1
                        Column = 1
                        GOSUB Display.Screen
                    END IF
                CASE 117 ' control-end
                    IF Max.Lines > 0 THEN
                        Line.Number = Max.Lines
                        IF Max.Lines < 22 THEN
                            Xcoor = Max.Lines
                        ELSE
                            Xcoor = 22
                        END IF
                        Ycoor = 1
                        Column = 1
                        GOSUB Display.Screen
                    END IF
            END SELECT
    END SELECT
LOOP
COLOR 15, 1
LOCATE 25, 1
IF Yes THEN
    GOSUB SaveFile
    VarZ = -1
END IF
Status = StoreStatus
IF CurrentMenu = 1 AND CurrentMenuSelection = 6 THEN
    IF QuitBox THEN
        CALL WriteConfig
        CALL WriteProfile
        COLOR 7, 0
        SYSTEM
    END IF
END IF
EXIT SUB

' write edit array to disk.
SaveFile:
Tempfile1 = FREEFILE
OPEN Filename FOR OUTPUT AS #Tempfile1
FOR Count! = 1! TO Max.Lines
    PRINT #Tempfile1, Temp.ArrayS(Count!)
NEXT
CLOSE #Tempfile1
RETURN

' dumps current screen in hex.
Screen.Dump:
IF Max.Lines = 0 THEN
    RETURN
END IF

VarX$ = DumpFile
IF VarX$ = "" THEN
    RETURN
END IF

Error.Flag = True
DataError = False
x = FREEFILE
OPEN VarX$ FOR OUTPUT AS #x
Error.Flag = False
IF DataError THEN
    RETURN
END IF

Count = 0
DO
    Count = Count + 1
    IF Count = 24 THEN
        EXIT DO
    END IF
    IF Count + Line.Number - Xcoor > Max.Lines THEN
        EXIT DO
    END IF
    Strng = MID$(Temp.ArrayS(Count + Line.Number - Xcoor), Column)
    IF LEN(Strng) > 80 THEN
        Strng = LEFT$(Strng, 80)
    END IF
    Strng2$ = ""
    FOR LineQ = 1 TO LEN(Strng)
        Strng2$ = Strng2$ + RIGHT$("00" + HEX$(ASC(MID$(Strng, LineQ, 1))), 2) + " "
    NEXT
    PRINT #x, Strng2$
LOOP
CLOSE #x
RETURN

' print current screen in hex.
Screen.Print:
IF Max.Lines = 0 THEN
    RETURN
END IF

Count = 0
DO
    Count = Count + 1
    IF Count = 24 THEN
        EXIT DO
    END IF
    IF Count + Line.Number - Xcoor > Max.Lines THEN
        EXIT DO
    END IF
    Strng = MID$(Temp.ArrayS(Count + Line.Number - Xcoor), Column)
    IF LEN(Strng) > 80 THEN
        Strng = LEFT$(Strng, 80)
    END IF
    Strng2$ = ""
    FOR LineQ = 1 TO LEN(Strng)
        Strng2$ = Strng2$ + RIGHT$("00" + HEX$(ASC(MID$(Strng, LineQ, 1))), 2) + " "
    NEXT
    LPRINT Strng2$
LOOP
LPRINT CHR$(12);
RETURN

' dumps current file in hex.
File.Dump:
IF Max.Lines = 0 THEN
    RETURN
END IF

VarX$ = DumpFile
IF VarX$ = "" THEN
    RETURN
END IF

Error.Flag = True
DataError = False
x = FREEFILE
OPEN VarX$ FOR OUTPUT AS #x
Error.Flag = False
IF DataError THEN
    RETURN
END IF

FOR Count& = 1 TO Max.Lines
    Strng = Temp.ArrayS(Count&)
    Strng2$ = ""
    FOR LineQ& = 1 TO LEN(Strng)
        Strng2$ = Strng2$ + RIGHT$("00" + HEX$(ASC(MID$(Strng, LineQ&, 1))), 2) + " "
    NEXT
    PRINT #x, Strng2$
NEXT
CLOSE #x
RETURN

' print current file in hex.
File.Print:
IF Max.Lines = 0 THEN
    RETURN
END IF
FOR Count& = 1 TO Max.Lines
    Strng = Temp.ArrayS(Count&)
    Strng2$ = ""
    FOR LineQ& = 1 TO LEN(Strng)
        Strng2$ = Strng2$ + RIGHT$("00" + HEX$(ASC(MID$(Strng, LineQ&, 1))), 2) + " "
    NEXT
    LPRINT Strng2$
NEXT
LPRINT CHR$(12);
RETURN

' display current edit screen
Display.Screen:
COLOR 15, 0
CLS
CALL Toolbar
LOCATE 2, 1

' display edit lines
Count = 0
IF Max.Lines = 0 THEN
    Count = 1
END IF
IF Max.Lines > 0 THEN
    DO
        Count = Count + 1
        IF Count = 24 THEN
            EXIT DO
        END IF
        IF Count + Line.Number - Xcoor > Max.Lines THEN
            EXIT DO
        END IF
        LOCATE Count + 1, 1
        Strng = MID$(Temp.ArrayS(Count + Line.Number - Xcoor), Column)
        IF LEN(Strng) > 80 THEN
            Strng = LEFT$(Strng, 80)
        END IF
        _CONTROLCHR OFF
        PRINT Strng;
        _CONTROLCHR ON
    LOOP
END IF

' eof marker
IF Count < 24 THEN
    LOCATE Count + 1, 1
    COLOR 10, 0
    Strng = "<end of file>"
    PRINT Strng;
END IF

' display filename
COLOR 15, 1
LOCATE 25, 1
Strng = SPACE$(80)
PRINT Strng;
LOCATE 25, 42
VarX$ = UCASE$(AmbiguateFile$(Filename))
IF LEN(VarX$) > 30 THEN
    VarX$ = LEFT$(VarX$, 27) + "..."
END IF
Strng = "File: " + VarX$
PRINT Strng;
RETURN

Jump.Line:
VarX$ = "File: " + UCASE$(AmbiguateFile$(Filename))
Var2$ = "Lines: " + FormatString$(Max.Lines)
X$ = JumpBox(2, VarX$, Var2$)
VarX = -1
IF LEN(X$) = 0 THEN
    VarX = 0
ELSE
    IF LEN(X$) > 7 THEN
        VarX = 0
    ELSE
        VarX = IsNumeric(X$)
    END IF
END IF
Count! = 0!
IF VarX THEN
    IF VAL(X$) > 0 AND VAL(X$) <= Max.Lines THEN
        Count! = VAL(X$)
    ELSE
        VarX = 0
    END IF
END IF
RETURN

Jump.Byte:
' count bytes in edit array
X# = 0#
FOR X! = 1 TO Max.Lines
    X# = X# + LEN(Temp.ArrayS(X!))
NEXT
VarX$ = "File: " + UCASE$(AmbiguateFile$(Filename))
Var2$ = "Bytes: " + FormatString$(X#)
X$ = JumpBox(1, VarX$, Var2$)
VarX = -1
IF LEN(X$) = 0 THEN
    VarX = 0
ELSE
    IF LEN(X$) > 15 THEN
        VarX = 0
    ELSE
        VarX = IsNumeric(X$)
    END IF
END IF
T# = 0#
IF VarX THEN
    Count# = VAL(X$)
    IF Count# > 0 AND Count# <= X# THEN
        FOR Count! = 1 TO Max.Lines
            IF T# + LEN(Temp.ArrayS(Count!)) >= Count# THEN
                EXIT FOR
            END IF
            T# = T# + LEN(Temp.ArrayS(Count!))
        NEXT
    ELSE
        VarX = 0
    END IF
END IF
RETURN

' display percentage scanned in title
Display.Percent:
IF File.Size > 0 THEN
    Count2# = Count2# + LEN(LineX$) + 2
    IF Count2# > File.Size THEN
        Count2# = File.Size
    END IF
    Percent = (Count2# / File.Size) * 100!
    Percent = INT(Percent)
    IF Percent <> Last.Percent THEN
        Last.Percent = Percent
        P$ = RIGHT$(STR$(Percent + 1000%), 3) + "%"
        V$ = Program + " - EDIT - " + P$
        _TITLE V$
    END IF
END IF
RETURN
END SUB

FUNCTION ValidFileChar (Var$)
V$ = "\/:*?<>|" + CHR$(34)

' trim trailing dots and spaces
DO
    IF RIGHT$(Var$, 1) = "." THEN
        Var$ = LEFT$(Var$, LEN(Var$) - 1)
    ELSE
        IF RIGHT$(Var$, 1) = " " THEN
            Var$ = LEFT$(Var$, LEN(Var$) - 1)
        ELSE
            EXIT DO
        END IF
    END IF
LOOP
IF Var$ = "" THEN
    ValidFileChar = -1
    EXIT FUNCTION
END IF

FOR V = 1 TO LEN(Var$)
    Z = ASC(MID$(Var$, V, 1))
    IF Z >= 1 AND Z <= 31 THEN
        ValidFileChar = -1
        EXIT FUNCTION
    END IF
NEXT
FOR V = 1 TO LEN(Var$)
    IF INSTR(V$, MID$(Var$, V, 1)) THEN
        ValidFileChar = -1
        EXIT FUNCTION
    END IF
NEXT
ValidFileChar = 0
END FUNCTION

FUNCTION TestFile (Var$)
' process error
DataError = 0
Error.Flag = -1
X = FREEFILE
OPEN Var$ FOR RANDOM AS #X
Error.Flag = 0
IF DataError THEN
    TestFile = 0
    EXIT FUNCTION
END IF
CLOSE #X
TestFile = -1
END FUNCTION

FUNCTION TestFile2 (Var$)
' process error
DataError = 0
Error.Flag = -1
X = FREEFILE
OPEN Var$ FOR INPUT AS #X
Error.Flag = 0
IF DataError THEN
    TestFile2 = 0
    EXIT FUNCTION
END IF
CLOSE #X
TestFile2 = -1
END FUNCTION

FUNCTION JumpBox$ (Var1, Var1$, Var2$)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare jump box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF JumpXcoor THEN Xcoor3 = JumpXcoor
IF JumpYcoor THEN Ycoor3 = JumpYcoor

XC = Xcoor3 + 4
YC = Ycoor3 + 8
LC = 22
IF Var1 = 1 THEN
    QZ = 15 ' double
ELSE
    QZ = 7 ' single
END IF

' draw box
BoxButton = 1
GOSUB DrawJumpBox

' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
Insert = -1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE UCASE$(I$)
                    CASE "O" ' ok
                        BoxButton = 1
                        EXIT DO
                    CASE "C" ' cancel
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(8) ' backspace
                        IF LEN(X$) THEN
                            IF Xposition2 > 0 THEN
                                IF Insert THEN ' destructive backspace
                                    X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                                END IF
                                Xposition2 = Xposition2 - 1
                                IF Xposition2 - LC <= Xposition1 THEN
                                    Xposition1 = Xposition1 - 1
                                    IF Xposition1 < 1 THEN
                                        Xposition1 = 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawJumpBoxButtons
                        GOSUB DisplayFileLine
                    CASE CHR$(1) ' ctrl-a
                        BackGround8 = BackGround8 + 1
                        IF BackGround8 = 8 THEN
                            BackGround8 = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(2) ' ctrl-b
                        BackGround9 = BackGround9 + 1
                        IF BackGround9 = 8 THEN
                            BackGround9 = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(4) ' ctrl-d
                        JumpBoxBorderColor = JumpBoxBorderColor + 1
                        IF JumpBoxBorderColor = 16 THEN
                            JumpBoxBorderColor = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(5) ' ctrl-e
                        JumpBoxTitleColor = JumpBoxTitleColor + 1
                        IF JumpBoxTitleColor = 16 THEN
                            JumpBoxTitleColor = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(6) ' ctrl-f
                        JumpBoxTextColor = JumpBoxTextColor + 1
                        IF JumpBoxTextColor = 16 THEN
                            JumpBoxTextColor = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(7) ' ctrl-g
                        JumpBoxButton1Color = JumpBoxButton1Color + 1
                        IF JumpBoxButton1Color = 16 THEN
                            JumpBoxButton1Color = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE CHR$(11) ' ctrl-k
                        JumpBoxButton2Color = JumpBoxButton2Color + 1
                        IF JumpBoxButton2Color = 16 THEN
                            JumpBoxButton2Color = 0
                        END IF
                        GOSUB DrawJumpBox
                    CASE ELSE
                        ' append/insert character into edit area.
                        IF I$ >= "0" AND I$ <= "9" THEN
                            IF Insert THEN
                                IF LEN(X$) < QZ THEN
                                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > LC THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                END IF
                            ELSE
                                IF Xposition2 + 1 > LEN(X$) THEN
                                    IF LEN(X$) < QZ THEN
                                        X$ = X$ + I$ ' append
                                        Xposition2 = Xposition2 + 1
                                        IF Xposition2 > LC THEN
                                            Xposition1 = Xposition1 + 1
                                        END IF
                                    END IF
                                ELSE ' insert
                                    MID$(X$, Xposition2 + 1, 1) = I$
                                    Xposition2 = Xposition2 + 1
                                END IF
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        Insert = NOT Insert
                        IF Insert THEN
                            LOCATE , , 1, 8, 8 ' line cursor
                        ELSE
                            LOCATE , , 1, 0, 8 ' block cursor
                        END IF
                    CASE 83 ' Delete
                        IF LEN(X$) > 0 THEN
                            X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                            GOSUB DisplayFileLine
                        END IF
                    CASE 71 ' Home
                        IF Xposition2 > 0 THEN
                            Xposition1 = 1
                            Xposition2 = 0
                            GOSUB DisplayFileLine
                        END IF
                    CASE 79 ' End
                        IF Xposition2 < LEN(X$) THEN
                            Xposition2 = LEN(X$)
                            Xposition1 = Xposition2 - LC + 1
                            IF Xposition1 < 1 THEN
                                Xposition1 = 1
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                    CASE 75 ' Left
                        IF LEN(X$) THEN
                            IF Xposition2 > 0 THEN
                                Xposition2 = Xposition2 - 1
                                IF Xposition2 - LC <= Xposition1 THEN
                                    Xposition1 = Xposition1 - 1
                                    IF Xposition1 < 1 THEN
                                        Xposition1 = 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF LEN(X$) THEN
                            IF Xposition2 < LEN(X$) THEN
                                Xposition2 = Xposition2 + 1
                                IF Xposition2 > Xposition1 + LC - 1 THEN
                                    Xposition1 = Xposition1 + 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 15 ' shift-tab
                        IF BoxButton = 2 THEN
                            BoxButton = 1
                        ELSE
                            BoxButton = 2
                        END IF
                        GOSUB DrawJumpBoxButtons
                        GOSUB DisplayFileLine
                    CASE 141 ' ctrl-up
                        I$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        I$ = ""
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        I$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        I$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        I$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawJumpBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        I$ = ""
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawJumpBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        I$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawJumpBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        I$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawJumpBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawJumpBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 6 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 6 THEN
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                IF BoxButton = 2 THEN
                    BoxButton = 1
                    GOSUB DrawJumpBoxButtons
                    GOSUB DisplayFileLine
                END IF
            END IF
            IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                IF BoxButton = 1 THEN
                    BoxButton = 2
                    GOSUB DrawJumpBoxButtons
                    GOSUB DisplayFileLine
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
JumpXcoor = Xcoor3
JumpYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN
    JumpBox = X$
ELSE
    JumpBox = ""
END IF
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawJumpBox:
COLOR JumpBoxBorderColor, BackGround8
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR JumpBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 12, 0
PRINT " Jump ";

' display jump text
COLOR JumpBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
PRINT Var1$;
LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
PRINT Var2$;
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
IF Var1 = 1 THEN
    PRINT "Enter byte in line to jump to:";
ELSE
    PRINT "Enter line to jump to:";
END IF
LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
PRINT "Entry: ";
GOSUB DrawJumpBoxButtons
XC = Xcoor3 + 4
YC = Ycoor3 + 8
GOSUB DisplayFileLine
LOCATE , , , 8, 8
RETURN

' display buttuns
DrawJumpBoxButtons:
IF BoxButton = 1 THEN
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR JumpBoxButton1Color, BackGround9
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR JumpBoxButton2Color, BackGround9
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
ELSE
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR JumpBoxButton2Color, BackGround9
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR JumpBoxButton1Color, BackGround9
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
END IF
COLOR 7, 0
RETURN

' display current entry area.
DisplayFileLine:
LOCATE XC, YC, 1
PRINT SPACE$(LC);

' display edit line and position cursor.
LOCATE XC, YC, 1
IF LEN(X$) THEN
    COLOR 15
    IF MaskInput THEN
        LV = LEN(MID$(X$, Xposition1, LC))
        PRINT STRING$(LV, MaskChar$);
    ELSE
        PRINT MID$(X$, Xposition1, LC);
    END IF
END IF
LOCATE XC, YC + Xposition2 - Xposition1 + 1, 1
COLOR 7, 0
RETURN

ToggleInsert:
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN
END FUNCTION

REM Append bytes:
REM   Var1 = 1 - Hex
REM   Var1 = 2 - Ascii
REM   Var1 = 3 - Scrnedit
FUNCTION AppendBox$ (Var1)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare append box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF AppendXcoor THEN Xcoor3 = AppendXcoor
IF AppendYcoor THEN Ycoor3 = AppendYcoor

XC = Xcoor3 + 4
YC = Ycoor3 + 8
LC = 22

' draw box
BoxButton = 1
GOSUB DrawAppendBox

' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
Insert = -1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE UCASE$(I$)
                    CASE CHR$(8) ' backspace
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 > 0 THEN
                                    IF Insert THEN ' destructive backspace
                                        X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                                    END IF
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE CHR$(13)
                        IF BoxButton = 3 THEN
                            BoxButton = 0
                        ELSE
                            BoxButton = -1
                        END IF
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 0
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        IF BoxButton = 1 THEN
                            GOSUB DrawAppendBoxButtons
                            GOSUB DisplayFileLine
                        ELSE
                            GOSUB DrawAppendBoxButtons
                        END IF
                    CASE CHR$(1) ' ctrl-a
                        BackGround12 = BackGround12 + 1
                        IF BackGround12 = 8 THEN
                            BackGround12 = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(2) ' ctrl-b
                        BackGround13 = BackGround13 + 1
                        IF BackGround13 = 8 THEN
                            BackGround13 = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(4) ' ctrl-d
                        AppendBoxBorderColor = AppendBoxBorderColor + 1
                        IF AppendBoxBorderColor = 16 THEN
                            AppendBoxBorderColor = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(5) ' ctrl-e
                        AppendBoxTitleColor = AppendBoxTitleColor + 1
                        IF AppendBoxTitleColor = 16 THEN
                            AppendBoxTitleColor = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(6) ' ctrl-f
                        AppendBoxTextColor = AppendBoxTextColor + 1
                        IF AppendBoxTextColor = 16 THEN
                            AppendBoxTextColor = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(7) ' ctrl-g
                        AppendBoxButton1Color = AppendBoxButton1Color + 1
                        IF AppendBoxButton1Color = 16 THEN
                            AppendBoxButton1Color = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE CHR$(11) ' ctrl-k
                        AppendBoxButton2Color = AppendBoxButton2Color + 1
                        IF AppendBoxButton2Color = 16 THEN
                            AppendBoxButton2Color = 0
                        END IF
                        GOSUB DrawAppendBox
                    CASE ELSE
                        IF BoxButton = 2 OR BoxButton = 3 THEN
                            IF UCASE$(I$) = "O" THEN
                                BoxButton = -1
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "C" THEN
                                BoxButton = 0
                                EXIT DO
                            END IF
                        END IF
                        IF BoxButton = 1 THEN
                            ' append/insert character into edit area.
                            OK = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                IF Var1 = 1 THEN
                                    OK = 0 ' disallow control charcters
                                END IF
                            END IF
                            IF OK THEN
                                IF Insert THEN
                                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > LC THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Xposition2 + 1 > LEN(X$) THEN
                                        X$ = X$ + I$ ' append
                                        Xposition2 = Xposition2 + 1
                                        IF Xposition2 > LC THEN
                                            Xposition1 = Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X$, Xposition2 + 1, 1) = I$
                                        Xposition2 = Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        IF BoxButton = 1 THEN
                            Insert = NOT Insert
                            IF Insert THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                    CASE 83 ' Delete
                        IF LEN(X$) > 0 THEN
                            IF BoxButton = 1 THEN
                                X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 71 ' Home
                        IF BoxButton = 1 THEN
                            IF Xposition2 > 0 THEN
                                Xposition1 = 1
                                Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 79 ' End
                        IF BoxButton = 1 THEN
                            IF Xposition2 < LEN(X$) THEN
                                Xposition2 = LEN(X$)
                                Xposition1 = Xposition2 - LC + 1
                                IF Xposition1 < 1 THEN
                                    Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 75 ' Left
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 > 0 THEN
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 < LEN(X$) THEN
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > Xposition1 + LC - 1 THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 15 ' shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 3
                        END IF
                        IF BoxButton = 1 THEN
                            GOSUB DrawAppendBoxButtons
                            GOSUB DisplayFileLine
                        ELSE
                            GOSUB DrawAppendBoxButtons
                        END IF
                    CASE 141 ' ctrl-up
                        I$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        I$ = ""
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        I$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        I$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        I$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawAppendBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        I$ = ""
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawAppendBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        I$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawAppendBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        I$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawAppendBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawAppendBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = XC THEN
                ' entry area
                IF MouseY >= YC AND MouseY <= YC + LC THEN
                    BoxButton = 1
                    GOSUB DrawAppendBox
                END IF
            END IF
            IF MouseX = Xcoor3 + 6 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = -1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 0
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
AppendXcoor = Xcoor3
AppendYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton THEN
    AppendBox = X$
ELSE
    AppendBox = ""
END IF
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawAppendBox:
COLOR AppendBoxBorderColor, BackGround12
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR AppendBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 11, 0
PRINT " Append ";

' display append text
COLOR AppendBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
IF Var1 = 1 THEN
    PRINT "Enter HEX bytes";
ELSE
    IF Var1 = 2 THEN
        PRINT "Enter ASCII string:";
    ELSE
        PRINT "Enter filename:";
    END IF
END IF
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
IF Var1 = 1 THEN
    PRINT "  (space separated):";
END IF

LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
PRINT "Entry: ";
GOSUB DrawAppendBoxButtons
LOCATE , , , 8, 8
XC = Xcoor3 + 4
YC = Ycoor3 + 8
GOSUB DisplayFileLine
RETURN

' display buttuns
DrawAppendBoxButtons:
IF BoxButton = 2 THEN
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR AppendBoxButton1Color, BackGround13
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR AppendBoxButton2Color, BackGround13
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";

    LOCATE Xcoor3 + 6, Ycoor3 + 2, 1, 8, 8 ' line cursor
ELSE
    IF BoxButton = 3 THEN
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR AppendBoxButton2Color, BackGround13
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
        COLOR AppendBoxButton1Color, BackGround13
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 1, 8, 8 ' line cursor
    ELSE
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR AppendBoxButton2Color, BackGround13
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
        COLOR AppendBoxButton2Color, BackGround13
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
        COLOR Magenta
        PRINT "C";
    END IF
END IF
COLOR 7, 0
RETURN

' display current entry area.
DisplayFileLine:
LOCATE XC, YC, 1
PRINT SPACE$(LC);

' display edit line and position cursor.
LOCATE XC, YC, 1
IF LEN(X$) THEN
    COLOR 15
    IF MaskInput THEN
        LV = LEN(MID$(X$, Xposition1, LC))
        PRINT STRING$(LV, MaskChar$);
    ELSE
        PRINT MID$(X$, Xposition1, LC);
    END IF
END IF
LOCATE XC, YC + Xposition2 - Xposition1 + 1, 1
COLOR 7, 0
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN

ToggleInsert:
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN
END FUNCTION

FUNCTION FilenameBox$ (Var1)
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen2

' declare append box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF FilenameXcoor THEN Xcoor3 = FilenameXcoor
IF FilenameYcoor THEN Ycoor3 = FilenameYcoor

' adjust xcoor
IF Status THEN
    Z = 23
    IF Xcoor3 + 7 > Z THEN
        Xcoor3 = Xcoor3 - 1
    END IF
END IF

XC = Xcoor3 + 4
YC = Ycoor3 + 8
LC = 22

' draw box
BoxButton = 1
GOSUB DrawFilenameBox

' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
Insert = -1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE UCASE$(I$)
                    CASE CHR$(8) ' backspace
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 > 0 THEN
                                    IF Insert THEN ' destructive backspace
                                        X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                                    END IF
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE CHR$(13)
                        IF BoxButton = 3 THEN
                            BoxButton = 0
                        ELSE
                            BoxButton = -1
                        END IF
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 0
                        EXIT DO
                    CASE CHR$(9) ' tab
                        BoxButton = BoxButton + 1
                        IF BoxButton = 4 THEN
                            BoxButton = 1
                        END IF
                        IF BoxButton = 1 THEN
                            GOSUB DrawFilenameBoxButtons
                            GOSUB DisplayFileLine
                        ELSE
                            GOSUB DrawFilenameBoxButtons
                        END IF
                    CASE CHR$(1) ' ctrl-a
                        BackGround14 = BackGround14 + 1
                        IF BackGround14 = 8 THEN
                            BackGround14 = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(2) ' ctrl-b
                        BackGround15 = BackGround15 + 1
                        IF BackGround15 = 8 THEN
                            BackGround15 = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(4) ' ctrl-d
                        FilenameBoxBorderColor = FilenameBoxBorderColor + 1
                        IF FilenameBoxBorderColor = 16 THEN
                            FilenameBoxBorderColor = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(5) ' ctrl-e
                        FilenameBoxTitleColor = FilenameBoxTitleColor + 1
                        IF FilenameBoxTitleColor = 16 THEN
                            FilenameBoxTitleColor = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(6) ' ctrl-f
                        FilenameBoxTextColor = FilenameBoxTextColor + 1
                        IF FilenameBoxTextColor = 16 THEN
                            FilenameBoxTextColor = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(7) ' ctrl-g
                        FilenameBoxButton1Color = FilenameBoxButton1Color + 1
                        IF FilenameBoxButton1Color = 16 THEN
                            FilenameBoxButton1Color = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE CHR$(11) ' ctrl-k
                        FilenameBoxButton2Color = FilenameBoxButton2Color + 1
                        IF FilenameBoxButton2Color = 16 THEN
                            FilenameBoxButton2Color = 0
                        END IF
                        GOSUB DrawFilenameBox
                    CASE ELSE
                        IF BoxButton = 2 OR BoxButton = 3 THEN
                            IF UCASE$(I$) = "O" THEN
                                BoxButton = -1
                                EXIT DO
                            END IF
                            IF UCASE$(I$) = "C" THEN
                                BoxButton = 0
                                EXIT DO
                            END IF
                        END IF
                        IF BoxButton = 1 THEN
                            ' append/insert character into edit area.
                            OK = -1
                            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                                IF Var1 = 1 THEN
                                    OK = 0 ' disallow control charcters
                                END IF
                            END IF
                            IF OK THEN
                                IF Insert THEN
                                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > LC THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                ELSE
                                    IF Xposition2 + 1 > LEN(X$) THEN
                                        X$ = X$ + I$ ' append
                                        Xposition2 = Xposition2 + 1
                                        IF Xposition2 > LC THEN
                                            Xposition1 = Xposition1 + 1
                                        END IF
                                    ELSE ' insert
                                        MID$(X$, Xposition2 + 1, 1) = I$
                                        Xposition2 = Xposition2 + 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        IF BoxButton = 1 THEN
                            Insert = NOT Insert
                            IF Insert THEN
                                LOCATE , , 1, 8, 8 ' line cursor
                            ELSE
                                LOCATE , , 1, 0, 8 ' block cursor
                            END IF
                        END IF
                    CASE 83 ' Delete
                        IF LEN(X$) > 0 THEN
                            IF BoxButton = 1 THEN
                                X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 71 ' Home
                        IF BoxButton = 1 THEN
                            IF Xposition2 > 0 THEN
                                Xposition1 = 1
                                Xposition2 = 0
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 79 ' End
                        IF BoxButton = 1 THEN
                            IF Xposition2 < LEN(X$) THEN
                                Xposition2 = LEN(X$)
                                Xposition1 = Xposition2 - LC + 1
                                IF Xposition1 < 1 THEN
                                    Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 75 ' Left
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 > 0 THEN
                                    Xposition2 = Xposition2 - 1
                                    IF Xposition2 - LC <= Xposition1 THEN
                                        Xposition1 = Xposition1 - 1
                                        IF Xposition1 < 1 THEN
                                            Xposition1 = 1
                                        END IF
                                        GOSUB DisplayFileLine
                                    END IF
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF BoxButton = 1 THEN
                            IF LEN(X$) THEN
                                IF Xposition2 < LEN(X$) THEN
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > Xposition1 + LC - 1 THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 15 ' shift-tab
                        BoxButton = BoxButton - 1
                        IF BoxButton = 0 THEN
                            BoxButton = 3
                        END IF
                        IF BoxButton = 1 THEN
                            GOSUB DrawFilenameBoxButtons
                            GOSUB DisplayFileLine
                        ELSE
                            GOSUB DrawFilenameBoxButtons
                        END IF
                    CASE 141 ' ctrl-up
                        I$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        I$ = ""
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        IF Xcoor3 + 7 < Z THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        I$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        I$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        I$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen2
                                GOSUB DrawFilenameBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        I$ = ""
                        IF Status THEN
                            Z = 23
                        ELSE
                            Z = 24
                        END IF
                        IF Xcoor3 + 7 < Z - 4 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        ELSE
                            IF Xcoor3 + 7 < Z THEN
                                Xcoor3 = Z - 7
                                CALL RestoreScreen2
                                GOSUB DrawFilenameBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        I$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen2
                                GOSUB DrawFilenameBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        I$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen2
                                GOSUB DrawFilenameBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    IF Status THEN
        Z = 23
    ELSE
        Z = 24
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= Z - 7 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen2
                            GOSUB DrawFilenameBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = XC THEN
                ' entry area
                IF MouseY >= YC AND MouseY <= YC + LC THEN
                    BoxButton = 1
                    GOSUB DrawFilenameBox
                END IF
            END IF
            IF MouseX = Xcoor3 + 6 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = -1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 0
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
FilenameXcoor = Xcoor3
FilenameYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen2
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton THEN
    FilenameBox = X$
ELSE
    FilenameBox = ""
END IF
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawFilenameBox:
COLOR FilenameBoxBorderColor, BackGround14
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR FilenameBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 10, 0
PRINT " Filename ";

' display append text
COLOR FilenameBoxTextColor
LOCATE Xcoor3 + 2, Ycoor3 + 1, 0
IF Var1 = 1 THEN
    PRINT "Enter HEX bytes";
ELSE
    IF Var1 = 2 THEN
        PRINT "Enter ASCII string:";
    ELSE
        PRINT "Enter filename:";
    END IF
END IF
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
IF Var1 = 1 THEN
    PRINT "  (space separated):";
END IF

LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
PRINT "Entry: ";
GOSUB DrawFilenameBoxButtons
LOCATE , , , 8, 8
XC = Xcoor3 + 4
YC = Ycoor3 + 8
GOSUB DisplayFileLine
RETURN

' display buttuns
DrawFilenameBoxButtons:
IF BoxButton = 2 THEN
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR FilenameBoxButton1Color, BackGround15
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR FilenameBoxButton2Color, BackGround15
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";

    LOCATE Xcoor3 + 6, Ycoor3 + 2, 1, 8, 8 ' line cursor
ELSE
    IF BoxButton = 3 THEN
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR FilenameBoxButton2Color, BackGround15
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
        COLOR FilenameBoxButton1Color, BackGround15
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
        COLOR Magenta
        PRINT "C";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 1, 8, 8 ' line cursor
    ELSE
        LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
        COLOR FilenameBoxButton2Color, BackGround15
        PRINT "<OK>";
        LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
        COLOR Magenta
        PRINT "O";

        LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
        COLOR FilenameBoxButton2Color, BackGround15
        PRINT "<Cancel>";
        LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
        COLOR Magenta
        PRINT "C";
    END IF
END IF
COLOR 7, 0
RETURN

' display current entry area.
DisplayFileLine:
LOCATE XC, YC, 1
PRINT SPACE$(LC);

' display edit line and position cursor.
LOCATE XC, YC, 1
IF LEN(X$) THEN
    COLOR 15
    IF MaskInput THEN
        LV = LEN(MID$(X$, Xposition1, LC))
        PRINT STRING$(LV, MaskChar$);
    ELSE
        PRINT MID$(X$, Xposition1, LC);
    END IF
END IF
LOCATE XC, YC + Xposition2 - Xposition1 + 1, 1
COLOR 7, 0
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN

ToggleInsert:
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN
END FUNCTION

FUNCTION AsciiBox$
' clear mouse activity.
MouseButton1 = 0
MouseButton2 = 0
MouseButton3 = 0
CALL ClearMouse

' store screen area.
CurrentX = CSRLIN
CurrentY = POS(0)
CALL SaveScreen3

' declare ascii box coordinates.
Xcoor3 = 10
Ycoor3 = 10
IF AsciiXcoor THEN Xcoor3 = AsciiXcoor
IF AsciiYcoor THEN Ycoor3 = AsciiYcoor

XC = Xcoor3 + 4
YC = Ycoor3 + 8
LC = 22

' draw box
BoxButton = 1
GOSUB DrawAsciiBox

' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
Insert = -1
GOSUB DisplayFileLine

' wait for keypress or mouse
_KEYCLEAR
Control.Break = False
Monitor.Box = False
DO
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        GOSUB ToggleInsert
    END IF

    _LIMIT 30
    I$ = INKEYx$
    IF LEN(I$) THEN
        SELECT CASE LEN(I$)
            CASE 1
                SELECT CASE UCASE$(I$)
                    CASE "O" ' ok
                        BoxButton = 1
                        EXIT DO
                    CASE "C" ' cancel
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(8) ' backspace
                        IF LEN(X$) THEN
                            IF Xposition2 > 0 THEN
                                IF Insert THEN ' destructive backspace
                                    X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                                END IF
                                Xposition2 = Xposition2 - 1
                                IF Xposition2 - LC <= Xposition1 THEN
                                    Xposition1 = Xposition1 - 1
                                    IF Xposition1 < 1 THEN
                                        Xposition1 = 1
                                    END IF
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE CHR$(13)
                        EXIT DO
                    CASE CHR$(27)
                        BoxButton = 2
                        EXIT DO
                    CASE CHR$(9) ' tab
                        IF BoxButton = 1 THEN
                            BoxButton = 2
                        ELSE
                            BoxButton = 1
                        END IF
                        GOSUB DrawAsciiBoxButtons
                        GOSUB DisplayFileLine
                    CASE CHR$(1) ' ctrl-a
                        BackGround10 = BackGround10 + 1
                        IF BackGround10 = 8 THEN
                            BackGround10 = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(2) ' ctrl-b
                        Background11 = Background11 + 1
                        IF Background11 = 8 THEN
                            Background11 = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(4) ' ctrl-d
                        AsciiBoxBorderColor = AsciiBoxBorderColor + 1
                        IF AsciiBoxBorderColor = 16 THEN
                            AsciiBoxBorderColor = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(5) ' ctrl-e
                        AsciiBoxTitleColor = AsciiBoxTitleColor + 1
                        IF AsciiBoxTitleColor = 16 THEN
                            AsciiBoxTitleColor = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(6) ' ctrl-f
                        AsciiBoxTextColor = AsciiBoxTextColor + 1
                        IF AsciiBoxTextColor = 16 THEN
                            AsciiBoxTextColor = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(7) ' ctrl-g
                        AsciiBoxButton1Color = AsciiBoxButton1Color + 1
                        IF AsciiBoxButton1Color = 16 THEN
                            AsciiBoxButton1Color = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE CHR$(11) ' ctrl-k
                        AsciiBoxButton2Color = AsciiBoxButton2Color + 1
                        IF AsciiBoxButton2Color = 16 THEN
                            AsciiBoxButton2Color = 0
                        END IF
                        GOSUB DrawAsciiBox
                    CASE ELSE
                        ' append/insert character into edit area.
                        IF I$ >= "0" AND I$ <= "9" THEN
                            IF Insert THEN
                                IF LEN(X$) < 3 THEN
                                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                                    Xposition2 = Xposition2 + 1
                                    IF Xposition2 > LC THEN
                                        Xposition1 = Xposition1 + 1
                                    END IF
                                END IF
                            ELSE
                                IF Xposition2 + 1 > LEN(X$) THEN
                                    IF LEN(X$) < 3 THEN
                                        X$ = X$ + I$ ' append
                                        Xposition2 = Xposition2 + 1
                                        IF Xposition2 > LC THEN
                                            Xposition1 = Xposition1 + 1
                                        END IF
                                    END IF
                                ELSE ' insert
                                    MID$(X$, Xposition2 + 1, 1) = I$
                                    Xposition2 = Xposition2 + 1
                                END IF
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(I$, 1))
                    CASE 82 ' Insert
                        Insert = NOT Insert
                        IF Insert THEN
                            LOCATE , , 1, 8, 8 ' line cursor
                        ELSE
                            LOCATE , , 1, 0, 8 ' block cursor
                        END IF
                    CASE 83 ' Delete
                        IF LEN(X$) > 0 THEN
                            X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                            GOSUB DisplayFileLine
                        END IF
                    CASE 71 ' Home
                        IF Xposition2 > 0 THEN
                            Xposition1 = 1
                            Xposition2 = 0
                            GOSUB DisplayFileLine
                        END IF
                    CASE 79 ' End
                        IF Xposition2 < LEN(X$) THEN
                            Xposition2 = LEN(X$)
                            Xposition1 = Xposition2 - LC + 1
                            IF Xposition1 < 1 THEN
                                Xposition1 = 1
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                    CASE 75 ' Left
                        IF LEN(X$) THEN
                            IF Xposition2 > 0 THEN
                                Xposition2 = Xposition2 - 1
                                IF Xposition2 - LC <= Xposition1 THEN
                                    Xposition1 = Xposition1 - 1
                                    IF Xposition1 < 1 THEN
                                        Xposition1 = 1
                                    END IF
                                    GOSUB DisplayFileLine
                                END IF
                            END IF
                        END IF
                    CASE 77 ' Right
                        IF LEN(X$) THEN
                            IF Xposition2 < LEN(X$) THEN
                                Xposition2 = Xposition2 + 1
                                IF Xposition2 > Xposition1 + LC - 1 THEN
                                    Xposition1 = Xposition1 + 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    CASE 15 ' shift-tab
                        IF BoxButton = 2 THEN
                            BoxButton = 1
                        ELSE
                            BoxButton = 2
                        END IF
                        GOSUB DrawAsciiBoxButtons
                        GOSUB DisplayFileLine
                    CASE 141 ' ctrl-up
                        I$ = ""
                        IF Xcoor3 > 2 THEN
                            Xcoor3 = Xcoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        END IF
                        _KEYCLEAR
                    CASE 145 ' ctrl-down
                        I$ = ""
                        IF Xcoor3 < 17 THEN
                            Xcoor3 = Xcoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        END IF
                        _KEYCLEAR
                    CASE 115 ' ctrl-left
                        I$ = ""
                        IF Ycoor3 > 1 THEN
                            Ycoor3 = Ycoor3 - 1
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        END IF
                        _KEYCLEAR
                    CASE 116 ' ctrl-right
                        I$ = ""
                        IF Ycoor3 < 49 THEN
                            Ycoor3 = Ycoor3 + 1
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        END IF
                        _KEYCLEAR
                    CASE 152 ' alt-up
                        I$ = ""
                        IF Xcoor3 > 5 THEN
                            Xcoor3 = Xcoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        ELSE
                            IF Xcoor3 > 2 THEN
                                Xcoor3 = 2
                                CALL RestoreScreen3
                                GOSUB DrawAsciiBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 160 ' alt-dn
                        I$ = ""
                        IF Xcoor3 < 13 THEN
                            Xcoor3 = Xcoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        ELSE
                            IF Xcoor3 < 17 THEN
                                Xcoor3 = 17
                                CALL RestoreScreen3
                                GOSUB DrawAsciiBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 155 ' alt-left
                        I$ = ""
                        IF Ycoor3 > 4 THEN
                            Ycoor3 = Ycoor3 - 4
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        ELSE
                            IF Ycoor3 > 1 THEN
                                Ycoor3 = 1
                                CALL RestoreScreen3
                                GOSUB DrawAsciiBox
                            END IF
                        END IF
                        _KEYCLEAR
                    CASE 157 ' alt-right
                        I$ = ""
                        IF Ycoor3 < 45 THEN
                            Ycoor3 = Ycoor3 + 4
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        ELSE
                            IF Ycoor3 < 49 THEN
                                Ycoor3 = 49
                                CALL RestoreScreen3
                                GOSUB DrawAsciiBox
                            END IF
                        END IF
                        _KEYCLEAR
                END SELECT
        END SELECT
    END IF
    X = MouseDriver
    IF MouseButton1 THEN
        ' hover over titlebar
        IF MouseX = Xcoor3 THEN
            IF MouseY >= Ycoor3 AND MouseY <= Ycoor3 + 31 THEN
                ' store mouse XY during click
                MouseTempX = MouseX
                MouseTempY = MouseY
                DO
                    X = MouseDriver
                    IF MouseX OR MouseY THEN ' drag
                        MoveBox = 0
                        ' difference in mouse X
                        IF MouseX <> MouseTempX THEN
                            IF MouseX >= 2 AND MouseX <= 17 THEN
                                Xcoor3 = MouseX
                                MouseTempX = MouseX
                                MoveBox = -1
                            END IF
                        END IF
                        ' difference in mouse Y
                        IF MouseY <> MouseTempY THEN
                            MoveY = Ycoor3 + (MouseY - MouseTempY)
                            IF MoveY >= 1 AND MoveY <= 49 THEN
                                Ycoor3 = MoveY
                                MouseTempY = MouseY
                                MoveBox = -1
                            END IF
                        END IF
                        ' move box
                        IF MoveBox THEN
                            CALL RestoreScreen3
                            GOSUB DrawAsciiBox
                        END IF
                    END IF
                LOOP UNTIL MouseButton1 = 0
            END IF
        ELSE
            IF MouseX = Xcoor3 + 6 THEN
                IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                    BoxButton = 1
                    EXIT DO
                END IF
                IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                    BoxButton = 2
                    EXIT DO
                END IF
            END IF
        END IF
    ELSE
        IF MouseX = Xcoor3 + 6 THEN
            IF MouseY >= Ycoor3 + 2 AND MouseY <= Ycoor3 + 5 THEN
                IF BoxButton = 2 THEN
                    BoxButton = 1
                    GOSUB DrawAsciiBoxButtons
                    GOSUB DisplayFileLine
                END IF
            END IF
            IF MouseY >= Ycoor3 + 8 AND MouseY <= Ycoor3 + 15 THEN
                IF BoxButton = 1 THEN
                    BoxButton = 2
                    GOSUB DrawAsciiBoxButtons
                    GOSUB DisplayFileLine
                END IF
            END IF
        END IF
    END IF
LOOP
_DELAY .2
_KEYCLEAR
_DELAY .2

' store box coordinates
AsciiXcoor = Xcoor3
AsciiYcoor = Ycoor3

' restore screen area.
CALL RestoreScreen3
COLOR 7, 0
LOCATE CurrentX, CurrentY, 1
IF BoxButton = 1 THEN
    AsciiBox = X$
ELSE
    AsciiBox = ""
END IF
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' draw box
DrawAsciiBox:
COLOR AsciiBoxBorderColor, BackGround10
LOCATE Xcoor3, Ycoor3, 0
PRINT CHR$(ULcorner) + STRING$(30, Hline) + CHR$(URcorner);
FOR RowX1 = Xcoor3 + 1 TO Xcoor3 + 6
    LOCATE RowX1, Ycoor3, 0
    PRINT CHR$(Vline) + SPACE$(30) + CHR$(Vline);
NEXT
LOCATE Xcoor3 + 7, Ycoor3, 0
PRINT CHR$(LLcorner) + STRING$(30, Hline) + CHR$(LRcorner);
COLOR AsciiBoxTitleColor
LOCATE Xcoor3, Ycoor3 + 12, 0
PRINT " ASCII ";

' display ascii text
COLOR AsciiBoxTextColor
LOCATE Xcoor3 + 1, Ycoor3 + 1, 0
PRINT "ASCII code entry box"
LOCATE Xcoor3 + 3, Ycoor3 + 1, 0
PRINT "Enter 3-digit ASCII code: "
LOCATE Xcoor3 + 4, Ycoor3 + 1, 0
PRINT "Entry: ";
GOSUB DrawAsciiBoxButtons
LOCATE , , , 8, 8
XC = Xcoor3 + 4
YC = Ycoor3 + 8
GOSUB DisplayFileLine
RETURN

' display buttuns
DrawAsciiBoxButtons:
IF BoxButton = 1 THEN
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR AsciiBoxButton1Color, Background11
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR AsciiBoxButton2Color, Background11
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
ELSE
    LOCATE Xcoor3 + 6, Ycoor3 + 2, 0
    COLOR AsciiBoxButton2Color, Background11
    PRINT "<OK>";
    LOCATE Xcoor3 + 6, Ycoor3 + 3, 0
    COLOR Magenta
    PRINT "O";

    LOCATE Xcoor3 + 6, Ycoor3 + 8, 0
    COLOR AsciiBoxButton1Color, Background11
    PRINT "<Cancel>";
    LOCATE Xcoor3 + 6, Ycoor3 + 9, 0
    COLOR Magenta
    PRINT "C";
END IF
COLOR 7, 0
RETURN

' display current entry area.
DisplayFileLine:
LOCATE XC, YC, 1
PRINT SPACE$(LC);

' display edit line and position cursor.
LOCATE XC, YC, 1
IF LEN(X$) THEN
    COLOR 15
    IF MaskInput THEN
        LV = LEN(MID$(X$, Xposition1, LC))
        PRINT STRING$(LV, MaskChar$);
    ELSE
        PRINT MID$(X$, Xposition1, LC);
    END IF
END IF
LOCATE XC, YC + Xposition2 - Xposition1 + 1, 1
COLOR 7, 0
RETURN

ToggleInsert:
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
RETURN
END FUNCTION

' move to drive
SUB SwitchDrive (Var$)
DIM ASCIIZ AS STRING * 260
Var$ = UCASE$(Var$)
IF Var$ >= "A" AND Var$ <= "Z" THEN
    ' check drive
    V = ASC(Var$) - 64
    IF DRIVEEXISTS(V) THEN
        CALL DisplayError("Drive exists error accessing drive " + CHR$(V + 64) + ":")
        EXIT SUB
    END IF
    IF MEDIAEXISTS(V) = 0 THEN
        CALL DisplayError("Media exists error accessing drive " + CHR$(V + 64) + ":")
        EXIT SUB
    END IF
    ' make directory name
    Var2$ = RTRIM$(DriveTable(V))
    IF LEN(Var2$) THEN
        IF RIGHT$(Var2$, 1) = "\" THEN
            ASCIIZ = Var2$ + CHR$(0)
        ELSE
            ASCIIZ = Var2$ + "\" + CHR$(0)
        END IF
    ELSE
        ASCIIZ = Var$ + ":\" + CHR$(0)
    END IF
    x = SetCurrentDirectoryA(ASCIIZ)
    ' check error flag
    IF x = 0 THEN
        CALL DisplayWinError(x)
    END IF
    EXIT SUB
END IF
CALL DisplayError("Drive letter error accessing drive.")
END SUB

' move to netpath
SUB SwitchNetpath (Var$)
DIM ASCIIZ AS STRING * 260
ASCIIZ = Var$ + CHR$(0)
x = SetCurrentDirectoryA(ASCIIZ)
' check error flag
IF x = 0 THEN
    CALL DisplayWinError(x)
    EXIT SUB
END IF
' store netpath
FOR V = 1 TO NetPathCount
    IF UCASE$(_CWD$) = UCASE$(RTRIM$(NetPathTable(V))) THEN
        EXIT SUB
    END IF
NEXT
NetPathCount = NetPathCount + 1
REDIM _PRESERVE NetPathTable(NetPathCount) AS STRING
NetPathTable(NetPathCount) = _CWD$
END SUB

SUB NewDir (ErrFlag)
DIM ASCIIZ AS STRING * 260

' get current drive
Current.Drive = LEFT$(_CWD$, 1)

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get switches from command line
Continuous.Display = ParseLine("/C")
Display.Drive = ParseLine("/D")
Display.Path = ParseLine("/V")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = UCASE$(LEFT$(Command.Work, 1))
            MID$(Command.Work, 1, 1) = UCASE$(MID$(Command.Work, 1, 1))
        ELSE
            Drive.Search = Current.Drive
        END IF

        ' display current path
        IF Command.Work = Nul THEN
            COLOR Yellow, Black
            PRINT #StdoutHandle, _CWD$
            EXIT SUB
        END IF

        ' check netpath
        Flag = -1
        IF LEFT$(Command.Work, 2) <> "\\" THEN
            ' check drive
            IF MID$(Command.Work, 2, 1) = ":" THEN
                V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
                IF DRIVEEXISTS(V) THEN
                    ' make directory name
                    ASCIIZ = Drive.Search + ":\" + CHR$(0)
                    x = SetCurrentDirectoryA(ASCIIZ)
                    CALL DisplayWinError(x)
                    Flag = 0
                ELSE
                    IF MEDIAEXISTS(V) = 0 THEN
                        ' display any errors
                        CALL DisplayError("Error accessing drive " + CHR$(V + 64) + ":")
                        Flag = 0
                    END IF
                END IF
            END IF
        END IF
        IF Flag THEN
            ' make directory name
            ASCIIZ = Command.Work + CHR$(0)
            x = SetCurrentDirectoryA(ASCIIZ)

            ' check error flag
            IF x = 0 THEN
                CALL DisplayWinError(x)
            ELSE
                ' store path in table
                PathTable(1) = _CWD$

                ' store path in netpath table
                Var$ = _CWD$
                IF LEFT$(Var$, 2) = "\\" THEN
                    Flag = 0
                    FOR V = 1 TO NetPathCount
                        IF UCASE$(_CWD$) = UCASE$(RTRIM$(NetPathTable(V))) THEN
                            Flag = -1
                            EXIT FOR
                        END IF
                    NEXT
                    IF Flag = 0 THEN
                        NetPathCount = NetPathCount + 1
                        REDIM _PRESERVE NetPathTable(NetPathCount) AS STRING
                        NetPathTable(NetPathCount) = _CWD$
                    END IF
                ELSE
                    ' store path in drive table
                    V = ASC(LEFT$(Var$, 1)) - 64
                    DriveTable(V) = _CWD$
                END IF

                ' store drive letter
                Outpt$ = Nul
                IF Display.Drive = False THEN
                    IF LEFT$(_CWD$, 2) <> "\\" THEN
                        Outpt$ = Drive.Search$ + ":"
                    END IF
                END IF

                ' store pathname
                IF Display.Path = False THEN
                    IF LEFT$(_CWD$, 2) = "\\" THEN
                        Outpt$ = _CWD$
                    ELSE
                        Outpt$ = Outpt$ + MID$(_CWD$, 3)
                    END IF
                END IF

                ' reset file menu box
                IF LEFT$(_CWD$, 2) = "\\" THEN
                    NetPath2$ = _CWD$
                ELSE
                    NetPath2$ = ""
                    OldDrive$ = Drive.Search
                END IF
                OldDirectory$ = ""
                FileSpec1$ = "*.*"

                ' display default drive/netpath
                IF LEN(Outpt$) THEN
                    COLOR Yellow, Black
                    CALL AddStdout2(Outpt$)
                END IF
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "CHDIR"
PRINT "   Directory change utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   CHDIR [d:][path] [/cdvz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  don't display drive letter"
PRINT "   /v  don't display pathname"
PRINT "   /z  suppress error messages"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type CHDIR /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory change utility; "
END IF
RETURN
END SUB

' command line switch position function.
FUNCTION LastSwitch (Var)
IF Last.Switch = 0 THEN
    Last.Switch = Var - 1
    Switch.Exist = -1
ELSE
    IF Var < Last.Switch THEN
        Last.Switch = Var - 1
        Switch.Exist = -1
    END IF
END IF
END FUNCTION

' command line parser
FUNCTION ParseLine (X$)
Imbedded = INSTR(Command.Line, LCASE$(X$))
IF Imbedded THEN
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + LEN(X$))
    Last.Switch = Imbedded - 1
    Switch.Exist = -1
    ParseLine = True
ELSE
    Imbedded = INSTR(Command.Line, UCASE$(X$))
    IF Imbedded THEN
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + LEN(X$))
        Last.Switch = Imbedded - 1
        Switch.Exist = -1
        ParseLine = True
    ELSE
        ParseLine = False
    END IF
END IF
END FUNCTION

' displays carry flag error
SUB DisplayError (Temp$)
Var$ = Temp$
' check display errors flag
IF Display.Errors = False THEN
    ' display error
    IF LCASE$(StderrFile$) = "scrn:" THEN
        COLOR Red, Black
        PRINT #StderrHandle, Var$
    ELSE
        PRINT #StderrHandle, DATE$ + " " + TIME$ + " " + Var$
    END IF
END IF
END SUB

' displays windows error
SUB DisplayWinError (x)
' define error message value
v& = GetLastError
' check display errors flag
IF Display.Errors = False THEN
    ' call windows error message routine
    x& = FormatMessageA&(&H1200, "", v&, 0, ErrorBuffer$, 260, "")
    ' display error
    IF x& THEN
        Var$ = LEFT$(ErrorBuffer$, x& - 2)
    ELSE
        Var$ = "Unknown error 0x" + HEX$(v&) + "."
    END IF
    ' display error
    IF LCASE$(StderrFile$) = "scrn:" THEN
        COLOR Red, Black
        PRINT #StderrHandle, Var$
    ELSE
        PRINT #StderrHandle, DATE$ + " " + TIME$ + " " + Var$
    END IF
END IF
END SUB

' test volume media inserted.
FUNCTION MEDIAEXISTS (V)
' check drive exists.
IF DRIVEEXISTS(V) THEN
    MEDIAEXISTS = False
    EXIT FUNCTION
END IF

' get drive info.
VarX$ = CHR$(V + 64) + ":\" + CHR$(0)
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    MEDIAEXISTS = True
ELSE
    MEDIAEXISTS = False
END IF
END FUNCTION

' check drive exists.
'  returns -1 if drive not detected.
FUNCTION DRIVEEXISTS (V)
VarX$ = CHR$(V + 64) + ":\" + CHR$(0)
VarX = GetDriveType(VarX$)
DriveType = Nul
SELECT CASE VarX
    CASE 0
        DriveType = "[UNKNOWN]"
    CASE 1
        DriveType = "[BADROOT]"
    CASE 2
        DriveType = "[REMOVABLE]"
    CASE 3
        DriveType = "[FIXED]"
    CASE 4
        DriveType = "[REMOTE]"
    CASE 5
        DriveType = "[CDROM]"
    CASE 6
        DriveType = "[RAMDISK]"
END SELECT
IF VarX > 1 THEN
    DRIVEEXISTS = False
ELSE
    DRIVEEXISTS = True
END IF
END FUNCTION

FUNCTION IsAlphaNum (x$)
z$ = LCASE$(x$)
IF z$ >= "0" AND z$ <= "9" THEN
    IsAlphaNum = -1
END IF
IF z$ >= "a" AND z$ <= "z" THEN
    IsAlphaNum = -1
END IF
END FUNCTION

FUNCTION IsNumeric (x$)
FOR V = 1 TO LEN(x$)
    IF MID$(x$, V, 1) >= "0" AND MID$(x$, V, 1) <= "9" THEN
        Eat$ = Nul
    ELSE
        IsNumeric = 0
        EXIT FUNCTION
    END IF
NEXT
IsNumeric = -1
END FUNCTION

FUNCTION IsSingle (x$)
FOR V = 1 TO LEN(x$)
    IF MID$(x$, V, 1) >= "0" AND MID$(x$, V, 1) <= "9" THEN
        Eat$ = Nul
    ELSE
        IF MID$(x$, V, 1) = "." THEN
            IF Dot THEN
                IsSingle = 0
                EXIT FUNCTION
            ELSE
                Dot = -1
            END IF
        ELSE
            IsSingle = 0
            EXIT FUNCTION
        END IF
    END IF
NEXT
IsSingle = -1
END FUNCTION

REM check hex byte string.
REM   input: ByteString$ is space-separated hex byte pair string.
REM   returns: ValidByteString equals true if string is valid,
REM     ByteString$ is concatenated hex byte string.
SUB CheckHexBytes (ByteString$)
ByteString$ = LTRIM$(ByteString$)
ByteString$ = RTRIM$(ByteString$)
IF LEN(ByteString$) = False THEN
    ValidByteString = False
    EXIT SUB
END IF
ValidByteString = True
FOR StringPosition = 1 TO LEN(ByteString$)
    Byte$ = MID$(ByteString$, StringPosition, 1)
    IF (StringPosition MOD 3) = False THEN
        IF Byte$ <> " " THEN
            ValidByteString = False
            EXIT SUB
        END IF
    ELSE
        SELECT CASE Byte$
            CASE "0" TO "9", "A" TO "F", "a" TO "f"
                MID$(ByteString$, StringPosition, 1) = UCASE$(Byte$)
            CASE ELSE
                ValidByteString = False
                EXIT SUB
        END SELECT
    END IF
NEXT
Var1$ = ByteString$
DO
    Parse = INSTR(Var1$, " ")
    IF Parse THEN
        Var1$ = LEFT$(Var1$, Parse - 1) + MID$(Var1$, Parse + 1)
    ELSE
        EXIT DO
    END IF
LOOP
ByteString$ = Var1$
END SUB

REM check hex byte string.
REM   input:
REM     ByteString$ is space-separated hex byte pair string w/ wildcards.
REM   returns: ValidByteString equals true if string is valid,
REM     ByteString$ is concatenated hex byte string converted to ASCII.
SUB CheckHexBytes2 (ByteString$)
ByteString$ = LTRIM$(ByteString$)
ByteString$ = RTRIM$(ByteString$)
IF LEN(ByteString$) = False THEN
    ValidByteString = False
    EXIT SUB
END IF
ValidByteString = True
FOR StringPosition = 1 TO LEN(ByteString$)
    Byte$ = MID$(ByteString$, StringPosition, 1)
    IF (StringPosition MOD 3) = False THEN
        IF Byte$ <> " " THEN
            ValidByteString = False
            EXIT SUB
        END IF
    ELSE
        SELECT CASE Byte$
            CASE "0" TO "9", "A" TO "F", "a" TO "f", "?", "*"
                MID$(ByteString$, StringPosition, 1) = UCASE$(Byte$)
            CASE ELSE
                ValidByteString = False
                EXIT SUB
        END SELECT
    END IF
NEXT
' remove spaces.
Var1$ = ByteString$
DO
    Parse = INSTR(Var1$, " ")
    IF Parse THEN
        Var1$ = LEFT$(Var1$, Parse - 1) + MID$(Var1$, Parse + 1)
    ELSE
        EXIT DO
    END IF
LOOP
' check wildcards
FOR Parse = 1 TO LEN(Var1$) STEP 2
    IF MID$(Var1$, Parse, 1) = "?" OR MID$(Var1$, Parse + 1, 1) = "?" THEN
        IF MID$(Var1$, Parse, 2) = "??" THEN
            ' nul
        ELSE
            ValidByteString = False
            EXIT SUB
        END IF
    END IF
    IF MID$(Var1$, Parse, 1) = "*" OR MID$(Var1$, Parse + 1, 1) = "*" THEN
        IF MID$(Var1$, Parse, 2) = "**" THEN
            ' nul
        ELSE
            ValidByteString = False
            EXIT SUB
        END IF
    END IF
NEXT
' convert hex to ascii
VarB$ = ""
DO
    VarA$ = LEFT$(Var1$, 2)
    IF VarA$ = "??" THEN
        VarB$ = VarB$ + "?"
    ELSE
        IF VarA$ = "**" THEN
            VarB$ = VarB$ + "*"
        ELSE
            VarB$ = VarB$ + CHR$(VAL("&H" + VarA$))
        END IF
    END IF
    Var1$ = MID$(Var1$, 3)
    IF Var1$ = "" THEN
        EXIT DO
    END IF
LOOP
ByteString$ = VarB$
END SUB

' function to match case-sensitive substring with ?, * characters in substring
'  input: VarQ = -1 for case-sensitive
'    Var1$ is search string
'    Var2$ is string being searched
'  output: Var = -1 search string matched
SUB InstrSUB1 (Var, Var1$, Var2$, VarQ)

' store case-sensitive string match variables
IF VarQ THEN
    S2$ = Var1$
    S3$ = Var2$
ELSE
    S2$ = LCASE$(Var1$)
    S3$ = LCASE$(Var2$)
END IF

' check default instr
IF INSTR(S2$, "*") = 0 THEN
    IF INSTR(S2$, "?") = 0 THEN
        Var = INSTR(S3$, S2$)
        EXIT SUB
    END IF
END IF

Var = -1 ' assume match

' asterick always matches
IF Var1$ = "*" THEN
    EXIT SUB
END IF

' see if S2$ matches in S3$ with substrings
FOR S3 = 1 TO LEN(S3$)
    S1$ = MID$(S3$, S3)
    P1 = 1 ' pointer to S1$
    P2 = 1 ' pointer to S2$
    DO
        ' check match
        IF P2 > LEN(S2$) THEN
            EXIT SUB
        END IF

        ' check character in S2$ at P2
        V$ = MID$(S2$, P2, 1)
        SELECT CASE V$
            CASE "*" ' global character
                ' scan to next char
                IF P2 > LEN(S2$) THEN
                    EXIT DO
                END IF
                S4$ = MID$(S2$, P2 + 1, 1)
                SELECT CASE S4$
                    CASE "*", "?"
                        P2 = P2 + 1
                    CASE ELSE
                        DO
                            IF MID$(S1$, P1, 1) = S4$ THEN
                                EXIT DO
                            END IF
                            IF P1 >= LEN(S1$) THEN
                                EXIT DO
                            END IF
                            P1 = P1 + 1
                        LOOP
                        P2 = P2 + 1
                END SELECT
            CASE "?" ' wildcard character
                P1 = P1 + 1
                P2 = P2 + 1
            CASE ELSE ' ascii character
                IF MID$(S1$, P1, 1) <> V$ THEN ' no match
                    EXIT DO
                END IF
                P1 = P1 + 1
                P2 = P2 + 1
        END SELECT
    LOOP
NEXT
Var = 0 ' no match
END SUB

SUB Capitalize (Var3$)
Var3$ = RTRIM$(LTRIM$(Var3$))
IF Var3$ = Nul THEN
    EXIT SUB
END IF
MID$(Var3$, 1, 1) = UCASE$(MID$(Var3$, 1, 1))
Var = INSTR(Var3$, " ")
IF Var < LEN(Var3$) THEN
    DO WHILE Var
        MID$(Var3$, Var + 1, 1) = UCASE$(MID$(Var3$, Var + 1, 1))
        Var = INSTR(Var + 1, Var3$, " ")
    LOOP
END IF
END SUB

SUB MakDir (ErrFlag)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' get current drive
Current.Drive = UCASE$(LEFT$(_CWD$, 1))

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get switches from command line
Continuous.Display = ParseLine("/C")
Display.Drive = ParseLine("/D")
Display.Path = ParseLine("/V")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = UCASE$(LEFT$(Command.Work, 1))
        ELSE
            Drive.Search = Current.Drive
        END IF

        Command.Work = LTRIM$(Command.Work)
        Command.Work = RTRIM$(Command.Work)

        ' check drive
        Flag = -1
        IF LEFT$(_CWD$, 2) <> "\\" THEN
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) = 0 THEN
                ' display any errors
                CALL DisplayError("Error accessing drive " + CHR$(V + 64) + ":")
                Flag = 0
            END IF
        END IF
        ' check directory exists
        ASCIIZ = Command.Work + CHR$(0)
        hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, FILE_FLAG_BACKUP_SEMANTICS, 0)
        IF hfind <> INVALID_HANDLE_VALUE THEN
            x = CloseHandle(hfind)
            ' display any errors
            CALL DisplayError("Directory exists.")
            Flag = 0
        END IF
        IF Command.Work = "" THEN
            Flag = False
        END IF
        IF Flag THEN
            ' make directory name
            f$ = Command.Work
            x = CreateDirectoryStructure(f$)
            IF x = 0 THEN
                IF NetPathError THEN
                    CALL DisplayError("Netpath error making directory.")
                ELSE
                    CALL DisplayError("Error 0x" + HEX$(GetLastError) + " making directory.")
                    CALL DisplayWinError(x)
                END IF
            ELSE
                ' store drive letter
                Outpt$ = Nul
                IF Display.Drive = False THEN
                    IF LEFT$(_CWD$, 2) <> "\\" THEN
                        Outpt$ = Drive.Search + ":"
                    END IF
                END IF

                ' store pathname
                IF Display.Path = False THEN
                    IF LEFT$(DirFilename$(0), 2) = "\\" THEN
                        Outpt$ = DirFilename$(0)
                    ELSE
                        IF MID$(DirFilename$(0), 2, 1) = ":" THEN
                            Outpt$ = Outpt$ + MID$(DirFilename$(0), 3)
                        ELSE
                            Outpt$ = DirFilename$(0)
                        END IF
                    END IF
                END IF

                ' display drive/netpath
                IF LEN(Outpt$) THEN
                    COLOR Yellow, Black
                    CALL AddStdout2(Outpt$)
                END IF
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "MKDIR"
PRINT "   Directory create utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   MKDIR [d:][path] [/cdvz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  don't display drive letter"
PRINT "   /v  don't display pathname"
PRINT "   /z  suppress error messages"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type MKDIR /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory create utility; "
END IF
RETURN
END SUB

' gets default directory
FUNCTION DirFilename$ (Var)

' fully qualified path
IF MID$(Command.Work, 2, 1) = ":" THEN
    DirFilename$ = Command.Work
    EXIT FUNCTION
END IF

' fully qualified netpath
IF LEFT$(Command.Work, 2) = "\\" THEN
    DirFilename$ = Command.Work
    EXIT FUNCTION
END IF

' fully qualified root path
IF LEFT$(Command.Work, 1) = "\" THEN
    Var$ = Drive.Search + ":"
    DirFilename$ = Var$ + Command.Work
    EXIT FUNCTION
END IF

' get current directory
Drive$ = Drive.Search + ":"
IF LEFT$(_CWD$, 2) = "\\" THEN
    Directory.Search$ = _CWD$
ELSE ' path\filename
    IF Drive.Search = Current.Drive THEN
        Directory.Search$ = _CWD$
    ELSE
        Original$ = _CWD$
        x = SetCurrentDirectoryA(Drive$ + CHR$(0))
        Directory.Search$ = _CWD$
        x = SetCurrentDirectoryA(Original$ + CHR$(0))
    END IF
END IF

' store default directory
Filename.Search$ = Command.Work
IF RIGHT$(Directory.Search$, 1) = "\" THEN
    Directory.Search$ = Directory.Search$ + Filename.Search$
ELSE
    Directory.Search$ = Directory.Search$ + "\" + Filename.Search$
END IF
DirFilename$ = Directory.Search$
END FUNCTION

' routine to create directory structure
FUNCTION CreateDirectoryStructure (NewDirectory$)
' construct path
NewDirectory$ = RTRIM$(NewDirectory$)
Next.Dest.Net$ = ""
NetPathError = 0
IF LEFT$(NewDirectory$, 2) = "\\" THEN ' \\serve\share\
    V = INSTR(3, NewDirectory$, "\")
    IF V THEN
        V = INSTR(V + 1, NewDirectory$, "\")
        IF V THEN
            Next.Dest.Net$ = LEFT$(NewDirectory$, V - 1)
            NewDirectory$ = MID$(NewDirectory$, V + 1)
        END IF
    END IF
    IF Next.Dest.Net$ = "" THEN
        NetPathError = -1
        CreateDirectory = 0
        EXIT FUNCTION
        NewDirectory$ = "\"
        Drive.Create$ = Current.Drive
    END IF
ELSE
    IF MID$(NewDirectory$, 2, 1) = ":" THEN ' c:\tempx
        Drive.Create$ = LEFT$(NewDirectory$, 1)
        NewDirectory$ = MID$(NewDirectory$, 3)
    ELSE
        IF LEFT$(NewDirectory$, 1) = "\" THEN ' \tempx
            Drive.Create$ = LEFT$(_CWD$, 1)
        ELSE
            IF LEFT$(_CWD$, 2) = "\\" THEN ' \\tempx
                Next.Dest.Net$ = _CWD$
            ELSE
                Drive.Create$ = LEFT$(_CWD$, 1) ' c:tempx
            END IF
        END IF
    END IF
END IF
IF RIGHT$(NewDirectory$, 1) <> "\" THEN
    NewDirectory$ = NewDirectory$ + "\"
END IF

' create path
x = 0
Next.Dir = INSTR(NewDirectory$, "\")
DO
    IF Next.Dir = False THEN
        EXIT DO
    END IF

    ' parse path left to right
    SubDir$ = LEFT$(NewDirectory$, Next.Dir - 1) ' \tempx\t1\t2\t3\
    Next.Dir = INSTR(Next.Dir + 1, NewDirectory$, "\")

    ' make directory name
    IF LEN(SubDir$) THEN
        ' create directory
        IF LEN(Next.Dest.Net$) THEN
            F$ = Next.Dest.Net$ + "\" + SubDir$
        ELSE
            F$ = Drive.Create$ + ":" + SubDir$
        END IF

        ' check directory
        IF _DIREXISTS(F$) = 0 THEN
            F$ = F$ + CHR$(0)
            x = CreateDirectoryA(F$, 0)

            ' check error
            IF x = 0 THEN
                IF GetLastError& = &HB7 THEN ' ignore already exists
                    ' nul
                ELSE
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
CreateDirectoryStructure = x
END FUNCTION

SUB DelDir (ErrFlag)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' get current drive
Current.Drive = LEFT$(_CWD$, 1)

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset counters
Dirs.Counted = False

' store command line
Command.Line = RTRIM$(Command.Line)

' get switches from command line
Continuous.Display = ParseLine("/C")
Display.Drive = ParseLine("/D")
Prompt.Delete = ParseLine("/P")
Display.Path = ParseLine("/V")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = UCASE$(LEFT$(Command.Work, 1))
        ELSE
            Drive.Search = Current.Drive
        END IF

        ' check drive
        Flag = -1
        IF LEFT$(_CWD$, 2) <> "\\" THEN
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) = 0 THEN
                ' display any errors
                CALL DisplayError("Error accessing drive " + CHR$(V + 64) + ":")
                Flag = 0
            END IF
        END IF
        IF Flag THEN
            ' check directory exists
            ASCIIZ = Command.Work + CHR$(0)
            hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, FILE_FLAG_BACKUP_SEMANTICS, 0)
            IF hfind <> INVALID_HANDLE_VALUE THEN
                x = CloseHandle(hfind)

                ' reset delete flag
                Delete.File = True

                ' check delete flag
                IF Prompt.Delete = 0 THEN
                    Prompt2$ = "Delete directory " + DirFilename$(0) + "(y/n/q/c)?"
                    CALL MorePrompt(Prompt2$, "ynqc", Outpt$)
                    SELECT CASE Outpt$
                        CASE "c"
                            Prompt.Delete = True
                        CASE "n"
                            Delete.File = False
                        CASE "q"
                            Delete.File = False
                            Quit.Searching = True
                    END SELECT
                END IF

                ' check delete flag
                IF Delete.File THEN

                    ' delete directory.
                    G$ = Command.Work
                    IF RIGHT$(G$, 1) <> "\" THEN G$ = G$ + "\"
                    CALL Silentdel_Directories(G$)

                    ' check error flag
                    IF _DIREXISTS(G$) THEN
                        CALL DisplayError("Error deleting directory.")
                    ELSE
                        ' display drive letter
                        Outpt$ = Nul
                        IF Display.Drive = False THEN
                            IF LEFT$(_CWD$, 2) <> "\\" THEN
                                Outpt$ = Drive.Search + ":"
                            END IF
                        END IF

                        ' store pathname
                        IF Display.Path = False THEN
                            IF LEFT$(DirFilename$(0), 2) = "\\" THEN
                                Outpt$ = DirFilename$(0)
                            ELSE
                                IF MID$(DirFilename$(0), 2, 1) = ":" THEN
                                    Outpt$ = Outpt$ + MID$(DirFilename$(0), 3)
                                ELSE
                                    Outpt$ = DirFilename$(0)
                                END IF
                            END IF
                        END IF

                        ' display drive/netpath
                        IF LEN(Outpt$) THEN
                            COLOR Yellow, Black
                            CALL AddStdout2(Outpt$)
                        END IF

                        ' count directories deleted
                        Dirs.Counted = Dirs.Counted + 1
                    END IF
                END IF
            ELSE
                CALL DisplayWinError(x)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Dirs.Counted)
    PRINT "Directories counted: "; Total$
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "RMDIR"
PRINT "   Directory delete utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   RMDIR [d:][path] [/cdpvz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  don't display drive letter"
PRINT "   /p  don't prompt before delete"
PRINT "   /v  don't display pathname"
PRINT "   /z  suppress error messages"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type RMDIR /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory delete utility; "
END IF
RETURN
END SUB

SUB MakeFile (ErrFlag)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' get current drive
Current.Drive = LEFT$(_CWD$, 1)

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get switches from command line
Continuous.Display = ParseLine("/C")
Display.Drive = ParseLine("/D")
Display.Path = ParseLine("/V")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = UCASE$(LEFT$(Command.Work, 1))
        ELSE
            Drive.Search = Current.Drive
        END IF

        ' check drive
        Flag = -1
        IF LEFT$(_CWD$, 2) <> "\\" THEN
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) = 0 THEN
                ' display any errors
                CALL DisplayError("Error accessing drive " + CHR$(V + 64) + ":")
                Flag = 0
            END IF
        END IF
        IF Flag THEN
            ' make filename
            ASCIIZ = Command.Work + CHR$(0)
            hfind = CreateFile(ASCIIZ, &H180, 0, 0, 1, 0, 0)
            IF hfind <> INVALID_HANDLE_VALUE THEN
                x = CloseHandle(hfind)

                ' store drive letter
                Outpt$ = Nul
                IF Display.Drive = False THEN
                    IF LEFT$(_CWD$, 2) <> "\\" THEN
                        Outpt$ = Drive.Search + ":"
                    END IF
                END IF

                ' store pathname
                IF Display.Path = False THEN
                    IF LEFT$(DirFilename$(0), 2) = "\\" THEN
                        Outpt$ = DirFilename$(0)
                    ELSE
                        IF MID$(DirFilename$(0), 2, 1) = ":" THEN
                            Outpt$ = Outpt$ + MID$(DirFilename$(0), 3)
                        ELSE
                            Outpt$ = DirFilename$(0)
                        END IF
                    END IF
                END IF

                ' display drive/netpath
                IF LEN(Outpt$) THEN
                    COLOR Yellow, Black
                    CALL AddStdout2(Outpt$)
                END IF
            ELSE
                CALL DisplayError("Error 0x" + HEX$(GetLastError) + " creating file.")
                CALL DisplayWinError(x)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "MKFILE"
PRINT "   File create utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   MKFILE [d:][path] [/cdvz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  don't display drive letter"
PRINT "   /v  don't display filename"
PRINT "   /z  suppress error messages"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type MKFILE /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File create utility; "
END IF
RETURN
END SUB

SUB Whereis (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' get exclude switch
Exclude.List = Nul
Imbedded = INSTR(Command.Line, "/(")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Exclude.List = Nul
    Next.Bracket = INSTR(Imbedded + 2, Command.Line, ")")
    IF Next.Bracket THEN
        Exclude.List = MID$(Command.Line, Imbedded + 2, Next.Bracket - Imbedded - 2)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Next.Bracket + 1)
    END IF
    IF Exclude.List = Nul THEN
        GOTO Boot.Error
    END IF
END IF
Command.Line = RTRIM$(Command.Line)

' get dos command
DOS.Command = Nul
Imbedded = INSTR(Command.Line, "/[")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    DOS.Command = Nul
    Next.Bracket = INSTR(Imbedded + 2, Command.Line, "]")
    IF Next.Bracket THEN
        DOS.Command = MID$(Command.Line, Imbedded + 2, Next.Bracket - Imbedded - 2)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Next.Bracket + 1)
    END IF
    IF DOS.Command = Nul THEN
        GOTO Boot.Error
    END IF
END IF
Command.Line = RTRIM$(Command.Line)

' edit special DOS command parameters
Imbedded = INSTR(DOS.Command, "//1")
WHILE Imbedded
    DOS.Command = LEFT$(DOS.Command, Imbedded - 1) + ">" + MID$(DOS.Command, Imbedded + 3)
    Imbedded = INSTR(DOS.Command, "//1")
WEND
Imbedded = INSTR(DOS.Command, "//2")
WHILE Imbedded
    DOS.Command = LEFT$(DOS.Command, Imbedded - 1) + "<" + MID$(DOS.Command, Imbedded + 3)
    Imbedded = INSTR(DOS.Command, "//2")
WEND
Imbedded = INSTR(DOS.Command, "//3")
WHILE Imbedded
    DOS.Command = LEFT$(DOS.Command, Imbedded - 1) + "|" + MID$(DOS.Command, Imbedded + 3)
    Imbedded = INSTR(DOS.Command, "//3")
WEND
Imbedded = INSTR(DOS.Command, "//4")
WHILE Imbedded
    DOS.Command = LEFT$(DOS.Command, Imbedded - 1) + "%" + MID$(DOS.Command, Imbedded + 3)
    Imbedded = INSTR(DOS.Command, "//4")
WEND
Imbedded = INSTR(DOS.Command, "//5")
WHILE Imbedded
    DOS.Command = LEFT$(DOS.Command, Imbedded - 1) + "]" + MID$(DOS.Command, Imbedded + 3)
    Imbedded = INSTR(DOS.Command, "//5")
WEND

' check command line switches
No.Display.Archive = ParseLine("//A")
No.Display.Hidden = ParseLine("//H")
No.Display.Directory = ParseLine("//I")
No.Display.Readonly = ParseLine("//O")
No.Display.System = ParseLine("//S")
No.Display.Any = ParseLine("//X")
No.Display.Compress = ParseLine("//M1")
No.Display.Encrypt = ParseLine("//M2")

Display.Archive = ParseLine("/A")
Display.Hidden = ParseLine("/H")
Display.Directory = ParseLine("/I")
Display.Readonly = ParseLine("/O")
Display.System = ParseLine("/S")
Display.Any = ParseLine("/X")
Display.Compress = ParseLine("/M1")
Display.Encrypt = ParseLine("/M2")

No.Display.Drive = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Short.Display = ParseLine("/E")
Display.Search = ParseLine("/G")
Skip.Current = ParseLine("/J")
Short.Filenames = ParseLine("/K")
Double.Line = ParseLine("/L")
Disable.Cmd = ParseLine("/P")
Display.Dirs.Counted = ParseLine("/Q")
Recurse.Directories = ParseLine("/R")
Remove.Slash = ParseLine("/U")
Display.Volume = ParseLine("/V")
Wide.Display = ParseLine("/W")
Display.Lowercase = ParseLine("/Y")
Display.Only.Errors = ParseLine("/Z1")
Display.Errors = ParseLine("/Z")
Enable.Redirect = ParseLine("/@")
Ignore.Prompts = ParseLine("/#")

' check wide display
IF Wide.Display THEN
    Double.Line = False
    Short.Display = True
    Short.Filenames = True
    Wide.List = 0
END IF

' set attribute display variable
Display.Attribute = False
IF Display.Archive OR No.Display.Archive THEN
    Display.Attribute = True
END IF
IF Display.Hidden OR No.Display.Hidden THEN
    Display.Attribute = True
END IF
IF Display.Readonly OR No.Display.Readonly THEN
    Display.Attribute = True
END IF
IF Display.System OR No.Display.System THEN
    Display.Attribute = True
END IF
IF Display.Any OR No.Display.Any THEN
    Display.Attribute = True
END IF
IF Display.Directory OR No.Display.Directory THEN
    Display.Attribute = True
END IF
IF Display.Compress OR No.Display.Compress THEN
    Display.Attribute = True
END IF
IF Display.Encrypt OR No.Display.Encrypt THEN
    Display.Attribute = True
END IF

' reset file counter variable
Dirs.Counted = FalseD
Dirs.Displayed = FalseD
Files.Counted = FalseD
More.Display = False
   
' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32 + T3!
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32 + T3!
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get file size from command line
Search.File.Size = False
Search.Size.From = FalseD
Search.Size.To = FalseD
Imbedded = INSTR(UCASE$(Command.Line), "/F")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.File.Size = True
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Search.Size.From = Var#
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    GOSUB Get.Numeric
    Search.Size.To = Var#
    Search.Size.From = Search.Size.From * 1024#
    Search.Size.To = Search.Size.To * 1024#
END IF

' get extended date\time switches
Creation.Time = ParseLine("/1")
Access.Time = ParseLine("/2")
Modified.Time = ParseLine("/3")
IF Creation.Time = False THEN
    IF Access.Time = False THEN
        IF Modified.Time = False THEN
            Modified.Time = True
        END IF
    END IF
END IF

' get nested switch from command line
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' reset work variables
Files.Counter = False
Nested.Levels = False
Quit.Searching = False
Total.Bytes = False
Files.Counted = False
Volumes.Counted = False

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' locate filelist character
        Flag = False
        Command.Out$ = Command.Work
        Command.Out2$ = Nul
        IF MID$(Command.Out$, 2, 1) = ":" THEN
            Command.Out2$ = LEFT$(Command.Out$, 2)
            Command.Out$ = MID$(Command.Out$, 3)
        END IF
        IF LEFT$(Command.Out$, 1) = "@" THEN
            Flag = True
            Command.Out$ = MID$(Command.Out$, 2)
        ELSE
            FOR Var = LEN(Command.Out$) TO 1 STEP -1
                IF MID$(Command.Out$, Var, 1) = "\" THEN
                    Char$ = MID$(Command.Out$, Var + 1)
                    IF LEFT$(Char$, 1) = "@" THEN
                        Flag = True
                        Command.Out$ = LEFT$(Command.Out$, Var) + MID$(Char$, 2)
                    END IF
                    EXIT FOR
                END IF
            NEXT
        END IF
        IF LEN(Command.Out2$) THEN
            Command.Out$ = Command.Out2$ + Command.Out$
        END IF

        ' process work filename
        IF Flag AND Enable.Redirect THEN

            ' concatenate work filename
            Var$ = Command.Work ' command line filename
            FOR Var = LEN(Var$) TO 1 STEP -1
                IF MID$(Var$, Var, 1) = "\" THEN
                    Var$ = MID$(Var$, Var + 1)
                    EXIT FOR
                END IF
            NEXT
            IF MID$(Var$, 2, 1) = ":" THEN
                Var$ = MID$(Var$, 3)
            END IF
            Filename2$ = Var$
            Var$ = MID$(Var$, 2)
            Var2$ = Command.Out$
            FOR Var = LEN(Var2$) TO 1 STEP -1
                IF MID$(Var2$, Var, 1) = "\" THEN
                    Var2$ = LEFT$(Var2$, Var)
                    Var$ = Var2$ + Var$
                    EXIT FOR
                END IF
            NEXT

            ' check excluded file
            IF LEN(Exclude.List) THEN
                Filename1$ = UCASE$(RTRIM$(Exclude.List))
                Filename2$ = UCASE$(RTRIM$(Filename2$))
                CALL CheckExcluded(Filename1$, Filename2$)
                IF Exclude.File THEN
                    Max.DirRecurse = 0
                    CALL Whereis_ProcessCommand
                    GOTO Next.Filename
                END IF
            END IF

            ' prompt for filelist
            IF Ignore.Prompts = False THEN

                ' prompt to use filelist
                Prompt2$ = "Use '" + LCASE$(Var$) + "' as filelist(y/n/q)?"
                CALL MorePrompt(Prompt2$, "ynq", Outpt$)
                SELECT CASE Outpt$
                    CASE "q"
                        Quit.Searching = True
                        GOTO Next.Filename
                    CASE "n"
                        Max.DirRecurse = 0
                        CALL Whereis_ProcessCommand
                        GOTO Next.Filename
                END SELECT
            END IF

            ' open work filename
            DataError = False
            Error.Flag = True
            V = FREEFILE
            OPEN Command.Out$ FOR INPUT SHARED AS #V
            Error.Flag = False

            ' check error flag
            IF DataError THEN
                CALL DisplayError("Error opening filelist.")
            ELSE
                ' process input filenames
                Max.DirRecurse = 0
                DO UNTIL EOF(V)
                    ' get next filename
                    LINE INPUT #V, Command.Work

                    ' process search filename
                    Command.Work = RTRIM$(Command.Work)
                    IF LEN(Command.Work) THEN
                        CALL Whereis_ProcessCommand
                        ' check recurse error
                        IF Max.DirRecurse >= RecurseLevel THEN
                            EXIT DO
                        END IF
                    END IF
                LOOP
                CLOSE #V
            END IF
        ELSE
            ' process remaining filenames
            Max.DirRecurse = 0
            CALL Whereis_ProcessCommand
        END IF

        ' continue label
        Next.Filename:

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Wide.List THEN
    Wide.List = False
    PRINT
END IF
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    IF Display.Volume THEN
        Total$ = FormatString$(CDBL(Volumes.Counted))
        PRINT "Volumes counted: "; Total$
    ELSE
        IF Display.Dirs.Counted = False THEN
            Total$ = FormatString$(Dirs.Counted)
            PRINT "Directories counted: "; Total$
        ELSE
            Total$ = FormatString$(Dirs.Displayed)
            PRINT "Directories displayed: "; Total$
        END IF
        Total$ = FormatString$(Files.Counted)
        PRINT "Files counted: "; Total$
        IF Double.Line THEN
            Total$ = FormatString$(Total.Bytes)
            PRINT "Bytes counted: "; Total$
        END IF
    END IF
END IF

' reset color
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File search utility;"
END IF
RETURN

' display program usage
Boot.Usage:
Display.Page = 1
DO
    SELECT CASE Display.Page
        CASE 1
            GOSUB Display.Page.One
        CASE 2
            GOSUB Display.Page.Two
    END SELECT
    Prompt2$ = "Press 1, 2, or q to quit:"
    CALL MorePrompt(Prompt2$, "12q", Outpt$)
    SELECT CASE Outpt$
        CASE "1"
            Display.Page = 1
        CASE "2"
            Display.Page = 2
        CASE "q"
            COLOR Plain, Black
            RETURN
    END SELECT
LOOP
COLOR Plain, Black
PRINT "Exiting to DOS:"
EXIT SUB

Display.Page.One:
GOSUB Display.Header
PRINT "  /b  suppress drive letter      /c  continuous display"
PRINT "  /e  short filename display     /g  display search directories"
PRINT "  /j  skip current directory     /k  use 8.3 short filenames"
PRINT "  /l  double line display        /nxxx  nested directories"
PRINT "  /q  directory count override   /r  recurse directories"
PRINT "  /u  remove trailing slash      /v  display volume label"
PRINT "  /w  wide file list             /y  display lowercase"
PRINT "  /z  no error messages          /z1 display only errors"
PRINT "display ranges:"
PRINT "  /1  creation, /2  last access, /3  modify time"
PRINT "  /d  is range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "  /t  is range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "  /f  is range of file sizes in form xxx-xxx in kilobytes"
RETURN

Display.Page.Two:
GOSUB Display.Header
PRINT "  filelist switches:"
PRINT "    /@  enable filelists   /#  ignore filelist prompts"
PRINT "  exclude file list switch:"
PRINT "    /(<filename.ext>)"
PRINT "      excludes files containing ? and * characters."
PRINT "  display file attributes: / prefix for files with, // prefix for files without"
PRINT "    a  archive, h  hidden, i  directory, o  read-only, s  system"
PRINT "    m1  compressed,  m2  encrypted, x  none"
PRINT "  DOS command: /[<command>]"
PRINT "    with command replacement parameters: (use /p for /c override);"
PRINT "      %1 = d:, %2 = d:\, %3 = d:\pathname, %4 = d:\pathname\"
PRINT "      %5 = d:\pathname\filename.ext, %6 = \pathname, %7 = \pathname\"
PRINT "      %8 = \pathname\filename.ext, %9 = filename.ext, %a = filename"
PRINT "      %b = .ext, %c = ext, //1 = >, //2 = <, //3 = |, //4 = %, //5 = ]"
RETURN

Display.Header:
COLOR White, Black
PRINT "DIR"
PRINT "  File search utility;"
COLOR Yellow, Black
PRINT "Usage:"
PRINT "  DIR [@][d:\path\filename.ext][//ahimosx][/bcdefgjklnpqrtuvwyz@#][/123]"
PRINT "Where:"
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type DIR /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

REM constructs and checks filelist filename exists.
REM   returns ambiguated filename.
FUNCTION FILELISTEXISTS$ (Var$)
' construct filelist filename.
VarX$ = Var$
V = INSTR(VarX$, "@")
IF V = 0 THEN
    FILELISTEXISTS$ = Nul
    EXIT FUNCTION
END IF
FOR V = LEN(VarX$) TO 1 STEP -1
    IF MID$(VarX$, V, 1) = "@" THEN
        EXIT FOR
    END IF
NEXT
IF V = 1 THEN ' @filelist.txt
    VarX$ = MID$(VarX$, V + 1)
ELSE
    IF V > 1 THEN
        IF MID$(VarX$, V - 1, 1) = "\" THEN ' \dos\@filelist.txt
            VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 1)
        ELSE
            IF MID$(VarX$, V - 1, 1) = ":" THEN ' c:@filelist.txt
                VarX$ = LEFT$(VarX$, V - 1) + MID$(VarX$, V + 1)
            END IF
        END IF
    END IF
END IF
' check filelist filename exists.
IF INSTR(V, VarX$, "\") THEN ' c:\@path\filename.ext
    FILELISTEXISTS$ = Nul
ELSE
    IF _FILEEXISTS(VarX$) = 0 THEN
        FILELISTEXISTS$ = Nul
    ELSE
        FILELISTEXISTS$ = VarX$
    END IF
END IF
END FUNCTION

SUB Whereis_ProcessCommand
' check recurse error
IF Max.DirRecurse >= RecurseLevel THEN
    EXIT SUB
END IF

' store current drive
IF MID$(Command.Work, 2, 1) = ":" THEN
    Drive.Search = LEFT$(Command.Work, 1)
    Command.Work = MID$(Command.Work, 3)
ELSE
    Drive.Search = Current.Drive
END IF
Drive.Search = UCASE$(Drive.Search)

' store current directory
Directory.Search$ = ""
Imbedded1 = INSTR(Command.Work, "\")
Imbedded2 = Imbedded1
WHILE Imbedded1
    Imbedded2 = Imbedded1
    Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
WEND
IF Imbedded2 THEN
    Directory.Search$ = LEFT$(Command.Work, Imbedded2)
    Command.Work = MID$(Command.Work, Imbedded2 + 1)
END IF
IF Directory.Search$ = "" THEN
    IF LEFT$(Current.Directory, 2) = "\\" THEN
        Directory.Search$ = Current.Directory
    ELSE
        IF Drive.Search = Current.Drive THEN
            Directory.Search$ = MID$(Current.Directory, 3)
        ELSE
            Directory.Search$ = "\"
        END IF
    END IF
END IF
IF RIGHT$(Directory.Search$, 1) <> "\" THEN
    Directory.Search$ = Directory.Search$ + "\"
END IF
   
' get filename spec
Filename.Search$ = RTRIM$(Command.Work)
V = INSTR(Filename.Search$, CHR$(0))
IF V THEN
    Filename.Search$ = LEFT$(Filename.Search$, V - 1)
END IF
IF Filename.Search$ = Nul THEN
    Filename.Search$ = "*.*"
END IF
Command.Work = Nul
   
' display search filename
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        PRINT "Searching: " + Directory.Search$ + Filename.Search$
    ELSE
        IF Display.Volume THEN
            PRINT "Searching: " + Drive.Search + ":"
        ELSE
            IF LEFT$(Directory.Search$, 1) = "@" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
        END IF
    END IF
    Files.Counter = Files.Counter + 1
END IF

' call routine to search for files
IF Display.Volume THEN
    CALL Whereis_DisplayVolume(Drive.Search)
ELSE
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        CALL Whereis_Directories(Directory.Search$, Filename.Search$)
    ELSE
        V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
        IF MEDIAEXISTS(V) THEN
            CALL Whereis_Directories(Directory.Search$, Filename.Search$)
        END IF
    END IF
END IF
END SUB

' subroutine to access directories
SUB Whereis_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM Recurse.Temp AS INTEGER

' check to quit
IF Quit.Searching THEN
    EXIT SUB
END IF

' check recurse error
IF Max.DirRecurse >= RecurseLevel THEN
    EXIT SUB
END IF

' display directories searched
IF Display.Search THEN
    ' check skip directory
    Flag = True
    IF Skip.Current THEN
        IF First.Directory = False THEN
            First.Directory = True
            Flag = False
        END IF
    END IF
    IF Flag THEN
        ' count only directories displayed
        Dirs.Displayed = Dirs.Displayed + 1#
        ' store directory
        Filename$ = Directory.Search$
        IF Filename$ <> "" THEN
            IF Remove.Slash THEN
                Filename$ = LEFT$(Filename$, LEN(Filename$) - 1)
            END IF
        END IF
        ' store drive letter
        IF No.Display.Drive = False THEN
            IF LEFT$(Filename$, 2) <> "\\" THEN
                Filename$ = Drive.Search + ":" + Filename$
            END IF
        END IF
        ' store directory
        Outpt$ = RTRIM$(Filename$)
        Length = LEN(Outpt$)
        IF Length THEN
            ' display directory
            COLOR Yellow, Black
            PRINT Outpt$
            ' check paginate
            Files.Counter = Files.Counter + INT(Length / 80) + 1
            CALL PagePrompt(Files.Counter, FlagX)
            IF FlagX THEN
                Files.Counter = 0
            END IF
            IF Quit.Searching THEN
                EXIT SUB
            END IF
        END IF
    END IF
END IF

' store directory asciiz
IF LEFT$(Directory.Search$, 2) <> "\\" THEN
    IF LEFT$(Directory.Search$, 1) = "@" THEN
        ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)
    ELSE
        ASCIIZ = Drive.Search + ":" + Directory.Search$ + "*.*" + CHR$(0)
    END IF
ELSE
    ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)
END IF

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search directory names
    CALL Whereis_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Searching THEN
                EXIT DO
            END IF

            ' check max recurse levevl
            IF Max.DirRecurse >= RecurseLevel THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Whereis_Directories(Next.Directory$, Filename.Search$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Whereis_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM Recurse.Temp AS INTEGER

' check to quit
IF Quit.Searching THEN
    EXIT SUB
END IF

' check recurse error
IF Max.DirRecurse >= RecurseLevel THEN
    EXIT SUB
END IF

' make filename
Var1$ = Directory.Search$
IF LEFT$(Var1$, 2) <> "\\" THEN
    IF LEFT$(Var1$, 1) = "\" THEN
        Var1$ = Drive.Search + ":" + Var1$
    END IF
END IF
Var2$ = Filename.Search$
IF LEFT$(Var2$, 1) = "@" AND Enable.Redirect THEN
    Var2$ = MID$(Var2$, 2)
END IF

Var1$ = RTRIM$(Var1$)
Var2$ = RTRIM$(Var2$)

ASCIIZ = Var1$ + Var2$ + CHR$(0)
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search filenames
    DO
        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF

        ' store file data
        finddatatemp = finddata

        ' check directory attribute
        Attribute = finddata.dwFileAttributes

        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check unicode
        IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
            Filename$ = finddata.cAlternateFileName
            V = INSTR(Filename$, CHR$(0))
            IF V THEN Filename$ = LEFT$(Filename$, V - 1)
        END IF

        ' store short filename
        Short.Directory$ = finddata.cAlternateFileName
        V = INSTR(Short.Directory$, CHR$(0))
        IF V THEN Short.Directory$ = LEFT$(Short.Directory$, V - 1)
        IF Short.Directory$ = Nul THEN
            Short.Directory$ = finddata.cFileName
            V = INSTR(Short.Directory$, CHR$(0))
            IF V THEN Short.Directory$ = LEFT$(Short.Directory$, V - 1)
        END IF

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN
                ' increment directories actually searched
                Dirs.Counted = Dirs.Counted + 1#
            END IF

            ' store file size
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' recursively call filelists
            IF LEFT$(Filename.Search$, 1) = "@" AND Enable.Redirect THEN
                Filelist.Filename$ = Directory.Search$ + Filename$

                ' check recursion level
                Max.DirRecurse = Max.DirRecurse + 1
                IF Max.DirRecurse >= RecurseLevel THEN
                    CALL DisplayError("Recursive filelist error.")
                    Flag = False
                    EXIT DO
                END IF

                ' reset prompt flag
                Flag = True

                ' concatenate work filename
                Var$ = Filename$ ' command line filename
                FOR Var = LEN(Var$) TO 1 STEP -1
                    IF MID$(Var$, Var, 1) = "\" THEN
                        Var$ = MID$(Var$, Var + 1)
                        EXIT FOR
                    END IF
                NEXT
                IF MID$(Var$, 2, 1) = ":" THEN
                    Var$ = MID$(Var$, 3)
                END IF
                Filename2$ = Var$
                Var2$ = Filelist.Filename$
                FOR Var = LEN(Var2$) TO 1 STEP -1
                    IF MID$(Var2$, Var, 1) = "\" THEN
                        Var2$ = LEFT$(Var2$, Var)
                        Var$ = Var2$ + Var$
                        EXIT FOR
                    END IF
                NEXT

                ' check excluded file
                IF LEN(Exclude.List) THEN
                    Filename1$ = UCASE$(RTRIM$(Exclude.List))
                    Filename2$ = UCASE$(RTRIM$(Filename2$))
                    CALL CheckExcluded(Filename1$, Filename2$)
                    IF Exclude.File THEN
                        Flag = False
                    END IF
                END IF

                ' prompt for filelist
                IF Ignore.Prompts = False THEN

                    ' prompt to use filelist
                    IF Flag THEN
                        Prompt2$ = "Use '" + LCASE$(Var$) + "' as filelist(y/n/q)?"
                        CALL MorePrompt(Prompt2$, "ynq", Outpt$)
                        SELECT CASE Outpt$
                            CASE "q"
                                Quit.Searching = True
                                Flag = False
                            CASE "n"
                                Flag = False
                        END SELECT
                    END IF
                END IF

                ' open work filename
                IF Flag THEN
                    V = FREEFILE
                    DataError = False
                    Error.Flag = True
                    OPEN Filelist.Filename$ FOR INPUT SHARED AS #V
                    Error.Flag = False

                    ' check error flag
                    IF DataError THEN
                        CALL DisplayError("Error opening filelist.")
                    ELSE
                        ' process input filenames
                        DO UNTIL EOF(V)
                            LINE INPUT #V, Next.Filename$
                            Next.Filename$ = RTRIM$(Next.Filename$)
                            IF LEN(Next.Filename$) THEN
                                ' process search filename
                                Command.Work = Next.Filename$
                                CALL Whereis_ProcessCommand
                                ' check recurse error
                                IF Max.DirRecurse >= RecurseLevel THEN
                                    EXIT DO
                                END IF
                            END IF
                        LOOP
                        CLOSE #V
                    END IF
                END IF
            ELSE
                ' display filename info
                CALL Whereis_DisplayFiles(Directory.Search$, Filename$)
            END IF
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF

        ' check max recurse levevl
        IF Max.DirRecurse >= RecurseLevel THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to display volume information
SUB Whereis_DisplayVolume (Var$)
' check drive exists
D = ASC(Var$) - 64
IF DRIVEEXISTS(D) THEN
    EXIT SUB
END IF

' increment volume counter
Volumes.Counted = Volumes.Counted + 1

' get drive info.
VarX$ = Var$ + ":\" + CHR$(0)
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)

' display volume information
Outpt$ = Var$ + ":"
IF CurrentCommand = CommandCount THEN
    COLOR Yellow, Black
    PRINT #StdoutHandle, Outpt$;
END IF
IF R = 0 THEN
    IF CurrentCommand = CommandCount THEN
        PRINT #StdoutHandle, DriveType;
    ELSE
        Outpt$ = Outpt$ + DriveType
    END IF
    IF Double.Line THEN
        IF CurrentCommand = CommandCount THEN
            COLOR Red, Black
            PRINT #StdoutHandle, " (????-????)";
            COLOR White, Black
            PRINT #StdoutHandle, " [????]";
        ELSE
            Outpt$ = Outpt$ + " (????-????) [????]"
        END IF
    END IF
    IF CurrentCommand = CommandCount THEN
        PRINT #StdoutHandle, ""
    ELSE
        CALL AddStdout2(Outpt$)
    END IF
    EXIT SUB
END IF

' volume label
Outpt2$ = RTRIM$(Vname$)
v = INSTR(Outpt2$, CHR$(0))
IF v THEN Outpt2$ = LEFT$(Outpt2$, v - 1)
IF Outpt2$ = "" THEN Outpt2$ = "<none>"

Outpt$ = Outpt$ + Outpt2$
IF CurrentCommand = CommandCount THEN
    PRINT #StdoutHandle, Outpt2$;
END IF
IF Double.Line = False THEN
    IF CurrentCommand = CommandCount THEN
        PRINT #StdoutHandle, ""
    ELSE
        CALL AddStdout2(Outpt$)
    END IF
    EXIT SUB
END IF

' display volume serial number
COLOR Red, Black
Outpt2$ = " (" + LEFT$(HEX$(serial~&), 4) + "-" + RIGHT$(HEX$(serial~&), 4) + ")"
Outpt$ = Outpt$ + Outpt2$
IF CurrentCommand = CommandCount THEN
    PRINT #StdoutHandle, Outpt2$;
END IF

' display volume system type.
COLOR White, Black
Var2$ = RTRIM$(Fname$)
v = INSTR(Var2$, CHR$(0))
IF v THEN Var2$ = LEFT$(Var2$, v - 1)
Outpt2$ = " [" + Var2$ + "]"
Outpt$ = Outpt$ + Outpt2$
IF CurrentCommand = CommandCount THEN
    PRINT #StdoutHandle, Outpt2$
ELSE
    CALL AddStdout2(Outpt$)
END IF
END SUB

' subroutine to display a filename
SUB Whereis_DisplayFiles (Search.Directory$, Search.Filename$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260

' get file attributes
Attribute = finddatatemp.dwFileAttributes

' make filename
File.List$ = Search.Directory$ + Short.Directory$
IF LEFT$(File.List$, 2) <> "\\" THEN
    File.List$ = Drive.Search$ + ":" + File.List$
END IF

' check file attribute
IF Display.Attribute THEN

    ' check for readonly file
    IF Display.Readonly THEN
        IF (Attribute AND &H1) <> &H1 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Readonly THEN
        IF (Attribute AND &H1) = &H1 THEN
            EXIT SUB
        END IF
    END IF

    ' check for hidden file
    IF Display.Hidden THEN
        IF (Attribute AND &H2) <> &H2 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Hidden THEN
        IF (Attribute AND &H2) = &H2 THEN
            EXIT SUB
        END IF
    END IF

    ' check for system file
    IF Display.System THEN
        IF (Attribute AND &H4) <> &H4 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.System THEN
        IF (Attribute AND &H4) = &H4 THEN
            EXIT SUB
        END IF
    END IF

    ' check for directory file
    IF Display.Directory THEN
        IF (Attribute AND &H10) <> &H10 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Directory THEN
        IF (Attribute AND &H10) = &H10 THEN
            EXIT SUB
        END IF
    END IF

    ' check for archive file
    IF Display.Archive THEN
        IF (Attribute AND &H20) <> &H20 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Archive THEN
        IF (Attribute AND &H20) = &H20 THEN
            EXIT SUB
        END IF
    END IF

    ' check for compressed file
    IF Display.Compress THEN
        IF (Attribute AND &H800) <> &H800 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Compress THEN
        IF (Attribute AND &H800) = &H800 THEN
            EXIT SUB
        END IF
    END IF

    ' check for encrypted file
    IF Display.Encrypt THEN
        IF (Attribute AND &H4000) <> &H4000 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Encrypt THEN
        IF (Attribute AND &H4000) = &H4000 THEN
            EXIT SUB
        END IF
    END IF

    ' check for no attributes
    IF Display.Any THEN
        IF (Attribute AND &H1) = &H1 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H2) = &H2 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H4) = &H4 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H10) = &H10 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H20) = &H20 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H800) = &H800 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H4000) = &H4000 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Any THEN
        IF (Attribute AND &H1) = &H0 THEN
            IF (Attribute AND &H2) = &H0 THEN
                IF (Attribute AND &H4) = &H0 THEN
                    IF (Attribute AND &H10) = &H0 THEN
                        IF (Attribute AND &H20) = &H0 THEN
                            IF (Attribute AND &H800) = &H0 THEN
                                IF (Attribute AND &H4000) = &H0 THEN
                                    EXIT SUB
                                END IF
                            END IF
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' check file size
IF Search.File.Size THEN
    IF Search.Size.From = FalseD AND Search.Size.To = FalseD THEN
        IF File.Size <> FalseD THEN
            EXIT SUB
        END IF
    ELSE
        IF Search.Size.From > FalseD OR Search.Size.To > FalseD THEN
            IF Search.Size.From > FalseD THEN
                IF Search.Size.To = FalseD THEN
                    IF File.Size < Search.Size.From THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From = FalseD THEN
                IF Search.Size.To > FalseD THEN
                    IF File.Size > Search.Size.To THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From <= Search.Size.To THEN
                IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
            IF Search.Size.From > Search.Size.To THEN
                IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        EXIT SUB
    END IF
    IF File.Work.Date > Search.To.Date THEN
        EXIT SUB
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        EXIT SUB
    END IF
    IF File.Work.Time > Search.To.Time THEN
        EXIT SUB
    END IF
END IF

' make filename
File.List$ = Search.Filename$
Display.Filename$ = Search.Filename$
IF Short.Filenames THEN
    Display.Filename$ = Short.Directory$
END IF
IF Short.Display THEN
    File.List$ = Display.Filename$
ELSE
    File.List$ = Search.Directory$ + Display.Filename$
END IF
IF No.Display.Drive = False THEN
    IF LEFT$(File.List$, 2) <> "\\" THEN
        File.List$ = Drive.Search + ":" + File.List$
    END IF
END IF
IF LEFT$(Search.Directory$, 2) = "\\" THEN
    IF MID$(File.List$, 2, 1) = ":" THEN
        File.List$ = MID$(File.List$, 3)
    END IF
END IF
IF (Attribute AND &H10) = &H10 THEN
    IF Display.Only.Errors = False THEN
        IF Remove.Slash = False THEN
            File.List$ = File.List$ + "\"
        END IF
    END IF
END IF

' check excluded file
IF LEN(Exclude.List) THEN
    Filename1$ = UCASE$(RTRIM$(Exclude.List))
    Filename2$ = UCASE$(RTRIM$(Search.Filename$))
    CALL CheckExcluded(Filename1$, Filename2$)
    IF Exclude.File THEN
        EXIT SUB
    END IF
END IF

' check for directory bit
IF (Attribute AND &H10) <> &H10 THEN
    ' increment files counted
    Files.Counted = Files.Counted + 1#
    Total.Bytes = Total.Bytes + File.Size
END IF

' display directory/filename
IF Display.Search = False THEN
    ' check for directory bit
    IF (Attribute AND &H10) = &H10 THEN
        Dirs.Displayed = Dirs.Displayed + 1#
    END IF

    ' format filename
    IF Display.Lowercase THEN
        Outpt$ = LCASE$(File.List$)
    ELSE
        IF Short.Filenames THEN
            Outpt$ = UCASE$(File.List$)
        ELSE
            Outpt$ = RTRIM$(File.List$)
        END IF
    END IF

    ' display output filename
    IF CurrentCommand < CommandCount THEN
        CALL AddStdout2(Outpt$)
        EXIT SUB ' exit after redirect to stream
    END IF

    ' display wide format
    IF Wide.Display THEN
        IF LEN(Outpt$) > 15 THEN
            Outpt$ = LEFT$(Outpt$, 14) + "~"
        END IF
        Outpt$ = LEFT$(Outpt$, 15)
        Outpt$ = Outpt$ + SPACE$(16 - LEN(Outpt$))
        COLOR Yellow, Black
        PRINT #StdoutHandle, Outpt$;
        Wide.List = Wide.List + 1
        IF Wide.List = 5 THEN
            Wide.List = False
            PRINT #StdoutHandle, ""
            Files.Counter = Files.Counter + 1
        END IF
    ELSE
        ' display filename
        IF LEN(Outpt$) THEN
            COLOR Yellow, Black
            PRINT #StdoutHandle, Outpt$
            Length = LEN(Outpt$)
            Files.Counter = Files.Counter + INT(Length / 80) + 1
        END IF
    END IF

    ' exit redirect to handle
    IF LCASE$(StdoutFile$) <> "scrn:" THEN
        EXIT SUB
    END IF

    ' check paginate
    CALL PagePrompt(Files.Counter, FlagX)
    IF FlagX THEN
        Files.Counter = 0
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT SUB
    END IF

    ' check max recurse levevl
    IF Max.DirRecurse >= RecurseLevel THEN
        EXIT SUB
    END IF

    ' reset line length
    Length = False

    ' check extended display info
    IF Double.Line THEN
        ' reset display flag
        Var = False

        ' display file date\time
        COLOR Green, Black
        IF Creation.Time THEN
            GOSUB Make.Date1
            GOSUB Convert.Date
            GOSUB Convert.Time
            Outpt$ = File.Date$ + " " + File.Time$
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
            Var = True
        END IF
        IF Access.Time THEN
            IF Creation.Time THEN
                Outpt$ = "\"
                Length = Length + LEN(Outpt$)
                COLOR White, Black
                PRINT Outpt$;
                COLOR Green, Black
            END IF
            GOSUB Make.Date2
            GOSUB Convert.Date
            GOSUB Convert.Time
            Outpt$ = File.Date$ + " " + File.Time$
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
            Var = True
        END IF
        IF Modified.Time THEN
            IF Creation.Time OR Access.Time THEN
                Outpt$ = "\"
                Length = Length + LEN(Outpt$)
                COLOR White, Black
                PRINT Outpt$;
                COLOR Green, Black
            END IF
            GOSUB Make.Date3
            GOSUB Convert.Date
            GOSUB Convert.Time
            Outpt$ = File.Date$ + " " + File.Time$
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
            Var = True
        END IF
        IF Var THEN ' check flag
            Outpt$ = " "
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for directory bit
        IF (Attribute AND &H10) = &H10 THEN
            Size$ = "<dir>"
        ELSE
            Size$ = FormatString$(File.Size)
        END IF

        ' display file size
        COLOR Red, Black
        Outpt$ = Size$
        Length = Length + LEN(Outpt$)
        PRINT Outpt$;
        COLOR White, Black

        ' check for directory file
        IF (Attribute AND &H10) = &H10 THEN
            Outpt$ = " Directory"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for read-only file
        IF (Attribute AND 1) = 1 THEN
            Outpt$ = " Read-only"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for hidden file
        IF (Attribute AND 2) = 2 THEN
            Outpt$ = " Hidden"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for system file
        IF (Attribute AND 4) = 4 THEN
            Outpt$ = " System"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for archive file
        IF (Attribute AND &H20) = &H20 THEN
            Outpt$ = " Archive"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for compressed
        IF (Attribute AND &H800) = &H800 THEN
            Outpt$ = " Compressed"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        ' check for encryption
        IF (Attribute AND &H4000) = &H4000 THEN
            Outpt$ = " Encrypted"
            Length = Length + LEN(Outpt$)
            PRINT Outpt$;
        END IF

        PRINT

        ' increment counter for extra line
        Files.Counter = Files.Counter + INT(Length / 80) + 1
    END IF

    ' check paginate
    CALL PagePrompt(Files.Counter, FlagX)
    IF FlagX THEN
        Files.Counter = 0
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT SUB
    END IF

    ' check max recurse levevl
    IF Max.DirRecurse >= RecurseLevel THEN
        EXIT SUB
    END IF
END IF

' check to run DOS command
IF LEN(DOS.Command) THEN
    CALL ExecuteCommand(Search.Directory$ + Display.Filename$)
END IF
EXIT SUB

Make.Date1:
x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
RETURN

Make.Date2:
x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
RETURN

Make.Date3:
x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
RETURN

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' check to paginate
SUB PagePrompt (Var, Flag)
Flag = False
IF LCASE$(StdoutFile$) <> "scrn:" THEN
    Flag = True
    EXIT SUB
END IF
IF Status THEN
    Z = 21
ELSE
    Z = 22
END IF
IF Var >= Z THEN
    Flag = True
    IF Continuous.Display = False THEN
        IF More.Display = False THEN
            Prompt2$ = "More(y/n/c)?"
            CALL MorePrompt(Prompt2$, "ync", Outpt$)
            SELECT CASE Outpt$
                CASE "c"
                    More.Display = True
                CASE "n"
                    Quit.Searching = True
            END SELECT
        END IF
    END IF
END IF
END SUB

' subroutine to shell to dos
SUB ShellSub (x$)
z$ = MID$(x$, 6, 1) ' shell
IF LEN(z$) THEN
    IF IsAlphaNum(z$) THEN
        PRINT "Type HELP for help."
        EXIT SUB
    END IF
END IF
IF LTRIM$(RTRIM$(MID$(x$, 6))) = "/?" THEN
    CALL DisplayHelp("shell")
    EXIT SUB
END IF
_TITLE Program + " - SHELL"
IF UCASE$(x$) = "SHELL" THEN
    SHELL
ELSE
    X = FREEFILE
    OPEN "QB64SHELL.BAT" FOR OUTPUT AS #X
    PRINT #X, "@ECHO OFF"
    PRINT #X, MID$(x$, 6)
    PRINT #X, "PAUSE"
    PRINT #X, "EXIT"
    CLOSE #X
    SHELL "QB64SHELL.BAT"
    f$ = "QB64SHELL.BAT" + CHR$(0)
    X = DeleteFileA(f$)
END IF
_TITLE Program
END SUB

' subroutine to execute DOS command
SUB ExecuteCommand (DOS.Filename$)
' make replacement parameters
DOS.Pathname$ = DOS.Filename$
Imbedded1 = INSTR(DOS.Pathname$, "\")
Imbedded2 = False
WHILE Imbedded1
    Imbedded2 = Imbedded1
    Imbedded1 = INSTR(Imbedded1 + 1, DOS.Pathname$, "\")
WEND
Filename$ = MID$(DOS.Pathname$, Imbedded2 + 1)
IF Imbedded2 THEN
    DOS.Pathname$ = LEFT$(DOS.Pathname$, Imbedded2 - 1)
ELSE
    DOS.Pathname$ = Nul
END IF
Extension$ = Nul
Imbedded = INSTR(Filename$, ".")
IF Imbedded THEN
    Extension$ = MID$(Filename$, Imbedded + 1)
    Filename$ = LEFT$(Filename$, Imbedded - 1)
END IF

' store replacement parameters
IF LEFT$(DOS.Filename$, 2) = "\\" THEN
    Parameters(1) = "\\"
    Parameters(2) = "\\"
    Parameters(3) = DOS.Pathname$
    Parameters(4) = DOS.Pathname$ + "\"
    Parameters(5) = DOS.Filename$
ELSE
    Parameters(1) = Drive.Search + ":"
    Parameters(2) = Drive.Search + ":\"
    Parameters(3) = Drive.Search + ":" + DOS.Pathname$
    Parameters(4) = Drive.Search + ":" + DOS.Pathname$ + "\"
    Parameters(5) = Drive.Search + ":" + DOS.Filename$
END IF
Parameters(6) = DOS.Pathname$
Parameters(7) = DOS.Pathname$ + "\"
Parameters(8) = DOS.Filename$
Parameters(9) = Filename$ + "." + Extension$
Parameters(10) = Filename$
Parameters(11) = "." + Extension$
Parameters(12) = Extension$

' create DOS command
Execute.Command$ = DOS.Command

' replace parameters
Count = 1
DO
    IF Count >= LEN(Execute.Command$) THEN
        EXIT DO
    END IF
    Imbed$ = MID$(Execute.Command$, Count, 1)
    IF Imbed$ = "%" THEN
        Imbed2$ = MID$(Execute.Command$, Count + 1, 1)
        SELECT CASE UCASE$(Imbed2$)
            CASE "1" TO "9", "A" TO "C"
                Count2 = VAL("&H" + Imbed2$)
                Execute.Command$ = LEFT$(Execute.Command$, Count - 1) + RTRIM$(Parameters(Count2)) + MID$(Execute.Command$, Count + 2)
                Count = Count + LEN(RTRIM$(Parameters(Count2)))
            CASE ELSE
                Count = Count + 1
        END SELECT
    ELSE
        Count = Count + 1
    END IF
LOOP

' start DOS command shell
IF LEN(Execute.Command$) THEN
    ' check display counter
    IF Wide.List THEN
        Wide.List = False
        PRINT
    END IF

    ' read command path from environment
    Comspec$ = ENVIRON$("COMSPEC")
    IF LEN(Comspec$) THEN
        ' call shell routine
        IF Disable.Cmd THEN
            Shell.Command$ = Comspec$ + " /C " + Execute.Command$
        ELSE
            Shell.Command$ = Comspec$ + " /K " + Execute.Command$
        END IF
        SHELL Shell.Command$
    END IF
END IF
END SUB

' routine compares occurrence of filename1$ in filename2$
' with pattern matching.
SUB CheckExcluded (Filename1$, Filename2$)
Exclude.File = True ' assume mask matches filename2.
Length1 = 1
Length2 = 1
DO
    ' global replacement.
    IF MID$(Filename1$, Length1, 1) = "*" THEN
        DO
            Length1 = Length1 + 1
            IF Length1 > LEN(Filename1$) THEN
                EXIT SUB
            END IF
            ' global replacement followed by exclusion character.
            ' searches remaining string until exclusion character found or not.
            IF MID$(Filename1$, Length1, 1) = "^" THEN
                Length1 = Length1 + 1
                Not.Include$ = MID$(Filename1$, Length1, 1)
                DO
                    IF Not.Include$ <> MID$(Filename2$, Length2, 1) THEN
                        Length2 = Length2 + 1
                    ELSE
                        Exclude.File = False
                        EXIT SUB
                    END IF
                    IF Length2 > LEN(Filename2$) THEN
                        EXIT SUB
                    END IF
                LOOP
            END IF
            ' global replacement followed by ? or another *
            ' skips to next character.
            IF MID$(Filename1$, Length1, 1) <> "*" THEN
                IF MID$(Filename1$, Length1, 1) <> "?" THEN
                    EXIT DO
                END IF
            END IF
        LOOP
        ' global replacement.
        ' searches for next matching character.
        DO
            IF MID$(Filename1$, Length1, 1) = MID$(Filename2$, Length2, 1) THEN
                EXIT DO
            ELSE
                Length2 = Length2 + 1
            END IF
            IF Length2 > LEN(Filename2$) THEN
                EXIT DO
            END IF
        LOOP
    ELSE
        ' character replacement.
        ' matches any next character.
        IF MID$(Filename1$, Length1, 1) = "?" THEN
            Length1 = Length1 + 1
            Length2 = Length2 + 1
        ELSE
            ' exclusion character.
            ' checks next character unmatched.
            IF MID$(Filename1$, Length1, 1) = "^" THEN
                Length1 = Length1 + 1
                Not.Include$ = MID$(Filename1$, Length1, 1)
                IF Not.Include$ <> MID$(Filename2$, Length2, 1) THEN
                    Length1 = Length1 + 1
                    Length2 = Length2 + 1
                ELSE
                    Exclude.File = False
                    EXIT DO
                END IF
            ELSE
                ' matches next character.
                IF MID$(Filename1$, Length1, 1) = MID$(Filename2$, Length2, 1) THEN
                    Length1 = Length1 + 1
                    Length2 = Length2 + 1
                ELSE
                    Exclude.File = False
                    EXIT DO
                END IF
                ' check string lengths.
                IF Length1 > LEN(Filename1$) THEN
                    IF Length2 <= LEN(Filename2$) THEN
                        Exclude.File = False
                    END IF
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
END SUB

SUB Zsort (ErrFlag)
' check command line
StdoutLines = 0
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOTO Boot.Usage
END IF

' check command line switches
Continuous.Display = ParseLine("/C")
Ignore.Case = ParseLine("/I")
Reverse.Sort = ParseLine("/R")
Strip.Blanks = ParseLine("/T")

' get sort column
Sort.Column = 1
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Imbedded2 = Imbedded + 2
    DO
        Switch$ = MID$(Command.Line, Imbedded2, 1)
        IF Switch$ >= "0" AND Switch$ <= "9" THEN
            Column$ = Column$ + Switch$
        ELSE
            EXIT DO
        END IF
        Imbedded2 = Imbedded2 + 1
    LOOP
    IF Column$ = Nul THEN
        GOTO Boot.Error
    END IF
    Sort.Column = INT(VAL(Column$))
    IF Sort.Column = False THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded2)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

IF Stdin THEN
    CALL SortStdin
END IF
IF Command.Line = "" THEN
    EXIT SUB
END IF

' filename processing loop
DO
    ' store entire command
    IF LEFT$(Command.Line, 1) = CHR$(34) THEN
        Imbedded = INSTR(2, Command.Line, CHR$(34))
        IF Imbedded THEN
            Command.Work = MID$(Command.Line, 2, Imbedded - 2)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    ELSE
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    END IF
    Command.Line = LTRIM$(Command.Line)
    Command.Line = RTRIM$(Command.Line)

    ' sort parsed command line
    CALL SortFiles

    ' check search filename
    IF Command.Line = "" THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP
EXIT SUB

Boot.Usage:
' make header
COLOR White, Black
PRINT "SORT"
PRINT "   Sort utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   SORT [\path\][filename] [/cinrt]"
PRINT "Where:"
PRINT "   [\path\][filename] are the filename(s) to sort."
PRINT "Switches:"
PRINT "   /c  continuous list"
PRINT "   /i  ignore case"
PRINT "   /n###  sort at column"
PRINT "   /r  reverse order"
PRINT "   /t  strip blank lines"
COLOR Plain, Black
EXIT SUB

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type SORT /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

SUB SortStdin
' shell sort
Sort.Swaps = False
Num = StdinLines
Span = INT(Num / 2)
DO WHILE Span > False
    FOR Start = Span TO Num - 1
        FOR Element = (Start - Span + 1) TO 1 STEP -Span
            Sort.Column1$ = RTRIM$(MID$(StdinArray(Element), Sort.Column))
            Sort.Column2$ = RTRIM$(MID$(StdinArray(Element + Span), Sort.Column))
            IF Ignore.Case THEN
                Sort.Column1$ = UCASE$(Sort.Column1$)
                Sort.Column2$ = UCASE$(Sort.Column2$)
            END IF
            IF Reverse.Sort THEN
                IF Sort.Column2$ = "" THEN
                    EXIT FOR
                ELSE
                    IF Sort.Column2$ <= Sort.Column1$ THEN
                        EXIT FOR
                    END IF
                END IF
            ELSE
                IF Sort.Column1$ = "" THEN
                    EXIT FOR
                ELSE
                    IF Sort.Column1$ <= Sort.Column2$ THEN
                        EXIT FOR
                    END IF
                END IF
            END IF
            SWAP StdinArray(Element), StdinArray(Element + Span)
            Sort.Swaps = Sort.Swaps + 1
        NEXT
    NEXT
    Span = INT(Span / 2)
LOOP

' output array
FOR Array.Line# = 1 TO StdinLines
    X$ = RTRIM$(StdinArray(Array.Line#))
    CALL AddStdout(X$)
NEXT
EXIT SUB

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    PRINT "Lines counted"; StdinLines
    PRINT "Sort swaps made"; Sort.Swaps
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
END SUB

SUB SortFiles
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' redimension sort array
REDIM Sort.Array(1) AS STRING

DIM hfind AS _OFFSET

' make filename
Filename1$ = Command.Work
ASCIIZ = Filename1$ + CHR$(0)

' check filename
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    r = FindClose(Wfile.Handle)
ELSE
    EXIT SUB
END IF

' test file open
f$ = Filename1$ + CHR$(0)
hfind = CreateFile(f$, &H180, 0, 0, 4, 0, 0)
IF hfind = 0 THEN
    EXIT SUB
END IF
r = CloseHandle(hfind)

' check file exists
IF _FILEEXISTS(Filename1$) = 0 THEN
    EXIT SUB
END IF

' reset count variables
Lines.Counted = 0
Max.Sort.Lines = 1

' read array from file
X = FREEFILE
OPEN Filename1$ FOR BINARY AS #X
Count# = 0#
IF LOF(X) > 0 THEN
    DO UNTIL EOF(X)
        LINE INPUT #X, line1$
        Count# = Count# + LEN(line1$) + 2#
        Lines.Counted = Lines.Counted + 1
        IF Lines.Counted > Max.Sort.Lines THEN
            Max.Sort.Lines = Max.Sort.Lines + 1
            REDIM _PRESERVE Sort.Array(Max.Sort.Lines) AS STRING
        END IF
        Sort.Array(Lines.Counted) = line1$
    LOOP
    IF Count# < LOF(X) THEN
        Lines.Counted = Lines.Counted + 1
        IF Lines.Counted > Max.Sort.Lines THEN
            Max.Sort.Lines = Max.Sort.Lines + 1
            REDIM _PRESERVE Sort.Array(Max.Sort.Lines) AS STRING
        END IF
        Sort.Array(Lines.Counted) = ""
    END IF
END IF
CLOSE #X

' shell sort
Sort.Swaps = False
Num = Lines.Counted
Span = INT(Num / 2)
DO WHILE Span > False
    FOR Start = Span TO Num - 1
        FOR Element = (Start - Span + 1) TO 1 STEP -Span
            Sort.Column1$ = RTRIM$(MID$(Sort.Array(Element), Sort.Column))
            Sort.Column2$ = RTRIM$(MID$(Sort.Array(Element + Span), Sort.Column))
            IF Ignore.Case THEN
                Sort.Column1$ = UCASE$(Sort.Column1$)
                Sort.Column2$ = UCASE$(Sort.Column2$)
            END IF
            IF Reverse.Sort THEN
                IF Sort.Column2$ = "" THEN
                    EXIT FOR
                ELSE
                    IF Sort.Column2$ <= Sort.Column1$ THEN
                        EXIT FOR
                    END IF
                END IF
            ELSE
                IF Sort.Column1$ = "" THEN
                    EXIT FOR
                ELSE
                    IF Sort.Column1$ <= Sort.Column2$ THEN
                        EXIT FOR
                    END IF
                END IF
            END IF
            SWAP Sort.Array(Element), Sort.Array(Element + Span)
            Sort.Swaps = Sort.Swaps + 1
        NEXT
    NEXT
    Span = INT(Span / 2)
LOOP

' output array
FOR Array.Line# = 1 TO Lines.Counted
    X$ = RTRIM$(Sort.Array(Array.Line#))
    CALL AddStdout(X$)
NEXT

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    PRINT "Lines counted"; Lines.Counted
    PRINT "Sort swaps made"; Sort.Swaps
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
END SUB

SUB TypeY (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
StdoutLines = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOTO Boot.Usage
END IF

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$

' get switches from command line
No.Display.Archive = ParseLine("//A")
No.Display.Hidden = ParseLine("//H")
No.Display.Readonly = ParseLine("//O")
No.Display.System = ParseLine("//S")
No.Display.Any = ParseLine("//X")
No.Display.Compress = ParseLine("//M1")
No.Display.Encrypt = ParseLine("//M2")

Display.Archive = ParseLine("/A")
Display.Hidden = ParseLine("/H")
Display.Readonly = ParseLine("/O")
Display.System = ParseLine("/S")
Display.Any = ParseLine("/X")
Display.Compress = ParseLine("/M1")
Display.Encrypt = ParseLine("/M2")

Continuous.Display = ParseLine("/C")
Prompt.More = ParseLine("/P")
Recurse.Directories = ParseLine("/R")
Display.Errors = ParseLine("/Z")
Strip.Blanks = ParseLine("/I")

' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32 + INT(T3! / 2)
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32 + INT(T3! / 2)
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get display column
Sort.Column = 1
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Imbedded2 = Imbedded + 2
    DO
        Switch$ = MID$(Command.Line, Imbedded2, 1)
        IF Switch$ >= "0" AND Switch$ <= "9" THEN
            Column$ = Column$ + Switch$
        ELSE
            EXIT DO
        END IF
        Imbedded2 = Imbedded2 + 1
    LOOP
    IF Column$ = Nul THEN
        GOTO Boot.Error
    END IF
    Sort.Column = INT(VAL(Column$))
    IF Sort.Column = False THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded2)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' initialize work variables
Bytes.Counted = False
Lines.Counted = False
More.Display = False
Quit.Searching = False

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header
IF Stdin THEN
    CALL TypeStdin
END IF
IF Command.Line = "" THEN
    EXIT SUB
END IF

' filename processing loop
DO
    ' store entire command
    IF LEFT$(Command.Line, 1) = CHR$(34) THEN
        Imbedded = INSTR(2, Command.Line, CHR$(34))
        IF Imbedded THEN
            Command.Work = MID$(Command.Line, 2, Imbedded - 2)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    ELSE
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    END IF
    Command.Line = LTRIM$(Command.Line)
    Command.Line = RTRIM$(Command.Line)

    ' store current drive
    IF MID$(Command.Work, 2, 1) = ":" THEN
        Drive.Search = LEFT$(Command.Work, 1)
        Command.Work = MID$(Command.Work, 3)
    ELSE
        Drive.Search = Current.Drive$
    END IF
    Drive.Search = UCASE$(Drive.Search)

    ' store current directory
    Directory.Search$ = ""
    Imbedded1 = INSTR(Command.Work, "\")
    Imbedded2 = Imbedded1
    WHILE Imbedded1
        Imbedded2 = Imbedded1
        Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
    WEND
    IF Imbedded2 THEN
        Directory.Search$ = LEFT$(Command.Work, Imbedded2)
        Command.Work = MID$(Command.Work, Imbedded2 + 1)
    END IF
    IF Directory.Search$ = "" THEN
        IF LEFT$(Current.Directory, 2) = "\\" THEN
            Directory.Search$ = Current.Directory
        ELSE
            IF Drive.Search$ = Current.Drive$ THEN
                Directory.Search$ = MID$(Current.Directory, 3)
            ELSE
                Directory.Search$ = "\"
            END IF
        END IF
    END IF
    IF RIGHT$(Directory.Search$, 1) <> "\" THEN
        Directory.Search$ = Directory.Search$ + "\"
    END IF
   
    ' get filename spec
    Filename.Search$ = Command.Work
    IF Filename.Search$ = Nul THEN
        Filename.Search$ = "*.*"
    END IF
    Command.Work = Nul

    ' display search filename
    IF Continuous.Display = False THEN
        COLOR Yellow, Black
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            PRINT "Searching: " + Directory.Search$ + Filename.Search$
        ELSE
            PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
        END IF
    END IF

    ' call routine to search for files
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        CALL TypeY_Directories(Directory.Search$, Filename.Search$)
    ELSE
        V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
        IF MEDIAEXISTS(V) THEN
            CALL TypeY_Directories(Directory.Search$, Filename.Search$)
        END IF
    END IF

    ' check search filename
    IF Command.Line = Nul THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "TYPE"
PRINT "   File display utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   TYPE [d:\path\filename.ext][//ahmosx][/cdfinprtz][/123]"
PRINT "Where:"
PRINT "   /c  continuous display   /i  strip blanks"
PRINT "   /p  prompt for more      /nxxx  column display"
PRINT "   /r  recurse directories"
PRINT "   /z  suppress error messages"
PRINT "   display ranges: (/1  creation, /2  last access  /3  modify time)"
PRINT "     /d  is range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "     /t  is range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "     /f  is range of file sizes in form xxx-xxx in kilobytes"
PRINT "   type files with attributes"
PRINT "     / prefix for files with, // prefix for files without,"
PRINT "       a  archive, h  hidden, o  read-only, s  system, x  none"
PRINT "       m1  compressed, m2  encrypted"
COLOR Plain, Black
EXIT SUB

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type TYPE /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File content display utility; "
END IF
RETURN
END SUB

' subroutine to access directories
SUB TypeY_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search directory names
    CALL TypeY_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Deleting THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL TypeY_Directories(Next.Directory$, Filename.Search$)
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB TypeY_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' check directory attribute
    Attribute = finddata.dwFileAttributes

    ' check for directory
    IF (Attribute AND &H10) = &H10 THEN
        ' increment directories actually searched
        Dirs.Counted = Dirs.Counted + 1#
    END IF

    ' recurse filename
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check unicode
        IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
            Filename$ = finddata.cAlternateFileName
            V = INSTR(Filename$, CHR$(0))
            IF V THEN Filename$ = LEFT$(Filename$, V - 1)
        END IF

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' store file data
            finddatatemp = finddata

            ' store file size
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) <> &H10 THEN
                CALL TypeY_DisplayFiles(Directory.Search$, Filename$)
            END IF
        END IF

        ' check to quit
        IF Quit.Deleting THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to count and display filenames
SUB TypeY_DisplayFiles (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG

' check directory attribute
Attribute = finddatatemp.dwFileAttributes

' check for readonly file
IF (Attribute AND &H1) = &H1 THEN
    IF No.Display.Readonly THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Readonly THEN
        EXIT SUB
    END IF
END IF

' check for hidden file
IF (Attribute AND &H2) = &H2 THEN
    IF No.Display.Hidden THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Hidden THEN
        EXIT SUB
    END IF
END IF

' check for system file
IF (Attribute AND &H4) = &H4 THEN
    IF No.Display.System THEN
        EXIT SUB
    END IF
ELSE
    IF Display.System THEN
        EXIT SUB
    END IF
END IF

' check for archive file
IF (Attribute AND &H20) = &H20 THEN
    IF No.Display.Archive THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Archive THEN
        EXIT SUB
    END IF
END IF

' check for compressed file
IF (Attribute AND &H800) = &H800 THEN
    IF No.Display.Compress THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Compress THEN
        EXIT SUB
    END IF
END IF

' check for encrypted file
IF (Attribute AND &H4000) = &H4000 THEN
    IF No.Display.Encrypt THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Encrypt THEN
        EXIT SUB
    END IF
END IF

' check all attributes
IF Display.Any THEN
    IF (Attribute AND &H1) = &H1 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H2) = &H2 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4) = &H4 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H20) = &H20 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H800) = &H800 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4000) = &H4000 THEN
        EXIT SUB
    END IF
END IF
IF No.Display.Any THEN
    IF (Attribute AND &H1) = False THEN
        IF (Attribute AND &H2) = False THEN
            IF (Attribute AND &H4) = False THEN
                IF (Attribute AND &H20) = False THEN
                    IF (Attribute AND &H800) = False THEN
                        IF (Attribute AND &H4000) = False THEN
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' check file size
IF Search.File.Size THEN
    IF Search.Size.From = FalseD AND Search.Size.To = FalseD THEN
        IF File.Size <> FalseD THEN
            EXIT SUB
        END IF
    ELSE
        IF Search.Size.From > FalseD OR Search.Size.To > FalseD THEN
            IF Search.Size.From > FalseD THEN
                IF Search.Size.To = FalseD THEN
                    IF File.Size < Search.Size.From THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From = FalseD THEN
                IF Search.Size.To > FalseD THEN
                    IF File.Size > Search.Size.To THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From <= Search.Size.To THEN
                IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
            IF Search.Size.From > Search.Size.To THEN
                IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        EXIT SUB
    END IF
    IF File.Work.Date > Search.To.Date THEN
        EXIT SUB
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        EXIT SUB
    END IF
    IF File.Work.Time > Search.To.Time THEN
        EXIT SUB
    END IF
END IF

' make directory filename
IF Display.Filenames THEN
    Filename$ = Directory.Search$ + Filename.Search$
    IF Prepend.Drive THEN
        IF LEFT$(Filename$, 2) <> "\\" THEN
            Filename$ = Drive.Search + ":" + Filename$
        END IF
    END IF
    COLOR Yellow, Black
    PRINT Filename$
END IF

' update bytes counted
Bytes.Counted = Bytes.Counted + File.Size

' update files counter
Files.Counted = Files.Counted + 1#

' send contents of file to screen
CALL TypeY_DisplayFileLines(Directory.Search$, Filename.Search$)
EXIT SUB

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' subroutine to display lines in a filename
SUB TypeY_DisplayFileLines (Directory.Search$, Filename.Search$)
' make filename
Filename.Count$ = Directory.Search$ + Filename.Search$
IF ValidFileChar(Filename.Search$) THEN
    Short.Filename$ = finddatatemp.cAlternateFileName
    V = INSTR(Short.Filename$, CHR$(0))
    IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
    Filename.Count$ = Directory.Search$ + Short.Filename$
    IF ValidFileChar(Short.Filename$) THEN
        EXIT SUB
    END IF
END IF
IF _FILEEXISTS(Filename.Count$) THEN
    X = FREEFILE
    OPEN Filename.Count$ FOR BINARY SHARED AS #X
    Count# = 0#
    IF LOF(X) > 0 THEN
        DO UNTIL EOF(X)
            COLOR 7, 0
            LINE INPUT #X, X$
            Count# = Count# + LEN(X$) + 2#
            X$ = MID$(X$, Sort.Column)
            CALL AddStdout(X$)
        LOOP
    END IF
    IF Count# < LOF(X) THEN
        CALL AddStdout("")
    END IF
    CLOSE #X
END IF
END SUB

' subroutine to display lines in standard input
SUB TypeStdin
COLOR 7, 0
FOR StdinCount = 1 TO StdinLines
    X$ = MID$(StdinArray(StdinCount), Sort.Column)
    CALL AddStdout(X$)
NEXT
EXIT SUB
' display end program
IF Continuous.Display = False THEN
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
END SUB

SUB AddStdout (X$)
IF RTRIM$(X$) = "" THEN
    IF Strip.Blanks THEN
        EXIT SUB
    END IF
END IF
IF CurrentCommand < CommandCount THEN
    StdoutLines = StdoutLines + 1
    IF StdoutLines > StdoutMax THEN
        StdoutMax = StdoutLines
        REDIM _PRESERVE StdoutArray(StdoutMax) AS STRING
    END IF
    StdoutArray(StdoutLines) = X$
ELSE
    IF LCASE$(StdoutFile$) = "scrn:" THEN
        _CONTROLCHR OFF
        PRINT X$
        _CONTROLCHR ON
    ELSE
        PRINT #StdoutHandle, X$
    END IF
END IF
END SUB

SUB AddStdout2 (X$)
IF CurrentCommand < CommandCount THEN
    StdoutLines = StdoutLines + 1
    IF StdoutLines > StdoutMax THEN
        StdoutMax = StdoutLines
        REDIM _PRESERVE StdoutArray(StdoutMax) AS STRING
    END IF
    StdoutArray(StdoutLines) = X$
    EXIT SUB
END IF
PRINT #StdoutHandle, X$
END SUB

REM file: Copyit.bas - Public Domain QB64 Utility v9.0a r4.0a 2017.

SUB Copyit (ErrFlag)
_TITLE "COPYIT"

' get command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = Nul THEN
    COLOR Yellow, Black
    PRINT "Type COPY /? for help. Or edit COPYIT.CFG file."
    COLOR Plain, Black
    EXIT SUB
END IF

' read command line switches
CALL ReadSwitches(VarX)
IF VarX > 0 THEN
    ErrFlag = -1
    COLOR Plain, Black
    EXIT SUB
END IF

' store skip file list
Skip.Filenames(1) = "WIN386.SWP"
Skip.Filenames(2) = "PAGEFILE.SYS"
Skip.Filenames(3) = "HIBERFIL.SYS"
Skip.Filenames(4) = "SWAPFILE.SYS"

' read current drive
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' store default directory
IF LEFT$(_CWD$, 2) = "\\" THEN
    Default.Dir = _CWD$
ELSE
    Default.Dir = MID$(_CWD$, 3)
END IF

' store default server name
Default.Net = Nul
Var$ = ENVIRON$("COMPUTERNAME")
IF LEN(Var$) THEN
    Default.Net = "\\" + Var$
END IF

' reset number of destination filenames
IF Number.Dest.Files = False THEN
    Number.Dest.Files = 1
    Destinate.Filename(1) = Nul
END IF

' reset number of destination directories
IF Number.Dest.Dirs = False OR Zero.Dirs THEN
    Zero.Dirs = True
    Number.Dest.Dirs = 1
    IF Number.Dest.Nets = False THEN
        Destinate.Directory(1) = Default.Dir
    ELSE
        Destinate.Directory(1) = "\"
    END IF
END IF

' reset and check netpaths
IF Zero.Nets THEN
    Var2$ = RTRIM$(Default.Net)
    IF LEN(Var2$) THEN
        Var2$ = Var2$ + LCASE$(Current.Drive) + "\"
        Number.Dest.Nets = 1
        Destinate.Netpaths(1) = Var2$
    END IF
END IF
IF Number.Dest.Nets = False THEN
    Number.Dest.Nets = 1
    Destinate.Netpaths(1) = Nul
END IF
FOR Count0 = 1 TO Number.Dest.Nets
    Next.Dest.Net$ = Destinate.Netpaths(Count0)
    Next.Dest.Net$ = RTRIM$(Next.Dest.Net$)
    IF Next.Dest.Net$ <> Nul THEN
        ' check netpath
        IF RIGHT$(Next.Dest.Net$, 1) = "\" THEN
            Next.Dest.Net$ = LEFT$(Next.Dest.Net$, LEN(Next.Dest.Net$) - 1)
        END IF
        IF LEFT$(Next.Dest.Net$, 2) <> "\\" THEN
            Next.Dest.Net$ = "\" + Next.Dest.Net$
        END IF
        IF LEFT$(Next.Dest.Net$, 2) <> "\\" THEN
            Next.Dest.Net$ = "\" + Next.Dest.Net$
        END IF
        Destinate.Netpaths(Count0) = Next.Dest.Net$
    END IF
NEXT

' check /d omitted or specified w/o parameters
IF Zero.Dirs THEN
    ' compare /g\\serv\c\ to default.net
    IF Number.Dest.Nets = 1 THEN
        Var1$ = RTRIM$(Destinate.Netpaths(1))
        IF LEN(Var1$) THEN
            Var1$ = UCASE$(Var1$) + "\"
            Var2$ = RTRIM$(Default.Net)
            IF LEN(Var2$) THEN
                Var2$ = UCASE$(Var2$) + Current.Drive + "\"
                IF Var1$ = Var2$ THEN
                    Destinate.Directory(1) = Default.Dir
                END IF
            END IF
        END IF
    END IF
END IF

' store time copying began
Start.Time = TIMER

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' search filespecs
    DO

        ' store input filename
        Stored.Command.Line$ = Command.Line

        ' create filenames
        CALL Copyit_MakeFilename

        ' locate filelist character
        Command.Out$ = Command.Work

        ' check filelist character
        Flag = FilelistIS(Command.Out$)
        IF Flag AND Enable.Redirect THEN

            ' reset recurse counter
            Max.CopyRecurse = 0

            ' process work filename containing filelist
            CALL Copyit_ProcessFilelist(Command.Out$, -1)

            ' check overwrite flag
            IF Quit.Searching THEN
                Error.Level = 1
                GOTO End.Copy
            END IF
        ELSE

            ' process filename
            CALL Copyit_ProcessCommand(Quit)

            ' check overwrite flag
            IF Quit OR Quit.Searching THEN
                Error.Level = 1
                GOTO End.Copy
            END IF
        END IF

        ' check command line
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' program end label
End.Copy:

' check append files
IF Append.Files THEN
    CALL Copyit_CloseFile(-1)
END IF

' calculate time elapsed during file copying
Stop.Time = TIMER
Time.Elapsed = Stop.Time - Start.Time

' check timing past midnight
IF Time.Elapsed < SFalse THEN
    Time.Elapsed = Time.Elapsed + 86400! ' add one day of seconds
END IF

' display end program
IF Wide.List THEN
    Wide.List = False
    PRINT
END IF
IF Continuous.Display = False THEN
    IF Continuous.Display2 = False THEN
        COLOR Yellow, Black
        PRINT "Directories copied: ";
        PRINT FormatString$(Dirs.Counted)
        IF Dirs.Created > 0 THEN
            PRINT "Directories created: ";
            PRINT FormatString$(Dirs.Created)
        END IF
        IF Double.Line >= 4 THEN
            PRINT "Directories deleted: ";
            PRINT FormatString$(Total.Deleted)
        END IF
        PRINT "Files copied: ";
        PRINT FormatString$(Files.Counted)
        IF Double.Line >= 4 THEN
            PRINT "Bytes counted: ";
            PRINT FormatString$(Bytes.Copied)
            PRINT "Bytes copied: ";
            PRINT FormatString$(Total.Bytes)
            IF Deleted.Files > 0# THEN
                PRINT "Files deleted: ";
                PRINT FormatString$(Deleted.Files)
            END IF
        END IF
    END IF

    ' check to display time/rate
    IF Rate.Mode THEN
        ' display transfer time
        CALL Time.Display1

        ' display transfer rate
        CALL Time.Display2
    END IF

    ' prompt to exit Copyit
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt2(Prompt2$, CHR$(13), Outpt$, CHR$(13))
END IF

' reset color
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "COPY " + CopyitVersion$ + " " + CopyitRelease$ + ": File copy utility; "
END IF
RETURN
END SUB

' clear display progress
SUB Copyit_ClearDisplay (Var)
' break wide list
IF Var = 0 THEN
    IF Wide.List THEN
        Wide.List = False
        PRINT
    END IF
END IF
' clear dots
FOR Count = 1 TO Dot.Count
    PRINT CHR$(29); " "; CHR$(29);
NEXT
' clear percent display
IF Percent.Display THEN
    IF Percent.Flag THEN
        FOR Imbedded = 1 TO 4
            PRINT CHR$(29); " "; CHR$(29);
        NEXT
    END IF
END IF
' clear progress bar
IF Progress.Bar THEN
    FOR Imbedded = 1 TO Current.Progress
        PRINT CHR$(29); " "; CHR$(29);
    NEXT
END IF
Dot.Count = False
Dot.Total = False
Count.Forward = True
Current.Progress = False
END SUB

' process filelist recursively
SUB Copyit_ProcessFilelist (Command.Out$, PromptFlag)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' check recurse error
IF Max.CopyRecurse >= RecurseLevel THEN
    EXIT SUB
END IF

' process work filename
Var$ = Nul
IF Command.Out$ <> Nul THEN
    IF _FILEEXISTS(Command.Out$) THEN
        Var$ = Command.Out$
    END IF
END IF

' check filelist exists
IF Var$ <> Nul THEN

    ' concatenate work filename
    Var$ = Command.Work ' command line filename
    FOR Var = LEN(Var$) TO 1 STEP -1
        IF MID$(Var$, Var, 1) = "\" THEN
            Var$ = MID$(Var$, Var + 1)
            EXIT FOR
        END IF
    NEXT
    IF MID$(Var$, 2, 1) = ":" THEN
        Var$ = MID$(Var$, 3)
    END IF
    Filename2$ = Var$
    Var$ = MID$(Var$, 2)
    Var2$ = Command.Out$
    FOR Var = LEN(Var2$) TO 1 STEP -1
        IF MID$(Var2$, Var, 1) = "\" THEN
            Var2$ = LEFT$(Var2$, Var)
            Var$ = Var2$ + Var$
            EXIT FOR
        END IF
    NEXT

    ' check filename
    CALL Copyit_CheckFile(Filename2$, True)
    IF Check.Flag = False THEN
        Var$ = Nul
    END IF

    ' prompt for filelist
    IF Disable.Prompts = False THEN
        IF Var$ <> Nul THEN
            IF PromptFlag THEN
                ' prompt to use filelist
                Prompt2$ = "Use '" + LCASE$(Var$) + "' as filelist(y)es,(n)o,(q)uit?"
                CALL MorePrompt2(Prompt2$, "ynq", Outpt$, "y")
                SELECT CASE Outpt$
                    CASE "q"
                        Quit.Searching = True
                        EXIT SUB
                    CASE "n"
                        Var$ = Nul
                        EXIT SUB
                END SELECT
            END IF
        END IF
    END IF
END IF

' check filelist exists
IF Var$ = Nul THEN

    ' restore original command line
    Command.Line = Stored.Command.Line$

    ' create filenames
    CALL Copyit_MakeFilename

    ' continue processing filename
    EXIT SUB
END IF

' open filename list
x = 0
ASCIIZ = Command.Out$ + CHR$(0)
hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, 0, 0)
IF hfind <> INVALID_HANDLE_VALUE THEN
    x = CloseHandle(hfind)
END IF
IF x = 0 THEN
    ' display filename error
    CALL Copyit_DisplayError("Error" + STR$(GetLastError&) + " opening listfile.")
    EXIT SUB
END IF

V = FREEFILE
DataError = False
Error.Flag = True
OPEN Command.Out$ FOR INPUT SHARED AS #V
Error.Flag = False

' check error flag
IF DataError THEN
    ' display filename error
    CALL Copyit_DisplayError("Error" + STR$(DataError) + " opening listfile.")
    EXIT SUB
END IF

' check recursion level
Max.CopyRecurse = Max.CopyRecurse + 1
IF Max.CopyRecurse >= RecurseLevel THEN
    CALL Copyit_DisplayError("Recursive filelist error.")
    EXIT SUB
END IF

' process input filenames
DO UNTIL EOF(V)

    ' read next filename from input file
    LINE INPUT #V, Command.Work
    Command.Work = RTRIM$(Command.Work)

    ' check input
    IF LEN(Command.Work) THEN

        ' locate filelist character
        Var2$ = Command.Work
        Flag = FilelistIS(Var2$)
        IF Flag THEN
            ' recursively call filelist
            CALL Copyit_ProcessFilelist(Var2$, 0)
        ELSE
            ' process filename
            CALL Copyit_ProcessCommand(Quit)
        END IF

        ' check overwrite flag
        IF Quit OR Quit.Searching THEN
            Error.Level = 1
            EXIT DO
        END IF
    END IF
LOOP
CLOSE #V

' check recursion level
Max.CopyRecurse = Max.CopyRecurse - 1
END SUB

' return -1 for filelist character
FUNCTION FilelistIS (Command.Out$)
Flag = False
Command.Out2$ = Nul
IF MID$(Command.Out$, 2, 1) = ":" THEN
    Command.Out2$ = LEFT$(Command.Out$, 2)
    Command.Out$ = MID$(Command.Out$, 3)
END IF
IF LEFT$(Command.Out$, 1) = "@" THEN
    Flag = True
    Command.Out$ = MID$(Command.Out$, 2)
ELSE
    FOR Var = LEN(Command.Out$) TO 1 STEP -1
        IF MID$(Command.Out$, Var, 1) = "\" THEN
            Char$ = MID$(Command.Out$, Var + 1)
            IF LEFT$(Char$, 1) = "@" THEN
                Flag = True
                Command.Out$ = LEFT$(Command.Out$, Var) + MID$(Char$, 2)
            END IF
            EXIT FOR
        END IF
    NEXT
END IF
IF LEN(Command.Out2$) THEN
    Command.Out$ = Command.Out2$ + Command.Out$
END IF
FilelistIS = Flag
END FUNCTION

' routine to copy a filename
SUB Copyit_Files (Next.Dest.Net$, Source.Directory$, Source.Filename$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260
DIM ASCIIZwrite AS STRING * 260
DIM Handle AS INTEGER
DIM Number.Bytes AS INTEGER
DIM Sum.Bytes AS DOUBLE
DIM Sum.Bytes2 AS DOUBLE

' make filename
Temp.Net$ = RTRIM$(Source.Net)
IF Temp.Net$ = Nul THEN
    ASCIIZ = Source.Directory$ + Source.Filename$ + CHR$(0)
ELSE
    IF LEFT$(Directory.Search$, 1) = "\" THEN
        ASCIIZ = Temp.Net$ + Source.Directory$ + Source.Filename$ + CHR$(0)
    ELSE
        ASCIIZ = Temp.Net$ + "\" + Source.Directory$ + Source.Filename$ + CHR$(0)
    END IF
END IF

' read file attributes
Attribute = GetFileAttributes(ASCIIZ)

' check directory
IF (Attribute AND &H10) = &H10 THEN
    EXIT SUB
END IF

' store file attribute
Source.Attribute = Attribute

' check file attribute
IF Display.Attribute THEN

    ' check for readonly file
    IF Display.Readonly = True THEN
        IF (Attribute AND 1) <> 1 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Readonly = True THEN
        IF (Attribute AND 1) = 1 THEN
            EXIT SUB
        END IF
    END IF

    ' check for hidden file
    IF Display.Hidden = True THEN
        IF (Attribute AND 2) <> 2 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Hidden = True THEN
        IF (Attribute AND 2) = 2 THEN
            EXIT SUB
        END IF
    END IF

    ' check for system file
    IF Display.System = True THEN
        IF (Attribute AND 4) <> 4 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.System = True THEN
        IF (Attribute AND 4) = 4 THEN
            EXIT SUB
        END IF
    END IF

    ' check for archive file
    IF Display.Archive = True THEN
        IF (Attribute AND &H20) <> &H20 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Archive = True THEN
        IF (Attribute AND &H20) = &H20 THEN
            EXIT SUB
        END IF
    END IF

    ' check for compressed file
    IF Display.Compressed = True THEN
        IF (Attribute AND &H800) <> &H800 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Compressed = True THEN
        IF (Attribute AND &H800) = &H800 THEN
            EXIT SUB
        END IF
    END IF

    ' check for encrypted file
    IF Display.Encrypted = True THEN
        IF (Attribute AND &H4000) <> &H4000 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Encrypted = True THEN
        IF (Attribute AND &H4000) = &H4000 THEN
            EXIT SUB
        END IF
    END IF

    ' check for no attributes
    IF Display.Any = True THEN
        IF (Attribute AND 1) = 1 THEN
            EXIT SUB
        END IF
        IF (Attribute AND 2) = 2 THEN
            EXIT SUB
        END IF
        IF (Attribute AND 4) = 4 THEN
            EXIT SUB
        END IF
        IF (Attribute AND &H20) = &H20 THEN
            EXIT SUB
        END IF
    END IF
    IF No.Display.Any = True THEN
        IF (Attribute AND 1) = False THEN
            IF (Attribute AND 2) = False THEN
                IF (Attribute AND 4) = False THEN
                    IF (Attribute AND &H20) = False THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' reset destination file attribute
IF Set.Dest.Readonly THEN
    Attribute = Attribute OR 1
END IF
IF Set.Dest.Hidden THEN
    Attribute = Attribute OR 2
END IF
IF Set.Dest.System THEN
    Attribute = Attribute OR 4
END IF
IF Set.Dest.Archive THEN
    Attribute = Attribute OR &H20
END IF
IF Set.Dest.Any THEN
    Attribute = Attribute OR &H27
END IF
IF Clear.Dest.Readonly THEN
    Attribute = Attribute AND NOT 1
END IF
IF Clear.Dest.Hidden THEN
    Attribute = Attribute AND NOT 2
END IF
IF Clear.Dest.System THEN
    Attribute = Attribute AND NOT 4
END IF
IF Clear.Dest.Archive THEN
    Attribute = Attribute AND NOT &H20
END IF
IF Clear.Dest.Any THEN
    Attribute = Attribute AND NOT &H27
END IF

' reset source file attribute
IF Set.Source.Readonly THEN
    Source.Attribute = Source.Attribute OR 1
END IF
IF Set.Source.Hidden THEN
    Source.Attribute = Source.Attribute OR 2
END IF
IF Set.Source.System THEN
    Source.Attribute = Source.Attribute OR 4
END IF
IF Set.Source.Archive THEN
    Source.Attribute = Source.Attribute OR &H20
END IF
IF Set.Source.Any THEN
    Source.Attribute = Source.Attribute OR &H27
END IF
IF Clear.Source.Readonly THEN
    Source.Attribute = Source.Attribute AND NOT 1
END IF
IF Clear.Source.Hidden THEN
    Source.Attribute = Source.Attribute AND NOT 2
END IF
IF Clear.Source.System THEN
    Source.Attribute = Source.Attribute AND NOT 4
END IF
IF Clear.Source.Archive THEN
    Source.Attribute = Source.Attribute AND NOT &H20
END IF
IF Clear.Source.Any THEN
    Source.Attribute = Source.Attribute AND NOT &H27
END IF

' check file size range
CALL Copyit_CheckSizeRange(Varble)

' check return flag
IF Range.Flag THEN
    EXIT SUB
END IF

' store file date and time
x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
GOSUB Convert.Date
Synch.Work.Create.Date = Work.Date.Temp!

x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
GOSUB Convert.Date
Synch.Work.Access.Date = Work.Date.Temp!

x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
GOSUB Convert.Date
Synch.Work.Modify.Date = Work.Date.Temp!

x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
GOSUB Convert.Time
Synch.Work.Create.Time = Work.Time.Temp!

x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
GOSUB Convert.Time
Synch.Work.Access.Time = Work.Time.Temp!

x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
GOSUB Convert.Time
Synch.Work.Modify.Time = Work.Time.Temp!

' check date/time range
CALL Copyit_CheckDateRange(Varble)

' check return flag
IF Range.Flag THEN
    EXIT SUB
END IF

' construct extended date/time string
FileInfo$ = Nul
IF Double.Line >= 1 THEN
    ' make file info
    FileInfo$ = Copyit_MakeDate(1)
    FileInfo$ = FileInfo$ + "\" + Copyit_MakeDate(2)
    FileInfo$ = FileInfo$ + "\" + Copyit_MakeDate(3)
END IF

' store filename being copying
Copy.File$ = Source.Directory$ + Source.Filename$
IF Append.Files THEN
    Copy.Dest$ = Dest.Dir
ELSE
    IF Copy.Directory = False THEN
        Copy.Dest$ = Dest.Dir + Source.Filename$
    ELSE
        Copy.Dest$ = Copy.Target + Source.Filename$
    END IF
    IF Next.Dest.Net$ = Nul THEN
        ASCIIZwrite = Copy.Dest$ + CHR$(0)
    ELSE
        IF LEFT$(Copy.Dest$, 1) = "\" THEN
            ASCIIZwrite = Next.Dest.Net$ + Copy.Dest$ + CHR$(0)
        ELSE
            ASCIIZwrite = Next.Dest.Net$ + "\" + Copy.Dest$ + CHR$(0)
        END IF
    END IF
END IF

' check filename
Filename2$ = Source.Filename$
CALL Copyit_CheckFile(Filename2$, False)
IF Check.Flag = False THEN
    EXIT SUB
END IF

' make target directory only if files exist
IF Skip.Directory THEN
    CALL Copyit_CreateDirectory(Next.Dest.Net$)
END IF

' check if appending files
IF Append.Files THEN
    ' store output file info
    IF Next.Dest.Net$ = Nul THEN
        ASCIIZfile = Dest.Dir + CHR$(0)
    ELSE
        IF LEFT$(Dest.Dir, 1) = "\" THEN
            ASCIIZfile = Next.Dest.Net$ + Dest.Dir + CHR$(0)
        ELSE
            ASCIIZfile = Next.Dest.Net$ + "\" + Dest.Dir + CHR$(0)
        END IF
    END IF

    ' check if files similar
    CALL Copyit_ReadFileSynch

    ' check files
    CALL Copyit_CheckSynched(Varble)
    IF Synched.Flag THEN
        EXIT SUB
    END IF
ELSE
    ' store output file info
    ASCIIZfile = ASCIIZwrite

    ' check if files similar
    CALL Copyit_ReadFileSynch

    ' check files
    CALL Copyit_CheckSynched(Varble)
    IF Synched.Flag THEN
        EXIT SUB
    END IF
END IF

' verify/display filename
CALL Copyit_CheckFileExists(Next.Dest.Net$, Copy.File$, Copy.Dest$, Response)
IF Response = False AND Skip.All THEN
    Eat$ = Nul
ELSE
    CALL Copyit_FilenameCopying(Next.Dest.Net$, Copy.File$, Copy.Dest$)
END IF

' check copying flag
IF Filename.Exists THEN
    EXIT SUB
END IF

' get filename date/time
ASCIIZcopy = ASCIIZ
CALL Copyit_ExtendedFile(1)

' reset flag error
Handle = 0

' open file for input
Open.Filename$ = ASCIIZ
V = INSTR(Open.Filename$, CHR$(0))
IF V THEN
    Open.Filename$ = LEFT$(Open.Filename$, V - 1)
END IF
Handle = FREEFILE

DataError = False
Error.Flag = True
OPEN Open.Filename$ FOR BINARY SHARED AS #Handle
Error.Flag = False

' check error flag
IF DataError THEN
    Handle = 0
END IF

' check open file error
IF Handle = 0 THEN
    ' display any error
    IF Debug.Mode THEN
        IF DataError <> 70 THEN ' ignore share violation
            CALL Copyit_DisplayError("Error" + STR$(DataError) + " opening input filename.")
        END IF
    END IF
    EXIT SUB
END IF

' set seek position to start of file
SeekPosition# = 1#

' check byte position override
IF Byte.Switch THEN
    ' offset from specified byte position
    SeekPosition# = Byte.Override
END IF

' check byte position override
IF Resume.File THEN
    ' offset from dest. filesize position
    SeekPosition# = Dest.File.Size
END IF

' seek position in input file
SEEK Handle, SeekPosition#

' reset delete file flag
Copy.Error = False

' check if appending files
IF Append.Files THEN
    ' store output file info
    IF Next.Dest.Net$ = Nul THEN
        ASCIIZfile = Dest.Dir + CHR$(0)
    ELSE
        IF LEFT$(Dest.Dir, 1) = "\" THEN
            ASCIIZfile = Next.Dest.Net$ + Dest.Dir + CHR$(0)
        ELSE
            ASCIIZfile = Next.Dest.Net$ + "\" + Dest.Dir + CHR$(0)
        END IF
    END IF
    File.Attribute = Attribute

    ' check to overwrite
    IF Overwrite.Prompt = False THEN
        Copy.File$ = Directory.Search$ + RTRIM$(Filename.Search)
        Copy.Dest$ = Dest.Dir
        CALL Copyit_PromptOverwrite(Next.Dest.Net$, Copy.File$, Copy.Dest$, Response)
        IF Response = False THEN
            GOTO End.Copy.File1
        END IF
    END IF

    ' open output file
    CALL Copyit_OpenFile(OpenError)

    ' check error opening output file
    IF OpenError THEN
        EXIT SUB
    END IF
ELSE
    ' store output file info
    ASCIIZfile = ASCIIZwrite
    File.Attribute = Attribute

    ' check to overwrite
    IF Overwrite.Prompt = False THEN
        CALL Copyit_PromptOverwrite(Next.Dest.Net$, Copy.File$, Copy.Dest$, Response)
        IF Response = False THEN
            GOTO End.Copy.File1
        END IF
    END IF

    ' open output file
    CALL Copyit_OpenFile(OpenError)

    ' check error opening output file
    IF OpenError THEN
        GOTO End.Copy.File1
    END IF
END IF

' reset error flag
Error.Type = False
Percent.Flag = False
Current.Progress = False

' copy file
Count.Forward = True
Dot.Count = False
Dot.Total = False

' file input loop
DO
    ' reset and store error flag
    DataError = False
    Error.Flag = True

    ' read from file
    Number.Bytes = buflen
    position# = LOC(Handle)
    GET Handle, , fbuffer

    ' restore and check error flag
    Error.Flag = False
    IF DataError THEN
        Error.Type = DataError
        EXIT DO
    END IF

    ' store buffer and length of buffer
    BufferOut$ = Nul
    IF position# + buflen > LOF(Handle) THEN
        Number.Bytes = LOF(Handle) - position#
    END IF
    IF Number.Bytes > 0 THEN
        BufferOut$ = LEFT$(fbuffer, Number.Bytes)
    END IF

    ' show disk activity
    IF Continuous.Display = False AND Dot.Mode = False THEN
        Dot.Total = Dot.Total + 1 ' count buflen blocks
        IF Dot.Total >= 256 THEN ' could be increased
            Dot.Total = 0
            IF Count.Forward THEN
                PRINT ".";
                Dot.Count = Dot.Count + 1
                IF Dot.Count = 16 THEN
                    Count.Forward = False
                END IF
            ELSE
                PRINT CHR$(29); " "; CHR$(29);
                Dot.Count = Dot.Count - 1
                IF Dot.Count = False THEN
                    Count.Forward = True
                END IF
            END IF
        END IF
    END IF

    ' add total bytes counted
    Sum.Bytes2 = Sum.Bytes2 + Number.Bytes

    ' check total filesize to copy
    IF File.Size > 0# THEN

        ' check to display percent copied
        P$ = Nul
        Percent = (Sum.Bytes2 / File.Size) * 100!
        Percent = INT(Percent)
        IF Percent = 0 THEN
            IF Percent.Flag = 0 THEN
                P$ = "000%"
                IF Percent.Display THEN
                    PRINT P$;
                    Percent.Flag = True
                END IF
            END IF
        ELSE
            IF Percent <> Last.Percent THEN
                Last.Percent = Percent
                P$ = RIGHT$(STR$(Percent + 1000%), 3) + "%"
                IF Percent.Display THEN
                    IF Percent.Flag THEN
                        FOR Imbedded = 1 TO 4
                            PRINT CHR$(29); " "; CHR$(29);
                        NEXT
                    END IF
                    PRINT P$;
                    Percent.Flag = True
                END IF
            END IF
        END IF

        ' display percent copied on title
        IF LEN(P$) THEN
            IF P$ <> Title.Percent THEN
                Title.Percent = P$
                _TITLE Title.Header + " " + P$
            END IF
        END IF

        ' check to display progress bar
        IF Progress.Bar THEN
            Percent = (Sum.Bytes2 / File.Size) * 100#
            Percent = INT(Percent / 10)
            DO WHILE Percent > Current.Progress
                IF Current.Progress >= 10 THEN
                    EXIT DO
                END IF
                PRINT "#";
                Current.Progress = Current.Progress + 1
            LOOP
        END IF
    END IF

    ' check binary copy
    IF Copy.Ascii = False THEN
        Ascii.Exit = False
    ELSE
        IF INSTR(BufferOut$, CHR$(26)) THEN ' eof char
            BufferOut$ = LEFT$(BufferOut$, INSTR(BufferOut$, CHR$(26)))
            Number.Bytes = LEN(BufferOut$)
            Ascii.Exit = True
        END IF
    END IF

    ' check number of bytes to copy
    IF Number.Bytes > False THEN

        ' check ascii filters
        FOR Count = 1 TO Ascii.Filters
            Ascii.From = Convert.Ascii(Count, 1)
            Ascii.To = Convert.Ascii(Count, 2)
            Next.Imbedded = 1
            DO
                Imbedded = INSTR(Next.Imbedded, BufferOut$, CHR$(Ascii.From))
                IF Imbedded THEN
                    Next.Imbedded = Imbedded + 1
                    MID$(BufferOut$, Imbedded, 1) = CHR$(Ascii.To)
                ELSE
                    EXIT DO
                END IF
                IF Next.Imbedded >= Number.Bytes THEN
                    EXIT DO
                END IF
            LOOP
        NEXT
    END IF

    ' check number of bytes to copy
    IF Number.Bytes > False THEN

        ' check ascii strips
        FOR Count = 1 TO Ascii.Strips
            ' check entire buffer is stripped
            IF LEFT$(BufferOut$, Number.Bytes) = STRING$(Number.Bytes, CHR$(Strip.Ascii(Count))) THEN
                Number.Bytes = 0
                EXIT FOR
            END IF

            ' check strip values
            DO
                Imbedded = INSTR(BufferOut$, CHR$(Strip.Ascii(Count)))
                IF Imbedded > False AND Imbedded <= Number.Bytes THEN
                    BufferOut$ = LEFT$(BufferOut$, Imbedded - 1) + MID$(BufferOut$, Imbedded + 1)
                    Number.Bytes = Number.Bytes - 1
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
    END IF

    ' check number of bytes to copy
    IF Number.Bytes > False THEN
        ' reset and store error flag
        DataError = False
        Error.Flag = True

        ' append to file
        PUT Append.Handle, , BufferOut$

        ' restore and check error flag
        Error.Flag = False
        IF DataError THEN
            Error.Type = DataError
            EXIT DO
        END IF

        ' increment bytes copied
        Sum.Bytes = Sum.Bytes + Number.Bytes

        ' check binary copy
        IF Ascii.Exit THEN
            EXIT DO
        END IF
    END IF

    ' check for more input
    IF EOF(Handle) THEN
        EXIT DO
    END IF
LOOP

' clear display progress
CALL Copyit_ClearDisplay(-1)

' update counter
Files.Counted = Files.Counted + 1#

' check copying file error
IF DataError THEN
    IF Wide.List THEN
        Wide.List = False
        PRINT
    END IF
    Copy.Error = True
    IF Display.Errors = False THEN
        COLOR Red, Black
        IF Error.Type = 61 THEN ' disk full
            Disk.Full = True
            Quit.Searching = True
        END IF
        CALL DisplayCriticalError(Error.Type)
    END IF
END IF

' close output file
CALL Copyit_CloseFile(0)

' pcocess end of input file parsing
End.Copy.File1:
' close input file
IF Handle THEN
    CLOSE #Handle
END IF

' display any errors
IF Debug.Mode THEN
    IF CloseError > False THEN
        CALL Copyit_DisplayError("Error closing input filename.")
    END IF
END IF

' check disk full flag
IF Disk.Full THEN
    ' delete destination file
    IF Append.Dest = False THEN
        CALL Copyit_DeleteDestFile(False, Next.Dest.Net$, Source.Filename$)
    END IF
    EXIT SUB
END IF

' check copy error flag
IF Copy.Error THEN
    ' delete destination file
    IF Append.Dest = False THEN
        CALL Copyit_DeleteDestFile(True, Next.Dest.Net$, Source.Filename$)
    END IF
    EXIT SUB
END IF

' check to change source attribute
Reset.Source = False
IF RTRIM$(Source.Net) <> Nul THEN
    Reset.Source = True
END IF
IF Set.Source.Attribute OR Reset.Source THEN
    ' reset source attribute
    x = SetFileAttributes&(ASCIIZ, Source.Attribute)
END IF

' store total bytes copied
IF OpenError = 0 THEN
    Total.Bytes2 = Total.Bytes2 + File.Size
END IF

' store actual bytes copied
IF OpenError = 0 THEN
    Total.Bytes = Total.Bytes + Sum.Bytes
END IF

' store virtual bytes copied
IF Resume.File THEN
    Var# = File.Size - Dest.File.Size
ELSE
    IF Byte.Switch THEN
        Var# = File.Size - Byte.Override
    ELSE
        Var# = File.Size
    END IF
END IF
IF Var# < 0# THEN
    Var# = 0#
END IF
IF OpenError = 0 THEN
    Bytes.Copied = Bytes.Copied + Var#
END IF

' check display type
Var = False
IF Double.Line >= 1 THEN
    ' display file date/time
    COLOR Green, Black
    PRINT FileInfo$ + " ";
    Var = True
END IF

' check display type
IF Double.Line >= 2 THEN
    ' display file size
    COLOR Red, Black
    PRINT FormatString$(File.Size);
    Var = True
END IF

' check display type
IF Double.Line >= 3 THEN
    ' check for archive file
    COLOR White, Black
    IF (Attribute AND &H20) = &H20 THEN
        PRINT " Archive";
    END IF

    ' check for read-only file
    IF (Attribute AND 1) = 1 THEN
        PRINT " Read-only";
    END IF

    ' check for hidden file
    IF (Attribute AND 2) = 2 THEN
        PRINT " Hidden";
    END IF

    ' check for system file
    IF (Attribute AND 4) = 4 THEN
        PRINT " System";
    END IF

    ' check for compressed file
    IF (Attribute AND &H800) = &H800 THEN
        PRINT " Compressed";
    END IF

    ' check for encrypted file
    IF (Attribute AND &H4000) = &H4000 THEN
        PRINT " Encrypted";
    END IF
    Var = True
END IF
IF Var THEN
    Wide.List = False
    PRINT
END IF

' check to delete source file
IF Delete.Copied THEN
    CALL Copyit_DeleteSourceFile(Source.Directory$, Source.Filename$)
END IF
EXIT SUB

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
Work.Date.Temp! = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
Work.Time.Temp! = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' display error trap message.
SUB DisplayCriticalError (Var)
SELECT CASE Var
    CASE 6
        PRINT #StderrHandle, "Overflow";
    CASE 9
        PRINT #StderrHandle, "Subscript out of range";
    CASE 14
        PRINT #StderrHandle, "Out of string space";
    CASE 24
        PRINT #StderrHandle, "Device timeout";
    CASE 25
        PRINT #StderrHandle, "Device fault";
    CASE 27
        PRINT #StderrHandle, "Out of paper";
    CASE 52
        PRINT #StderrHandle, "Bad file name or number";
    CASE 53
        PRINT #StderrHandle, "File not found";
    CASE 54
        PRINT #StderrHandle, "Bad file mode";
    CASE 55
        PRINT #StderrHandle, "File already open";
    CASE 57
        PRINT #StderrHandle, "Device I/O error";
    CASE 58
        PRINT #StderrHandle, "File already exists";
    CASE 59
        PRINT #StderrHandle, "Bad record length";
    CASE 61
        PRINT #StderrHandle, "Disk full";
    CASE 62
        PRINT #StderrHandle, "Input past eof";
    CASE 63
        PRINT #StderrHandle, "Bad record length";
    CASE 64
        PRINT #StderrHandle, "Bad file name";
    CASE 67
        PRINT #StderrHandle, "Not enough file handles";
    CASE 68
        PRINT #StderrHandle, "Device unavailable";
    CASE 70
        PRINT #StderrHandle, "Permission denied";
    CASE 71
        PRINT #StderrHandle, "Disk not ready";
    CASE 72
        PRINT #StderrHandle, "Disk-media error";
    CASE 75
        PRINT #StderrHandle, "Path/File access error";
    CASE 76
        PRINT #StderrHandle, "Pathname not found";
    CASE 81
        PRINT #StderrHandle, "Invalid name";
    CASE 90
        PRINT #StderrHandle, "Too many processes";
    CASE ELSE
        PRINT #StderrHandle, "Untrapped error" + STR$(Var);
END SELECT
PRINT #StderrHandle, " copying files."
END SUB

' routine to access directories
SUB Copyit_Directories (Next.Dest.Net$, Directory.Search$, Directory.Target$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ.Sub AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' declare wide functions
DIM ASCIIZ.FileW AS STRING * MAX_PATH2 ' a null terminated filename
DIM finddataw AS WIN32_FIND_DATAW ' the windows filename structure
DIM Wfile.HandleW AS _UNSIGNED _OFFSET ' windows file handle for FindFile

' make directory filename
Temp.Net$ = RTRIM$(Source.Net)
IF Temp.Net$ = Nul THEN
    ASCIIZ.Sub = Directory.Search$ + "*.*" + CHR$(0)
    Source.Search$ = Directory.Search$
ELSE
    IF LEFT$(Directory.Search$, 1) = "\" THEN
        ASCIIZ.Sub = Temp.Net$ + Directory.Search$ + "*.*" + CHR$(0)
        Source.Search$ = Temp.Net$ + Directory.Search$
    ELSE
        ASCIIZ.Sub = Temp.Net$ + "\" + Directory.Search$ + "*.*" + CHR$(0)
        Source.Search$ = Temp.Net$ + "\" + Directory.Search$
    END IF
END IF
IF RIGHT$(Source.Search$, 1) = "\" THEN
    Source.Search$ = LEFT$(Source.Search$, LEN(Source.Search$) - 1)
END IF

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ.Sub), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' store source copied directory filename
    Copy.Target = Directory.Target$

    ' check to copy directory structure
    Directory.Exists = False
    IF Copy.Directory THEN
        Dir.Copy.Name = Directory.Search$
        Dir.Copy.Target = Directory.Target$
        IF Skip.Directory = 0 THEN
            CALL Copyit_DirectoryStructure(Source.Search$, Next.Dest.Net$)
        END IF
    END IF

    ' check directory exists
    IF Directory.Exists = False THEN

        ' search directory names
        CALL Copyit_Filenames(Directory.Target$, Next.Dest.Net$, Directory.Search$)

        ' check to recurse directories
        IF Recurse.Directories THEN

            ' recursive subdirectories
            DO
                ' check to quit
                IF Quit.Searching THEN
                    EXIT DO
                END IF

                ' check directory attribute
                Attribute = finddata.dwFileAttributes
                IF (Attribute AND &H10) = &H10 THEN

                    ' store directory name
                    Next.Search$ = finddata.cFileName
                    Imbedded = INSTR(Next.Search$, CHR$(0))
                    IF Imbedded THEN
                        Next.Search$ = LEFT$(Next.Search$, Imbedded - 1)
                    END IF

                    ' check unicode
                    DirIsUnicode = 0
                    IF INSTR(Next.Search$, "?") THEN
                        Next.Search$ = finddata.cAlternateFileName
                        Imbedded = INSTR(Next.Search$, CHR$(0))
                        IF Imbedded THEN
                            Next.Search$ = LEFT$(Next.Search$, Imbedded - 1)
                        END IF
                        Unicode$ = Directory.Search$ + Next.Search$
                        WideZ$ = AsciiToWide$(Unicode$) + CHR$(0)
                        Wfile.HandleW = FindFirstFileW(WideZ$, _OFFSET(finddataw))
                        IF Wfile.HandleW <> INVALID_HANDLE_VALUE THEN
                            DirIsUnicode = -1
                            DirUnicodeShort = WideZ$
                            DirUnicodeLong = finddataw.cFileName
                            x = FindClose(Wfile.HandleW)
                        END IF
                    END IF

                    ' check directory name
                    IF Next.Search$ <> "." AND Next.Search$ <> ".." THEN
                        ' check recursion levels
                        Recursion = True
                        IF Nested.Recurse > False THEN
                            Nested.Levels = Nested.Levels + 1
                            IF Nested.Levels >= Nested.Recurse THEN
                                Recursion = False
                            END IF
                        END IF

                        ' recursively search subdirectories
                        IF Recursion THEN
                            Temp1$ = Directory.Search$ + Next.Search$ + "\"
                            Temp2$ = Directory.Target$ + Next.Search$ + "\"
                            CALL Copyit_Directories(Next.Dest.Net$, Temp1$, Temp2$)
                        END IF
                        IF Nested.Recurse > False THEN
                            Nested.Levels = Nested.Levels - 1
                        END IF
                    END IF
                END IF
            LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
        END IF
    END IF
    x = FindClose(Wfile.Handle)
END IF

' check to delete subdirectory
IF Recurse.Error = False THEN
    IF Delete.Directory THEN
        CALL Copyit_DeleteSubdirectory(Directory.Search$)
    END IF
END IF
END SUB

' routine to access filenames in directory
SUB Copyit_Filenames (Source.Search$, Next.Dest.Net$, Directory.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM ASCIIZ.File AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM Recurse.Temp AS INTEGER

' declare wide functions
DIM ASCIIZ.FileW AS STRING * MAX_PATH2 ' a null terminated filename
DIM finddataw AS WIN32_FIND_DATAW ' the windows filename structure
DIM Wfile.HandleW AS _UNSIGNED _OFFSET ' windows file handle for FindFile

' make filename
Temp.Net$ = RTRIM$(Source.Net)
Var$ = Filename.Search
Var$ = RTRIM$(Var$)
IF Temp.Net$ = Nul THEN
    ASCIIZ.File = Directory.Search$ + Var$ + CHR$(0)
ELSE
    IF LEFT$(Directory.Search$, 1) = "\" THEN
        ASCIIZ.File = Temp.Net$ + Directory.Search$ + Var$ + CHR$(0)
    ELSE
        ASCIIZ.File = Temp.Net$ + "\" + Directory.Search$ + Var$ + CHR$(0)
    END IF
END IF

' find first long filename
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ.File), _OFFSET(finddata))
 
' check findirst error
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    DO
        ' store filename
        Filename$ = finddata.cFileName
        Imbedded = INSTR(Filename$, CHR$(0))
        IF Imbedded THEN
            Filename$ = LEFT$(Filename$, Imbedded - 1)
        END IF

        ' check unicode
        FileIsUnicode = 0
        IF INSTR(Filename$, "?") THEN
            Filename$ = finddata.cAlternateFileName
            Imbedded = INSTR(Filename$, CHR$(0))
            IF Imbedded THEN
                Filename$ = LEFT$(Filename$, Imbedded - 1)
            END IF
            WideZ$ = AsciiToWide$(Filename$) + CHR$(0)
            Wfile.HandleW = FindFirstFileW(WideZ$, _OFFSET(finddataw))
            IF Wfile.HandleW <> INVALID_HANDLE_VALUE THEN
                FileIsUnicode = -1
                UnicodeShort = WideZ$
                UnicodeLong = finddataw.cFileName
                x = FindClose(Wfile.HandleW)
            END IF
        END IF

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' store file info
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' store filename structure buffer
            finddatatemp = finddata

            ' store target directory
            Directory.Path = Source.Search$

            ' call subroutine to copy files
            CALL Copyit_Files(Next.Dest.Net$, Directory.Search$, Filename$)
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' routine to open output file
SUB Copyit_OpenFile (OpenError)
DIM hfind AS _OFFSET

' reset file open error flags
Disk.Full = False
OpenError = False

' declare filename to open
Open.Filename$ = ASCIIZfile
V = INSTR(Open.Filename$, CHR$(0))
IF V THEN
    Open.Filename$ = LEFT$(Open.Filename$, V - 1)
END IF
IF LEN(Open.Filename$) = 0 THEN
    IF Debug.Mode THEN
        Var$ = "Error" + STR$(DataError) + " opening output filename."
        CALL Copyit_DisplayError(Var$)
    END IF
    OpenError = -1
    EXIT SUB
END IF

' change filename attribute
AttrX& = GetFileAttributes(ASCIIZfile)
AttrX& = AttrX& AND NOT &H1 ' reset read-only bit
x = SetFileAttributes&(ASCIIZfile, AttrX&)

' overwrite destination file
IF Append.Dest = 0 THEN
    hfind = CreateFile(ASCIIZfile, &H180, &H3, 0, 3, 0, 0)
    IF hfind <> INVALID_HANDLE_VALUE THEN
        x = CloseHandle(hfind)

        ' delete destination filename
        x = DeleteFileA(ASCIIZfile)

        ' display any errors
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") deleting output filename."
                CALL Copyit_DisplayError(Var$)
            END IF
            OpenError = -1
            EXIT SUB
        END IF
    END IF
END IF

' open destination file
DataError = False
Error.Flag = True
Append.Handle = FREEFILE
OPEN Open.Filename$ FOR BINARY AS #Append.Handle
Error.Flag = False

' check error flag
IF DataError THEN
    IF Debug.Mode THEN
        Var$ = "Error" + STR$(DataError) + " opening output filename."
        CALL Copyit_DisplayError(Var$)
    END IF
    OpenError = -1
    EXIT SUB
END IF

' reset and store error flag
DataError = False
Error.Flag = True

' seek file position to eof
SeekPosition2# = LOF(Append.Handle) + 1#
SEEK Append.Handle, SeekPosition2#

' reset and check error flag
Error.Flag = False
IF DataError THEN
    IF Debug.Mode THEN
        Var$ = "Error" + STR$(DataError) + " opening output filename."
        CALL Copyit_DisplayError(Var$)
    END IF
    OpenError = -1
END IF
END SUB

' routine to close output file
SUB Copyit_CloseFile (CloseType%)
' check file handle
IF Append.Handle = False THEN
    EXIT SUB
END IF

' close file
CLOSE #Append.Handle

' reset file handle
Append.Handle = False

' write dest. filename attributes
IF CloseType% = 0 THEN
    x = SetFileAttributes&(ASCIIZfile, File.Attribute)
END IF

' set filename date/time
CALL Copyit_ExtendedFile(2)

' restore unicode long filename
CALL RenameUnicode
END SUB

SUB RenameUnicode
' check if file is unicode
IF FileIsUnicode THEN
    UnicodeShort = AsciiToWide$(ASCIIZfile) + CHR$(0)

    ' parse unicode filename
    Var$ = ASCIIZfile
    IF INSTR(Var$, CHR$(0)) THEN
        Var$ = LEFT$(Var$, INSTR(Var$, CHR$(0)) - 1)
    END IF
    Var = 0
    FOR X = LEN(Var$) TO 1 STEP -1
        IF MID$(Var$, X, 1) = "\" THEN
            Var$ = AsciiToWide$(LEFT$(Var$, X)) + UnicodeLong
            Var = -1
            EXIT FOR
        END IF
    NEXT
    IF Var = 0 THEN
        Var$ = UnicodeLong
    END IF

    ' rename the filename
    r = MoveFileW(UnicodeShort, Var$ + CHR$(0))
    IF r = 0 THEN
        IF Debug.ModeX THEN
            PRINT "Unicode file rename error: "; HEX$(GetLastError)
        END IF
    END IF
END IF
END SUB

' routine overrides destination file date/time.
'   1 = get filename date/time
'   2 = set filename date/time
SUB Copyit_ExtendedFile (T)
DIM hfind AS _OFFSET

' get source filename date/time.
IF T = 1 THEN
    hfind = CreateFile(ASCIIZcopy, &H180, &H3, 0, 3, 0, 0)
    IF hfind <> INVALID_HANDLE_VALUE THEN
        x = GetFileTime&(hfind, SourceCreateDate, SourceAccessDate, SourceModifyDate)
        ' display any errors.
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") reading filename date/time."
                CALL Copyit_DisplayError(Var$)
            END IF
        END IF
        x = CloseHandle(hfind)
    END IF
    EXIT SUB
END IF

' override filename date/time.
IF Override.Create.Date OR Override.Access.Date OR Override.Modify.Date THEN
    ' store default date/time.
    CreateDate = SourceCreateDate
    AccessDate = SourceAccessDate
    ModifyDate = SourceModifyDate

    ' store override date/time.
    IF Override.Date AND Override.Time THEN
        GOSUB GetSysDate
        IF Override.Create.Date THEN
            x = SystemTimeToFileTime&(SysTime, CreateDate)
        END IF
        IF Override.Access.Date THEN
            x = SystemTimeToFileTime&(SysTime, AccessDate)
        END IF
        IF Override.Modify.Date THEN
            x = SystemTimeToFileTime&(SysTime, ModifyDate)
        END IF
    END IF

    ' set the destination filename date/time.
    hfind = CreateFile(ASCIIZfile, &H180, &H3, 0, 3, 0, 0)
    IF hfind <> INVALID_HANDLE_VALUE THEN
        x = SetFileTime&(hfind, CreateDate, AccessDate, ModifyDate)
        ' display any errors.
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename date/time."
                CALL Copyit_DisplayError(Var$)
            END IF
        END IF
        x = CloseHandle(hfind)
    END IF
END IF
EXIT SUB

GetSysDate:
' store override variables
SysTime.wMonth = Override.Date.Month
SysTime.wDay = Override.Date.Day
SysTime.wYear = Override.Date.Year
SysTime.wHour = Override.Time.Hour
SysTime.wMinute = Override.Time.Minute
SysTime.wSecond = Override.Time.Second
RETURN
END SUB

' routine to read file info
SUB Copyit_ReadFileSynch
' declare subroutine variables
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZfile), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' store file size
    Dest.File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

    ' store file date and time
    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
    GOSUB Convert.Date
    Synch.File.Create.Date = Work.Date.Temp!

    x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
    GOSUB Convert.Date
    Synch.File.Access.Date = Work.Date.Temp!

    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    GOSUB Convert.Date
    Synch.File.Modify.Date = Work.Date.Temp!

    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
    GOSUB Convert.Time
    Synch.File.Create.Time = Work.Time.Temp!

    x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
    GOSUB Convert.Time
    Synch.File.Access.Time = Work.Time.Temp!

    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    GOSUB Convert.Time
    Synch.File.Modify.Time = Work.Time.Temp!

    x = FindClose(Wfile.Handle)
END IF
EXIT SUB

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
Work.Date.Temp! = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
Work.Time.Temp! = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' routine to preprocess filenames in command line
SUB Copyit_ProcessCommand (Quit)
' declare work variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM hfind AS _OFFSET

' reset quit flag
Quit = False

' store current drive
IF MID$(Command.Work, 2, 1) = ":" THEN
    Drive.Search = LEFT$(UCASE$(Command.Work), 1)
    Command.Work = MID$(Command.Work, 3)
ELSE
    Drive.Search = UCASE$(Current.Drive)
END IF

' read current directory
Temp.Search$ = ""
Drive.Temp = ASC(Drive.Search) - 64
IF MEDIAEXISTS(Drive.Temp) THEN
    IF _DIREXISTS(Drive.Search + ":") THEN
        x = SetCurrentDirectoryA(Drive.Search + ":" + CHR$(0))
        Temp.Search$ = RTRIM$(MID$(_CWD$, 4)) ' C:\TEMP
    END IF
END IF

' store source netpath
Source.Net = Nul
IF LEFT$(_CWD$, 2) = "\\" THEN
    Source.Net = _CWD$
END IF

' store current directory
Directory.Search$ = Nul
IF LEN(Command.Work) THEN
    Imbedded = INSTR(Command.Work, "\")
    Imbedded.Char = Imbedded
    WHILE Imbedded
        Imbedded.Char = Imbedded
        Imbedded = INSTR(Imbedded + 1, Command.Work, "\")
    WEND
    IF Imbedded.Char THEN
        Directory.Search$ = LEFT$(Command.Work, Imbedded.Char)
        Command.Work = MID$(Command.Work, Imbedded.Char + 1)
    ELSE
        IF RTRIM$(Source.Net) <> Nul THEN
            Directory.Search$ = "\"
        END IF
    END IF
END IF
IF Directory.Search$ = Nul THEN
    IF Temp.Search$ = "\" THEN
        Directory.Search$ = "\"
    ELSE
        Directory.Search$ = "\" + Temp.Search$
    END IF
END IF
IF RIGHT$(Directory.Search$, 1) <> "\" THEN
    Directory.Search$ = Directory.Search$ + "\"
END IF
Directory.Search$ = RTRIM$(Directory.Search$)

' read filename spec
Filename.Search = Command.Work
IF RTRIM$(Filename.Search) = Nul THEN
    Filename.Search = "*.*"
END IF
Command.Work = Nul
   
' check search filename
Target.Dir$ = Nul
Imbedded1 = INSTR(Filename.Search, "?")
Imbedded2 = INSTR(Filename.Search, "*")
IF Imbedded1 = False AND Imbedded2 = False THEN
    ' check search directory
    ASCIIZ = Directory.Search$ + RTRIM$(Filename.Search) + CHR$(0)

    ' find first directory
    Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
    IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

        ' check source is directory or filename
        Attribute = finddata.dwFileAttributes
        IF (Attribute AND &H10) = &H10 THEN
            Directory.Search$ = Directory.Search$ + RTRIM$(Filename.Search) + "\"
            Target.Dir$ = RTRIM$(Filename.Search)
            Filename.Search = "*.*"
        END IF
        x = FindClose(Wfile.Handle)
    END IF
END IF

' display search filename
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    IF RTRIM$(Source.Net) = Nul THEN
        PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + RTRIM$(Filename.Search)
    ELSE
        IF LEFT$(Directory.Search$, 1) = "\" THEN
            PRINT "Searching: " + RTRIM$(Source.Net) + Directory.Search$ + RTRIM$(Filename.Search)
        ELSE
            PRINT "Searching: " + RTRIM$(Source.Net) + "\" + Directory.Search$ + RTRIM$(Filename.Search)
        END IF
    END IF
END IF

' process destination netpaths
FOR Count0 = 1 TO Number.Dest.Nets

    ' store netpath
    Next.Dest.Net$ = Destinate.Netpaths(Count0)
    Next.Dest.Net$ = RTRIM$(Next.Dest.Net$)

    ' process destination directories
    FOR Count1 = 1 TO Number.Dest.Dirs
        ' store directory
        Next.Dest.Dir$ = Destinate.Directory(Count1)
        Next.Dest.Dir$ = RTRIM$(Next.Dest.Dir$)

        ' check destination drive
        IF MID$(Next.Dest.Dir$, 2, 1) = ":" THEN
            Drive.Number = ASC(UCASE$(MID$(Next.Dest.Dir$, 1, 1)))
            Next.Dest.Dir$ = MID$(Next.Dest.Dir$, 3)
        ELSE
            Drive.Number = ASC(UCASE$(Current.Drive))
        END IF

        ' store destination directory
        MediaError = 0
        IF Next.Dest.Net$ = Nul THEN
            GOSUB Store.Dest.Dir
        END IF

        ' recheck destination directory
        IF LEFT$(Next.Dest.Dir$, 1) <> "\" THEN
            Next.Dest.Dir$ = Default.Dest + Next.Dest.Dir$
        END IF
        IF RIGHT$(Next.Dest.Dir$, 1) <> "\" THEN
            Next.Dest.Dir$ = Next.Dest.Dir$ + "\"
        END IF

        ' store destination drive/netpath
        IF Next.Dest.Net$ = Nul THEN
            IF LEFT$(Next.Dest.Dir$, 2) <> "\\" THEN
                Next.Dest.Dir$ = CHR$(Drive.Number) + ":" + Next.Dest.Dir$
            END IF
        END IF

        ' process destination filenames
        FOR Count2 = 1 TO Number.Dest.Files

            ' store filename
            Dest.File = Destinate.Filename(Count2)
            Dest.File = RTRIM$(Dest.File)

            ' check filelist filename
            IF LEFT$(Dest.File, 1) = "@" AND Enable.Redirect THEN
                Dest.Out$ = MID$(Dest.File$, 2)

                ' open filename list
                x = 0
                ASCIIZ2 = Dest.Out$ + CHR$(0)
                hfind = CreateFile(ASCIIZ2, &H180, &H3, 0, 3, 0, 0)
                IF hfind <> INVALID_HANDLE_VALUE THEN
                    x = CloseHandle(hfind)
                END IF
                IF x = 0 THEN
                    ' display filename error
                    CALL Copyit_DisplayError("Error" + STR$(GetLastError&) + " opening listfile.")
                ELSE

                    V = FREEFILE
                    DataError = False
                    Error.Flag = True
                    OPEN Dest.Out$ FOR INPUT SHARED AS #V
                    Error.Flag = False

                    ' check error flag
                    IF DataError THEN
                        CALL Copyit_DisplayError("Error" + STR$(DataError) + " opening filelist.")
                    ELSE
                        ' process input filenames
                        DO UNTIL EOF(V)
                            LINE INPUT #V, Next.Filename$
                            Next.Filename$ = RTRIM$(Next.Filename$)
                            IF LEN(Next.Filename$) THEN
                                ' process search filename
                                Dest.File$ = Next.Filename$
                                GOSUB ProcessDestinate
                            END IF
                            ' check to quit
                            IF Quit.Searching THEN
                                EXIT DO
                            END IF
                        LOOP
                        CLOSE #V
                    END IF
                END IF
            ELSE
                GOSUB ProcessDestinate
            END IF

            ' check quit flag
            IF Quit.Searching THEN
                Quit = True
                EXIT FOR
            END IF
        NEXT

        ' check quit flag
        IF Quit.Searching THEN
            EXIT FOR
        END IF
    NEXT

    ' check netpaths
    IF Next.Dest.Net$ = Nul THEN
        EXIT FOR
    END IF
NEXT
EXIT SUB

ProcessDestinate:
' check destination drive exists.
IF MediaError THEN
    RETURN
END IF

' reset work variable
Append.Files = False

' check destination filename
Dest.Dir = Next.Dest.Dir$
IF Dest.File <> Nul THEN
    Dest.Dir = Dest.Dir + Dest.File
    Append.Files = True
END IF

' store target directory for directory structure copy
Directory.Target$ = Dest.Dir
IF Zero.Nest = False THEN
    IF LEN(Target.Dir$) THEN
        Directory.Target$ = Directory.Target$ + Target.Dir$ + "\"
    END IF
END IF

' construct target directory
IF Copy.Directory THEN
    Directory.Path = Directory.Target$
    CALL Copyit_CreateDirectory(Next.Dest.Net$)
END IF

' reset recurse flag
Recurse.Error = False

' recursively copy files
CALL Copyit_Directories(Next.Dest.Net$, Directory.Search$, Directory.Target$)

' close output file
IF Append.Files THEN
    CALL Copyit_CloseFile(-1)
END IF
RETURN

' get default directory of target drive
Store.Dest.Dir:
Default.Dest = ""
IF LEFT$(_CWD$, 2) = "\\" THEN
    Default.Dest = _CWD$
    IF RIGHT$(Default.Dest, 1) <> "\" THEN
        Default.Dest = Default.Dest + "\"
    END IF
    RETURN
END IF
IF MEDIAEXISTS(Drive.Number - 64) = 0 THEN
    IF Debug.Mode THEN
        CALL DisplayError("Error accessing target drive.")
    END IF
    MediaError = -1
    RETURN
END IF
IF _DIREXISTS(CHR$(Drive.Number) + ":") = 0 THEN
    IF Debug.Mode THEN
        CALL DisplayError("Error accessing target directory.")
    END IF
    MediaError = -1
    RETURN
END IF

' store directory
x = SetCurrentDirectoryA(CHR$(Drive.Number) + ":" + CHR$(0))
Default.Dest = RTRIM$(MID$(_CWD$, 4)) ' C:\TEMP

' restore directory
Drive.Temp = ASC(Drive.Search) - 64
IF MEDIAEXISTS(Drive.Temp) THEN
    x = SetCurrentDirectoryA(Drive.Search + ":" + CHR$(0))
END IF

' parse directory
IF Default.Dest = Nul THEN
    Default.Dest = "\"
END IF
IF Default.Dest <> "\" THEN
    Default.Dest = "\" + Default.Dest
END IF
IF RIGHT$(Default.Dest, 1) <> "\" THEN
    Default.Dest = Default.Dest + "\"
END IF
RETURN
END SUB

' routine to see if destination file exists
SUB Copyit_CheckFileExists (Dest.Net$, Source.File$, Destination.File$, Response)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM PromptASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' reset continue searching flag
Response = True

' make filename
IF Dest.Net$ = Nul THEN
    PromptASCIIZ = Destination.File$ + CHR$(0)
ELSE
    IF LEFT$(Destination.File$, 1) = "\" THEN
        PromptASCIIZ = Dest.Net$ + Destination.File$ + CHR$(0)
    ELSE
        PromptASCIIZ = Dest.Net$ + "\" + Destination.File$ + CHR$(0)
    END IF
END IF

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(PromptASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    Response = False
    x = FindClose(Wfile.Handle)
END IF
END SUB

' routine to prompt before overwriting files
SUB Copyit_PromptOverwrite (Dest.Net$, Source.File$, Destination.File$, Response)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM PromptASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' reset continue searching flag
Response = True

' make filename
IF Dest.Net$ = Nul THEN
    PromptASCIIZ = Destination.File$ + CHR$(0)
ELSE
    IF LEFT$(Destination.File$, 1) = "\" THEN
        PromptASCIIZ = Dest.Net$ + Destination.File$ + CHR$(0)
    ELSE
        PromptASCIIZ = Dest.Net$ + "\" + Destination.File$ + CHR$(0)
    END IF
END IF

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(PromptASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' make drive
    Start.Dir = Dest.Dir
    IF MID$(Start.Dir, 2, 1) = ":" THEN
        Drive.Number = ASC(UCASE$(LEFT$(Start.Dir, 1)))
        Start.Dir = MID$(Start.Dir, 3)
    ELSE
        Drive.Number = ASC(UCASE$(Drive.Search))
    END IF

    ' make filename
    To.File$ = Destination.File$
    IF MID$(To.File$, 2, 1) = ":" THEN
        To.File$ = MID$(To.File$, 3)
    END IF

    ' make filename
    From.File$ = Source.File$
    IF MID$(From.File$, 2, 1) = ":" THEN
        From.File$ = MID$(From.File$, 3)
    END IF

    ' check overwrite flag
    IF Dont.Overwrite THEN
        ' display file
        Response = False

        ' check display flags
        IF Short.Display OR Short.Display4 THEN
            Eat$ = Nul
        ELSE
            ' make filename
            Outpt$ = RTRIM$(From.File$)
            Temp.Net$ = RTRIM$(Source.Net)
            IF Temp.Net$ = Nul THEN
                Outpt$ = CHR$(Drive.Number) + ":" + Outpt$
            ELSE
                IF LEFT$(Outpt$, 1) = "\" THEN
                    Outpt$ = Temp.Net$ + Outpt$
                ELSE
                    Outpt$ = Temp.Net$ + "\" + Outpt$
                END IF
            END IF

            IF Display.Lowercase THEN
                Outpt$ = LCASE$(Outpt$)
            END IF

            ' check continuous display
            IF Continuous.Display = False THEN
                Outpt$ = "Not copying file " + Outpt$
            END IF

            ' display filename
            IF Wide.List THEN
                Wide.List = False
                PRINT
            END IF
            COLOR White, Black
            PRINT Outpt$
        END IF
    ELSE
        ' check continuous copying
        IF Continue.Searching = False THEN

            ' check to skip all copied files
            IF Skip.All THEN
                Response = False
            ELSE

                ' make filename
                Prompt2$ = RTRIM$(To.File$)
                IF Dest.Net$ = Nul THEN
                    Prompt2$ = CHR$(Drive.Number) + ":" + Prompt2$
                ELSE
                    IF LEFT$(Prompt2$, 1) = "\" THEN
                        Prompt2$ = Dest.Net$ + Prompt2$
                    ELSE
                        Prompt2$ = Dest.Net$ + "\" + Prompt2$
                    END IF
                END IF
                IF Display.Lowercase THEN
                    Prompt2$ = LCASE$(Prompt2$)
                END IF

                ' make prompt
                IF Short.Display4 THEN
                    Response = True
                ELSE
                    IF Continuous.Display3 THEN
                        Response = True
                    ELSE
                        IF Short.Display3 THEN
                            Prompt2$ = "File exists. "
                        ELSE
                            Prompt2$ = "File " + Prompt2$ + " exists. "
                            PRINT Prompt2$
                            Prompt2$ = Nul
                        END IF
                        IF Append.Dest THEN
                            IF Resume.File THEN
                                Prompt2$ = Prompt2$ + "Resume"
                            ELSE
                                Prompt2$ = Prompt2$ + "Append"
                            END IF
                        ELSE
                            Prompt2$ = Prompt2$ + "Overwrite"
                        END IF
                        Prompt2$ = Prompt2$ + "(y)es,(n)o,(c)opyall,(s)kipall,(q)uit)?"
                        CALL MorePrompt2(Prompt2$, "ynqsc", Outpt$, "y")
                        SELECT CASE Outpt$
                            CASE "c"
                                Response = True
                                Continue.Searching = True
                            CASE "q"
                                Response = False
                                Quit.Searching = True
                            CASE "s"
                                Skip.All = True
                                Response = False
                            CASE "y"
                                Response = True
                            CASE "n"
                                Response = False
                        END SELECT
                    END IF
                END IF
            END IF
        END IF
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' routine to create directory,
'   and copy directory attributes.
SUB Copyit_DirectoryStructure (Source.Path$, Target.Net$)
DIM ASCIIZ AS STRING * 260

' check for copying directory onto itself
CALL Copyit_DirectoryCopying(Target.Net$)

' process directory copying
IF Directory.Exists THEN
    Recurse.Error = True
    IF Display.Errors = False THEN
        IF Wide.List THEN
            Wide.List = False
            PRINT
        END IF
        COLOR White, Black
        IF Directory.Exists = -1 THEN
            CALL Copyit_DisplayError("Recursive directory error.")
        ELSE
            IF Directory.Exists = -2 THEN
                CALL Copyit_DisplayError("Self-Recursive directory error.")
            END IF
        END IF
    END IF
    EXIT SUB
END IF

' make directory
Filename$ = RTRIM$(Dir.Copy.Target)
Filename$ = LEFT$(Filename$, LEN(Filename$) - 1)

' check directory name
IF Filename$ <> "." AND Filename$ <> ".." THEN

    ' create directory
    Directory.Path = Filename$
    CALL Copyit_CreateDirectory(Target.Net$)

    ' increment directory counter
    Dirs.Counted = Dirs.Counted + 1!

    ' display directory created
    IF Debug.ModeX THEN
        PRINT "Verified: " + Filename$
    END IF
    EXIT SUB

    ' make directory name
    Directory$ = RTRIM$(Dir.Copy.Name)
    Directory$ = LEFT$(Directory$, LEN(Directory$) - 1)

    ' make source directory name
    IF Source.Path$ = Nul THEN
        ASCIIZ = Directory$ + CHR$(0)
    ELSE
        ASCIIZ = Source.Path$ + CHR$(0)
    END IF

    ' store source directory attribute
    Dir.Attribute = GetFileAttributes(ASCIIZ)

    ' make target directory name
    IF Target.Net$ = Nul THEN
        ASCIIZ = Filename$ + CHR$(0)
    ELSE
        IF LEFT$(Filename$, 1) = "\" THEN
            IF RIGHT$(Target.Net$, 1) = "\" THEN
                ASCIIZ = Target.Net$ + MID$(Filename$, 2) + CHR$(0)
            ELSE
                ASCIIZ = Target.Net$ + Filename$ + CHR$(0)
            END IF
        ELSE
            IF RIGHT$(Target.Net$, 1) = "\" THEN
                ASCIIZ = Target.Net$ + Filename$ + CHR$(0)
            ELSE
                ASCIIZ = Target.Net$ + "\" + Filename$ + CHR$(0)
            END IF
        END IF
    END IF

    ' copy directory attribute
    x = SetFileAttributes&(ASCIIZ, Dir.Attribute)
END IF
END SUB

' routine to create directory structure
SUB Copyit_CreateDirectory (Next.Dest.Net$)
' construct path
Directory$ = RTRIM$(Directory.Path$)
IF Next.Dest.Net$ = Nul THEN
    IF MID$(Directory$, 2, 1) = ":" THEN
        Drive.Create$ = LEFT$(Directory$, 1)
        Directory$ = MID$(Directory$, 3)
    ELSE
        Drive.Create$ = CHR$(Drive.Number)
    END IF
END IF
IF RIGHT$(Directory$, 1) <> "\" THEN
    Directory$ = Directory$ + "\"
END IF

' create path
Next.Dir = INSTR(Directory$, "\")
DO
    IF Next.Dir = False THEN
        EXIT DO
    END IF

    Filename$ = LEFT$(Directory$, Next.Dir - 1)
    Next.Dir = INSTR(Next.Dir + 1, Directory$, "\")

    ' make directory name
    IF LEN(Filename$) THEN
        IF LEN(Next.Dest.Net$) THEN
            Filename$ = Next.Dest.Net$ + "\" + Filename$
        ELSE
            Filename$ = Drive.Create$ + ":" + Filename$
        END IF

        ' check directory
        IF _DIREXISTS(Filename$) = 0 THEN
            ' create directory
            f$ = Filename$ + CHR$(0)
            x = CreateDirectoryA(f$, 0)

            ' increment directory counter
            IF x THEN
                Dirs.Created = Dirs.Created + 1!
                IF Debug.ModeX THEN
                    PRINT "Created: "; Filename$
                END IF
            END IF

            ' display any error
            IF x = 0 THEN
                IF Debug.ModeX THEN
                    IF GetLastError& = &HB7 THEN ' ignore already exists
                        ' nul
                    ELSE
                        Var$ = "Error (0x" + HEX$(GetLastError&) + ") creating directory."
                        CALL Copyit_DisplayError(Var$)
                    END IF
                END IF
            END IF
        END IF
    END IF
LOOP

' check if file is unicode
IF DirIsUnicode THEN
    IF LEN(Next.Dest.Net$) THEN
        Filename$ = Next.Dest.Net$ + "\" + Directory$
    ELSE
        Filename$ = Drive.Create$ + ":" + Directory$
    END IF
    DirUnicodeShort = AsciiToWide$(Filename$) + CHR$(0)

    ' parse unicode filename
    Var$ = Filename$
    IF RIGHT$(Var$, 1) = "\" THEN
        Var$ = LEFT$(Var$, LEN(Var$) - 1)
    END IF
    Var = 0
    FOR x = LEN(Var$) TO 1 STEP -1
        IF MID$(Var$, x, 1) = "\" THEN
            Var$ = AsciiToWide$(LEFT$(Var$, x)) + DirUnicodeLong
            Var = -1
            EXIT FOR
        END IF
    NEXT
    IF Var = 0 THEN
        Var$ = DirUnicodeLong
    END IF

    ' rename the filename
    r = MoveFileW(DirUnicodeShort, Var$ + CHR$(0))
    IF r = 0 THEN
        IF Debug.ModeX THEN
            PRINT "Unicode directory rename error: "; HEX$(GetLastError)
        END IF
    END IF
END IF
END SUB

' subroutine to delete destination filename
SUB Copyit_DeleteDestFile (Delete.Prompt, Work.Net$, Filename$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260

' close output file
IF Append.Files THEN
    CALL Copyit_CloseFile(-1)
END IF

' make drive
Start.Dir = Dest.Dir
IF MID$(Start.Dir, 2, 1) = ":" THEN
    Drive.Number = ASC(UCASE$(LEFT$(Start.Dir, 1)))
    Start.Dir = MID$(Start.Dir, 3)
ELSE
    Drive.Number = ASC(UCASE$(Drive.Search))
END IF
 
' display file being deleted
IF Dest.File = Nul THEN
    IF Work.Net$ = Nul THEN
        Source.File$ = CHR$(Drive.Number) + ":" + Start.Dir + Filename$
    ELSE
        Source.File$ = Work.Net$ + "\" + Start.Dir + Filename$
    END IF
ELSE
    IF Work.Net$ = Nul THEN
        Source.File$ = CHR$(Drive.Number) + ":" + Start.Dir
    ELSE
        Source.File$ = Work.Net$ + "\" + Start.Dir
    END IF
END IF

' check prompt to delete
IF Delete.Prompt THEN
    Prompt2$ = Nul
    IF Display.Errors THEN
        Prompt2$ = "Error copying file. "
    END IF
    Prompt2$ = Prompt2$ + "Delete partially copied destination file: " + Source.File$ + "(y)es,(n)o,(q)uit?"
    CALL MorePrompt2(Prompt2$, "ynq", Respond$, "y")
    SELECT CASE Respond$
        CASE "n"
            EXIT SUB
        CASE "q"
            Quit.Searching = True
            EXIT SUB
    END SELECT
END IF

' check to display file
Deleted.Files = Deleted.Files + 1
IF Continuous.Display = False THEN
    IF Wide.List THEN
        Wide.List = False
        PRINT
    END IF
    COLOR White, Black
    PRINT "Deleting file: " + Source.File$
END IF

' delete filename
ASCIIZ = Source.File$ + CHR$(0)

' change filename attribute
AttrX& = GetFileAttributes(ASCIIZ)
AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
x = SetFileAttributes&(ASCIIZ, AttrX&)

' delete long filename
x = DeleteFileA(ASCIIZ)

' check delete error
IF x = 0 THEN
    IF Debug.Mode THEN
        Var$ = "Error (0x" + HEX$(GetLastError&) + ") deleting file."
        CALL Copyit_DisplayError(Var$)
    END IF
END IF
END SUB

' subroutine to delete source filename
SUB Copyit_DeleteSourceFile (Directory$, Filename$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260

' make filename
Source.File$ = Directory$ + Filename$

' make directory name
Temp.Dir$ = Directory$
CALL Copyit_MakeFile(Nul, Source.File$, Temp.Dir$)

' reset delete flag
Response = False

' check continue flag
IF Continue.Deleting THEN
    Response = True
END IF

' check delete prompting
IF Prompt.Delete1 THEN
    Response = True
END IF

' prompt to delete
IF Response = False THEN
    Outpt$ = Source.File$
    IF Display.Lowercase THEN
        Outpt$ = LCASE$(Outpt$)
    END IF
    Prompt2$ = "Delete copied source file: " + Outpt$ + "(y)es,(n)o,(d)elete all,(q)uit?"
    CALL MorePrompt2(Prompt2$, "ynqd", Respond$, "y")
    SELECT CASE Respond$
        CASE "d"
            Response = True
            Continue.Deleting = True
        CASE "q"
            Response = False
            Quit.Searching = True
        CASE "y"
            Response = True
        CASE "n"
            Response = False
    END SELECT
END IF

' check delete flag
IF Response = True THEN
    Deleted.Files = Deleted.Files + 1
    ' display file being deleted
    IF Continuous.Display = False THEN
        IF Wide.List THEN
            Wide.List = False
            PRINT
        END IF
        COLOR White, Black
        Prompt2$ = "Deleting file: " + Outpt$
        PRINT Prompt2$
    END IF

    ' delete filename
    ASCIIZ = Source.File$ + CHR$(0)

    ' change filename attribute
    AttrX& = GetFileAttributes(ASCIIZ)
    AttrX& = AttrX& AND NOT &H1 ' reset read-only bit
    x = SetFileAttributes&(ASCIIZ, AttrX&)

    ' check error and delete 8.3 filename
    IF x = 0 THEN
        Short.Filename$ = finddatatemp.cAlternateFileName
        V = INSTR(Short.Filename$, CHR$(0))
        IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
        ASCIIZ = Directory$ + Short.Filename$ + CHR$(0)

        ' change filename attribute
        AttrX& = GetFileAttributes(ASCIIZ)
        AttrX& = AttrX& AND NOT &H1
        x = SetFileAttributes&(ASCIIZ, AttrX&)

        ' delete short filename
        x = DeleteFileA(ASCIIZ)
    END IF

    ' delete long filename
    x = DeleteFileA(ASCIIZ)
    IF x = 0 THEN
        IF Debug.Mode THEN
            Var$ = "Error (0x" + HEX$(GetLastError&) + ") deleting file."
            CALL Copyit_DisplayError(Var$)
        END IF
    END IF
END IF
END SUB

' subroutine to delete subdirectory
SUB Copyit_DeleteSubdirectory (Source.Directory$)
' declare work variables
DIM Attribute AS _UNSIGNED LONG
DIM DeleteASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory name
Directory$ = Source.Directory$
Temp.Dir$ = Directory$
CALL Copyit_MakeFile(Nul, Directory$, Temp.Dir$)

' find first directory
IF RIGHT$(Source.Directory$, 1) = "\" THEN
    DeleteASCIIZ = LEFT$(Source.Directory$, LEN(Source.Directory$) - 1) + CHR$(0)
ELSE
    DeleteASCIIZ = Source.Directory$ + CHR$(0)
END IF
Wfile.Handle = FindFirstFileA(_OFFSET(DeleteASCIIZ), _OFFSET(finddata))
IF Wfile.Handle = 0 THEN
    EXIT SUB
END IF
x = FindClose(Wfile.Handle)

' reset delete flag
Response = False

' check to continue
IF Continue.Subdelete THEN
    Response = True
END IF

' check for prompting
IF Prompt.Delete2 THEN
    Response = True
END IF

' make directory name
Outpt$ = Directory$
IF Display.Lowercase THEN
    Outpt$ = LCASE$(Outpt$)
END IF

' check delete flag
IF Response = False THEN
    ' prompt to delete
    Prompt2$ = "Delete copied source directory: " + Outpt$ + "(y)es,(n)o,(d)elete all,(q)uit?"
    CALL MorePrompt2(Prompt2$, "ynqd", Respond$, "y")
    SELECT CASE Respond$
        CASE "d"
            Response = True
            Continue.Subdelete = True
        CASE "q"
            Response = False
            Quit.Searching = True
        CASE "y"
            Response = True
        CASE "n"
            Response = False
    END SELECT
END IF

' check delete flag
IF Response THEN
    Deleted.Dirs = Deleted.Dirs + 1
    IF Continuous.Display = False THEN
        IF Wide.List THEN
            Wide.List = False
            PRINT
        END IF
        COLOR White, Black
        PRINT "Deleting directory: " + Outpt$
    END IF
    CALL Copyit_TreeDirectories(Directory$)
END IF
END SUB

' subroutine to delete subdirectories
SUB Copyit_TreeDirectories (Directory.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' delete filenames
    CALL Copyit_DeleteFiles(Directory.Search$)

    ' recurse subdirectories
    DO
        ' check directory attribute
        Attribute = finddata.dwFileAttributes

        ' check for directory
        IF (Attribute AND &H10) = &H10 THEN

            ' store directory name
            Directory$ = finddata.cFileName
            Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

            ' check directory name
            IF Directory$ <> "." AND Directory$ <> ".." THEN
                ' check unicode
                IF INSTR(Directory$, "?") THEN
                    Directory$ = finddata.cAlternateFileName
                    V = INSTR(Directory$, CHR$(0))
                    IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                END IF

                ' make next search directory
                Next.Directory$ = Directory.Search$ + Directory$ + "\"

                ' recursively search subdirectories
                CALL Copyit_TreeDirectories(Next.Directory$)
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF

' delete directory
CALL Copyit_DeleteDirectory(Directory.Search$)
END SUB

' subroutine to delete files in a directory
SUB Copyit_DeleteFiles (Directory$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search filenames
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN
            ' delete file
            ASCIIZ2 = Directory$ + Filename$ + CHR$(0)

            ' change filename attribute
            AttrX& = GetFileAttributes(ASCIIZ2)
            AttrX& = AttrX& AND NOT &H1 ' reset read-only bit
            x = SetFileAttributes&(ASCIIZ2, AttrX&)

            ' delete long filename
            x = DeleteFileA(ASCIIZ2)

            ' check error and delete 8.3 filename
            IF x = 0 THEN
                Short.Filename$ = finddata.cAlternateFileName
                V = INSTR(Short.Filename$, CHR$(0))
                IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
                ASCIIZ2 = Directory$ + Short.Filename$ + CHR$(0)

                ' change filename attribute
                AttrX& = GetFileAttributes(ASCIIZ2)
                AttrX& = AttrX& AND NOT &H1
                x = SetFileAttributes&(ASCIIZ2, AttrX&)

                ' delete short filename
                x = DeleteFileA(ASCIIZ2)
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to delete an empty directory
SUB Copyit_DeleteDirectory (Directory$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260

' make filename
Temp.Dir$ = Directory$
IF RIGHT$(Temp.Dir$, 1) = "\" THEN
    Temp.Dir$ = LEFT$(Temp.Dir$, LEN(Temp.Dir$) - 1)
END IF
ASCIIZ = Temp.Dir$ + CHR$(0)

' change directory attribute
AttrX& = GetFileAttributes(ASCIIZ)
AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
x = SetFileAttributes&(ASCIIZ, AttrX&)

' remove an empty directory
x = RemoveDirectoryA(ASCIIZ)

' check delete error flag
IF x = 0 THEN
    IF Debug.Mode THEN
        Var$ = "Error (0x" + HEX$(GetLastError&) + ") deleting directory."
        CALL Copyit_DisplayError(Var$)
    END IF
END IF

' check delete error flag
IF x THEN
    Total.Deleted = Total.Deleted + 1
END IF
END SUB

' verifies duplicate filenames,
'   displays filenames being copied.
SUB Copyit_FilenameCopying (Copy.Net$, From$, To$)
' reset flag
Filename.Exists = False

' make filenames
From.File$ = From$
To.File$ = To$
CALL Copyit_MakeFile(Copy.Net$, From.File$, To.File$)

' check filename length
IF LEN(To.File$) > 259 THEN
    Filename.Exists = True
    EXIT SUB
END IF

' check to run DOS commands
FOR Count = 1 TO Number.Commands
    Next.Command$ = RTRIM$(DOS.Command(Count))
    IF LEN(Next.Command$) THEN
        DOS.Command = Next.Command$
        CALL ExecuteCommand(From.File$)
    END IF
NEXT

' store filenames
From.Filename$ = From.File$
To.Filename$ = To.File$

' display filename
GOSUB Display.File

' make source netpath
IF Copy.Net$ <> Nul THEN
    From.Filename$ = From$
    CALL Copyit_MakeFile2(From.Filename$)
END IF

' compare filenames
IF LCASE$(From.Filename$) = LCASE$(To.Filename$) THEN
    Filename.Exists = True
    IF Continuous.Display = False THEN
        IF Wide.List THEN
            Wide.List = False
            PRINT
        END IF
        COLOR White, Black
        PRINT "File in list cannot be copied onto itself."
    END IF
END IF
EXIT SUB

' display filename being copied
Display.File:

' display file being copied in title
Var1$ = AmbiguateFile$(From.File$)
Var2$ = AmbiguateFile$(To.File$)
Title.Header = "COPY - " + Var1$ + " -> " + Var2$
_TITLE Title.Header

' ambiguate filenames
IF Display.Wide OR Ambiguate.Display THEN
    From.File$ = Ambiguate1$(From.File$)
    To.File$ = Ambiguate1$(To.File$)
END IF

' convert to lowercase
IF Display.Lowercase THEN
    From.File$ = LCASE$(From.File$)
    To.File$ = LCASE$(To.File$)
END IF

' truncate pathname
IF Display.Path OR Display.Wide THEN
    FOR Imbedded = LEN(From.File$) TO 1 STEP -1
        IF MID$(From.File$, Imbedded, 1) = "\" THEN
            From.File$ = LEFT$(From.File$, 2) + MID$(From.File$, Imbedded + 1)
            EXIT FOR
        END IF
    NEXT
    FOR Imbedded = LEN(To.File$) TO 1 STEP -1
        IF MID$(To.File$, Imbedded, 1) = "\" THEN
            To.File$ = LEFT$(To.File$, 2) + MID$(To.File$, Imbedded + 1)
            EXIT FOR
        END IF
    NEXT
END IF

' truncate drive letter
IF Drive.Letter THEN
    IF MID$(From.File$, 2, 1) = ":" THEN
        From.File$ = MID$(From.File$, 3)
    END IF
    IF MID$(To.File$, 2, 1) = ":" THEN
        To.File$ = MID$(To.File$, 3)
    END IF
END IF

' store filename
COLOR Yellow, Black
Outpt$ = RTRIM$(From.File$)

' truncate drive letter
IF Copy.Net$ <> Nul THEN
    IF MID$(Outpt$, 2, 1) = ":" THEN
        Outpt$ = MID$(Outpt$, 3)
    END IF
END IF

' store only source filename
IF CurrentCommand < CommandCount THEN
    StdoutLines = StdoutLines + 1
    IF StdoutLines > StdoutMax THEN
        StdoutMax = StdoutLines
        REDIM _PRESERVE StdoutArray(StdoutMax) AS STRING
    END IF
    StdoutArray(StdoutLines) = Outpt$
END IF

' display filename in wide list format
IF Display.Wide THEN
    IF Short.Display4 = 0 THEN
        ' check filename
        IF LEN(Outpt$) > 14 THEN
            Outpt$ = LEFT$(Outpt$, 13) + "~"
        END IF

        ' display filename in wide format
        IF LEFT$(Outpt$, 2) = "\\" THEN
            Outpt$ = MID$(Outpt$, 3)
        END IF
        PRINT #StdoutHandle, Outpt$ + SPACE$(15 - LEN(Outpt$));

        ' count wide display filenames
        Wide.List = Wide.List + 1
        IF Wide.List = 5 THEN
            Wide.List = False
            PRINT #StdoutHandle, ""
        END IF
    END IF
    RETURN
END IF

' prepend copying type
IF Short.Display2 = False THEN
    IF Append.Dest THEN
        IF Resume.File THEN
            Outpt$ = "Resuming: " + Outpt$
        ELSE
            Outpt$ = "Appending: " + Outpt$
        END IF
    ELSE
        Outpt$ = "Copying: " + Outpt$
    END IF
END IF

' display only source filename
IF Short.Display4 = 0 THEN
    IF Short.Display THEN
        PRINT #StdoutHandle, Outpt$
    ELSE
        ' display source/destination filenames
        Outpt$ = Outpt$ + " - " + To.File$
        PRINT #StdoutHandle, Outpt$
    END IF
END IF
RETURN
END SUB

' ambiguate filename
FUNCTION Ambiguate1$ (Var$)
' declare subroutine variables
DIM ASCIIZ1 AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
VarX$ = Var$

' read 8.3 filename
ASCIIZ1 = VarX$ + CHR$(0)
ret = GetShortPathName(ASCIIZ1, ASCIIZ2, MAX_PATH)

' check function error
IF ret > 0 THEN
    ' store 8.3 filename
    VarX$ = ASCIIZ2
    Imbedded = INSTR(VarX$, CHR$(0))
    IF Imbedded THEN
        VarX$ = LEFT$(VarX$, Imbedded - 1)
    END IF
END IF
Ambiguate1$ = VarX$
END FUNCTION

' ambiguate filename
FUNCTION AmbiguateFile$ (Var$)
' declare subroutine variables
DIM ASCIIZ1 AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
VarX$ = Var$

' read 8.3 filename
ASCIIZ1 = VarX$ + CHR$(0)
ret = GetShortPathName(ASCIIZ1, ASCIIZ2, MAX_PATH)

' check function error
IF ret > 0 THEN
    ' store 8.3 filename
    VarX$ = ASCIIZ2
    Imbedded = INSTR(VarX$, CHR$(0))
    IF Imbedded THEN
        VarX$ = LEFT$(VarX$, Imbedded - 1)
    END IF
END IF

' strip path
FOR Imbedded = LEN(VarX$) TO 1 STEP -1
    IF MID$(VarX$, Imbedded, 1) = "\" THEN
        VarX$ = MID$(VarX$, Imbedded + 1)
        EXIT FOR
    END IF
NEXT

' strip drive
IF MID$(VarX$, 2, 1) = ":" THEN
    VarX$ = MID$(VarX$, 3)
END IF
AmbiguateFile$ = VarX$
END FUNCTION

' unambiguate filename
FUNCTION UnAmbiguateFile$ (Var$)
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _UNSIGNED _OFFSET
File$ = Var$
TempASCIIZ = Var$ + CHR$(0)
hfind = FindFirstFileA(_OFFSET(TempASCIIZ), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    File$ = finddata.cFileName
    V = INSTR(File$, CHR$(0))
    IF V THEN File$ = LEFT$(File$, V - 1)
    x = FindClose(hfind)
END IF
UnAmbiguateFile$ = File$
END FUNCTION

' routine parses command line for switches, filenames, and additional info.
SUB ReadSwitches (VarX)
' assign a quote.
Quote = CHR$(34)
VarX = 0

' check command line.
IF Command.Line = "/?" THEN
    CALL Copyit_BootUsage
    COLOR Plain, Black
    PRINT "Exiting to DOS:"
    VarX = 2
    EXIT SUB
END IF

' concatenate environment variable.
IF INSTR(Command.Line, "/J") OR INSTR(Command.Line, "/j") THEN
    Force.LowerCase = True
END IF
IF Force.LowerCase THEN
    Command.Line = Command.Line + LCASE$(ENVIRON$("COPYIT"))
ELSE
    Command.Line = Command.Line + ENVIRON$("COPYIT")
END IF

' read config switches.
Ignore.Config = ParseLine("/.")
Ignore.Skip = ParseLine("/:")

' check config switch.
IF Ignore.Config = False THEN

    ' search for config file in dos path.
    CALL Copyit_ReadConfig(StartDir$, File.Found)

    ' search for config file in dosutils path.
    IF File.Found = False THEN
        CALL Copyit_ReadConfig(ENVIRON$("DOSUTILS"), File.Found)
    END IF

    ' search config file in path variable.
    IF File.Found = False THEN
        Var$ = ENVIRON$("PATH")
        IF Var$ <> Nul THEN
            DO
                Imbedded = INSTR(Var$, ";")
                IF Imbedded THEN
                    Path$ = LEFT$(Var$, Imbedded - 1)
                    Var$ = MID$(Var$, Imbedded + 1)
                ELSE
                    Path$ = Var$
                    Var$ = Nul
                END IF
                IF LEN(Path$) THEN
                    CALL Copyit_ReadConfig(Path$, File.Found)
                    IF File.Found THEN
                        EXIT DO
                    END IF
                END IF
                IF Var$ = Nul THEN
                    EXIT DO
                END IF
            LOOP
        END IF
    END IF
END IF

' check command line.
IF Command.Line = "" THEN
    Boot.List$ = "Error specifying command line."
    GOTO Boot.Error
END IF

' read command line display switch.
IF INSTR(Command.Line, "/P") OR INSTR(Command.Line, "/p") THEN
    Display.CommandLine = True
END IF
IF Display.CommandLine THEN
    PRINT Command.Line
END IF

' read startup command line switches.
Force.LowerCase = ParseLine("/J")
Display.CommandLine = ParseLine("/P")
Ignore.Sort = ParseLine("/-1")
IF ParseLine("/-2") THEN
    Sort.Method = 1
END IF
IF Sort.Method = 0 THEN
    IF ParseLine("/-3") THEN
        Sort.Method = 2
    END IF
END IF

' read attribute switches.
Set.Source.Archive = ParseLine("///XA")
Set.Source.Hidden = ParseLine("///XH")
Set.Source.Readonly = ParseLine("///XO")
Set.Source.System = ParseLine("///XS")
Set.Source.Any = ParseLine("///XX")

Clear.Source.Archive = ParseLine("///ZA")
Clear.Source.Hidden = ParseLine("///ZH")
Clear.Source.Readonly = ParseLine("///ZO")
Clear.Source.System = ParseLine("///ZS")
Clear.Source.Any = ParseLine("///ZX")

Set.Dest.Archive = ParseLine("//XA")
Set.Dest.Hidden = ParseLine("//XH")
Set.Dest.Readonly = ParseLine("//XO")
Set.Dest.System = ParseLine("//XS")
Set.Dest.Any = ParseLine("//XX")

Clear.Dest.Archive = ParseLine("//ZA")
Clear.Dest.Hidden = ParseLine("//ZH")
Clear.Dest.Readonly = ParseLine("//ZO")
Clear.Dest.System = ParseLine("//ZS")
Clear.Dest.Any = ParseLine("//ZX")

Display.Archive = ParseLine("/XA")
Display.Hidden = ParseLine("/XH")
Display.Readonly = ParseLine("/XO")
Display.System = ParseLine("/XS")
Display.Any = ParseLine("/XX")

Display.Compressed = ParseLine("/XC")
Display.Encrypted = ParseLine("/XE")

No.Display.Archive = ParseLine("/ZA")
No.Display.Hidden = ParseLine("/ZH")
No.Display.Readonly = ParseLine("/ZO")
No.Display.System = ParseLine("/ZS")
No.Display.Any = ParseLine("/ZX")

No.Display.Compressed = ParseLine("/ZC")
No.Display.Encrypted = ParseLine("/ZE")

' read synchronize switches.
Synch.Files1 = ParseLine("/A1")
Synch.Files2 = ParseLine("/A2")
Synch.Files3 = ParseLine("/A3")
Synch.Files4 = ParseLine("/A4")
Synch.Files5 = ParseLine("/A5")
Synch.Files6 = ParseLine("/A6")
Synch.Files7 = ParseLine("/A7")
Synch.Files8 = ParseLine("/A8")
Synch.Files9 = ParseLine("/A9")
Synch.FilesA = ParseLine("/AA")
Synch.FilesB = ParseLine("/AB")
Synch.FilesC = ParseLine("/AC")

' read synchronize override switches.
Synch.Type1 = ParseLine("/Z1")
Synch.Type2 = ParseLine("/Z2")
Synch.Type3 = ParseLine("/Z3")

' reset default date/time switch.
IF Synch.Type1 = False THEN
    IF Synch.Type2 = False THEN
        IF Synch.Type3 = False THEN
            Synch.Type3 = True
        END IF
    END IF
END IF

' read synchronize override switches.
Compare.Create.Date = ParseLine("/S1")
Compare.Access.Date = ParseLine("/S2")
Compare.Modify.Date = ParseLine("/S3")

' reset default synchronize override switch.
IF Compare.Create.Date = False THEN
    IF Compare.Access.Date = False THEN
        IF Compare.Modify.Date = False THEN
            Compare.Modify.Date = True
        END IF
    END IF
END IF

' read remaining alphabetic switches.
Synch.Files = ParseLine("/A")
Copy.Directory = ParseLine("/B1")
Zero.Nest = ParseLine("/B2")
Skip.Directory = ParseLine("/B3")
Continuous.Display2 = ParseLine("/C1")
Continuous.Display3 = ParseLine("/C2")
Continuous.Display = ParseLine("/C")
Drive.Letter = ParseLine("/H")
Delete.Copied = ParseLine("/I")
Delete.Directory = ParseLine("/K")
Dont.Overwrite = ParseLine("/M")
Prompt.Delete1 = ParseLine("/N")
Display.Wide = ParseLine("/O1")
Ambiguate.Display = ParseLine("/O2")
Prompt.Delete2 = ParseLine("/Q")
Recurse.Directories = ParseLine("/R")
Display.Path = ParseLine("/S")
Overwrite.Prompt = ParseLine("/W")
Short.Display = ParseLine("/X1")
Short.Display2 = ParseLine("/X2")
Short.Display3 = ParseLine("/X3")
Short.Display4 = ParseLine("/X4")
Display.Lowercase = ParseLine("/Y")
Display.Errors = ParseLine("/Z")
Disable.Cmd = ParseLine("/_")
Extended.Time = ParseLine("/-4")
Skip.All = ParseLine("/-5")

' parse append switches
Imbedded = INSTR(UCASE$(Command.Line), "/V2")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 3)
    CALL GetNumeric(Var#)
    IF Var# > 4294967295# THEN
        Boot.List$ = "Error specifying append value."
        GOTO Boot.Error
    END IF
    Byte.Switch = True
    Byte.Override = Var#
    Append.Dest = ParseLine("/V")
ELSE
    Resume.File = ParseLine("/V1")
    IF Resume.File THEN
        Append.Dest = True
    ELSE
        Append.Dest = ParseLine("/V")
    END IF
END IF

' read more switches
Pattern.Match = ParseLine("/@2")
Enable.Redirect = ParseLine("/@")
Disable.Prompts = ParseLine("/#")

' read extended display switches.
IF ParseLine("/L1") THEN
    Double.Line = 1
END IF
IF ParseLine("/L2") THEN
    Double.Line = 2
END IF
IF ParseLine("/L3") THEN
    Double.Line = 3
END IF
IF ParseLine("/L4") THEN
    Double.Line = 4
END IF
IF ParseLine("/L") THEN
    Double.Line = 4
END IF
 
' init first copied directory flag.
IF Zero.Nest THEN
    Copy.Directory = True
END IF
IF Skip.Directory THEN
    Copy.Directory = True
END IF

' parse special debug switch.
Debug.ModeX = ParseLine("/-=")

' parse progress switches.
Progress.Bar = ParseLine("/*")
Percent.Display = ParseLine("/!")
Dot.Mode = ParseLine("/-")

' init display switches.
IF Percent.Display THEN
    Dot.Mode = True
END IF
IF Progress.Bar THEN
    Percent.Display = False
    Dot.Mode = True
END IF

' read timing switch.
Rate.Mode = ParseLine("/+")

' read debug mode switches
DebugDelay! = 3!
Var2 = 0
FOR Var = 0 TO 9
    Var2 = ParseLine("/=" + MID$(STR$(Var), 2))
    IF Var2 THEN
        DebugDelay! = CSNG(Var)
        Debug.Mode = -1
        EXIT FOR
    END IF
NEXT
IF Var2 = 0 THEN
    Debug.Mode = ParseLine("/=")
END IF

' set attribute display variable.
Display.Attribute = False
IF Display.Archive OR No.Display.Archive THEN
    Display.Attribute = True
END IF
IF Display.Hidden OR No.Display.Hidden THEN
    Display.Attribute = True
END IF
IF Display.Readonly OR No.Display.Readonly THEN
    Display.Attribute = True
END IF
IF Display.System OR No.Display.System THEN
    Display.Attribute = True
END IF
IF Display.Any OR No.Display.Any THEN
    Display.Attribute = True
END IF
IF Display.Compressed OR Display.Encrypted THEN
    Display.Attribute = True
END IF
IF No.Display.Compressed OR No.Display.Encrypted THEN
    Display.Attribute = True
END IF

' set source attribute flag.
Set.Source.Attribute = False
IF Clear.Source.Archive OR Set.Source.Archive THEN
    Set.Source.Attribute = True
END IF
IF Clear.Source.Hidden OR Set.Source.Hidden THEN
    Set.Source.Attribute = True
END IF
IF Clear.Source.Readonly OR Set.Source.Readonly THEN
    Set.Source.Attribute = True
END IF
IF Clear.Source.System OR Set.Source.System THEN
    Set.Source.Attribute = True
END IF
IF Clear.Source.Any OR Set.Source.Any THEN
    Set.Source.Attribute = True
END IF

' reset file counter variables.
Bytes.Copied = DFalse
Error.Level = False
Dirs.Counted = SFalse
Dirs.Created = SFalse
Files.Counted = DFalse
More.Display = False
Total.Bytes = DFalse
Total.Bytes2 = DFalse
Total.Deleted = False
Wide.List = False

' read excluded filename specs from command line.
Max.Excluded = 10
Number.Excluded = False
DO
    Excluded = INSTR(Command.Line, "/(")
    IF Excluded = False THEN
        EXIT DO
    END IF
    Var = LastSwitch(Excluded)
    Excluded.Spec$ = Nul
    Next.Bracket = INSTR(Excluded + 2, Command.Line, ")")
    IF Next.Bracket THEN
        Excluded.Spec$ = MID$(Command.Line, Excluded + 2, Next.Bracket - Excluded - 2)
        Command.Line = LEFT$(Command.Line, Excluded - 1) + MID$(Command.Line, Next.Bracket + 1)
    END IF
    IF Excluded.Spec$ = Nul THEN
        Boot.List$ = "Error specifying excluded list."
        GOTO Boot.Error
    END IF
    Number.Excluded = Number.Excluded + 1
    IF Number.Excluded > Max.Excluded THEN
        Max.Excluded = Max.Excluded + 10
        REDIM _PRESERVE Excluded.Files(1 TO Max.Excluded) AS STRING * 128
    END IF
    Excluded.Files(Number.Excluded) = UCASE$(Excluded.Spec$)
    Command.Line = RTRIM$(Command.Line)
LOOP

' read DOS commands from command line.
Max.Commands = 10
Number.Commands = False
DO
    DOSswitch = INSTR(Command.Line, "/[")
    IF DOSswitch = False THEN
        EXIT DO
    END IF
    Var = LastSwitch(DOSswitch)
    DOS.Spec$ = Nul
    Next.Bracket = INSTR(DOSswitch + 2, Command.Line, "]")
    IF Next.Bracket THEN
        DOS.Spec$ = MID$(Command.Line, DOSswitch + 2, Next.Bracket - DOSswitch - 2)
        Command.Line = LEFT$(Command.Line, DOSswitch - 1) + MID$(Command.Line, Next.Bracket + 1)
    END IF
    IF DOS.Spec$ = Nul THEN
        Boot.List$ = "Error specifying DOS command."
        GOTO Boot.Error
    END IF
    Number.Commands = Number.Commands + 1
    IF Number.Commands > Max.Commands THEN
        Max.Commands = Max.Commands + 10
        REDIM _PRESERVE DOS.Command(1 TO Max.Commands) AS STRING * 128
    END IF
    DOS.Command(Number.Commands) = DOS.Spec$
    Command.Line = RTRIM$(Command.Line)
LOOP

' parse dos command parameters.
FOR Count = 1 TO Number.Commands
    Next.Command$ = DOS.Command(Count)
    Imbedded = INSTR(Next.Command$, "//1")
    WHILE Imbedded
        Next.Command$ = LEFT$(Next.Command$, Imbedded - 1) + ">" + MID$(Next.Command$, Imbedded + 3)
        Imbedded = INSTR(Next.Command$, "//1")
    WEND
    Imbedded = INSTR(Next.Command$, "//2")
    WHILE Imbedded
        Next.Command$ = LEFT$(Next.Command$, Imbedded - 1) + "<" + MID$(Next.Command$, Imbedded + 3)
        Imbedded = INSTR(Next.Command$, "//2")
    WEND
    Imbedded = INSTR(Next.Command$, "//3")
    WHILE Imbedded
        Next.Command$ = LEFT$(Next.Command$, Imbedded - 1) + "|" + MID$(Next.Command$, Imbedded + 3)
        Imbedded = INSTR(Next.Command$, "//3")
    WEND
    Imbedded = INSTR(Next.Command$, "//4")
    WHILE Imbedded
        Next.Command$ = LEFT$(Next.Command$, Imbedded - 1) + "%" + MID$(Next.Command$, Imbedded + 3)
        Imbedded = INSTR(Next.Command$, "//4")
    WEND
    Imbedded = INSTR(Next.Command$, "//5")
    WHILE Imbedded
        Next.Command$ = LEFT$(Next.Command$, Imbedded - 1) + "]" + MID$(Next.Command$, Imbedded + 3)
        Imbedded = INSTR(Next.Command$, "//5")
    WEND
    DOS.Command(Count) = Next.Command$
NEXT

' read destination file date override.
Override.Date.Month = False
Override.Date.Day = False
Override.Date.Year = False
Override.Date = False
Imbedded = INSTR(Command.Line, "/5")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Date.Override$ = MID$(Command.Line, Imbedded + 2, 10)
    IF MID$(Date.Override$, 3, 1) <> "/" THEN
        Boot.List$ = "Error specifying override date."
        GOTO Boot.Error
    END IF
    IF MID$(Date.Override$, 6, 1) <> "/" THEN
        Boot.List$ = "Error specifying override date."
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 12)
    Override.Date.Month = INT(VAL(MID$(Date.Override$, 1, 2)))
    Override.Date.Day = INT(VAL(MID$(Date.Override$, 4, 2)))
    Override.Date.Year = INT(VAL(MID$(Date.Override$, 7, 4)))
    IF Override.Date.Month = 99 AND Override.Date.Day = 99 AND Override.Date.Year = 9999 THEN
        Override.Date.Month = VAL(LEFT$(DATE$, 2))
        Override.Date.Day = VAL(MID$(DATE$, 4, 2))
        Override.Date.Year = VAL(RIGHT$(DATE$, 4))
    END IF
    Override.Date = True
END IF

' read destination file time override.
Override.Time.Hour = False
Override.Time.Minute = False
Override.Time.Second = False
Override.Time = False
Imbedded = INSTR(Command.Line, "/6")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Time.Override$ = MID$(Command.Line, Imbedded + 2, 8)
    IF MID$(Time.Override$, 3, 1) <> ":" THEN
        Boot.List$ = "Error specifying override time."
        GOTO Boot.Error
    END IF
    IF MID$(Time.Override$, 6, 1) <> ":" THEN
        Boot.List$ = "Error specifying override time."
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 10)
    Override.Time.Hour = INT(VAL(MID$(Time.Override$, 1, 2)))
    Override.Time.Minute = INT(VAL(MID$(Time.Override$, 4, 2)))
    Override.Time.Second = INT(VAL(MID$(Time.Override$, 7, 2)))
    IF Override.Time.Hour = 99 AND Override.Time.Minute = 99 AND Override.Time.Second = 99 THEN
        Override.Time.Hour = INT(VAL(LEFT$(TIME$, 2)))
        Override.Time.Minute = INT(VAL(MID$(TIME$, 4, 2)))
        Override.Time.Second = INT(VAL(RIGHT$(TIME$, 2)))
    END IF
    Override.Time = True
END IF

' read date from command line.
Search.Date = False
Search.From.Date = False
Search.To.Date = False
Imbedded = INSTR(UCASE$(Command.Line), "/E")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.Date = True
    ' /e01/01/1997-01/01/1997
    IF MID$(Command.Line, Imbedded + 4, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 7, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 12, 1) = "-" AND _
    MID$(Command.Line, Imbedded + 15, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 18, 1) = "/" THEN
        D$ = MID$(Command.Line, Imbedded + 2, 21)
        S$ = LEFT$(D$, 10)
        D1! = INT(VAL(MID$(S$, 1, 2))) ' month
        D2! = INT(VAL(MID$(S$, 4, 2))) ' day
        D3! = INT(VAL(MID$(S$, 7, 4))) ' year
        IF D1! = 99! AND D2! = 99! AND D3! = 9999! THEN ' current date
            D1! = VAL(LEFT$(DATE$, 2))
            D2! = VAL(MID$(DATE$, 4, 2))
            D3! = VAL(RIGHT$(DATE$, 4))
        END IF
        Search.From.Date = (D3! - 1980) * 512 + D1! * 32 + D2!
        S$ = RIGHT$(D$, 10)
        D1! = INT(VAL(MID$(S$, 1, 2))) ' month
        D2! = INT(VAL(MID$(S$, 4, 2))) ' day
        D3! = INT(VAL(MID$(S$, 7, 4))) ' year
        IF D1! = 99! AND D2! = 99! AND D3! = 9999! THEN ' current date
            D1! = VAL(LEFT$(DATE$, 2))
            D2! = VAL(MID$(DATE$, 4, 2))
            D3! = VAL(RIGHT$(DATE$, 4))
        END IF
        Search.To.Date = (D3! - 1980) * 512 + D1! * 32 + D2!
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    ELSE
        ' /e-01/01/1997
       IF MID$(Command.Line, Imbedded + 2, 1) = "-" AND _
       MID$(Command.Line, Imbedded + 5, 1) = "/" AND _
       MID$(Command.Line, Imbedded + 8, 1) = "/" THEN
            Search.From.Date = False
            S$ = MID$(Command.Line, Imbedded + 3, 10)
            D1! = INT(VAL(MID$(S$, 1, 2))) ' month
            D2! = INT(VAL(MID$(S$, 4, 2))) ' day
            D3! = INT(VAL(MID$(S$, 7, 4))) ' year
            IF D1! = 99! AND D2! = 99! AND D3! = 9999! THEN ' current date
                D1! = VAL(LEFT$(DATE$, 2))
                D2! = VAL(MID$(DATE$, 4, 2))
                D3! = VAL(RIGHT$(DATE$, 4))
            END IF
            Search.To.Date = (D3! - 1980) * 512 + D1! * 32 + D2!
            Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 13)
        ELSE
            ' /e01/01/1997-
          IF MID$(Command.Line, Imbedded + 4, 1) = "/" AND _
          MID$(Command.Line, Imbedded + 7, 1) = "/" AND _
          MID$(Command.Line, Imbedded + 12, 1) = "-" THEN
                Search.To.Date = False
                S$ = MID$(Command.Line, Imbedded + 2, 10)
                D1! = INT(VAL(MID$(S$, 1, 2))) ' month
                D2! = INT(VAL(MID$(S$, 4, 2))) ' day
                D3! = INT(VAL(MID$(S$, 7, 4))) ' year
                IF D1! = 99! AND D2! = 99! AND D3! = 9999! THEN ' current date
                    D1! = VAL(LEFT$(DATE$, 2))
                    D2! = VAL(MID$(DATE$, 4, 2))
                    D3! = VAL(RIGHT$(DATE$, 4))
                END IF
                Search.From.Date = (D3! - 1980) * 512 + D1! * 32 + D2!
                Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 13)
            ELSE
                Boot.List$ = "Error specifying search date."
                GOTO Boot.Error
            END IF
        END IF
    END IF
    IF Search.From.Date < False OR Search.To.Date < False THEN
        Boot.List$ = "Error specifying search date."
        GOTO Boot.Error
    END IF
END IF

' read time from command line.
Search.Time = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.Time = True
    ' /t12:00:00-12:00:00
    IF MID$(Command.Line, Imbedded + 4, 1) = ":" AND _
    MID$(Command.Line, Imbedded + 7, 1) = ":" AND _
    MID$(Command.Line, Imbedded + 10, 1) = "-" AND _
    MID$(Command.Line, Imbedded + 13, 1) = ":" AND _
    MID$(Command.Line, Imbedded + 16, 1) = ":" THEN
        D$ = MID$(Command.Line, Imbedded + 2, 17)
        S$ = LEFT$(D$, 8)
        T1! = INT(VAL(MID$(S$, 1, 2))) ' hours
        T2! = INT(VAL(MID$(S$, 4, 2))) ' minutes
        T3! = INT(VAL(MID$(S$, 7, 2))) ' seconds
        IF T1! = 99 AND T2! = 99 AND T3! = 99 THEN
            T1! = INT(VAL(LEFT$(TIME$, 2)))
            T2! = INT(VAL(MID$(TIME$, 4, 2)))
            T3! = INT(VAL(RIGHT$(TIME$, 2)))
        END IF
        Search.From.Time = T1! * 2048 + T2! * 32 + T3!
        S$ = RIGHT$(D$, 8)
        T1! = INT(VAL(MID$(S$, 1, 2))) ' hours
        T2! = INT(VAL(MID$(S$, 4, 2))) ' minutes
        T3! = INT(VAL(MID$(S$, 7, 2))) ' seconds
        IF T1! = 99 AND T2! = 99 AND T3! = 99 THEN
            T1! = INT(VAL(LEFT$(TIME$, 2)))
            T2! = INT(VAL(MID$(TIME$, 4, 2)))
            T3! = INT(VAL(RIGHT$(TIME$, 2)))
        END IF
        Search.To.Time = T1! * 2048 + T2! * 32 + T3!
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    ELSE
        ' /t-12:00:00
       IF MID$(Command.Line, Imbedded + 2, 1) = "-" AND _
       MID$(Command.Line, Imbedded + 5, 1) = ":" AND _
       MID$(Command.Line, Imbedded + 8, 1) = ":" THEN
            Search.From.Time = False
            S$ = MID$(Command.Line, Imbedded + 3, 8)
            T1! = INT(VAL(MID$(S$, 1, 2))) ' hours
            T2! = INT(VAL(MID$(S$, 4, 2))) ' minutes
            T3! = INT(VAL(MID$(S$, 7, 4))) ' seconds
            IF T1! = 99 AND T2! = 99 AND T3! = 99 THEN
                T1! = INT(VAL(LEFT$(TIME$, 2)))
                T2! = INT(VAL(MID$(TIME$, 4, 2)))
                T3! = INT(VAL(RIGHT$(TIME$, 2)))
            END IF
            Search.To.Time = T1! * 2048 + T2! * 32 + T3!
            Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 11)
        ELSE
            ' /t12:00:00-
          IF MID$(Command.Line, Imbedded + 4, 1) = ":" AND _
          MID$(Command.Line, Imbedded + 7, 1) = ":" AND _
          MID$(Command.Line, Imbedded + 10, 1) = "-" THEN
                Search.To.Time = False
                S$ = MID$(Command.Line, Imbedded + 2, 10)
                T1! = INT(VAL(MID$(S$, 1, 2))) ' hours
                T2! = INT(VAL(MID$(S$, 4, 2))) ' minutes
                T3! = INT(VAL(MID$(S$, 7, 4))) ' seconds
                IF T1! = 99 AND T2! = 99 AND T3! = 99 THEN
                    T1! = INT(VAL(LEFT$(TIME$, 2)))
                    T2! = INT(VAL(MID$(TIME$, 4, 2)))
                    T3! = INT(VAL(RIGHT$(TIME$, 2)))
                END IF
                Search.From.Time = T1! * 2048 + T2! * 32 + T3!
                Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 11)
            ELSE
                Boot.List$ = "Error specifying search time."
                GOTO Boot.Error
            END IF
        END IF
    END IF
    IF Search.From.Time < False OR Search.To.Time < False THEN
        Boot.List$ = "Error specifying search time."
        GOTO Boot.Error
    END IF
END IF

' read file size from command line.
Search.File.Size = False
Search.Size.From = DFalse
Search.Size.To = DFalse
Imbedded = INSTR(UCASE$(Command.Line), "/U")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.File.Size = True
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    CALL GetNumeric(Var#)
    Search.Size.From = Var#
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        Boot.List$ = "Error specifying search size."
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    CALL GetNumeric(Var#)
    Search.Size.To = Var#
    Search.Size.From = Search.Size.From * 1024#
    Search.Size.To = Search.Size.To * 1024#
END IF

' read pattern date from command line.
Pattern.Search.Date = False
Pattern.Search.From.Date = Nul
Pattern.Search.To.Date = Nul
Imbedded = INSTR(UCASE$(Command.Line), "/E1:")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Pattern.Search.Date = True
    ' /e1:01/01/2000-01/01/2010
    IF MID$(Command.Line, Imbedded + 6, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 9, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 14, 1) = "-" AND _
    MID$(Command.Line, Imbedded + 17, 1) = "/" AND _
    MID$(Command.Line, Imbedded + 20, 1) = "/" THEN
        D$ = MID$(Command.Line, Imbedded + 4, 21)
        Pattern.Search.From.Date = LEFT$(D$, 10)
        IF Pattern.Search.From.Date = "99/99/9999" THEN ' current date
            Pattern.Search.From.Date = DATE$
            MID$(Pattern.Search.From.Date, 3, 1) = "/"
            MID$(Pattern.Search.From.Date, 6, 1) = "/"
        END IF
        Pattern.Search.To.Date = RIGHT$(D$, 10)
        IF Pattern.Search.To.Date = "99/99/9999" THEN ' current date
            Pattern.Search.To.Date = DATE$
            MID$(Pattern.Search.To.Date, 3, 1) = "/"
            MID$(Pattern.Search.To.Date, 6, 1) = "/"
        END IF
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 25)
    ELSE
        Boot.List$ = "Error specifying pattern search date."
        GOTO Boot.Error
    END IF
END IF

' read filename pattern from command line.
Pattern.Search = False
Pattern.Filename = Nul
Imbedded = INSTR(UCASE$(Command.Line), "/E2:")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    IF MID$(Command.Line, Imbedded + 4, 1) = Quote THEN
        FOR Var = Imbedded + 5 TO LEN(Command.Line)
            IF MID$(Command.Line, Var, 1) = Quote THEN
                Pattern.Search = True
                Pattern.Filename = UCASE$(MID$(Command.Line, Imbedded + 5, Var - Imbedded - 5))
                Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Var + 1)
                EXIT FOR
            END IF
        NEXT
        IF Pattern.Search = False THEN
            Boot.List$ = "Error specifying filename pattern."
            GOTO Boot.Error
        END IF
    END IF
END IF

' verify both parameters set.
IF Pattern.Search OR Pattern.Search.Date THEN
    IF Pattern.Search = False OR Pattern.Search.Date = False THEN
        Boot.List$ = "Error specifying filename pattern."
        GOTO Boot.Error
    END IF
END IF

' read remaining numeric switches.
Copy.Zero = ParseLine("/0")
Copy.Ascii = ParseLine("/1")

' read ascii replacement filters from command line.
Ascii.Filters = False
Max.Filters = 10
DO
    Imbedded = INSTR(Command.Line, "/2")
    IF Imbedded = False THEN
        EXIT DO
    END IF
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    CALL GetNumeric(Var#)
    IF Var# > 255# THEN
        Boot.List$ = "Error specifying search filter."
        GOTO Boot.Error
    END IF
    Ascii.From = CINT(Var#)
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        Boot.List$ = "Error specifying search filter."
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    CALL GetNumeric(Var#)
    IF Var# > 255# THEN
        Boot.List$ = "Error specifying search filter."
        GOTO Boot.Error
    END IF
    Ascii.To = CINT(Var#)
    IF Ascii.To <> Ascii.From THEN
        Ascii.Filters = Ascii.Filters + 1
        IF Ascii.Filters > Max.Filters THEN
            Max.Filters = Max.Filters + 10
            REDIM _PRESERVE Convert.Ascii(1 TO Max.Filters, 1 TO 2) AS INTEGER
        END IF
        Convert.Ascii(Ascii.Filters, 1) = Ascii.From
        Convert.Ascii(Ascii.Filters, 2) = Ascii.To
    END IF
LOOP

' read ascii strip filters from command line.
Ascii.Strips = False
Max.Strips = 10
DO
    Imbedded = INSTR(Command.Line, "/3")
    IF Imbedded = False THEN
        EXIT DO
    END IF
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    CALL GetNumeric(Var#)
    IF Var# > 255# THEN
        Boot.List$ = "Error specifying strip filter."
        GOTO Boot.Error
    END IF
    Ascii.Strips = Ascii.Strips + 1
    IF Ascii.Strips > Max.Strips THEN
        Max.Strips = Max.Strips + 10
        REDIM _PRESERVE Strip.Ascii(1 TO Max.Strips) AS INTEGER
    END IF
    Strip.Ascii(Ascii.Strips) = CINT(Var#)
LOOP

' read nest override switch.
Imbedded = INSTR(Command.Line, "/4")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    CALL GetNumeric(Var#)
    IF Var# > 32767# THEN
        Boot.List$ = "Error specifying nest override."
        GOTO Boot.Error
    END IF
    Nested.Recurse = CINT(Var#)
END IF

' read override switches.
Override.Create.Date = ParseLine("/7")
Override.Access.Date = ParseLine("/8")
Override.Modify.Date = ParseLine("/9")

' reset remaining variables.
Dest.Dir = Nul
Max.Dest.Dirs = 10
Number.Dest.Dirs = False

Dest.File = Nul
Max.Dest.Files = 10
Number.Dest.Files = False

Max.Dest.Nets = 10
Number.Dest.Nets = False

Temp.Drive.Dir = Nul

' check command line.
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' read remaining filename switches.
DO
    IF INSTR(Command.Line, "/") = False THEN
        ' check trailing command line.
        Command.Line = RTRIM$(Command.Line)
        IF Switch.Exist THEN
            IF LEN(Command.Line) > Last.Switch THEN
                Boot.List$ = "Unknown command line characters."
                GOTO Boot.Error
            END IF
        END IF
        EXIT DO
    END IF
    FOR Count = LEN(Command.Line) TO 1 STEP -1
        IF MID$(Command.Line, Count, 1) = "/" THEN
            ' store switch letter.
            Switch$ = MID$(Command.Line, Count + 1, 1)

            ' store switch value.
            Var = LastSwitch(Count)
            Switch.Line$ = MID$(Command.Line, Count + 2)
            Switch.Line$ = LTRIM$(Switch.Line$)
            Switch.Line$ = RTRIM$(Switch.Line$)

            ' check quoted string.
            File.Quotes = False
            IF RIGHT$(Switch.Line$, 1) = Quote THEN
                IF LEFT$(Switch.Line$, 1) = Quote THEN
                    Switch.Line$ = MID$(Switch.Line$, 2)
                    Switch.Line$ = LEFT$(Switch.Line$, LEN(Switch.Line$) - 1)
                    File.Quotes = True
                END IF
            END IF

            ' process switch.
            SELECT CASE UCASE$(Switch$)
                CASE "F" ' filename
                    Var = LastSwitch(Count)
                    IF LEN(Switch.Line$) = False THEN
                        Boot.List$ = "Missing command line characters."
                        GOTO Boot.Error
                    END IF
                    IF INSTR(Switch.Line$, " ") THEN
                        IF File.Quotes = False THEN
                            Boot.List$ = "Unknown command line characters."
                            GOTO Boot.Error
                        END IF
                    END IF
                    Number.Dest.Files = Number.Dest.Files + 1
                    IF Number.Dest.Files > Max.Dest.Files THEN
                        Max.Dest.Files = Max.Dest.Files + 10
                        REDIM _PRESERVE Destinate.Filename(1 TO Max.Dest.Files) AS STRING * 260
                    END IF
                    Destinate.Filename(Number.Dest.Files) = Switch.Line$
                    Dest.File = Switch.Line$
                CASE "D" ' directory
                    Var = LastSwitch(Count)
                    IF LEN(Switch.Line$) = False THEN
                        ' check /d w/o parameters specified
                        IF Zero.Dirs THEN
                            Boot.List$ = "Multiple nul directories not allowed."
                            GOTO Boot.Error
                        END IF
                        ' set /d w/o parameters specified
                        Zero.Dirs = True
                    END IF
                    IF INSTR(Switch.Line$, " ") THEN
                        IF File.Quotes = False THEN
                            Boot.List$ = "Unknown command line characters."
                            GOTO Boot.Error
                        END IF
                    END IF
                    Number.Dest.Dirs = Number.Dest.Dirs + 1
                    IF Number.Dest.Dirs > Max.Dest.Dirs THEN
                        Max.Dest.Dirs = Max.Dest.Dirs + 10
                        REDIM _PRESERVE Destinate.Directory(1 TO Max.Dest.Dirs) AS STRING * 260
                    END IF
                    ' check /d w/o parameters specified
                    IF Number.Dest.Dirs > 1 THEN
                        IF Zero.Dirs THEN
                            Boot.List$ = "Combined specified and nul directory not allowed."
                            GOTO Boot.Error
                        END IF
                    END IF
                    Destinate.Directory(Number.Dest.Dirs) = Switch.Line$
                CASE "G" ' netpath
                    Var = LastSwitch(Count)
                    IF LEN(Switch.Line$) = False THEN
                        ' check /g w/o parameters specified
                        IF Zero.Nets THEN
                            Boot.List$ = "Multiple nul netpaths not allowed."
                            GOTO Boot.Error
                        END IF
                        ' set /g w/o parameters specified
                        Zero.Nets = True
                    END IF
                    IF INSTR(Switch.Line$, " ") THEN
                        IF File.Quotes = False THEN
                            Boot.List$ = "Unknown command line characters."
                            GOTO Boot.Error
                        END IF
                    END IF
                    Number.Dest.Nets = Number.Dest.Nets + 1
                    IF Number.Dest.Nets > Max.Dest.Nets THEN
                        Max.Dest.Nets = Max.Dest.Nets + 10
                        REDIM _PRESERVE Destinate.Netpaths(1 TO Max.Dest.Nets) AS STRING * 260
                    END IF
                    ' check /g w/o parameters specified
                    IF Number.Dest.Nets > 1 THEN
                        IF Zero.Nets THEN
                            Boot.List$ = "Combined specified and nul netpaths not allowed."
                            GOTO Boot.Error
                        END IF
                    END IF
                    IF LEFT$(Switch.Line$, 2) <> "\\" THEN
                        IF Zero.Nets = False THEN
                            Boot.List$ = "Unknown netpath characters."
                            GOTO Boot.Error
                        END IF
                    END IF
                    Destinate.Netpaths(Number.Dest.Nets) = Switch.Line$
                CASE ELSE ' unknown switch
                    EXIT DO
            END SELECT
            Command.Line = LEFT$(Command.Line, Count - 1)
            Command.Line = RTRIM$(Command.Line)
            EXIT FOR
        END IF
    NEXT
LOOP

' check trailing command line.
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        Boot.List$ = "Unknown command line characters."
        GOTO Boot.Error
    END IF
END IF

' check zero dirs specified
IF Number.Dest.Nets > 1 THEN
    IF Zero.Dirs THEN
        Boot.List$ = "Nul directory and multiple netpaths not allowed."
        GOTO Boot.Error
    END IF
END IF

' check zero nets specified
IF Number.Dest.Nets > 1 THEN
    IF Zero.Nets THEN
        Boot.List$ = "Nul netpath and multiple netpaths not allowed."
        GOTO Boot.Error
    END IF
END IF

' reset temporary drive.
IF Temp.Drive.Dir = Nul THEN
    Temp$ = ENVIRON$("TEMP")
    IF LEN(Temp$) THEN
        Temp.Drive.Dir = Temp$
    ELSE
        Temp$ = ENVIRON$("TMP")
        IF LEN(Temp$) THEN
            Temp.Drive.Dir = Temp$
        END IF
    END IF
END IF

' recheck command line.
IF INSTR(Command.Line, "/") THEN
    Boot.List$ = "Unknown / switch."
    GOTO Boot.Error
END IF

' reset work variables.
Copy.Target = Nul
Continue.Searching = False
Nested.Levels = False
Quit.Searching = False
   
' store remaining command line.
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' leave arrays in reverse order.
IF Ignore.Sort THEN
    EXIT SUB
END IF

' sort arrays ascending.
IF Sort.Method = 1 THEN
    IF Number.Dest.Files > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Files
            FOR Var2 = Var1 + 1 TO Number.Dest.Files
                IF Destinate.Filename(Var1) > Destinate.Filename(Var2) THEN
                    SWAP Destinate.Filename(Var1), Destinate.Filename(Var2)
                END IF
            NEXT
        NEXT
    END IF
    IF Number.Dest.Dirs > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Dirs
            FOR Var2 = Var1 + 1 TO Number.Dest.Dirs
                IF Destinate.Directory(Var1) > Destinate.Directory(Var2) THEN
                    SWAP Destinate.Directory(Var1), Destinate.Directory(Var2)
                END IF
            NEXT
        NEXT
    END IF
    IF Number.Dest.Nets > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Nets
            FOR Var2 = Var1 + 1 TO Number.Dest.Nets
                IF Destinate.Netpaths(Var1) > Destinate.Netpaths(Var2) THEN
                    SWAP Destinate.Netpaths(Var1), Destinate.Netpaths(Var2)
                END IF
            NEXT
        NEXT
    END IF
    EXIT SUB
END IF

' sort arrays descending.
IF Sort.Method = 2 THEN
    IF Number.Dest.Files > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Files
            FOR Var2 = Var1 + 1 TO Number.Dest.Files
                IF Destinate.Filename(Var1) < Destinate.Filename(Var2) THEN
                    SWAP Destinate.Filename(Var1), Destinate.Filename(Var2)
                END IF
            NEXT
        NEXT
    END IF
    IF Number.Dest.Dirs > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Dirs
            FOR Var2 = Var1 + 1 TO Number.Dest.Dirs
                IF Destinate.Directory(Var1) < Destinate.Directory(Var2) THEN
                    SWAP Destinate.Directory(Var1), Destinate.Directory(Var2)
                END IF
            NEXT
        NEXT
    END IF
    IF Number.Dest.Nets > 1 THEN
        FOR Var1 = 1 TO Number.Dest.Nets
            FOR Var2 = Var1 + 1 TO Number.Dest.Nets
                IF Destinate.Netpaths(Var1) < Destinate.Netpaths(Var2) THEN
                    SWAP Destinate.Netpaths(Var1), Destinate.Netpaths(Var2)
                END IF
            NEXT
        NEXT
    END IF
    EXIT SUB
END IF

' restore arrays to original order.
IF Number.Dest.Files > 1 THEN
    Last.Array = Number.Dest.Files
    Pivot.Array = INT(Last.Array / 2)
    FOR Var = 1 TO Pivot.Array
        SWAP Destinate.Filename(Var), Destinate.Filename(Last.Array)
        Last.Array = Last.Array - 1
    NEXT
END IF
IF Number.Dest.Dirs > 1 THEN
    Last.Array = Number.Dest.Dirs
    Pivot.Array = INT(Last.Array / 2)
    FOR Var = 1 TO Pivot.Array
        SWAP Destinate.Directory(Var), Destinate.Directory(Last.Array)
        Last.Array = Last.Array - 1
    NEXT
END IF
IF Number.Dest.Nets > 1 THEN
    Last.Array = Number.Dest.Nets
    Pivot.Array = INT(Last.Array / 2)
    FOR Var = 1 TO Pivot.Array
        SWAP Destinate.Netpaths(Var), Destinate.Netpaths(Last.Array)
        Last.Array = Last.Array - 1
    NEXT
END IF
EXIT SUB

Boot.Error:
IF Display.Header = 0 THEN
    COLOR White, Black
    PRINT "COPY " + CopyitVersion$ + " " + CopyitRelease$ + ": File copy utility; "
END IF
COLOR Yellow, Black
PRINT Boot.List$
PRINT "Type COPY /? for help. Or edit COPYIT.CFG file."
COLOR Plain, Black
VarX = 2
END SUB

' subroutines searchs for and parses config file.
SUB Copyit_ReadConfig (Var$, File.Found)
VarX$ = Var$
File.Found = False
IF LEFT$(VarX$, 1) = Quote THEN
    VarX$ = MID$(VarX$, 2)
END IF
IF RIGHT$(VarX$, 1) = Quote THEN
    VarX$ = LEFT$(VarX$, LEN(VarX$) - 1)
END IF
IF RIGHT$(VarX$, 1) <> "\" THEN
    VarX$ = VarX$ + "\"
END IF
VarX$ = VarX$ + "COPYIT.CFG"
IF _FILEEXISTS(VarX$) THEN
    X = FREEFILE
    CLOSE #X
    DataError = 0
    Error.Flag = -1
    OPEN VarX$ FOR INPUT AS #X
    Error.Flag = 0
    IF DataError THEN
        EXIT SUB
    END IF
    File.Found = True
    DO UNTIL EOF(X)
        LINE INPUT #X, Param$
        Var2$ = LTRIM$(RTRIM$(Param$))
        IF Var2$ <> Nul THEN
            IF LEFT$(Var2$, 1) <> ";" THEN
                Command.Line = Command.Line + Param$
            END IF
        END IF
    LOOP
    CLOSE #X
END IF
END SUB

' subroutine converts string to numeric value
SUB GetNumeric (Value#)
Value# = DFalse
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Value# = Value# * 10# + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT SUB
    END IF
LOOP
END SUB

' display program usage
SUB Copyit_BootUsage
Display.Page = 1
DO
    SELECT CASE Display.Page
        CASE 1
            GOSUB Display.Page.One
        CASE 2
            GOSUB Display.Page.Two
        CASE 3
            GOSUB Display.Page.Three
        CASE 4
            GOSUB Display.Page.Four
        CASE 5
            GOSUB Display.Page.Five
        CASE 6
            GOSUB Display.Page.Six
        CASE 7
            GOSUB Display.Page.Seven
        CASE 8
            GOSUB Display.Page.Eight
        CASE 9
            GOSUB Display.Page.Nine
        CASE 10
            GOSUB Display.Page.Ten
        CASE 11
            GOSUB Display.Page.Eleven
    END SELECT
    Prompt2$ = "Press 1, 2, 3, 4, 5, 6, a, b, n, s, ?, or q to quit:"
    CALL MorePrompt2(Prompt2$, "123456abns?q", Outpt$, "q")
    SELECT CASE Outpt$
        CASE "1"
            Display.Page = 1
        CASE "2"
            Display.Page = 2
        CASE "3"
            Display.Page = 3
        CASE "4"
            Display.Page = 4
        CASE "5"
            Display.Page = 5
        CASE "6"
            Display.Page = 6
        CASE "a"
            Display.Page = 7
        CASE "b"
            Display.Page = 8
        CASE "n"
            Display.Page = 9
        CASE "s"
            Display.Page = 10
        CASE "?"
            Display.Page = 11
        CASE "q"
            EXIT SUB
    END SELECT
LOOP
EXIT SUB

Display.Page.One:
GOSUB Display.Header
PRINT "Where:"
PRINT "   <filelist> is multiple source files to copy and are:"
PRINT "       [@][d:][\path\][filename.ext]"
PRINT "     or with netpath specified \\server\share\"
PRINT "       @ prefix specifies filename containing filelist to copy"
PRINT "   (destination switches):"
PRINT "     /d[d:][\pathname\]  is destination directory"
PRINT "     /f[filename.ext]    is destination filename"
PRINT "     /g[\\server\share\] is destination netpath"
PRINT "   (directory/file switches):"
PRINT "     /a  synchronize dest. files"
PRINT "     /b1 copy directory structure w\o rename"
PRINT "     /b2 copy directory structure w\ rename"
PRINT "     /b3 only copy directories with files found"
PRINT "     /i  delete files after copy"
PRINT "     /k  delete directory after copy"
PRINT "     /n  don't prompt to delete file after copy"
PRINT "     /q  don't prompt to delete directory after copy"
RETURN

Display.Page.Two:
GOSUB Display.Header
PRINT "Where:"
PRINT "   (standard copying switches):"
PRINT "     /j  force environment variable to lowercase"
PRINT "     /m  don't overwrite dest. file"
PRINT "     /p  display command line"
PRINT "     /r  recurse directories"
PRINT "     /v  append to dest. file"
PRINT "     /v1 append/resume to dest. file"
PRINT "     /v2### copy from byte position"
PRINT "     /w  don't prompt before overwrite"
PRINT "   (display switches):"
PRINT "     /c  continuous display         /x1 don't display dest. filename"
PRINT "     /c1  continuous output display /x2 don't prepend display type"
PRINT "     /c2  skip file overwrite       /x3 don't display override display"
PRINT "     /h  don't display drive letter /x4 don't display any filenames"
PRINT "     /o1 display wide list          /y  display files in lowercase"
PRINT "     /o2 ambiguate filename         /z  suppress error messages"
PRINT "     /s  don't display pathname"
RETURN

Display.Page.Three:
GOSUB Display.Header
PRINT "Where:"
PRINT "   (filelist switches):"
PRINT "     /@  enable filelists"
PRINT "     /#  disable filelist prompts"
PRINT "     /@2 ignore pattern matching for filelist filenames"
PRINT "   (copy filter switches):"
PRINT "     /0  don't copy zero-length files"
PRINT "     /1  copy file to eof"
PRINT "     /2xxx-xxx  convert ascii filter (fast)"
PRINT "     /3xxx  strip ascii values (slow)"
PRINT "     /4xxx nested directories"
PRINT "   (file override switches):"
PRINT "     /5mm/dd/yyyy  override destination file date"
PRINT "     /6hh:mm:ss    override destination file time"
PRINT "     /7  creation date/time"
PRINT "     /8  last access date/time"
PRINT "     /9  last modified (write) date/time"
RETURN

Display.Page.Four:
GOSUB Display.Header
PRINT "Where:"
PRINT "   (file override ranges):"
PRINT "     /e  search date in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "     /t  search time in form hh:mm:ss-hh:mm:ss"
PRINT "     /u  search file size in form xxx-xxx in kilobytes"
PRINT "   (compare switches used with /e, /t):"
PRINT "     /s1  compare source creation date/time"
PRINT "     /s2  compare source last access date/time"
PRINT "     /s3  compare source last modified date/time"
PRINT "   (date range switches):"
PRINT "     /e1:mm/dd/yyyy-mm/dd/yyyy  pattern match date range"
PRINT "     /e2:" + Quote + "filename.ext" + Quote + "  pattern match filename, include:"
PRINT "       MM, DD, YY, or YYYY, ?, to match with /e1:"
PRINT "   (file attributes switches):"
PRINT "     [prefix1][prefix2][ahosx]"
PRINT "       prefix1: / copy files, // reset dest. file, /// reset source file;"
PRINT "       prefix2: x  with\clear, z  without\set;"
PRINT "         attribute: a  archive, h  hidden, o  read-only, s  system, x  none"
RETURN

Display.Page.Five:
GOSUB Display.Header
PRINT "Where:"
PRINT "   (additional display switches):"
PRINT "     /l1  display file creation\last access\last modified date/time."
PRINT "     /l2  display /l1 plus filesize."
PRINT "     /l3  display /l2 plus file attributes."
PRINT "     /l4  display /l3 plus bytes counted\copied, directories deleted."
PRINT "     /l   same as /l4"
PRINT "   (extended display switches):"
PRINT "     /*  display progress bar  /!  display percent copied"
PRINT "     /-  disable copying dots  /+  display transfer rate"
PRINT "   (debug switches):"
PRINT "     /=<x>  set /= and more prompt to <x> seconds"
PRINT "     /=  enable extended error messages  /:  ignore skipped files"
PRINT "     /.  disable reading configure file"
PRINT "   (exclude file list switch):"
PRINT "     /(<filename.ext>)"
PRINT "       Excludes files containing ? and * characters when"
PRINT "       enclosed in parenthesis. Multiple exclusions allowed."
RETURN

Display.Page.Six:
GOSUB Display.Header
PRINT "Where:"
PRINT "   (DOS command switch): (use /_ for /K in command call);"
PRINT "     /[<command>]  with command replacement parameters:"
PRINT "       %1 = d:, %2 = d:\, %3 = d:\pathname, %4 = d:\pathname\"
PRINT "       %5 = d:\pathname\filename.ext, %6 = \pathname, %7 = \pathname\"
PRINT "       %8 = \pathname\filename.ext, %9 = filename.ext, %a = filename"
PRINT "       %b = .ext, %c = ext, //1 = >, //2 = <, //3 = |, //4 = %, //5 = ]"
PRINT "   (exit errorlevel codes):"
PRINT "     0 = files copied successfully   2 = no files were copied"
PRINT "     4 = disk full/disk not ready    8 = copy interrupted by user"
RETURN

Display.Page.Seven:
COLOR White, Black
PRINT "Alphabetic list of COPY command line switches (page 'a'):"
COLOR Yellow, Black
PRINT "/a - synchronize dest. files           /p - display command line"
PRINT "/b - copy directory structure          /q - don't prompt to delete directory"
PRINT "/c - continuous list                   /r - recursive directory search"
PRINT "/c1 - continuous output display        /s - don't display pathname"
PRINT "/c2 - skip file overwrite              /t - time range hh:mm:ss-hh:mm:ss"
PRINT "/d - destination directory [d:][\path] /u - file size range in KB"
PRINT "/e - date range mm/dd/yyyy-mm/dd/yyyy  /v - append to dest. file"
PRINT "/f - destination file [filename.ext]   /v1 - append/resume to dest. file"
PRINT "/g - specifies dest. \\server\share\   /v2 - copy from byte position"
PRINT "/h - don't display drive letter        /w - don't prompt before overwrite"
PRINT "/i - delete files after copy           /x1 - don't display dest. filename"
PRINT "/j - force environment to lowercase    /x2 - don't prepend display type"
PRINT "/k - delete directory after copy       /x3 - don't display override display"
PRINT "/lx - added info switches x=1-4        /x4 - don't display any filenames"
PRINT "/m - don't overwrite dest. file        /y - display files in lowercase"
PRINT "/n - don't prompt to delete file       /z - don't display errors"
PRINT "/o1 - display wide list"
PRINT "/o2 - ambiguate filename"
RETURN

Display.Page.Eight:
COLOR White, Black
PRINT "Bit copying list of COPY command line switches (page 'b'):"
COLOR Yellow, Black
PRINT "Source copy bits:"
PRINT "/xa - with archive bit                  /xh - with hidden bit"
PRINT "/xo - with read-only bit                /xs - with system bit"
PRINT "/xx - with no bits (without all bits)"
PRINT "/za - without archive bit               /zh - without hidden bit"
PRINT "/zo - without read-only bit             /zs - without system bit"
PRINT "/zx - without no bits (with only all bits)"
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
COLOR Yellow, Black
PRINT "Destination reset bits:"
PRINT "//xa - set dest. archive bit            //xh - set dest. hidden bit"
PRINT "//xo - set dest. read-only bit          //xs - set dest. system bit"
PRINT "//xx - set all dest. bits"
PRINT "//za - clear dest. archive bit          //zh - clear dest. hidden bit"
PRINT "//zo - clear dest. read-only bit        //zs - clear dest. system bit"
PRINT "//zx - clear all dest. bits"
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
COLOR Yellow, Black
PRINT "Source reset bits:"
PRINT "///xa - set source archive bit          ///xh - set source hidden bit"
PRINT "///xo - set source read-only bit        ///xs - set source system bit"
PRINT "///xx - set all source bits"
PRINT "///za - clear source archive bit        ///zh - clear source hidden bit"
PRINT "///zo - clear source read-only bit      ///zs - clear source system bit"
PRINT "///zx - clear all source bits"
COLOR White, Black
PRINT "Special switches:"
COLOR Yellow, Black
PRINT "/xc - copy only compressed files,  /xe  copy only encrypted files."
PRINT "/zc - don't copy compressed files,  /ze  don't copy encrypted files."
CALL MorePrompt("-more-", " " + CHR$(13), Outpt$)
RETURN

Display.Page.Nine:
COLOR White, Black
PRINT "Numeric list of COPY command line switches (page 'n'):"
COLOR Yellow, Black
PRINT "/0 - don't copy zero-length files"
PRINT "/1 - copy ascii file to eof"
PRINT "/2 - convert ascii filter xxx-xxx (fast)"
PRINT "/3 - strip ascii filter (slow)"
PRINT "/4 - override nested directories"
PRINT "/5 - specify dest. file date mm/dd/yyyy override"
PRINT "/6 - specify dest. file time hh:mm:ss override"
PRINT "/7 - file creation date/time override"
PRINT "/8 - file last access date/time override"
PRINT "/9 - file last modified date/time override"
PRINT "/-1 - ignore file sort during command line parsing"
PRINT "/-2 - sort command line files ascending"
PRINT "/-3 - sort command line files descending"
PRINT "/-4 - display extended time for /L"
PRINT "/-5 - skips all duplicate files from being copied"
RETURN

Display.Page.Ten:
COLOR White, Black
PRINT "List of extended /a synchronize numeric switches (page 's'):"
COLOR Yellow, Black
PRINT "/a1  copy source files with sizes less than destination file sizes."
PRINT "/a2  copy source files with sizes greater than destination file sizes."
PRINT "/a3  copy source files with sizes equal to destination file sizes."
PRINT "/a4  copy source files with sizes not equal to destination file sizes."
PRINT "/a5  copy source files with dates earlier than destination file dates."
PRINT "/a6  copy source files with dates greater than destination file dates."
PRINT "/a7  copy source files with dates equal to destination file dates."
PRINT "/a8  copy source files with dates not equal to destination file dates."
PRINT "/a9  copy source files with times earlier than destination file times."
PRINT "/aa  copy source files with times greater than destination file times."
PRINT "/ab  copy source files with times equal to destination file times."
PRINT "/ac  copy source files with times not equal to destination file times."
COLOR White, Black
PRINT "Synchronize switches for date/time comparing: (used with /a)"
COLOR Yellow, Black
PRINT "/z1  compare source creation date/time"
PRINT "/z2  compare source last access date/time"
PRINT "/z3  compare source last modified date/time"
RETURN

Display.Page.Eleven:
COLOR White, Black
PRINT "List of reserved characters:"
COLOR Yellow, Black
PRINT "In DOS the 8.3 filename cannot contain any of the following characters:"
COLOR White, Black
PRINT "\/:*?<>|" + Quote
COLOR Yellow, Black
PRINT "In Windows command line the following characters are reserved and should be"
PRINT "enclosed in quotes when used in filenames on the command line:"
COLOR White, Black
PRINT "<space>"
PRINT "&()[]{}^=;!'+,`~"
RETURN

' make header.
Display.Header:
COLOR White, Black
PRINT "COPY " + CopyitVersion$ + " " + CopyitRelease$ + ": File copy utility help page";
PRINT STR$(Display.Page) + ";"
COLOR Yellow, Black
PRINT "Usage: COPY <filelist>[/dfg][<switches>]"
RETURN
END SUB

' routine prompts user for an option
SUB MorePrompt2 (Input.String$, Input.Mask$, Output.String$, Default$)
CALL Copyit_ClearDisplay(0)
Rate.Mode = False
Start.Timer! = TIMER
Input.Char$ = Nul
COLOR White, Black
PRINT Input.String$ + " ";
DO
    LOCATE , , 1
    DO
        _LIMIT 50
        Input.Char$ = INKEYx$
        IF LEN(Input.Char$) THEN
            EXIT DO
        END IF

        ' exits with default prompt in debug mode after timer.
        IF Debug.Mode THEN
            IF DebugDelay! = 0! THEN
                Input.Char$ = Default$
                EXIT DO
            END IF
            IF DebugDelay! > 0! THEN
                Elapsed.Time! = TIMER - Start.Timer!
                IF Elapsed.Time! < SFalse THEN
                    Elapsed.Time! = Elapsed.Time! + 86400!
                END IF
                IF Elapsed.Time! >= DebugDelay! THEN
                    Input.Char$ = Default$
                    EXIT DO
                END IF
            END IF
        END IF
    LOOP

    ' check keyboard input.
    Input.Char$ = LCASE$(Input.Char$)
    IF INSTR(Input.Mask$, Input.Char$) THEN
        IF Input.Char$ = CHR$(13) THEN
            PRINT
        ELSE
            PRINT Input.Char$
        END IF
        Output.String$ = Input.Char$
        EXIT DO
    END IF
LOOP
x = ClearKeyboard
END SUB

' displays function error
SUB Copyit_DisplayError (Temp$)
DIM hfind AS _OFFSET

' check display errors flag.
IF Display.Errors = False THEN
    ' clear display progress
    CALL Copyit_ClearDisplay(0)

    ' display error
    IF LCASE$(StderrFile$) = "scrn:" THEN
        COLOR Red, Black
        PRINT #StderrHandle, Temp$
    ELSE
        PRINT #StderrHandle, DATE$ + " " + TIME$ + " " + Temp$
    END IF
END IF

' check debug flag
IF Debug.ModeX = 0 THEN
    EXIT SUB
END IF

' test open file for append
f$ = StartDir$ + ErrorLogFile$ + CHR$(0)
hfind = CreateFile(f$, &H180, 0, 0, 4, 0, 0)
IF hfind <> INVALID_HANDLE_VALUE THEN
    x = CloseHandle(hfind)

    ' write error to file
    V = FREEFILE
    DataError = 0
    Error.Flag = -1
    OPEN StartDir$ + ErrorLogFile$ FOR APPEND AS #V
    Error.Flag = 0
    IF DataError THEN
        EXIT SUB
    END IF
    PRINT #V, DATE$ + " " + TIME$ + " " + Temp$
    CLOSE #V
END IF
END SUB

' subroutine to check for copying directory structure onto itself
SUB Copyit_DirectoryCopying (Work.Net$)
' reset return flag.
Directory.Exists = False

' read source drive.
From.Drive = ASC(UCASE$(Drive.Search)) - 64

' read source directory.
Copy.From$ = RTRIM$(Dir.Copy.Name)
Copy.From2$ = Copy.From$
Work.Net3$ = RTRIM$(Source.Net)
Work.Net4$ = RTRIM$(Default.Net)
IF LEFT$(Copy.From$, 1) = "\" THEN
    IF Work.Net3$ <> Nul THEN
        Copy.From$ = RTRIM$(Work.Net3$) + Copy.From$
    ELSE
        IF Work.Net4$ <> Nul THEN
            Copy.From$ = RTRIM$(Work.Net4$) + Drive.Search + Copy.From$
        END IF
    END IF
ELSE
    IF Work.Net3$ <> Nul THEN
        Copy.From$ = RTRIM$(Work.Net3$) + "\" + Copy.From$
    ELSE
        IF Work.Net4$ <> Nul THEN
            Copy.From$ = RTRIM$(Work.Net4$) + Drive.Search + "\" + Copy.From$
        END IF
    END IF
END IF

' read target drive.
Copy.To$ = Copy.Target
IF MID$(Copy.To$, 2, 1) = ":" THEN
    Drive.Number = ASC(UCASE$(LEFT$(Copy.To$, 1))) - 64
    Copy.To$ = MID$(Copy.To$, 3)
ELSE
    Drive.Number = ASC(UCASE$(Drive.Search)) - 64
END IF

' read target directory.
IF LEFT$(Copy.To$, 1) <> "\" THEN
    Copy.To$ = Dest.Dir
    IF MID$(Copy.To$, 2, 1) = ":" THEN
        Drive.Number = ASC(UCASE$(LEFT$(Copy.To$, 1))) - 64
        Copy.To$ = MID$(Copy.To$, 3)
    ELSE
        Drive.Number = ASC(UCASE$(Drive.Search)) - 64
    END IF
    IF Work.Net$ = Nul THEN
        Copy.To$ = Default.Dir + "\" + Copy.To$
    END IF
END IF
IF Work.Net$ <> Nul THEN
    IF LEFT$(Copy.To$, 1) = "\" THEN
        Copy.To$ = Work.Net$ + Copy.To$
    ELSE
        Copy.To$ = Work.Net$ + "\" + Copy.To$
    END IF
END IF
Copy.From$ = LCASE$(Copy.From$)
Copy.To$ = LCASE$(Copy.To$)

' compare drives.
IF From.Drive = Drive.Number THEN
    ' check directories.
    IF Copy.From$ = Copy.To$ THEN
        Directory.Exists = -1
        EXIT SUB
    END IF

    ' check recursing directories.
    IF Recurse.Directories THEN
        IF Copy.From2$ <> "\" THEN
            ' check recursive copying.
            IF Copy.From$ = LEFT$(Copy.To$, LEN(Copy.From$)) THEN
                Directory.Exists = -2
            END IF
        END IF
    END IF
END IF
END SUB

' subroutine constructs base filenames
SUB Copyit_MakeFile (Work.Net$, From.File$, To.File$)
' make destination filename.
IF MID$(To.File$, 2, 1) = ":" THEN
    To.File$ = MID$(To.File$, 3)
END IF
IF LEFT$(To.File$, 1) <> "\" THEN
    IF Work.Net$ = Nul THEN
        IF Default.Dir = "\" THEN
            To.File$ = Default.Dir + To.File$
        ELSE
            To.File$ = Default.Dir + "\" + To.File$
        END IF
    END IF
    IF MID$(To.File$, 2, 1) = ":" THEN
        To.File$ = MID$(To.File$, 3)
    END IF
END IF
IF MID$(Dest.Dir, 2, 1) = ":" THEN
    Drive.Number = ASC(UCASE$(LEFT$(Dest.Dir, 1)))
ELSE
    Drive.Number = ASC(UCASE$(Drive.Search))
END IF
IF Work.Net$ = Nul THEN
    IF LEFT$(To.File$, 2) <> "\\" THEN
        To.File$ = CHR$(Drive.Number) + ":" + To.File$
    END IF
ELSE
    IF LEFT$(To.File$, 1) = "\" THEN
        To.File$ = Work.Net$ + To.File$
    ELSE
        To.File$ = Work.Net$ + "\" + To.File$
    END IF
END IF

' make source filename.
Work.Net2$ = RTRIM$(Source.Net)
IF MID$(From.File$, 2, 1) = ":" THEN
    From.File$ = MID$(From.File$, 3)
END IF
IF LEFT$(From.File$, 1) <> "\" THEN
    IF Work.Net2$ = Nul THEN
        IF Default.Dir = "\" THEN
            From.File$ = Default.Dir + From.File$
        ELSE
            From.File$ = Default.Dir + "\" + From.File$
        END IF
    END IF
    IF MID$(From.File$, 2, 1) = ":" THEN
        From.File$ = MID$(From.File$, 3)
    END IF
END IF
IF Work.Net2$ = Nul THEN
    IF LEFT$(From.File$, 2) <> "\\" THEN
        From.File$ = Drive.Search + ":" + From.File$
    END IF
ELSE
    IF LEFT$(From.File$, 1) = "\" THEN
        From.File$ = Work.Net2$ + From.File$
    ELSE
        From.File$ = Work.Net2$ + "\" + From.File$
    END IF
END IF

' check filenames.
IF Display.Lowercase THEN
    From.File$ = LCASE$(From.File$)
    To.File$ = LCASE$(To.File$)
END IF
END SUB

' subroutine constructs base netpath
SUB Copyit_MakeFile2 (From.File$)
Work.Net$ = RTRIM$(Source.Net)
Work.Net2$ = RTRIM$(Default.Net)
IF MID$(From.File$, 2, 1) = ":" THEN
    From.File$ = MID$(From.File$, 3)
END IF
IF LEFT$(From.File$, 1) <> "\" THEN
    IF Work.Net$ = Nul AND Work.Net2$ = Nul THEN
        IF Default.Dir = "\" THEN
            From.File$ = Default.Dir + From.File$
        ELSE
            From.File$ = Default.Dir + "\" + From.File$
        END IF
    END IF
    IF MID$(From.File$, 2, 1) = ":" THEN
        From.File$ = MID$(From.File$, 3)
    END IF
END IF
IF Work.Net$ = Nul AND Work.Net2$ = Nul THEN
    From.File$ = Drive.Search + ":" + From.File$
ELSE
    IF LEFT$(From.File$, 1) = "\" THEN
        IF Work.Net$ <> Nul THEN
            From.File$ = RTRIM$(Work.Net$) + From.File$
        ELSE
            From.File$ = RTRIM$(Work.Net2$) + Drive.Search + From.File$
        END IF
    ELSE
        IF Work.Net$ <> Nul THEN
            From.File$ = RTRIM$(Work.Net$) + "\" + From.File$
        ELSE
            From.File$ = RTRIM$(Work.Net2$) + Drive.Search + "\" + From.File$
        END IF
    END IF
END IF
From.File$ = LCASE$(From.File$)
END SUB

' subroutine to create source filename from command line
SUB Copyit_MakeFilename
' store redirected input.
Standard.Input$ = RTRIM$(Standard.Input$)
Standard.Input$ = LTRIM$(Standard.Input$)
IF LEFT$(Standard.Input$, 1) = Quote THEN
    Standard.Input$ = MID$(Standard.Input$, 2)
END IF
IF RIGHT$(Standard.Input$, 1) = Quote THEN
    Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
END IF

' store entire command.
IF LEFT$(Command.Line, 1) = Quote THEN
    Imbedded = INSTR(2, Command.Line, Quote)
    IF Imbedded THEN
        Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
        Command.Line = MID$(Command.Line, Imbedded + 1)
    ELSE
        Command.Work = Standard.Input$ + Command.Line
        Command.Line = Nul
    END IF
ELSE
    Imbedded = INSTR(Command.Line, " ")
    IF Imbedded THEN
        Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
        Command.Line = MID$(Command.Line, Imbedded + 1)
    ELSE
        Command.Work = Standard.Input$ + Command.Line
        Command.Line = Nul
    END IF
END IF
Command.Line = LTRIM$(Command.Line)
Command.Line = RTRIM$(Command.Line)
END SUB

' routine constructs file date string
FUNCTION Copyit_MakeDate$ (Var)
SELECT CASE Var
    CASE 1
        x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    CASE 2
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
    CASE 3
        x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
END SELECT
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))

IF Extended.Time THEN
    HourTemp! = SysTime.wHour
    MinuteTemp! = SysTime.wMinute
    SecondsTemp! = SysTime.wSecond
    File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
    File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
    File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
    File.Date$ = File.Date$ + " " + File.Time$
END IF

Copyit_MakeDate$ = File.Date$
END FUNCTION

' routine checks file size range
SUB Copyit_CheckSizeRange (Range)
' reset return flag.
Size.Flag = False

' check for zero-byte files.
IF Copy.Zero THEN
    IF File.Size = DFalse THEN
        Size.Flag = True
        EXIT SUB
    END IF
END IF

' check file size range.
IF Search.File.Size = False THEN
    EXIT SUB
END IF

' files of zero-byte are copied.
IF Search.Size.From = DFalse AND Search.Size.To = DFalse THEN
    IF File.Size <> DFalse THEN
        Size.Flag = True
    END IF
    EXIT SUB
END IF

' check /u10-
IF Search.Size.From > DFalse THEN
    IF Search.Size.To = DFalse THEN
        IF File.Size < Search.Size.From THEN
            Size.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /u-10
IF Search.Size.From = DFalse THEN
    IF Search.Size.To > DFalse THEN
        IF File.Size > Search.Size.To THEN
            Size.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /u10-20
IF Search.Size.From <= Search.Size.To THEN
    IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
        Size.Flag = True
        EXIT SUB
    END IF
END IF

' check /u20-10
IF Search.Size.From > Search.Size.To THEN
    IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
        Size.Flag = True
        EXIT SUB
    END IF
END IF
END SUB

' routine checks date/time ranges
'   checks creation, last access, last modified date/time.
SUB Copyit_CheckDateRange (Range)
' reset return flag.
Range.Flag = False

' check date range.
IF Search.Date THEN
    IF Compare.Create.Date THEN ' /s1
        IF Search.From.Date > False OR Search.To.Date > False THEN
            ' /e01/01/1980-
            IF Search.From.Date > False THEN
                IF Search.To.Date = False THEN
                    IF Synch.Work.Create.Date < Search.From.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e-12/31/2010
            IF Search.From.Date = False THEN
                IF Search.To.Date > False THEN
                    IF Synch.Work.Create.Date > Search.To.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e01/01/1980-12/31/2010
            IF Search.From.Date <= Search.To.Date THEN
                IF Synch.Work.Create.Date < Search.From.Date OR Synch.Work.Create.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /e12/31/2010-01/01/1980
            IF Search.From.Date > Search.To.Date THEN
                IF Synch.Work.Create.Date < Search.From.Date AND Synch.Work.Create.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Compare.Access.Date THEN ' /s2
        IF Search.From.Date > False OR Search.To.Date > False THEN
            ' /e01/01/1980-
            IF Search.From.Date > False THEN
                IF Search.To.Date = False THEN
                    IF Synch.Work.Access.Date < Search.From.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e-12/31/2010
            IF Search.From.Date = False THEN
                IF Search.To.Date > False THEN
                    IF Synch.Work.Access.Date > Search.To.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e01/01/1980-12/31/2010
            IF Search.From.Date <= Search.To.Date THEN
                IF Synch.Work.Access.Date < Search.From.Date OR Synch.Work.Access.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /e12/31/2010-01/01/1980
            IF Search.From.Date > Search.To.Date THEN
                IF Synch.Work.Access.Date < Search.From.Date AND Synch.Work.Access.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Compare.Modify.Date THEN ' /s3
        IF Search.From.Date > False OR Search.To.Date > False THEN
            ' /e01/01/1980-
            IF Search.From.Date > False THEN
                IF Search.To.Date = False THEN
                    IF Synch.Work.Modify.Date < Search.From.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e-12/31/2010
            IF Search.From.Date = False THEN
                IF Search.To.Date > False THEN
                    IF Synch.Work.Modify.Date > Search.To.Date THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /e01/01/1980-12/31/2010
            IF Search.From.Date <= Search.To.Date THEN
                IF Synch.Work.Modify.Date < Search.From.Date OR Synch.Work.Modify.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /e12/31/2010-01/01/1980
            IF Search.From.Date > Search.To.Date THEN
                IF Synch.Work.Modify.Date < Search.From.Date AND Synch.Work.Modify.Date > Search.To.Date THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' check time range.
IF Search.Time THEN
    IF Compare.Create.Date THEN ' /s1
        IF Search.From.Time > False OR Search.To.Time > False THEN
            ' /t00:00:01-
            IF Search.From.Time > False THEN
                IF Search.To.Time = False THEN
                    IF Synch.Work.Create.Time < Search.From.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t-23:59:59
            IF Search.From.Time = False THEN
                IF Search.To.Time > False THEN
                    IF Synch.Work.Create.Time > Search.To.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t00:00:01-23:59:59
            IF Search.From.Time <= Search.To.Time THEN
                IF Synch.Work.Create.Time < Search.From.Time OR Synch.Work.Create.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /t23:59:59-00:00:01
            IF Search.From.Time > Search.To.Time THEN
                IF Synch.Work.Create.Time < Search.From.Time AND Synch.Work.Create.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Compare.Access.Date THEN ' /s2
        IF Search.From.Time > False OR Search.To.Time > False THEN
            ' /t00:00:01-
            IF Search.From.Time > False THEN
                IF Search.To.Time = False THEN
                    IF Synch.Work.Access.Time < Search.From.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t-23:59:59
            IF Search.From.Time = False THEN
                IF Search.To.Time > False THEN
                    IF Synch.Work.Access.Time > Search.To.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t00:00:01-23:59:59
            IF Search.From.Time <= Search.To.Time THEN
                IF Synch.Work.Access.Time < Search.From.Time OR Synch.Work.Access.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /t23:59:59-00:00:01
            IF Search.From.Time > Search.To.Time THEN
                IF Synch.Work.Access.Time < Search.From.Time AND Synch.Work.Access.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Compare.Modify.Date THEN ' /s3
        IF Search.From.Time > SFalse OR Search.To.Time > SFalse THEN
            ' /t00:00:01-
            IF Search.From.Time > SFalse THEN
                IF Search.To.Time = SFalse THEN
                    IF Synch.Work.Modify.Time < Search.From.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t-23:59:59
            IF Search.From.Time = SFalse THEN
                IF Search.To.Time > SFalse THEN
                    IF Synch.Work.Modify.Time > Search.To.Time THEN
                        Range.Flag = True
                        EXIT SUB
                    END IF
                END IF
            END IF
            ' /t00:00:01-23:59:59
            IF Search.From.Time <= Search.To.Time THEN
                IF Synch.Work.Modify.Time < Search.From.Time OR Synch.Work.Modify.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
            ' /t23:59:59-00:00:01
            IF Search.From.Time > Search.To.Time THEN
                IF Synch.Work.Modify.Time < Search.From.Time AND Synch.Work.Modify.Time > Search.To.Time THEN
                    Range.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF
END SUB

' routine verifies source - destination synchronized file size, date/time.
'   checks creation, last access, last modified date/time.
SUB Copyit_CheckSynched (Synched)
' reset return value.
Synched.Flag = False

' check /a
IF Synch.Files THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF File.Size = Dest.File.Size THEN
            IF Synch.Work.Create.Date = Synch.File.Create.Date THEN
                IF Synch.Work.Create.Time = Synch.File.Create.Time THEN
                    Synched.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF File.Size = Dest.File.Size THEN
            IF Synch.Work.Access.Date = Synch.File.Access.Date THEN
                IF Synch.Work.Access.Time = Synch.File.Access.Time THEN
                    Synched.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF File.Size = Dest.File.Size THEN
            IF Synch.Work.Modify.Date = Synch.File.Modify.Date THEN
                IF Synch.Work.Modify.Time = Synch.File.Modify.Time THEN
                    Synched.Flag = True
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' check /a1
IF Synch.Files1 THEN
    IF File.Size >= Dest.File.Size THEN
        Synched.Flag = True
        EXIT SUB
    END IF
END IF

' check /a2
IF Synch.Files2 THEN
    IF File.Size <= Dest.File.Size THEN
        Synched.Flag = True
        EXIT SUB
    END IF
END IF

' check /a3
IF Synch.Files3 THEN
    IF File.Size <> Dest.File.Size THEN
        Synched.Flag = True
        EXIT SUB
    END IF
END IF

' check /a4
IF Synch.Files4 THEN
    IF File.Size = Dest.File.Size THEN
        Synched.Flag = True
        EXIT SUB
    END IF
END IF

' check /a5
IF Synch.Files5 THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Date >= Synch.File.Create.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Date >= Synch.File.Access.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Date >= Synch.File.Modify.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /a6
IF Synch.Files6 THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Date <= Synch.File.Create.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Date <= Synch.File.Access.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Date <= Synch.File.Modify.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /a7
IF Synch.Files7 THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Date <> Synch.File.Create.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Date <> Synch.File.Access.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Date <> Synch.File.Modify.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /a8
IF Synch.Files8 THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Date = Synch.File.Create.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Date = Synch.File.Access.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Date = Synch.File.Modify.Date THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /a9
IF Synch.Files9 THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Time >= Synch.File.Create.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Time >= Synch.File.Access.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Time >= Synch.File.Modify.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /aa
IF Synch.FilesA THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Time <= Synch.File.Create.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Time <= Synch.File.Access.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Time <= Synch.File.Modify.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /ab
IF Synch.FilesB THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Time <> Synch.File.Create.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Time <> Synch.File.Access.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Time <> Synch.File.Modify.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF

' check /ac
IF Synch.FilesC THEN
    IF Synch.Type1 THEN ' creation date/time.
        IF Synch.Work.Create.Time = Synch.File.Create.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type2 THEN ' last access date/time.
        IF Synch.Work.Access.Time = Synch.File.Access.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
    IF Synch.Type3 THEN ' last modified date/time.
        IF Synch.Work.Modify.Time = Synch.File.Modify.Time THEN
            Synched.Flag = True
            EXIT SUB
        END IF
    END IF
END IF
END SUB

' routine checks filename
SUB Copyit_CheckFile (Var$, Check)
' store test flag
Check.Flag = True

' store input filename
Var2$ = UCASE$(Var$)
Var2$ = RTRIM$(Var2$)

' check skip files
IF Ignore.Skip = 0 THEN
    FOR Skip.File = 1 TO 10
        Var1$ = Skip.Filenames(Skip.File)
        Var1$ = UCASE$(Var1$)
        Var1$ = RTRIM$(Var1$)
        IF LEN(Var1$) THEN
            IF Var1$ = Var2$ THEN
                Check.Flag = False
                EXIT SUB
            END IF
        END IF
    NEXT
END IF

' check excluded files
FOR File.Number = 1 TO Number.Excluded
    Var1$ = Excluded.Files(File.Number)
    Var1$ = UCASE$(Var1$)
    Var1$ = RTRIM$(Var1$)
    IF LEN(Var1$) THEN
        CALL Copyit_CheckExcluded(Var1$, Var2$)
        IF Exclude.Flag THEN
            Check.Flag = False
            EXIT SUB
        END IF
    END IF
NEXT

' check pattern files
IF Check THEN
    IF Pattern.Match THEN
        EXIT SUB
    END IF
END IF
CALL Copyit_CheckPattern(Var2$)
IF Pattern.Flag = False THEN
    Check.Flag = False
END IF
END SUB

' routine compares occurrence of filename1$ in filename2$
'  with pattern matching.
SUB Copyit_CheckExcluded (Filename1$, Filename2$)
Exclude.Flag = True ' assume mask matches filename2.
Length1 = 1
Length2 = 1
DO
    ' global replacement.
    IF MID$(Filename1$, Length1, 1) = "*" THEN
        DO
            Length1 = Length1 + 1
            IF Length1 > LEN(Filename1$) THEN
                EXIT SUB
            END IF
            ' global replacement followed by exclusion character.
            ' searches remaining string until exclusion character found or not.
            IF MID$(Filename1$, Length1, 1) = "^" THEN
                Length1 = Length1 + 1
                Not.Include$ = MID$(Filename1$, Length1, 1)
                DO
                    IF Not.Include$ <> MID$(Filename2$, Length2, 1) THEN
                        Length2 = Length2 + 1
                    ELSE
                        Exclude.Flag = False
                        EXIT SUB
                    END IF
                    IF Length2 > LEN(Filename2$) THEN
                        EXIT SUB
                    END IF
                LOOP
            END IF
            ' global replacement followed by ? or another *
            ' skips to next character.
            IF MID$(Filename1$, Length1, 1) <> "*" THEN
                IF MID$(Filename1$, Length1, 1) <> "?" THEN
                    EXIT DO
                END IF
            END IF
        LOOP
        ' global replacement.
        ' searches for next matching character.
        DO
            IF MID$(Filename1$, Length1, 1) = MID$(Filename2$, Length2, 1) THEN
                EXIT DO
            ELSE
                Length2 = Length2 + 1
            END IF
            IF Length2 > LEN(Filename2$) THEN
                EXIT DO
            END IF
        LOOP
    ELSE
        ' character replacement.
        ' matches any next character.
        IF MID$(Filename1$, Length1, 1) = "?" THEN
            Length1 = Length1 + 1
            Length2 = Length2 + 1
        ELSE
            ' exclusion character.
            ' checks next character unmatched.
            IF MID$(Filename1$, Length1, 1) = "^" THEN
                Length1 = Length1 + 1
                Not.Include$ = MID$(Filename1$, Length1, 1)
                IF Not.Include$ <> MID$(Filename2$, Length2, 1) THEN
                    Length1 = Length1 + 1
                    Length2 = Length2 + 1
                ELSE
                    Exclude.Flag = False
                    EXIT DO
                END IF
            ELSE
                ' matches next character.
                IF MID$(Filename1$, Length1, 1) = MID$(Filename2$, Length2, 1) THEN
                    Length1 = Length1 + 1
                    Length2 = Length2 + 1
                ELSE
                    Exclude.Flag = False
                    EXIT DO
                END IF
                ' check string lengths.
                IF Length1 > LEN(Filename1$) THEN
                    IF Length2 <= LEN(Filename2$) THEN
                        Exclude.Flag = False
                    END IF
                    EXIT DO
                END IF
            END IF
        END IF
    END IF
LOOP
END SUB

' routine checks filename date pattern
SUB Copyit_CheckPattern (Filename2$)
' reset to match.
Pattern.Flag = True
' exit for no matching.
IF Pattern.Search = False THEN
    EXIT SUB
END IF
' compare exact filename lengths.
IF LEN(Filename2$) <> LEN(Pattern.Filename) THEN
    Pattern.Flag = False
    EXIT SUB
END IF
' scan pattern.
FOR Var = 1 TO LEN(Pattern.Filename)
    Var$ = MID$(Pattern.Filename, Var, 1)
    SELECT CASE Var$
        CASE "?" ' nul
            Var0$ = Nul
        CASE "M", "D", "Y" ' skip
            Var0$ = Nul
        CASE ELSE
            ' match exact filename character.
            IF MID$(Filename2$, Var, 1) <> Var$ THEN
                Pattern.Flag = False
                EXIT SUB
            END IF
    END SELECT
NEXT
' scan year pattern type 1.
Year1 = False
FOR Var = 1 TO LEN(Pattern.Filename)
    IF MID$(Pattern.Filename, Var, 4) = "YYYY" THEN
        Year1 = VAL(MID$(Filename2$, Var, 4))
        EXIT FOR
    END IF
NEXT
' scan year pattern type 2.
IF Year1 = False THEN
    FOR Var = 1 TO LEN(Pattern.Filename)
        IF MID$(Pattern.Filename, Var, 2) = "YY" THEN
            Year1 = VAL(MID$(Filename2$, Var, 2))
            EXIT FOR
        END IF
    NEXT
END IF
' scan month pattern.
Month1 = False
FOR Var = 1 TO LEN(Pattern.Filename)
    IF MID$(Pattern.Filename, Var, 2) = "MM" THEN
        Month1 = VAL(MID$(Filename2$, Var, 2))
        EXIT FOR
    END IF
NEXT
' scan day pattern.
Day1 = False
FOR Var = 1 TO LEN(Pattern.Filename)
    IF MID$(Pattern.Filename, Var, 2) = "DD" THEN
        Day1 = VAL(MID$(Filename2$, Var, 2))
        EXIT FOR
    END IF
NEXT

' store date patterns.
Day.From = VAL(MID$(Pattern.Search.From.Date, 4, 2))
Month.From = VAL(LEFT$(Pattern.Search.From.Date, 2))
Year.From = VAL(RIGHT$(Pattern.Search.From.Date, 4))
Day.To = VAL(MID$(Pattern.Search.To.Date, 4, 2))
Month.To = VAL(LEFT$(Pattern.Search.To.Date, 2))
Year.To = VAL(RIGHT$(Pattern.Search.To.Date, 4))

' match date patterns. checks all possible patterned combinations.
IF Day1 > False AND Month1 > False AND Year1 > False THEN ' DD-MM-YYYY
    ' for a complete date, serial function can be used.
    From.Serial# = DateSerial(Year.From, Month.From, Day.From)
    To.Serial# = DateSerial(Year.To, Month.To, Day.To)
    Now.Serial# = DateSerial(Year1, Month1, Day1)
    IF From.Serial# <= To.Serial# THEN
        IF Now.Serial# >= From.Serial# AND Now.Serial# <= To.Serial# THEN
            Pattern.Flag = True
        ELSE
            Pattern.Flag = False
        END IF
    ELSE
        Pattern.Flag = False
    END IF
ELSE
    ' for all remaining combinations, specific ranges must be checked.
    IF Day1 > False AND Month1 > False AND Year1 = False THEN ' DD-MM
        IF Month.From = Month.To THEN
            IF Month1 = Month.From THEN
                IF Day.From <= Day.To THEN
                    IF Day1 >= Day.From AND Day1 <= Day.To THEN
                        Pattern.Flag = True
                    ELSE
                        Pattern.Flag = False
                    END IF
                ELSE
                    Pattern.Flag = False
                END IF
            ELSE
                Pattern.Flag = False
            END IF
        ELSE
            IF Month.From < Month.To THEN
                IF Month1 >= Month.From AND Month1 <= Month.To THEN
                    IF Day.From > Day.To THEN
                        IF Day1 >= Day.From OR Day1 <= Day.To THEN
                            Pattern.Flag = True
                        ELSE
                            Pattern.Flag = False
                        END IF
                    ELSE
                        IF Day1 >= Day.From AND Day1 <= Day.To THEN
                            Pattern.Flag = True
                        ELSE
                            Pattern.Flag = False
                        END IF
                    END IF
                ELSE
                    Pattern.Flag = False
                END IF
            ELSE
                Pattern.Flag = False
            END IF
        END IF
    ELSE
        IF Day1 > False AND Month1 = False AND Year1 > False THEN ' DD-YYYY
            IF Year.From = Year.To THEN
                IF Year1 = Year.From THEN
                    IF Day.From <= Day.To THEN
                        IF Day1 >= Day.From AND Day1 <= Day.To THEN
                            Pattern.Flag = True
                        ELSE
                            Pattern.Flag = False
                        END IF
                    ELSE
                        Pattern.Flag = False
                    END IF
                ELSE
                    Pattern.Flag = False
                END IF
            ELSE
                IF Year.From < Year.To THEN
                    IF Year1 >= Year.From AND Year1 <= Year.To THEN
                        IF Day.From > Day.To THEN
                            IF Day1 >= Day.From OR Day1 <= Day.To THEN
                                Pattern.Flag = True
                            ELSE
                                Pattern.Flag = False
                            END IF
                        ELSE
                            IF Day1 >= Day.From AND Day1 <= Day.To THEN
                                Pattern.Flag = True
                            ELSE
                                Pattern.Flag = False
                            END IF
                        END IF
                    ELSE
                        Pattern.Flag = False
                    END IF
                ELSE
                    Pattern.Flag = False
                END IF
            END IF
        ELSE
            IF Day1 > False AND Month1 = False AND Year1 = False THEN ' DD
                IF Day1 >= Day.From AND Day1 <= Day.To THEN
                    Pattern.Flag = True
                ELSE
                    Pattern.Flag = False
                END IF
            ELSE
                IF Day1 = False AND Month1 > False AND Year1 > False THEN ' MM-YYYY
                    IF Year.From = Year.To THEN
                        IF Year1 = Year.From THEN
                            IF Month.From <= Month.To THEN
                                IF Month1 >= Month.From AND Month1 <= Month.To THEN
                                    Pattern.Flag = True
                                ELSE
                                    Pattern.Flag = False
                                END IF
                            ELSE
                                Pattern.Flag = False
                            END IF
                        ELSE
                            Pattern.Flag = False
                        END IF
                    ELSE
                        IF Year.From < Year.To THEN
                            IF Year1 >= Year.From AND Year1 <= Year.To THEN
                                IF Month.From > Month.To THEN
                                    IF Month1 >= Month.From OR Month1 <= Month.To THEN
                                        Pattern.Flag = True
                                    ELSE
                                        Pattern.Flag = False
                                    END IF
                                ELSE
                                    IF Month1 >= Month.From AND Month1 <= Month.To THEN
                                        Pattern.Flag = True
                                    ELSE
                                        Pattern.Flag = False
                                    END IF
                                END IF
                            ELSE
                                Pattern.Flag = False
                            END IF
                        ELSE
                            Pattern.Flag = False
                        END IF
                    END IF
                ELSE
                    IF Day1 = False AND Month1 > False AND Year1 = False THEN ' MM
                        IF Month1 >= Month.From AND Month1 <= Month.To THEN
                            Pattern.Flag = True
                        ELSE
                            Pattern.Flag = False
                        END IF
                    ELSE
                        IF Day1 = False AND Month1 = False AND Year1 > False THEN ' YYYY
                            IF Year1 >= Year.From AND Year1 <= Year.To THEN
                                Pattern.Flag = True
                            ELSE
                                Pattern.Flag = False
                            END IF
                        ELSE
                            IF Day1 = False AND Month1 = False AND Year1 = False THEN ' nul
                                ' no date in pattern, so filename matches.
                                EXIT SUB
                            END IF
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF
END SUB

' calculate byte suffix
SUB Suffix (Var#, Var3$)
REM B  (Byte) = 00x - 0FFx (hexidecimal zero-based)
REM KB (Kilobyte) = 1024 B
REM MB (Megabyte) = 1024 KB (1 MB B)
REM GB (Gigabyte) = 1024 MB
REM TB (Terabyte) = 1024 GB (1 MB MB)
REM PB (Petabyte) = 1024 TB
REM EB (Exabyte) = 1024 PB (1 MB TB)

' check double
VarX# = Var#
s$ = STR$(VarX#)
IF INSTR(s$, "D") THEN
    Var3$ = s$
    EXIT SUB
END IF

' get sign
IF VarX# < 0# THEN
    Sign = True
    VarX# = ABS(VarX#)
END IF

' calculate bytes
TempA = False
DO
    IF VarX# >= 1024 THEN
        VarX# = VarX# / 1024
        TempA = TempA + 1
        IF TempA = 6 THEN
            EXIT DO
        END IF
    ELSE
        EXIT DO
    END IF
LOOP

' calculate byte string
Var3$ = FormatString$(VarX#)
IF INSTR(Var3$, ".") THEN
    Var3$ = LEFT$(Var3$, INSTR(Var3$, ".") + 1)
ELSE
    Var3$ = Var3$ + ".0"
END IF

' calculate byte suffix
Var$ = Nul
IF TempA > 0 THEN
    Var$ = MID$("KMGTPE", TempA, 1)
END IF
Var3$ = Var3$ + " " + Var$ + "B"

' calculate byte sign
IF Sign THEN
    Var3$ = "-" + Var3$
END IF
END SUB

' format and display time elapsed copying file
SUB Time.Display1
Var1! = Time.Elapsed
Hours = INT(Var1! / 3600!)
Var1! = Var1! - Hours * 3600!
Minutes = INT(Var1! / 60!)
Var1! = Var1! - Minutes * 60!
Seconds = INT(Var1!)
PRINT "Transfer time: ";
IF Hours > 23 THEN
    Days = INT(Hours / 24)
    PRINT RIGHT$("00" + MID$(STR$(Days), 2), 2); ":";
END IF
PRINT RIGHT$("00" + MID$(STR$(Hours), 2), 2); ":";
PRINT RIGHT$("00" + MID$(STR$(Minutes), 2), 2); ":";
PRINT RIGHT$("00" + MID$(STR$(Seconds), 2), 2);
Var1$ = STR$(Var1!)
Var = INSTR(Var1$, ".")
IF Var THEN
    PRINT MID$(Var1$, Var, 3)
ELSE
    PRINT ".00"
END IF
END SUB

' format and display time rate copying file
SUB Time.Display2
REM 1 B (Byte) = 00x - 0FFx (hexidecimal zero-based)
REM KB (Kilobyte) = 1024 B
REM MB (Megabyte) = 1024 KB (1 MB B)
REM GB (Gigabyte) = 1024 MB
REM TB (Terabyte) = 1024 GB (1 MB MB)
REM PB (Petabyte) = 1024 TB
REM EB (Exabyte) = 1024 PB (1 MB TB)

' division by zero
IF Time.Elapsed = 0! THEN
    Var# = 0#
    S$ = "B"
    GOTO EndCalc
END IF

' rate of bytes copied/sec.
Total.Bytes2 = Total.Bytes2 / Time.Elapsed

' zero bytes copied
IF Total.Bytes2 = 0# THEN
    Var# = 0#
    S$ = "B"
    GOTO EndCalc
END IF

' calculate suffix
TempA = False
Var# = Total.Bytes2
DO
    IF Var# < 1024# THEN
        EXIT DO
    END IF
    Var# = Var# / 1024#
    TempA = TempA + 1
    IF TempA = 6 THEN
        EXIT DO
    END IF
LOOP

' determine suffix
S$ = ""
IF TempA > 0 THEN
    S$ = MID$("KMGTPE", TempA, 1)
END IF
S$ = S$ + "B"
EndCalc:
' display rate
PRINT "Transfer rate: " + FormatString$(Var#) + " " + S$ + "/second."
END SUB

SUB Nameit (ErrFlag)

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset count variables
Display.Lines = False
Continuous.Display = False
Quit.Searching = False
Dirs.Counted = False
Files.Counted = False
Lines.Counted = False

' store command line
Command.Line = RTRIM$(Command.Line)

' get switches from command line
Continuous.Display = ParseLine("/C")
Dir.Search = ParseLine("/D")
Lower.Case = ParseLine("/L")
Sensitive = ParseLine("/S")
Display.Errors = ParseLine("/Z")
List.Only.Short = ParseLine("/0")
List.Only.Source = ParseLine("/1")
List.Only.Dest = ParseLine("/2")
List.Only.Not = ParseLine("/3")
More.Display = False

' get new filename from command line
Command.Line = RTRIM$(Command.Line)
New.Name = False
New.File$ = ""
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    IF INSTR(Imbedded + 1, Command.Line, "/") THEN
        GOTO Boot.Error
    END IF
    New.Name = True
    New.File$ = MID$(Command.Line, Imbedded + 2)
    IF LEFT$(New.File$, 1) = CHR$(34) AND RIGHT$(New.File$, 1) = CHR$(34) THEN
        New.File$ = MID$(New.File$, 2)
        New.File$ = LEFT$(New.File$, LEN(New.File$) - 1)
    ELSE
        IF INSTR(New.File$, " ") THEN
            GOTO Boot.Error
        END IF
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current filename
        Old.File$ = Command.Work
        Command.Work = Nul

        ' rename filename
        IF New.Name = False THEN
            New.File$ = Old.File$
        END IF
        IF Lower.Case THEN
            New.File$ = LCASE$(New.File$)
        END IF

        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            PRINT "Searching: " + Old.File$
        END IF

        ' rename filenames
        CALL Nameit_Filenames(Old.File$, New.File$)

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Dirs.Counted)
    PRINT "Directories counted: "; Total$
    Total$ = FormatString$(Files.Counted)
    PRINT "Files counted: "; Total$
    COLOR White, Black
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "RENAME"
PRINT "   File rename utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   RENAME [\path\]filename.ext [/n][/0123cdlsz]"
PRINT "Where:"
PRINT "   /nfilename.ext  new filename"
PRINT "   /c  continuous display"
PRINT "   /d  rename only directories"
PRINT "   /l  force to lowercase"
PRINT "   /s  force case-sensitive matching"
PRINT "   /z  suppress error messages"
PRINT "   /0  only display short source filename"
PRINT "   /1  only display source filename"
PRINT "   /2  only display dest filename"
PRINT "   /3  don't display filenames"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type RENAME /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, False
    PRINT "File rename utility; "
END IF
RETURN
END SUB

' subroutine to access filenames
SUB Nameit_Filenames (OldPath$, NewPath$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' declare wide functions
DIM ASCIIZ.FileW AS STRING * MAX_PATH2 ' a null terminated filename
DIM finddataw AS WIN32_FIND_DATAW ' the windows filename structure
DIM Wfile.HandleW AS _UNSIGNED _OFFSET ' windows file handle for FindFile

' get drive from source
Drive$ = ""
IF MID$(OldPath$, 2, 1) = ":" THEN
    Drive$ = LEFT$(OldPath$, 2)
    V = ASC(UCASE$(LEFT$(Drive$, 1))) - 64
    IF MEDIAEXISTS(V) = 0 THEN
        CALL DisplayError("Error accessing source drive.")
        EXIT SUB
    END IF
END IF

' get path from source
Path$ = ""
FOR I = LEN(OldPath$) TO 1 STEP -1
    IF MID$(OldPath$, I, 1) = "\" THEN
        Path$ = LEFT$(OldPath$, I)
        EXIT FOR
    END IF
NEXT
IF MID$(Path$, 2, 1) = ":" THEN
    Path$ = MID$(Path$, 3)
END IF

' make filename
Var$ = RTRIM$(OldPath$)
ASCIIZ = Var$ + CHR$(0)
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search filenames
    DO
        ' check directory attribute
        Attribute = finddata.dwFileAttributes
        Flag = 0
        IF (Attribute AND &H10) = &H10 THEN
            IF Dir.Search THEN
                Dirs.Counted = Dirs.Counted + 1
                Flag = -1
            END IF
        ELSE
            IF Dir.Search = 0 THEN
                Files.Counted = Files.Counted + 1
                Flag = -1
            END IF
        END IF
        IF Flag THEN

            ' store filename
            Filename$ = finddata.cFileName
            Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

            ' store short filename
            Short.Filename$ = finddata.cAlternateFileName
            V = INSTR(Short.Filename$, CHR$(0))
            IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
            IF Short.Filename$ = Nul THEN
                Short.Filename$ = finddata.cFileName
                V = INSTR(Short.Filename$, CHR$(0))
                IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
            END IF

            ' check filename
            IF Filename$ <> "." AND Filename$ <> ".." THEN

                ' check unicode
                IF INSTR(Filename$, "?") THEN
                    WideZ$ = AsciiToWide$(Short.Filename$) + CHR$(0)
                    Wfile.HandleW = FindFirstFileW(WideZ$ + CHR$(0), _OFFSET(finddataw))
                    IF Wfile.HandleW <> INVALID_HANDLE_VALUE THEN
                        UnicodeLong = finddataw.cFileName
                        UnicodeLong = LEFT$(UnicodeLong, wzLength(UnicodeLong))
                        r = FindClose(Wfile.HandleW)
                        CALL RenameWideFile(Drive$, Path$, UnicodeLong, NewPath$)
                    ELSE
                        PRINT "Error 0x"; HEX$(GetLastError); " "; WideZ$
                    END IF
                ELSE
                    CALL RenameFile(Drive$, Path$, Filename$, NewPath$)
                END IF
            END IF
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to rename file
SUB RenameFile (Drive1$, Path1$, File1$, File2$)
' initialize filename buffer
DIM OldASCIIZ AS STRING * 260
DIM NewASCIIZ AS STRING * 260
OldDrive$ = Drive1$
OldPath$ = Path1$
OldFile$ = File1$
NewFile$ = File2$

' check case sensiive
IF Sensitive = 0 THEN
    OldFile$ = LCASE$(OldFile$)
    NewFile$ = LCASE$(NewFile$)
END IF

' get drive from destination
Drive$ = ""
IF MID$(NewFile$, 2, 1) = ":" THEN
    Drive$ = LEFT$(NewFile$, 2)
    V = ASC(UCASE$(LEFT$(Drive$, 1))) - 64
    IF MEDIAEXISTS(V) = 0 THEN
        CALL DisplayError("Error accessing destination drive.")
        EXIT SUB
    END IF
    NewFile$ = MID$(NewFile$, 3)
END IF

' get path from destination
Path$ = ""
FOR I = LEN(NewFile$) TO 1 STEP -1
    IF MID$(NewFile$, I, 1) = "\" THEN
        Path$ = LEFT$(NewFile$, I)
        NewFile$ = MID$(NewFile$, I + 1)
        EXIT FOR
    END IF
NEXT

' remove unicode
DO
    Var = INSTR(OldFile$, "?")
    IF Var THEN
        MID$(OldFile$, Var, 1) = " "
    ELSE
        EXIT DO
    END IF
LOOP

' call matching function
Mask$ = MaskNewName$(OldFile$, NewFile$)
IF LEN(Mask$) THEN
    ' rename file
    OldASCIIZ = OldDrive$ + OldPath$ + OldFile$ + CHR$(0)
    NewASCIIZ = Drive$ + Path$ + Mask$ + CHR$(0)
    r = MoveFile(OldASCIIZ, NewASCIIZ)

    ' check error and try to rename short 8.3 filename
    IF r = 0 THEN
        OldASCIIZ = OldDrive$ + OldPath$ + Short.Filename$ + CHR$(0)
        r = MoveFile(OldASCIIZ, NewASCIIZ)
    END IF

    ' display search filename
    IF List.Only.Short THEN
        Output.Line$ = Short.Filename$
    ELSE
        IF List.Only.Source THEN
            Output.Line$ = OldFile$
        ELSE
            IF List.Only.Dest THEN
                Output.Line$ = Mask$
            ELSE
                Output.Line$ = OldFile$ + " - " + Mask$
            END IF
        END IF
    END IF
    IF List.Only.Not = False THEN
        CALL Nameit_DisplayFilename
    END IF
    IF Quit.Searching THEN
        EXIT SUB
    END IF

    ' display any errors
    IF r = 0 THEN
        CALL DisplayError("Error 0x" + HEX$(GetLastError&) + " renaming filename.")
    END IF
END IF
END SUB

' displays output line with prompt
SUB Nameit_DisplayFilename
' display filename
COLOR Yellow, Black
CALL AddStdout2(Output.Line)

' check line length
Length = LEN(Output.Line)
Lines.Counted = Lines.Counted + INT(Length / 80) + 1
CALL PagePrompt(Lines.Counted, FlagX)
IF FlagX THEN
    Lines.Counted = 0
END IF
END SUB

' subroutine to rename unicode file
SUB RenameWideFile (Drive1$, Path1$, File1$, File2$)
OldDrive$ = Drive1$
OldPath$ = Path1$
OldFile$ = File1$
NewFile$ = File2$

' get drive from destination
Drive$ = ""
IF MID$(NewFile$, 2, 1) = ":" THEN
    Drive$ = LEFT$(NewFile$, 2)
    V = ASC(UCASE$(LEFT$(Drive$, 1))) - 64
    IF MEDIAEXISTS(V) = 0 THEN
        CALL DisplayError("Error accessing destination drive.")
        EXIT SUB
    END IF
    NewFile$ = MID$(NewFile$, 3)
END IF

' get path from destination
Path$ = ""
FOR I = LEN(NewFile$) TO 1 STEP -1
    IF MID$(NewFile$, I, 1) = "\" THEN
        Path$ = LEFT$(NewFile$, I)
        NewFile$ = MID$(NewFile$, I + 1)
        EXIT FOR
    END IF
NEXT

' call matching function
NewFile$ = AsciiToWide$(NewFile$)
Mask$ = MaskNewWideName(OldFile$, NewFile$)

IF LEN(Mask$) THEN
    ' rename file
    OldASCIIZ$ = AsciiToWide$(OldDrive$) + AsciiToWide$(OldPath$) + OldFile$ + CHR$(0) + CHR$(0)
    NewASCIIZ$ = AsciiToWide$(Drive$) + AsciiToWide$(Path$) + Mask$ + CHR$(0) + CHR$(0)
    r = MoveFileW(OldASCIIZ$, NewASCIIZ$)

    ' display search filename
    NewMask$ = WideToAscii$(Mask$)
    IF List.Only.Short THEN
        Output.Line$ = Short.Filename$
    ELSE
        IF List.Only.Source THEN
            Output.Line$ = Short.Filename$
        ELSE
            IF List.Only.Dest THEN
                Output.Line$ = NewMask$
            ELSE
                Output.Line$ = Short.Filename$ + " - " + NewMask$
            END IF
        END IF
    END IF
    IF List.Only.Not = False THEN
        CALL Nameit_DisplayFilename
    END IF
    IF Quit.Searching THEN
        EXIT SUB
    END IF

    ' display any errors
    IF r = 0 THEN
        CALL DisplayError("Error 0x" + HEX$(GetLastError&) + " renaming filename.")
    END IF
END IF
END SUB

FUNCTION WideCharToAscii (v$)
P = 1
DO
    IF MID$(v$, P, 1) = CHR$(0) THEN
        IF MID$(v$, P + 1, 1) = CHR$(0) THEN
            v$ = LEFT$(v$, P - 1)
            EXIT FUNCTION
        END IF
        v$ = LEFT$(v$, P - 1) + MID$(v$, P + 1)
    ELSE
        P = P + 1
    END IF
    IF P > 28 THEN
        EXIT FUNCTION
    END IF
LOOP
END FUNCTION

' make a wide ascii filename
FUNCTION AsciiToWide$ (Var$)
Var2$ = Var$
FOR X = 1 TO LEN(Var2$)
    VarX$ = VarX$ + MID$(Var2$, X, 1) + CHR$(0)
NEXT
AsciiToWide$ = VarX$
END FUNCTION

' make an ansi filename
FUNCTION WideToAscii$ (Var$)
Var2$ = Var$
FOR X = 1 TO LEN(Var2$) STEP 2
    VarX$ = VarX$ + MID$(Var2$, X, 1)
NEXT
WideToAscii$ = VarX$
END FUNCTION

REM inputs a filename and matches wildcards returning masked output filename.
FUNCTION MaskNewName$ (path$, mask$)
IF path$ = "" THEN
    EXIT FUNCTION
END IF
IF INSTR(path$, "?") OR INSTR(path$, "*") THEN
    EXIT FUNCTION
END IF
FOR m = 0 TO LEN(mask$) - 1
    ch$ = MID$(mask$, m + 1, 1)
    q$ = MID$(path$, x + 1, 1)
    z$ = MID$(mask$, m + 2, 1)
    SELECT CASE ch$
        CASE "?"
            IF LEN(q$) AND q$ <> "." THEN
                R$ = R$ + q$
                x = x + 1
            END IF
        CASE "*"
            IF m = LEN(mask$) - 1 THEN
                WHILE x < LEN(path$)
                    R$ = R$ + MID$(path$, x + 1, 1)
                    x = x + 1
                WEND
            ELSE
                SELECT CASE z$
                    CASE "."
                        FOR i = LEN(path$) - 1 TO 0 STEP -1
                            IF MID$(path$, i + 1, 1) = "." THEN
                                EXIT FOR
                            END IF
                        NEXT
                        IF i < 0 THEN
                            R$ = R$ + MID$(path$, x + 1) + "."
                            i = LEN(path$)
                        ELSE
                            R$ = R$ + MID$(path$, x + 1, i - x + 1)
                        END IF
                        x = i + 1
                        m = m + 1
                    CASE "?"
                        R$ = R$ + MID$(path$, x + 1)
                        m = m + 1
                        x = LEN(path$)
                    CASE ELSE
                        FOR i = LEN(path$) - 1 TO 0 STEP -1
                            x$ = MID$(path$, i + 1, 1)
                            IF UCASE$(x$) = UCASE$(z$) THEN
                                EXIT FOR
                            END IF
                        NEXT
                        IF i < 0 THEN
                            R$ = R$ + MID$(path$, x + 1) + z$
                            x = LEN(path$)
                            m = m + 1
                        ELSE
                            R$ = R$ + MID$(path$, x + 1, i - x)
                            x = i + 1
                        END IF
                END SELECT
            END IF
        CASE "."
            DO WHILE x < LEN(path$)
                IF MID$(path$, x + 1, 1) = "." THEN
                    x = x + 1
                    EXIT DO
                END IF
                x = x + 1
            LOOP
            R$ = R$ + ch$
        CASE ELSE
            IF LEN(q$) AND q$ <> "." THEN
                x = x + 1
            END IF
            R$ = R$ + ch$
    END SELECT
NEXT

' trim trailing dots and spaces
DO
    IF RIGHT$(R$, 1) = "." THEN
        R$ = LEFT$(R$, LEN(R$) - 1)
    ELSE
        IF RIGHT$(R$, 1) = " " THEN
            R$ = LEFT$(R$, LEN(R$) - 1)
        ELSE
            EXIT DO
        END IF
    END IF
LOOP

' return masked filename
MaskNewName$ = R$
END FUNCTION

REM inputs a unicode filename and matches wildcards returning masked output filename.
FUNCTION MaskNewWideName$ (path$, mask$) ' * . M P 4
FOR m = 0 TO LEN(mask$) - 1 STEP 2
    ch$ = MID$(mask$, m + 1, 1)
    ch2$ = MID$(mask$, m + 1, 2)
    q$ = MID$(path$, x + 1, 1)
    q2$ = MID$(path$, x + 1, 2)
    z$ = MID$(mask$, m + 3, 1)
    z2$ = MID$(mask$, m + 3, 2)
    SELECT CASE ch$
        CASE "?"
            IF LEN(q$) AND q$ <> "." THEN
                R$ = R$ + q2$
                x = x + 2
            END IF
        CASE "*"
            IF m >= LEN(mask$) - 2 THEN ' asterick at eol
                WHILE x < LEN(path$)
                    R$ = R$ + MID$(path$, x + 1, 1)
                    x = x + 1
                WEND
            ELSE
                SELECT CASE z$
                    CASE "."
                        FOR i = LEN(path$) TO 0 STEP -2
                            IF MID$(path$, i + 1, 1) = "." THEN
                                EXIT FOR
                            END IF
                        NEXT
                        IF i < 0 THEN
                            R$ = R$ + MID$(path$, x + 1) + "." + CHR$(0)
                            i = LEN(path$)
                        ELSE
                            R$ = R$ + MID$(path$, x + 1, i - x + 2)
                        END IF
                        x = i + 2
                        m = m + 2
                    CASE "?"
                        R$ = R$ + MID$(path$, x + 1)
                        m = m + 2
                        x = LEN(path$)
                    CASE ELSE
                        FOR i = LEN(path$) TO 0 STEP -2
                            x$ = MID$(path$, i + 1, 1)
                            IF UCASE$(x$) = UCASE$(z$) THEN
                                EXIT FOR
                            END IF
                        NEXT
                        IF i < 0 THEN
                            R$ = R$ + MID$(path$, x + 1) + z2$
                            x = LEN(path$)
                            m = m + 2
                        ELSE
                            R$ = R$ + MID$(path$, x + 1, i - x + 1)
                            x = i + 2
                        END IF
                END SELECT
            END IF
        CASE "."
            DO WHILE x < LEN(path$)
                IF MID$(path$, x + 1, 1) = "." THEN
                    x = x + 2
                    EXIT DO
                END IF
                x = x + 2
            LOOP
            R$ = R$ + ch2$
        CASE ELSE
            IF LEN(q$) AND q$ <> "." THEN
                x = x + 2
            END IF
            R$ = R$ + ch2$
    END SELECT
NEXT

' trim trailing dots and spaces
DO
    IF RIGHT$(R$, 1) = "." THEN
        R$ = LEFT$(R$, LEN(R$) - 1)
    ELSE
        IF RIGHT$(R$, 1) = " " THEN
            R$ = LEFT$(R$, LEN(R$) - 1)
        ELSE
            EXIT DO
        END IF
    END IF
LOOP

' return masked filename
MaskNewWideName$ = R$
END FUNCTION

FUNCTION wzLength~& (ustr$)
FOR i% = 1 TO LEN(ustr$) STEP 2
    IF MID$(ustr$, i%, 2) = STRING$(2, CHR$(0)) THEN
        wzLength~& = i% - 1
        EXIT FUNCTION
    END IF
NEXT i%
wzLength~& = LEN(ustr$) - 1
END FUNCTION

SUB MoreX (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOTO Boot.Usage
END IF

' read command line switches
Continuous.Display = ParseLine("/C")
Strip.Blanks = ParseLine("/T")

' read start line override switch.
Line.Override = 0#
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Line.Override = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$

' initialize work variables
Lines.Counted = False
More.Display = False
Quit.Searching = False
Total.Lines = 0#

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

IF Stdin THEN
    CALL MoreStdin
END IF
IF Command.Line = "" THEN
    EXIT SUB
END IF

' filename processing loop
DO
    ' store entire command
    IF LEFT$(Command.Line, 1) = CHR$(34) THEN
        Imbedded = INSTR(2, Command.Line, CHR$(34))
        IF Imbedded THEN
            Command.Work = MID$(Command.Line, 2, Imbedded - 2)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    ELSE
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    END IF
    Command.Line = LTRIM$(Command.Line)
    Command.Line = RTRIM$(Command.Line)

    ' store current drive
    IF MID$(Command.Work, 2, 1) = ":" THEN
        Drive.Search = LEFT$(Command.Work, 1)
        Command.Work = MID$(Command.Work, 3)
    ELSE
        Drive.Search = Current.Drive$
    END IF
    Drive.Search = UCASE$(Drive.Search)

    ' store current directory
    Directory.Search$ = ""
    Imbedded1 = INSTR(Command.Work, "\")
    Imbedded2 = Imbedded1
    WHILE Imbedded1
        Imbedded2 = Imbedded1
        Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
    WEND
    IF Imbedded2 THEN
        Directory.Search$ = LEFT$(Command.Work, Imbedded2)
        Command.Work = MID$(Command.Work, Imbedded2 + 1)
    END IF
    IF Directory.Search$ = "" THEN
        IF LEFT$(Current.Directory, 2) = "\\" THEN
            Directory.Search$ = Current.Directory
        ELSE
            IF Drive.Search$ = Current.Drive$ THEN
                Directory.Search$ = MID$(Current.Directory, 3)
            ELSE
                Directory.Search$ = "\"
            END IF
        END IF
    END IF
    IF RIGHT$(Directory.Search$, 1) <> "\" THEN
        Directory.Search$ = Directory.Search$ + "\"
    END IF
   
    ' get filename spec
    Filename.Search$ = Command.Work
    IF Filename.Search$ = Nul THEN
        Filename.Search$ = "*.*"
    END IF
    Command.Work = Nul

    ' display search filename
    IF Continuous.Display = False THEN
        COLOR Yellow, Black
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            PRINT "Searching: " + Directory.Search$ + Filename.Search$
        ELSE
            PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
        END IF
    END IF

    ' call routine to search for files
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        CALL More_DisplayFileLines(Directory.Search$, Filename.Search$)
    ELSE
        V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
        IF MEDIAEXISTS(V) THEN
            CALL More_DisplayFileLines(Directory.Search$, Filename.Search$)
        END IF
    END IF

    ' check search filename
    IF Command.Line = Nul THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP
COLOR Plain, Black
EXIT SUB

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type MORE /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "MORE"
PRINT "   Filename display utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   MORE [d:][path][filename.ext] [/cnt]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /nxxx  start display line number"
PRINT "   /t  skip blank lines"
COLOR Plain, Black
END SUB

' subroutine to display lines in standard input
SUB MoreStdin
COLOR 7, 0
Advance.Line = 0
Lines.Counter = 0
Total.Lines = 0#
FOR StdinCount = 1 TO StdinLines
    COLOR 7, 0
    Outpt.Length = 0
    X$ = StdinArray(StdinCount)
    IF RTRIM$(X$) = "" THEN
        IF Strip.Blanks = 0 THEN
            PRINT #StdoutHandle, X$
            Outpt.Length = LEN(X$)
        END IF
    ELSE
        IF LCASE$(StdoutFile$) = "scrn:" THEN
            _CONTROLCHR OFF
            PRINT X$
            _CONTROLCHR ON
            Outpt.Length = LEN(X$)
        ELSE
            PRINT #StdoutHandle, X$
            Outpt.Length = LEN(X$)
        END IF
    END IF

    Lines.Counter = Lines.Counter + INT(Outpt.Length / 80) + 1
    CALL MoreDisplayPrompt(Lines.Counter, FlagX)
    IF FlagX THEN
        Lines.Counter = 0
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT FOR
    END IF
NEXT
COLOR Plain, Black
END SUB

' subroutine to display lines in a filename
SUB More_DisplayFileLines (Directory.Search$, Filename.Search$)

' reset counters
Advance.Line = 0
Lines.Counter = 0
Total.Lines = 0#

' make filename
Filename.Count$ = Directory.Search$ + Filename.Search$
IF _FILEEXISTS(Filename.Count$) THEN
    X = FREEFILE
    OPEN Filename.Count$ FOR BINARY SHARED AS #X
    Count# = 0#
    IF LOF(X) > 0 THEN
        DO UNTIL EOF(X)
            LINE INPUT #X, X$
            Count# = Count# + LEN(X$) + 2#
            Total.Lines = Total.Lines + 1
            DisplayLine = -1
            IF Line.Override > 0 THEN
                IF Line.Override > Total.Lines THEN
                    DisplayLine = 0
                END IF
            END IF
            Outpt.Length = 0
            IF DisplayLine THEN
                IF RTRIM$(X$) = "" THEN
                    IF Strip.Blanks = 0 THEN
                        Outpt.Length = LEN(X$)
                        COLOR 7, 0
                        PRINT #StdoutHandle, X$
                    END IF
                ELSE
                    IF LCASE$(StdoutFile$) = "scrn:" THEN
                        Outpt.Length = LEN(X$)
                        _CONTROLCHR OFF
                        COLOR 7, 0
                        PRINT X$
                        _CONTROLCHR ON
                    ELSE
                        Outpt.Length = LEN(X$)
                        COLOR 7, 0
                        PRINT #StdoutHandle, X$
                    END IF
                END IF
                Lines.Counter = Lines.Counter + INT(Outpt.Length / 80) + 1
                CALL MoreDisplayPrompt(Lines.Counter, FlagX)
                IF FlagX THEN
                    Lines.Counter = 0
                END IF
            END IF

            ' check quit searching
            IF Quit.Searching THEN
                EXIT DO
            END IF
        LOOP
    END IF
    IF Count# < LOF(X) THEN
        IF Quit.Searching = False THEN
            IF Line.Override <= Total.Lines THEN
                PRINT #StdoutHandle, ""
                Lines.Counter = Lines.Counter + 1
                CALL MoreDisplayPrompt(Lines.Counter, FlagX)
                IF FlagX THEN
                    Lines.Counter = 0
                END IF
            END IF
        END IF
    END IF
    CLOSE #X
END IF
END SUB

SUB MoreDisplayPrompt (Var, Flag)
Flag = 0
IF Continuous.Display = False THEN
    IF LCASE$(StdoutFile$) = "scrn:" THEN
        IF Status THEN
            Z = 21
        ELSE
            Z = 22
        END IF
        IF Advance.Line THEN
            Z = Advance.Line
        END IF
        IF Var >= Z THEN
            Flag = -1
            Advance.Line = False
            Prompt2$ = "More(y/n/c/s)?"
            CALL MorePrompt(Prompt2$, " yncs" + CHR$(13), Outpt2$)
            SELECT CASE Outpt2$
                CASE "s"
                    DO
                        _LIMIT 30
                        X$ = INKEYx$
                        IF X$ >= "1" AND X$ <= "9" THEN
                            Advance.Line = VAL(X$)
                            EXIT DO
                        END IF
                    LOOP
                CASE CHR$(13)
                    Advance.Line = 1
                CASE "c"
                    Continuous.Display = True
                CASE "n"
                    Quit.Searching = True
            END SELECT
        END IF
    END IF
END IF
END SUB

SUB Stree (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset counters
Directories.Counted = FalseS
Display.Lines = False
More.Display = False
Continuous.Display = False
Quit.Searching = False

' get current drive/directory
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' store command line
Command.Line = RTRIM$(Command.Line)

' check command line switches
Append.Slash = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Extended.List = ParseLine("/E")
Short.Display = ParseLine("/F")

Search.Archive = ParseLine("/A")
Search.Hidden = ParseLine("/H")
Search.Readonly = ParseLine("/O")
Search.System = ParseLine("/S")
Search.Compressed = ParseLine("/J")
Search.Encrypted = ParseLine("/K")

Display.Lowercase = ParseLine("/Y")
Strip.Drive = ParseLine("/U")
Short.Filenames = ParseLine("/V")
Wide.Display2 = ParseLine("/W1")
Wide.Display = ParseLine("/W")
Truncate.Slash = ParseLine("/X")
Display.Errors = ParseLine("/Z")
Recurse.Directories = ParseLine("/R")

' reset some display variables
IF Wide.Display THEN
    Extended.List = False
    Short.Display = True
    Short.Filenames = True
    Display.Length = 0
END IF

' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32 + T3!
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32 + T3!
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get extended date\time switches
Creation.Time = ParseLine("/1")
Access.Time = ParseLine("/2")
Modified.Time = ParseLine("/3")
IF Creation.Time = False THEN
    IF Access.Time = False THEN
        IF Modified.Time = False THEN
            Modified.Time = True
        END IF
    END IF
END IF

' get nested switch from command line
Nested.Levels = 256%
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = Var%
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "" THEN
    Command.Line = "*.*"
END IF
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store search drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
        ELSE
            Drive.Search = Current.Drive
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = Command.Work

        ' parse drive letter
        Temp.Directory$ = Directory.Search$
        IF MID$(Temp.Directory$, 2, 1) = ":" THEN
            Temp.Drive$ = UCASE$(LEFT$(Temp.Directory$, 1))
            Temp.Directory$ = MID$(Temp.Directory$, 3)
        ELSE
            Temp.Drive$ = Drive.Search
        END IF

        ' parse directory
        IF LEFT$(Temp.Directory$, 2) <> "\\" THEN
            IF LEFT$(Temp.Directory$, 1) <> "\" THEN
                IF Temp.Drive$ <> LEFT$(Current.Directory, 1) THEN
                    Directory.Search$ = Drive.Search + ":\" + Temp.Directory$
                ELSE
                    Directory.Search$ = Current.Directory + Temp.Directory$
                END IF
            ELSE
                Directory.Search$ = Temp.Drive$ + ":" + Temp.Directory$
            END IF
        END IF

        ' display search header
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            Display.Lines = Display.Lines + 1
            PRINT "Searching: " + Directory.Search$
        END IF

        ' call subroutine to search directories
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Stree_Directories(Directory.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Temp.Drive$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Stree_Directories(Directory.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Display.Length THEN
    PRINT
END IF
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Directories.Counted)
    PRINT "Directories counted: "; Total$
END IF
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory search utility;"
END IF
RETURN

Get.Numeric:
Var% = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var% = Var% * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "TREEDIR"
PRINT "   Directory search utility;"
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   TREEDIR [d:\path\][/ahosjk][/bcdefnrtuvwxyz][/123]"
PRINT "Where:"
PRINT "   /b  append slash         /c  continuous display"
PRINT "   /e  extended display     /f  short filename display"
PRINT "   /r  recurse directories  /u  strip leading drive letter"
PRINT "   /v  use 8.3 filenames    /w  wide list display"
PRINT "   /w1 prepend slash        /x  truncate slash"
PRINT "   /y  lowercase display    /z  suppress errors"
PRINT "   /nxxx  recurse levels override"
PRINT "   display directory ranges:"
PRINT "      /1  creation, /2 accessed, /3 modified"
PRINT "      /d  is range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "      /t  is range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "   display directory attributes:"
PRINT "      /a  archive, /h  hidden, /o  read-only, /s  system"
PRINT "      /j  compressed,  /k  encrypted"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type TREEDIR /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

' subroutine to access directories
SUB Stree_Directories (Directory.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM Recurse.Temp AS INTEGER

' reset wildcard flag
Wild = 0

' make directory filename
Temp.Dir$ = Directory.Search$
Recurse.Temp = Recurse.Directories
IF Recurse.Directories THEN
    IF INSTR(Temp.Dir$, "?") OR INSTR(Temp.Dir$, "*") THEN
        ' Recurse.Temp = False
    ELSE
        IF RIGHT$(Temp.Dir$, 1) <> "\" THEN
            Temp.Dir$ = Temp.Dir$ + "\"
        END IF
        Temp.Dir$ = Temp.Dir$ + "*.*"
    END IF
END IF
ASCIIZ = Temp.Dir$ + CHR$(0)

' find first long filename
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))

' check findirst error
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' filename/directory loop
    DO

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF

        ' get file attributes
        Attribute = finddata.dwFileAttributes

        ' check directory attribute
        IF (Attribute AND &H10) = &H10 THEN

            ' check to recurse directories
            IF Recurse.Temp = 0 THEN
                ' store directory name
                Directory$ = finddata.cFileName
                V = INSTR(Directory$, CHR$(0))
                IF V THEN Directory$ = LEFT$(Directory$, V - 1)

                ' display directory
                GOSUB Display.Directory
            END IF

            ' check to recurse directories
            IF Recurse.Temp THEN

                ' store directory name
                Directory$ = finddata.cFileName
                V = INSTR(Directory$, CHR$(0))
                IF V THEN Directory$ = LEFT$(Directory$, V - 1)

                ' check unicode
                Unicode = 0
                IF INSTR(Directory$, "?") THEN
                    Unicode = -1
                    Directory$ = finddata.cAlternateFileName
                    V = INSTR(Directory$, CHR$(0))
                    IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                END IF

                ' check directory name
                Valid.Flag = -1
                IF Directory$ = "." THEN
                    Valid.Flag = 0
                END IF
                IF Directory$ = ".." THEN
                    Valid.Flag = 0
                END IF
                IF Directory$ = Nul THEN
                    Valid.Flag = 0
                END IF
                IF Valid.Flag THEN

                    ' display directory
                    GOSUB Display.Directory

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Stree_Directories(Next.Directory$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
EXIT SUB

Display.Directory:
' retract one directory if wildcard/global characters in trailing path
IF INSTR(Directory.Search$, "?") OR INSTR(Directory.Search$, "*") THEN
    Wild = -1
    FOR V = LEN(Directory.Search$) TO 1 STEP -1
        IF MID$(Directory.Search$, V, 1) = "\" THEN
            Directory.Search$ = LEFT$(Directory.Search$, V)
            EXIT FOR
        END IF
    NEXT
END IF

' make next search directory
IF RIGHT$(Directory.Search$, 1) <> "\" THEN
    Next.Directory$ = Directory.Search$ + "\" + Directory$
ELSE
    Next.Directory$ = Directory.Search$ + Directory$
END IF

' store directory name
Temp.Directory$ = finddata.cFileName
V = INSTR(Temp.Directory$, CHR$(0))
IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)

IF Temp.Directory$ = Nul THEN
    RETURN
END IF
IF Temp.Directory$ = "." THEN
    RETURN
END IF
IF Temp.Directory$ = ".." THEN
    RETURN
END IF

' construct directory for display
IF Recurse.Directories = 0 THEN
    IF Wild THEN
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Temp.Directory$ = Directory.Search$ + "\" + Directory$
        ELSE
            Temp.Directory$ = Directory.Search$ + Directory$
        END IF
    ELSE
        Temp.Directory$ = Directory.Search$
    END IF
ELSE
    IF Unicode AND CurrentCommand = CommandCount THEN
        Unicode.Directory$ = finddata.cFileName
        V = INSTR(Unicode.Directory$, CHR$(0))
        IF V THEN Unicode.Directory$ = LEFT$(Unicode.Directory$, V - 1)
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Temp.Directory$ = Directory.Search$ + "\" + Unicode.Directory$
        ELSE
            Temp.Directory$ = Directory.Search$ + Unicode.Directory$
        END IF
    ELSE
        Temp.Directory$ = Next.Directory$
    END IF
    Wild = -1
END IF

' get short filename
IF Short.Filenames THEN
    Temp.Directory$ = finddata.cAlternateFileName
    V = INSTR(Temp.Directory$, CHR$(0))
    IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)
    IF Temp.Directory$ = Nul THEN
        Temp.Directory$ = finddata.cFileName
        V = INSTR(Temp.Directory$, CHR$(0))
        IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)
    END IF
END IF

' check for read-only file
IF Search.Readonly THEN
    IF (Attribute AND &H1) <> &H1 THEN
        RETURN
    END IF
END IF

' check for hidden file
IF Search.Hidden THEN
    IF (Attribute AND &H2) <> &H2 THEN
        RETURN
    END IF
END IF

' check for system file
IF Search.System THEN
    IF (Attribute AND &H4) <> &H4 THEN
        RETURN
    END IF
END IF

' check for archive file
IF Search.Archive THEN
    IF (Attribute AND &H20) <> &H20 THEN
        RETURN
    END IF
END IF

' check for compressed file
IF Search.Compressed THEN
    IF (Attribute AND &H800) <> &H800 THEN
        RETURN
    END IF
END IF

' check for encrypted file
IF Search.Encrypted THEN
    IF (Attribute AND &H4000) <> &H4000 THEN
        RETURN
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        RETURN
    END IF
    IF File.Work.Date > Search.To.Date THEN
        RETURN
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        RETURN
    END IF
    IF File.Work.Time > Search.To.Time THEN
        RETURN
    END IF
END IF

' store directory name
Outpt$ = RTRIM$(Temp.Directory$)
IF Short.Display THEN
    IF RIGHT$(Outpt$, 1) = "\" THEN
        Outpt$ = LEFT$(Outpt$, LEN(Outpt$) - 1)
    END IF
    FOR Imbedded = LEN(Outpt$) TO 1 STEP -1
        IF MID$(Outpt$, Imbedded, 1) = "\" THEN
            Outpt$ = MID$(Outpt$, Imbedded + 1)
            EXIT FOR
        END IF
    NEXT
END IF
IF Truncate.Slash THEN
    IF Outpt$ <> "\" THEN
        IF RIGHT$(Outpt$, 1) = "\" THEN
            Outpt$ = LEFT$(Outpt$, LEN(Outpt$) - 1)
        END IF
    END IF
END IF
IF Short.Filenames = False OR Short.Display THEN
    IF MID$(Outpt$, 2, 1) <> ":" THEN
        IF LEFT$(Directory.Search$, 2) <> "\\" THEN
            Outpt$ = Drive.Search + ":" + Outpt$
        END IF
    END IF
END IF
IF LEFT$(Directory.Search$, 2) = "\\" THEN
    IF MID$(Outpt$, 2, 1) = ":" THEN
        Outpt$ = MID$(Outpt$, 3)
    END IF
END IF
IF Strip.Drive THEN
    IF MID$(Outpt$, 2, 1) = ":" THEN
        Outpt$ = MID$(Outpt$, 3)
    END IF
END IF
IF Display.Lowercase THEN
    Outpt$ = LCASE$(Outpt$)
END IF

' increment directories counted
Directories.Counted = Directories.Counted + 1!

' check wide display output
IF Wide.Display THEN
    COLOR Yellow, Black
    ' append/prepend slash
    IF Strip.Drive THEN
        IF Wide.Display2 THEN ' prepend
            IF LEFT$(Outpt$, 1) <> "\" THEN
                Outpt$ = "\" + Outpt$
            END IF
        END IF
        IF Append.Slash THEN ' append
            IF RIGHT$(Outpt$, 1) <> "\" THEN
                Outpt$ = Outpt$ + "\"
            END IF
        END IF
    END IF
    IF CurrentCommand < CommandCount THEN
        CALL AddStdout2(Outpt$)
        RETURN
    END IF
    Outpt$ = LEFT$(Outpt$, 14)
    IF CurrentCommand = CommandCount THEN
        PRINT #StdoutHandle, Outpt$;
        PRINT #StdoutHandle, SPACE$(15 - LEN(Outpt$));
        ' count directories in one line
        Display.Length = Display.Length + 1
        IF Display.Length = 5 THEN
            ' reset counters
            PRINT #StdoutHandle, ""
            Display.Length = False
            Display.Lines = Display.Lines + 1
            CALL PagePrompt(Display.Lines, FlagX)
            IF FlagX THEN
                Display.Lines = 0
            END IF
            IF Quit.Searching THEN
                EXIT SUB
            END IF
        END IF
    END IF
ELSE
    ' check slash switch
    IF Append.Slash THEN
        IF RIGHT$(Outpt$, 1) <> "\" THEN
            Outpt$ = Outpt$ + "\"
        END IF
    END IF
    IF CurrentCommand < CommandCount THEN
        CALL AddStdout2(Outpt$)
        RETURN
    END IF
    IF CurrentCommand = CommandCount THEN

        ' display full directory pathname
        COLOR Yellow, Black
        PRINT #StdoutHandle, Outpt$

        ' store length of directory name
        Length = LEN(Outpt$)

        ' calculate length variable
        Display.Lines = Display.Lines + INT(Length / 80) + 1
        CALL PagePrompt(Display.Lines, FlagX)
        IF FlagX THEN
            Display.Lines = 0
        END IF

        ' check output type
        IF Extended.List THEN
            ' display extended directory attributes
            IF LCASE$(StdoutFile$) = "scrn:" THEN
                GOSUB Display.List
            END IF
        END IF
    END IF
END IF
RETURN

' routine to display extended directory attributes
Display.List:
Length = False
' store file creation date\time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
    Outpt$ = File.Date$ + " " + File.Time$
    Length = Length + LEN(Outpt$)
    COLOR Red, Black
    PRINT Outpt$;
END IF
IF Access.Time THEN
    x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
    IF Creation.Time THEN
        Length = Length + 1
        COLOR White, Black
        PRINT "\";
    END IF
    Outpt$ = File.Date$ + " " + File.Time$
    Length = Length + LEN(Outpt$)
    COLOR Red, Black
    PRINT Outpt$;
END IF
IF Modified.Time THEN
    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
    IF Creation.Time OR Access.Time THEN
        Length = Length + 1
        COLOR White, Black
        PRINT "\";
    END IF
    Outpt$ = File.Date$ + " " + File.Time$
    Length = Length + LEN(Outpt$)
    COLOR Red, Black
    PRINT Outpt$;
END IF

' display file attributes
IF (Attribute AND &H1) = &H1 THEN
    Outpt$ = " Read-only"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF (Attribute AND &H2) = &H2 THEN
    Outpt$ = " Hidden"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF (Attribute AND &H4) = &H4 THEN
    Outpt$ = " System"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF (Attribute AND &H20) = &H20 THEN
    Outpt$ = " Archive"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF (Attribute AND &H800) = &H800 THEN
    Outpt$ = " Compressed"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF (Attribute AND &H4000) = &H4000 THEN
    Outpt$ = " Encrypted"
    Length = Length + LEN(Outpt$)
    COLOR White, Black
    PRINT Outpt$;
END IF
IF Length THEN
    PRINT
END IF
Display.Lines = Display.Lines + INT(Length / 80) + 1
CALL PagePrompt(Display.Lines, FlagX)
IF FlagX THEN
    Display.Lines = 0
END IF
RETURN

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

SUB FindY (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOTO Boot.Usage
END IF

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$

' check command line switches
Boolean.Search = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Display.Lines = ParseLine("/D")
Ignore.Case = ParseLine("/I")
List.Lines = ParseLine("/L")
Reverse.Search = ParseLine("/R")
WildCard.Search = ParseLine("/W")
Strip.Blanks = ParseLine("/T")

' get display column
Search.Column = 1
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Imbedded2 = Imbedded + 2
    DO
        Switch$ = MID$(Command.Line, Imbedded2, 1)
        IF Switch$ >= "0" AND Switch$ <= "9" THEN
            Column$ = Column$ + Switch$
        ELSE
            EXIT DO
        END IF
        Imbedded2 = Imbedded2 + 1
    LOOP
    IF Column$ = Nul THEN
        GOTO Boot.Error
    END IF
    Search.Column = INT(VAL(Column$))
    IF Search.Column = False THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded2)
END IF

' get search string
Search.String = Nul
Imbedded = INSTR(Command.Line, "/[")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.String = Nul
    Next.Bracket = INSTR(Imbedded + 2, Command.Line, "]")
    IF Next.Bracket THEN
        Search.String = MID$(Command.Line, Imbedded + 2, Next.Bracket - Imbedded - 2)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Next.Bracket + 1)
    END IF
    IF Search.String = Nul THEN
        GOTO Boot.Error
    END IF
END IF
IF Search.String = Nul THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' initialize work variables
Lines.Counted = False
More.Display = False
Quit.Searching = False

' check for case sensitive
IF Ignore.Case THEN
    Search.String = UCASE$(Search.String)
END IF

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header
IF Stdin THEN
    CALL FindStdin
END IF
IF Command.Line = "" THEN
    EXIT SUB
END IF

' filename processing loop
DO
    ' store entire command
    IF LEFT$(Command.Line, 1) = CHR$(34) THEN
        Imbedded = INSTR(2, Command.Line, CHR$(34))
        IF Imbedded THEN
            Command.Work = MID$(Command.Line, 2, Imbedded - 2)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    ELSE
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
    END IF
    Command.Line = LTRIM$(Command.Line)
    Command.Line = RTRIM$(Command.Line)

    ' store current drive
    IF MID$(Command.Work, 2, 1) = ":" THEN
        Drive.Search = LEFT$(Command.Work, 1)
        Command.Work = MID$(Command.Work, 3)
    ELSE
        Drive.Search = Current.Drive$
    END IF
    Drive.Search = UCASE$(Drive.Search)

    ' store current directory
    Directory.Search$ = ""
    Imbedded1 = INSTR(Command.Work, "\")
    Imbedded2 = Imbedded1
    WHILE Imbedded1
        Imbedded2 = Imbedded1
        Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
    WEND
    IF Imbedded2 THEN
        Directory.Search$ = LEFT$(Command.Work, Imbedded2)
        Command.Work = MID$(Command.Work, Imbedded2 + 1)
    END IF
    IF Directory.Search$ = "" THEN
        IF LEFT$(Current.Directory, 2) = "\\" THEN
            Directory.Search$ = Current.Directory
        ELSE
            IF Drive.Search$ = Current.Drive$ THEN
                Directory.Search$ = MID$(Current.Directory, 3)
            ELSE
                Directory.Search$ = "\"
            END IF
        END IF
    END IF
    IF RIGHT$(Directory.Search$, 1) <> "\" THEN
        Directory.Search$ = Directory.Search$ + "\"
    END IF
   
    ' get filename spec
    Filename.Search$ = Command.Work
    IF Filename.Search$ = Nul THEN
        Filename.Search$ = "*.*"
    END IF
    Command.Work = Nul

    ' display search filename
    IF Continuous.Display = False THEN
        COLOR Yellow, Black
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            PRINT "Searching: " + Directory.Search$ + Filename.Search$
        ELSE
            PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
        END IF
    END IF

    ' call routine to search for files
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        CALL FindY_DisplayFileLines(Directory.Search$, Filename.Search$)
    ELSE
        V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
        IF MEDIAEXISTS(V) THEN
            CALL FindY_DisplayFileLines(Directory.Search$, Filename.Search$)
        END IF
    END IF

    ' check search filename
    IF Command.Line = Nul THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    IF List.Lines THEN
        COLOR Yellow, Black
        PRINT "Lines counted: "; FormatString$(Lines.Counted)
    END IF
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "FIND"
PRINT "   Find utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   FIND [d:\path\filename.ext] /[string] [/bcdilnrw]"
PRINT "Where:"
PRINT "   /[string]  is required search string"
PRINT "   /b  boolean search"
PRINT "   /c  continuous list"
PRINT "   /d  display line number"
PRINT "   /i  ignore case"
PRINT "   /l  only list lines counted"
PRINT "   /n###  search at column"
PRINT "   /r  unmatched search"
PRINT "   /t  strip blank lines"
PRINT "   /w  wildcard search"
COLOR Plain, Black
EXIT SUB

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type FIND /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Find utility; "
END IF
RETURN
END SUB

' subroutine to find lines in standard input
SUB FindStdin
COLOR 7, 0
FOR StdinCount = 1 TO StdinLines
    X$ = StdinArray(StdinCount)
    Line.Number = StdinCount
    Input.Line = X$
    IF Ignore.Case THEN
        Input.Line = UCASE$(Input.Line)
    END IF
    Input.Line = MID$(Input.Line, Search.Column)
    CALL Search.Pattern(VarX)
    IF VarX THEN
        IF Reverse.Search = 0 THEN
            CALL FindY_DisplayLine(X$)
        END IF
    ELSE
        IF Reverse.Search THEN
            CALL FindY_DisplayLine(X$)
        END IF
    END IF
NEXT

' display end program
IF Continuous.Display = False THEN
    IF List.Lines THEN
        COLOR Yellow, Black
        PRINT "Lines counted: "; FormatString$(Lines.Counted)
    END IF
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
END SUB

' subroutine to find lines in a filename
SUB FindY_DisplayFileLines (Directory.Search$, Filename.Search$)
Line.Number = 0
' make filename
Filename.Count$ = Directory.Search$ + Filename.Search$
IF _FILEEXISTS(Filename.Count$) THEN
    X = FREEFILE
    OPEN Filename.Count$ FOR BINARY SHARED AS #X
    IF LOF(X) > 0 THEN
        DO UNTIL EOF(X)
            COLOR 7, 0
            LINE INPUT #X, X$
            Line.Number = Line.Number + 1
            Input.Line = X$
            IF Ignore.Case THEN
                Input.Line = UCASE$(Input.Line)
            END IF
            Input.Line = MID$(Input.Line, Search.Column)
            CALL Search.Pattern(VarX)
            IF VarX THEN
                IF Reverse.Search = 0 THEN
                    CALL FindY_DisplayLine(X$)
                END IF
            ELSE
                IF Reverse.Search THEN
                    CALL FindY_DisplayLine(X$)
                END IF
            END IF
        LOOP
    END IF
    CLOSE #X
END IF
END SUB

' search string with wildcard and boolean matching
SUB Search.Pattern (VarX)
VarX = False
IF Boolean.Search THEN
    IF INSTR(Search.String, "&") THEN
        Search.String2$ = Search.String
        CommandLine2$ = Search.String
        First = INSTR(CommandLine2$, "&")
        Search.String = LEFT$(CommandLine2$, First - 1)
        CommandLine2$ = MID$(CommandLine2$, First + 1)
        DO
            GOSUB Get.Match
            IF VarX = False THEN
                EXIT DO
            END IF
            IF CommandLine2$ = Nul THEN
                EXIT DO
            END IF
            First = INSTR(CommandLine2$, "&")
            IF First THEN
                Search.String = LEFT$(CommandLine2$, First - 1)
                CommandLine2$ = MID$(CommandLine2$, First + 1)
            ELSE
                Search.String = CommandLine2$
                CommandLine2$ = Nul
            END IF
        LOOP
        Search.String = Search.String2$
        EXIT SUB
    END IF
END IF
GOSUB Get.Match
EXIT SUB

Get.Match:
VarX = False
IF Search.String = Nul THEN
    IF Input.Line = Nul THEN
        VarX = True
    END IF
    EXIT SUB
END IF
IF WildCard.Search THEN
    GOSUB CheckWildcard
    IF Match THEN
        VarX = True
    END IF
ELSE
    IF INSTR(Input.Line, Search.String) THEN
        VarX = True
    END IF
END IF
EXIT SUB

CheckWildcard:
Match = False
Length1 = LEN(Search.String)
FOR I1 = 1 TO LEN(Input.Line)
    S$ = MID$(Input.Line, I1)
    S$ = LEFT$(S$, Length1)
    IF LEN(S$) <> Length1 THEN
        EXIT FOR
    END IF
    ' compare search string to wildcard
    VarQ = True
    FOR I2 = 1 TO Length1
        IF MID$(Search.String, I2, 1) <> "?" THEN
            IF MID$(Search.String, I2, 1) <> MID$(S$, I2, 1) THEN
                VarQ = False
                EXIT FOR
            END IF
        END IF
    NEXT
    IF VarQ THEN
        Match = True
        EXIT FOR
    END IF
NEXT
EXIT SUB
END SUB

' display search found string
SUB FindY_DisplayLine (X$)
Lines.Counted = Lines.Counted + 1
IF List.Lines = False THEN
    COLOR Yellow, Black
    IF RTRIM$(X$) = "" THEN
        IF Strip.Blanks = 0 THEN
            IF Display.Lines THEN
                X$ = MID$(STR$(Line.Number), 2) + " " + X$
            END IF
            PRINT #StdoutHandle, X$
        END IF
    ELSE
        IF LCASE$(StdoutFile$) = "scrn:" THEN
            IF Display.Lines THEN
                X$ = MID$(STR$(Line.Number), 2) + " " + X$
            END IF
            _CONTROLCHR OFF
            PRINT X$
            _CONTROLCHR ON
        ELSE
            IF Display.Lines THEN
                X$ = MID$(STR$(Line.Number), 2) + " " + X$
            END IF
            PRINT #StdoutHandle, X$
        END IF
    END IF
END IF
END SUB

SUB Delete (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' get attribute bit switches
Dont.Delete.Archive = ParseLine("//A")
Dont.Delete.Hidden = ParseLine("//H")
Dont.Delete.Readonly = ParseLine("//O")
Dont.Delete.System = ParseLine("//S")
Dont.Delete.Any = ParseLine("//X")
Dont.Delete.Compress = ParseLine("//M1")
Dont.Delete.Encrypt = ParseLine("//M2")

Delete.Archive = ParseLine("/A")
Delete.Hidden = ParseLine("/H")
Delete.Readonly = ParseLine("/O")
Delete.System = ParseLine("/S")
Delete.Any = ParseLine("/X")
Delete.Compress = ParseLine("/M1")
Delete.Encrypt = ParseLine("/M2")

' get switches from command line
Continuous.Display = ParseLine("/C")
Prompt.Delete = ParseLine("/P")
Recurse.Directories = ParseLine("/R")
Display.Lowercase = ParseLine("/U")
Display.Errors = ParseLine("/Z")

' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32 + T3!
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32 + T3!
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get file size from command line
Search.File.Size = False
Search.Size.From = FalseD
Search.Size.To = FalseD
Imbedded = INSTR(UCASE$(Command.Line), "/F")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.File.Size = True
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Search.Size.From = Var#
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    GOSUB Get.Numeric
    Search.Size.To = Var#
    Search.Size.From = Search.Size.From * 1024#
    Search.Size.To = Search.Size.To * 1024#
END IF

' get extended date\time switches
Creation.Time = ParseLine("/1")
Access.Time = ParseLine("/2")
Modified.Time = ParseLine("/3")
IF Creation.Time = False THEN
    IF Access.Time = False THEN
        IF Modified.Time = False THEN
            Modified.Time = True
        END IF
    END IF
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' initialize some variables
Continuous.Delete = False
Files.Deleted = FalseD
Total.Bytes = FalseD
Quit.Deleting = False

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = UCASE$(LEFT$(Command.Work, 1))
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive
        END IF

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul
   
        ' display search filename
        IF Continuous.Display = False THEN
            Prompt2$ = Directory.Search$ + Filename.Search$
            IF Display.Lowercase THEN
                Prompt2$ = LCASE$(Prompt2$)
            END IF
            COLOR Yellow, Black
            IF LEFT$(Prompt2$, 2) <> "\\" THEN
                Prompt2$ = Drive.Search + ":" + Prompt2$
            END IF
            Prompt2$ = "Searching: " + Prompt2$
            PRINT Prompt2$
        END IF

        ' call routine to search for files
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Delete_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Delete_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Deleting THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Deleting THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = "Files deleted "
    Total$ = Total$ + FormatString$(Files.Deleted)
    Total$ = Total$ + " Total bytes deleted "
    Total$ = Total$ + FormatString$(Total.Bytes)
    PRINT Total$
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "DELETE"
PRINT "   File delete utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   DELETE [d:\path\filename.ext][//ahmosx][/cdfprtuz]"
PRINT "Where:"
PRINT "   /c  continuous display   /p  do not prompt before deleting"
PRINT "   /r  recurse directories  /u  display files in lowercase"
PRINT "   /z  suppress error messages"
PRINT "   display ranges: (/1  creation, /2  last access  /3  modify time)"
PRINT "     /d  range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "     /t  range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "     /f  range of file size xxx-xxx in kilobytes"
PRINT "   delete file attributes:"
PRINT "     / prefix for files with, // prefix for files without,"
PRINT "       a  archive, h  hidden, o  read-only, s  system, x  none"
PRINT "       m1  compressed, m2  encrypted"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type DELETE /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File delete utility; "
END IF
RETURN
END SUB
   
' subroutine to delete a filename
SUB Delete_DeleteFiles (Directory$, Filename$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260
DIM Attribute AS _UNSIGNED LONG

' get file attributes
Attribute = finddatatemp.dwFileAttributes

' reset delete flag
Delete.File = True

' check for directory
IF (Attribute AND &H10) = &H10 THEN
    EXIT SUB
END IF

' check for read-only file
IF (Attribute AND &H1) = &H1 THEN
    IF Dont.Delete.Readonly THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.Readonly THEN
        EXIT SUB
    END IF
END IF

' check for hidden file
IF (Attribute AND &H2) = &H2 THEN
    IF Dont.Delete.Hidden THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.Hidden THEN
        EXIT SUB
    END IF
END IF

' check for system file
IF (Attribute AND &H4) = &H4 THEN
    IF Dont.Delete.System THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.System THEN
        EXIT SUB
    END IF
END IF

' check for archive file
IF (Attribute AND &H20) = &H20 THEN
    IF Dont.Delete.Archive THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.Archive THEN
        EXIT SUB
    END IF
END IF

' check for compressed file
IF (Attribute AND &H800) = &H800 THEN
    IF Dont.Delete.Compress THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.Compress THEN
        EXIT SUB
    END IF
END IF

' check for encrypted file
IF (Attribute AND &H4000) = &H4000 THEN
    IF Dont.Delete.Encrypt THEN
        EXIT SUB
    END IF
ELSE
    IF Delete.Encrypt THEN
        EXIT SUB
    END IF
END IF

' check all bits
IF Delete.Any THEN
    IF (Attribute AND &H1) = &H1 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H2) = &H2 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4) = &H4 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H20) = &H20 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H800) = &H800 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4000) = &H4000 THEN
        EXIT SUB
    END IF
END IF
IF Dont.Delete.Any THEN
    IF (Attribute AND &H1) = False THEN
        IF (Attribute AND &H2) = False THEN
            IF (Attribute AND &H4) = False THEN
                IF (Attribute AND &H20) = False THEN
                    IF (Attribute AND &H800) = False THEN
                        IF (Attribute AND &H4000) = False THEN
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' check file size
IF Search.File.Size THEN
    IF Search.Size.From = FalseD AND Search.Size.To = FalseD THEN
        IF File.Size <> FalseD THEN
            EXIT SUB
        END IF
    ELSE
        IF Search.Size.From > FalseD OR Search.Size.To > FalseD THEN
            IF Search.Size.From > FalseD THEN
                IF Search.Size.To = FalseD THEN
                    IF File.Size < Search.Size.From THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From = FalseD THEN
                IF Search.Size.To > FalseD THEN
                    IF File.Size > Search.Size.To THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From <= Search.Size.To THEN
                IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
            IF Search.Size.From > Search.Size.To THEN
                IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        Delete.File = False
    END IF
    IF File.Work.Date > Search.To.Date THEN
        Delete.File = False
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        Delete.File = False
    END IF
    IF File.Work.Time > Search.To.Time THEN
        Delete.File = False
    END IF
END IF

' check delete flag
IF Delete.File THEN
    ' check delete type
    IF Continuous.Delete = False THEN
        ' check for prompt to delete
        IF Prompt.Delete THEN
            ' display filename
            Outpt$ = Directory$ + Filename$
            IF Display.Lowercase THEN
                Outpt$ = LCASE$(Outpt$)
            END IF
            COLOR Yellow, Black
            CALL AddStdout2(Outpt$)
        ELSE
            ' prompt to delete
            Prompt2$ = Directory$ + Filename$
            IF Display.Lowercase THEN
                Prompt2$ = LCASE$(Prompt2$)
            END IF
            IF LEFT$(Prompt2$, 2) <> "\\" THEN
                Prompt2$ = Drive.Search + ":" + Prompt2$
            END IF
            Prompt3$ = ""
            IF (Attribute AND &H20) = &H20 THEN
                Prompt3$ = Prompt3$ + "/a"
            END IF
            IF (Attribute AND &H1) = &H1 THEN
                Prompt3$ = Prompt3$ + "/o"
            END IF
            IF (Attribute AND &H2) = &H2 THEN
                Prompt3$ = Prompt3$ + "/h"
            END IF
            IF (Attribute AND &H4) = &H4 THEN
                Prompt3$ = Prompt3$ + "/s"
            END IF
            IF (Attribute AND &H800) = &H800 THEN
                Prompt3$ = Prompt3$ + "/c"
            END IF
            IF (Attribute AND &H4000) = &H4000 THEN
                Prompt3$ = Prompt3$ + "/e"
            END IF
            IF LEN(Prompt3$) THEN
                Prompt2$ = Prompt2$ + "{" + MID$(Prompt3$, 2) + "}"
            END IF
            Prompt2$ = "Delete file " + Prompt2$ + "(y/n/q/c)?"
            CALL MorePrompt(Prompt2$, "ynqc", Outpt$)
            SELECT CASE Outpt$
                CASE "n"
                    Delete.File = False
                CASE "q"
                    Delete.File = False
                    Quit.Deleting = True
                CASE "c"
                    Continuous.Delete = True
            END SELECT
        END IF
    END IF

    ' check delete file flag
    IF Delete.File THEN
        ' make filename
        ASCIIZ = Directory$ + Filename$ + CHR$(0)

        ' change filename attribute
        AttrX& = GetFileAttributes(ASCIIZ)
        AttrX& = AttrX& AND NOT &H1
        x = SetFileAttributes&(ASCIIZ, AttrX&)

        ' delete long filename
        x = DeleteFileA(ASCIIZ)

        ' check error and delete 8.3 filename
        IF x = 0 THEN
            Short.Filename$ = finddatatemp.cAlternateFileName
            V = INSTR(Short.Filename$, CHR$(0))
            IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
            ASCIIZ = Directory$ + Short.Filename$ + CHR$(0)

            ' change filename attribute
            AttrX& = GetFileAttributes(ASCIIZ)
            AttrX& = AttrX& AND NOT &H1
            x = SetFileAttributes&(ASCIIZ, AttrX&)

            ' delete short filename
            x = DeleteFileA(ASCIIZ)
        END IF

        ' display any errors
        IF x = 0 THEN
            Var1$ = "Error 0x" + HEX$(GetLastError) + " deleting filename."
            CALL DisplayError(Var1$)
            CALL DisplayWinError(x)
        ELSE
            ' check display type
            IF Prompt.Delete = False THEN
                IF Continuous.Display = False THEN
                    Prompt2$ = Directory$ + Filename$
                    IF Display.Lowercase THEN
                        Prompt2$ = LCASE$(Prompt2$)
                    END IF
                    COLOR Yellow, Black
                    IF LEFT$(Prompt2$, 2) <> "\\" THEN
                        Prompt2$ = Drive.Search + ":" + Prompt2$
                    END IF
                    Prompt2$ = "Deleting: " + Prompt2$
                    PRINT #StdoutHandle, Prompt2$
                END IF
            END IF
            ' update counters
            Files.Deleted = Files.Deleted + 1#
            Total.Bytes = Total.Bytes + File.Size
        END IF
    END IF
END IF
EXIT SUB

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' subroutine to access directories
SUB Delete_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search directory names
    CALL Delete_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Deleting THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            IF (Attribute AND &H10) = &H10 THEN
                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL Delete_Directories(Next.Directory$, Filename.Search$)
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to delete filenames in directory
SUB Delete_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' recurse filename
    DO

        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN
            ' store file size
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' store file data
            finddatatemp = finddata

            ' delete filename
            CALL Delete_DeleteFiles(Directory.Search$, Filename$)
        END IF

        ' check to quit
        IF Quit.Deleting THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

SUB Treedel (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' get switches from command line
No.Touch.Archive = ParseLine("//A")
No.Touch.Hidden = ParseLine("//H")
No.Touch.Readonly = ParseLine("//O")
No.Touch.System = ParseLine("//S")
No.Touch.Any = ParseLine("//Y")
No.Touch.Compress = ParseLine("//M1")
No.Touch.Encrypt = ParseLine("//M2")

Touch.Archive = ParseLine("/A")
Touch.Hidden = ParseLine("/H")
Touch.Readonly = ParseLine("/O")
Touch.System = ParseLine("/S")
Touch.Any = ParseLine("/Y")
Touch.Compress = ParseLine("/M1")
Touch.Encrypt = ParseLine("/M2")

Continuous.Display = ParseLine("/C")
Short.Display = ParseLine("/D")
Prompt.Delete = ParseLine("/P")
Display.Lowercase = ParseLine("/U")
Add.Slash = ParseLine("/X")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' set searching work variables
Directories.Counted = 0
Quit.Searching = False

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO

        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store search drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search$ = LEFT$(Command.Work, 1)
        ELSE
            Drive.Search$ = Current.Drive
        END IF
        Drive.Search$ = UCASE$(Drive.Search$)

        ' store current directory
        Directory.Search$ = Command.Work

        ' parse drive letter
        Temp.Directory$ = Directory.Search$
        IF MID$(Temp.Directory$, 2, 1) = ":" THEN
            Temp.Drive$ = UCASE$(LEFT$(Temp.Directory$, 1))
            Temp.Directory$ = MID$(Temp.Directory$, 3)
        ELSE
            Temp.Drive$ = Drive.Search$
        END IF

        ' parse directory
        IF LEFT$(Temp.Directory$, 2) <> "\\" THEN
            IF LEFT$(Temp.Directory$, 1) <> "\" THEN
                IF Temp.Drive$ <> LEFT$(Current.Directory, 1) THEN
                    Directory.Search$ = Drive.Search$ + ":\" + Temp.Directory$
                ELSE
                    Directory.Search$ = Current.Directory + Temp.Directory$
                END IF
            ELSE
                Directory.Search$ = Temp.Drive$ + ":" + Temp.Directory$
            END IF
        END IF

        ' display search header
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            Display.Lines = Display.Lines + 1
            PRINT "Searching: " + Directory.Search$
        END IF

        ' call subroutine to search directories
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Treedel_TreeDirectories(Directory.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Temp.Drive$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Treedel_TreeDirectories(Directory.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Deleting THEN
        EXIT DO
    END IF
LOOP
' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Directories.Counted)
    PRINT "Directories counted: "; Total$
END IF
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory delete utility; "
END IF
RETURN

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "TREEDEL"
PRINT "   Directory delete utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   TREEDEL [d:]\path\ [//ahmosy][/cdpuyz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  short file display"
PRINT "   /p  don't prompt before deleting"
PRINT "   /u  display lowercase"
PRINT "   /x  add trailing slash"
PRINT "   /z  suppress error messages"
PRINT "   delete directories with attributes:"
PRINT "     / prefix to delete directories with,"
PRINT "     // prefix to delete directories without,"
PRINT "       a  archive, h  hidden, o  read-only, s  system, y  none"
PRINT "       m1  compressed, m2  encrypted"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type TREEDEL /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

' subroutine to access directories
SUB Treedel_TreeDirectories (Directory.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' loop directories
    DO
        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF

        ' store directory attribute
        Attribute = finddata.dwFileAttributes

        ' check directory
        IF (Attribute AND &H10) = &H10 THEN

            ' check directory attributes
            CALL Treedel_CheckAttribute(Attribute&, Check)
            IF Check = 0 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' make directory
                    FOR Imbedded = LEN(Directory.Search$) TO 1 STEP -1
                        IF MID$(Directory.Search$, Imbedded, 1) = "\" THEN
                            Directory$ = LEFT$(Directory.Search$, Imbedded) + Directory$
                            EXIT FOR
                        END IF
                    NEXT

                    ' store short filename
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                        IF Directory$ = Nul THEN
                            EXIT DO
                        END IF
                    END IF

                    ' prepare directory
                    IF RIGHT$(Directory$, 1) <> "\" THEN
                        Directory$ = Directory$ + "\"
                    END IF

                    ' check to prompt
                    Delete.Flag = False
                    IF Prompt.Delete = False THEN
                        Prompt2$ = Directory$
                        IF Display.Lowercase THEN
                            Prompt2$ = LCASE$(Prompt2$)
                        END IF
                        Prompt2$ = "Delete subdirectories in " + Prompt2$
                        Prompt3$ = ""
                        IF (Attribute AND &H20) = &H20 THEN
                            Prompt3$ = Prompt3$ + "/a"
                        END IF
                        IF (Attribute AND &H1) = &H1 THEN
                            Prompt3$ = Prompt3$ + "/o"
                        END IF
                        IF (Attribute AND &H2) = &H2 THEN
                            Prompt3$ = Prompt3$ + "/h"
                        END IF
                        IF (Attribute AND &H4) = &H4 THEN
                            Prompt3$ = Prompt3$ + "/s"
                        END IF
                        IF (Attribute AND &H800) = &H800 THEN
                            Prompt3$ = Prompt3$ + "/c"
                        END IF
                        IF (Attribute AND &H4000) = &H4000 THEN
                            Prompt3$ = Prompt3$ + "/e"
                        END IF
                        IF LEN(Prompt3$) THEN
                            Prompt2$ = Prompt2$ + "{" + MID$(Prompt3$, 2) + "}"
                        END IF
                        Prompt2$ = Prompt2$ + "(y/n/c/q)?"
                        CALL MorePrompt(Prompt2$, "yncq", Outpt$)
                        SELECT CASE Outpt$
                            CASE "c"
                                Prompt.Delete = True
                            CASE "n"
                                Delete.Flag = True
                            CASE "q"
                                Quit.Searching = True
                                EXIT DO
                        END SELECT
                    END IF

                    ' check to continue deleteing
                    IF Delete.Flag = False THEN
                        ' store directory filename
                        Temp.Dir$ = LEFT$(Directory$, LEN(Directory$) - 1)

                        ' display directory
                        IF Short.Display THEN
                            IF LEFT$(Temp.Dir$, 2) <> "\\" THEN
                                Temp.Dir$ = MID$(Temp.Dir$, 3)
                            END IF
                        END IF
                        IF Display.Lowercase THEN
                            Temp.Dir$ = LCASE$(Temp.Dir$)
                        END IF
                        IF Add.Slash THEN
                            Temp.Dir$ = Temp.Dir$ + "\"
                        END IF
                        COLOR Yellow, Black
                        IF Continuous.Display = False THEN
                            IF CurrentCommand = CommandCount THEN
                                Outpt$ = "Deleting: " + Temp.Dir$
                                PRINT #StdoutHandle, Outpt$
                            ELSE
                                CALL AddStdout2(Temp.Dir$)
                            END IF
                        ELSE
                            CALL AddStdout(Temp.Dir$)
                        END IF

                        ' routine to delete directories
                        CALL Treedel_Directories(Directory$)
                    END IF
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access subdirectories
SUB Treedel_Directories (Directory.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' delete filenames
    CALL Treedel_DeleteFiles(Directory.Search$)

    ' recurse subdirectories
    DO
        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF

        ' check directory attribute
        Attribute = finddata.dwFileAttributes

        ' check directory
        IF (Attribute AND &H10) = &H10 THEN

            ' store directory name
            Directory$ = finddata.cFileName
            Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

            ' check directory name
            IF Directory$ <> "." AND Directory$ <> ".." THEN

                ' store short filename
                IF INSTR(Directory$, "?") THEN
                    Directory$ = finddata.cAlternateFileName
                    V = INSTR(Directory$, CHR$(0))
                    IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                END IF

                ' make next search directory
                IF LEN(Directory$) THEN
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL Treedel_Directories(Next.Directory$)
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF

' delete directory
CALL Treedel_DeleteDirectory(Directory.Search$)
END SUB

' subroutine to delete an empty directory
SUB Treedel_DeleteDirectory (Directory$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260

' store directory filename
Var$ = Directory$
Var$ = LEFT$(Var$, LEN(Var$) - 1)
ASCIIZ = Var$ + CHR$(0)

' change directory attribute
AttrX& = GetFileAttributes(ASCIIZ)
AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
x = SetFileAttributes&(ASCIIZ, AttrX&)

' delete directory
x = RemoveDirectoryA(ASCIIZ)

' count directories deleted
IF x THEN
    Directories.Counted = Directories.Counted + 1
END IF

' display any errors
IF x = 0 THEN
    IF GetLastError = &H91 THEN ' ignore directory not empty
        Eat$ = Nul
    ELSE
        Var1$ = "Error 0x" + HEX$(GetLastError) + " deleting directory: " + Var$
        CALL DisplayError(Var1$)
        CALL DisplayWinError(x)
    END IF
END IF
END SUB

' subroutine to delete files in a directory
SUB Treedel_DeleteFiles (Directory$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory$ + "*.*" + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' filename loop
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN
            ' store filename
            ASCIIZ2 = Directory$ + Filename$ + CHR$(0)

            ' change filename attribute
            AttrX& = GetFileAttributes(ASCIIZ2)
            AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
            x = SetFileAttributes&(ASCIIZ2, AttrX&)

            ' delete long filename
            x = DeleteFileA(ASCIIZ2)

            ' check error and delete 8.3 filename
            IF x = 0 THEN
                Short.Filename$ = finddata.cAlternateFileName
                V = INSTR(Short.Filename$, CHR$(0))
                IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
                ASCIIZ2 = Directory$ + Short.Filename$ + CHR$(0)

                ' change filename attribute
                AttrX& = GetFileAttributes(ASCIIZ2)
                AttrX& = AttrX& AND NOT &H1
                x = SetFileAttributes&(ASCIIZ2, AttrX&)

                ' delete short filename
                x = DeleteFileA(ASCIIZ2)
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

SUB Treedel_CheckAttribute (Attribute&, Check)
Check = False
' check for readonly file
IF Touch.Readonly THEN
    IF (Attribute& AND &H1) <> &H1 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Readonly THEN
    IF (Attribute& AND &H1) = &H1 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for hidden file
IF Touch.Hidden THEN
    IF (Attribute& AND &H2) <> &H2 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Hidden THEN
    IF (Attribute& AND &H2) = &H2 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for system file
IF Touch.System THEN
    IF (Attribute& AND &H4) <> &H4 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.System THEN
    IF (Attribute& AND &H4) = &H4 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for archive file
IF Touch.Archive THEN
    IF (Attribute& AND &H20) <> &H20 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Archive THEN
    IF (Attribute& AND &H20) = &H20 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for compress file
IF Touch.Compress THEN
    IF (Attribute& AND &H800) <> &H800 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Compress THEN
    IF (Attribute& AND &H800) = &H800 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for encyrpted file
IF Touch.Encrypt THEN
    IF (Attribute& AND &H4000) <> &H4000 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Encrypt THEN
    IF (Attribute& AND &H4000) = &H4000 THEN
        Check = True
        EXIT SUB
    END IF
END IF

' check for no attributes
IF Touch.Any THEN
    IF (Attribute& AND &H1) = &H1 THEN
        Check = True
        EXIT SUB
    END IF
    IF (Attribute& AND &H2) = &H2 THEN
        Check = True
        EXIT SUB
    END IF
    IF (Attribute& AND &H4) = &H4 THEN
        Check = True
        EXIT SUB
    END IF
    IF (Attribute& AND &H20) = &H20 THEN
        Check = True
        EXIT SUB
    END IF
    IF (Attribute& AND &H800) = &H800 THEN
        Check = True
        EXIT SUB
    END IF
    IF (Attribute& AND &H4000) = &H4000 THEN
        Check = True
        EXIT SUB
    END IF
END IF
IF No.Touch.Any THEN
    IF (Attribute& AND &H1) = &H0 THEN
        IF (Attribute& AND &H2) = &H0 THEN
            IF (Attribute& AND &H4) = &H0 THEN
                IF (Attribute& AND &H20) = &H0 THEN
                    IF (Attribute& AND &H800) = &H0 THEN
                        IF (Attribute& AND &H4000) = &H0 THEN
                            Check = True
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF
END SUB

' subroutine to access subdirectories
SUB Silentdel_Directories (Directory.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' delete filenames
    CALL Silentdel_DeleteFiles(Directory.Search$)

    ' recurse subdirectories
    DO
        ' check directory attribute
        Attribute = finddata.dwFileAttributes

        ' check directory
        IF (Attribute AND &H10) = &H10 THEN

            ' store directory name
            Directory$ = finddata.cFileName
            Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

            ' check directory name
            IF Directory$ <> "." AND Directory$ <> ".." THEN

                ' store short filename
                IF INSTR(Directory$, "?") THEN
                    Directory$ = finddata.cAlternateFileName
                    V = INSTR(Directory$, CHR$(0))
                    IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                END IF

                ' make next search directory
                IF LEN(Directory$) THEN
                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL Silentdel_Directories(Next.Directory$)
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF

' delete directory
CALL Silentdel_DeleteDirectory(Directory.Search$)
END SUB

' subroutine to delete an empty directory
SUB Silentdel_DeleteDirectory (Directory$)
' declare subroutine variables
DIM ASCIIZ AS STRING * 260

' store directory filename
ASCIIZ = Directory$ + CHR$(0)

' change directory attribute
AttrX& = GetFileAttributes(ASCIIZ)
AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
x = SetFileAttributes&(ASCIIZ, AttrX&)

' delete directory
x = RemoveDirectoryA(ASCIIZ)
END SUB

' subroutine to delete files in a directory
SUB Silentdel_DeleteFiles (Directory$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM ASCIIZ2 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory$ + "*.*" + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' filename loop
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN
            ' store filename
            ASCIIZ2 = Directory$ + Filename$ + CHR$(0)

            ' change filename attribute
            AttrX& = GetFileAttributes(ASCIIZ2)
            AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
            x = SetFileAttributes&(ASCIIZ2, AttrX&)

            ' delete long filename
            x = DeleteFileA(ASCIIZ2)

            ' check error and delete 8.3 filename
            IF x = 0 THEN
                Short.Filename$ = finddata.cAlternateFileName
                V = INSTR(Short.Filename$, CHR$(0))
                IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
                ASCIIZ2 = Directory$ + Short.Filename$ + CHR$(0)

                ' change filename attribute
                AttrX& = GetFileAttributes(ASCIIZ2)
                AttrX& = AttrX& AND NOT &H1
                x = SetFileAttributes&(ASCIIZ2, AttrX&)

                ' delete short filename
                x = DeleteFileA(ASCIIZ2)
            END IF
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

SUB Drives (ErrFlag)
' dimension work variables
REDIM Skip.Drives(1 TO 26) AS INTEGER

' get current drive
Default.Drive = ASC(LEFT$(_CWD$, 1)) - 64

' get maximum drives
Last.Drive = 26

' get command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = UCASE$(COMMAND2$)
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    COLOR White, Black
    PRINT "DRIVES"
    PRINT "   Drive letter display utility; "
    GOSUB Boot.Usage
    COLOR Plain, Black
    EXIT SUB
END IF

' check command line switches
Skip.Drives(1) = ParseLine("/A")
Skip.Drives(2) = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Display.Colon = ParseLine("/W")
Display.Current = ParseLine("/X")
Display.Errors = ParseLine("/Z")

' get skip drives list
FOR Var = 3 TO 26
    Skip.Drives(Var) = ParseLine("/1:" + CHR$(Var + 64))
NEXT

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' check command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF LEN(Command.Line) THEN
    GOTO Boot.Error
END IF

' make header
IF Continuous.Display = False THEN
    IF Display.Header = False THEN
        COLOR White, Black
        PRINT "Drive letter display utility;"
    END IF
END IF

' display drives
COLOR Yellow, Black
IF Display.Current THEN
    IF LEFT$(_CWD$, 2) <> "\\" THEN
        IF Skip.Drives(Default.Drive) = False THEN
            Outpt$ = CHR$(Default.Drive + 64)
            IF Display.Colon = 0 THEN
                Outpt$ = Outpt$ + ":"
            END IF
            CALL AddStdout2(Outpt$)
        END IF
    END IF
ELSE
    FOR Drives.Count = 1 TO Last.Drive
        IF Skip.Drives(Drives.Count) = False THEN
            IF DRIVEEXISTS(Drives.Count) = 0 THEN
                Outpt$ = CHR$(Drives.Count + 64)
                IF Display.Colon = 0 THEN
                    Outpt$ = Outpt$ + ":"
                END IF
                CALL AddStdout2(Outpt$)
            END IF
        END IF
    NEXT
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   DRIVES [/abcwxz1]"
PRINT "Where:"
PRINT "   /a  ignore drive A:"
PRINT "   /b  ignore drive B:"
PRINT "   /c  continuous display"
PRINT "   /w  don't display colon"
PRINT "   /x  list only current drive"
PRINT "   /z  suppress errors"
PRINT "   /1:d  skip drive d (d is C to Z)"
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type DRIVES /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

SUB Label (ErrFlag)

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOTO Boot.Usage
END IF

' get switches from command line
Continuous.Display = ParseLine("/C")
Display.Drive = ParseLine("/D")
Display.Volume = ParseLine("/X")
Display.Errors = ParseLine("/Z")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' display header
Header.Flag = False
GOSUB Header

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    DO
        ' store entire command
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Standard.Input$ + Command.Line
            Command.Line = Nul
        END IF

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' display drive searched
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            PRINT "Searching: "; Drive.Search$; ":"
        END IF

        ' check drive exists
        V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
        IF MEDIAEXISTS(V) THEN
            ' store new volume label
            Volume.Label$ = Command.Work + CHR$(0)
            Drive$ = Drive.Search$ + ":\" + CHR$(0)

            ' change volume label
            x = SetVolumeLabelA%(Drive$, Volume.Label$)

            ' check volume flag
            IF x = 0 THEN
                Temp$ = "Error 0x" + HEX$(GetLastError) + " naming volume."
                CALL DisplayError(Temp$)
                CALL DisplayWinError(x)
            ELSE
                GOSUB Display.Label
            END IF
        ELSE
            Temp$ = "Error accessing drive " + Drive.Search$ + ":\"
            CALL DisplayError(Temp$)
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    Prompt2$ = "Press <enter> to exit to DOS:"
    CALL MorePrompt(Prompt2$, CHR$(13), Outpt$)
END IF
COLOR Plain, Black
EXIT SUB

' display volume label
Display.Label:
' display drive letter
Outpt$ = Nul
IF Display.Drive THEN
    Outpt$ = Drive.Search + ":"
END IF
' display volume label
IF Display.Volume = False THEN
    Outpt$ = Outpt$ + Volume.Label$
END IF
IF LEN(Outpt$) THEN
    COLOR Yellow, Black
    CALL AddStdout2(Outpt$)
END IF
RETURN

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "LABEL"
PRINT "   Volume update utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   LABEL [d:][volumename][/cdxz]"
PRINT "Where:"
PRINT "   /c  continuous display"
PRINT "   /d  display drive letter"
PRINT "   /x  don't list volume label"
PRINT "   /z  suppress errors"
COLOR Plain, Black
EXIT SUB

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type LABEL /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Volume update utility;"
END IF
RETURN
END SUB

SUB ListVols (ErrFlag)
' dimension work variables
REDIM Skip.Drives(1 TO 26) AS INTEGER

' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Default.Drive = ASC(LEFT$(_CWD$, 1)) - 64

' reset counter
Header.Flag = 0
Volumes.Counted = 0

' check command line switches
Skip.Drives(1) = ParseLine("/A")
Skip.Drives(2) = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Display.Serial = ParseLine("/E")
Display.FatType = ParseLine("/F")
Display.Volumes = ParseLine("/W")
Display.Drive.Letter = ParseLine("/Y")
Display.Errors = ParseLine("/Z")

' get skip drives list
FOR Var = 3 TO 26
    Skip.Drives(Var) = ParseLine("/1:" + CHR$(Var + 64))
NEXT

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' make header
GOSUB Header

' check command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
Command.Line = UCASE$(Command.Line)

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF
    IF Standard.Input$ = "" THEN
        Command.Line = Command.Line.Redirect$
    ELSE
        Command.Line = Standard.Input$
    END IF

    ' filename processing loop
    DO
        Imbedded = INSTR(Command.Line, " ")
        IF Imbedded THEN
            Command.Work = LEFT$(Command.Line, Imbedded - 1)
            Command.Line = MID$(Command.Line, Imbedded + 1)
        ELSE
            Command.Work = Command.Line
            Command.Line = Nul
        END IF
        Command.Work = RTRIM$(Command.Work)
        Command.Work = LTRIM$(Command.Work)
        Command.Work = UCASE$(Command.Work)

        ' get drive letter
        Drive.Number = 0
        IF Command.Work = "" THEN
            IF LEFT$(_CWD$, 2) = "\\" THEN
                PRINT #StdoutHandle, _CWD$
                EXIT DO
            ELSE
                Drive.Number = Default.Drive
            END IF
        ELSE
            IF LEN(Command.Work) = 2 THEN
                IF RIGHT$(Command.Work, 1) = ":" THEN
                    Drive.Number = ASC(LEFT$(Command.Work, 1)) - 64
                END IF
            END IF
        END IF
        IF Drive.Number >= 1 AND Drive.Number <= 26 THEN
            ' display volume
            GOSUB Get.Volume.Label
        ELSE
            ' display any errors
            CALL DisplayError("Error accessing drive.")
            EXIT SUB
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    PRINT "Volumes counted"; Volumes.Counted
END IF
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Volume display utility;"
END IF
RETURN

' display volume information for drive number
Get.Volume.Label:
' check skip drive letter array
IF Skip.Drives(Drive.Number) THEN
    RETURN
END IF

' check drive
IF DRIVEEXISTS(Drive.Number) THEN
    RETURN
END IF

' increment counter
Volumes.Counted = Volumes.Counted + 1

' display drive letter
COLOR Yellow, Black
Outpt$ = Nul
IF Display.Drive.Letter = False THEN
    Outpt$ = CHR$(Drive.Number + 64) + ":"
END IF

' display volume label
Out3 = CHR$(Drive.Number + 64)
CALL Vlabel(Out3)
IF RTRIM$(Out3) = Nul THEN
    Outpt$ = Outpt$ + DriveType
ELSE
    Outpt$ = Outpt$ + RTRIM$(Out3)
END IF

IF CurrentCommand = CommandCount THEN
    PRINT #StdoutHandle, Outpt$ + " ";
ELSE
    CALL AddStdout2(Outpt$)
END IF

' display volume serial number
IF Display.Serial THEN
    COLOR Red, Black
    Out3 = CHR$(Drive.Number + 64)
    CALL Vserial(Out3)
    PRINT #StdoutHandle, Out3 + " ";
END IF

' display volume file system type
IF Display.FatType THEN
    COLOR White, Black
    Out3 = CHR$(Drive.Number + 64)
    CALL Vtype(Out3)
    PRINT #StdoutHandle, Out3;
END IF
PRINT #StdoutHandle, ""
RETURN

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "VOL"
PRINT "   Volume display utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   VOL [d:][/abcefwyz1]"
PRINT "Where:"
PRINT "   d:  list drive D: volume only"
PRINT "   /a  ignore drive A: volume"
PRINT "   /b  ignore drive B: volume"
PRINT "   /c  continuous display"
PRINT "   /e  display volume serial number"
PRINT "   /f  display volume file system type"
PRINT "   /w  don't list invalid volumes"
PRINT "   /y  don't display drive letter"
PRINT "   /z  suppress errors"
PRINT "   /1:d  skip drive d (d is C to Z)"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type VOL /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

' get drive freespace
SUB FreeSpace (Var$)
VarX$ = Var$ + ":\" + CHR$(0)
Var$ = Nul
IF DriveType = "[CDROM]" THEN
    EXIT SUB
END IF
IF DriveType = "[REMOVABLE]" THEN
    EXIT SUB
END IF
r = GetDiskFreeSpaceExA(VarX$, free~&&, total~&&, free2~&&)
IF r THEN
    Var$ = LTRIM$(STR$(free~&&))
END IF
EXIT SUB

r = GetDiskFreeSpaceA(VarX$, sectors&, bytes&, free&, total&)
IF r THEN
    ' sectors per cluster * bytes per sector * free clusters
    x1# = CDBL(sectors&) * CDBL(bytes&) * CDBL(free&)
    Var$ = LTRIM$(STR$(x1#))
END IF
END SUB

' get drive totalspace
SUB TotalSpace (Var$)
VarX$ = Var$ + ":\" + CHR$(0)
Var$ = Nul
IF DriveType = "[CDROM]" THEN
    EXIT SUB
END IF
IF DriveType = "[REMOVABLE]" THEN
    EXIT SUB
END IF
r = GetDiskFreeSpaceExA(VarX$, free~&&, total~&&, free2~&&)
IF r THEN
    Var$ = LTRIM$(STR$(total~&&))
END IF
EXIT SUB

r = GetDiskFreeSpaceA(VarX$, sectors&, bytes&, free&, total&)
IF r THEN
    ' sectors per cluster * bytes per sector * total clusters
    x1# = CDBL(sectors&) * CDBL(bytes&) * CDBL(total&)
    Var$ = LTRIM$(STR$(x1#))
END IF
END SUB

' get volume label
SUB Vlabel (Var$)
' Note: in DOS the volume label was 8.3 format,
'  however, in windows XP+ it is 32 char.

' get drive info.
VarX$ = Var$ + ":\" + CHR$(0)
Var$ = Nul
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    ' get volume label.
    Var$ = RTRIM$(Vname$)
    v = INSTR(Var$, CHR$(0))
    IF v THEN Var$ = LEFT$(Var$, v - 1)
END IF
END SUB

' get volume serial number
SUB Vserial (Var$)

' get drive info.
VarX$ = Var$ + ":\" + CHR$(0)
Var$ = Nul
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    ' serial number.
    Var$ = LEFT$(HEX$(serial~&), 4) + "-" + RIGHT$(HEX$(serial~&), 4)
END IF
END SUB

' get volume system type
SUB Vtype (Var$)

' get drive info.
VarX$ = Var$ + ":\" + CHR$(0)
Var$ = Nul
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    ' get volume system type.
    Var$ = RTRIM$(Fname$)
    v = INSTR(Var$, CHR$(0))
    IF v THEN Var$ = LEFT$(Var$, v - 1)
END IF
END SUB

SUB Encrypt (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset count variables
Crypted.Count = 0#
Crypted.Count2 = 0#
Dirs.Counted = 0#
Files.Counted = 0#
Display.Lines = False
More.Display = False
Continuous.Display = False
Quit.Searching = False

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' store command line
Command.Line = RTRIM$(Command.Line)

' check command line switches
Set.Archive = ParseLine("//A")
Set.Hidden = ParseLine("//H")
Set.Readonly = ParseLine("//O")
Set.System = ParseLine("//S")
Clear.Archive = ParseLine("/A")
Clear.Hidden = ParseLine("/H")
Clear.Readonly = ParseLine("/O")
Clear.System = ParseLine("/S")
Continuous.Display = ParseLine("/C")
Recurse.Directories = ParseLine("/R")
Display.Filenames = ParseLine("/X")
Display.Errors = ParseLine("/Z")
More.Display = Continuous.Display

' get nested switch from command line
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive$ THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul
   
        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            IF LEFT$(Directory.Search$, 2) = "\\" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
            Lines.Counted = Lines.Counted + 1
        END IF

        ' call subroutine to search directories
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Encrypt_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Encrypt_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Dirs.Counted)
    PRINT "Directories counted " + Total$
    Total$ = FormatString$(Files.Counted)
    PRINT "Files counted " + Total$
    IF Encrypt.Call = 0 THEN
        Total$ = FormatString$(Crypted.Count2)
        PRINT "Directories encrypted " + Total$
        Total$ = FormatString$(Crypted.Count)
        PRINT "Files encrypted " + Total$
    ELSE
        Total$ = FormatString$(Crypted.Count2)
        PRINT "Directories decrypted " + Total$
        Total$ = FormatString$(Crypted.Count)
        PRINT "Files decrypted " + Total$
    END IF
END IF
COLOR Plain, Black
EXIT SUB

Boot.Usage:
' make header
COLOR White, Black
IF Encrypt.Call = 0 THEN
    PRINT "ENCRYPT"
    PRINT "   File/directory encrypt utility; "
ELSE
    PRINT "DECRYPT"
    PRINT "   File/directory decrypt utility; "
END IF
COLOR Yellow, Black
IF Encrypt.Call = 0 THEN
    PRINT "Usage:"
    PRINT "   Encrypt [d:][\path\]filename.ext [//ahos][/cnrxz]"
ELSE
    PRINT "Usage:"
    PRINT "   Decrypt [d:][\path\]filename.ext [//ahos][/cnrxz]"
END IF
PRINT "Where:"
PRINT "   Set file bits:"
PRINT "     //a  archive   //h  hidden   //o  read-only   //s  system"
PRINT "   Clear file bits:"
PRINT "     /a  archive    /h  hidden    /o  read-only    /s  system"
PRINT "Remaining switches:"
PRINT "     /c  continuous display       /x  don't display filename"
PRINT "     /r  recurse directories      /z  don't display errors"
PRINT "     /n### override nested directories"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
IF Encrypt.Call = 0 THEN
    PRINT "Command line error. Type ENCRYPT /? for help."
ELSE
    PRINT "Command line error. Type DECRYPT /? for help."
END IF
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    IF Encrypt.Call = 0 THEN
        PRINT "File/directory encrypt utility;"
    ELSE
        PRINT "File/directory decrypt utility;"
    END IF
END IF
RETURN
END SUB

' subroutine to access directories
SUB Encrypt_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search directory names
    CALL Encrypt_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Searching THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Encrypt_Directories(Next.Directory$, Filename.Search$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Encrypt_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search filenames
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check unicode
        IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
            Filename$ = finddata.cAlternateFileName
            V = INSTR(Filename$, CHR$(0))
            IF V THEN Filename$ = LEFT$(Filename$, V - 1)
        END IF

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' store file data
            finddatatemp = finddata

            ' display filename
            CALL Encrypt_DisplayFiles(Directory.Search$, Filename$)
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to encrypt a filename
SUB Encrypt_DisplayFiles (Search.Directory$, Display.Filename$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260

' make filename
Search.Filename$ = Search.Directory$ + Display.Filename$

' check file attribute
Attribute = finddatatemp.dwFileAttributes

' check directory bit
IF (Attribute AND &H10) = &H10 THEN
    Directory.Set = True
ELSE
    Directory.Set = False
END IF

' clear directory bit to set directory attribute
Attribute = Attribute AND NOT &H10

' check to clear read-only bit
IF Clear.Readonly THEN
    Attribute = Attribute AND NOT &H1
END IF

' check to clear hidden bit
IF Clear.Hidden THEN
    Attribute = Attribute AND NOT &H2
END IF

' check to clear system bit
IF Clear.System THEN
    Attribute = Attribute AND NOT &H4
END IF

' check to clear archive bit
IF Clear.Archive THEN
    Attribute = Attribute AND NOT &H20
END IF

' check to set read-only bit
IF Set.Readonly THEN
    Attribute = Attribute OR &H1
END IF

' check to set hidden bit
IF Set.Hidden THEN
    Attribute = Attribute OR &H2
END IF

' check to set system bit
IF Set.System THEN
    Attribute = Attribute OR &H4
END IF

' check to set archive bit
IF Set.Archive THEN
    Attribute = Attribute OR &H20
END IF

' change filename attribute
ASCIIZ = Search.Filename$ + CHR$(0)
Old.Attribute& = Attribute
Attribute = Attribute AND NOT &H1 ' reset read-only bit
x = SetFileAttributes&(ASCIIZ, Attribute)

' encrypt filename
ASCIIZ = Search.Filename$ + CHR$(0)
IF Encrypt.Call = 0 THEN
    V = EncryptFileA(ASCIIZ)
ELSE
    V = DecryptFileA(ASCIIZ, 0)
END IF

' count files reset
IF Directory.Set THEN
    Dirs.Counted = Dirs.Counted + 1#
ELSE
    Files.Counted = Files.Counted + 1#
END IF

' display filename
IF Display.Filenames = False THEN
    IF Directory.Set THEN
        Outpt$ = Drive.Search + ":" + Search.Filename$
    ELSE
        IF LEFT$(Search.Directory$, 2) <> "\\" THEN
            Outpt$ = Drive.Search + ":" + Search.Filename$
        ELSE
            Outpt$ = Search.Filename$
        END IF
    END IF
    COLOR Yellow, Black
    CALL AddStdout2(Outpt$)

    ' display error accessing file
    IF V = 0 THEN
        IF Encrypt.Call = 0 THEN
            CALL DisplayError("Error 0x" + HEX$(GetLastError) + " encrypting file.")
            CALL DisplayWinError(x)
        ELSE
            CALL DisplayError("Error 0x" + HEX$(GetLastError) + " decrypting file.")
            CALL DisplayWinError(x)
        END IF
    END IF

    ' check line length
    Length = LEN(Outpt$)
    Lines.Counted = Lines.Counted + INT(Length / 80) + 1
    CALL PagePrompt(Lines.Counted, FlagX)
    IF FlagX THEN
        Lines.Counted = 0
    END IF
END IF

' restore filename attribute
ASCIIZ = Search.Filename$ + CHR$(0)
Attribute = Old.Attribute& ' restore attribute
x = SetFileAttributes&(ASCIIZ, Attribute)

' increment counters
IF V THEN
    IF Directory.Set THEN
        Crypted.Count2 = Crypted.Count2 + 1#
    ELSE
        Crypted.Count = Crypted.Count + 1#
    END IF
END IF
END SUB

SUB Dirattr (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset count variables
Dirs.Counted = 0#
Files.Counted = 0#
Errors.Counted = 0#
Display.Lines = False
Continuous.Display = False
Lines.Counted = False
Quit.Searching = False

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' store command line
Command.Line = RTRIM$(Command.Line)

' get command line switches
No.List.Archive = ParseLine("//A")
No.List.Hidden = ParseLine("//H")
No.List.Directory = ParseLine("//I")
No.List.Readonly = ParseLine("//O")
No.List.System = ParseLine("//S")
No.List.Compressed = ParseLine("//J")
No.List.Encrypted = ParseLine("//K")

List.Archive = ParseLine("/A")
List.Hidden = ParseLine("/H")
List.Directory = ParseLine("/I")
List.Readonly = ParseLine("/O")
List.System = ParseLine("/S")
List.Compressed = ParseLine("/J")
List.Encrypted = ParseLine("/K")

List.Filename = ParseLine("/E")
Continuous.Display = ParseLine("/C")
List.Lowercase = ParseLine("/U")
Recurse.Directories = ParseLine("/R")
Display.Filenames = ParseLine("/X")
Display.Descrip = ParseLine("/Y")
Display.Errors = ParseLine("/Z")
Short.File = ParseLine("/W2")
Short.Attr = ParseLine("/W")
More.Display = Continuous.Display

' get nested switch from command line
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive$ THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul
   
        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            IF LEFT$(Directory.Search$, 2) = "\\" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
            Lines.Counted = Lines.Counted + 1
        END IF

        ' call routine to search for files
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Dirattr_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Dirattr_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Dirs.Counted)
    PRINT "Directories counted " + Total$
    Total$ = FormatString$(Files.Counted)
    PRINT "Files counted " + Total$
END IF
COLOR Plain, Black
EXIT SUB

Boot.Usage:
' make header
COLOR White, Black
PRINT "GETATTR"
PRINT "   File/directory attribute display utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   GETATTR [[d:\path\]filename.ext][//ahijkos][/cdenqruwxyz]"
PRINT "Where:"
PRINT "   Don't list files with bits:"
PRINT "     //a  archive  //h  hidden  //i  directory  //o  read-only"
PRINT "     //s  system   //j  compressed   //k  encrypted"
PRINT "   Only list files with bits:"
PRINT "     /a  archive   /h  hidden   /i  directory   /o  read-only"
PRINT "     /s  system    /j  compressed   /k  encrypted"
PRINT "   Remaining switches:"
PRINT "     /c  continuous display     /u  display files in lowercase"
PRINT "     /d  prepend drive letter   /x  don't display filename"
PRINT "     /e  don't display drive    /y  don't display attributes"
PRINT "     /r  recurse directories    /z  don't display errors"
PRINT "     /w  short attribute display  /w2  strip pathname"
PRINT "     /n###  override nested directories"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type GETATTR /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File/directory attribute display utility;"
END IF
RETURN
END SUB

' subroutine to access directories
SUB Dirattr_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search directory names
    CALL Dirattr_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Searching THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Dirattr_Directories(Next.Directory$, Filename.Search$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Dirattr_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search filenames
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' check unicode
            IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
                Filename$ = finddata.cAlternateFileName
                V = INSTR(Filename$, CHR$(0))
                IF V THEN Filename$ = LEFT$(Filename$, V - 1)
            END IF

            ' store file data
            finddatatemp = finddata

            ' display filename
            CALL Dirattr_DisplayFiles(Directory.Search$, Filename$)
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to display a filename description
SUB Dirattr_DisplayFiles (Search.Directory$, Display.Filename$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG

' make filename
Search.Filename$ = Search.Directory$ + Display.Filename$

' check file attribute
Attribute = finddatatemp.dwFileAttributes

' get full pathname
Var$ = Search.Filename$
IF LEFT$(Var$, 2) <> "\\" THEN
    IF List.Filename = 0 THEN
        Var$ = Drive.Search + ":" + Var$
    END IF
END IF

' store list work variable
List.File = True

' check list attribute
IF List.Readonly THEN
    IF (Attribute AND &H1) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.Hidden THEN
    IF (Attribute AND &H2) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.System THEN
    IF (Attribute AND &H4) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.Directory THEN
    IF (Attribute AND &H10) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.Archive THEN
    IF (Attribute AND &H20) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.Compressed THEN
    IF (Attribute AND &H800) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF List.Encrypted THEN
    IF (Attribute AND &H4000) = &H0 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Readonly THEN
    IF (Attribute AND &H1) = &H1 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Hidden THEN
    IF (Attribute AND &H2) = &H2 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.System THEN
    IF (Attribute AND &H4) = &H4 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Directory THEN
    IF (Attribute AND &H10) = &H10 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Archive THEN
    IF (Attribute AND &H20) = &H20 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Compressed THEN
    IF (Attribute AND &H800) = &H800 THEN
        List.File = False
    END IF
END IF

' check list attribute
IF No.List.Encrypted THEN
    IF (Attribute AND &H4000) = &H4000 THEN
        List.File = False
    END IF
END IF

' check list work variable
IF List.File THEN

    ' count files displayed
    IF (Attribute AND &H10) = &H10 THEN
        Dirs.Counted = Dirs.Counted + 1#
    ELSE
        Files.Counted = Files.Counted + 1#
    END IF

    ' calculate lines counted
    IF Display.Filenames THEN
        IF Display.Descrip = False THEN
            Flag = False
            IF (Attribute AND &H1) = &H1 THEN ' read-only
                Flag = True
            END IF
            IF (Attribute AND &H2) = &H2 THEN ' hidden
                Flag = True
            END IF
            IF (Attribute AND &H4) = &H4 THEN ' system
                Flag = True
            END IF
            IF (Attribute AND &H10) = &H10 THEN ' directory
                Flag = True
            END IF
            IF (Attribute AND &H20) = &H20 THEN ' archive
                Flag = True
            END IF
            IF (Attribute AND &H800) = &H800 THEN ' compressed
                Flag = True
            END IF
            IF (Attribute AND &H4000) = &H4000 THEN ' encrypted
                Flag = True
            END IF
            IF Flag THEN
                Lines.Counted = Lines.Counted + 1
            END IF
        END IF
    END IF

    ' display filename
    IF Display.Filenames = False THEN
        IF LEFT$(Search.Directory$, 2) <> "\\" THEN
            IF List.Filename = False THEN
                IF Short.File = 0 THEN
                    Outpt$ = Drive.Search + ":" + Search.Filename$
                ELSE
                    Outpt$ = Search.Filename$
                END IF
            ELSE
                Outpt$ = Search.Filename$
            END IF
            IF Short.File THEN
                FOR V = LEN(Outpt$) TO 1 STEP -1
                    IF MID$(Outpt$, V, 1) = "\" THEN
                        Outpt$ = MID$(Outpt$, V + 1)
                        EXIT FOR
                    END IF
                NEXT
            END IF
        ELSE
            Outpt$ = Search.Filename$
        END IF
        Outpt$ = RTRIM$(Outpt$)

        ' calculate line length
        Outpt.Length = LEN(Outpt$)
        IF Display.Descrip = False THEN
            Outpt.Length = Outpt.Length + 1
            IF (Attribute AND &H1) = &H1 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Read-only ")
                END IF
            END IF
            IF (Attribute AND &H2) = &H2 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Hidden ")
                END IF
            END IF
            IF (Attribute AND &H4) = &H4 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("System ")
                END IF
            END IF
            IF (Attribute AND &H10) = &H10 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Directory ")
                END IF
            END IF
            IF (Attribute AND &H20) = &H20 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Archive ")
                END IF
            END IF
            IF (Attribute AND &H800) = &H800 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Compressed ")
                END IF
            END IF
            IF (Attribute AND &H4000) = &H4000 THEN
                IF Short.Attr THEN
                    Outpt.Length = Outpt.Length + 2
                ELSE
                    Outpt.Length = Outpt.Length + LEN("Encrypted ")
                END IF
            END IF
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT SUB
        END IF

        ' display filename
        COLOR Yellow, Black
        IF List.Lowercase THEN
            IF CurrentCommand = CommandCount THEN
                PRINT #StdoutHandle, LCASE$(Outpt$);
            ELSE
                CALL AddStdout2(LCASE$(Outpt$))
            END IF
        ELSE
            IF CurrentCommand = CommandCount THEN
                PRINT #StdoutHandle, Outpt$;
            ELSE
                CALL AddStdout2(Outpt$)
            END IF
        END IF
        IF CurrentCommand < CommandCount THEN
            EXIT SUB
        END IF

        ' check suppress display
        IF Display.Descrip THEN
            IF CurrentCommand = CommandCount THEN
                PRINT #StdoutHandle, ""
            END IF
            ' prompt for more
            Lines.Counted = Lines.Counted + INT(Outpt.Length / 80) + 1
            CALL PagePrompt(Lines.Counted, FlagX)
            IF FlagX THEN
                Lines.Counted = 0
            END IF
            EXIT SUB
        END IF
        IF CurrentCommand = CommandCount THEN
            PRINT #StdoutHandle, " ";
        END IF
    END IF

    ' check suppress display
    IF Display.Descrip = False THEN

        ' display attributes
        COLOR White, Black
        Flag = False

        ' check for read-only file
        IF (Attribute AND &H1) = &H1 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "O ";
            ELSE
                PRINT #StdoutHandle, "Read-only ";
            END IF
            Flag = True
        END IF

        ' check for hidden file
        IF (Attribute AND &H2) = &H2 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "H ";
            ELSE
                PRINT #StdoutHandle, "Hidden ";
            END IF
            Flag = True
        END IF

        ' check for system file
        IF (Attribute AND &H4) = &H4 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "S ";
            ELSE
                PRINT #StdoutHandle, "System ";
            END IF
            Flag = True
        END IF

        ' check for directory file
        IF (Attribute AND &H10) = &H10 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "D ";
            ELSE
                PRINT #StdoutHandle, "Directory ";
            END IF
            Flag = True
        END IF

        ' check for archive file
        IF (Attribute AND &H20) = &H20 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "A ";
            ELSE
                PRINT #StdoutHandle, "Archive ";
            END IF
            Flag = True
        END IF

        ' check for compressed file
        IF (Attribute AND &H800) = &H800 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "C ";
            ELSE
                PRINT #StdoutHandle, "Compressed ";
            END IF
            Flag = True
        END IF

        ' check for encrypted file
        IF (Attribute AND &H4000) = &H4000 THEN
            IF Short.Attr THEN
                PRINT #StdoutHandle, "E ";
            ELSE
                PRINT #StdoutHandle, "Encrypted ";
            END IF
            Flag = True
        END IF

        ' check display
        IF Display.Filenames = False THEN
            PRINT #StdoutHandle, ""
        ELSE
            IF Flag THEN
                PRINT #StdoutHandle, ""
            END IF
        END IF

        ' prompt for more
        Lines.Counted = Lines.Counted + INT(Outpt.Length / 80) + 1
        CALL PagePrompt(Lines.Counted, FlagX)
        IF FlagX THEN
            Lines.Counted = 0
        END IF
    END IF
END IF
END SUB

SUB Setattr (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' reset count variables
Dirs.Counted = 0#
Files.Counted = 0#
Display.Lines = False
More.Display = False
Continuous.Display = False
Lines.Counted = False
Quit.Searching = False

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' store command line
Command.Line = RTRIM$(Command.Line)

' check command line switches
Set.Archive = ParseLine("//A")
Set.Hidden = ParseLine("//H")
Set.Readonly = ParseLine("//O")
Set.System = ParseLine("//S")
Clear.Archive = ParseLine("/A")
Clear.Hidden = ParseLine("/H")
Clear.Readonly = ParseLine("/O")
Clear.System = ParseLine("/S")
Continuous.Display = ParseLine("/C")
Set.Directory = ParseLine("/I")
Only.Set.Directory = ParseLine("/J")
Recurse.Directories = ParseLine("/R")
Display.Filenames = ParseLine("/X")
Display.Errors = ParseLine("/Z")
More.Display = Continuous.Display

' get nested switch from command line
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive$ THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul
   
        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            IF LEFT$(Directory.Search$, 2) = "\\" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
            Lines.Counted = Lines.Counted + 1
        END IF

        ' call routine to search for files
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Setattr_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Setattr_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    Total$ = FormatString$(Dirs.Counted)
    PRINT "Directories counted " + Total$
    Total$ = FormatString$(Files.Counted)
    PRINT "Files counted " + Total$
END IF
COLOR Plain, Black
EXIT SUB

Boot.Usage:
' make header
COLOR White, Black
PRINT "SETATTR"
PRINT "   File/directory attribute change utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   SETATTR [d:][\path\]filename.ext [//ahos][/ij][/cnrxz]"
PRINT "Where:"
PRINT "   Set file bits:"
PRINT "     //a  archive   //h  hidden   //o  read-only   //s  system"
PRINT "   Clear file bits:"
PRINT "     /a  archive    /h  hidden    /o  read-only    /s  system"
PRINT "   Directory bits:"
PRINT "     /i  set directory attributes  /j  only set directory attributes"
PRINT "Remaining switches:"
PRINT "     /c  continuous display        /x  don't display filename"
PRINT "     /r  recurse directories       /z  don't display errors"
PRINT "     /n### override nested directories"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type SETATTR /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File/directory attribute change utility;"
END IF
RETURN
END SUB

' subroutine to access directories
SUB Setattr_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search directory names
    CALL Setattr_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Searching THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN

                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Setattr_Directories(Next.Directory$, Filename.Search$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Setattr_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' search filenames
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' check unicode
            IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
                Filename$ = finddata.cAlternateFileName
                V = INSTR(Filename$, CHR$(0))
                IF V THEN Filename$ = LEFT$(Filename$, V - 1)
            END IF

            ' store file data
            finddatatemp = finddata

            ' display filename
            CALL Setattr_DisplayFiles(Directory.Search$, Filename$)
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to change a filename attribute
SUB Setattr_DisplayFiles (Search.Directory$, Display.Filename$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260

' make filename
Search.Filename$ = Search.Directory$ + Display.Filename$

' check file attribute
Attribute = finddatatemp.dwFileAttributes

' check directory bit
IF (Attribute AND &H10) = &H10 THEN
    Directory.Set = True
ELSE
    Directory.Set = False
END IF

' check to set directories
IF (Attribute AND &H10) <> &H10 THEN
    IF Only.Set.Directory THEN
        EXIT SUB
    END IF
END IF
IF (Attribute AND &H10) = &H10 THEN
    IF Set.Directory = False AND Only.Set.Directory = False THEN
        EXIT SUB
    END IF
END IF

' clear directory bit to set directory attribute
Attribute = Attribute AND NOT &H10

' check to clear read-only bit
IF Clear.Readonly THEN
    Attribute = Attribute AND NOT &H1
END IF

' check to clear hidden bit
IF Clear.Hidden THEN
    Attribute = Attribute AND NOT &H2
END IF

' check to clear system bit
IF Clear.System THEN
    Attribute = Attribute AND NOT &H4
END IF

' check to clear archive bit
IF Clear.Archive THEN
    Attribute = Attribute AND NOT &H20
END IF

' check to set read-only bit
IF Set.Readonly THEN
    Attribute = Attribute OR &H1
END IF

' check to set hidden bit
IF Set.Hidden THEN
    Attribute = Attribute OR &H2
END IF

' check to set system bit
IF Set.System THEN
    Attribute = Attribute OR &H4
END IF

' check to set archive bit
IF Set.Archive THEN
    Attribute = Attribute OR &H20
END IF

' change filename attribute
ASCIIZ = Search.Filename$ + CHR$(0)
x = SetFileAttributes&(ASCIIZ, Attribute)

' check set error
IF x = 0 THEN
    Short.Filename$ = finddatatemp.cAlternateFileName
    V = INSTR(Short.Filename$, CHR$(0))
    IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
    ' make filename
    Search.Filename$ = Search.Directory$ + Short.Filename$

    ' change filename attribute
    ASCIIZ = Search.Filename$ + CHR$(0)
    x = SetFileAttributes&(ASCIIZ, Attribute)
END IF

' count files reset
IF x THEN
    IF Directory.Set THEN
        Dirs.Counted = Dirs.Counted + 1#
    ELSE
        Files.Counted = Files.Counted + 1#
    END IF
END IF

' display filename
IF Display.Filenames = False THEN
    IF LEFT$(Search.Directory$, 2) <> "\\" THEN
        Outpt$ = Drive.Search + ":" + Search.Filename$
    ELSE
        Outpt$ = Search.Filename$
    END IF
    COLOR Yellow, Black
    CALL AddStdout2(Outpt$)

    ' display any errors
    IF x = 0 THEN
        CALL DisplayError("Error 0x" + HEX$(GetLastError) + " setting attribute.")
        CALL DisplayWinError(x)
    END IF

    ' calculate length of display
    Outpt.Length = LEN(Outpt$)
    Lines.Counted = Lines.Counted + INT(Outpt.Length / 80) + 1
    CALL PagePrompt(Lines.Counted, FlagX)
    IF FlagX THEN
        Lines.Counted = 0
    END IF
END IF
END SUB

SUB Touch (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' check command line switches
No.Display.Archive = ParseLine("//A")
No.Display.Hidden = ParseLine("//H")
No.Display.Readonly = ParseLine("//O")
No.Display.System = ParseLine("//S")
No.Display.Compress = ParseLine("//M1")
No.Display.Encrypt = ParseLine("//M2")
No.Display.Any = ParseLine("//X")

Display.Archive = ParseLine("/A")
Display.Hidden = ParseLine("/H")
Display.Readonly = ParseLine("/O")
Display.System = ParseLine("/S")
Display.Compress = ParseLine("/M1")
Display.Encrypt = ParseLine("/M2")
Display.Any = ParseLine("/X")

Continuous.Display = ParseLine("/C")
Display.Search = ParseLine("/J")
Prompt.Files = ParseLine("/P")
Recurse.Directories = ParseLine("/R")
Prepend.Drive = ParseLine("/U")
Display.Filenames = ParseLine("/V")
Remove.Slash = ParseLine("/Y")
Display.Errors = ParseLine("/Z")

' read override switches.
Override.Create.Date = ParseLine("/7")
Override.Access.Date = ParseLine("/8")
Override.Modify.Date = ParseLine("/9")

' reset file counter variable
Continue = False
Files.Counted = 0#
   
' read date override switch
File.Date$ = DATE$
Imbedded = INSTR(UCASE$(Command.Line), "/5")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    File.Date$ = MID$(Command.Line, Imbedded + 2, 10)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 12)
    IF LEN(File.Date$) <> 10 THEN
        GOTO Boot.Error
    END IF
END IF

' store date values
MonthTemp = VAL(MID$(File.Date$, 1, 2))
DayTemp = VAL(MID$(File.Date$, 4, 2))
YearTemp = VAL(MID$(File.Date$, 7, 4))

' read time override switch
File.Time$ = TIME$
Imbedded = INSTR(UCASE$(Command.Line), "/6")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    File.Time$ = MID$(Command.Line, Imbedded + 2, 8)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 10)
    IF LEN(File.Time$) <> 8 THEN
        GOTO Boot.Error
    END IF
END IF

' store time values hh:mm:ss
HourTemp = VAL(MID$(File.Time$, 1, 2))
MinuteTemp = VAL(MID$(File.Time$, 4, 2))
SecondTemp = VAL(MID$(File.Time$, 7, 2))

' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get file size from command line
Search.File.Size = False
Search.Size.From = FalseD
Search.Size.To = FalseD
Imbedded = INSTR(UCASE$(Command.Line), "/F")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.File.Size = True
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Search.Size.From = Var#
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    GOSUB Get.Numeric
    Search.Size.To = Var#
    Search.Size.From = Search.Size.From * 1024#
    Search.Size.To = Search.Size.To * 1024#
END IF

' get extended date\time switches
Creation.Time = ParseLine("/1")
Access.Time = ParseLine("/2")
Modified.Time = ParseLine("/3")
IF Creation.Time = False THEN
    IF Access.Time = False THEN
        IF Modified.Time = False THEN
            Modified.Time = True
        END IF
    END IF
END IF

' check command line switch
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' parse special debug switch.
Debug.Mode = ParseLine("/=")

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' reset work variables
Files.Counter = False
Quit.Searching = False
   
' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive$ THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul

        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            IF LEFT$(Directory.Search$, 2) = "\\" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
        END IF

        ' call routine to search for files
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Touch_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Touch_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    PRINT "Files touched ";
    PRINT FormatString$(Files.Counted)
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "TOUCHFILE"
PRINT "   File date/time update utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   TOUCHFILE [d:][\pathname\][filename][//ahmosx][/cdfjnprtuvyz][/123][/56789]"
PRINT "Where:"
PRINT "   d:\pathname\filename is the search directory."
PRINT "   /c  continuous display       /u  prepend drive letter"
PRINT "   /j  display directories      /v  don't display filenames"
PRINT "   /nxxx  recurse override      /y  remove trailing slash"
PRINT "   /p  don't prompt to touch    /z  suppress error messages"
PRINT "   /r  recurse directories      /=  set debug mode"
PRINT "   /0  create file if not exist"
PRINT "   file override switches:"
PRINT "     /5mm/dd/yyyy  override touch file date"
PRINT "     /6hh:mm:ss    override touch file time"
PRINT "     /7  touch creation date/time"
PRINT "     /8  touch last access date/time"
PRINT "     /9  touch last modified (write) date/time"
PRINT "   display ranges: (/1  creation, /2  last access  /3  modify time)"
PRINT "     /d  is range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "     /t  is range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "     /f  is range of file sizes in form xxx-xxx in kilobytes"
PRINT "   display file attributes:"
PRINT "     / prefix for files with, // prefix for files without,"
PRINT "       a  archive, h  hidden, o  read-only, s  system, x  none"
PRINT "       m1  compressed, m2  encrypted"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type TOUCHFILE /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "File date/time update utility; "
END IF
RETURN
END SUB

' subroutine to access directories
SUB Touch_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make directory filename
ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search directory names
    CALL Touch_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            IF (Attribute AND &H10) = &H10 THEN

                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN
                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL Touch_Directories(Next.Directory$, Filename.Search$)
                END IF
            END IF

            ' check to quit
            IF Quit.Searching THEN
                EXIT DO
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Touch_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' check directory attribute
    Attribute = finddata.dwFileAttributes

    ' check for directory
    IF (Attribute AND &H10) = &H10 THEN
        ' display directory being searched
        CALL Touch_DisplayDirectory(Directory.Search$)
    END IF

    ' recurse filename
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' check unicode
            IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
                Filename$ = finddata.cAlternateFileName
                V = INSTR(Filename$, CHR$(0))
                IF V THEN Filename$ = LEFT$(Filename$, V - 1)
            END IF

            ' store file data
            finddatatemp = finddata

            ' store file size
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) <> &H10 THEN
                CALL Touch_DisplayFiles(Directory.Search$, Filename$)
            END IF
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to display directory
SUB Touch_DisplayDirectory (Directory.Search$)
IF Display.Search THEN
    Filename$ = Directory.Search$
    IF Filename$ <> "\" THEN
        IF Remove.Slash THEN
            Slash = LEN(Filename$)
            Filename$ = LEFT$(Filename$, Slash - 1)
        END IF
    END IF
    IF Prepend.Drive THEN
        IF LEFT$(Filename$, 2) <> "\\" THEN
            Filename$ = Drive.Search + ":" + Filename$
        END IF
    END IF
    COLOR Green, Black
    CALL AddStdout2(Filename$)
END IF
END SUB

' subroutine to count and display filenames
SUB Touch_DisplayFiles (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG

' check directory attribute
Attribute = finddatatemp.dwFileAttributes

' check for readonly file
IF (Attribute AND &H1) = &H1 THEN
    IF No.Display.Readonly THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Readonly THEN
        EXIT SUB
    END IF
END IF

' check for hidden file
IF (Attribute AND &H2) = &H2 THEN
    IF No.Display.Hidden THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Hidden THEN
        EXIT SUB
    END IF
END IF

' check for system file
IF (Attribute AND &H4) = &H4 THEN
    IF No.Display.System THEN
        EXIT SUB
    END IF
ELSE
    IF Display.System THEN
        EXIT SUB
    END IF
END IF

' check for archive file
IF (Attribute AND &H20) = &H20 THEN
    IF No.Display.Archive THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Archive THEN
        EXIT SUB
    END IF
END IF

' check for compressed file
IF (Attribute AND &H800) = &H800 THEN
    IF No.Display.Compress THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Compress THEN
        EXIT SUB
    END IF
END IF

' check for encrypted file
IF (Attribute AND &H4000) = &H4000 THEN
    IF No.Display.Encrypt THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Encrypt THEN
        EXIT SUB
    END IF
END IF

' check all attributes
IF Display.Any THEN
    IF (Attribute AND &H1) = &H1 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H2) = &H2 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4) = &H4 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H20) = &H20 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H800) = &H800 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4000) = &H4000 THEN
        EXIT SUB
    END IF
END IF
IF No.Display.Any THEN
    IF (Attribute AND &H1) = False THEN
        IF (Attribute AND &H2) = False THEN
            IF (Attribute AND &H4) = False THEN
                IF (Attribute AND &H20) = False THEN
                    IF (Attribute AND &H800) = False THEN
                        IF (Attribute AND &H4000) = False THEN
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' check file size
IF Search.File.Size THEN
    IF Search.Size.From = FalseD AND Search.Size.To = FalseD THEN
        IF File.Size <> FalseD THEN
            EXIT SUB
        END IF
    ELSE
        IF Search.Size.From > FalseD OR Search.Size.To > FalseD THEN
            IF Search.Size.From > FalseD THEN
                IF Search.Size.To = FalseD THEN
                    IF File.Size < Search.Size.From THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From = FalseD THEN
                IF Search.Size.To > FalseD THEN
                    IF File.Size > Search.Size.To THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From <= Search.Size.To THEN
                IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
            IF Search.Size.From > Search.Size.To THEN
                IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        EXIT SUB
    END IF
    IF File.Work.Date > Search.To.Date THEN
        EXIT SUB
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        EXIT SUB
    END IF
    IF File.Work.Time > Search.To.Time THEN
        EXIT SUB
    END IF
END IF

' touch filename
CALL TouchFile(Directory.Search$, Filename.Search$)
EXIT SUB

Convert.Date:
YearTempX! = SysTime.wYear
MonthTempX! = SysTime.wMonth
DayTempX! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTempX! - 1980) * 512) + MonthTempX! * 32 + DayTempX!
RETURN

Convert.Time:
HourTempX! = SysTime.wHour
MinuteTempX! = SysTime.wMinute
SecondsTempX! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTempX! * 2048 + MinuteTempX! * 32 + SecondsTempX!
RETURN
END SUB

' subroutine to touch date/time of a file
SUB TouchFile (Directory.Search$, Filename.Search$)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

' check for prompting
IF Prompt.Files = False THEN
    Prompt2$ = "Touch " + Filename.Search$ + "(y/n/c/q)?"
    CALL MorePrompt(Prompt2$, "yncq", Outpt$)
    SELECT CASE Outpt$
        CASE "c"
            Prompt.Files = True
        CASE "n"
            EXIT SUB
        CASE "q"
            Quit.Searching = True
            EXIT SUB
    END SELECT
END IF

' display filename
IF Display.Filenames = 0 THEN
    Filename$ = Directory.Search$ + Filename.Search$
    IF Prepend.Drive THEN
        IF LEFT$(Filename$, 2) <> "\\" THEN
            Filename$ = Drive.Search + ":" + Filename$
        END IF
    END IF
    COLOR Yellow, Black
    CALL AddStdout2(Filename$)
END IF

' get source filename date/time.
hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, 0, 0)
IF hfind <> INVALID_HANDLE_VALUE THEN
    x = GetFileTime&(hfind, SourceCreateDate, SourceAccessDate, SourceModifyDate)
    ' display any errors.
    IF x = 0 THEN
        IF Debug.Mode THEN
            Var$ = "Error (0x" + HEX$(GetLastError&) + ") reading filename date/time."
            CALL DisplayError(Var$)
            CALL DisplayWinError(x)
        END IF
    END IF

    ' store default date/time.
    CreateDate = SourceCreateDate
    AccessDate = SourceAccessDate
    ModifyDate = SourceModifyDate

    ' store override date/time.
    GOSUB GetSysDate
    IF Override.Create.Date THEN
        x = SystemTimeToFileTime&(SysTime, CreateDate)
    END IF
    IF Override.Access.Date THEN
        x = SystemTimeToFileTime&(SysTime, AccessDate)
    END IF
    IF Override.Modify.Date THEN
        x = SystemTimeToFileTime&(SysTime, ModifyDate)
    END IF

    ' set the destination filename date/time.
    x = SetFileTime&(hfind, CreateDate, AccessDate, ModifyDate)
    ' display any errors.
    IF x = 0 THEN
        IF Debug.Mode THEN
            Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename date/time."
            CALL DisplayError(Var$)
            CALL DisplayWinError(x)
        END IF
    END IF
    x = CloseHandle(hfind)
ELSE
    Var$ = "Error (0x" + HEX$(GetLastError&) + ") opening filename."
    CALL DisplayError(Var$)
    CALL DisplayWinError(x)
END IF

' update files counter
Files.Counted = Files.Counted + 1#
EXIT SUB

GetSysDate:
' store override variables
SysTime.wMonth = MonthTemp
SysTime.wDay = DayTemp
SysTime.wYear = YearTemp
SysTime.wHour = HourTemp
SysTime.wMinute = MinuteTemp
SysTime.wSecond = SecondTemp
RETURN
END SUB

SUB TouchDir (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive/directory
Current.Drive = LEFT$(_CWD$, 1)
Current.Directory = _CWD$
IF RIGHT$(Current.Directory, 1) <> "\" THEN
    Current.Directory = Current.Directory + "\"
END IF

' reset counters
Dirs.Counted = False
Display.Lines = False
Continuous.Display = False
Quit.Searching = False

' check command line switches
Search.Archive = ParseLine("/A")
Search.Hidden = ParseLine("/H")
Search.Readonly = ParseLine("/O")
Search.System = ParseLine("/S")
Search.Compress = ParseLine("/J")
Search.Encrypt = ParseLine("/K")

Display.Filenames = ParseLine("/B")
Continuous.Display = ParseLine("/C")
Short.Display = ParseLine("/E")
Truncate.Slash = ParseLine("/G")
Prompt.Files = ParseLine("/P")
Strip.Drive = ParseLine("/Q")
Recurse.Directories = ParseLine("/R")
Prepend.Drive = ParseLine("/U")
Short.Filenames = ParseLine("/V")
Display.Lowercase = ParseLine("/Y")
Display.Errors = ParseLine("/Z")

' read date override switch
File.Date$ = DATE$
Imbedded = INSTR(UCASE$(Command.Line), "/5")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    File.Date$ = MID$(Command.Line, Imbedded + 2, 10)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 12)
    IF LEN(File.Date$) <> 10 THEN
        GOTO Boot.Error
    END IF
END IF

' store date values
MonthTemp = VAL(MID$(File.Date$, 1, 2))
DayTemp = VAL(MID$(File.Date$, 4, 2))
YearTemp = VAL(MID$(File.Date$, 7, 4))

' read time override switch
File.Time$ = TIME$
Imbedded = INSTR(UCASE$(Command.Line), "/6")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    File.Time$ = MID$(Command.Line, Imbedded + 2, 8)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 10)
    IF LEN(File.Time$) <> 8 THEN
        GOTO Boot.Error
    END IF
END IF

' store time values hh:mm:ss
HourTemp = VAL(MID$(File.Time$, 1, 2))
MinuteTemp = VAL(MID$(File.Time$, 4, 2))
SecondTemp = VAL(MID$(File.Time$, 7, 2))

' read override switches.
Override.Create.Date = ParseLine("/7")
Override.Access.Date = ParseLine("/8")
Override.Modify.Date = ParseLine("/9")

' get nested switch from command line
Nested.Levels = 256%
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = Var%
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF

        ' store search drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
        ELSE
            Drive.Search = Current.Drive
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = Command.Work

        ' parse drive letter
        Temp.Directory$ = Directory.Search$
        IF MID$(Temp.Directory$, 2, 1) = ":" THEN
            Temp.Drive$ = UCASE$(LEFT$(Temp.Directory$, 1))
            Temp.Directory$ = MID$(Temp.Directory$, 3)
        ELSE
            Temp.Drive$ = Drive.Search
        END IF

        ' parse directory
        IF LEFT$(Temp.Directory$, 2) <> "\\" THEN
            IF LEFT$(Temp.Directory$, 1) <> "\" THEN
                IF Temp.Drive$ <> LEFT$(Current.Directory, 1) THEN
                    Directory.Search$ = Drive.Search + ":\" + Temp.Directory$
                ELSE
                    Directory.Search$ = Current.Directory + Temp.Directory$
                END IF
            ELSE
                Directory.Search$ = Temp.Drive$ + ":" + Temp.Directory$
            END IF
        END IF

        ' display search header
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            Display.Lines = Display.Lines + 1
            PRINT "Searching: " + Directory.Search$
        END IF

        ' call subroutine to search directories
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Touchdir_Directories(Directory.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Temp.Drive$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Touchdir_Directories(Directory.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display counters
IF Continuous.Display = False THEN
    COLOR Yellow, Black
    PRINT "Directories counted ";
    PRINT FormatString$(Dirs.Counted)
END IF
COLOR Plain, Black
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory date/time update utility;"
END IF
RETURN

Get.Numeric:
Var% = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var% = Var% * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "TOUCHDIR"
PRINT "   Directory date/time update utility;"
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   TOUCHDIR [d:]\path\ [/ahosjk][/cdptxz][/56789]"
PRINT "Where:"
PRINT "   /b  don't display files   /p  don't prompt to touch"
PRINT "   /c  continuous display    /q  strip drive letter"
PRINT "   /e  don't display path    /u  prepend drive letter"
PRINT "   /g  remove slash          /v  display short filenames"
PRINT "   /r  recurse directories   /y  display lowercase file"
PRINT "   /z  suppress errors       /nxxx  recurse levels override"
PRINT "   file override switches:"
PRINT "     /5mm/dd/yyyy  override touch file date"
PRINT "     /6hh:mm:ss    override touch file time"
PRINT "     /7  touch creation date/time"
PRINT "     /8  touch last access date/time"
PRINT "     /9  touch last modified (write) date/time"
PRINT "   Only touch files with bits:"
PRINT "     /a  archive   /h  hidden   /o  read-only   /s  system"
PRINT "     /j  compressed  /k  encrypted"
COLOR Plain, Black
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type TOUCHDIR /? for help."
COLOR Plain, Black
ErrFlag = -1
END SUB

' subroutine to access directories
SUB Touchdir_Directories (Directory.Search$)
' declare subroutine variables
'  local only to this subroutine for recursion.
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET
DIM Recurse.Temp AS INTEGER

' reset wildcard flag
Wild = 0

' make directory filename
Temp.Dir$ = Directory.Search$
Recurse.Temp = Recurse.Directories
IF Recurse.Directories THEN
    IF INSTR(Temp.Dir$, "?") OR INSTR(Temp.Dir$, "*") THEN
        ' Recurse.Temp = False
    ELSE
        IF RIGHT$(Temp.Dir$, 1) <> "\" THEN
            Temp.Dir$ = Temp.Dir$ + "\"
        END IF
        Temp.Dir$ = Temp.Dir$ + "*.*"
    END IF
END IF
ASCIIZ = Temp.Dir$ + CHR$(0)

' find first long filename
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))

' check findirst error
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN

    ' filename/directory loop
    DO
        ' get file attributes
        Attribute = finddata.dwFileAttributes

        ' check directory attribute
        IF (Attribute AND &H10) = &H10 THEN

            ' check to recurse directories
            IF Recurse.Temp = 0 THEN
                ' store directory name
                Directory$ = finddata.cFileName
                V = INSTR(Directory$, CHR$(0))
                IF V THEN Directory$ = LEFT$(Directory$, V - 1)

                ' display directory
                GOSUB Display.Directory
            END IF

            ' check to recurse directories
            IF Recurse.Temp THEN

                ' store directory name
                Directory$ = finddata.cFileName
                V = INSTR(Directory$, CHR$(0))
                IF V THEN Directory$ = LEFT$(Directory$, V - 1)

                ' check directory name
                Valid.Flag = -1
                IF Directory$ = "." THEN
                    Valid.Flag = 0
                END IF
                IF Directory$ = ".." THEN
                    Valid.Flag = 0
                END IF
                IF Directory$ = Nul THEN
                    Valid.Flag = 0
                END IF
                IF Valid.Flag THEN

                    ' display directory
                    GOSUB Display.Directory

                    ' check recursion levels
                    Recursion% = True
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels + 1
                        IF Nested.Levels >= Nested.Recurse THEN
                            Recursion% = False
                        END IF
                    END IF

                    ' recursively search subdirectories
                    IF Recursion% THEN
                        CALL Touchdir_Directories(Next.Directory$)
                    END IF
                    IF Nested.Recurse > False THEN
                        Nested.Levels = Nested.Levels - 1
                    END IF
                END IF
            END IF
        END IF

        ' check to quit
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
EXIT SUB

Display.Directory:
' retract one directory if wildcard/global characters in trailing path
IF INSTR(Directory.Search$, "?") OR INSTR(Directory.Search$, "*") THEN
    Wild = -1
    FOR V = LEN(Directory.Search$) TO 1 STEP -1
        IF MID$(Directory.Search$, V, 1) = "\" THEN
            Directory.Search$ = LEFT$(Directory.Search$, V)
            EXIT FOR
        END IF
    NEXT
END IF

' make next search directory
IF RIGHT$(Directory.Search$, 1) <> "\" THEN
    Next.Directory$ = Directory.Search$ + "\" + Directory$
ELSE
    Next.Directory$ = Directory.Search$ + Directory$
END IF

' store directory name
Temp.Directory$ = finddata.cFileName
V = INSTR(Temp.Directory$, CHR$(0))
IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)

IF Temp.Directory$ = Nul THEN
    RETURN
END IF
IF Temp.Directory$ = "." THEN
    RETURN
END IF
IF Temp.Directory$ = ".." THEN
    RETURN
END IF

' construct directory for display
IF Recurse.Directories = 0 THEN
    IF Wild THEN
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Temp.Directory$ = Directory.Search$ + "\" + Directory$
        ELSE
            Temp.Directory$ = Directory.Search$ + Directory$
        END IF
    ELSE
        Temp.Directory$ = Directory.Search$
    END IF
ELSE
    Temp.Directory$ = Next.Directory$
    Wild = -1
END IF

' get short filename
IF Short.Filenames THEN
    Temp.Directory$ = finddata.cAlternateFileName
    V = INSTR(Temp.Directory$, CHR$(0))
    IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)
    IF Temp.Directory$ = Nul THEN
        Temp.Directory$ = finddata.cFileName
        V = INSTR(Temp.Directory$, CHR$(0))
        IF V THEN Temp.Directory$ = LEFT$(Temp.Directory$, V - 1)
    END IF
END IF

' check directory attribute
Valid.Attribute = True

' check for read-only file
IF Search.Readonly THEN
    IF (Attribute AND &H1) <> &H1 THEN
        Valid.Attribute = False
    END IF
END IF

' check for hidden file
IF Search.Hidden THEN
    IF (Attribute AND &H2) <> &H2 THEN
        Valid.Attribute = False
    END IF
END IF

' check for system file
IF Search.System THEN
    IF (Attribute AND &H4) <> &H4 THEN
        Valid.Attribute = False
    END IF
END IF

' check for archive file
IF Search.Archive THEN
    IF (Attribute AND &H20) <> &H20 THEN
        Valid.Attribute = False
    END IF
END IF

' check for compressed file
IF Search.Compress THEN
    IF (Attribute AND &H800) <> &H800 THEN
        Valid.Attribute = False
    END IF
END IF

' check for compressed file
IF Search.Encrypt THEN
    IF (Attribute AND &H4000) <> &H4000 THEN
        Valid.Attribute = False
    END IF
END IF

' check for valid directory
IF Valid.Attribute THEN

    ' store directory name
    Outpt$ = RTRIM$(Temp.Directory$)
    IF Short.Display THEN
        IF RIGHT$(Outpt$, 1) = "\" THEN
            Outpt$ = LEFT$(Outpt$, LEN(Outpt$) - 1)
        END IF
        FOR Imbedded = LEN(Outpt$) TO 1 STEP -1
            IF MID$(Outpt$, Imbedded, 1) = "\" THEN
                Outpt$ = MID$(Outpt$, Imbedded + 1)
                EXIT FOR
            END IF
        NEXT
    END IF
    IF Truncate.Slash THEN
        IF Outpt$ <> "\" THEN
            IF RIGHT$(Outpt$, 1) = "\" THEN
                Outpt$ = LEFT$(Outpt$, LEN(Outpt$) - 1)
            END IF
        END IF
    END IF
    IF Short.Filenames = False OR Short.Display THEN
        IF MID$(Outpt$, 2, 1) <> ":" THEN
            IF LEFT$(Directory.Search$, 2) <> "\\" THEN
                Outpt$ = Drive.Search + ":" + Outpt$
            END IF
        END IF
    END IF
    IF LEFT$(Directory.Search$, 2) = "\\" THEN
        IF MID$(Outpt$, 2, 1) = ":" THEN
            Outpt$ = MID$(Outpt$, 3)
        END IF
    END IF
    IF Strip.Drive THEN
        IF MID$(Outpt$, 2, 1) = ":" THEN
            Outpt$ = MID$(Outpt$, 3)
        END IF
    END IF
    IF Prepend.Drive THEN
        IF LEFT$(Outpt$, 2) <> "\\" THEN
            IF MID$(Outpt$, 2, 1) <> ":" THEN
                Outpt$ = Drive.Search + ":" + Outpt$
            END IF
        END IF
    END IF
    IF Display.Lowercase THEN
        Outpt$ = LCASE$(Outpt$)
    END IF
    IF Append.Slash THEN
        IF RIGHT$(Outpt$, 1) <> "\" THEN
            Outpt$ = Outpt$ + "\"
        END IF
    END IF

    ' check slash switch
    IF Append.Slash THEN
        IF RIGHT$(Outpt$, 1) <> "\" THEN
            Outpt$ = Outpt$ + "\"
        END IF
    END IF

    ' store full directory pathname
    CALL TouchDirectory(Temp.Directory$, Outpt$)
END IF
RETURN
END SUB

' subroutine to touch date/time of a file
SUB TouchDirectory (Directory.Search$, Filename.Search$)
DIM ASCIIZ AS STRING * 260
DIM hfind AS _OFFSET

' make filename
ASCIIZ = Directory.Search$ + CHR$(0)

' check for prompting
IF Prompt.Files = False THEN
    Prompt2$ = "Touch " + Filename.Search$ + "(y/n/c/q)?"
    CALL MorePrompt(Prompt2$, "yncq", OutptX$)
    SELECT CASE OutptX$
        CASE "c"
            Prompt.Files = True
        CASE "n"
            EXIT SUB
        CASE "q"
            Quit.Searching = True
            EXIT SUB
    END SELECT
END IF

' display filename
IF Display.Filenames = 0 THEN
    COLOR Yellow, Black
    CALL AddStdout2(Filename.Search$)
END IF

' get source filename date/time.
hfind = CreateFile(ASCIIZ, &H180, &H3, 0, 3, FILE_FLAG_BACKUP_SEMANTICS, 0)
IF hfind <> INVALID_HANDLE_VALUE THEN
    x = GetFileTime&(hfind, SourceCreateDate, SourceAccessDate, SourceModifyDate)
    ' display any errors.
    IF x = 0 THEN
        IF Debug.Mode THEN
            Var$ = "Error (0x" + HEX$(GetLastError&) + ") reading filename date/time."
            CALL DisplayError(Var$)
            CALL DisplayWinError(x)
        END IF
    END IF

    IF x THEN
        ' increment directories counted
        Dirs.Counted = Dirs.Counted + 1!
    END IF

    ' store default date/time.
    CreateDate = SourceCreateDate
    AccessDate = SourceAccessDate
    ModifyDate = SourceModifyDate

    ' store override date/time.
    IF Override.Create.Date THEN
        GOSUB GetSysDate
        x = SystemTimeToFileTime&(SysTime, CreateDate)
        ' display any errors.
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename create date/time."
                CALL DisplayError(Var$)
                CALL DisplayWinError(x)
            END IF
        END IF
    END IF
    IF Override.Access.Date THEN
        GOSUB GetSysDate
        x = SystemTimeToFileTime&(SysTime, AccessDate)
        ' display any errors.
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename access date/time."
                CALL DisplayError(Var$)
                CALL DisplayWinError(x)
            END IF
        END IF
    END IF
    IF Override.Modify.Date THEN
        GOSUB GetSysDate
        x = SystemTimeToFileTime&(SysTime, ModifyDate)
        ' display any errors.
        IF x = 0 THEN
            IF Debug.Mode THEN
                Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename modify date/time."
                CALL DisplayError(Var$)
                CALL DisplayWinError(x)
            END IF
        END IF
    END IF

    ' set the destination filename date/time.
    x = SetFileTime&(hfind, CreateDate, AccessDate, ModifyDate)
    ' display any errors.
    IF x = 0 THEN
        IF Debug.Mode THEN
            Var$ = "Error (0x" + HEX$(GetLastError&) + ") touching filename date/time."
            CALL DisplayError(Var$)
            CALL DisplayWinError(x)
        END IF
    END IF
    x = CloseHandle(hfind)
ELSE
    Var$ = "Error (0x" + HEX$(GetLastError&) + ") opening filename."
    CALL DisplayError(Var$)
    CALL DisplayWinError(x)
END IF
EXIT SUB

GetSysDate:
' store override variables
SysTime.wMonth = MonthTemp
SysTime.wDay = DayTemp
SysTime.wYear = YearTemp
SysTime.wHour = HourTemp
SysTime.wMinute = MinuteTemp
SysTime.wSecond = SecondTemp
RETURN
END SUB

SUB Xcount (ErrFlag)
' check command line
Last.Switch = 0
Switch.Exist = 0
Command.Line = COMMAND2$
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)
IF Command.Line = "/?" THEN
    GOSUB Boot.Usage
    EXIT SUB
END IF

' get current drive
Current.Drive$ = LEFT$(_CWD$, 1)
Current.Directory = _CWD$

' check command line switches
No.Display.Archive = ParseLine("//A")
No.Display.Hidden = ParseLine("//H")
No.Display.Readonly = ParseLine("//O")
No.Display.System = ParseLine("//S")
No.Display.Any = ParseLine("//X")
No.Display.Compress = ParseLine("//M1")
No.Display.Encrypt = ParseLine("//M2")

Display.Archive = ParseLine("/A")
Display.Hidden = ParseLine("/H")
Display.Readonly = ParseLine("/O")
Display.System = ParseLine("/S")
Display.Any = ParseLine("/X")
Display.Compress = ParseLine("/M1")
Display.Encrypt = ParseLine("/M2")

Continuous.Display = ParseLine("/C")
Count.Dirs = ParseLine("/E")
Count.Files = ParseLine("/G")
Count.Bytes = ParseLine("/I")
Count.Lines = ParseLine("/K")
Display.Search = ParseLine("/J")
Recurse.Directories = ParseLine("/R")
Prepend.Drive = ParseLine("/U")
Display.Filenames = ParseLine("/V")
Remove.Slash = ParseLine("/Y")
Display.Errors2 = ParseLine("/Z1")
Display.Errors = ParseLine("/Z")

' reset file counter variable
Continue = False
Dirs.Counted = 0#
Files.Counted = 0#
Lines.Counted = 0#
Total.Bytes = 0#
   
' get date\time from command line
Search.From.Date = False
Search.To.Date = False
Search.From.Time = False
Search.To.Time = False
Imbedded = INSTR(UCASE$(Command.Line), "/D")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    D$ = MID$(Command.Line, Imbedded + 2, 21)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 23)
    IF LEN(D$) <> 21 THEN
        GOTO Boot.Error
    END IF
    IF MID$(D$, 11, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.From.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    S$ = RIGHT$(D$, 10)
    D1! = INT(VAL(MID$(S$, 1, 2)))
    D2! = INT(VAL(MID$(S$, 4, 2)))
    D3! = INT(VAL(MID$(S$, 7, 4)))
    Search.To.Date = ((D3! - 1980) * 512) + D1! * 32 + D2!
    IF Search.From.Date < False OR Search.To.Date < False THEN
        GOTO Boot.Error
    END IF
END IF
Imbedded = INSTR(UCASE$(Command.Line), "/T")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    T$ = MID$(Command.Line, Imbedded + 2, 17)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 19)
    IF LEN(T$) <> 17 THEN
        GOTO Boot.Error
    END IF
    IF MID$(T$, 9, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    S$ = LEFT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.From.Time = T1! * 2048 + T2! * 32
    S$ = RIGHT$(T$, 8)
    T1! = INT(VAL(MID$(S$, 1, 2)))
    T2! = INT(VAL(MID$(S$, 4, 2)))
    T3! = INT(VAL(MID$(S$, 7, 2)))
    Search.To.Time = T1! * 2048 + T2! * 32
    IF Search.From.Time < False OR Search.To.Time < False THEN
        GOTO Boot.Error
    END IF
END IF

' get file size from command line
Search.File.Size = False
Search.Size.From = FalseD
Search.Size.To = FalseD
Imbedded = INSTR(UCASE$(Command.Line), "/F")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Search.File.Size = True
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Search.Size.From = Var#
    IF MID$(Command.Line, Imbedded, 1) <> "-" THEN
        GOTO Boot.Error
    END IF
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    GOSUB Get.Numeric
    Search.Size.To = Var#
    Search.Size.From = Search.Size.From * 1024#
    Search.Size.To = Search.Size.To * 1024#
END IF

' get extended date\time switches
Creation.Time = ParseLine("/1")
Access.Time = ParseLine("/2")
Modified.Time = ParseLine("/3")
IF Creation.Time = False THEN
    IF Access.Time = False THEN
        IF Modified.Time = False THEN
            Modified.Time = True
        END IF
    END IF
END IF

' check command line switch
Imbedded = INSTR(UCASE$(Command.Line), "/N")
IF Imbedded THEN
    Var = LastSwitch(Imbedded)
    Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 2)
    GOSUB Get.Numeric
    Nested.Recurse = CINT(Var#)
END IF

' recheck command line
IF INSTR(Command.Line, "/") THEN
    GOTO Boot.Error
END IF
Command.Line = RTRIM$(Command.Line)
IF Switch.Exist THEN
    IF LEN(Command.Line) > Last.Switch THEN
        GOTO Boot.Error
    END IF
END IF

' reset work variables
Bytes.Counted = False
Dirs.Counted = False
Files.Counter = False
Lines.Counted = False
Quit.Searching = False
   
' remove blanks from command line
Command.Line = RTRIM$(Command.Line)
Command.Line = LTRIM$(Command.Line)

' store entire command
Command.Work = Command.Line

' display header
Header.Flag = False
GOSUB Header

Command.Line.Redirect$ = Command.Line
StdinCount = 1
DO
    ' get standard input
    Standard.Input$ = ""
    IF Stdin THEN
        IF StdinLines > 0 THEN ' nul:
            DO UNTIL StdinCount > StdinLines
                Standard.Input$ = StdinArray(StdinCount)
                Standard.Input$ = RTRIM$(LTRIM$(Standard.Input$))
                IF Standard.Input$ <> "" THEN
                    EXIT DO
                END IF
                StdinCount = StdinCount + 1
            LOOP
            IF StdinCount > StdinLines THEN
                EXIT DO
            END IF
        END IF
    END IF

    ' store entire command
    Command.Line = Command.Line.Redirect$

    ' filename processing loop
    DO
        ' store redirected input
        Standard.Input$ = RTRIM$(Standard.Input$)
        Standard.Input$ = LTRIM$(Standard.Input$)
        IF LEFT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = MID$(Standard.Input$, 2)
        END IF
        IF RIGHT$(Standard.Input$, 1) = CHR$(34) THEN
            Standard.Input$ = LEFT$(Standard.Input$, LEN(Standard.Input$) - 1)
        END IF

        ' store entire command
        IF LEFT$(Command.Line, 1) = CHR$(34) THEN
            Imbedded = INSTR(2, Command.Line, CHR$(34))
            IF Imbedded THEN
                Command.Work = Standard.Input$ + MID$(Command.Line, 2, Imbedded - 2)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        ELSE
            Imbedded = INSTR(Command.Line, " ")
            IF Imbedded THEN
                Command.Work = Standard.Input$ + LEFT$(Command.Line, Imbedded - 1)
                Command.Line = MID$(Command.Line, Imbedded + 1)
            ELSE
                Command.Work = Standard.Input$ + Command.Line
                Command.Line = Nul
            END IF
        END IF
        Command.Line = LTRIM$(Command.Line)
        Command.Line = RTRIM$(Command.Line)

        ' store current drive
        IF MID$(Command.Work, 2, 1) = ":" THEN
            Drive.Search = LEFT$(Command.Work, 1)
            Command.Work = MID$(Command.Work, 3)
        ELSE
            Drive.Search = Current.Drive$
        END IF
        Drive.Search = UCASE$(Drive.Search)

        ' store current directory
        Directory.Search$ = ""
        Imbedded1 = INSTR(Command.Work, "\")
        Imbedded2 = Imbedded1
        WHILE Imbedded1
            Imbedded2 = Imbedded1
            Imbedded1 = INSTR(Imbedded1 + 1, Command.Work, "\")
        WEND
        IF Imbedded2 THEN
            Directory.Search$ = LEFT$(Command.Work, Imbedded2)
            Command.Work = MID$(Command.Work, Imbedded2 + 1)
        END IF
        IF Directory.Search$ = "" THEN
            IF LEFT$(Current.Directory, 2) = "\\" THEN
                Directory.Search$ = Current.Directory
            ELSE
                IF Drive.Search = Current.Drive$ THEN
                    Directory.Search$ = MID$(Current.Directory, 3)
                ELSE
                    Directory.Search$ = "\"
                END IF
            END IF
        END IF
        IF RIGHT$(Directory.Search$, 1) <> "\" THEN
            Directory.Search$ = Directory.Search$ + "\"
        END IF
   
        ' get filename spec
        Filename.Search$ = Command.Work
        IF Filename.Search$ = Nul THEN
            Filename.Search$ = "*.*"
        END IF
        Command.Work = Nul

        ' display search filename
        IF Continuous.Display = False THEN
            COLOR Yellow, Black
            IF LEFT$(Directory.Search$, 2) = "\\" THEN
                PRINT "Searching: " + Directory.Search$ + Filename.Search$
            ELSE
                PRINT "Searching: " + Drive.Search + ":" + Directory.Search$ + Filename.Search$
            END IF
        END IF

        ' call routine to search for files
        IF LEFT$(Directory.Search$, 2) = "\\" THEN
            CALL Xcount_Directories(Directory.Search$, Filename.Search$)
        ELSE
            V = ASC(UCASE$(LEFT$(Drive.Search$, 1))) - 64
            IF MEDIAEXISTS(V) THEN
                CALL Xcount_Directories(Directory.Search$, Filename.Search$)
            END IF
        END IF

        ' check search filename
        IF Command.Line = Nul THEN
            EXIT DO
        END IF

        ' check quit searching
        IF Quit.Searching THEN
            EXIT DO
        END IF
    LOOP

    ' check stdin
    StdinCount = StdinCount + 1
    IF StdinCount > StdinLines THEN
        EXIT DO
    END IF

    ' check quit searching
    IF Quit.Searching THEN
        EXIT DO
    END IF
LOOP

' display end program
IF Continuous.Display = False THEN
    IF Display.Errors2 = 0 THEN
        COLOR Yellow, Black
        IF Count.Dirs THEN
            PRINT "Directories counted ";
            PRINT FormatString$(Dirs.Counted)
        END IF
        IF Count.Files THEN
            PRINT "Files counted ";
            PRINT FormatString$(Files.Counted)
        END IF
        IF Count.Lines THEN
            PRINT "Lines counted ";
            PRINT FormatString$(Lines.Counted)
        END IF
        IF Count.Bytes THEN
            PRINT "Bytes counted ";
            PRINT FormatString$(Bytes.Counted)
        END IF
    END IF
END IF
COLOR Plain, Black
EXIT SUB

' display program usage
Boot.Usage:
' make header
COLOR White, Black
PRINT "COUNT"
PRINT "   Directory/file count utility; "
COLOR Yellow, Black
PRINT "Usage:"
PRINT "   COUNT [d:][\pathname\][filename][//ahosx][/cdefgijknrtuvyz][/123]"
PRINT "Where:"
PRINT "   d:\pathname\filename is the search directory."
PRINT "   /c  continuous display    /e  count directories"
PRINT "   /g  count files           /i  count total bytes"
PRINT "   /j  display directories   /k  count total lines"
PRINT "   /nxxx  recurse override   /r  recurse directories"
PRINT "   /u  prepend drive letter  /v  display filenames"
PRINT "   /y  remove trailing slash /z  suppress error messages"
PRINT "   /z1 don't display output"
PRINT "   display ranges: (/1  creation, /2  last access  /3  modify time)"
PRINT "     /d  is range of file dates in form mm/dd/yyyy-mm/dd/yyyy"
PRINT "     /t  is range of file times in form hh:mm:ss-hh:mm:ss"
PRINT "     /f  is range of file sizes in form xxx-xxx in kilobytes"
PRINT "   display file attributes:"
PRINT "     / prefix for files with, // prefix for files without,"
PRINT "       a  archive, h  hidden, o  read-only, s  system, x  none"
PRINT "       m1  compressed, m2  encrypted"
COLOR Plain, Black
RETURN

Get.Numeric:
Var# = False
DO
    Temp$ = MID$(Command.Line, Imbedded, 1)
    IF Temp$ >= "0" AND Temp$ <= "9" THEN
        Var# = Var# * 10 + VAL(Temp$)
        Command.Line = LEFT$(Command.Line, Imbedded - 1) + MID$(Command.Line, Imbedded + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

Boot.Error:
COLOR White, Black
PRINT "Command line error. Type COUNT /? for help."
COLOR Plain, Black
ErrFlag = -1
EXIT SUB

' make header
Header:
IF Header.Flag THEN
    RETURN
END IF
Header.Flag = True
IF Continuous.Display = False THEN
    COLOR White, Black
    PRINT "Directory/file count utility; "
END IF
RETURN
END SUB

' subroutine to access directories
SUB Xcount_Directories (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' store directory asciiz
IF LEFT$(Directory.Search$, 2) <> "\\" THEN
    ASCIIZ = Drive.Search + ":" + Directory.Search$ + "*.*" + CHR$(0)
ELSE
    ASCIIZ = Directory.Search$ + "*.*" + CHR$(0)
END IF

' start directory search
Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' search directory names
    CALL Xcount_Filenames(Directory.Search$, Filename.Search$)

    ' check to recurse directories
    IF Recurse.Directories THEN

        ' recurse directories
        DO
            ' check to quit
            IF Quit.Deleting THEN
                EXIT DO
            END IF

            ' check directory attribute
            Attribute = finddata.dwFileAttributes
            IF (Attribute AND &H10) = &H10 THEN
                ' store directory name
                Directory$ = finddata.cFileName
                Directory$ = LEFT$(Directory$, INSTR(Directory$, CHR$(0)) - 1)

                ' check directory name
                IF Directory$ <> "." AND Directory$ <> ".." THEN
                    ' check unicode
                    IF INSTR(Directory$, "?") THEN
                        Directory$ = finddata.cAlternateFileName
                        V = INSTR(Directory$, CHR$(0))
                        IF V THEN Directory$ = LEFT$(Directory$, V - 1)
                    END IF

                    ' make next search directory
                    Next.Directory$ = Directory.Search$ + Directory$ + "\"

                    ' recursively search subdirectories
                    CALL Xcount_Directories(Next.Directory$, Filename.Search$)
                END IF
            END IF
        LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    END IF
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to access filenames in directory
SUB Xcount_Filenames (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG
DIM ASCIIZ AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM Wfile.Handle AS _UNSIGNED _OFFSET

' make filename
ASCIIZ = Directory.Search$ + Filename.Search$ + CHR$(0)

Wfile.Handle = FindFirstFileA(_OFFSET(ASCIIZ), _OFFSET(finddata))
IF Wfile.Handle <> INVALID_HANDLE_VALUE THEN
    ' recurse filename
    DO
        ' store filename
        Filename$ = finddata.cFileName
        Filename$ = LEFT$(Filename$, INSTR(Filename$, CHR$(0)) - 1)

        ' check filename
        IF Filename$ <> "." AND Filename$ <> ".." THEN

            ' check unicode
            IF INSTR(Filename$, "?") AND CurrentCommand < CommandCount THEN
                Filename$ = finddata.cAlternateFileName
                V = INSTR(Filename$, CHR$(0))
                IF V THEN Filename$ = LEFT$(Filename$, V - 1)
            END IF

            ' store file data
            finddatatemp = finddata

            ' store file size
            File.Size = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow

            ' check directory attribute
            Attribute = finddata.dwFileAttributes

            ' check for directory
            IF (Attribute AND &H10) <> &H10 THEN
                CALL Xcount_DisplayFiles(Directory.Search$, Filename$)
            END IF

            ' check for directory
            IF (Attribute AND &H10) = &H10 THEN
                ' increment directories actually searched
                Dirs.Counted = Dirs.Counted + 1#

                ' display directory being searched
                CALL Xcount_DisplayDirectory(Directory.Search$, Filename$)
            END IF
        END IF

        ' check to quit
        IF Quit.Deleting THEN
            EXIT DO
        END IF
    LOOP WHILE FindNextFileA(Wfile.Handle, _OFFSET(finddata))
    x = FindClose(Wfile.Handle)
END IF
END SUB

' subroutine to display directory
SUB Xcount_DisplayDirectory (Directory.Search$, Filename.Search$)
' display directory
IF Display.Search THEN
    Filename$ = Directory.Search$ + Filename.Search$
    IF RIGHT$(Filename$, 1) <> "\" THEN
        Filename$ = Filename$ + "\"
    END IF
    IF Remove.Slash THEN
        IF Filename$ <> "\" THEN
            IF RIGHT$(Filename$, 1) = "\" THEN
                Slash = LEN(Filename$)
                Filename$ = LEFT$(Filename$, Slash - 1)
            END IF
        END IF
    END IF
    IF Prepend.Drive THEN
        IF LEFT$(Filename$, 2) <> "\\" THEN
            Filename$ = Drive.Search + ":" + Filename$
        END IF
    END IF
    COLOR Green, Black
    CALL AddStdout2(Filename$)
END IF
END SUB

' subroutine to count and display filenames
SUB Xcount_DisplayFiles (Directory.Search$, Filename.Search$)
' declare subroutine variables
DIM Attribute AS _UNSIGNED LONG

' check directory attribute
Attribute = finddatatemp.dwFileAttributes

' check for readonly file
IF (Attribute AND &H1) = &H1 THEN
    IF No.Display.Readonly THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Readonly THEN
        EXIT SUB
    END IF
END IF

' check for hidden file
IF (Attribute AND &H2) = &H2 THEN
    IF No.Display.Hidden THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Hidden THEN
        EXIT SUB
    END IF
END IF

' check for system file
IF (Attribute AND &H4) = &H4 THEN
    IF No.Display.System THEN
        EXIT SUB
    END IF
ELSE
    IF Display.System THEN
        EXIT SUB
    END IF
END IF

' check for archive file
IF (Attribute AND &H20) = &H20 THEN
    IF No.Display.Archive THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Archive THEN
        EXIT SUB
    END IF
END IF

' check for compressed file
IF (Attribute AND &H800) = &H800 THEN
    IF No.Display.Compress THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Compress THEN
        EXIT SUB
    END IF
END IF

' check for encrypted file
IF (Attribute AND &H4000) = &H4000 THEN
    IF No.Display.Encrypt THEN
        EXIT SUB
    END IF
ELSE
    IF Display.Encrypt THEN
        EXIT SUB
    END IF
END IF

' check all attributes
IF Display.Any THEN
    IF (Attribute AND &H1) = &H1 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H2) = &H2 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4) = &H4 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H20) = &H20 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H800) = &H800 THEN
        EXIT SUB
    END IF
    IF (Attribute AND &H4000) = &H4000 THEN
        EXIT SUB
    END IF
END IF
IF No.Display.Any THEN
    IF (Attribute AND &H1) = False THEN
        IF (Attribute AND &H2) = False THEN
            IF (Attribute AND &H4) = False THEN
                IF (Attribute AND &H20) = False THEN
                    IF (Attribute AND &H800) = False THEN
                        IF (Attribute AND &H4000) = False THEN
                            EXIT SUB
                        END IF
                    END IF
                END IF
            END IF
        END IF
    END IF
END IF

' check file size
IF Search.File.Size THEN
    IF Search.Size.From = FalseD AND Search.Size.To = FalseD THEN
        IF File.Size <> FalseD THEN
            EXIT SUB
        END IF
    ELSE
        IF Search.Size.From > FalseD OR Search.Size.To > FalseD THEN
            IF Search.Size.From > FalseD THEN
                IF Search.Size.To = FalseD THEN
                    IF File.Size < Search.Size.From THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From = FalseD THEN
                IF Search.Size.To > FalseD THEN
                    IF File.Size > Search.Size.To THEN
                        EXIT SUB
                    END IF
                END IF
            END IF
            IF Search.Size.From <= Search.Size.To THEN
                IF File.Size < Search.Size.From OR File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
            IF Search.Size.From > Search.Size.To THEN
                IF File.Size < Search.Size.From AND File.Size > Search.Size.To THEN
                    EXIT SUB
                END IF
            END IF
        END IF
    END IF
END IF

' store file date and time
IF Creation.Time THEN
    x& = FileTimeToSystemTime&(finddatatemp.ftCreationTime, SysTime)
    GOSUB Convert.Date
    GOSUB Convert.Time
ELSE
    IF Access.Time THEN
        x& = FileTimeToSystemTime&(finddatatemp.ftLastAccessTime, SysTime)
        GOSUB Convert.Date
        GOSUB Convert.Time
    ELSE
        IF Modified.Time THEN
            x& = FileTimeToSystemTime&(finddatatemp.ftLastWriteTime, SysTime)
            GOSUB Convert.Date
            GOSUB Convert.Time
        END IF
    END IF
END IF

' check date\time range
IF Search.From.Date OR Search.To.Date THEN
    IF File.Work.Date < Search.From.Date THEN
        EXIT SUB
    END IF
    IF File.Work.Date > Search.To.Date THEN
        EXIT SUB
    END IF
END IF
IF Search.From.Time OR Search.To.Time THEN
    IF File.Work.Time < Search.From.Time THEN
        EXIT SUB
    END IF
    IF File.Work.Time > Search.To.Time THEN
        EXIT SUB
    END IF
END IF

' make directory filename
IF Display.Filenames THEN
    Filename$ = Directory.Search$ + Filename.Search$
    IF Prepend.Drive THEN
        IF LEFT$(Filename$, 2) <> "\\" THEN
            Filename$ = Drive.Search + ":" + Filename$
        END IF
    END IF
    COLOR Yellow, Black
    CALL AddStdout2(Filename$)
END IF

' update bytes counted
Bytes.Counted = Bytes.Counted + File.Size

' update files counter
Files.Counted = Files.Counted + 1#

' check count flag
IF Count.Lines THEN
    CALL Xcount_CountFileLines(Directory.Search$, Filename.Search$)
END IF
EXIT SUB

Convert.Date:
YearTemp! = SysTime.wYear
MonthTemp! = SysTime.wMonth
DayTemp! = SysTime.wDay
File.Date$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
File.Date$ = File.Date$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
File.Date$ = File.Date$ + LTRIM$(STR$(SysTime.wYear))
File.Work.Date = ((YearTemp! - 1980) * 512) + MonthTemp! * 32 + DayTemp!
RETURN

Convert.Time:
HourTemp! = SysTime.wHour
MinuteTemp! = SysTime.wMinute
SecondsTemp! = SysTime.wSecond
File.Time$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
File.Time$ = File.Time$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
File.Work.Time = HourTemp! * 2048 + MinuteTemp! * 32 + SecondsTemp!
RETURN
END SUB

' subroutine to count lines in a filename
SUB Xcount_CountFileLines (Directory.Search$, Filename.Search$)
' make filename
Filename.Count$ = Directory.Search$ + Filename.Search$
IF ValidFileChar(Filename.Search$) THEN
    Short.Filename$ = finddatatemp.cAlternateFileName
    V = INSTR(Short.Filename$, CHR$(0))
    IF V THEN Short.Filename$ = LEFT$(Short.Filename$, V - 1)
    Filename.Count$ = Directory.Search$ + Short.Filename$
    IF ValidFileChar(Short.Filename$) THEN
        EXIT SUB
    END IF
END IF
Var$ = Ambiguate1$(Filename.Search$)

' start file count
Control.Break = False
Count# = 0#
Percent = 0
Last.Percent = 0
IF _FILEEXISTS(Filename.Count$) THEN
    X = FREEFILE
    OPEN Filename.Count$ FOR BINARY SHARED AS #X LEN = xbuflen
    IF LOF(X) > 0 THEN
        File.Size = LOF(X)
        DO UNTIL EOF(X)
            ' get file buffer
            GOSUB Display.Percent
            GET X, , xbuffer
            FOR L = 1 TO xbuflen
                IF MID$(xbuffer, L, 1) = CHR$(10) THEN
                    Lines.Counted = Lines.Counted + 1#
                END IF
            NEXT
            IF Control.Break THEN
                Control.Break = False
                EXIT DO
            END IF
            IF INKEYx$ = CHR$(27) THEN
                EXIT DO
            END IF
        LOOP
    END IF
    CLOSE #X
    _TITLE Program + " - COUNT"
END IF
EXIT SUB

' display percentage scanned in title
Display.Percent:
IF File.Size > 0 THEN
    Count# = Count# + xbuflen
    IF Count# > LOF(X) THEN
        Count# = LOF(X)
    END IF
    Percent = (Count# / File.Size) * 100!
    Percent = INT(Percent)
    IF Percent <> Last.Percent THEN
        Last.Percent = Percent
        P$ = RIGHT$(STR$(Percent + 1000%), 3) + "%"
        V$ = Program + " - COUNT - " + Var$ + " " + P$
        _TITLE V$
    END IF
END IF
RETURN
END SUB

' screen save
SUB SaveScreen
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        TempZ1 = SCREEN(Var1, Var2) ' screen char
        TempZ2 = SCREEN(Var1, Var2, 1) ' char color
        TempArrayY((Var1 - 1) * 80 + Var2) = TempZ1
        TempArrayZ((Var1 - 1) * 80 + Var2) = TempZ2
    NEXT
NEXT
END SUB

' screen save
SUB SaveScreen2
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        TempZ1 = SCREEN(Var1, Var2) ' screen char
        TempZ2 = SCREEN(Var1, Var2, 1) ' char color
        TempArrayY2((Var1 - 1) * 80 + Var2) = TempZ1
        TempArrayZ2((Var1 - 1) * 80 + Var2) = TempZ2
    NEXT
NEXT
END SUB

' screen save only used with full screen editor
SUB SaveScreen3
FOR Var1 = 2 TO 24
    FOR Var2 = 1 TO 80
        TempZ1 = SCREEN(Var1, Var2) ' screen char
        TempZ2 = SCREEN(Var1, Var2, 1) ' char color
        TempArrayY2((Var1 - 1) * 80 + Var2) = TempZ1
        TempArrayZ2((Var1 - 1) * 80 + Var2) = TempZ2
    NEXT
NEXT
END SUB

' screen restore
SUB RestoreScreen
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        VarB = INT(TempArrayZ((Var1 - 1) * 80 + Var2) / 16)
        VarF = TempArrayZ((Var1 - 1) * 80 + Var2) MOD 16
        TempZ1 = TempArrayY((Var1 - 1) * 80 + Var2)
        LOCATE Var1, Var2, 1
        COLOR VarF, VarB
        _CONTROLCHR OFF
        PRINT CHR$(TempZ1);
        _CONTROLCHR ON
    NEXT
NEXT
END SUB

' screen restore
SUB RestoreScreen2
IF Status THEN
    V = 23
ELSE
    V = 24
END IF
FOR Var1 = 2 TO V
    FOR Var2 = 1 TO 80
        VarB = INT(TempArrayZ2((Var1 - 1) * 80 + Var2) / 16)
        VarF = TempArrayZ2((Var1 - 1) * 80 + Var2) MOD 16
        TempZ1 = TempArrayY2((Var1 - 1) * 80 + Var2)
        LOCATE Var1, Var2, 1
        COLOR VarF, VarB
        _CONTROLCHR OFF
        PRINT CHR$(TempZ1);
        _CONTROLCHR ON
    NEXT
NEXT
END SUB

' screen restore only used with full screen editor
SUB RestoreScreen3
FOR Var1 = 2 TO 24
    FOR Var2 = 1 TO 80
        VarB = INT(TempArrayZ2((Var1 - 1) * 80 + Var2) / 16)
        VarF = TempArrayZ2((Var1 - 1) * 80 + Var2) MOD 16
        TempZ1 = TempArrayY2((Var1 - 1) * 80 + Var2)
        LOCATE Var1, Var2, 1
        COLOR VarF, VarB
        _CONTROLCHR OFF
        PRINT CHR$(TempZ1);
        _CONTROLCHR ON
    NEXT
NEXT
END SUB

' Start drop down init menu.
SUB Menu2
IF Status = 0 THEN
    VIEW PRINT 1 TO 24
ELSE
    VIEW PRINT 1 TO 23
END IF
IF CurrentMenu = 0 THEN
    CurrentMenu = 1
END IF
CurrentMenuSelection = 1
COLOR Yellow, 0
CALL DisplayTab2(CurrentMenu)

' keyboard/mouse input loop.
Control.Break = False
Monitor.Box = False
DO
    ' check alt key depress
    CharInput$ = ""
    Q& = _KEYHIT
    IF Q& = KEY_LALT& THEN
        _DELAY .75
        CharInput$ = INKEYx$
        IF CharInput$ = "" THEN
            COLOR White, 1
            CALL DisplayTab(CurrentMenu)
            CurrentMenu = False
            EXIT DO
        END IF
    END IF
    IF Q& = KEY_RALT& THEN
        _DELAY .75
        CharInput$ = INKEYx$
        IF CharInput$ = "" THEN
            COLOR White, 1
            CALL DisplayTab(CurrentMenu)
            CurrentMenu = False
            EXIT DO
        END IF
    END IF

    ' check mouse activity
    Z = MouseDriver
    IF MouseButton1 THEN
        ' select drop down menu.
        IF MouseX = 1 THEN
            NewMenuSelection = 0
            SELECT CASE MouseY
                CASE 1 TO 3 ' cpu
                    CPUOFF = NOT CPUOFF
                    IF CPUOFF THEN
                        TIMER OFF
                        _TITLE Program
                        CALL DisplayToolBar2
                    ELSE
                        TIMER ON
                        CALL DisplayToolBar2
                    END IF
                CASE 74 TO 78 ' about
                    CALL HelpAbout
                CASE 6 TO 9 ' File
                    NewMenuSelection = 1
                CASE 16 TO 21 ' Charts
                    NewMenuSelection = 2
                CASE 26 TO 29 ' Dump
                    NewMenuSelection = 3
                CASE 36 TO 39 ' Edit
                    NewMenuSelection = 4
                CASE 46 TO 49 ' Jump
                    NewMenuSelection = 5
                CASE 56 TO 60 ' Print
                    NewMenuSelection = 6
                CASE 64 TO 69 ' Search
                    NewMenuSelection = 7
            END SELECT
            IF NewMenuSelection THEN
                COLOR White, 1
                CALL DisplayTab(CurrentMenu)
                CurrentMenu = NewMenuSelection
                NewMenuSelection = 0
                CurrentMenuSelection = 1
                CALL DropDownMenu(Var)
                EXIT DO
            END IF
        END IF
    ELSE
        ' check mouse position.
        IF MouseX OR MouseY THEN
            ' select drop down menu.
            IF MouseX = 1 THEN
                NewMenuSelection = False
                SELECT CASE MouseY
                    CASE 6 TO 9 ' File
                        NewMenuSelection = 1
                    CASE 16 TO 21 ' Charts
                        NewMenuSelection = 2
                    CASE 26 TO 29 ' Dump
                        NewMenuSelection = 3
                    CASE 36 TO 39 ' Edit
                        NewMenuSelection = 4
                    CASE 46 TO 49 ' Jump
                        NewMenuSelection = 5
                    CASE 56 TO 60 ' Print
                        NewMenuSelection = 6
                    CASE 64 TO 69 ' Search
                        NewMenuSelection = 7
                    CASE ELSE
                        NewMenuSelection = False
                END SELECT
                IF NewMenuSelection > False THEN
                    COLOR White, 1
                    CALL DisplayTab(CurrentMenu)
                    CurrentMenu = NewMenuSelection
                    NewMenuSelection = False
                    COLOR Yellow, 0
                    CALL DisplayTab2(CurrentMenu)
                END IF
            END IF
        END IF
    END IF

    ' check mouse wheel activity
    IF MouseWheel THEN
        MousePressed = -1
        MouseWheel = 0
        CurrentMenuSelection = 1
        CALL DropDownMenu(Var)
        IF Var THEN
            EXIT DO
        END IF
    END IF

    ' check key press
    _LIMIT 100
    IF CharInput$ = "" THEN
        CharInput$ = INKEYx$
    END IF

    ' check break
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        LOCATE , , 0
    END IF

    Var = False
    IF LEN(CharInput$) THEN
        SELECT CASE LEN(CharInput$)
            CASE 1
                CharInput$ = UCASE$(CharInput$)
                SELECT CASE ASC(CharInput$)
                    CASE 13 ' Enter
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 27 ' Escape
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = False
                        EXIT DO
                    CASE 65 TO 90 ' A - Z
                        IF CharInput$ = "A" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CALL HelpAbout
                            _KEYCLEAR
                            CurrentMenu = False
                            EXIT DO
                        END IF
                        IF CharInput$ = "F" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 1
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "C" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 2
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "D" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 3
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "E" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 4
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "J" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 5
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "P" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 6
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                        IF CharInput$ = "S" THEN
                            COLOR White, 1
                            CALL DisplayTab(CurrentMenu)
                            CurrentMenu = 7
                            CurrentMenuSelection = 1
                            CALL DropDownMenu(Var)
                        END IF
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(CharInput$, 1))
                    CASE 80, 145, 160 ' Down/Ctrl-Down/Alt-Down
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 75, 15, 115, 155 ' Left/Shift-Tab/Ctrl-Left/Alt-Left
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        IF CurrentMenu = 1 THEN
                            CurrentMenu = 7
                        ELSE
                            CurrentMenu = CurrentMenu - 1
                        END IF
                        COLOR Yellow, 0
                        CALL DisplayTab2(CurrentMenu)
                    CASE 77, 116, 157 ' Right/Ctrl-Right/Alt-Right
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        IF CurrentMenu = 7 THEN
                            CurrentMenu = 1
                        ELSE
                            CurrentMenu = CurrentMenu + 1
                        END IF
                        COLOR Yellow, 0
                        CALL DisplayTab2(CurrentMenu)
                    CASE 30 ' Alt-A
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CALL HelpAbout
                        _KEYCLEAR
                        CurrentMenu = False
                        EXIT DO
                    CASE 33 ' Alt-F
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 1
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 46 ' Alt-C
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 2
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 32 ' Alt-D
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 3
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 18 ' Alt-E
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 4
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 36 ' Alt-J
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 5
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 25 ' Alt-P
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 6
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                    CASE 31 ' Alt-S
                        COLOR White, 1
                        CALL DisplayTab(CurrentMenu)
                        CurrentMenu = 7
                        CurrentMenuSelection = 1
                        CALL DropDownMenu(Var)
                END SELECT
        END SELECT
    END IF
    IF CurrentMenu = False THEN
        EXIT DO
    END IF
    IF Var THEN
        EXIT DO
    END IF
LOOP
IF Status = 0 THEN
    VIEW PRINT 1 TO 24
ELSE
    VIEW PRINT 1 TO 23
END IF
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
ELSE
    VIEW PRINT 2 TO 23
END IF
END SUB

REM Drop down menu subroutine:

REM Returns: variable CurrentMenu equals 1 to 7, or 0 if exit,
REM  variable CurrentMenuSelection is menu selection in CurrentMenu.

SUB DropDownMenu (VarX)
IF Status = 0 THEN
    VIEW PRINT 1 TO 24
ELSE
    VIEW PRINT 1 TO 23
END IF
VarX = 0
GOSUB StoreArea
GOSUB DrawMenu
GOSUB DrawCurrentMenuSelection

' keyboard/mouse input loop.
Control.Break = False
Monitor.Box = False
DO
    _LIMIT 100
    CharInput$ = ""

    ' call mouse subroutine.
    Z = MouseDriver

    ' check left mouse button.
    IF MouseButton1 THEN
        SaveMenu = CurrentMenu
        GOSUB MouseButton1X
        IF CurrentMenu = False THEN
            CurrentMenu = SaveMenu
            GOSUB RestoreArea
            CurrentMenu = 0
            EXIT DO
        END IF
    ELSE
        ' check right mouse button.
        IF MouseButton2 THEN
            GOSUB RestoreArea
            CurrentMenu = False
            EXIT DO
        ELSE
            ' check mouse position.
            IF MouseX OR MouseY THEN
                GOSUB MoveMouse
            END IF
        END IF
    END IF

    ' check mouse wheel.
    I$ = ""
    IF MouseWheel = -1 THEN
        I$ = CHR$(0) + CHR$(72) ' up
        MousePressed = -1
        MouseWheel = 0
    END IF
    IF MouseWheel = 1 THEN
        I$ = CHR$(0) + CHR$(80) ' down
        MousePressed = -1
        MouseWheel = 0
    END IF

    ' store keyboard buffer.
    IF I$ <> "" THEN
        CharInput$ = I$
    ELSE
        CharInput$ = INKEYx$
    END IF

    ' check break
    IF Control.Break OR Monitor.Box THEN
        Control.Break = False
        Monitor.Box = False
        LOCATE , , 0
    END IF

    IF LEN(CharInput$) THEN
        SELECT CASE LEN(CharInput$)
            CASE 1
                CharInput$ = UCASE$(CharInput$)
                SELECT CASE ASC(CharInput$)
                    CASE 65 TO 90 ' A - Z
                        SELECT CASE CurrentMenu
                            CASE 1
                                VarZ = INSTR("OCPVTX", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 2
                                VarZ = INSTR("AH", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 3
                                VarZ = INSTR("SF", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 4
                                VarZ = INSTR("ASR", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 5
                                VarZ = INSTR("BL", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 6
                                VarZ = INSTR("SF", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                            CASE 7
                                VarZ = INSTR("SBR", CharInput$)
                                IF VarZ THEN
                                    GOSUB RestoreArea
                                    CurrentMenuSelection = VarZ
                                    VarX = -1
                                    EXIT DO
                                END IF
                        END SELECT
                    CASE 9 ' Tab
                        GOSUB MenuRight
                    CASE 13 ' Enter
                        GOSUB RestoreArea
                        VarX = -1
                        EXIT DO
                    CASE 27 ' Escape
                        GOSUB RestoreArea
                        CurrentMenu = False
                        _KEYCLEAR
                        EXIT DO
                END SELECT
            CASE 2
                SELECT CASE ASC(RIGHT$(CharInput$, 1))
                    CASE 72, 141, 152 ' Up/Ctrl-Up/Alt-Up
                        GOSUB MenuUp
                    CASE 80, 145, 160 ' Down/Ctrl-Down/Alt-Down
                        GOSUB MenuDown
                    CASE 75, 15, 115, 155 ' Left/Shift-Tab/Ctrl-Left/Alt-Left
                        GOSUB MenuLeft
                    CASE 142, 74 ' Ctrl-Keypad-/Alt-Keypad-
                        GOSUB MenuLeft
                    CASE 77, 116, 157 ' Right/Ctrl-Right/Alt-Right
                        GOSUB MenuRight
                    CASE 144, 78 ' Ctrl-Keypad+/Alt-Keypad+
                        GOSUB MenuRight
                    CASE 71, 73, 119, 132 ' Home/PageUp/Ctrl-Home/Ctrl-PageUp
                        GOSUB MenuFirst
                    CASE 79, 81, 117, 118 ' End/PageDown/Ctrl-End/Ctrl-PageDown
                        GOSUB MenuLast
                END SELECT
        END SELECT
    END IF
LOOP
CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
ELSE
    VIEW PRINT 2 TO 23
END IF
EXIT SUB

' moves to menu first left
MenuFirst:
IF CurrentMenu <> 1 THEN
    GOSUB RestoreArea
    CurrentMenu = 1
    GOSUB StoreArea
    GOSUB DrawMenu
    GOSUB DrawCurrentMenuSelection
END IF
RETURN

' moves to menu last right
MenuLast:
IF CurrentMenu <> 7 THEN
    GOSUB RestoreArea
    CurrentMenu = 7
    GOSUB StoreArea
    GOSUB DrawMenu
    GOSUB DrawCurrentMenuSelection
END IF
RETURN

' moves one menu left, wrapping.
MenuLeft:
GOSUB RestoreArea
IF CurrentMenu = 1 THEN
    CurrentMenu = 7
ELSE
    CurrentMenu = CurrentMenu - 1
END IF
GOSUB StoreArea
GOSUB DrawMenu
GOSUB DrawCurrentMenuSelection
RETURN

' moves one menu right, wrapping.
MenuRight:
GOSUB RestoreArea
IF CurrentMenu = 7 THEN
    CurrentMenu = 1
ELSE
    CurrentMenu = CurrentMenu + 1
END IF
GOSUB StoreArea
GOSUB DrawMenu
GOSUB DrawCurrentMenuSelection
RETURN

' moves menu up, wrapping.
' selects CurrentMenuSelection variable.
MenuUp:
SELECT CASE CurrentMenu
    CASE 1
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 6
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 2
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 2
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 3
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 2
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 4
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 3
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 5
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 2
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 6
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 2
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 7
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 1 THEN
            CurrentMenuSelection = 3
        ELSE
            CurrentMenuSelection = CurrentMenuSelection - 1
        END IF
        GOSUB DrawCurrentMenuSelection
END SELECT
RETURN

' moves menu down, wrapping.
' selects CurrentMenuSelection variable.
MenuDown:
SELECT CASE CurrentMenu
    CASE 1
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 6 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 2
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 2 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 3
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 2 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 4
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 3 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 5
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 2 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 6
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 2 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
    CASE 7
        GOSUB EraseCurrentMenuSelection
        IF CurrentMenuSelection = 3 THEN
            CurrentMenuSelection = 1
        ELSE
            CurrentMenuSelection = CurrentMenuSelection + 1
        END IF
        GOSUB DrawCurrentMenuSelection
END SELECT
RETURN

' removes hilight from current menu selection.
EraseCurrentMenuSelection:
COLOR White, 0
SelectColor = 0
GOSUB DisplayCurrentMenuSelection
COLOR White, 0
RETURN

' adds hilight to current menu selection.
DrawCurrentMenuSelection:
COLOR White, 1
SelectColor = 1
GOSUB DisplayCurrentMenuSelection
COLOR White, 0
RETURN

' draws current menu selection.
DisplayCurrentMenuSelection:
SELECT CASE CurrentMenu
    CASE 1
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 7, 0
                PRINT "Open File      ";
                COLOR Magenta, SelectColor
                LOCATE 3, 7, 0
                PRINT "O";
            CASE 2
                LOCATE 4, 7, 0
                PRINT "Close File     ";
                COLOR Magenta, SelectColor
                LOCATE 4, 7, 0
                PRINT "C";
            CASE 3
                LOCATE 5, 7, 0
                PRINT "Print File     ";
                COLOR Magenta, SelectColor
                LOCATE 5, 7, 0
                PRINT "P";
            CASE 4
                LOCATE 6, 7, 0
                PRINT "View File Info ";
                COLOR Magenta, SelectColor
                LOCATE 6, 7, 0
                PRINT "V";
            CASE 5
                LOCATE 7, 7, 0
                PRINT "Type File      ";
                COLOR Magenta, SelectColor
                LOCATE 7, 7, 0
                PRINT "T";
            CASE 6
                LOCATE 8, 7, 0
                PRINT "Exit Program   ";
                COLOR Magenta, SelectColor
                LOCATE 8, 8, 0
                PRINT "x";
        END SELECT
    CASE 2
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 17, 0
                PRINT "ANSI Chart";
                COLOR Magenta, SelectColor
                LOCATE 3, 17, 0
                PRINT "A";
            CASE 2
                LOCATE 4, 17, 0
                PRINT "HEX Chart ";
                COLOR Magenta, SelectColor
                LOCATE 4, 17, 0
                PRINT "H";
        END SELECT
    CASE 3
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 27, 0
                PRINT "HEX Screen Dump";
                COLOR Magenta, SelectColor
                LOCATE 3, 31, 0
                PRINT "S";
            CASE 2
                LOCATE 4, 27, 0
                PRINT "HEX File Dump  ";
                COLOR Magenta, SelectColor
                LOCATE 4, 31, 0
                PRINT "F";
        END SELECT
    CASE 4
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 37, 0
                PRINT "Append Bytes       ";
                COLOR Magenta, SelectColor
                LOCATE 3, 37, 0
                PRINT "A";
            CASE 2
                LOCATE 4, 37, 0
                PRINT "Append ASCII String";
                COLOR Magenta, SelectColor
                LOCATE 4, 50, 0
                PRINT "S";
            CASE 3
                LOCATE 5, 37, 0
                PRINT "Redraw Screen      ";
                COLOR Magenta, SelectColor
                LOCATE 5, 37, 0
                PRINT "R";
        END SELECT
    CASE 5
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 47, 0
                PRINT "Jump To Byte";
                COLOR Magenta, SelectColor
                LOCATE 3, 55, 0
                PRINT "B";
            CASE 2
                LOCATE 4, 47, 0
                PRINT "Jump To Line";
                COLOR Magenta, SelectColor
                LOCATE 4, 55, 0
                PRINT "L";
        END SELECT
    CASE 6
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 57, 0
                PRINT "HEX Screen Print";
                COLOR Magenta, SelectColor
                LOCATE 3, 61, 0
                PRINT "S";
            CASE 2
                LOCATE 4, 57, 0
                PRINT "HEX File Print  ";
                COLOR Magenta, SelectColor
                LOCATE 4, 61, 0
                PRINT "F";
        END SELECT
    CASE 7
        SELECT CASE CurrentMenuSelection
            CASE 1
                LOCATE 3, 65, 0
                PRINT "Search String ";
                COLOR Magenta, SelectColor
                LOCATE 3, 72, 0
                PRINT "S";
            CASE 2
                LOCATE 4, 65, 0
                PRINT "Search Bytes  ";
                COLOR Magenta, SelectColor
                LOCATE 4, 72, 0
                PRINT "B";
            CASE 3
                LOCATE 5, 65, 0
                PRINT "Search Replace";
                COLOR Magenta, SelectColor
                LOCATE 5, 72, 0
                PRINT "R";
        END SELECT
END SELECT
RETURN

' draws menu.
DrawMenu:
CurrentMenuSelection = 1
GOSUB BoxBorder
BoxDrawX1 = Xstore1
BoxDrawX2 = Xstore2
BoxDrawY1 = Ystore1
BoxDrawY2 = Ystore2
SELECT CASE CurrentMenu
    CASE 1
        BoxDrawLength = 17
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 7, 0
        PRINT "Open File      ";
        LOCATE 4, 7, 0
        PRINT "Close File     ";
        LOCATE 5, 7, 0
        PRINT "Print File     ";
        LOCATE 6, 7, 0
        PRINT "View File Info ";
        LOCATE 7, 7, 0
        PRINT "Type File      ";
        LOCATE 8, 7, 0
        PRINT "Exit Program   ";
        COLOR Magenta
        LOCATE 3, 7, 0
        PRINT "O";
        LOCATE 4, 7, 0
        PRINT "C";
        LOCATE 5, 7, 0
        PRINT "P";
        LOCATE 6, 7, 0
        PRINT "V";
        LOCATE 7, 7, 0
        PRINT "T";
        LOCATE 8, 8, 0
        PRINT "x";
    CASE 2
        BoxDrawLength = 12
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 17, 0
        PRINT "ANSI Chart";
        LOCATE 4, 17, 0
        PRINT "HEX Chart ";
        COLOR Magenta
        LOCATE 3, 17, 0
        PRINT "A";
        LOCATE 4, 17, 0
        PRINT "H";
    CASE 3
        BoxDrawLength = 17
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 27, 0
        PRINT "HEX Screen Dump";
        LOCATE 4, 27, 0
        PRINT "HEX File Dump  ";
        COLOR Magenta
        LOCATE 3, 31, 0
        PRINT "S";
        LOCATE 4, 31, 0
        PRINT "F";
    CASE 4
        BoxDrawLength = 21
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 37, 0
        PRINT "Append Bytes       ";
        LOCATE 4, 37, 0
        PRINT "Append ASCII String";
        LOCATE 5, 37, 0
        PRINT "Redraw Screen      ";
        COLOR Magenta
        LOCATE 3, 37, 0
        PRINT "A";
        LOCATE 4, 50, 0
        PRINT "S";
        LOCATE 5, 37, 0
        PRINT "R";
    CASE 5
        BoxDrawLength = 14
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 47, 0
        PRINT "Jump To Byte";
        LOCATE 4, 47, 0
        PRINT "Jump To Line";
        COLOR Magenta
        LOCATE 3, 55, 0
        PRINT "B";
        LOCATE 4, 55, 0
        PRINT "L";
    CASE 6
        BoxDrawLength = 18
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 57, 0
        PRINT "HEX Screen Print";
        LOCATE 4, 57, 0
        PRINT "HEX File Print  ";
        COLOR Magenta
        LOCATE 3, 61, 0
        PRINT "S";
        LOCATE 4, 61, 0
        PRINT "F";
    CASE 7
        BoxDrawLength = 16
        GOSUB DrawBox
        COLOR White
        LOCATE 3, 65, 0
        PRINT "Search String ";
        LOCATE 4, 65, 0
        PRINT "Search Bytes  ";
        LOCATE 5, 65, 0
        PRINT "Search Replace";
        COLOR Magenta
        LOCATE 3, 72, 0
        PRINT "S";
        LOCATE 4, 72, 0
        PRINT "B";
        LOCATE 5, 72, 0
        PRINT "R";
END SELECT
RETURN

' stores area under menu.
StoreArea:
IF CurrentMenu = False THEN
    RETURN
END IF
GOSUB BoxBorder
RowX1 = 0
ColumnY1 = 0
FOR RowX2 = Xstore1 TO Xstore2
    RowX1 = RowX1 + 1
    ColumnY1 = 0
    FOR ColumnY2 = Ystore1 TO Ystore2
        ColumnY1 = ColumnY1 + 1
        ' store ascii character.
        Area1(RowX1, ColumnY1) = SCREEN(RowX2, ColumnY2)
        ' store color. (undocumented: also stores background color).
        Area2(RowX1, ColumnY1) = SCREEN(RowX2, ColumnY2, 1)
    NEXT
NEXT
' hilight menu tab.
COLOR Yellow, 0
CALL DisplayTab2(CurrentMenu)
RETURN

' restores area under menu.
RestoreArea:
IF CurrentMenu = False THEN
    RETURN
END IF
GOSUB BoxBorder
RowX1 = 0
ColumnY1 = 0
FOR RowX2 = Xstore1 TO Xstore2
    RowX1 = RowX1 + 1
    ColumnY1 = 0
    FOR ColumnY2 = Ystore1 TO Ystore2
        ColumnY1 = ColumnY1 + 1
        LOCATE RowX2, ColumnY2, 1
        ' restore color.
        TempZ = Area2(RowX1, ColumnY1)
        ' undocumented: also stores background color.
        VarB = INT(TempZ / 16)
        VarF = TempZ MOD 16
        COLOR VarF, VarB
        ' restore ascii character.
        PRINT CHR$(Area1(RowX1, ColumnY1));
    NEXT
NEXT
' remove hilight menu tab.
COLOR White, 1
CALL DisplayTab(CurrentMenu)
RETURN

' returns borders of box.
BoxBorder:
SELECT CASE CurrentMenu
    CASE 1
        Xstore1 = 2
        Xstore2 = 9
        Ystore1 = 6
        Ystore2 = 22
    CASE 2
        Xstore1 = 2
        Xstore2 = 5
        Ystore1 = 16
        Ystore2 = 27
    CASE 3
        Xstore1 = 2
        Xstore2 = 5
        Ystore1 = 26
        Ystore2 = 42
    CASE 4
        Xstore1 = 2
        Xstore2 = 6
        Ystore1 = 36
        Ystore2 = 56
    CASE 5
        Xstore1 = 2
        Xstore2 = 5
        Ystore1 = 46
        Ystore2 = 59
    CASE 6
        Xstore1 = 2
        Xstore2 = 5
        Ystore1 = 56
        Ystore2 = 73
    CASE 7
        Xstore1 = 2
        Xstore2 = 6
        Ystore1 = 64
        Ystore2 = 79
END SELECT
RETURN

' draws box from (BoxDrawX1,BoxDrawY1) To (BoxDrawX2,BoxDrawY2),
'  length of box from left to right being BoxDrawLength.
DrawBox:
COLOR Yellow
LOCATE BoxDrawX1, BoxDrawY1, 0
PRINT CHR$(ULcorner) + STRING$(BoxDrawLength - 2, Hline) + CHR$(URcorner);
FOR RowX1 = BoxDrawX1 + 1 TO BoxDrawX2 - 1
    LOCATE RowX1, BoxDrawY1, 0
    PRINT CHR$(Vline);
    LOCATE RowX1, BoxDrawY2, 0
    PRINT CHR$(Vline);
NEXT
LOCATE BoxDrawX2, BoxDrawY1, 0
PRINT CHR$(LLcorner) + STRING$(BoxDrawLength - 2, Hline) + CHR$(LRcorner);
RETURN

' process mouse move.
MoveMouse:
' select drop down menu.
IF MouseX = 1 THEN
    NewMenuSelection = False
    SELECT CASE MouseY
        CASE 6 TO 9 ' File
            NewMenuSelection = 1
        CASE 16 TO 21 ' Charts
            NewMenuSelection = 2
        CASE 26 TO 29 ' Dump
            NewMenuSelection = 3
        CASE 36 TO 39 ' Edit
            NewMenuSelection = 4
        CASE 46 TO 49 ' Jump
            NewMenuSelection = 5
        CASE 56 TO 60 ' Print
            NewMenuSelection = 6
        CASE 64 TO 69 ' Search
            NewMenuSelection = 7
    END SELECT
    IF NewMenuSelection > False THEN
        IF NewMenuSelection <> CurrentMenu THEN
            IF CurrentMenu > False THEN
                GOSUB RestoreArea
            END IF
            CurrentMenu = NewMenuSelection
            CurrentMenuSelection = 1
            GOSUB StoreArea
            GOSUB DrawMenu
            GOSUB DrawCurrentMenuSelection
        END IF
    END IF
    IF CurrentMenu > False THEN
        COLOR Yellow, 0
        CALL DisplayTab2(CurrentMenu)
    END IF
    RETURN
END IF

' check mouse selection boundaries.
SELECT CASE CurrentMenu
    CASE 1
        IF MouseX >= 3 AND MouseX <= 8 THEN
            IF MouseY >= 7 AND MouseY <= 21 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 2
        IF MouseX >= 3 AND MouseX <= 4 THEN
            IF MouseY >= 17 AND MouseY <= 26 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 3
        IF MouseX >= 3 AND MouseX <= 4 THEN
            IF MouseY >= 27 AND MouseY <= 41 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 4
        IF MouseX >= 3 AND MouseX <= 5 THEN
            IF MouseY >= 37 AND MouseY <= 55 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 5
        IF MouseX >= 3 AND MouseX <= 5 THEN
            IF MouseY >= 47 AND MouseY <= 58 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 6
        IF MouseX >= 3 AND MouseX <= 5 THEN
            IF MouseY >= 57 AND MouseY <= 72 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
    CASE 7
        IF MouseX >= 3 AND MouseX <= 6 THEN
            IF MouseY >= 65 AND MouseY <= 78 THEN
                IF MouseX - 2 <> CurrentMenuSelection THEN
                    GOSUB EraseCurrentMenuSelection
                    CurrentMenuSelection = MouseX - 2
                END IF
                GOSUB DrawCurrentMenuSelection
            END IF
        END IF
END SELECT
RETURN

' process left mouse button.
MouseButton1X:

' test mouse click above menu and clear.
IF MouseX = 1 THEN
    IF MouseY >= 1 AND MouseY <= 3 THEN
        CPUOFF = NOT CPUOFF
        IF CPUOFF THEN
            TIMER OFF
            _TITLE Program
            CALL DisplayToolBar2
        ELSE
            TIMER ON
            CALL DisplayToolBar2
        END IF
        RETURN
    END IF
    IF MouseY >= 74 AND MouseY <= 78 THEN
        CALL HelpAbout
        RETURN
    END IF
    CurrentMenu = False
    RETURN
END IF

' test mouse click below menu and clear.
SELECT CASE CurrentMenu
    CASE 1
        IF MouseX > 9 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 2
        IF MouseX > 5 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 3
        IF MouseX > 5 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 4
        IF MouseX > 6 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 5
        IF MouseX > 6 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 6
        IF MouseX > 6 THEN
            CurrentMenu = False
            RETURN
        END IF
    CASE 7
        IF MouseX > 7 THEN
            CurrentMenu = False
            RETURN
        END IF
END SELECT

' test mouse click outside edges of menu.
IF MouseX > 1 THEN
    SELECT CASE CurrentMenu
        CASE 1
            IF MouseY < 6 OR MouseY > 22 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 2
            IF MouseY < 16 OR MouseY > 27 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 3
            IF MouseY < 26 OR MouseY > 42 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 4
            IF MouseY < 36 OR MouseY > 56 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 5
            IF MouseY < 46 OR MouseY > 59 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 6
            IF MouseY < 56 OR MouseY > 73 THEN
                CurrentMenu = False
                RETURN
            END IF
        CASE 7
            IF MouseY < 66 OR MouseY > 79 THEN
                CurrentMenu = False
                RETURN
            END IF
    END SELECT
END IF

' test mouse click within menu.
IF MouseX - 2 = CurrentMenuSelection THEN
    SELECT CASE CurrentMenu
        CASE 1
            IF MouseY >= 7 AND MouseY <= 21 THEN
                GOTO MouseExit
            END IF
        CASE 2
            IF MouseY >= 17 AND MouseY <= 26 THEN
                GOTO MouseExit
            END IF
        CASE 3
            IF MouseY >= 27 AND MouseY <= 41 THEN
                GOTO MouseExit
            END IF
        CASE 4
            IF MouseY >= 37 AND MouseY <= 55 THEN
                GOTO MouseExit
            END IF
        CASE 5
            IF MouseY >= 47 AND MouseY <= 58 THEN
                GOTO MouseExit
            END IF
        CASE 6
            IF MouseY >= 57 AND MouseY <= 72 THEN
                GOTO MouseExit
            END IF
        CASE 7
            IF MouseY >= 65 AND MouseY <= 78 THEN
                GOTO MouseExit
            END IF
    END SELECT
END IF
RETURN

' exit with mouse click
MouseExit:
VarX = -1
GOSUB RestoreArea
EXIT SUB

CALL Toolbar
IF Status = 0 THEN
    VIEW PRINT 2 TO 24
ELSE
    VIEW PRINT 2 TO 23
END IF
END SUB

' redraws menu tab.
SUB DisplayTab (Var)
SELECT CASE Var
    CASE 1 ' File - 33
        LOCATE 1, 6, 0
        PRINT "File";
    CASE 2 ' Charts - 46
        LOCATE 1, 16, 0
        PRINT "Charts";
    CASE 3 ' Dump - 32
        LOCATE 1, 26, 0
        PRINT "Dump";
    CASE 4 ' Edit - 18
        LOCATE 1, 36, 0
        PRINT "Edit";
    CASE 5 ' Jump - 36
        LOCATE 1, 46, 0
        PRINT "Jump";
    CASE 6 ' Print- 25
        LOCATE 1, 56, 0
        PRINT "Print";
    CASE 7 ' Search - 31
        LOCATE 1, 64, 0
        PRINT "Search";
END SELECT
COLOR White, 0
END SUB

' redraws menu tab.
SUB DisplayTab2 (Var)
SELECT CASE Var
    CASE 1 ' File - 33
        LOCATE 1, 6, 0
        PRINT "File";
        COLOR White, 0
        LOCATE 1, 6, 0
        PRINT "F";
    CASE 2 ' Charts - 46
        LOCATE 1, 16, 0
        PRINT "Charts";
        COLOR White, 0
        LOCATE 1, 16, 0
        PRINT "C";
    CASE 3 ' Dump - 32
        LOCATE 1, 26, 0
        PRINT "Dump";
        COLOR White, 0
        LOCATE 1, 26, 0
        PRINT "D";
    CASE 4 ' Edit - 18
        LOCATE 1, 36, 0
        PRINT "Edit";
        COLOR White, 0
        LOCATE 1, 36, 0
        PRINT "E";
    CASE 5 ' Jump - 36
        LOCATE 1, 46, 0
        PRINT "Jump";
        COLOR White, 0
        LOCATE 1, 46, 0
        PRINT "J";
    CASE 6 ' Print- 25
        LOCATE 1, 56, 0
        PRINT "Print";
        COLOR White, 0
        LOCATE 1, 56, 0
        PRINT "P";
    CASE 7 ' Search - 31
        LOCATE 1, 64, 0
        PRINT "Search";
        COLOR White, 0
        LOCATE 1, 64, 0
        PRINT "S";
END SELECT
COLOR White, 0
END SUB

REM processes mouse activity.
FUNCTION MouseDriver
STATIC X1 AS INTEGER, Y1 AS INTEGER ' store old values
MouseX = 0: MouseY = 0
IF _MOUSEINPUT THEN
    X = CINT(_MOUSEX): Y = CINT(_MOUSEY) ' X,Y return single
    IF X <> X1 OR Y <> Y1 THEN
        X1 = X: Y1 = Y
        MouseX = Y: MouseY = X ' X,Y are reversed
        WHILE _MOUSEINPUT: WEND ' empty buffer
        MousePressed = -1
    END IF
    MouseButton1 = _MOUSEBUTTON(1)
    IF MouseButton1 THEN
        MouseX = Y1
        MouseY = X1
        MousePressed = -1
    END IF
    MouseButton2 = _MOUSEBUTTON(2)
    IF MouseButton2 THEN
        MouseX = Y1
        MouseY = X1
        MousePressed = -1
    END IF
    MouseButton3 = _MOUSEBUTTON(3)
    IF MouseButton3 THEN
        MousePressed = -1
    END IF
    MouseWheel = _MOUSEWHEEL
END IF
END FUNCTION

FUNCTION INKEYx$
X$ = INKEY$
IF LEN(X$) THEN
    INKEYx$ = X$
    KeyPressed = -1
END IF
END FUNCTION

FUNCTION ClearKeyboard
WHILE LEN(INKEYx$): WEND
END FUNCTION

REM Code for File Menu Box taken from Hexedit 2017

REM File menu box routine:

SUB Menu3 (N$, C$, D$, X$, E$)
DIM hfind AS _OFFSET

' read config file.
CALL ReadConfigFile

REM Call to Menu() Returns:
REM   N$  -  Network path
REM       \\server\share
REM       Without trailing slash
REM         Or Nul for no net
REM   C$  -  Drive letter
REM       Normally C without : or \
REM   D$  -  Directory
REM       Without leading or trailing \ (can be null)
REM       Can be long pathname. No trailing spaces.
REM   X$  -  Filespec
REM       Normally *.*
REM   E$  -  Filename.ext
REM       Can be long filename. No trailing spaces.
REM
REM Concatenate upon return:
REM   IF N$ = Nul THEN
REM      IF D$ = Nul THEN
REM         Filename = C$ + ":\" + E$
REM      ELSE
REM         Filename = C$ + ":\" + D$ + "\" + E$
REM      END IF
REM   ELSE
REM      IF D$ = Nul THEN
REM         Filename = N$ + "\" + E$
REM      ELSE
REM         Filename = N$ + "\" + D$ + "\" + E$
REM      END IF
REM   END IF
REM
REM Although netpath may be specified as c:\\server\share\dir1\filename.ext
REM this box does not fully support mixed drive letter and netpath.
REM
REM During editing,
REM   I$  -  Current keystroke.
REM   F$  -  Fully qualified pathname with filename and drive letter.
REM   Q$  -  Current subdirectory.
REM   D1$ -  Temporary directory storage.
REM   X1$ -  Temporary file entry area path/file.
REM   X2$ -  Temporary filename storage.
REM   G$  -  Pathname constructed for KILL, MKDIR, RMDIR.
REM
REM Mouse variables,
REM   IsMenu =
REM     1  -  Click in file box.
REM     2  -  Click in dir box.
REM     3  -  Click in drive box.
REM     4  -  Click in entry area.
REM     5  -  Click on 'x' to exit.

' reset flags.
Ambiguate = AmbiguateSwitch
FileSort = False
DirSort = False

' display menu screen.
CALL FileMainMenu

' initialize drives.
Current.Drive2 = False
GOSUB InitDrive

' get current filespec.
GOSUB InitFileSpec

' get current directory spec.
GOSUB InitDirSpec

' read all filenames.
GOSUB ClearFileBox
FileLoadMessage = "Initing files.."
CALL LoadFileSpec(FileY$, F$)
GOSUB DisplayFileSpec
GOSUB DisplayCurrentPath

' read all subdirectories.
GOSUB ClearDirBox
DirLoadMessage = "Initing dirs.."
CALL LoadDirSpec(DirY$, N$, C$, D$)
GOSUB DisplayDirSpec
GOSUB DisplayCurrentDir
GOSUB DisplayCurrentPath

' display first directory.
Current.Dir! = 1!
Dir.Box.Line = 1
GOSUB DisplayInitDir

' display all drives.
GOSUB InitDrives

' display first file.
GOSUB ClearLongName2
IF Num.Files > 0 THEN
    Current.File! = 1!
    Box.Line = 1
    GOSUB InitFiles
END IF

' reset filename attribute overrides.
RequireReadOnly = False
RequireHidden = False
RequireSystem = False
RequireArchive = False
RequireDirReadOnly = False
RequireDirHidden = False
RequireDirSystem = False
RequireDirArchive = False

REM File entry area variables:
REM   ProcessEditLine - processes X$ edit line when set to True.
REM   Xstart     - hilights file entry area for overstrike when set to True.
REM   Xposition1 - left position in file entry area editing string.
REM   Xposition2 - position of character in file entry area editing string.
REM

REM main input loop..

Xstart = True
Xposition1 = 1
Insert = -1
DO
    ' reset filespec.
    IF X$ = Nul THEN
        X$ = "*.*"
        Xposition1 = 1
        Xposition2 = 3
    END IF

    ' display file entry area.
    GOSUB DisplayFileLine

    ' start input loop.
    DO
        ' reset input.
        I$ = Nul

        ' reset edit line.
        ProcessEditLine = False

        ' process input.
        Control.Break = False
        Monitor.Box = False
        DO
            ' check break
            IF Control.Break OR Monitor.Box THEN
                Control.Break = False
                Monitor.Box = False
                IF Insert THEN
                    LOCATE , , 1, 8, 8 ' line cursor
                ELSE
                    LOCATE , , 1, 0, 8 ' block cursor
                END IF
            END IF

            ' get keypress.
            _LIMIT 100
            I$ = INKEYx$

            ' process keypress.
            IF LEN(I$) THEN
                IsMenu = False
                EXIT DO
            END IF

            ' call mouse subroutine.
            Z = MouseDriver

            ' check exit symbol.
            IF IsMenu = 5 THEN
                I$ = CHR$(27)
                EXIT DO
            END IF

            ' check mouse wheel down
            IF MouseWheel = -1 THEN
                MousePressed = -1
                MouseWheel = 0
                IF WheelReverse THEN
                    I$ = CHR$(9) ' down
                    GOTO TabFile
                END IF
            END IF
            IF MouseWheel = 1 THEN
                MousePressed = -1
                MouseWheel = 0
                IF WheelReverse = 0 THEN
                    I$ = CHR$(9) ' down
                    GOTO TabFile
                END IF
            END IF

            ' check left mouse button.
            IF MouseButton1 THEN
                GOSUB MouseButton1Z
                SELECT CASE IsMenu
                    CASE 1 ' file box.
                        IF I$ = CHR$(13) THEN
                            E$ = RTRIM$(Filenames(Current.File!))
                            GOTO EndLoop2
                        ELSE
                            GOSUB FileBoxKey
                        END IF
                    CASE 2 ' dir box.
                        IF I$ = CHR$(13) THEN
                            GOSUB GetNewPath
                            EXIT DO
                        ELSE
                            GOSUB DirBoxKey
                        END IF
                    CASE 3 ' drive box.
                        IF I$ = CHR$(13) THEN
                            X$ = CHR$(Current.Drive2 + 64) + ":\*.*"
                            ProcessEditLine = True
                            EXIT DO
                        ELSE
                            GOSUB DriveBoxKey
                        END IF
                END SELECT
            ELSE
                ' check mouse position.
                IF MouseX OR MouseY THEN
                    GOSUB MouseOverX1
                    GOSUB DisplayFileLine
                END IF
            END IF
        LOOP

        ' ignore slash
        IF I$ = "/" THEN
            I$ = Nul
        END IF

        StartLoop0:

        ' check alt-<key>
        IF LEN(I$) = 2 THEN
            GOSUB GetAltKey
            IF VarZ THEN
                ' specify drive letter.
                X$ = CHR$(Var2 + 64) + ":\*.*"
                I$ = Nul
                ProcessEditLine = True
                EXIT DO
            END IF
        END IF

        StartLoop1:

        ' check extended key
        IF LEN(I$) = 2 THEN
            SELECT CASE ASC(RIGHT$(I$, 1))
                CASE 71, 119 ' Home/Ctrl-Home
                    IF Xposition2 > 0 THEN
                        Xstart = False
                        Xposition1 = 1
                        Xposition2 = 0
                        GOSUB DisplayFileLine
                    END IF
                CASE 80, 145, 160 ' Down/Ctrl-Down/Alt-Down
                    I$ = CHR$(9)
                    GOTO TabFile
                CASE 79, 117 ' End/Ctrl-End
                    IF Xposition2 < LEN(X$) THEN
                        Xstart = False
                        Xposition2 = LEN(X$)
                        Xposition1 = Xposition2 - 25
                        IF Xposition1 < 1 THEN
                            Xposition1 = 1
                        END IF
                        GOSUB DisplayFileLine
                    END IF
                CASE 82 ' insert
                    Insert = NOT Insert
                    IF Insert THEN
                        LOCATE , , 1, 8, 8 ' line cursor
                    ELSE
                        LOCATE , , 1, 0, 8 ' block cursor
                    END IF
                CASE 83 ' Delete
                    IF LEN(X$) > 0 THEN
                        Xstart = False
                        X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                        GOSUB DisplayFileLine
                    END IF
                CASE 75, 115, 155 ' Left/Ctrl-Left/Alt-Left
                    IF LEN(X$) THEN
                        IF Xposition2 > 0 THEN
                            Xposition2 = Xposition2 - 1
                            IF Xposition2 - 25 <= Xposition1 THEN
                                Xposition1 = Xposition1 - 1
                                IF Xposition1 < 1 THEN
                                    Xposition1 = 1
                                END IF
                                GOSUB DisplayFileLine
                            END IF
                        END IF
                    END IF
                CASE 77, 116, 157 ' Right/Ctrl-Right/Alt-Right
                    IF LEN(X$) THEN
                        IF Xposition2 < LEN(X$) THEN
                            Xposition2 = Xposition2 + 1
                            IF Xposition2 > Xposition1 + 25 THEN
                                Xposition1 = Xposition1 + 1
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                    END IF
                CASE 59 ' F1  - change drive letter.
                    I$ = Nul
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 Yellow, 0
                    Printf "Drive letter: "
                    DO
                        _LIMIT 100
                        Z$ = INKEYx$
                        IF LEN(Z$) = 1 THEN
                            Z$ = UCASE$(Z$)
                            IF Z$ >= "A" AND Z$ <= "Z" THEN
                                EXIT DO
                            END IF
                        END IF
                    LOOP
                    GOSUB DisplayFilename6
                    I$ = Nul
                    X$ = Z$ + ":\*.*"
                    ProcessEditLine = True
                    EXIT DO
                CASE 60 ' F2  - toggle heap sorting.
                    I$ = Nul
                    HeapSortOff = NOT HeapSortOff
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    IF HeapSortOff THEN
                        Printf "Sorting Off. Press <esc>:"
                    ELSE
                        Printf "Sorting On. Press <esc>:"
                    END IF
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 61 ' F3 - move file menu box.
                    I$ = Nul
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Use cursor keys. Press <esc> to quit."
                    CALL MoveMenu
                    CALL BottomRow
                    GOSUB DisplayCurrentPath
                    IF Num.Files > 0 THEN
                        GOSUB DisplayLongName2
                    END IF
                    GOSUB DisplayFilename6
                    GOSUB DisplayFileLine
                CASE 62 ' F4 - backward 1 drive letter.
                    I$ = Nul
                    GOSUB MoveDriveUp
                    X$ = CHR$(Current.Drive2 + 64) + ":\*.*"
                    ProcessEditLine = True
                    EXIT DO
                CASE 63 ' F5 - forward 1 drive letter.
                    I$ = Nul
                    GOSUB MoveDriveDown
                    X$ = CHR$(Current.Drive2 + 64) + ":\*.*"
                    ProcessEditLine = True
                    EXIT DO
                CASE 64 ' F6 - toggle ambiguate.
                    I$ = Nul
                    Ambiguate = NOT Ambiguate
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    IF Ambiguate = False THEN
                        Printf "Ambiguation Off. Press <esc>:"
                    ELSE
                        Printf "Ambiguation On. Press <esc>:"
                    END IF
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 65 ' F7 - redraw screen
                    GOSUB RedrawScreen
                    EXIT DO
                CASE 66 ' F8 - attribute override
                    I$ = Nul
                    CALL OverridePrompt
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 67 ' F9 - help screen
                    I$ = Nul
                    CALL DisplayMenuHelp
                    GOSUB RedrawScreen
                CASE 68 ' F10 - exit program
                    I$ = CHR$(27)
                    EXIT DO
                CASE 133 ' F11 - drop to DOS
                    I$ = Nul
                    SHELL
                    EXIT DO
                CASE 134 ' F12 - enter filespec
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Enter file spec?"
                    X$ = LineInput$(Xcoor + 1, Ycoor + 19, 21)
                    I$ = Nul
                    IF LEN(X$) THEN
                        E$ = X$
                        IF MID$(E$, 2, 1) = ":" THEN
                            C$ = LEFT$(E$, 1)
                            E$ = MID$(E$, 3)
                            N$ = ""
                        END IF
                        FOR VarQ = LEN(E$) TO 1 STEP -1
                            IF MID$(E$, VarQ, 1) = "\" THEN
                                D$ = LEFT$(E$, VarQ)
                                E$ = MID$(E$, VarQ + 1)
                                EXIT FOR
                            END IF
                        NEXT
                        IF RIGHT$(D$, 1) = "\" THEN
                            D$ = LEFT$(D$, LEN(D$) - 1)
                        END IF
                        IF LEFT$(D$, 1) = "\" THEN
                            D$ = MID$(D$, 2)
                        END IF
                        X$ = "*.*"
                        I$ = CHR$(13)
                        EXIT DO
                    END IF
                    GOSUB DisplayFilename6
                    GOSUB DisplayFileLine
                    EXIT DO
                CASE 104 ' Alt-F1 - toggle quiet mode.
                    I$ = Nul
                    SELECT CASE QuietSwitch
                        CASE -1
                            QuietSwitch = 0
                        CASE 0
                            QuietSwitch = 1
                        CASE 1
                            QuietSwitch = 2
                        CASE 2
                            QuietSwitch = 3
                        CASE 3
                            QuietSwitch = 4
                        CASE 4
                            QuietSwitch = 5
                        CASE 5
                            QuietSwitch = 6
                        CASE 6
                            QuietSwitch = -1
                    END SELECT
                    SELECT CASE QuietSwitch
                        CASE 0
                            P$ = "Quiet switch off."
                        CASE -1
                            P$ = "Quiet switch on."
                        CASE ELSE
                            P$ = "Quiet switch level" + STR$(QuietSwitch) + "."
                    END SELECT
                    P$ = P$ + " Press <esc>."
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf P$
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 105 ' Alt-F2 - file exclusion entry.
                    I$ = Nul
                    CALL ExcludeList(1, Y$)
                    FileY$ = Y$
                    GOSUB RedrawScreen
                    EXIT DO
                CASE 106 ' Alt-F3 - dir exclusion entry.
                    I$ = Nul
                    CALL ExcludeList(2, Y$)
                    DirY$ = Y$
                    GOSUB RedrawScreen
                    EXIT DO
                CASE 107 ' Alt-F4 - toggle file sort ascending/descending.
                    I$ = Nul
                    FileSort = NOT FileSort
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    IF FileSort THEN
                        Printf "File sort descending. Press <esc>:"
                    ELSE
                        Printf "File sort ascending. Press <esc>:"
                    END IF
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 108 ' Alt-F5 - toggle dir sort ascending/descending.
                    I$ = Nul
                    DirSort = NOT DirSort
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    IF DirSort THEN
                        Printf "Dir sort descending. Press <esc>:"
                    ELSE
                        Printf "Dir sort ascending. Press <esc>:"
                    END IF
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 109 ' Alt-F6 - toggle file date/time switch.
                    I$ = Nul
                    SELECT CASE FileSwitch
                        CASE 0
                            FileSwitch = 1
                        CASE 1
                            FileSwitch = 2
                        CASE 2
                            FileSwitch = 0
                    END SELECT
                    P$ = "File switch level" + STR$(FileSwitch) + "."
                    P$ = P$ + " Press <esc>."
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf P$
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 110 ' Alt-F7 - toggle file size switch.
                    I$ = Nul
                    SELECT CASE FileSizeType
                        CASE 0
                            FileSizeType = 1
                            Var$ = "KB."
                        CASE 1
                            FileSizeType = 2
                            Var$ = "MB."
                        CASE 2
                            FileSizeType = 3
                            Var$ = "GB."
                        CASE 3
                            FileSizeType = 0
                            Var$ = "B."
                    END SELECT
                    P$ = "File size switch: " + Var$
                    P$ = P$ + " Press <esc>."
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf P$
                    CALL PromptInkey
                    GOSUB UpdateDisplay
                    EXIT DO
                CASE 111 ' Alt-F8 - network path.
                    I$ = Nul
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Net path?"
                    Z$ = LineInput$(Xcoor + 1, Ycoor + 12, 28)
                    X$ = Nul
                    IF LEN(Z$) THEN
                        IF LEFT$(Z$, 2) = "\\" THEN
                            X$ = Z$
                        END IF
                    END IF
                    GOSUB DisplayFilename6
                    GOSUB DisplayFileLine
                    ProcessEditLine = True
                    EXIT DO
                CASE 112 ' Alt-F9 - default network path.
                    I$ = Nul
                    X$ = Nul
                    ' store default server name
                    IF RTRIM$(DefaultNetPath) <> Nul THEN
                        X$ = "\\" + RTRIM$(DefaultNetPath) + "\C\"
                        GOSUB DisplayFilename6
                        GOSUB DisplayFileLine
                        ProcessEditLine = True
                        EXIT DO
                    END IF
                CASE 113 ' Alt-F10 - load default net w/ drive letter
                    I$ = Nul
                    X$ = Nul
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Drive letter: "
                    DO
                        _LIMIT 100
                        Z$ = INKEYx$
                        IF LEN(Z$) = 1 THEN
                            Z$ = UCASE$(Z$)
                            IF Z$ >= "A" AND Z$ <= "Z" THEN
                                EXIT DO
                            END IF
                        END IF
                    LOOP
                    Printf Z$
                    ' store default server name
                    IF RTRIM$(DefaultNetPath) <> Nul THEN
                        X$ = "\\" + RTRIM$(DefaultNetPath) + "\" + UCASE$(Z$) + "\"
                        GOSUB DisplayFilename6
                        GOSUB DisplayFileLine
                        ProcessEditLine = True
                        EXIT DO
                    END IF
                CASE ELSE
                    I$ = Nul ' nul activity
            END SELECT
            I$ = Nul ' nul keystroke
        END IF

        REM tab key label.
        TabFile:

        ' Tab moves to file loop, or
        '   directory loop if there are no files.
        IF I$ = CHR$(9) THEN
            IF Num.Files > 0 THEN
                GOSUB FileInputLoop

                ' check function key.
                IF LEN(I$) = 2 THEN
                    VarQ = ASC(RIGHT$(I$, 1))
                    SELECT CASE VarQ
                        CASE 59 TO 68, 104 TO 113, 133, 134
                            GOTO StartLoop1
                    END SELECT
                    GOSUB GetAltKey
                    IF VarZ THEN
                        GOTO StartLoop0
                    END IF
                END IF

                ' tab from drive input loop.
                IF I$ = CHR$(9) THEN
                    I$ = Nul
                    EXIT DO
                END IF

                ' shift-tab from dir input loop.
                IF I$ = CHR$(0) + CHR$(15) THEN
                    I$ = Nul
                    EXIT DO
                END IF

                ' click on file entry area.
                IF IsMenu = 4 THEN
                    GOSUB DisplayCurrentPath
                    IF Num.Files > 0 THEN
                        GOSUB DisplayLongName2
                    END IF
                    EXIT DO
                END IF
                IF I$ = CHR$(13) THEN
                    SELECT CASE IsMenu
                        CASE 1 ' file box.
                            E$ = RTRIM$(Filenames(Current.File!))
                            GOTO EndLoop2
                        CASE 2 ' dir box.
                            GOSUB GetNewPath
                        CASE 3 ' drive box.
                            X$ = CHR$(Current.Drive2 + 64) + ":\*.*"
                            I$ = Nul
                            ProcessEditLine = True
                            EXIT DO
                        CASE 4 ' entry area.
                            GOSUB DisplayCurrentPath
                            IF Num.Files > 0 THEN
                                GOSUB DisplayLongName2
                            END IF
                            EXIT DO
                        CASE 5
                            I$ = CHR$(27)
                            EXIT DO
                        CASE ELSE
                            E$ = RTRIM$(Filenames(Current.File!))
                            GOTO EndLoop2
                    END SELECT
                END IF
            ELSE
                GOSUB DirInputLoop
                ' check function key.
                IF LEN(I$) = 2 THEN
                    VarQ = ASC(RIGHT$(I$, 1))
                    SELECT CASE VarQ
                        CASE 59 TO 68, 104 TO 113, 133, 134
                            GOTO StartLoop1
                    END SELECT
                    GOSUB GetAltKey
                    IF VarZ THEN
                        GOTO StartLoop0
                    END IF
                END IF
                ' tab from drive input loop.
                IF I$ = CHR$(9) THEN
                    I$ = Nul
                    EXIT DO
                END IF
                ' shift-tab from drive input loop.
                IF I$ = CHR$(0) + CHR$(15) THEN
                    I$ = Nul
                    EXIT DO
                END IF
                IF IsMenu = 4 THEN
                    GOSUB DisplayCurrentPath
                    IF Num.Files > 0 THEN
                        GOSUB DisplayLongName2
                    END IF
                    EXIT DO
                END IF
                IF I$ = CHR$(13) THEN
                    SELECT CASE IsMenu
                        CASE 1 ' file box.
                            E$ = RTRIM$(Filenames(Current.File!))
                            GOTO EndLoop2
                        CASE 2 ' dir box.
                            GOSUB GetNewPath
                        CASE 3 ' drive box.
                            X$ = CHR$(Current.Drive2 + 64) + ":\*.*"
                            I$ = Nul
                            ProcessEditLine = True
                            EXIT DO
                        CASE 4 ' entry area.
                            GOSUB DisplayCurrentPath
                            IF Num.Files > 0 THEN
                                GOSUB DisplayLongName2
                            END IF
                            EXIT DO
                        CASE 5
                            I$ = CHR$(27)
                            EXIT DO
                        CASE ELSE
                            GOSUB GetNewPath
                    END SELECT
                END IF
            END IF
        END IF

        ' test break key.
        IF I$ = CHR$(3) THEN ' Ctrl-C
            I$ = "" ' ignore break
        END IF

        ' enter key.
        IF I$ = CHR$(13) THEN
            I$ = Nul
            ProcessEditLine = True
            EXIT DO
        END IF

        ' escape key.
        IF I$ = CHR$(27) THEN
            EXIT DO ' exit box.
        END IF

        ' backspace.
        IF I$ = CHR$(8) THEN
            IF LEN(X$) THEN
                IF Xstart = True AND Xposition2 = 0 THEN
                    X$ = Nul
                    Xstart = False
                    Xposition1 = 1
                    Xposition2 = 0
                    GOSUB DisplayFileLine
                ELSE
                    IF Xposition2 > 0 THEN
                        Xstart = False
                        X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                        Xposition2 = Xposition2 - 1
                        IF Xposition2 - 25 <= Xposition1 THEN
                            Xposition1 = Xposition1 - 1
                            IF Xposition1 < 1 THEN
                                Xposition1 = 1
                            END IF
                        END IF
                        GOSUB DisplayFileLine
                    END IF
                END IF
            END IF
        ELSE
            ' append/insert character into filename area.
            IF LEN(I$) THEN
                ' check input length.
                IF LEN(X$) >= 259 THEN
                    X$ = Nul
                    Xstart = False
                    Xposition1 = 1
                    Xposition2 = 0
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Edit line error. Press <esc>:"
                    CALL PromptInkey
                ELSE
                    IF Xstart THEN
                        X$ = Nul
                        Xstart = False
                        Xposition1 = 1
                        Xposition2 = 0
                        GOSUB DisplayFileLine
                    END IF
                    IF Insert THEN
                        X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                    ELSE
                        MID$(X$, Xposition2 + 1, 1) = I$
                    END IF
                    Xposition2 = Xposition2 + 1
                    IF Xposition2 > 26 THEN
                        Xposition1 = Xposition1 + 1
                    END IF
                    GOSUB DisplayFileLine
                END IF
            END IF
        END IF
    LOOP

    ' exit menu box w/o filename.
    IF I$ = CHR$(27) THEN
        E$ = Nul
        EXIT DO
    END IF

    ' exit menu box with filename.
    IF I$ = CHR$(13) THEN
        EXIT DO
    END IF

    REM Process edit line in X$
    REM
    REM Parses
    REM   c:\path\filename.ext
    REM     or
    REM   \\netpath\share\path
    REM
    REM With partial pathname override,
    REM filename ambiguation detection, and
    REM multi-dot imbedded path retracting.

    ' check process flag.
    IF ProcessEditLine THEN

        ' store values.
        D1$ = D$
        X2$ = F$

        ' change netpath.
        IF LEFT$(X$, 2) = "\\" THEN
            IF RIGHT$(X$, 1) <> "\" THEN
                X$ = X$ + "\"
            END IF
            N$ = X$

            ' display message.
            Colorf White
            Locatef Xcoor + 1, Ycoor + 12, 1
            Printf SPACE$(26)
            Locatef Xcoor + 1, Ycoor + 12, 1
            Printf "Checking path.."
            Colorf Plain

            ' check netpath not available.
            DataError = NETPATHEXISTS(N$)
            IF DataError THEN
                N$ = Nul
                GOSUB PromptNet
                GOSUB InitFileSpec
                Xstart = True
                GOTO EndLoop1
            END IF

            ' store new netpath.
            N$ = LEFT$(N$, LEN(N$) - 1)
            D$ = Nul ' root.
            X$ = "*.*"

            ' parse dospath from netpath
            VX = INSTR(3, N$, "\") + 1
            VX = INSTR(VX + 1, N$, "\")
            IF VX THEN
                D$ = MID$(N$, VX)
                N$ = LEFT$(N$, VX - 1)
            END IF
            New.Net = True
        END IF

        ' change drive.
        IF MID$(X$, 2, 1) = ":" THEN

            ' display message.
            Colorf White
            Locatef Xcoor + 1, Ycoor + 12, 1
            Printf SPACE$(26)
            Locatef Xcoor + 1, Ycoor + 12, 1
            Printf "Loading drive.."
            Colorf Plain

            ' check drive.
            Z$ = UCASE$(LEFT$(X$, 1))
            X$ = MID$(X$, 3)

            ' check last drive.
            V = ASC(Z$) - 64
            IF V > Last.Drive THEN
                GOSUB PromptDisk
                GOSUB InitFileSpec
                Xstart = True
                GOTO EndLoop1
            END IF

            ' check disk available.
            DataError = DRIVEEXISTS(V)

            ' check disk ready.
            IF DataError THEN
                GOSUB PromptDisk
                GOSUB InitFileSpec
                Xstart = True
                GOTO EndLoop1
            END IF

            ' check media exists.
            IF DataError = False THEN
                IF MEDIAEXISTS2(V) = False THEN
                    DataError = True
                END IF
            END IF

            ' check disk ready.
            IF DataError THEN
                GOSUB PromptMedia
                GOSUB InitFileSpec
                Xstart = True
                GOTO EndLoop1
            END IF

            ' store new drive.
            IF C$ <> Z$ OR N$ <> Nul THEN
                N$ = Nul
                C$ = Z$
                New.Drive = V
                IF X$ = Nul THEN
                    X$ = "\"
                END IF
                IF RIGHT$(X$, 1) = "\" THEN
                    X$ = X$ + "*.*"
                END IF
            END IF
        END IF

        ' parse out path.
        CALL Parse.Path(D$, X$)

        ' parse off directory.
        IF INSTR(X$, "\") THEN
            FOR Var = LEN(X$) TO 1 STEP -1
                IF MID$(X$, Var, 1) = "\" THEN
                    EXIT FOR
                END IF
            NEXT

            ' parse filename.
            SELECT CASE Var
                CASE 1 ' check root.
                    D$ = Nul
                    X$ = MID$(X$, Var + 1)
                CASE LEN(X$) ' check nul filename.
                    D$ = LEFT$(X$, Var - 1)
                    X$ = Nul
                    ' parse directory/filename.
                CASE ELSE
                    IF LEFT$(X$, 1) = "\" THEN
                        D$ = LEFT$(X$, Var - 1)
                    ELSE
                        D$ = D$ + "\" + LEFT$(X$, Var - 1)
                    END IF
                    X$ = MID$(X$, Var + 1)
            END SELECT
        END IF

        ' process directory.
        IF LEFT$(D$, 1) = "\" THEN
            D$ = MID$(D$, 2)
        END IF
        IF RIGHT$(D$, 1) = "\" THEN
            D$ = LEFT$(D$, LEN(D$) - 1)
        END IF
        IF LEFT$(D$, 1) = "\" THEN
            D$ = Nul
        END IF
        IF RIGHT$(D$, 1) = "\" THEN
            D$ = Nul
        END IF
        IF D$ = "\" THEN
            D$ = Nul
        END IF

        ' check directory.
        GOSUB CheckPath
        IF ValidPath = False THEN
            D$ = D1$
            X$ = X1$
            Xposition1 = 1
            Xposition2 = 0
            GOTO EndLoop1
        END IF

        ' store directory.
        D3$ = D$
        X3$ = X$

        ' prevent directory from loading twice.
        LoadedDir = False

        ' check filename.
        IF X$ = STRING$(LEN(X$), ".") THEN
            ' strip multiple dots from path.
            CALL Parse.Path2(D$, X$)
        END IF

        ' construct partial path.
        IF D$ = Nul THEN
            D$ = X$
        ELSE
            IF X$ <> Nul THEN
                D$ = D$ + "\" + X$
            END IF
        END IF

        ' check path.
        GOSUB CheckPath

        ' check valid path.
        IF ValidPath = False THEN
            D$ = D3$
            X$ = X3$
        END IF

        ' check valid path.
        IF ValidPath THEN
            ' check directory already loaded.
            IF D$ <> D1$ OR New.Drive > 0 OR New.Net THEN
                ' load current directory.
                DirLoadMessage = "Loading dirs.."
                CALL LoadDirSpec(DirY$, N$, C$, D$)
                New.Drive = True
                New.Net = False
            END IF
            ' reset filename.
            X$ = Nul
            ' set loaded flag.
            LoadedDir = True
            GOSUB DisplayCurrentDir
        END IF

        ' check filename.
        IF X$ = Nul THEN
            GOSUB InitFileSpec
        END IF

        ' check loaded dir.
        IF LoadedDir = False THEN
            ' check filename characters.
            IF INSTR(X$, "?") = False THEN
                IF INSTR(X$, "*") = False THEN
                    ' search for directory spec override.
                    FOR Var! = 1! TO Num.Dirs
                        S$ = UCASE$(RTRIM$(Directories(Var!)))
                        ' check X$ is directory in current one.
                        IF D$ = Nul THEN
                            TempASCIIZ = "\" + S$ + CHR$(0)
                        ELSE
                            TempASCIIZ = "\" + D$ + "\" + S$ + CHR$(0)
                        END IF
                        Flag = False
                        ' get conanicalized directory name.
                        V4 = False
                        GOSUB ShortFilenameX
                        ' check directory name.
                        IF UCASE$(X$) = Short.Filename$ THEN
                            Flag = True
                        END IF
                        ' check directory name.
                        IF Flag = False THEN
                            IF UCASE$(X$) = S$ THEN
                                Flag = True
                            END IF
                        END IF
                        IF Flag THEN
                            ' store directory.
                            S$ = RTRIM$(Directories(Var!))
                            IF D$ = Nul THEN
                                D$ = S$
                            ELSE
                                D$ = D$ + "\" + S$
                            END IF
                            ' restore filename override.
                            GOSUB InitFileSpec
                            EXIT FOR
                        END IF
                    NEXT
                END IF
            END IF
        END IF

        ' store current filespec.
        IF X$ = "." THEN
            X$ = Nul
        END IF

        ' store and load directories.
        IF LoadedDir = False THEN
            ' check directory already loaded.
            IF D$ <> D1$ OR New.Drive > 0 OR New.Net THEN
                ' load directories.
                GOSUB LoadNewDir
                New.Drive = True
                New.Net = False
            END IF
        ELSE
            ' init directories.
            GOSUB LoadNewDir2
            GOSUB DisplayCurrentDir
        END IF

        ' store filename.
        GOSUB InitDirSpec

        ' store filename.
        X1$ = X$
        Xposition1 = 1
        Xposition2 = 0

        ' compare filespecs.
        IF UCASE$(F$) <> UCASE$(X2$) THEN
            ' store filespec.
            X2$ = F$
            ' load file box.
            GOSUB LoadNewFiles
        END IF

        ' check filename characters.
        IF INSTR(X$, "?") = False THEN
            IF INSTR(X$, "*") = False THEN
                ' search for filename spec override.
                FOR Var! = 1! TO Num.Files
                    S$ = UCASE$(RTRIM$(Filenames(Var!)))
                    ' check X$ is filename in current directory.
                    IF D$ = Nul THEN
                        TempASCIIZ = "\" + S$ + CHR$(0)
                    ELSE
                        TempASCIIZ = "\" + D$ + "\" + S$ + CHR$(0)
                    END IF
                    Flag = False
                    ' get conanicalized filename.
                    V4 = False
                    GOSUB ShortFilenameX
                    ' check filename.
                    IF UCASE$(X$) = Short.Filename$ THEN
                        Flag = True
                    END IF
                    ' check filename.
                    IF Flag = False THEN
                        IF UCASE$(X$) = S$ THEN
                            Flag = True
                        END IF
                    END IF
                    IF Flag THEN
                        ' select file for editing.
                        E$ = RTRIM$(Filenames(Var!))
                        EXIT DO
                    END IF
                NEXT
            END IF
        END IF

        ' construct dir/filename.
        V$ = MakeFilename$(N$, C$, D$, X$)

        ' check filename characters.
        IF INSTR(V$, "?") = False THEN
            IF INSTR(V$, "*") = False THEN
                ' select file for editing.
                E$ = X$
                EXIT DO
            END IF
        END IF

        ' display all drives.
        IF New.Drive = True THEN
            Current.Drive2 = ASC(C$) - 64
            GOSUB InitDrives
            New.Drive = False
        END IF

        ' restart editing loop.
        Xstart = True
    END IF
    EndLoop1:
LOOP

' exit editing loop.
EndLoop2:

' exit file menu box.
Colorf2 7, 0
CLS

' return to editor.
_KEYCLEAR
EXIT SUB

' reset file menu box.
UpdateDisplay:
GOSUB DisplayFilename6
Colorf2 Plain, 0
' init dir box.
GOSUB ClearDirBox
DirLoadMessage = "Updating dirs.."
CALL LoadDirSpec(DirY$, N$, C$, D$)
GOSUB DisplayDirSpec
Current.Dir! = 1!
Dir.Box.Line = 1
GOSUB DisplayCurrentDir
' init file box.
GOSUB ClearFileBox
FileLoadMessage = "Updating files.."
CALL LoadFileSpec(FileY$, F$)
GOSUB DisplayFileSpec
GOSUB DisplayCurrentPath
' display first file.
GOSUB ClearLongName2
IF Num.Files > 0 THEN
    Current.File! = 1!
    Box.Line = 1
    GOSUB InitFiles
END IF
' init drive box.
GOSUB InitDrives
RETURN

' init filespec.
InitFileSpec:
' compare stored filename.
IF X1$ = Nul THEN
    X$ = "*.*"
    Xposition1 = 1
    Xposition2 = 3
ELSE
    ' restore filename spec.
    X$ = X1$
    Xposition1 = 1
    Xposition2 = 0
END IF
' check filename.
IF X$ = "." THEN
    X$ = Nul
END IF
RETURN

' init current directory.
InitDirSpec:
F$ = MakeFilename$(N$, C$, D$, X$)
RETURN

' inits dir box with new dirspec.
LoadNewDir:
Current.Dir! = 1!
Dir.Box.Line = 1
GOSUB ClearDirBox
GOSUB CheckPath
IF ValidPath = False THEN
    D$ = D1$
    X$ = X1$
    Xposition1 = 1
    Xposition2 = 0
END IF
DirLoadMessage = "Reloading dirs.."
CALL LoadDirSpec(DirY$, N$, C$, D$)
GOSUB DisplayDirSpec
GOSUB DisplayCurrentDir
GOSUB DisplayCurrentPath
GOSUB DisplayDirectory
RETURN

' inits dir box without new dirspec.
LoadNewDir2:
Current.Dir! = 1!
Dir.Box.Line = 1
GOSUB ClearDirBox
GOSUB CheckPath
IF ValidPath = False THEN
    D$ = D1$
    X$ = X1$
    Xposition1 = 1
    Xposition2 = 0
END IF
GOSUB DisplayDirSpec
GOSUB DisplayCurrentDir
GOSUB DisplayCurrentPath2
GOSUB DisplayDirectory
RETURN

' inits file box.
LoadNewFiles:
GOSUB ClearFileBox
FileLoadMessage = "Reloading files.."
CALL LoadFileSpec(FileY$, F$)
GOSUB DisplayFileSpec
IF Num.Files > 0 THEN
    Current.File! = 1!
    Box.Line = 1
    GOSUB InitFiles
ELSE
    GOSUB ClearLongName2
END IF
RETURN

' display current filename entry area.
DisplayFileLine:
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
' check hilighting.
IF Xstart THEN
    Colorf2 Black, Plain
END IF
' display edit line and position cursor.
Locatef Xcoor + 1, Ycoor + 12, 1
Printf MID$(X$, Xposition1, 26)
Locatef Xcoor + 1, Ycoor + Xposition2 - Xposition1 + 13, 1
IF Insert THEN
    LOCATE , , 1, 8, 8 ' line cursor
ELSE
    LOCATE , , 1, 0, 8 ' block cursor
END IF
Colorf2 Plain, 0
RETURN

' display invalid disk message.
PromptMedia:
Z2$ = "Media not ready. Press <esc>:"
GOTO NextPrompt
PromptDisk:
Z2$ = "Disk not ready. Press <esc>:"
GOTO NextPrompt
' display invalid netpath message.
PromptNet:
Z2$ = "Net not ready. Press <esc>:"
NextPrompt:
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf Z2$
CALL PromptInkey
GOSUB DisplayFilename6
Colorf2 Plain, 0
RETURN

' init filename input loop.
InitFiles:
E$ = RTRIM$(Filenames(1!))
Colorf2 Black, Plain
Locatef Xcoor + 5, Ycoor + 3, 1
Printf LEFT$(E$, 12)
Colorf2 Plain, 0
GOSUB DisplayLongName2
RETURN

' filename input loop.
FileInputLoop:
I$ = Nul
DO
    IF Current.File! > 0 THEN
        GOSUB DisplayLongName2
        GOSUB ClearFilename
        GOSUB DisplayFilename1
    END IF
    DO
        ' get keystroke.
        _LIMIT 100
        I$ = INKEYx$
        IF LEN(I$) THEN
            IsMenu = False
            EXIT DO
        END IF

        ' call mouse subroutine.
        Z = MouseDriver

        ' check file entry area.
        IF IsMenu >= 4 THEN
            RETURN
        END IF

        ' check mouse wheel.
        IF MouseWheel = -1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(80) ' down
            ELSE
                I$ = CHR$(0) + CHR$(72) ' up
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
        IF MouseWheel = 1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(72) ' up
            ELSE
                I$ = CHR$(0) + CHR$(80) ' down
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF

        ' check left mouse button.
        IF MouseButton1 THEN
            GOSUB MouseButton1Z
            SELECT CASE IsMenu
                CASE 1 ' file box.
                    IF I$ = CHR$(13) THEN
                        EXIT DO
                    ELSE
                        GOSUB FileBoxKey
                    END IF
                CASE 2 ' dir box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DirBoxKey
                    END IF
                CASE 3 ' drive box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DriveBoxKey
                    END IF
            END SELECT
        ELSE
            ' check mouse position.
            IF MouseX OR MouseY THEN
                GOSUB MouseOverX1
            END IF
        END IF
    LOOP

    ' check function key.
    IF LEN(I$) = 2 THEN
        VarQ = ASC(RIGHT$(I$, 1))
        SELECT CASE VarQ
            CASE 59 TO 68, 104 TO 113, 133, 134
                RETURN
        END SELECT
        GOSUB GetAltKey
        IF VarZ THEN
            RETURN
        END IF
    END IF

    ' up returns shift-tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 72 THEN
            IF Current.File! = 1! THEN
                IF Box.Line = 1 THEN
                    I$ = CHR$(0) + CHR$(15)
                END IF
            END IF
        END IF
    END IF

    ' shift-tab returns to top.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 15 THEN
            RETURN
        END IF
    END IF

    ' right returns tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 77 THEN
            I$ = CHR$(9)
        END IF
    END IF

    ' Tab moves to directory loop.
    IF I$ = CHR$(9) THEN
        GOSUB DirInputLoop
        ' check function key.
        IF LEN(I$) = 2 THEN
            VarQ = ASC(RIGHT$(I$, 1))
            SELECT CASE VarQ
                CASE 59 TO 68, 104 TO 113, 133, 134
                    RETURN
            END SELECT
            GOSUB GetAltKey
            IF VarZ THEN
                RETURN
            END IF
        END IF
        IF IsMenu >= 4 THEN
            RETURN
        END IF
        IF I$ = CHR$(9) THEN
            GOSUB DisplayCurrentPath
            RETURN
        END IF
    END IF
    IF I$ = CHR$(13) THEN
        RETURN
    END IF
    IF I$ = CHR$(27) THEN
        RETURN
    END IF
    IF LEN(I$) = 1 THEN
        SELECT CASE I$
            CASE "A" TO "Z", "a" TO "z", "0" TO "9"
                ' scan the first character of the filenames.
                J$ = I$
                DO
                    IF Current.File! + 1! <= Num.Files THEN
                        IF Box.Line < 10 THEN
                            GOSUB ClearFilename
                            Box.Line = Box.Line + 1
                            Current.File! = Current.File! + 1!
                            GOSUB DisplayFilename1
                            GOSUB DisplayLongName2
                        ELSE
                            Current.File! = Current.File! + 1!
                            GOSUB DisplayFilename2
                            GOSUB DisplayFileBox1
                            GOSUB DisplayFilename3
                            GOSUB DisplayLongName2
                        END IF
                        E$ = RTRIM$(Filenames(Current.File!))
                        IF LEFT$(E$, 1) = J$ THEN
                            EXIT DO
                        END IF
                        _LIMIT 100
                        I$ = INKEYx$
                        IF I$ = CHR$(27) THEN
                            EXIT DO
                        END IF
                    ELSE
                        EXIT DO
                    END IF
                LOOP
        END SELECT
        I$ = Nul
    END IF
    IF LEN(I$) = 2 THEN
        GOSUB FileBoxKey
        IF Num.Files = 0 THEN
            I$ = CHR$(0) + CHR$(15)
            RETURN
        END IF
    END IF
LOOP
RETURN

' process key in file box.
FileBoxKey:
V = ASC(RIGHT$(I$, 1))
SELECT CASE V
    CASE 82 ' insert
        Locatef Xcoor + 1, Ycoor + 2, 1
        Printf SPACE$(38)
        Locatef Xcoor + 1, Ycoor + 2, 1
        Colorf2 White, 0

        ' get new file.
        Printf "Filename?"
        ZQ$ = LineInput$(Xcoor + 1, Ycoor + 12, 28)

        ' parse new file.
        IF LEN(ZQ$) THEN

            ' create filename.
            G$ = MakeFilename$(N$, C$, D$, ZQ$)
            G$ = G$ + CHR$(0)
            hfind = CreateFile(G$, &H180, 0, 0, 1, 0, 0)
            IF hfind <> INVALID_HANDLE_VALUE THEN
                x = CloseHandle(hfind)
                ' init file box.
                GOSUB DisplayFilename6
                GOSUB ClearFileBox
                FileLoadMessage = "Updating files.."
                CALL LoadFileSpec(FileY$, F$)
                GOSUB DisplayFileSpec
                GOSUB DisplayCurrentPath
                ' display first file.
                GOSUB ClearLongName2
                IF Num.Files > 0 THEN
                    Current.File! = 1!
                    Box.Line = 1
                    GOSUB InitFiles
                END IF
            ELSE
                Locatef Xcoor + 1, Ycoor + 2, 1
                Printf SPACE$(38)
                Locatef Xcoor + 1, Ycoor + 2, 1
                Colorf2 White, 0
                IF GetLastError& = &H50 THEN ' check already exists
                    Printf "Filename already exists. Press <esc>:"
                ELSE
                    Printf "Error 0x" + HEX$(GetLastError) + " Press <esc>:"
                END IF
                CALL PromptInkey
            END IF
        END IF
        GOSUB DisplayFilename6
        GOSUB DisplayFileLine
        GOSUB DisplayLongName2
    CASE 83 ' delete
        ' check file.
        IF Num.Files > 0 THEN
            E$ = RTRIM$(Filenames(Current.File!))
            ErrorRespond$ = "Y"
            IF AutoConfirm = 0 THEN
                Locatef Xcoor + 1, Ycoor + 2, 1
                Printf SPACE$(38)
                Locatef Xcoor + 1, Ycoor + 2, 1
                Colorf2 White, 0
                Printf "Delete current file(y/n)?"
                ErrorRespond$ = Nul
                DO
                    _LIMIT 100
                    ErrorRespond$ = INKEYx$
                    IF LEN(ErrorRespond$) THEN
                        IF UCASE$(ErrorRespond$) = "Y" THEN
                            EXIT DO
                        END IF
                        IF UCASE$(ErrorRespond$) = "N" THEN
                            EXIT DO
                        END IF
                    END IF
                LOOP
            END IF

            ' parse key.
            IF UCASE$(ErrorRespond$) = "Y" THEN

                ' delete filename.
                G$ = MakeFilename$(N$, C$, D$, E$)
                G$ = G$ + CHR$(0)

                ' change filename attribute
                AttrX& = GetFileAttributes(G$)
                AttrX& = AttrX& AND NOT &H1 ' remove read-only bit
                x = SetFileAttributes&(G$, AttrX&)
                x = DeleteFileA(G$)

                ' check carry flag error
                IF x = 0 THEN
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Error 0x" + HEX$(GetLastError) + " Press <esc>:"
                    CALL PromptInkey
                ELSE
                    ' init file box.
                    GOSUB DisplayFilename6
                    GOSUB ClearFileBox
                    FileLoadMessage = "Updating files.."
                    CALL LoadFileSpec(FileY$, F$)
                    GOSUB DisplayFileSpec
                    GOSUB DisplayCurrentPath
                    ' display first file.
                    GOSUB ClearLongName2
                    IF Num.Files > 0 THEN
                        Current.File! = 1!
                        Box.Line = 1
                        GOSUB InitFiles
                    END IF
                END IF
            END IF
            GOSUB DisplayFilename6
            GOSUB DisplayFileLine
            GOSUB DisplayLongName2
        END IF
    CASE 72, 141, 152 ' up/ctrl-up/alt-up
        IF Current.File! - 1 > 0 THEN
            IF Box.Line > 1 THEN
                GOSUB ClearFilename
                Box.Line = Box.Line - 1
                Current.File! = Current.File! - 1!
                GOSUB DisplayFilename1
                GOSUB DisplayLongName2
            ELSE
                Current.File! = Current.File! - 1!
                GOSUB DisplayFilename2
                GOSUB DisplayFileBox2
                GOSUB DisplayFilename4
                GOSUB DisplayLongName2
            END IF
        END IF
    CASE 80, 145, 160 ' down/ctrl-down/alt-down
        IF Current.File! + 1 <= Num.Files THEN
            IF Box.Line < 10 THEN
                GOSUB ClearFilename
                Box.Line = Box.Line + 1
                Current.File! = Current.File! + 1!
                GOSUB DisplayFilename1
                GOSUB DisplayLongName2
            ELSE
                Current.File! = Current.File! + 1!
                GOSUB DisplayFilename2
                GOSUB DisplayFileBox1
                GOSUB DisplayFilename3
                GOSUB DisplayLongName2
            END IF
        END IF
    CASE 73, 132 ' pageup/ctrl-pageup
        IF Current.File! - 10 > 0 THEN
            Box.Line = 1
            Current.File! = Current.File! - 10!
        ELSE
            Box.Line = 1
            Current.File! = 1!
        END IF
        GOSUB DisplayFilename2
        GOSUB DisplayFileBox2
        GOSUB DisplayFilename4
        GOSUB DisplayLongName2
    CASE 81, 118 ' pagedown/ctrl-pagedown
        IF Current.File! + 10 <= Num.Files THEN
            Box.Line = 10
            Current.File! = Current.File! + 10!
            GOSUB DisplayFilename2
            GOSUB DisplayFileBox1
            GOSUB DisplayFilename3
            GOSUB DisplayLongName2
        ELSE
            IF Current.File! + 1 <= Num.Files THEN
                DO UNTIL Current.File! = Num.Files
                    IF Box.Line < 10 THEN
                        GOSUB ClearFilename
                        Box.Line = Box.Line + 1
                        Current.File! = Current.File! + 1!
                        GOSUB DisplayFilename1
                        GOSUB DisplayLongName2
                    ELSE
                        Current.File! = Current.File! + 1!
                        GOSUB DisplayFilename2
                        GOSUB DisplayFileBox1
                        GOSUB DisplayFilename3
                        GOSUB DisplayLongName2
                    END IF
                LOOP
            END IF
        END IF
    CASE 71, 119 ' home/ctrl-home
        IF Current.File! > 1 THEN
            Box.Line = 1
            Current.File! = 1!
        END IF
        GOSUB DisplayFilename2
        GOSUB DisplayFileBox2
        GOSUB DisplayFilename4
        GOSUB DisplayLongName2
    CASE 79, 117 ' end/ctrl-end
        IF Current.File! + 1 <= Num.Files THEN
            DO UNTIL Current.File! = Num.Files
                IF Box.Line < 10 THEN
                    GOSUB ClearFilename
                    Box.Line = Box.Line + 1
                    Current.File! = Current.File! + 1!
                    GOSUB DisplayFilename1
                    GOSUB DisplayLongName2
                ELSE
                    Current.File! = Current.File! + 1!
                    GOSUB DisplayFilename2
                    GOSUB DisplayFileBox1
                    GOSUB DisplayFilename3
                    GOSUB DisplayLongName2
                END IF
                _LIMIT 100
                I$ = INKEYx$
                IF I$ = CHR$(27) THEN
                    EXIT DO
                END IF
            LOOP
        END IF
END SELECT
RETURN

' clear filename.
ClearFilename:
Colorf2 Plain, 0
Locatef Box.Line + Xcoor + 4, Ycoor + 3, 1
Printf LEFT$(RTRIM$(Filenames(Current.File!)), 12)
RETURN

' display filename.
DisplayFilename1:
E$ = RTRIM$(Filenames(Current.File!))
Colorf2 Plain, 0
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
GOSUB TruncateFilename
Printf A$
Colorf2 Black, Plain
Locatef Box.Line + Xcoor + 4, Ycoor + 3, 1
Printf LEFT$(E$, 12)
V9 = LEN(RTRIM$(LEFT$(E$, 12)))
Locatef Box.Line + Xcoor + 4, Ycoor + V9 + 3, 1
Colorf2 Plain, 0
RETURN

' display filename.
DisplayFilename2:
E$ = RTRIM$(Filenames(Current.File!))
Colorf2 Plain, 0
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
GOSUB TruncateFilename
Printf A$
GOSUB ClearFileBox
Colorf2 Plain, 0
RETURN

' display filename.
DisplayFilename3:
E$ = RTRIM$(Filenames(Current.File!))
Colorf2 Black, Plain
Locatef Xcoor + 14, Ycoor + 3, 1
Printf LEFT$(E$, 12)
V9 = LEN(RTRIM$(LEFT$(E$, 12)))
Locatef Xcoor + 14, Ycoor + V9 + 3, 1
RETURN

' display filename.
DisplayFilename4:
E$ = RTRIM$(Filenames(Current.File!))
Colorf2 Black, Plain
Locatef Xcoor + 5, Ycoor + 3, 1
Printf LEFT$(E$, 12)
V9 = LEN(RTRIM$(LEFT$(E$, 12)))
Locatef Xcoor + 5, Ycoor + V9 + 3, 1
RETURN

' display filename.
DisplayFilename5:
E$ = RTRIM$(Filenames(Current.File!))
Colorf2 Black, Plain
Locatef Box.Line + Xcoor + 4, Ycoor + 3, 1
Printf LEFT$(E$, 12)
Colorf2 Plain, 0
RETURN

' display filename line.
DisplayFilename6:
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Filename: "
Colorf2 Plain, 0
RETURN

' update filename box.
DisplayFileBox1:
FOR V = 1 TO 9
    Locatef Xcoor - V + 14, Ycoor + 3, 1
    Printf LEFT$(RTRIM$(Filenames(Current.File! - V)), 12)
NEXT
RETURN

' update filename box.
DisplayFileBox2:
FOR V = 1 TO 9
    IF Current.File! + V <= Num.Files THEN
        Locatef Xcoor + V + 5, Ycoor + 3, 1
        Printf LEFT$(RTRIM$(Filenames(Current.File! + V)), 12)
    END IF
NEXT
RETURN

' display init directory.
DisplayInitDir:
Q$ = RTRIM$(Directories(Current.Dir!))
Locatef Dir.Box.Line + Xcoor + 4, Ycoor + 19, 1
Colorf2 Black, Plain
Printf LEFT$(Q$, 12)
Colorf2 Plain, 0
RETURN

' directory input loop.
DirInputLoop:
GOSUB DisplayInitDir
GOSUB DisplayLongName0
I$ = Nul
DO
    DO
        ' get keystroke.
        _LIMIT 100
        I$ = INKEYx$
        IF LEN(I$) THEN
            IsMenu = False
            EXIT DO
        END IF

        ' call mouse subroutine.
        Z = MouseDriver

        ' check file entry area.
        IF IsMenu >= 4 THEN
            RETURN
        END IF

        ' check mouse wheel.
        IF MouseWheel = -1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(80) ' down
            ELSE
                I$ = CHR$(0) + CHR$(72) ' up
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
        IF MouseWheel = 1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(72) ' up
            ELSE
                I$ = CHR$(0) + CHR$(80) ' down
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF

        ' check left mouse button.
        IF MouseButton1 THEN
            GOSUB MouseButton1Z
            SELECT CASE IsMenu
                CASE 1 ' file box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB FileBoxKey
                    END IF
                CASE 2 ' dir box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DirBoxKey
                    END IF
                CASE 3 ' drive box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DriveBoxKey
                    END IF
            END SELECT
        ELSE
            ' check mouse position.
            IF MouseX OR MouseY THEN
                GOSUB MouseOverX1
            END IF
        END IF
    LOOP

    ' check function key.
    IF LEN(I$) = 2 THEN
        VarQ = ASC(RIGHT$(I$, 1))
        SELECT CASE VarQ
            CASE 59 TO 68, 104 TO 113, 133, 134
                RETURN
        END SELECT
        GOSUB GetAltKey
        IF VarZ THEN
            RETURN
        END IF
    END IF

    ' up returns shift-tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 72 THEN
            IF Current.Dir! = 1! THEN
                IF Dir.Box.Line = 1 THEN
                    I$ = CHR$(0) + CHR$(15)
                END IF
            END IF
        END IF
    END IF

    ' left returns shift-tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 75 THEN
            I$ = CHR$(0) + CHR$(15)
        END IF
    END IF

    ' shift-tab returns to file loop.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 15 THEN
            ' display current filename.
            GOSUB ClearLongName2
            IF Num.Files > 0 THEN
                GOSUB DisplayLongName2
                GOSUB ClearFilename
                GOSUB DisplayFilename1
            END IF
            ' display current directory.
            GOSUB ClearDirectory
            GOSUB DisplayCurrentPath
            RETURN
        END IF
    END IF

    ' right returns tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 77 THEN
            I$ = CHR$(9)
        END IF
    END IF

    ' Tab moves to drive input loop.
    IF I$ = CHR$(9) THEN
        GOSUB DriveInputLoop
        ' check function key.
        IF LEN(I$) = 2 THEN
            VarQ = ASC(RIGHT$(I$, 1))
            SELECT CASE VarQ
                CASE 59 TO 68, 104 TO 113, 133, 134
                    RETURN
            END SELECT
            GOSUB GetAltKey
            IF VarZ THEN
                RETURN
            END IF
        END IF
        IF IsMenu >= 3 THEN
            GOSUB DisplayCurrentPath
            IF Num.Files > 0 THEN
                GOSUB DisplayLongName2
            END IF
            RETURN
        END IF
        IF I$ = CHR$(9) THEN
            GOSUB DisplayCurrentPath
            IF Num.Files > 0 THEN
                GOSUB DisplayLongName2
            END IF
            RETURN
        END IF
        Q$ = RTRIM$(Directories(Current.Dir!))
        GOSUB DisplayLongName1
    END IF
    IF I$ = CHR$(13) THEN
        IsMenu = 2
        RETURN
    END IF
    IF I$ = CHR$(27) THEN
        RETURN
    END IF
    IF LEN(I$) = 1 THEN
        SELECT CASE I$
            CASE "A" TO "Z", "a" TO "z", "0" TO "9"
                ' scan the first character of the directories.
                J$ = I$
                DO
                    IF Current.Dir! + 1! <= Num.Dirs THEN
                        IF Dir.Box.Line < 10 THEN
                            GOSUB ClearDirectory
                            Dir.Box.Line = Dir.Box.Line + 1
                            Current.Dir! = Current.Dir! + 1!
                            GOSUB DisplayDirectory
                        ELSE
                            GOSUB ClearDirBox
                            Current.Dir! = Current.Dir! + 1!
                            GOSUB DisplayDirBox1
                            GOSUB DisplayDirectory2
                        END IF
                        Q$ = RTRIM$(Directories(Current.Dir!))
                        GOSUB DisplayLongName1
                        IF LEFT$(Q$, 1) = J$ THEN
                            EXIT DO
                        END IF
                        _LIMIT 100
                        I$ = INKEYx$
                        IF I$ = CHR$(27) THEN
                            EXIT DO
                        END IF
                    ELSE
                        EXIT DO
                    END IF
                LOOP
        END SELECT
        I$ = Nul
    END IF
    IF LEN(I$) = 2 THEN
        GOSUB DirBoxKey
    END IF
LOOP
RETURN

' clear directory.
ClearDirectory:
Locatef Dir.Box.Line + Xcoor + 4, Ycoor + 19, 1
Colorf2 Plain, 0
Printf LEFT$(RTRIM$(Directories(Current.Dir!)), 12)
RETURN

' display directory.
DisplayDirectory:
Locatef Dir.Box.Line + Xcoor + 4, Ycoor + 19, 1
Colorf2 Black, Plain
Printf LEFT$(RTRIM$(Directories(Current.Dir!)), 12)
Colorf2 Plain, 0
RETURN

' display directory.
DisplayDirectory2:
Colorf2 Black, Plain
Locatef Xcoor + 14, Ycoor + 19, 1
Printf LEFT$(RTRIM$(Directories(Current.Dir!)), 12)
Colorf2 Plain, 0
RETURN

' display directory.
DisplayDirectory3:
Colorf2 Black, Plain
Locatef Xcoor + 5, Ycoor + 19, 1
Printf LEFT$(RTRIM$(Directories(Current.Dir!)), 12)
Colorf2 Plain, 0
RETURN

' update dir box.
DisplayDirBox1:
Colorf2 Plain, 0
FOR V = 1 TO 9
    Locatef Xcoor - V + 14, Ycoor + 19, 1
    Printf LEFT$(RTRIM$(Directories(Current.Dir! - V)), 12)
NEXT
RETURN

' update dir box.
DisplayDirBox2:
Colorf2 Plain, 0
FOR V = 1 TO 9
    IF Current.Dir! + V <= Num.Dirs THEN
        Locatef Xcoor + V + 5, Ycoor + 19, 1
        Printf LEFT$(RTRIM$(Directories(Current.Dir! + V)), 12)
    END IF
NEXT
RETURN

' process dir box key.
DirBoxKey:
V = ASC(RIGHT$(I$, 1))
SELECT CASE V
    CASE 82 ' insert
        Locatef Xcoor + 1, Ycoor + 2, 1
        Printf SPACE$(38)
        Locatef Xcoor + 1, Ycoor + 2, 1
        Colorf2 White, 0

        ' get new path.
        Printf "Dir path?"
        ZQ$ = LineInput$(Xcoor + 1, Ycoor + 12, 28)

        ' parse new path.
        IF LEN(ZQ$) THEN

            ' create directory.
            G$ = MakeFilename$(N$, C$, D$, ZQ$)
            G$ = G$ + CHR$(0)
            x = CreateDirectoryA(G$, 0)

            ' check carry flag error
            IF x = 0 THEN
                Locatef Xcoor + 1, Ycoor + 2, 1
                Printf SPACE$(38)
                Locatef Xcoor + 1, Ycoor + 2, 1
                Colorf2 White, 0
                IF GetLastError& = &HB7 THEN ' check already exists
                    Printf "Directory already exists. Press <esc>:"
                ELSE
                    Printf "Error 0x" + HEX$(GetLastError) + " Press <esc>:"
                END IF
                CALL PromptInkey
            ELSE
                ' redisplay dir box.
                GOSUB DisplayFilename6
                GOSUB ClearDirBox
                DirLoadMessage = "Loading dirs.."
                CALL LoadDirSpec(DirY$, N$, C$, D$)
                GOSUB DisplayDirSpec
                GOSUB DisplayCurrentDir
                GOSUB DisplayCurrentPath
                ' display first directory.
                Current.Dir! = 1!
                Dir.Box.Line = 1
                GOSUB DisplayInitDir
            END IF
        END IF
        GOSUB DisplayFilename6
        GOSUB DisplayFileLine
        GOSUB DisplayInitDir
        GOSUB DisplayLongName1
    CASE 83 ' delete
        ' check path.
        Q$ = RTRIM$(Directories(Current.Dir!))
        IF Q$ = "." OR Q$ = ".." THEN
            Locatef Xcoor + 1, Ycoor + 2, 1
            Printf SPACE$(38)
            Locatef Xcoor + 1, Ycoor + 2, 1
            Colorf2 White, 0
            Printf "Can't delete directory. Press <esc>:"
            CALL PromptInkey
        ELSE
            ErrorRespond$ = "Y"
            IF AutoConfirm = 0 THEN
                Locatef Xcoor + 1, Ycoor + 2, 1
                Printf SPACE$(38)
                Locatef Xcoor + 1, Ycoor + 2, 1
                Colorf2 White, 0
                Printf "Delete current path(y/n)?"
                ErrorRespond$ = Nul
                DO
                    _LIMIT 100
                    ErrorRespond$ = INKEYx$
                    IF LEN(ErrorRespond$) THEN
                        IF UCASE$(ErrorRespond$) = "Y" THEN
                            EXIT DO
                        END IF
                        IF UCASE$(ErrorRespond$) = "N" THEN
                            EXIT DO
                        END IF
                    END IF
                LOOP
            END IF

            ' parse key.
            IF UCASE$(ErrorRespond$) = "Y" THEN

                ' delete directory.
                G$ = MakeFilename$(N$, C$, D$, Q$) + "\"
                CALL Silentdel_Directories(G$)

                ' check carry flag error
                IF _DIREXISTS(G$) THEN
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Printf SPACE$(38)
                    Locatef Xcoor + 1, Ycoor + 2, 1
                    Colorf2 White, 0
                    Printf "Error deleting directory. Press <esc>:"
                    CALL PromptInkey
                ELSE
                    ' redisplay dir box.
                    GOSUB DisplayFilename6
                    GOSUB ClearDirBox
                    DirLoadMessage = "Loading dirs.."
                    CALL LoadDirSpec(DirY$, N$, C$, D$)
                    GOSUB DisplayDirSpec
                    GOSUB DisplayCurrentDir
                    GOSUB DisplayCurrentPath
                    ' display first directory.
                    Current.Dir! = 1!
                    Dir.Box.Line = 1
                    GOSUB DisplayInitDir
                END IF
            END IF
        END IF
        GOSUB DisplayFilename6
        GOSUB DisplayFileLine
        GOSUB DisplayInitDir
        GOSUB DisplayLongName1
    CASE 72, 141, 152 ' up/ctrl-up/alt-up
        IF Current.Dir! - 1 > 0 THEN
            IF Dir.Box.Line > 1 THEN
                GOSUB ClearDirectory
                Dir.Box.Line = Dir.Box.Line - 1
                Current.Dir! = Current.Dir! - 1!
                GOSUB DisplayDirectory
            ELSE
                GOSUB ClearDirBox
                Current.Dir! = Current.Dir! - 1!
                GOSUB DisplayDirBox2
                GOSUB DisplayDirectory3
            END IF
            Q$ = RTRIM$(Directories(Current.Dir!))
            GOSUB DisplayLongName1
        END IF
    CASE 80, 145, 160 ' down/ctrl-down/alt-down
        IF Current.Dir! + 1 <= Num.Dirs THEN
            IF Dir.Box.Line < 10 THEN
                GOSUB ClearDirectory
                Dir.Box.Line = Dir.Box.Line + 1
                Current.Dir! = Current.Dir! + 1!
                GOSUB DisplayDirectory
            ELSE
                GOSUB ClearDirBox
                Current.Dir! = Current.Dir! + 1!
                GOSUB DisplayDirBox1
                GOSUB DisplayDirectory2
            END IF
            Q$ = RTRIM$(Directories(Current.Dir!))
            GOSUB DisplayLongName1
        END IF
    CASE 73, 132 ' pageup/ctrl-pageup
        IF Current.Dir! - 10 > 0 THEN
            Dir.Box.Line = 1
            Current.Dir! = Current.Dir! - 10!
        ELSE
            Dir.Box.Line = 1
            Current.Dir! = 1!
        END IF
        GOSUB ClearDirBox
        GOSUB DisplayDirBox2
        GOSUB DisplayDirectory3
        Q$ = RTRIM$(Directories(Current.Dir!))
        GOSUB DisplayLongName1
    CASE 81, 118 ' pagedown/ctrl-pagedown
        IF Current.Dir! + 10 <= Num.Dirs THEN
            GOSUB ClearDirBox
            Dir.Box.Line = 10
            Current.Dir! = Current.Dir! + 10!
            GOSUB DisplayDirBox1
            GOSUB DisplayDirectory2
            Q$ = RTRIM$(Directories(Current.Dir!))
            GOSUB DisplayLongName1
        ELSE
            IF Current.Dir! + 1 <= Num.Dirs THEN
                DO UNTIL Current.Dir! = Num.Dirs
                    IF Dir.Box.Line < 10 THEN
                        GOSUB ClearDirectory
                        Dir.Box.Line = Dir.Box.Line + 1
                        Current.Dir! = Current.Dir! + 1!
                        GOSUB DisplayDirectory
                    ELSE
                        GOSUB ClearDirBox
                        Current.Dir! = Current.Dir! + 1!
                        GOSUB DisplayDirBox1
                        GOSUB DisplayDirectory2
                    END IF
                LOOP
                Q$ = RTRIM$(Directories(Current.Dir!))
                GOSUB DisplayLongName1
            END IF
        END IF
    CASE 71, 119 ' home/ctrl-home
        IF Current.Dir! > 1 THEN
            Dir.Box.Line = 1
            Current.Dir! = 1!
        END IF
        GOSUB ClearDirBox
        GOSUB DisplayDirBox2
        GOSUB DisplayDirectory3
        Q$ = RTRIM$(Directories(Current.Dir!))
        GOSUB DisplayLongName1
    CASE 79, 117 ' end/ctrl-end
        IF Current.Dir! + 1 <= Num.Dirs THEN
            DO UNTIL Current.Dir! = Num.Dirs
                IF Dir.Box.Line < 10 THEN
                    GOSUB ClearDirectory
                    Dir.Box.Line = Dir.Box.Line + 1
                    Current.Dir! = Current.Dir! + 1!
                    GOSUB DisplayDirectory
                ELSE
                    GOSUB ClearDirBox
                    Current.Dir! = Current.Dir! + 1!
                    GOSUB DisplayDirBox1
                    GOSUB DisplayDirectory2
                END IF
                _LIMIT 100
                I$ = INKEYx$
                IF I$ = CHR$(27) THEN
                    EXIT DO
                END IF
            LOOP
            Q$ = RTRIM$(Directories(Current.Dir!))
            GOSUB DisplayLongName1
        END IF
END SELECT
RETURN

' drive input loop.
DriveInputLoop:
I$ = Nul
DO
    GOSUB DisplayDrive
    DO
        ' get keystroke.
        _LIMIT 100
        I$ = INKEYx$
        IF LEN(I$) THEN
            IsMenu = False
            EXIT DO
        END IF

        ' call mouse subroutine.
        Z = MouseDriver

        ' check file entry area.
        IF IsMenu >= 4 THEN
            RETURN
        END IF

        ' check mouse wheel.
        IF MouseWheel = -1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(80) ' down
            ELSE
                I$ = CHR$(0) + CHR$(72) ' up
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
        IF MouseWheel = 1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(72) ' up
            ELSE
                I$ = CHR$(0) + CHR$(80) ' down
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF

        ' check left mouse button.
        IF MouseButton1 THEN
            GOSUB MouseButton1Z
            SELECT CASE IsMenu
                CASE 1 ' file box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB FileBoxKey
                    END IF
                CASE 2 ' dir box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DirBoxKey
                    END IF
                CASE 3 ' drive box.
                    IF I$ = CHR$(13) THEN
                        RETURN
                    ELSE
                        GOSUB DriveBoxKey
                    END IF
            END SELECT
        ELSE
            ' check mouse position.
            IF MouseX OR MouseY THEN
                GOSUB MouseOverX1
                GOSUB DisplayDrive
            END IF
        END IF
    LOOP

    ' check function key.
    IF LEN(I$) = 2 THEN
        VarQ = ASC(RIGHT$(I$, 1))
        SELECT CASE VarQ
            CASE 59 TO 68, 104 TO 113, 133, 134
                RETURN
        END SELECT
        GOSUB GetAltKey
        IF VarZ THEN
            RETURN
        END IF
    END IF

    ' left returns shift-tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 75 THEN
            I$ = CHR$(0) + CHR$(15)
        END IF
    END IF

    ' shift-tab returns to directory loop.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 15 THEN
            I$ = Nul
            GOSUB DisplayDirectory
            RETURN
        END IF
    END IF

    ' up returns tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 72 THEN
            IF Current.Drive2 = 1 THEN
                IF Drive.Box.Line = 1 THEN
                    I$ = CHR$(9)
                END IF
            END IF
        END IF
    END IF

    ' right returns tab.
    IF LEN(I$) = 2 THEN
        IF ASC(RIGHT$(I$, 1)) = 77 THEN
            I$ = CHR$(9)
        END IF
    END IF

    ' Tab returns to top loop.
    IF I$ = CHR$(9) THEN
        RETURN
    END IF

    ' select a drive.
    IF I$ = CHR$(13) THEN
        IsMenu = 3
        RETURN
    END IF

    ' exit from drive box.
    IF I$ = CHR$(27) THEN
        RETURN
    END IF

    ' process drive box keystroke.
    IF LEN(I$) THEN
        GOSUB DriveBoxKey
    END IF
LOOP
RETURN

' process drive letter.
GetAltKey:
' get alt keycode.
Var = ASC(RIGHT$(I$, 1))
VarZ = False
' compare all Alt-n keys.
FOR Var2 = 1 TO 26 ' A - Z
    ' check key.
    IF Keys(Var2) = Var THEN
        IF Var2 <= Last.Drive THEN
            VarZ = True
        END IF
        RETURN
    END IF
NEXT
RETURN

' process drive box key.
DriveBoxKey:
SELECT CASE LEN(I$)
    CASE 1
        ' select a drive letter.
        SELECT CASE I$
            CASE "A" TO "Z"
                V0 = ASC(I$) - 64
                GOSUB MoveSelectDrive
            CASE "a" TO "z"
                V0 = ASC(I$) - 96
                GOSUB MoveSelectDrive
            CASE "<" ' up
                GOSUB MoveDriveUp
            CASE ">" ' down
                V6 = False
                GOSUB MoveDriveDown
            CASE "|" ' drive A:
                GOSUB MoveFirstDrive
            CASE "#" ' drive Z:
                GOSUB MoveLastDrive
        END SELECT
    CASE 2
        V = ASC(RIGHT$(I$, 1))
        SELECT CASE V
            CASE 71, 119 ' home/ctrl-home
                GOSUB MoveFirstDrive
            CASE 79, 117 ' end/ctrl-end
                GOSUB MoveLastDrive
            CASE 72, 141, 152 ' up/ctrl-up/alt-up
                GOSUB MoveDriveUp
            CASE 80, 145, 160 ' down/ctrl-down/alt-down
                V6 = False
                GOSUB MoveDriveDown
            CASE 73, 132 ' pageup/ctrl-pageup
                FOR V7 = 1 TO 5 ' scroll up one-half page
                    GOSUB MoveDriveUp
                NEXT
            CASE 81, 118 ' pagedown/ctrl-pagedown
                FOR V7 = 1 TO 5 ' scroll down one-half page
                    V6 = False
                    GOSUB MoveDriveDown
                NEXT
        END SELECT
END SELECT
RETURN

' move to selected drive.
MoveSelectDrive:
IF V0 > Last.Drive THEN
    V0 = Last.Drive
END IF
GOSUB MoveFirstDrive ' assume A:
' scroll down to drive letter.
FOR V7 = 2 TO V0 ' B: to Z:
    V6 = False
    GOSUB MoveDriveDown
NEXT
RETURN

' move to drive A:
MoveFirstDrive:
DO
    IF Current.Drive2 > 1 THEN
        GOSUB MoveDriveUp
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' move to last drive.
MoveLastDrive:
DO
    IF Current.Drive2 + 1 <= Last.Drive THEN
        V6 = False
        GOSUB MoveDriveDown
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' move up one drive letter.
MoveDriveUp:
IF Current.Drive2 - 1 > 0 THEN
    IF Drive.Box.Line > 1 THEN
        GOSUB ClearDrive
        Drive.Box.Line = Drive.Box.Line - 1
        Current.Drive2 = Current.Drive2 - 1
        GOSUB DisplayDrive
    ELSE
        Current.Drive2 = Current.Drive2 - 1
        GOSUB ClearDriveBox
        Colorf2 Plain, 0
        FOR VX = 1 TO 9
            Locatef Xcoor + VX + 5, Ycoor + 35, 0
            Printf "[" + CHR$(Current.Drive2 + VX + 64) + "]"
        NEXT
        Drive.Box.Line = 1
        GOSUB DisplayDrive
    END IF
END IF
RETURN

' move down one drive letter with drive info.
MoveDriveDown:
IF Current.Drive2 + 1 <= Last.Drive THEN
    IF Drive.Box.Line < 10 THEN
        GOSUB ClearDrive
        Drive.Box.Line = Drive.Box.Line + 1
        Current.Drive2 = Current.Drive2 + 1
        IF V6 = False THEN
            GOSUB DisplayDrive
        ELSE
            GOSUB DisplayDrive2
        END IF
    ELSE
        Current.Drive2 = Current.Drive2 + 1
        GOSUB ClearDriveBox
        Colorf2 Plain, 0
        FOR VX = 1 TO 9
            Locatef Xcoor - VX + 14, Ycoor + 35, 1
            Printf "[" + CHR$(Current.Drive2 - VX + 64) + "]"
        NEXT
        Drive.Box.Line = 10
        IF V6 = False THEN
            GOSUB DisplayDrive
        ELSE
            GOSUB DisplayDrive2
        END IF
    END IF
END IF
RETURN

' display path with dir info.
DisplayCurrentPath:
GOSUB DisplayCurrentPath3
GOSUB DisplayLongName0
RETURN

' display path w/o dir info.
DisplayCurrentPath2:
GOSUB DisplayCurrentPath3
GOSUB DisplayLongName3
RETURN

' format pathname for display.
DisplayCurrentPath3:
Q$ = D$
FOR Var = LEN(Q$) TO 1 STEP -1
    IF MID$(Q$, Var, 1) = "\" THEN
        Q$ = MID$(Q$, Var + 1)
        EXIT FOR
    END IF
NEXT
IF Q$ = Nul THEN
    Q$ = "\"
END IF
RETURN

' display long directory w/o dir info.
DisplayLongName0:
Colorf2 Green, BackGround
IF Xcoor + 17 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 17, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 17, Ycoor, 0
    IF Q$ = "." OR Q$ = ".." THEN
        G$ = Q$
    ELSE
        G$ = MakeFilename2$(N$, C$, D$)
        G$ = UnAmbiguateFile$(G$)
    END IF
    P$ = "Path: " + G$
    IF LEN(P$) > Length THEN
        P$ = LEFT$(P$, Length - 3) + "..."
    END IF
    Printf P$
END IF
Colorf2 Plain, 0
RETURN

' display long directory with dir info.
DisplayLongName1:
Colorf2 Green, BackGround
IF Xcoor + 17 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 17, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 17, Ycoor, 0
    IF Q$ = "." OR Q$ = ".." THEN
        G$ = Q$
    ELSE
        G$ = MakeFilename$(N$, C$, D$, Q$)
        G$ = UnAmbiguateFile$(G$)
    END IF
    P$ = "Path: " + G$
    IF LEN(P$) > Length THEN
        P$ = LEFT$(P$, Length - 3) + "..."
    END IF
    Printf P$
END IF
Colorf2 Plain, 0
GOSUB DisplayDirInfo
V = LEN(RTRIM$(Directories(Current.Dir!)))
IF V > 12 THEN
    V = 12
END IF
Locatef Dir.Box.Line + Xcoor + 4, Ycoor + V + 19, 1
RETURN

' display long filename with file info.
DisplayLongName2:
Colorf2 Green, BackGround
IF Xcoor + 18 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 18, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 18, Ycoor, 0
    G$ = MakeFilename$(N$, C$, D$, E$)
    G$ = UnAmbiguateFile$(G$)
    P$ = "File: " + G$
    IF LEN(P$) > Length THEN
        P$ = LEFT$(P$, Length - 3) + "..."
    END IF
    Printf P$
END IF
Colorf2 Plain, 0
GOSUB DisplayFileInfo
V = LEN(RTRIM$(Filenames(Current.File!)))
IF V > 12 THEN
    V = 12
END IF
Locatef Box.Line + Xcoor + 4, Ycoor + V + 3, 1
RETURN

' display long directory w/o dir info.
DisplayLongName3:
Colorf2 Green, BackGround
IF Xcoor + 17 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 17, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 17, Ycoor, 0
    P$ = "Path: " + Q$
    IF LEN(P$) > Length THEN
        P$ = LEFT$(P$, Length - 3) + "..."
    END IF
    Printf P$
END IF
Colorf2 Plain, 0
RETURN

' display filename information.
DisplayFileInfo:
Colorf2 Green, BackGround
IF Xcoor + 19 <= 24 THEN
    GOSUB GetFileInfo
    Var# = FileSize
    P$ = "Info: " + FormatX$(Var#, -1)
    SELECT CASE QuietSwitch
        CASE -1, 1, 3, 5
            ' nul
        CASE ELSE
            IF LEN(FileDateTime) THEN
                P$ = P$ + FileDateTime
            END IF
    END SELECT
    SELECT CASE QuietSwitch
        CASE -1, 2, 3, 6
            ' nul
        CASE ELSE
            GOSUB GetAttribute
    END SELECT
    GOSUB DisplayInfo
END IF
Colorf2 Plain, 0
RETURN

' display directory information.
DisplayDirInfo:
Colorf2 Green, BackGround
IF Xcoor + 19 <= 24 THEN
    GOSUB GetDirInfo
    P$ = "Info: <dir> "
    SELECT CASE QuietSwitch
        CASE -1, 1, 3, 5
            ' nul
        CASE ELSE
            IF LEN(FileDateTime) THEN
                P$ = P$ + FileDateTime
            END IF
    END SELECT
    SELECT CASE QuietSwitch
        CASE -1, 2, 3, 6
            ' nul
        CASE ELSE
            GOSUB GetAttribute
    END SELECT
    GOSUB DisplayInfo
END IF
Colorf2 Plain, 0
RETURN

' display drive attribute.
DisplayDriveInfo:
Colorf2 Green, BackGround
IF Xcoor + 19 <= 24 THEN
    P$ = "Info: " + CHR$(Current.Drive2 + 64) + ":\"
    CALL GetDriveInfo(DriveAttribute$)
    P$ = P$ + DriveAttribute$
    GOSUB DisplayInfo
END IF
Colorf2 Plain, 0
RETURN

' display file info.
DisplayInfo:
Length = 80 - Ycoor - 1
Locatef Xcoor + 19, Ycoor, 0
Printf SPACE$(Length)
Locatef Xcoor + 19, Ycoor, 0
IF Ycoor + LEN(P$) >= 80 THEN
    P$ = LEFT$(P$, 80 - Ycoor - 1)
END IF
Printf P$
RETURN

' get filesize, date/time, and attribute.
GetFileInfo:
' store long filename.
TempASCIIZ = MakeFilename$(N$, C$, D$, E$) + CHR$(0)
' get file info
CALL GetInfo(-1, V$)
RETURN

' get dir date/time, attribute.
GetDirInfo:
' store long filename.
TempASCIIZ = MakeFilename$(N$, C$, D$, Q$) + CHR$(0)
' get file info
CALL GetInfo(-1, V$)
RETURN

' get file/dir attribute.
GetAttribute:
CALL GetInfo(0, A$)
IF A$ <> Nul THEN
    IF Ycoor + LEN(P$) + LEN(A$) < 80 THEN
        P$ = P$ + A$
    END IF
END IF
RETURN

' clear long directory area.
ClearLongName1:
Colorf2 Green, BackGround
IF Xcoor + 17 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 17, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 17, Ycoor, 0
    Printf "Path: "
END IF
Colorf2 Plain, 0
RETURN

' clear long filename area.
ClearLongName2:
Colorf2 Green, BackGround
IF Xcoor + 18 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 18, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 18, Ycoor, 0
    Printf "File: "
END IF
' clear size area.
IF Xcoor + 19 <= 24 THEN
    Length = 80 - Ycoor - 1
    Locatef Xcoor + 19, Ycoor, 0
    Printf SPACE$(Length)
    Locatef Xcoor + 19, Ycoor, 0
    Printf "Info: "
END IF
Colorf2 Plain, 0
RETURN

' construct new path.
GetNewPath:
' check filespec.
IF INSTR(X$, "?") = False THEN
    IF INSTR(X$, "*") = False THEN
        X$ = "*.*"
        Xposition1 = 1
        Xposition2 = 3
    END IF
END IF
' check file.
IF X$ = Nul THEN
    X$ = "*.*"
    Xposition1 = 1
    Xposition2 = 3
END IF
' store new path.
Q$ = RTRIM$(Directories(Current.Dir!))
' store path to be parsed.
X$ = Q$ + "\" + X$
RETURN

' truncate filename for display.
TruncateFilename:
A$ = RTRIM$(E$)
IF LEN(A$) > 26 THEN
    A$ = LEFT$(A$, 23) + "..."
END IF
RETURN

' return current drive.
InitDrive:
IF C$ = Nul THEN
    D1$ = _CWD$
    C$ = UCASE$(LEFT$(D1$, 1))
    Drive.Number = ASC(C$) - 65 ' 0=a, 1=b,..
    RETURN
END IF
Drive.Number = ASC(C$) - 65
RETURN

' check path.
CheckPath:
' reset flag.
ValidPath = True

' check path.
IF D$ = Nul THEN
    RETURN
END IF
IF D$ = "\" THEN
    RETURN
END IF

' store directory.
TempASCIIZ = MakeFilename2$(N$, C$, D$) + CHR$(0)
AttributeX = GetFileAttr(TempASCIIZ)

' check filename is not path.
IF (AttributeX AND &H10) <> &H10 THEN
    ValidPath = False
END IF
RETURN

' display directory box.
DisplayDirSpec:
VarX! = SFalse
DO
    VarX! = VarX! + 1!
    IF VarX! > 10! THEN
        EXIT DO
    END IF
    IF VarX! > Num.Dirs THEN
        EXIT DO
    END IF
    Locatef CINT(VarX!) + Xcoor + 4, Ycoor + 19, 1
    Q$ = Directories(VarX!)
    Q$ = RTRIM$(Q$)
    Printf LEFT$(Q$, 12)
LOOP
RETURN

' clear directory box.
ClearDirBox:
FOR VarX = 1 TO 10
    Locatef VarX + Xcoor + 4, Ycoor + 19, 1
    Printf SPACE$(12)
NEXT
RETURN

' display directory.
DisplayCurrentDir:

' ambiguate dospath.
IF RIGHT$(D$, 2) = "\." THEN
    D$ = LEFT$(D$, LEN(D$) - 2)
END IF
TempASCIIZ = MakeFilename4$(N$, C$, D$) + CHR$(0)
V4 = True
GOSUB ShortFilenameX
IF Short.Filename$ = Nul THEN
    IF N$ = Nul THEN
        Short.Filename$ = C$ + ":\"
    ELSE
        Short.Filename$ = N$
    END IF
END IF
Z$ = Short.Filename$
IF N$ = Nul THEN
    IF MID$(Z$, 2, 1) <> ":" THEN
        Z$ = C$ + ":" + Z$
    END IF
ELSE
    IF MID$(Z$, 2, 1) = ":" THEN
        Z$ = MID$(Z$, 3)
    END IF
END IF

' get deconcatenated filename.
N1$ = N$
CALL Deconcatenate(N$, Z$, 37)
N$ = N1$

' display filename.
Locatef Xcoor + 2, Ycoor + 2, 1
Printf SPACE$(37)
Locatef Xcoor + 2, Ycoor + 2, 1
Printf UCASE$(Z$)
RETURN

' display filename box.
DisplayFileSpec:
VarX! = SFalse
DO
    VarX! = VarX! + 1!
    IF VarX! > 10! THEN
        EXIT DO
    END IF
    IF VarX! > Num.Files THEN
        EXIT DO
    END IF
    Locatef CINT(VarX!) + Xcoor + 4, Ycoor + 3, 1
    Q$ = Filenames(VarX!)
    Q$ = RTRIM$(Q$)
    Printf LEFT$(Q$, 12)
LOOP
RETURN

' clear filename box.
ClearFileBox:
FOR VarX = 1 TO 10
    Locatef VarX + Xcoor + 4, Ycoor + 3, 1
    Printf SPACE$(12)
NEXT
RETURN

' format filename for display and search purposes.
ShortFilenameX:
IF D$ = "\" THEN
    Short.Filename$ = "\"
    RETURN
END IF
IF D$ = Nul THEN
    Short.Filename$ = "\"
    RETURN
END IF

' read ambiguated 8.3 filename.
CALL GetShortFilename(Short.Filename$)
IF Short.Filename$ = Nul THEN
    RETURN
END IF

' check 8.3 flag.
IF V4 THEN
    RETURN
END IF

' store 8.3 filename.
FOR V = LEN(Short.Filename$) TO 1 STEP -1
    IF MID$(Short.Filename$, V, 1) = "\" THEN
        Short.Filename$ = MID$(Short.Filename$, V + 1)
        EXIT FOR
    END IF
NEXT
RETURN

' clear current drive.
ClearDrive:
Colorf2 Plain, 0
Locatef Xcoor + Drive.Box.Line + 4, Ycoor + 35, 0
Printf "[" + CHR$(Current.Drive2 + 64) + "]"
RETURN

' clear drive box.
ClearDriveBox:
FOR Var = 1 TO 10
    Locatef Xcoor + 4 + Var, Ycoor + 35, 1
    Printf "   "
NEXT
RETURN

' display current drive with info.
DisplayDrive:
GOSUB DisplayDriveInfo
Colorf2 Black, Plain
Locatef Xcoor + Drive.Box.Line + 4, Ycoor + 35, 1
Printf "[" + CHR$(Current.Drive2 + 64) + "]"
Locatef Xcoor + Drive.Box.Line + 4, Ycoor + 38, 1
Colorf2 Plain, 0
RETURN

' display current drive w/o info.
DisplayDrive2:
Colorf2 Black, Plain
Locatef Xcoor + Drive.Box.Line + 4, Ycoor + 35, 1
Printf "[" + CHR$(Current.Drive2 + 64) + "]"
Locatef Xcoor + Drive.Box.Line + 4, Ycoor + 38, 1
Colorf2 Plain, 0
RETURN

' display available drives.
DisplayDrives:
FOR Var = 1 TO Last.Drive
    IF Var > 10 THEN
        EXIT FOR
    END IF
    Locatef Xcoor + 4 + Var, Ycoor + 35, 1
    Printf "[" + CHR$(Var + 64) + "]"
NEXT
RETURN

' display current drives.
InitDrives:
' init drive letter.
IF Current.Drive2 = False THEN
    Current.Drive2 = Drive.Number + 1
END IF
' display drives.
V3 = Current.Drive2
IF V3 > Last.Drive THEN
    V3 = Last.Drive
END IF
GOSUB ClearDriveBox
Drive.Box.Line = 1
Current.Drive2 = 1
GOSUB DisplayDrives
GOSUB DisplayDrive2
' scroll to drive.
DO UNTIL Current.Drive2 >= V3
    V6 = True
    GOSUB MoveDriveDown
LOOP
GOSUB DisplayDriveInfo
RETURN

' process left mouse button single-click.
MouseButton1Z:
' store mouse variables.
IsMenu = False
IF MouseX = 1 THEN
    IF MouseY >= 1 AND MouseY <= 3 THEN
        CPUOFF = NOT CPUOFF
        IF CPUOFF THEN
            TIMER OFF
            _TITLE Program
            CALL DisplayToolBar
        ELSE
            TIMER ON
            CALL DisplayToolBar
        END IF
        RETURN
    END IF
    IF MouseY >= 74 AND MouseY <= 78 THEN
        CALL HelpAbout
        RETURN
    END IF
END IF
' file box pageup
IF MouseX = Xcoor + 4 THEN
    IF MouseY = Ycoor + 16 THEN
        I$ = CHR$(0) + CHR$(73)
        IsMenu = 1
        RETURN
    END IF
END IF
' file box up
IF MouseX = Xcoor + 5 THEN
    IF MouseY = Ycoor + 16 THEN
        I$ = CHR$(0) + CHR$(72)
        IsMenu = 1
        RETURN
    END IF
END IF
' file box down
IF MouseX = Xcoor + 14 THEN
    IF MouseY = Ycoor + 16 THEN
        I$ = CHR$(0) + CHR$(80)
        IsMenu = 1
        RETURN
    END IF
END IF
' file box pagedown
IF MouseX = Xcoor + 15 THEN
    IF MouseY = Ycoor + 16 THEN
        I$ = CHR$(0) + CHR$(81)
        IsMenu = 1
        RETURN
    END IF
END IF
' directory box pageup
IF MouseX = Xcoor + 4 THEN
    IF MouseY = Ycoor + 32 THEN
        I$ = CHR$(0) + CHR$(73)
        IsMenu = 2
        RETURN
    END IF
END IF
' directory box up
IF MouseX = Xcoor + 5 THEN
    IF MouseY = Ycoor + 32 THEN
        I$ = CHR$(0) + CHR$(72)
        IsMenu = 2
        RETURN
    END IF
END IF
' directory box down
IF MouseX = Xcoor + 14 THEN
    IF MouseY = Ycoor + 32 THEN
        I$ = CHR$(0) + CHR$(80)
        IsMenu = 2
        RETURN
    END IF
END IF
' directory box pagedown
IF MouseX = Xcoor + 15 THEN
    IF MouseY = Ycoor + 32 THEN
        I$ = CHR$(0) + CHR$(81)
        IsMenu = 2
        RETURN
    END IF
END IF
' drive box up
IF MouseX = Xcoor + 5 THEN
    IF MouseY = Ycoor + 39 THEN
        I$ = CHR$(0) + CHR$(72)
        IsMenu = 3
        RETURN
    END IF
END IF
' drive box down
IF MouseX = Xcoor + 14 THEN
    IF MouseY = Ycoor + 39 THEN
        I$ = CHR$(0) + CHR$(80)
        IsMenu = 3
        RETURN
    END IF
END IF
' check mouse selection boundaries for file box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 3 AND MouseY <= Ycoor + 14 THEN
        IF Num.Files > 0 THEN
            I$ = Nul
            Valid = False
            V = MouseX - (Xcoor + 4)
            IF V = Box.Line THEN
                Valid = True
            END IF
            IF V > Box.Line THEN
                IF Current.File! + (V - Box.Line) <= Num.Files THEN
                    Current.File! = Current.File! + (V - Box.Line)
                    Valid = True
                END IF
            END IF
            IF V < Box.Line THEN
                IF Current.File! - (Box.Line - V) > 0 THEN
                    Current.File! = Current.File! - (Box.Line - V)
                    Valid = True
                END IF
            END IF
            IF Valid THEN
                I$ = CHR$(13)
                IsMenu = 1
            END IF
        END IF
        RETURN
    END IF
END IF
' check mouse selection boundaries for dir box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 19 AND MouseY <= Ycoor + 31 THEN
        I$ = Nul
        Valid = False
        V = MouseX - (Xcoor + 4)
        IF V = Dir.Box.Line THEN
            Valid = True
        END IF
        IF V > Dir.Box.Line THEN
            IF Current.Dir! + (V - Dir.Box.Line) <= Num.Dirs THEN
                Current.Dir! = Current.Dir! + (V - Dir.Box.Line)
                Valid = True
            END IF
        END IF
        IF V < Dir.Box.Line THEN
            IF Current.Dir! - (Dir.Box.Line - V) > 0 THEN
                Current.Dir! = Current.Dir! - (Dir.Box.Line - V)
                Valid = True
            END IF
        END IF
        IF Valid THEN
            I$ = CHR$(13)
            IsMenu = 2
        END IF
        RETURN
    END IF
END IF
' check mouse selection boundaries for drive box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 35 AND MouseY <= Ycoor + 37 THEN
        I$ = Nul
        Valid = False
        V = MouseX - (Xcoor + 4)
        IF V = Drive.Box.Line THEN
            Valid = True
        END IF
        IF V > Drive.Box.Line THEN
            IF Current.Drive2 + (V - Drive.Box.Line) <= Last.Drive THEN
                GOSUB ClearDrive
                Current.Drive2 = Current.Drive2 + (V - Drive.Box.Line)
                Drive.Box.Line = Drive.Box.Line + (V - Drive.Box.Line)
                GOSUB DisplayDrive
                Valid = True
            END IF
        END IF
        IF V < Drive.Box.Line THEN
            IF Current.Drive2 - (Drive.Box.Line - V) > 0 THEN
                GOSUB ClearDrive
                Current.Drive2 = Current.Drive2 - (Drive.Box.Line - V)
                Drive.Box.Line = Drive.Box.Line - (Drive.Box.Line - V)
                GOSUB DisplayDrive
                Valid = True
            END IF
        END IF
        IF Valid THEN
            I$ = CHR$(13)
            IsMenu = 3
        END IF
        RETURN
    END IF
END IF
' check mouse selection boundaries for file entry area.
IF MouseX = Xcoor + 1 THEN
    IF MouseY >= Ycoor + 2 AND MouseY <= Ycoor + 39 THEN
        I$ = Nul
        IsMenu = 4
        RETURN
    END IF
END IF
' check mouse selection boundaries for file box help.
IF MouseX = Xcoor THEN
    IF MouseY = Ycoor + 39 THEN
        I$ = Nul
        CALL DisplayMenuHelp
        GOSUB RedrawScreen
        RETURN
    END IF
END IF
' check mouse selection boundaries for file box exit.
IF MouseX = Xcoor THEN
    IF MouseY = Ycoor + 40 THEN
        I$ = Nul
        IsMenu = 5
        RETURN
    END IF
END IF
RETURN

' mouseover for file/dir/drive boxes.
MouseOverX1:
' check mouse selection boundaries for file box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 3 AND MouseY <= Ycoor + 14 THEN
        IF Num.Files > 0 THEN
            B1 = MouseX - (Xcoor + 4)
            IF B1 > Box.Line THEN
                B2 = B1 - Box.Line
                IF B2 <= Num.Files THEN
                    FOR B3 = 1 TO B2
                        IF Current.File! + 1 <= Num.Files THEN
                            IF Box.Line < 10 THEN
                                GOSUB ClearFilename
                                Box.Line = Box.Line + 1
                                Current.File! = Current.File! + 1!
                                GOSUB DisplayFilename5
                                GOSUB DisplayLongName2
                            END IF
                        END IF
                    NEXT
                END IF
            ELSE
                IF B1 < Box.Line THEN
                    B2 = Box.Line - B1
                    IF B2 > 0 THEN
                        FOR B3 = 1 TO B2
                            IF Current.File! - 1 > 0 THEN
                                IF Box.Line > 1 THEN
                                    GOSUB ClearFilename
                                    Box.Line = Box.Line - 1
                                    Current.File! = Current.File! - 1!
                                    GOSUB DisplayFilename5
                                    GOSUB DisplayLongName2
                                END IF
                            END IF
                        NEXT
                    END IF
                ELSE
                    IF B1 = Box.Line THEN
                        GOSUB ClearFilename
                        GOSUB DisplayFilename5
                        GOSUB DisplayLongName2
                    END IF
                END IF
            END IF
        END IF
        RETURN
    END IF
END IF
' check mouse selection boundaries for dir box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 19 AND MouseY <= Ycoor + 31 THEN
        IF Num.Dirs > 0 THEN
            D1 = MouseX - (Xcoor + 4)
            IF D1 > Dir.Box.Line THEN
                D2 = D1 - Dir.Box.Line
                IF D2 > 0 THEN
                    IF D2 <= Num.Dirs THEN
                        FOR D3 = 1 TO D2
                            IF Current.Dir! + 1 <= Num.Dirs THEN
                                IF Dir.Box.Line < 10 THEN
                                    GOSUB ClearDirectory
                                    Dir.Box.Line = Dir.Box.Line + 1
                                    Current.Dir! = Current.Dir! + 1!
                                    GOSUB DisplayDirectory
                                    Q$ = RTRIM$(Directories(Current.Dir!))
                                    GOSUB DisplayLongName1
                                END IF
                            END IF
                        NEXT
                    END IF
                END IF
            ELSE
                IF D1 < Dir.Box.Line THEN
                    D2 = Dir.Box.Line - D1
                    IF D2 > 0 THEN
                        IF D2 <= Num.Dirs THEN
                            FOR D3 = 1 TO D2
                                IF Current.Dir! - 1 > 0 THEN
                                    IF Dir.Box.Line > 1 THEN
                                        GOSUB ClearDirectory
                                        Dir.Box.Line = Dir.Box.Line - 1
                                        Current.Dir! = Current.Dir! - 1!
                                        GOSUB DisplayDirectory
                                        Q$ = RTRIM$(Directories(Current.Dir!))
                                        GOSUB DisplayLongName1
                                    END IF
                                END IF
                            NEXT
                        END IF
                    END IF
                ELSE
                    IF D1 = Dir.Box.Line THEN
                        GOSUB DisplayDirectory
                        Q$ = RTRIM$(Directories(Current.Dir!))
                        GOSUB DisplayLongName1
                    END IF
                END IF
            END IF
        END IF
        RETURN
    END IF
END IF
' check mouse selection boundaries for drive box.
IF MouseX >= Xcoor + 5 AND MouseX <= Xcoor + 14 THEN
    IF MouseY >= Ycoor + 35 AND MouseY <= Ycoor + 37 THEN
        X0 = MouseX - (Xcoor + 4)
        IF X0 > Drive.Box.Line THEN
            X1 = Current.Drive2 + (X0 - Drive.Box.Line)
            IF X1 <= Last.Drive THEN
                GOSUB ClearDrive
                Current.Drive2 = Current.Drive2 + (X0 - Drive.Box.Line)
                Drive.Box.Line = Drive.Box.Line + (X0 - Drive.Box.Line)
                GOSUB DisplayDrive
            END IF
        ELSE
            IF X0 < Drive.Box.Line THEN
                X1 = Current.Drive2 - (Drive.Box.Line - X0)
                IF X1 > 0 THEN
                    GOSUB ClearDrive
                    Current.Drive2 = Current.Drive2 - (Drive.Box.Line - X0)
                    Drive.Box.Line = Drive.Box.Line - (Drive.Box.Line - X0)
                    GOSUB DisplayDrive
                END IF
            ELSE
                IF X0 = Drive.Box.Line THEN
                    GOSUB DisplayDrive
                END IF
            END IF
        END IF
    END IF
END IF
RETURN

' redraws editing screen.
RedrawScreen:
Colorf2 White, BackGropund
CLS
CALL FileMainMenu
GOSUB UpdateDisplay
GOSUB DisplayCurrentPath
IF Num.Files > 0 THEN
    GOSUB DisplayLongName2
END IF
GOSUB DisplayFilename6
GOSUB DisplayFileLine
RETURN
END SUB

' read basic commands
SUB InitData
RESTORE CommandSetData
FOR V = 1 TO MaxSimple
    READ SimpleCommands(V)
NEXT
FOR V = 1 TO MaxSimple
    READ SimpleCommandsN(V)
NEXT
FOR V = 1 TO MaxComplex
    READ ComplexCommands(V)
NEXT

' read Alt-<key> data.
FOR Var = 1 TO 26
    READ Keys(Var)
NEXT

CommandSetData:
DATA "prompt","redraw","cpu","mem","cls","edit","debug","ver","help","color"
DATA "quit","system","key","status","menu","date","time","clock","hexcalc","ascii"
DATA "hexchart","whatis","toggle","about","fullscreen","window","blank","screensaver"
DATA "profiles","monitor","pushd","popd","compdir","compfile","listfile","listdir","listdrive"

' 0=no paramater
DATA 1,0,1,1,0,1,1,0,1,1,0,0,1,1,0,0,0,0,0,0,0,1,0,0,0,0,0,1,1,1,1,0,0,1,1,1,0

DATA "count","touchdir","touchfile","getattr","setattr"
DATA "encrypt","decrypt","treedel","treedir","delete"
DATA "rename","drives","label"
DATA "vol","dir","copy","parse","sort"
DATA "type","more","find","cd","chdir"
DATA "md","mkdir","rd","rmdir","mkfile"

' scancodes for Alt-A to Alt-Z.
DATA 30,48,46,32,18,33,34,35,23,36
DATA 37,38,50,49,24,25,16,19,31,20
DATA 22,47,17,45,21,44
END SUB

REM Subroutine to init variables.
SUB InitVars
' init standard error.
StderrHandle = FREEFILE
StderrFile$ = "scrn:"
OPEN "scrn:" FOR OUTPUT AS StderrHandle

' init a quote symbol.
Quote = CHR$(34)

' reset default printer port.
PrinterPort = "LPT1:"

' reset default string store variable.
StoreSearchMulti = True

' set background colors.
BackGround = 1
BackGround2 = 0
BackGround3 = 0
BackGround4 = 1
BackGround5 = 1
BackGround6 = 1
BackGround7 = 0
BackGround8 = 1
BackGround9 = 0
BackGround10 = 1
Background11 = 0
BackGround12 = 1
BackGround13 = 0
BackGround14 = 1
BackGround15 = 0

' get quit box colors
QuitBoxBorderColor = Yellow
QuitBoxTitleColor = White
QuitBoxTextColor = White
QuitBoxButton1Color = White
QuitBoxButton2Color = Plain

' get quit box coordinates
QuitXcoor = 10
QuitYcoor = 10
QuitXcoor2 = 10
QuitYcoor2 = 10

' get append box colors
AppendBoxBorderColor = Yellow
AppendBoxTitleColor = White
AppendBoxTextColor = White
AppendBoxButton1Color = White
AppendBoxButton2Color = Plain

' get append box coordinates
AppendXcoor = 10
AppendYcoor = 10

' get filename box colors
FilenameBoxBorderColor = Yellow
FilenameBoxTitleColor = White
FilenameBoxTextColor = White
FilenameBoxButton1Color = White
FilenameBoxButton2Color = Plain

' get filename box coordinates
FilenameXcoor = 10
FilenameYcoor = 10

' get jump box colors
JumpBoxBorderColor = Yellow
JumpBoxTitleColor = White
JumpBoxTextColor = White
JumpBoxButton1Color = White
JumpBoxButton2Color = Plain

' get jump box coordinates
JumpXcoor = 10
JumpYcoor = 10

' get search box colors
SearchBoxBorderColor = Yellow
SearchBoxTitleColor = White
SearchBoxTextColor = White
SearchBoxButton1Color = White
SearchBoxButton2Color = Plain
SearchBoxBackGround = 1
SearchBoxButtonBackGround = 0

' get search box coordinates
SearchXcoor = 10
SearchYcoor = 10

' get ascii box colors
AsciiBoxBorderColor = Yellow
AsciiBoxTitleColor = White
AsciiBoxTextColor = White
AsciiBoxButton1Color = White
AsciiBoxButton2Color = Plain

' get ascii box coordinates
AsciiXcoor = 10
AsciiYcoor = 10

' get about box colors
AboutBoxBorderColor = Yellow
AboutBoxTitleColor = White
AboutBoxTextColor = White

' get about box coordinates
AboutXcoor = 10
AboutYcoor = 10

' file menu box colors
FileBoxBorderColor = White
FileBoxTitleColor = White
FileBoxTextColor = White

' file menu box colors
CalcBoxBorderColor = Yellow
CalcBoxTitleColor = White
CalcBoxTextColor = White

' hexcalc box color
HexCalcBackground1 = 0 ' black
HexCalcBackground2 = 1 ' blue
HexCalcForeground1 = Yellow
HexCalcForeground2 = White

' hexcalc help box colors
HexCalcHelpBackground1 = 0 ' black
HexCalcHelpBackground2 = 1 ' blue
HexCalcHelpForeground1 = Yellow
HexCalcHelpForeground2 = White

' init ascii characters.
Hline = 205
Vline = 186
ULcorner = 201
URcorner = 187
LLcorner = 200
LRcorner = 188

' store max file menu box files.
SortMax = 32768!
SortMax2 = 16384!

' get default drives.
CALL GetDrives

' check temp directory.
FX$ = "C:\TEMP"
IF Node THEN
    FX$ = "C:\TEMP" + LTRIM$(STR$(Node))
END IF
IF _DIREXISTS(FX$) = 0 THEN
    IF _FILEEXISTS(FX$) THEN
        f$ = FX$ + CHR$(0)
        r = DeleteFileA(f$)
    END IF
    f$ = FX$ + CHR$(0)
    r = CreateDirectoryA(f$, 0)
END IF
DRX = FX$ + "\"

' init filename variables.
DumpFile = StartDir$ + "hexedit.dmp"
Var$ = ENVIRON$("DUMPFILE")
IF LEN(Var$) THEN
    DumpFile = Var$
END IF
ExcludeFile = StartDir$ + "hexedit.exd"
Var$ = ENVIRON$("EXCLUDEFILE")
IF LEN(Var$) THEN
    ExcludeFile = Var$
END IF
END SUB

REM get default drives.
SUB GetDrives
' get current drive.
Default.Drive = ASC(LEFT$(_CWD$, 1)) - 65 ' 0=a, 1=b,..

' set maximum drive.
Last.Drive = 26
END SUB

REM Store default server name.
SUB StoreServerName
' store default server name.
Var$ = ENVIRON$("COMPUTERNAME")
IF LEN(Var$) THEN
    DefaultNetPath = Var$
END IF
END SUB

REM Gets default drive/directory.
SUB GetDrive (Drive.Number, Directory$)
' store current directory.
Directory$ = _CWD$
Drive.Number = ASC(LEFT$(Directory$, 1)) - 65 ' 0=a, 1=b, ..
IF MID$(Directory$, 2, 1) = ":" THEN
    Directory$ = MID$(Directory$, 3)
END IF
IF LEFT$(Directory$, 1) = "\" THEN
    Directory$ = MID$(Directory$, 2)
END IF
END SUB

REM Subroutine to get environment variable settings,
REM   and read custom config file variables.

SUB ReadConfigFile
' reset autoconfirm for file menu box delete.
AutoConfirm = 0

' check for ambiguation variable.
IF ENVIRON$("AMBIGUATE") <> Nul THEN
    AmbiguateSwitch = True
END IF

' check for heap sort variable.
IF ENVIRON$("HEXSORT") <> Nul THEN
    HeapSortOff = True
END IF

' check for mouse wheel reversal variable.
IF ENVIRON$("REVERSEWHEEL") <> Nul THEN
    WheelReverse = True
END IF

' check for file date/time variable.
IF ENVIRON$("FILESWITCH") <> Nul THEN
    X = INT(VAL(ENVIRON$("FILESWITCH")))
    SELECT CASE X
        CASE 0, 1, 2
            FileSwitch = X
    END SELECT
END IF

' check for file menu box filesize display.
IF ENVIRON$("FILESIZE") <> Nul THEN
    X$ = ENVIRON$("FILESIZE")
    SELECT CASE X$
        CASE "0", "B"
            FileSizeType = 0
        CASE "1", "K"
            FileSizeType = 1
        CASE "2", "M"
            FileSizeType = 2
        CASE "3", "G"
            FileSizeType = 3
    END SELECT
END IF

' check for quiet variable.
IF ENVIRON$("QUIET") <> Nul THEN
    X = INT(VAL(ENVIRON$("QUIET")))
    SELECT CASE X
        CASE -1, 0, 1, 2, 3, 4, 5, 6
            QuietSwitch = X
    END SELECT
END IF

' init alternate characters.
IF ENVIRON$("HEXCHARS") <> Nul THEN
    Hline = 45
    Vline = 124
    ULcorner = 43
    URcorner = 43
    LLcorner = 43
    LRcorner = 43
END IF

' define screen offsets for file menu box.
' allows menu repositioning.
Xcoor = 3
Ycoor = 20

' check config variable.
IF ENVIRON$("HEXCONFIG") <> Nul THEN
    EXIT SUB
END IF

REM get and read file menu editor config filename.

' get config filename
Config.Filename$ = GetMenuConfigFilename$
IF _FILEEXISTS(Config.Filename$) = 0 THEN
    CALL WriteMenuConfig
    EXIT SUB
END IF

' open filename.
ErrorTrap = False
ValidFile = False
x2 = FREEFILE
OPEN Config.Filename$ FOR INPUT SHARED AS #x2
IF ErrorTrap THEN
    EXIT SUB
END IF
ValidFile = True

' read input lines.
DO WHILE NOT EOF(x2)
    LINE INPUT #x2, FileLine$

    ' remove appended comment.
    Var = INSTR(FileLine$, "'")
    IF Var THEN
        IF Var = 1 THEN
            FileLine$ = Nul
        ELSE
            FileLine$ = LEFT$(FileLine$, Var - 1)
        END IF
    END IF

    ' store config line.
    FileLine2$ = FileLine$

    ' trim config line.
    GOSUB Remove.Spaces

    ' check for prepended comment.
    IF LEFT$(FileLine$, 1) = ";" THEN
        FileLine$ = Nul
    END IF

    ' check assignment value.
    Parse = INSTR(FileLine$, "=")

    ' parse off left/right sides.
    IF Parse THEN

        ' get left/right side values.
        Char$ = LEFT$(FileLine$, Parse - 1)
        Setting$ = MID$(FileLine$, Parse + 1)

        ' get right side value for filenames
        ' enclosed in quotes containing spaces.
        Parse = INSTR(FileLine2$, "=")
        Setting2$ = MID$(FileLine2$, Parse + 1)
        Setting2$ = RTRIM$(Setting2$)
        Setting2$ = LTRIM$(Setting2$)

        ' check length of values.
        IF LEN(Char$) > 0 AND LEN(Setting$) > 0 THEN
            ' compare values.
            SELECT CASE UCASE$(Char$)
                CASE "AUTOCONFIRM"
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            AutoConfirm = False
                        CASE "ON", "-1", "TRUE"
                            AutoConfirm = True
                    END SELECT

                CASE "REVERSEWHEEL"
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            WheelReverse = False
                        CASE "ON", "-1", "TRUE"
                            WheelReverse = True
                    END SELECT
                    ' check file sort maximum
                CASE "SORTMAX"
                    Var! = 0!
                    GOSUB Remove.Commas
                    SELECT CASE UCASE$(RIGHT$(Var3$, 1))
                        CASE "K"
                            Var3$ = LEFT$(Var3$, LEN(Var3$) - 1)
                            Var3! = INT(VAL(Var3$))
                            IF Var3! >= 1! AND Var3! <= MaxPrecision1 THEN
                                Var! = Var3! * 1024!
                            END IF
                        CASE "M"
                            Var3$ = LEFT$(Var3$, LEN(Var3$) - 1)
                            Var3! = INT(VAL(Var3$))
                            IF Var3! <= MaxPrecision2 THEN
                                Var! = Var3! * 1048576!
                            END IF
                        CASE ELSE
                            Var! = INT(VAL(Var3$))
                    END SELECT
                    Var! = INT(Var!)
                    IF Var! = -1! THEN ' disable
                        SortMax = -1!
                    END IF
                    IF Var! > MaxPrecision THEN
                        Var! = MaxPrecision
                    END IF
                    IF Var! >= 1024! AND Var! <= MaxPrecision THEN
                        SortMax = Var!
                    END IF
                    ' check dir sort maximum
                CASE "SORTMAX2"
                    Var! = 0!
                    GOSUB Remove.Commas
                    SELECT CASE UCASE$(RIGHT$(Var3$, 1))
                        CASE "K"
                            Var3$ = LEFT$(Var3$, LEN(Var3$) - 1)
                            Var3! = INT(VAL(Var3$))
                            IF Var3! >= 1! AND Var3! <= MaxPrecision1 THEN
                                Var! = Var3! * 1024!
                            END IF
                        CASE "M"
                            Var3$ = LEFT$(Var3$, LEN(Var3$) - 1)
                            Var3! = INT(VAL(Var3$))
                            IF Var3! <= MaxPrecision2 THEN
                                Var! = Var3! * 1048576!
                            END IF
                        CASE ELSE
                            Var! = INT(VAL(Var3$))
                    END SELECT
                    Var! = INT(Var!)
                    IF Var! = -1! THEN ' disable
                        SortMax2 = -1!
                    END IF
                    IF Var! > MaxPrecision THEN
                        Var! = MaxPrecision
                    END IF
                    IF Var! >= 1024! AND Var! <= MaxPrecision THEN
                        SortMax2 = Var!
                    END IF
                    ' check dump filename.
                CASE "DUMPFILE"
                    DumpFile = Setting$
                    ' check exclude filename.
                CASE "EXCLUDEFILE"
                    ExcludeFile = Setting$
                CASE "HEXSORT" ' file\dir sort.
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            HeapSortOff = True
                        CASE "ON", "-1", "TRUE"
                            HeapSortOff = False
                    END SELECT
                CASE "AMBIGUATE" ' file menu box ambiguation.
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            AmbiguateSwitch = False
                        CASE "ON", "-1", "TRUE"
                            AmbiguateSwitch = True
                    END SELECT
                CASE "QUIET" ' file menu box display.
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            QuietSwitch = False
                        CASE "ON", "-1", "TRUE"
                            QuietSwitch = True
                        CASE ELSE
                            SELECT CASE INT(VAL(Setting$))
                                CASE 1 TO 6
                                    QuietSwitch = INT(VAL(Setting$))
                            END SELECT
                    END SELECT
                CASE "FILESWITCH" ' file menu box date\time display.
                    SELECT CASE INT(VAL(Setting$))
                        CASE 0, 1, 2
                            FileSwitch = INT(VAL(Setting$))
                    END SELECT
                CASE "FILESIZE" ' file menu box filesize display.
                    SELECT CASE UCASE$(Setting$)
                        CASE "0", "B"
                            FileSizeType = 0
                        CASE "1", "K"
                            FileSizeType = 1
                        CASE "2", "M"
                            FileSizeType = 2
                        CASE "3", "G"
                            FileSizeType = 3
                    END SELECT
                CASE "PORT" ' default printer port.
                    PrinterPort = Setting$
                CASE "SEARCH" ' default search variable.
                    SELECT CASE UCASE$(Setting$)
                        CASE "OFF", "0", "FALSE"
                            StoreSearchMulti = False
                        CASE "ON", "-1", "TRUE"
                            StoreSearchMulti = True
                    END SELECT
                CASE "LASTDRIVE" ' override last drive.
                    SELECT CASE UCASE$(Setting$)
                        CASE "C" TO "Z"
                            Last.Drive = ASC(UCASE$(Setting$)) - 64
                    END SELECT

                CASE "XCOOR" ' file menu box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 3 THEN
                        IF NewValue <= 8 THEN
                            Xcoor = NewValue
                        END IF
                    END IF
                CASE "YCOOR" ' file menu box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 1 THEN
                        IF NewValue <= 39 THEN
                            Ycoor = NewValue
                        END IF
                    END IF

                CASE "ABOUTXCOOR" ' about box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 2 THEN
                        IF NewValue <= 16 THEN
                            AboutXcoor = NewValue
                        END IF
                    END IF
                CASE "ABOUTYCOOR" ' about box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 1 THEN
                        IF NewValue <= 48 THEN
                            AboutYcoor = NewValue
                        END IF
                    END IF

                CASE "QUITXCOOR" ' quit box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 2 THEN
                        IF NewValue <= 16 THEN
                            QuitXcoor = NewValue
                        END IF
                    END IF
                CASE "QUITYCOOR" ' quit box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 1 THEN
                        IF NewValue <= 48 THEN
                            QuitYcoor = NewValue
                        END IF
                    END IF

                CASE "JUMPXCOOR" ' jump box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 2 THEN
                        IF NewValue <= 16 THEN
                            JumpXcoor = NewValue
                        END IF
                    END IF
                CASE "JUMPYCOOR" ' jump box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 1 THEN
                        IF NewValue <= 48 THEN
                            JumpYcoor = NewValue
                        END IF
                    END IF

                CASE "ASCIIXCOOR" ' ascii box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 2 THEN
                        IF NewValue <= 16 THEN
                            AsciiXcoor = NewValue
                        END IF
                    END IF
                CASE "ASCIIYCOOR" ' ascii box location.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 1 THEN
                        IF NewValue <= 48 THEN
                            AsciiYcoor = NewValue
                        END IF
                    END IF

                CASE "BACKGROUND", "MENUBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround = 0
                        CASE "BLUE"
                            BackGround = 1
                        CASE "GREEN"
                            BackGround = 2
                        CASE "CYAN"
                            BackGround = 3
                        CASE "RED"
                            BackGround = 4
                        CASE "MAGENTA"
                            BackGround = 5
                        CASE "BROWN"
                            BackGround = 6
                        CASE "WHITE"
                            BackGround = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND2", "HELPBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround2 = 0
                        CASE "BLUE"
                            BackGround2 = 1
                        CASE "GREEN"
                            BackGround2 = 2
                        CASE "CYAN"
                            BackGround2 = 3
                        CASE "RED"
                            BackGround2 = 4
                        CASE "MAGENTA"
                            BackGround2 = 5
                        CASE "BROWN"
                            BackGround2 = 6
                        CASE "WHITE"
                            BackGround2 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround2 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND3", "UNUSEDBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround3 = 0
                        CASE "BLUE"
                            BackGround3 = 1
                        CASE "GREEN"
                            BackGround3 = 2
                        CASE "CYAN"
                            BackGround3 = 3
                        CASE "RED"
                            BackGround3 = 4
                        CASE "MAGENTA"
                            BackGround3 = 5
                        CASE "BROWN"
                            BackGround3 = 6
                        CASE "WHITE"
                            BackGround3 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround3 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND4", "ABOUTBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround4 = 0
                        CASE "BLUE"
                            BackGround4 = 1
                        CASE "GREEN"
                            BackGround4 = 2
                        CASE "CYAN"
                            BackGround4 = 3
                        CASE "RED"
                            BackGround4 = 4
                        CASE "MAGENTA"
                            BackGround4 = 5
                        CASE "BROWN"
                            BackGround4 = 6
                        CASE "WHITE"
                            BackGround4 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround4 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND5", "HEXCALCBACKROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround5 = 0
                        CASE "BLUE"
                            BackGround5 = 1
                        CASE "GREEN"
                            BackGround5 = 2
                        CASE "CYAN"
                            BackGround5 = 3
                        CASE "RED"
                            BackGround5 = 4
                        CASE "MAGENTA"
                            BackGround5 = 5
                        CASE "BROWN"
                            BackGround5 = 6
                        CASE "WHITE"
                            BackGround5 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround5 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND6", "QUITBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround6 = 0
                        CASE "BLUE"
                            BackGround6 = 1
                        CASE "GREEN"
                            BackGround6 = 2
                        CASE "CYAN"
                            BackGround6 = 3
                        CASE "RED"
                            BackGround6 = 4
                        CASE "MAGENTA"
                            BackGround6 = 5
                        CASE "BROWN"
                            BackGround6 = 6
                        CASE "WHITE"
                            BackGround6 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround6 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND7", "QUITBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround7 = 0
                        CASE "BLUE"
                            BackGround7 = 1
                        CASE "GREEN"
                            BackGround7 = 2
                        CASE "CYAN"
                            BackGround7 = 3
                        CASE "RED"
                            BackGround7 = 4
                        CASE "MAGENTA"
                            BackGround7 = 5
                        CASE "BROWN"
                            BackGround7 = 6
                        CASE "WHITE"
                            BackGround7 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround7 = NewValue
                            END IF
                    END SELECT

                CASE "BACKGROUND8", "JUMPBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround8 = 0
                        CASE "BLUE"
                            BackGround8 = 1
                        CASE "GREEN"
                            BackGround8 = 2
                        CASE "CYAN"
                            BackGround8 = 3
                        CASE "RED"
                            BackGround8 = 4
                        CASE "MAGENTA"
                            BackGround8 = 5
                        CASE "BROWN"
                            BackGround8 = 6
                        CASE "WHITE"
                            BackGround8 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround8 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND9", "JUMPBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround9 = 0
                        CASE "BLUE"
                            BackGround9 = 1
                        CASE "GREEN"
                            BackGround9 = 2
                        CASE "CYAN"
                            BackGround9 = 3
                        CASE "RED"
                            BackGround9 = 4
                        CASE "MAGENTA"
                            BackGround9 = 5
                        CASE "BROWN"
                            BackGround9 = 6
                        CASE "WHITE"
                            BackGround9 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround9 = NewValue
                            END IF
                    END SELECT

                CASE "BACKGROUND14", "SEARCHBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            SearchBoxBackGround = 0
                        CASE "BLUE"
                            SearchBoxBackGround = 1
                        CASE "GREEN"
                            SearchBoxBackGround = 2
                        CASE "CYAN"
                            SearchBoxBackGround = 3
                        CASE "RED"
                            SearchBoxBackGround = 4
                        CASE "MAGENTA"
                            SearchBoxBackGround = 5
                        CASE "BROWN"
                            SearchBoxBackGround = 6
                        CASE "WHITE"
                            SearchBoxBackGround = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                SearchBoxBackGround = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND15", "SEARCHBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            SearchBoxButtonBackGround = 0
                        CASE "BLUE"
                            SearchBoxButtonBackGround = 1
                        CASE "GREEN"
                            SearchBoxButtonBackGround = 2
                        CASE "CYAN"
                            SearchBoxButtonBackGround = 3
                        CASE "RED"
                            SearchBoxButtonBackGround = 4
                        CASE "MAGENTA"
                            SearchBoxButtonBackGround = 5
                        CASE "BROWN"
                            SearchBoxButtonBackGround = 6
                        CASE "WHITE"
                            SearchBoxButtonBackGround = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                SearchBoxButtonBackGround = NewValue
                            END IF
                    END SELECT

                CASE "BACKGROUND10", "ASCIIBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround10 = 0
                        CASE "BLUE"
                            BackGround10 = 1
                        CASE "GREEN"
                            BackGround10 = 2
                        CASE "CYAN"
                            BackGround10 = 3
                        CASE "RED"
                            BackGround10 = 4
                        CASE "MAGENTA"
                            BackGround10 = 5
                        CASE "BROWN"
                            BackGround10 = 6
                        CASE "WHITE"
                            BackGround10 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround10 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND11", "ASCIIBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            Background11 = 0
                        CASE "BLUE"
                            Background11 = 1
                        CASE "GREEN"
                            Background11 = 2
                        CASE "CYAN"
                            Background11 = 3
                        CASE "RED"
                            Background11 = 4
                        CASE "MAGENTA"
                            Background11 = 5
                        CASE "BROWN"
                            Background11 = 6
                        CASE "WHITE"
                            Background11 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                Background11 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND12", "APPENDBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround12 = 0
                        CASE "BLUE"
                            BackGround12 = 1
                        CASE "GREEN"
                            BackGround12 = 2
                        CASE "CYAN"
                            BackGround12 = 3
                        CASE "RED"
                            BackGround12 = 4
                        CASE "MAGENTA"
                            BackGround12 = 5
                        CASE "BROWN"
                            BackGround12 = 6
                        CASE "WHITE"
                            BackGround12 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround12 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND13", "APPENDBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround13 = 0
                        CASE "BLUE"
                            BackGround13 = 1
                        CASE "GREEN"
                            BackGround13 = 2
                        CASE "CYAN"
                            BackGround13 = 3
                        CASE "RED"
                            BackGround13 = 4
                        CASE "MAGENTA"
                            BackGround13 = 5
                        CASE "BROWN"
                            BackGround13 = 6
                        CASE "WHITE"
                            BackGround13 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround13 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND14", "FILENAMEBOXBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround14 = 0
                        CASE "BLUE"
                            BackGround14 = 1
                        CASE "GREEN"
                            BackGround14 = 2
                        CASE "CYAN"
                            BackGround14 = 3
                        CASE "RED"
                            BackGround14 = 4
                        CASE "MAGENTA"
                            BackGround14 = 5
                        CASE "BROWN"
                            BackGround14 = 6
                        CASE "WHITE"
                            BackGround14 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround14 = NewValue
                            END IF
                    END SELECT
                CASE "BACKGROUND15", "FILENAMEBOXBUTTONBACKGROUND"
                    SELECT CASE UCASE$(Setting$)
                        CASE "BLACK"
                            BackGround15 = 0
                        CASE "BLUE"
                            BackGround15 = 1
                        CASE "GREEN"
                            BackGround15 = 2
                        CASE "CYAN"
                            BackGround15 = 3
                        CASE "RED"
                            BackGround15 = 4
                        CASE "MAGENTA"
                            BackGround15 = 5
                        CASE "BROWN"
                            BackGround15 = 6
                        CASE "WHITE"
                            BackGround15 = 7
                        CASE ELSE
                            NewValue = INT(VAL(Setting$))
                            IF NewValue >= 0 AND NewValue <= 7 THEN
                                BackGround15 = NewValue
                            END IF
                    END SELECT
                CASE "JUMPTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        JumpBoxTextColor = NewValue
                    END IF
                CASE "JUMPTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        JumpBoxTitleColor = NewValue
                    END IF
                CASE "JUMPBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        JumpBoxBorderColor = NewValue
                    END IF
                CASE "JUMPBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        JumpBoxButton1Color = NewValue
                    END IF
                CASE "JUMPBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        JumpBoxButton2Color = NewValue
                    END IF

                CASE "SEARCHTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        SearchBoxTextColor = NewValue
                    END IF
                CASE "SEARCHTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        SearchBoxTitleColor = NewValue
                    END IF
                CASE "SEARCHBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        SearchBoxBorderColor = NewValue
                    END IF
                CASE "SEARCHBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        SearchBoxButton1Color = NewValue
                    END IF
                CASE "SEARCHBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        SearchBoxButton2Color = NewValue
                    END IF

                CASE "ASCIITEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AsciiBoxTextColor = NewValue
                    END IF
                CASE "ASCIITITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AsciiBoxTitleColor = NewValue
                    END IF
                CASE "ASCIIBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AsciiBoxBorderColor = NewValue
                    END IF
                CASE "ASCIIBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AsciiBoxButton1Color = NewValue
                    END IF
                CASE "ASCIIBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AsciiBoxButton2Color = NewValue
                    END IF

                CASE "APPENDTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AppendBoxTextColor = NewValue
                    END IF
                CASE "APPENDTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AppendBoxTitleColor = NewValue
                    END IF
                CASE "APPENDBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AppendBoxBorderColor = NewValue
                    END IF
                CASE "APPENDBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AppendBoxButton1Color = NewValue
                    END IF
                CASE "APPENDBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AppendBoxButton2Color = NewValue
                    END IF

                CASE "FILENAMETEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FilenameBoxTextColor = NewValue
                    END IF
                CASE "FILENAMETITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FilenameBoxTitleColor = NewValue
                    END IF
                CASE "FILENAMEBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FilenameBoxBorderColor = NewValue
                    END IF
                CASE "FILENAMEBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FilenameBoxButton1Color = NewValue
                    END IF
                CASE "FILENAMEBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FilenameBoxButton2Color = NewValue
                    END IF

                CASE "QUITTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        QuitBoxTextColor = NewValue
                    END IF
                CASE "QUITTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        QuitBoxTitleColor = NewValue
                    END IF
                CASE "QUITBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        QuitBoxBorderColor = NewValue
                    END IF
                CASE "QUITBOX1"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        QuitBoxButton1Color = NewValue
                    END IF
                CASE "QUITBOX2"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        QuitBoxButton2Color = NewValue
                    END IF

                CASE "ABOUTTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AboutBoxTextColor = NewValue
                    END IF
                CASE "ABOUTTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AboutBoxTitleColor = NewValue
                    END IF
                CASE "ABOUTBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        AboutBoxBorderColor = NewValue
                    END IF

                CASE "BOXTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FileBoxTextColor = NewValue
                    END IF
                CASE "BOXTTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FileBoxTitleColor = NewValue
                    END IF
                CASE "BOXBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        FileBoxBorderColor = NewValue
                    END IF

                CASE "CALCBOXTEXT"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        CalcBoxTextColor = NewValue
                    END IF
                CASE "CALCBOXTTITLE"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        CalcBoxTitleColor = NewValue
                    END IF
                CASE "CALCBOXBORDER"
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 15 THEN
                        CalcBoxBorderColor = NewValue
                    END IF
                CASE ELSE
                    ' menu box ansi character overrides.
                    NewValue = INT(VAL(Setting$))
                    IF NewValue >= 0 AND NewValue <= 255 THEN
                        SELECT CASE UCASE$(Char$)
                            CASE "HLINE"
                                Hline = NewValue
                            CASE "VLINE"
                                Vline = NewValue
                            CASE "ULCORNER"
                                ULcorner = NewValue
                            CASE "URCORNER"
                                URcorner = NewValue
                            CASE "LLCORNER"
                                LLcorner = NewValue
                            CASE "LRCORNER"
                                LRcorner = NewValue
                        END SELECT
                    END IF
            END SELECT
        END IF
    END IF
LOOP
CLOSE #x2
EXIT SUB

' remove spaces.
Remove.Spaces:
Var1$ = FileLine$
DO
    Parse = INSTR(Var1$, " ")
    IF Parse THEN
        Var1$ = LEFT$(Var1$, Parse - 1) + MID$(Var1$, Parse + 1)
    ELSE
        EXIT DO
    END IF
LOOP
FileLine$ = Var1$
RETURN

' removes commas
Remove.Commas:
Var3$ = Setting$
DO
    Var = INSTR(Var3$, ",")
    IF Var THEN
        Var3$ = LEFT$(Var3$, Var - 1) + MID$(Var3$, Var + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN
END SUB

SUB WriteMenuConfig
DIM hfind AS _OFFSET

' force file creation
f$ = GetMenuConfigFilename$ + CHR$(0)
hfind = CreateFile(f$, &H180, 0, 0, 2, 0, 0)
IF hfind = 0 THEN
    EXIT SUB
END IF
r = CloseHandle(hfind)

' open file for output
X = FREEFILE
DataError = 0
Error.Flag = -1
f$ = GetMenuConfigFilename$
OPEN f$ FOR OUTPUT AS #X
Error.Flag = 0
IF DataError THEN
    EXIT SUB
END IF

' write values to file
PRINT #X, ";config file created " + DATE$ + " " + TIME$
RESTORE MenuData
DO
    READ X$
    IF X$ = "-eof-" THEN
        EXIT DO
    END IF
    PRINT #X, X$
LOOP
CLOSE #X
EXIT SUB

MenuData:
DATA ""
DATA "; sample config file for QB64shell File Menu Box"
DATA ""
DATA "; remove semi-colons from commented lines to use variables."
DATA ""
DATA "; high ascii characters."
DATA ";Hline = 205"
DATA ";Vline = 186"
DATA ";ULcorner = 201"
DATA ";URcorner = 187"
DATA ";LLcorner = 200"
DATA ";LRcorner = 188"
DATA ""
DATA "; low ascii characters."
DATA ";Hline = 45"
DATA ";Vline = 124"
DATA ";ULcorner = 43"
DATA ";URcorner = 43"
DATA ";LLcorner = 43"
DATA ";LRcorner = 43"
DATA ""
DATA "; color overrides:"
DATA "; background   foreground"
DATA ";  Black=0      Gray=8"
DATA ";  Blue=1       LightBlue=9"
DATA ";  Green=2      LightGreen=10"
DATA ";  Cyan=3       LightCyan=11"
DATA ";  Red=4        LightRed=12"
DATA ";  Magenta=5    LightMagenta=13"
DATA ";  Brown=6      Yellow=14"
DATA ";  White=7      HighIntensityWhite=15"
DATA ""
DATA "; File menu box"
DATA "MenuBoxBackground = 1"
DATA "; File menu help box"
DATA "HelpBoxBackGround = 0"

DATA "; Jump menu box"
DATA "JumpBoxBackGround = 1"
DATA "; Jump menu box button"
DATA "JumpBoxButtonBackGround = 0"

DATA "; Search menu box"
DATA "SearchBoxBackGround = 1"
DATA "; Search menu box button"
DATA "SearchBoxButtonBackGround = 0"

DATA "; Quit menu box"
DATA "QuitBoxBackGround = 1"
DATA "; Quit menu box button"
DATA "QuitBoxButtonBackGround = 0"

DATA "; Append menu box"
DATA "AppendBoxBackGround = 1"
DATA "; Append menu box button"
DATA "AppendBoxButtonBackGround = 0"

DATA "; Filename menu box"
DATA "FilenameBoxBackGround = 1"
DATA "; Filename menu box button"
DATA "FilenameBoxButtonBackGround = 0"

DATA "; Ascii menu box"
DATA "AsciiBoxBackGround = 1"
DATA "; Ascii menu box button"
DATA "AsciiBoxButtonBackGround = 0"

DATA "; About menu box"
DATA "AboutBoxBackGround = 1"
DATA "; Hexcalc menu box"
DATA "HexCalcBackGround = 1"
DATA ""
DATA "; Jump box foregrounds"
DATA "JumpText = 14"
DATA "JumpTitle = 15"
DATA "JumpBorder = 15"
DATA "JumpBox1 = 15"
DATA "JumpBox2 = 7"

DATA "; Search box foregrounds"
DATA "SearchText = 14"
DATA "SearchTitle = 15"
DATA "SearchBorder = 15"
DATA "SearchBox1 = 15"
DATA "SearchBox2 = 7"

DATA "; Quit box foregrounds"
DATA "QuitText = 14"
DATA "QuitTitle = 15"
DATA "QuitBorder = 15"
DATA "QuitBox1 = 15"
DATA "QuitBox2 = 7"

DATA "; Append box foregrounds"
DATA "AppendText = 14"
DATA "AppendTitle = 15"
DATA "AppendBorder = 15"
DATA "AppendBox1 = 15"
DATA "AppendBox2 = 7"

DATA "; Filename box foregrounds"
DATA "FilenameText = 14"
DATA "FilenameTitle = 15"
DATA "FilenameBorder = 15"
DATA "FilenameBox1 = 15"
DATA "FilenameBox2 = 7"

DATA "; Ascii box foregrounds"
DATA "AsciiText = 14"
DATA "AsciiTitle = 15"
DATA "AsciiBorder = 15"
DATA "AsciiBox1 = 15"
DATA "AsciiBox2 = 7"

DATA "; About box foregrounds"
DATA "AboutText = 14"
DATA "AboutTitle = 15"
DATA "AboutBorder = 15"

DATA "; File edit jump box foregrounds"
DATA "BoxText = 15"
DATA "BoxTitle = 15"
DATA "BoxBorder = 15"

DATA "; Hex calc box foregrounds"
DATA "CalcBoxText = 14"
DATA "CalcBoxTotle = 15"
DATA "CalcBoxBorder = 15"
DATA ""
DATA "; force ambiguation in file menu box."
DATA ";Ambiguate = -1"
DATA ""
DATA "; hex dump filename."
DATA ";DumpFile = filename.dmp"
DATA ""
DATA "; file exclusion."
DATA ";ExcludeFile = filename.exd"
DATA ""
DATA "; disable filename\directory sort."
DATA ";Hexsort = Off"
DATA ""
DATA "; file date/time switch (0, 1, 2) for file menu box."
DATA ";  (0=last modified 1=creation 2=last access)"
DATA ";FileSwitch = 1"
DATA ""
DATA "; filesize switch (0, 1, 2, 3) for file menu box."
DATA ";  (0=B. 1=KB. 2=MB. 3=GB.) (also: B, K. M, G)."
DATA "FileSize = 1"
DATA ""
DATA "; quiet switch for file menu box (-1, 0, 1-6)."
DATA ";  -1 = on (display no info)"
DATA ";   0 = off (display all info)"
DATA ";   1 = don't display date\time"
DATA ";   2 = don't display attribute"
DATA ";   3 = don't display date\time\attribute"
DATA ";   4 = don't display volume serial number\fat type"
DATA ";     after volume label in drive box."
DATA ";   5 = don't display date\time,don't display"
DATA ";     volume serial number\fat type after volume label in drive box."
DATA ";   6 = don't display attribute, don't display volume serial\"
DATA ";     fat type after volume label in drive box."
DATA ";Quiet = -1"
DATA ""
DATA "; file menu box position."
DATA ";  Xcoor ranges from 3 to 8,"
DATA ";  Ycoor ranges from 1 to 39."
DATA ";Xcoor = 2"
DATA ";Ycoor = 1"
DATA ""
DATA "; about box position."
DATA ";  Xcoor ranges from 2 to 16,"
DATA ";  Ycoor ranges from 1 to 48."
DATA ";AboutXcoor = 10"
DATA ";AboutYcoor = 10"
DATA ""

DATA "; quit box position."
DATA ";  Xcoor ranges from 2 to 16,"
DATA ";  Ycoor ranges from 1 to 48."
DATA ";QuitXcoor = 10"
DATA ";QuitYcoor = 10"
DATA ""

DATA "; jump box position."
DATA ";  Xcoor ranges from 2 to 16,"
DATA ";  Ycoor ranges from 1 to 48."
DATA ";JumpXcoor = 10"
DATA ";JumpYcoor = 10"
DATA ""

DATA "; default string storage override."
DATA "Search=-1"
DATA ""
DATA "; maximum files list"
DATA ";Sortmax=1024K"
DATA ""
DATA "; maximum dirs list"
DATA ";Sortmax2=1M"
DATA ""
DATA "; Override lastdrive"
DATA ";Lastdrive=H"
DATA "-eof-"
END SUB

' load current directories.
SUB LoadDirSpec (DirY$, N$, C$, D$)
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _UNSIGNED _OFFSET
REDIM DirectoryList(1) AS STRING
IF RequireDirReadOnly OR RequireDirHidden OR RequireDirSystem OR RequireDirArchive THEN
    Required = True
ELSE
    Required = False
END IF
Colorf White
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
Printf DirLoadMessage
Colorf Plain

' reset directory counter.
IF D$ = "" OR D$ = "\" THEN
    Num.Dirs = 0
ELSE
    Num.Dirs = 1
    DirectoryList(1) = ".."
END IF

' find directories.
F$ = MakeFilename3$(N$, C$, D$)
F$ = F$ + CHR$(0)
hfind = FindFirstFileA(_OFFSET(F$), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    DO
        X$ = finddata.cFileName
        VarX = INSTR(X$, CHR$(0))
        IF VarX THEN
            X$ = LEFT$(X$, VarX - 1)
        END IF

        IF INSTR(X$, "?") THEN ' unicode
            X$ = finddata.cAlternateFileName
            VarX = INSTR(X$, CHR$(0))
            IF VarX THEN
                X$ = LEFT$(X$, VarX - 1)
            END IF
        END IF

        IF LEN(X$) THEN
            Attr& = finddata.dwFileAttributes
            IF (Attr& AND &H10) = &H10 THEN
                IF X$ = "." OR X$ = ".." THEN
                    Eat$ = Nul
                ELSE
                    ' check file counter.
                    IF Num.Dirs >= MaxPrecision THEN
                        EXIT DO
                    END IF

                    ' reset file flag.
                    Flag = True
                    IF Required THEN
                        GOSUB Check.Required1
                    END IF

                    ' check flag.
                    IF Flag THEN
                        ' check excluded list.
                        IF LEN(DirY$) THEN
                            GOSUB Check.List1
                        END IF
                        IF Flag THEN
                            ' store directory.
                            Num.Dirs = Num.Dirs + 1!
                            REDIM _PRESERVE DirectoryList(Num.Dirs) AS STRING
                            DirectoryList(Num.Dirs) = X$
                        END IF
                    END IF
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(hfind, _OFFSET(finddata))
    x = FindClose(hfind)
END IF

' display number of directories.
Locatef Xcoor + 3, Ycoor + 24, 0
Printf SPACE$(8)
Locatef Xcoor + 3, Ycoor + 24, 0
Colorf White

' format display
Var# = CDBL(Num.Dirs)
IF Var# > 32767# THEN
    Var# = INT(Var# / 1024#)
END IF
Value$ = FormatX$(Var#, 0)
IF Num.Dirs > 32767! THEN
    Value$ = Value$ + "K"
END IF
Printf Value$
Colorf Plain

' sort directories.
IF HeapSortOff THEN
    EXIT SUB
END IF
IF Num.Dirs < 2 THEN
    EXIT SUB
END IF
IF SortMax2 > 0! THEN
    IF Num.Dirs >= SortMax2 THEN
        Locatef Xcoor + 1, Ycoor + 2, 1
        Printf SPACE$(38)
        Locatef Xcoor + 1, Ycoor + 2, 1
        Colorf2 White, 0
        Var! = INT(SortMax2 / 1024!)
        Printf "Dirs exceed" + STR$(Var!) + "K. Sort anyway<y/n>:"
        DO
            M$ = Nul
            WHILE M$ = Nul
                _LIMIT 100
                M$ = INKEYx$
            WEND
            IF M$ = CHR$(27) THEN
                GOSUB FileDisplay
                EXIT SUB
            END IF
            M$ = UCASE$(M$)
            IF M$ = "Y" THEN
                GOSUB FileDisplay
                EXIT DO
            END IF
            IF M$ = "N" THEN
                GOSUB FileDisplay
                EXIT SUB
            END IF
        LOOP
    END IF
END IF
Colorf White
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
Printf "Sorting dirs.."
Colorf Plain
CALL SortMenuFiles(2)
EXIT SUB

' display filename
FileDisplay:
Locatef Xcoor + 1, Ycoor + 2, 1
Printf "Filename: "
RETURN

' check dir exclusion list.
Check.List1:
Y2$ = UCASE$(DirY$)
L1$ = UCASE$(X$)
DO
    ' parse list.
    Y1 = INSTR(Y2$, " ")
    IF Y1 THEN
        Y1$ = LEFT$(Y2$, Y1 - 1)
        Y2$ = MID$(Y2$, Y1 + 1)
    ELSE
        Y1$ = Y2$
        Y2$ = Nul
    END IF
    Y1$ = LTRIM$(Y1$)
    Y1$ = RTRIM$(Y1$)
    Y2$ = LTRIM$(Y2$)
    Y2$ = RTRIM$(Y2$)
    ' check excluded file in filename.
    IF LEN(Y1$) THEN
        CALL CheckExcluded(Y1$, L1$)
        IF Exclude.File THEN
            Flag = False
            EXIT DO
        END IF
    END IF
    IF Y2$ = Nul THEN
        EXIT DO
    END IF
LOOP
RETURN

' check requirements.
Check.Required1:
IF RequireDirReadOnly THEN
    IF (Attr AND 1) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireDirHidden THEN
    IF (Attr AND 2) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireDirSystem THEN
    IF (Attr AND 4) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireDirArchive THEN
    IF (Attr AND 32) = 0 THEN
        Flag = False
    END IF
END IF
RETURN
END SUB

' load specified files.
SUB LoadFileSpec (FileY$, F$)
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _UNSIGNED _OFFSET
REDIM FilenameList(1) AS STRING
IF RequireReadOnly OR RequireHidden OR RequireSystem OR RequireArchive THEN
    Required = True
ELSE
    Required = False
END IF
Colorf White
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
Printf FileLoadMessage
Colorf Plain
Num.Files = False

' find files.
IF RIGHT$(F$, 1) = "\" THEN
    F$ = F$ + "*.*"
END IF
F$ = F$ + CHR$(0)
hfind = FindFirstFileA(_OFFSET(F$), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    DO
        X$ = finddata.cFileName
        VarX = INSTR(X$, CHR$(0))
        IF VarX THEN
            X$ = LEFT$(X$, VarX - 1)
        END IF

        IF INSTR(X$, "?") THEN ' unicode
            X$ = finddata.cAlternateFileName
            VarX = INSTR(X$, CHR$(0))
            IF VarX THEN
                X$ = LEFT$(X$, VarX - 1)
            END IF
        END IF

        IF LEN(X$) THEN
            Attr& = finddata.dwFileAttributes
            IF (Attr& AND &H10) = &H0 THEN
                ' check file counter.
                IF Num.Files >= MaxPrecision THEN
                    EXIT DO
                END IF

                ' reset file flag.
                Flag = True
                IF Required THEN
                    GOSUB Check.Required2
                END IF

                ' check flag.
                IF Flag THEN
                    ' check excluded list.
                    IF LEN(FileY$) THEN
                        GOSUB Check.List2
                    END IF

                    ' check flag.
                    IF Flag THEN
                        ' store filename.
                        Num.Files = Num.Files + 1!
                        REDIM _PRESERVE FilenameList(Num.Files) AS STRING
                        FilenameList(Num.Files) = X$
                    END IF
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(hfind, _OFFSET(finddata))
    x = FindClose(hfind)
END IF

' display number of files.
Locatef Xcoor + 3, Ycoor + 9, 0
Printf SPACE$(9)
Locatef Xcoor + 3, Ycoor + 9, 0
Colorf White

' format display
Var# = CDBL(Num.Files)
IF Var# > 32767# THEN
    Var# = INT(Var# / 1024#)
END IF
Value$ = FormatX$(Var#, 0)
IF Num.Files > 32767! THEN
    Value$ = Value$ + "K"
END IF
Printf Value$
Colorf Plain

' sort files.
IF HeapSortOff THEN
    EXIT SUB
END IF
IF Num.Files < 2 THEN
    EXIT SUB
END IF
IF SortMax > 0! THEN
    IF Num.Files >= SortMax THEN
        Locatef Xcoor + 1, Ycoor + 2, 1
        Printf SPACE$(38)
        Locatef Xcoor + 1, Ycoor + 2, 1
        Colorf2 White, 0
        Var! = INT(SortMax / 1024!)
        Printf "Files exceed" + STR$(Var!) + "K. Sort anyway<y/n>:"
        DO
            M$ = Nul
            WHILE M$ = Nul
                _LIMIT 100
                M$ = INKEYx$
            WEND
            IF M$ = CHR$(27) THEN
                GOSUB FileDisplay2
                EXIT SUB
            END IF
            M$ = UCASE$(M$)
            IF M$ = "Y" THEN
                GOSUB FileDisplay2
                EXIT DO
            END IF
            IF M$ = "N" THEN
                GOSUB FileDisplay2
                EXIT SUB
            END IF
        LOOP
    END IF
END IF
Colorf White
Locatef Xcoor + 1, Ycoor + 12, 1
Printf SPACE$(26)
Locatef Xcoor + 1, Ycoor + 12, 1
Printf "Sorting files.."
Colorf Plain
CALL SortMenuFiles(1)
EXIT SUB

' display filename
FileDisplay2:
Locatef Xcoor + 1, Ycoor + 2, 1
Printf "Filename: "
RETURN

' check file exclusion list.
Check.List2:
Y2$ = UCASE$(FileY$)
L1$ = UCASE$(X$)
DO
    ' parse list.
    Y1 = INSTR(Y2$, " ")
    IF Y1 THEN
        Y1$ = LEFT$(Y2$, Y1 - 1)
        Y2$ = MID$(Y2$, Y1 + 1)
    ELSE
        Y1$ = Y2$
        Y2$ = Nul
    END IF
    Y1$ = LTRIM$(Y1$)
    Y1$ = RTRIM$(Y1$)
    Y2$ = LTRIM$(Y2$)
    Y2$ = RTRIM$(Y2$)
    ' check excluded file in filename.
    IF LEN(Y1$) THEN
        CALL CheckExcluded(Y1$, L1$)
        IF Exclude.File THEN
            Flag = False
            EXIT DO
        END IF
    END IF
    IF Y2$ = Nul THEN
        EXIT DO
    END IF
LOOP
RETURN

' check requirements.
Check.Required2:
IF RequireReadOnly THEN
    IF (Attr AND 1) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireHidden THEN
    IF (Attr AND 2) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireSystem THEN
    IF (Attr AND 4) = 0 THEN
        Flag = False
    END IF
END IF
IF RequireArchive THEN
    IF (Attr AND 32) = 0 THEN
        Flag = False
    END IF
END IF
RETURN
END SUB

REM filename structure function.
FUNCTION Filenames$ (Var1!)
IF Var1! > 0! THEN
    Filenames$ = FilenameList(Var1!)
ELSE
    Filenames$ = Nul
END IF
END FUNCTION

REM directory structure function.
FUNCTION Directories$ (Var1!)
IF Num.Dirs = 0! THEN
    Directories$ = "."
    EXIT FUNCTION
END IF
IF Var1! > 0! THEN
    Directories$ = DirectoryList(Var1!)
ELSE
    Directories$ = Nul
END IF
END FUNCTION

' sorts directories/filenames
SUB SortMenuFiles (Var1)
DIM I1 AS STRING, I2 AS STRING
' sort windows filenames.
IF Var1 = 1 THEN
    Num! = Num.Files
    ' perform heap sort.
    IF FileSort THEN
        ' descending.
        FOR I! = 2! TO Num!
            J! = I!
            DO UNTIL J! = 1!
                P! = J! \ 2!
                I1 = RTRIM$(Filenames(J!))
                I2 = RTRIM$(Filenames(P!))
                IF I1 < I2 THEN
                    FilenameList(J!) = I2
                    FilenameList(P!) = I1
                    J! = P!
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
        FOR I! = Num! TO 2! STEP -1!
            I1 = RTRIM$(Filenames(1!))
            I2 = RTRIM$(Filenames(I!))
            FilenameList(1!) = I2
            FilenameList(I!) = I1
            J! = 1!
            DO
                IF 2! * J! > I! - 1! THEN
                    EXIT DO
                END IF
                C! = 2! * J!
                IF C! + 1! <= I! - 1! THEN
                    IF Filenames(C! + 1!) < Filenames(C!) THEN
                        C! = C! + 1!
                    END IF
                END IF
                I1 = RTRIM$(Filenames(J!))
                I2 = RTRIM$(Filenames(C!))
                IF I1 >= I2 THEN
                    FilenameList(J!) = I2
                    FilenameList(C!) = I1
                    J! = C!
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
    ELSE
        ' ascending.
        FOR I! = 2! TO Num!
            J! = I!
            DO UNTIL J! = 1!
                P! = J! \ 2!
                I1 = RTRIM$(Filenames(J!))
                I2 = RTRIM$(Filenames(P!))
                IF I1 > I2 THEN
                    FilenameList(J!) = I2
                    FilenameList(P!) = I1
                    J! = P!
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
        FOR I! = Num! TO 2! STEP -1!
            I1 = RTRIM$(Filenames(1!))
            I2 = RTRIM$(Filenames(I!))
            FilenameList(1!) = I2
            FilenameList(I!) = I1
            J! = 1!
            DO
                IF 2! * J! > I! - 1! THEN
                    EXIT DO
                END IF
                C! = 2! * J!
                IF C! + 1! <= I! - 1! THEN
                    IF Filenames(C! + 1!) > Filenames(C!) THEN
                        C! = C! + 1!
                    END IF
                END IF
                I1 = RTRIM$(Filenames(J!))
                I2 = RTRIM$(Filenames(C!))
                IF I1 < I2 THEN
                    FilenameList(J!) = I2
                    FilenameList(C!) = I1
                    J! = C!
                ELSE
                    EXIT DO
                END IF
            LOOP
        NEXT
    END IF
    ' sort windows directories.
ELSE
    Num! = Num.Dirs
    Start! = 0!
    IF RTRIM$(Directories(1)) = ".." THEN
        Start! = 1!
    END IF
    IF Start! >= Num! THEN
        EXIT SUB
    END IF
    ' perform insertion sort.
    FOR R! = 2! + Start! TO Num!
        I1 = RTRIM$(Directories(R!))
        FOR J! = R! TO 2! + Start! STEP -1!
            I2 = RTRIM$(Directories(J! - 1!))
            ' descending.
            IF DirSort THEN
                IF I2 < I1 THEN
                    DirectoryList(J!) = I2
                ELSE
                    EXIT FOR
                END IF
            ELSE
                ' ascending.
                IF I2 > I1 THEN
                    DirectoryList(J!) = I2
                ELSE
                    EXIT FOR
                END IF
            END IF
        NEXT
        DirectoryList(J!) = I1
    NEXT
END IF
END SUB

REM subroutine to send string of characters to screen.
SUB Printf (Var1$)
FOR VarX = 1 TO LEN(Var1$)
    VarY = ASC(MID$(Var1$, VarX, 1))
    PRINT CHR$(VarY);
NEXT
END SUB

REM subroutine stores recent color change.
SUB Colorf (Var)
COLOR Var
END SUB

REM subroutine stores recent color change.
SUB Colorf2 (Var1, Var2)
COLOR Var1, Var2
END SUB

REM subroutine sets cursor position.
SUB Locatef (Var1, Var2, Var3)
' test end of page.
IF Var1 > 24 THEN
    ' scroll screen.
    PRINT ;
END IF
LOCATE Var1, Var2, Var3
END SUB

REM File box main menu.

SUB FileMainMenu
Colorf2 White, BackGround
CALL DisplayLowerLines
Locatef Xcoor, Ycoor, 0
Colorf2 White, 0
Colorf Yellow
Printf CHR$(ULcorner) + STRING$(10, Hline)
Colorf Green
Printf "(<esc>=Quit,F9=Help)"
Colorf Yellow
Printf STRING$(9, Hline) + CHR$(URcorner)
Locatef Xcoor + 1, Ycoor, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "Filename:"
Colorf Yellow
Printf SPACE$(29) + CHR$(Vline)
Locatef Xcoor + 2, Ycoor, 0
Printf CHR$(Vline) + SPACE$(39) + CHR$(Vline)
Locatef Xcoor + 3, Ycoor, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "Files:          Dirs:          Drives:"
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor + 4, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(ULcorner) + STRING$(12, Hline) + CHR$(URcorner) + "  "
Printf CHR$(ULcorner) + STRING$(12, Hline) + CHR$(URcorner) + "  "
Printf CHR$(ULcorner) + STRING$(3, Hline) + CHR$(URcorner) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 5, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 6, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 7, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 8, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 9, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 10, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 11, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 12, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 13, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 14, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(12) + CHR$(Vline) + "  "
Printf CHR$(Vline) + SPACE$(3) + CHR$(Vline) + " "
Colorf Yellow
Printf CHR$(Vline)
Colorf Yellow
Locatef Xcoor + 15, Ycoor, 0
Printf CHR$(Vline)
Colorf Magenta
Printf " " + CHR$(LLcorner) + STRING$(12, Hline) + CHR$(LRcorner) + "  "
Printf CHR$(LLcorner) + STRING$(12, Hline) + CHR$(LRcorner) + "  "
Printf CHR$(LLcorner) + STRING$(3, Hline) + CHR$(LRcorner) + " "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor + 16, Ycoor, 0
Printf CHR$(LLcorner) + STRING$(39, Hline) + CHR$(LRcorner)
Colorf White
' make file input box mouse symbols.
Locatef Xcoor + 4, Ycoor + 16, 0
Printf "<"
Locatef Xcoor + 5, Ycoor + 16, 0
Printf "|"
Locatef Xcoor + 14, Ycoor + 16, 0
Printf "|"
Locatef Xcoor + 15, Ycoor + 16, 0
Printf ">"
' make directory input box mouse symbols.
Locatef Xcoor + 4, Ycoor + 32, 0
Printf "<"
Locatef Xcoor + 5, Ycoor + 32, 0
Printf "|"
Locatef Xcoor + 14, Ycoor + 32, 0
Printf "|"
Locatef Xcoor + 15, Ycoor + 32, 0
Printf ">"
' make drive input box mouse symbols.
Locatef Xcoor + 5, Ycoor + 39, 0
Printf "|"
Locatef Xcoor + 14, Ycoor + 39, 0
Printf "|"
' make file menu box question symbol.
Locatef Xcoor, Ycoor + 39, 0
Printf "?"
' make file menu box exit symbol.
Locatef Xcoor, Ycoor + 40, 0
Printf "x"
' make remaining menu box areas.
IF Xcoor - 1 > 1 THEN
    Colorf2 Green, BackGround
    Locatef Xcoor - 1, Ycoor, 0
    Printf "  File Menu Box For QB64 Windows Shell."
END IF
CALL BottomRow
END SUB

REM File box bottom row menu.

SUB BottomRow
IF Xcoor + 17 <= 24 THEN
    Colorf2 Green, BackGround
    Locatef Xcoor + 17, Ycoor, 0
    Printf "Path:"
END IF
IF Xcoor + 18 <= 24 THEN
    Colorf2 Green, BackGround
    Locatef Xcoor + 18, Ycoor, 0
    Printf "File:"
END IF
IF Xcoor + 19 <= 24 THEN
    Colorf2 Green, BackGround
    Locatef Xcoor + 19, Ycoor, 0
    Printf "Size:"
END IF
Colorf2 Plain, 0
END SUB

' display softkeys when off.
SUB DisplayLowerLines
KEY OFF
VIEW PRINT 2 TO 25
CLS
IF Status THEN
    COLOR 15, 0
    LOCATE 24, 1, 1
    PRINT SPACE$(80);
    COLOR 14, 0
    LOCATE 24, 1, 1
    CALL DisplayVersion
END IF
IF KEYOFF = 0 THEN
    COLOR 15, 0
    LOCATE 25, 1, 1
    PRINT SPACE$(80);
    LOCATE 25, 1, 1
    FOR V = 1 TO 10
        COLOR 7, 0
        PRINT LTRIM$(STR$(V));
        IF V < 10 THEN
            V$ = LEFT$(FunctionKeys(V), 6)
            V$ = V$ + SPACE$(6 - LEN(V$))
            GOSUB ReplaceChars
            COLOR 0, 7
            _CONTROLCHR OFF
            PRINT V$;
            _CONTROLCHR ON
            COLOR 7, 0
            PRINT " ";
        ELSE
            V$ = LEFT$(FunctionKeys(V), 5)
            V$ = V$ + SPACE$(5 - LEN(V$))
            GOSUB ReplaceChars
            COLOR 0, 7
            _CONTROLCHR OFF
            PRINT V$;
            _CONTROLCHR ON
        END IF
    NEXT
END IF
COLOR 7, 0
EXIT SUB

ReplaceChars:
FOR X = 1 TO LEN(V$)
    Z = ASC(MID$(V$, X, 1))
    IF Z >= 0 AND Z <= 31 THEN
        IF Z = 13 THEN
            MID$(V$, X, 1) = CHR$(27)
        ELSE
            MID$(V$, X, 1) = " "
        END IF
    END IF
NEXT
RETURN
END SUB

REM display filename border area.

SUB DisplayScreen3X
Locatef 3, 2, 0
Colorf Magenta
Printf CHR$(Vline) + " "
Colorf Green
Printf CHR$(ULcorner) + STRING$(46, Hline) + CHR$(URcorner) + " "
Printf CHR$(ULcorner) + STRING$(22, Hline) + CHR$(URcorner)
Colorf White
Printf "x"
Colorf Magenta
Printf CHR$(Vline)
END SUB

REM prompts for key press.
SUB PressKey
Colorf White
Printf "Press a key:"
DO
    _LIMIT 100
    IF INKEYx$ <> "" THEN
        EXIT DO
    END IF
LOOP
END SUB

' prompt for escape key.
SUB PromptInkey
Z$ = Nul
DO
    _LIMIT 100
    Z$ = INKEYx$
    ' escape
    IF Z$ = CHR$(27) THEN
        EXIT DO
    END IF
LOOP
END SUB

REM prompts for any key or left mouse click.
SUB PromptKey
DO
    _LIMIT 100
    IF INKEYx$ <> Nul THEN
        EXIT DO
    END IF

    ' call mouse subroutine.
    Z = MouseDriver

    ' check left mouse button.
    IF MouseButton1 THEN
        EXIT DO
    END IF
LOOP
END SUB

REM prompts for escape key or left mouse click.
SUB PromptEscKey
I$ = Nul
DO
    _LIMIT 100
    I$ = INKEYx$
    IF LEN(I$) THEN
        ' check escape key pressed.
        IF I$ = CHR$(27) THEN
            EXIT DO
        END IF
    END IF

    ' call mouse subroutine.
    Z = MouseDriver

    ' check left mouse button.
    IF MouseButton1 THEN
        EXIT DO
    END IF
LOOP
END SUB

' file menu box move subroutine.
SUB MoveMenu
GOSUB StoreArea
GOSUB RestoreArea
I$ = Nul
DO
    I$ = Nul
    DO
        _LIMIT 100
        I$ = INKEYx$
        IF LEN(I$) THEN
            EXIT DO
        END IF
        Z = MouseDriver ' get mouse
        IF MouseButton1 THEN
            MouseButton1 = False
            CALL ClearMouse
            IF Xcoor - 1 >= 3 THEN
                Xcoor = Xcoor - 1
                GOSUB RestoreArea
            END IF
        END IF
        IF MouseButton2 THEN
            MouseButton2 = False
            CALL ClearMouse
            IF Xcoor + 1 <= 8 THEN
                Xcoor = Xcoor + 1
                GOSUB RestoreArea
            END IF
        END IF
        IF MouseButton3 THEN
            I$ = CHR$(0) + CHR$(76) ' center box
            EXIT DO
        END IF
        IF MouseWheel = -1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(80) ' down
            ELSE
                I$ = CHR$(0) + CHR$(72) ' up
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
        IF MouseWheel = 1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(72) ' up
            ELSE
                I$ = CHR$(0) + CHR$(80) ' down
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
    LOOP
    Var = False
    SELECT CASE LEN(I$)
        CASE 1
            SELECT CASE I$
                CASE CHR$(27) ' exit
                    I$ = Nul
                    EXIT DO
                CASE "|"
                    Xcoor = 3
                    Var = True
                CASE "#"
                    Xcoor = 8
                    Var = True
                CASE "<" ' Up
                    IF Xcoor - 1 >= 3 THEN
                        Xcoor = Xcoor - 1
                        Var = True
                    END IF
                CASE ">" ' Down
                    IF Xcoor + 1 <= 8 THEN
                        Xcoor = Xcoor + 1
                        Var = True
                    END IF
                CASE "1", "2", "`", "~" ' left
                    IF Ycoor - 1 >= 1 THEN
                        Ycoor = Ycoor - 1
                        Var = True
                    END IF
                CASE "3" TO "8" ' up/down
                    Xcoor = INT(VAL(I$))
                    Var = True
                CASE "9", "0", "-", "=", "_", "+" ' right
                    IF Ycoor + 1 <= 39 THEN
                        Ycoor = Ycoor + 1
                        Var = True
                    END IF
                CASE "!", "@", "#", "$", "%", "^", "&", "*", "(", ")" ' center box
                    Xcoor = 3
                    Ycoor = 20
                    Var = True
            END SELECT
        CASE 2
            VarQ = ASC(RIGHT$(I$, 1))
            SELECT CASE VarQ
                CASE 59, 60 ' F1, F2
                    Xcoor = 3
                    Var = True
                CASE 61 TO 66 ' F3 to F8
                    Xcoor = VarQ - 58
                    Var = True
                CASE 67, 68 ' F9, F10
                    Xcoor = 8
                    Var = True

                CASE 94, 95 ' Ctrl-F1, Ctrl-F2
                    Xcoor = 3
                    Var = True
                CASE 96 TO 101 ' Ctrl-F3 to Ctrl-F8
                    Xcoor = VarQ - 93
                    Var = True
                CASE 102, 103 ' Ctrl-F9, Ctrl-F10
                    Xcoor = 8
                    Var = True

                CASE 104, 105 ' Alt-F1, Alt-F2
                    Xcoor = 3
                    Var = True
                CASE 106 TO 111 ' Alt-F3 to Alt-F8
                    Xcoor = VarQ - 103
                    Var = True
                CASE 112, 113 ' Alt-F9, Alt-F10
                    Xcoor = 8
                    Var = True

                CASE 120, 121 ' Alt-1, Alt-2
                    Xcoor = 3
                    Var = True
                CASE 122 TO 127 ' Alt-3 to Alt-8
                    Xcoor = VarQ - 119
                    Var = True
                CASE 128, 129 ' Alt-9, Alt-0
                    Xcoor = 8
                    Var = True

                CASE 133, 134 ' F11, F12
                    Xcoor = 8
                    Var = True
                CASE 137, 138 ' Ctrl-F11, Ctrl-F12
                    Xcoor = 8
                    Var = True
                CASE 139, 140 ' Alt-F11, Alt-F12
                    Xcoor = 8
                    Var = True

                CASE 152, 72, 142, 74, 130 ' Alt-Up/Up/Ctrl--/Alt--
                    IF Xcoor - 1 >= 3 THEN
                        Xcoor = Xcoor - 1
                        Var = True
                    END IF
                CASE 160, 80, 144, 78, 131 ' Alt-Down/Down/Ctrl-+/Alt-+
                    IF Xcoor + 1 <= 8 THEN
                        Xcoor = Xcoor + 1
                        Var = True
                    END IF
                CASE 75, 155 ' Left/Alt-Left
                    IF Ycoor - 1 >= 1 THEN
                        Ycoor = Ycoor - 1
                        Var = True
                    END IF
                CASE 77, 157 ' Right/Alt-Right
                    IF Ycoor + 1 <= 39 THEN
                        Ycoor = Ycoor + 1
                        Var = True
                    END IF
                CASE 71, 119, 132 ' Home/Ctrl-Home/Ctrl-PageUp = move to upper-left
                    Xcoor = 3
                    Ycoor = 1
                    Var = True
                CASE 79, 117, 118 ' End/Ctrl-End/Ctrl-PageDown = move to lower-right
                    Xcoor = 8
                    Ycoor = 39
                    Var = True
                CASE 73 ' PageUp = move to upper-right
                    Xcoor = 3
                    Ycoor = 39
                    Var = True
                CASE 81 ' PageDown = move to lower-left
                    Xcoor = 8
                    Ycoor = 1
                    Var = True
                CASE 115, 155 ' Ctrl-Left/Alt-Left = move to left-middle
                    Xcoor = 3
                    Ycoor = 1
                    Var = True
                CASE 116, 157 ' Ctrl-Right/Alt-Right = move to right-middle
                    Xcoor = 3
                    Ycoor = 39
                    Var = True
                CASE 141 ' Ctrl-Up = move to upper-middle
                    Xcoor = 3
                    Ycoor = 20
                    Var = True
                CASE 145 ' Ctrl-Down = move to lower-middle
                    Xcoor = 8
                    Ycoor = 20
                    Var = True
                CASE 76, 143 ' Keypad-5/Ctrl-Keypad-5 = center box
                    Xcoor = 5
                    Ycoor = 20
                    Var = True
                CASE 74, 142 ' Alt-Keypad-/Ctrl-Keypad- = move to left-middle
                    Xcoor = 3
                    Ycoor = 1
                    Var = True
                CASE 78, 144 ' Alt-Keypad+/Ctrl-Keypad+ = move to right-middle
                    Xcoor = 3
                    Ycoor = 39
                    Var = True
            END SELECT
    END SELECT
    IF Var THEN
        GOSUB RestoreArea
    END IF
LOOP
EXIT SUB

' store area.
StoreArea:
RowX1 = False
ColumnY1 = False
FOR RowX2 = Xcoor - 1 TO Xcoor + 16
    RowX1 = RowX1 + 1
    ColumnY1 = False
    FOR ColumnY2 = Ycoor TO Ycoor + 41
        ColumnY1 = ColumnY1 + 1
        ' store ascii character.
        Area1(RowX1, ColumnY1) = SCREEN(RowX2, ColumnY2)
        ' store color. (undocumented: also stores background color).
        Area2(RowX1, ColumnY1) = SCREEN(RowX2, ColumnY2, 1)
    NEXT
NEXT
RETURN

' restore area.
RestoreArea:
Colorf2 White, BackGround
CALL DisplayLowerLines
RowX1 = False
ColumnY1 = False
FOR RowX2 = Xcoor - 1 TO Xcoor + 16
    RowX1 = RowX1 + 1
    ColumnY1 = False
    FOR ColumnY2 = Ycoor TO Ycoor + 41
        ColumnY1 = ColumnY1 + 1
        Locatef RowX2, ColumnY2, 1
        ' restore color.
        TempZ = Area2(RowX1, ColumnY1)
        ' undocumented: also stores background color.
        VarB = INT(TempZ / 16)
        VarF = TempZ MOD 16
        Colorf2 VarF, VarB
        ' restore ascii character.
        Printf CHR$(Area1(RowX1, ColumnY1))
    NEXT
NEXT
RETURN
END SUB

SUB OverridePrompt
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require readonly bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireReadOnly = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireReadOnly = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require hidden bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireHidden = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireHidden = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require system bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireSystem = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireSystem = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require archive bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireArchive = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireArchive = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require dir readonly bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireDirReadOnly = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireDirReadOnly = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require dir hidden bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireDirHidden = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireDirHidden = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require dir system bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireDirSystem = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireDirSystem = False
        EXIT DO
    END IF
LOOP
Locatef Xcoor + 1, Ycoor + 2, 1
Printf SPACE$(38)
Locatef Xcoor + 1, Ycoor + 2, 1
Colorf2 White, 0
Printf "Require dir archive bit. Press <y/n>:"
DO
    M$ = Nul
    WHILE M$ = Nul
        _LIMIT 100
        M$ = INKEYx$
    WEND
    M$ = UCASE$(M$)
    IF M$ = "Y" THEN
        RequireDirArchive = True
        EXIT DO
    END IF
    IF M$ = "N" THEN
        RequireDirArchive = False
        EXIT DO
    END IF
LOOP
END SUB

REM Help screen subroutines:

' display help for file menu input box.
SUB DisplayMenuHelp
Xcoor2 = 5
Ycoor2 = 3
GOSUB RedrawHelpScreen
I$ = Nul
DO
    I$ = Nul
    DO
        _LIMIT 100
        I$ = INKEYx$
        IF I$ <> Nul THEN
            EXIT DO
        END IF
        Z = MouseDriver ' get mouse
        IF MouseButton1 THEN
            MouseButton1 = False
            CALL ClearMouse
            IF Xcoor2 - 1 >= 2 THEN
                Xcoor2 = Xcoor2 - 1
                GOSUB RedrawHelpScreen
                Var = False
            END IF
        END IF
        IF MouseButton2 THEN
            MouseButton2 = False
            CALL ClearMouse
            IF Xcoor2 + 1 <= 13 THEN
                Xcoor2 = Xcoor2 + 1
                GOSUB RedrawHelpScreen
                Var = False
            END IF
        END IF
        IF MouseButton3 THEN
            I$ = CHR$(0) + CHR$(76) ' center box
            EXIT DO
        END IF
        IF MouseWheel = -1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(80) ' down
            ELSE
                I$ = CHR$(0) + CHR$(72) ' up
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
        IF MouseWheel = 1 THEN
            IF WheelReverse THEN
                I$ = CHR$(0) + CHR$(72) ' up
            ELSE
                I$ = CHR$(0) + CHR$(80) ' down
            END IF
            MousePressed = -1
            MouseWheel = 0
            EXIT DO
        END IF
    LOOP
    Var = False
    SELECT CASE LEN(I$)
        CASE 1
            SELECT CASE I$
                CASE CHR$(27) ' ExitBox
                    EXIT DO
                CASE "|" ' SelectBox
                    IF Xcoor2 <> 2 THEN
                        Xcoor2 = 2
                        Var = True
                    END IF
                CASE "#" ' SelectBox
                    IF Xcoor2 <> 13 THEN
                        Xcoor2 = 13
                        Var = True
                    END IF
                CASE "1" TO "9" ' SelectBox
                    IF INT(VAL(I$)) + 1 <> Xcoor2 THEN
                        Xcoor2 = INT(VAL(I$)) + 1
                        Var = True
                    END IF
                CASE "0" ' SelectBox
                    IF Xcoor2 <> 11 THEN
                        Xcoor2 = 11
                        Var = True
                    END IF
                CASE "-" ' SelectBox
                    IF Xcoor2 <> 12 THEN
                        Xcoor2 = 12
                        Var = True
                    END IF
                CASE "=" ' SelectBox
                    IF Xcoor2 <> 13 THEN
                        Xcoor2 = 13
                        Var = True
                    END IF
            END SELECT
        CASE 2
            VarQ = ASC(RIGHT$(I$, 1))
            SELECT CASE VarQ
                ' Up/Ctrl-Up/Alt-Up/Ctrl--,Alt-- = MoveUp
                CASE 72, 141, 152, 142, 74, 130
                    IF Xcoor2 - 1 >= 2 THEN
                        Xcoor2 = Xcoor2 - 1
                        Var = True
                    END IF
                    ' Down/Ctrl-Down/Alt-Down/Ctrl-+/Alt-+ = MoveDown
                CASE 80, 145, 160, 144, 78, 131
                    IF Xcoor2 + 1 <= 13 THEN
                        Xcoor2 = Xcoor2 + 1
                        Var = True
                    END IF
                CASE 76, 143 ' Keypad-5/Ctrl-Keypad-5 = CenterBox
                    IF Xcoor2 <> 5 THEN
                        Xcoor2 = 5
                        Var = True
                    END IF
                    ' Home/PageUp/Ctrl-Home/Ctrl-PageUp = MoveTop
                CASE 71, 73, 119, 132
                    IF Xcoor2 <> 2 THEN
                        Xcoor2 = 2
                        Var = True
                    END IF
                    ' End/PageDown/Ctrl-End/Ctrl-PageDown = MoveBottom
                CASE 79, 81, 117, 118
                    IF Xcoor2 <> 13 THEN
                        Xcoor2 = 13
                        Var = True
                    END IF
            END SELECT
    END SELECT
    IF Var THEN
        Var = False
        GOSUB RedrawHelpScreen
    END IF
LOOP
EXIT SUB

RedrawHelpScreen:
Colorf2 White, BackGround2
CALL DisplayLowerLines
Locatef Xcoor2, Ycoor2, 0
Printf "                 File Menu Box Function Keys Help Screen"
Locatef Xcoor2 + 1, Ycoor2, 0
Colorf2 White, 0
Colorf Yellow
Printf CHR$(ULcorner) + STRING$(74, Hline) + CHR$(URcorner)
Locatef Xcoor2 + 2, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "                  Press <escape> To Return To Menu:                      "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 3, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F1 - Change Drive  F8 - Attribute Specs Alt-F3 Dir Exclude List          "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 4, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F2 - Toggle Sort   F9 - Display Help    Alt-F4 Toggle File Sort Direction"
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 5, Ycoor2, 0
Colorf Yellow
Printf CHR$(Vline) + " "
Colorf White
Printf "F3 - Move Menu     F10 - Exit Editor    Alt-F5 Toggle Dir Sort Direction "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 6, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F4 - Back Drive    F11 - Shell To DOS   Alt-F6 Toggle File Date/Time Info"
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 7, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F5 - Forward Drive F12 - Load FileSpec  Alt-F7 Toggle File Size Info     "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 8, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F6 - Toggle Ambiguate  Alt-F1 Toggle Quiet   Alt-F8 Load Network Path    "
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 9, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "F7 - Redraw Menu Box   Alt-F2 Exclude List   Alt-F9 Load Default Net Path"
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 10, Ycoor2, 0
Printf CHR$(Vline) + " "
Colorf White
Printf "Alt-A to Alt-Z  Select Drive Letter   Alt-F10 Load Default Net Path Drive"
Colorf Yellow
Printf CHR$(Vline)
Locatef Xcoor2 + 11, Ycoor2, 0
Printf CHR$(LLcorner) + STRING$(74, Hline) + CHR$(LRcorner)
Colorf White
RETURN
END SUB

' construct file/dir exclusion list.
SUB ExcludeList (V1, Y$)
IF V1 = 1 THEN
    Locatef Xcoor + 1, Ycoor + 2, 1
    Printf SPACE$(38)
    Locatef Xcoor + 1, Ycoor + 2, 1
    Colorf2 White, 0
    Printf "Enter file spec: "
    Y$ = LineInput$(Xcoor + 1, Ycoor + 19, 21)
ELSE
    Locatef Xcoor + 1, Ycoor + 2, 1
    Printf SPACE$(38)
    Locatef Xcoor + 1, Ycoor + 2, 1
    Colorf2 White, 0
    Printf "Enter dir spec: "
    Y$ = LineInput$(Xcoor + 1, Ycoor + 18, 22)
END IF
IF LEN(Y$) THEN
    ' process filelist filename.
    Filename2$ = FILELISTEXISTS(Y$)
    IF LEN(Filename2$) THEN
        Y$ = Nul

        ' open filelist filename.
        Error.Flag = True
        DataError = False
        x2 = FREEFILE
        OPEN Filename2$ FOR INPUT AS #x2
        Error.Flag = False
        IF DataError THEN
            Locatef Xcoor + 1, Ycoor + 2, 1
            Printf SPACE$(38)
            Locatef Xcoor + 1, Ycoor + 2, 1
            Colorf2 White, 0
            Printf "Filelist not found. Press <esc>:"
            CALL PromptInkey
            EXIT SUB
        END IF
        DO UNTIL EOF(x2)
            LINE INPUT #x2, Y1$
            Y1$ = LTRIM$(Y1$)
            Y1$ = RTRIM$(Y1$)
            Y1$ = UCASE$(Y1$)
            IF LEN(Y1$) THEN
                IF LEFT$(Y1$, 1) = Quote THEN
                    Y1$ = MID$(Y1$, 2)
                    IF RIGHT$(Y1$, 1) = Quote THEN
                        Y1$ = LEFT$(Y1$, LEN(Y1$) - 1)
                    END IF
                END IF
                IF LEN(Y1$) THEN
                    IF INSTR(Y1$, " ") = False THEN
                        ' add excluded filename.
                        Y$ = Y$ + Y1$ + " "
                        IF LEN(Y$) > 4096 THEN
                            Y$ = Nul
                            EXIT DO
                        END IF
                    END IF
                END IF
            END IF
        LOOP
        CLOSE #x2
        Locatef Xcoor + 1, Ycoor + 2, 1
        Printf SPACE$(38)
        Locatef Xcoor + 1, Ycoor + 2, 1
        Colorf2 White, 0
        IF LEN(Y$) THEN
            Printf "Filelist loaded. Press <esc>:"
        ELSE
            Printf "Filelist not loaded. Press <esc>:"
        END IF
        CALL PromptInkey
    END IF
END IF
Y$ = LTRIM$(Y$)
Y$ = RTRIM$(Y$)
END SUB

' check netpath exists.
'  returns -1 if netpath not detected.
FUNCTION NETPATHEXISTS (Z$)
VarX$ = Z$ + CHR$(0)
VarX = GetDriveType(VarX$)
IF VarX > 1 THEN
    NETPATHEXISTS = 0
ELSE
    NETPATHEXISTS = -1
END IF
END FUNCTION

SUB Parse.Path (D$, X$)
' parse out imbedded \.\ path.
GOSUB PathOverride1
' parse out trailing \.. path with multiple dots.
GOSUB PathOverride2
' parse out leading ..\ path with multiple dots.
GOSUB PathOverride3
' parse out leading \..\ path with multiple dots.
GOSUB PathOverride4
' parse out imbedded \..\ path with multiple dots.
GOSUB PathOverride5
' parse out imbedded \\ from path.
GOSUB PathOverride6
EXIT SUB

' parse out imbedded \.\ path.
PathOverride1:
IF LEN(X$) THEN
    DO
        V = INSTR(X$, "\.\")
        IF V THEN
            X$ = LEFT$(X$, V) + MID$(X$, V + 3)
        ELSE
            EXIT DO
        END IF
    LOOP
END IF
RETURN

' parse out trailing \.. path with multiple dots.
PathOverride2:
Z$ = Nul
IF LEN(X$) THEN
    ' search for trailing path.
    DO
        Z = False
        L = False
        IF RIGHT$(X$, 1) = "." THEN
            FOR V = LEN(X$) TO 1 STEP -1
                IF MID$(X$, V, 1) = "." THEN
                    L = L + 1
                ELSE
                    IF MID$(X$, V, 1) = "\" THEN
                        Z = True
                        EXIT FOR
                    ELSE
                        EXIT FOR
                    END IF
                END IF
            NEXT
        END IF
        IF Z = False THEN
            EXIT DO
        END IF
        Z$ = MID$(X$, V) + Z$
        X$ = LEFT$(X$, V - 1)
    LOOP
    IF Z$ = Nul THEN
        ' strip trailing dots.
        IF INSTR(X$, "\") THEN
            DO
                IF RIGHT$(X$, 1) = "." THEN
                    X$ = LEFT$(X$, LEN(X$) - 1)
                ELSE
                    EXIT DO
                END IF
            LOOP
        END IF
        RETURN
    END IF
    DO
        ' search leading dots.
        L = False
        FOR V = 2 TO LEN(Z$)
            IF MID$(Z$, V, 1) = "." THEN
                L = L + 1
                IF V = LEN(Z$) THEN
                    Z$ = Nul
                    EXIT FOR
                END IF
            ELSE
                IF MID$(Z$, V, 1) = "\" THEN
                    Z$ = MID$(Z$, V)
                    EXIT FOR
                END IF
            END IF
        NEXT
        ' strip trailing path.
        DO
            L2 = 1
            Z = False
            FOR V = LEN(X$) TO 1 STEP -1
                IF MID$(X$, V, 1) = "\" THEN
                    L2 = L2 + 1
                    IF L = L2 THEN
                        Z = True
                        EXIT DO
                    END IF
                END IF
            NEXT
            IF Z = False THEN
                EXIT DO
            END IF
        LOOP
        IF Z = False THEN
            X$ = Nul
            EXIT DO
        END IF
        IF V = 1 THEN
            IF LEFT$(X$, 1) = "\" THEN
                X$ = "\"
                EXIT DO
            END IF
        END IF
        X$ = LEFT$(X$, V - 1)
        IF Z$ = Nul THEN
            EXIT DO
        END IF
    LOOP
END IF
' strip trailing dots.
IF INSTR(X$, "\") THEN
    DO
        IF RIGHT$(X$, 1) = "." THEN
            X$ = LEFT$(X$, LEN(X$) - 1)
        ELSE
            EXIT DO
        END IF
    LOOP
END IF
RETURN

' parse out leading ..\ path with multiple dots.
PathOverride3:
Z$ = D$
DO
    ' search for leading dots.
    Z = False
    L = False
    IF LEFT$(X$, 1) = "." THEN
        FOR V = 1 TO LEN(X$)
            IF MID$(X$, V, 1) = "." THEN
                L = L + 1
            ELSE
                IF MID$(X$, V, 1) = "\" THEN
                    Z = True
                    EXIT FOR
                ELSE
                    EXIT FOR
                END IF
            END IF
        NEXT
    END IF
    IF Z = False THEN
        EXIT DO
    END IF
    X$ = MID$(X$, V + 1)
    ' parse off trailing path.
    IF L > 1 THEN
        L2 = 1
        DO
            Z = False
            FOR V = LEN(Z$) TO 1 STEP -1
                IF MID$(Z$, V, 1) = "\" THEN
                    Z$ = LEFT$(Z$, V - 1)
                    L2 = L2 + 1
                    Z = True
                    EXIT FOR
                END IF
            NEXT
            IF Z = False THEN
                Z$ = Nul
                EXIT DO
            END IF
            IF L2 = L THEN
                EXIT DO
            END IF
        LOOP
    END IF
LOOP
D$ = Z$
RETURN

' parse out leading \..\ path with multiple dots.
PathOverride4:
DO
    IF LEFT$(X$, 1) = "\" THEN
        IF LEN(X$) = 1 THEN
            EXIT DO
        END IF
        L = False
        FOR V = 2 TO LEN(X$)
            IF MID$(X$, V, 1) = "." THEN
                L = L + 1
            ELSE
                IF MID$(X$, V, 1) = "\" THEN
                    IF L > 0 THEN
                        X$ = MID$(X$, V)
                        EXIT FOR
                    END IF
                END IF
                EXIT DO
            END IF
        NEXT
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' parse out imbedded \..\ path with multiple dots.
PathOverride5:
IF LEN(X$) THEN
    DO
        ' search for \.\
        Z = False
        FOR V = 1 TO LEN(X$)
            IF MID$(X$, V, 1) = "\" THEN
                L = False
                FOR V2 = V + 1 TO LEN(X$)
                    IF MID$(X$, V2, 1) = "." THEN
                        L = L + 1
                    ELSE
                        IF MID$(X$, V2, 1) = "\" THEN
                            IF L > 0 THEN
                                Z = True
                            END IF
                            EXIT FOR
                        ELSE
                            EXIT FOR
                        END IF
                    END IF
                NEXT
            END IF
            IF Z = True THEN
                EXIT FOR
            END IF
        NEXT
        IF Z = False THEN
            EXIT DO
        END IF
        Z$ = "\" + STRING$(L, ".") + "\"
        V4 = INSTR(X$, Z$)
        A$ = MID$(X$, V4 + L + 2)
        L2 = 1
        DO
            V4 = V4 - 1
            IF V4 = False THEN
                EXIT DO
            END IF
            IF MID$(X$, V4, 1) = "\" THEN
                L2 = L2 + 1
                IF L2 = L THEN
                    EXIT DO
                END IF
            END IF
        LOOP
        IF V4 = False THEN
            X$ = A$
        ELSE
            X$ = LEFT$(X$, V4) + A$
        END IF
    LOOP
END IF

' search for imbedded ..\
DO
    V = INSTR(X$, ".\")
    IF V THEN
        L = False
        FOR V2 = V TO 1 STEP -1
            IF MID$(X$, V2, 1) = "." THEN
                L = L + 1
            ELSE
                X$ = LEFT$(X$, V2) + MID$(X$, V + 1)
                EXIT FOR
            END IF
        NEXT
    ELSE
        EXIT DO
    END IF
LOOP
RETURN

' strip imbedded \\ from path.
PathOverride6:
DO
    V = INSTR(X$, "\\")
    IF V THEN
        X$ = LEFT$(X$, V - 1) + MID$(X$, V + 1)
    ELSE
        EXIT DO
    END IF
LOOP
RETURN
END SUB

' strip multiple dots from path.
SUB Parse.Path2 (D$, X$)
DO
    IF LEFT$(X$, 2) = ".." THEN
        Z = False
        FOR Var = LEN(D$) TO 1 STEP -1
            IF MID$(D$, Var, 1) = "\" THEN
                D$ = LEFT$(D$, Var - 1)
                X$ = MID$(X$, 2)
                Z = True
                EXIT FOR
            END IF
        NEXT
        IF Z = False THEN
            D$ = Nul
            X$ = Nul
            EXIT DO
        END IF
    ELSE
        X$ = Nul
        EXIT DO
    END IF
LOOP
END SUB

REM Functions:
REM   MakeFilename?$  --  constructs filename for file menu box.

FUNCTION MakeFilename$ (N$, C$, D$, X$)
IF N$ = Nul THEN
    IF D$ = Nul THEN
        MakeFilename$ = C$ + ":\" + X$
    ELSE
        MakeFilename$ = C$ + ":\" + D$ + "\" + X$
    END IF
ELSE
    IF D$ = Nul THEN
        MakeFilename$ = N$ + "\" + X$
    ELSE
        MakeFilename$ = N$ + "\" + D$ + "\" + X$
    END IF
END IF
END FUNCTION

FUNCTION MakeFilename2$ (N$, C$, D$)
IF N$ = Nul THEN
    IF D$ = Nul THEN
        MakeFilename2$ = C$ + ":\"
    ELSE
        IF D$ = "\" THEN
            MakeFilename2$ = C$ + ":\"
        ELSE
            MakeFilename2$ = C$ + ":\" + D$
        END IF
    END IF
ELSE
    IF D$ = Nul THEN
        MakeFilename2$ = N$ + "\"
    ELSE
        IF D$ = "\" THEN
            MakeFilename2$ = N$ + "\"
        ELSE
            MakeFilename2$ = N$ + "\" + D$
        END IF
    END IF
END IF
END FUNCTION

FUNCTION MakeFilename3$ (N$, C$, D$)
IF N$ = Nul THEN
    IF D$ = Nul THEN
        MakeFilename3$ = C$ + ":\*.*"
    ELSE
        IF D$ = "\" THEN
            MakeFilename3$ = C$ + ":\*.*"
        ELSE
            MakeFilename3$ = C$ + ":\" + D$ + "\*.*"
        END IF
    END IF
ELSE
    IF D$ = Nul THEN
        MakeFilename3$ = N$ + "\*.*"
    ELSE
        IF D$ = "\" THEN
            MakeFilename3$ = N$ + "\*.*"
        ELSE
            MakeFilename3$ = N$ + "\" + D$ + "\*.*"
        END IF
    END IF
END IF
END FUNCTION

FUNCTION MakeFilename4$ (N$, C$, D$)
IF N$ = Nul THEN
    MakeFilename4$ = C$ + ":\" + D$
ELSE
    IF D$ = "\" THEN
        MakeFilename4$ = N$ + "\"
    ELSE
        MakeFilename4$ = N$ + "\" + D$
    END IF
END IF
END FUNCTION

' get drive information.
SUB GetDriveInfo (DriveAttribute$)
' reset drive info.
DriveAttribute$ = Nul

' check drive exists.
IF DRIVEEXISTS(Current.Drive2) THEN
    EXIT SUB
END IF

' get drive info.
VarX$ = CHR$(Current.Drive2 + 64) + ":\" + CHR$(0)
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    ' get volume label.
    tmp1$ = RTRIM$(Vname$)
    v = INSTR(tmp1$, CHR$(0))
    IF v THEN tmp1$ = LEFT$(tmp1$, v - 1)
    DriveAttribute$ = tmp1$

    ' check quiet switch.
    SELECT CASE QuietSwitch
        CASE -1, 4, 5, 6
            EXIT SUB
    END SELECT

    ' serial number.
    Sname$ = LEFT$(HEX$(serial~&), 4) + "-" + RIGHT$(HEX$(serial~&), 4)
    DriveAttribute$ = DriveAttribute$ + " (" + Sname$ + ")"

    ' file system type.
    tmp1$ = RTRIM$(Fname$)
    v = INSTR(tmp1$, CHR$(0))
    IF v THEN tmp1$ = LEFT$(tmp1$, v - 1)
    DriveAttribute$ = DriveAttribute$ + " [" + tmp1$ + "]"
    EXIT SUB
END IF
DriveAttribute$ = DriveType
END SUB

SUB GetInfo (VarZ, VarZ$)
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _UNSIGNED _OFFSET
IF VarZ = 0 THEN
    X$ = Nul
    IF (FileAttribute AND 1) = 1 THEN
        X$ = X$ + "o"
    END IF
    IF (FileAttribute AND 2) = 2 THEN
        X$ = X$ + "h"
    END IF
    IF (FileAttribute AND 4) = 4 THEN
        X$ = X$ + "s"
    END IF
    IF (FileAttribute AND 32) = 32 THEN
        X$ = X$ + "a"
    END IF
    IF X$ <> Nul THEN
        X$ = " (" + X$ + ")"
    END IF
    VarZ$ = X$
    EXIT SUB
END IF
 
' construct file date and time for display.
FileDateTime = Nul
FileSize = False
hfind = FindFirstFileA(_OFFSET(TempASCIIZ), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    FileAttribute = finddata.dwFileAttributes
    FileSize = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow
    SELECT CASE FileSwitch
        CASE 0 ' last modified /TW
            x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
        CASE 1 ' creation /TC
            x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
        CASE 2 ' last accessed /TA
            x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
        CASE ELSE ' /TW
            x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    END SELECT
    FileDateTime = "[" + GetDateTime$(z) + "]"
    x = FindClose(hfind)
    EXIT SUB
END IF

' get file attribute.
FileAttribute = GetFileAttr(TempASCIIZ)
END SUB

' returns MM-DD-YYYY HH:MM:SS
FUNCTION GetDateTime$ (z)
GetDateTime$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
GetDateTime$ = GetDateTime$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
GetDateTime$ = GetDateTime$ + LTRIM$(STR$(SysTime.wYear)) + " "

GetDateTime$ = GetDateTime$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
GetDateTime$ = GetDateTime$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
GetDateTime$ = GetDateTime$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
END FUNCTION

REM get file attributes.
FUNCTION GetFileAttr (Var$)
Var = GetFileAttributes(TempASCIIZ)
IF Var < 0 THEN
    Var = 0
END IF
GetFileAttr = Var
END FUNCTION

REM Deconcatenates a filename to a specified length
REM  with imbedded \..\ character strings.
REM
REM Input:
REM   Y$ - Netpath (full path)
REM   Z$ - DosPath (ambiguated)
REM   Z% - Length (32-64)
REM
REM Returns:
REM   Z$ - Concatenated path
REM     C:\path\..\path\
REM   or,
REM     For remote or default server:
REM       \\server name\share name\path\..\path\

SUB Deconcatenate (Y$, Z$, Z%)
DIM StoreASCIIZ AS STRING * 260
Y$ = UCASE$(Y$)
Z$ = UCASE$(Z$)
' check netpath.
IF LEN(Y$) THEN
    ' store net path.
    Y1$ = UCASE$(Y$)
    IF RIGHT$(Y1$, 1) <> "\" THEN
        Y1$ = Y1$ + "\"
    END IF

    ' check default net path.
    IF RTRIM$(DefaultNetPath) <> Nul THEN

        ' store default netpath.
        Z1$ = "\\" + UCASE$(RTRIM$(DefaultNetPath)) + "\"

        ' check netpath is in default path.
        IF LEFT$(Y1$, LEN(Z1$)) = Z1$ THEN

            ' ambiguate netpath.
            StoreASCIIZ = TempASCIIZ
            TempASCIIZ = Y1$ + CHR$(0)
            CALL GetShortFilename(Short.Filename$)

            ' store ambiguated netpath.
            TempASCIIZ = StoreASCIIZ
            N1$ = Short.Filename$
        END IF
    END IF
    IF N1$ = Nul THEN
        N1$ = Y1$
    END IF
    N1$ = UCASE$(N1$)

    ' truncate drive letter.
    IF MID$(N1$, 2, 1) = ":" THEN
        N1$ = MID$(N1$, 3)
    END IF

    ' truncate netpath from dospath.
    IF LEFT$(Z$, LEN(N1$)) = N1$ THEN
        Z$ = MID$(Z$, LEN(N1$) + 1)
    END IF

    ' get ending position of server name.
    Y2 = INSTR(3, Y1$, "\")

    ' get ending position of share name.
    Y3 = INSTR(Y2 + 1, Y1$, "\")

    ' get server\share name.
    Y1$ = LEFT$(Y1$, Y3)
END IF

' check path
IF INSTR(Z$, "\") = False THEN
    IF LEN(Y$) THEN
        Z$ = Y1$ + Z$
    END IF
    IF LEN(Z$) > Z% THEN
        Z$ = LEFT$(Z$, Z% - 3) + "..."
    END IF
    EXIT SUB
END IF

' decrement length of netpath
Z1% = Z%
IF LEN(Y1$) THEN
    Z1% = Z1% - LEN(Y1$)
    IF Z1% <= 0 THEN
        Z$ = LEFT$(Y1$, Z% - 3) + "..."
        EXIT SUB
    END IF
    Z1% = Z1% + 1
END IF

' deconcatenate path
IF Y$ <> Nul THEN
    IF LEFT$(Z$, 1) <> "\" THEN
        Z$ = "\" + Z$
    END IF
END IF
DO
    IF LEN(Z$) > Z1% THEN
        V1 = INSTR(Z$, "\")
        IF V1 > 0 THEN
            DO
                IF MID$(Z$, V1, 4) = "\..\" THEN
                    V1 = INSTR(V1 + 1, Z$, "\")
                ELSE
                    EXIT DO
                END IF
            LOOP
            V2 = INSTR(V1 + 1, Z$, "\")
            IF V2 > 0 THEN
                Z$ = LEFT$(Z$, V1) + ".." + MID$(Z$, V2)
            ELSE
                EXIT DO
            END IF
        END IF
    ELSE
        EXIT DO
    END IF
    ' remove redundant paths
    DO
        V = INSTR(Z$, "\..\..\")
        IF V > 0 THEN
            Z$ = LEFT$(Z$, V - 1) + "\..\" + MID$(Z$, V + 7)
        ELSE
            EXIT DO
        END IF
    LOOP
LOOP

' remove redundant paths
DO
    V = INSTR(Z$, "\..\..\")
    IF V > 0 THEN
        Z$ = LEFT$(Z$, V - 1) + "\..\" + MID$(Z$, V + 7)
    ELSE
        EXIT DO
    END IF
LOOP

' prepend net path
IF LEN(Y1$) THEN
    IF Z$ = "\" THEN
        Z$ = Y1$
    ELSE
        IF LEFT$(Z$, 2) = "\\" THEN
            N1$ = Z$
        ELSE
            IF LEFT$(Z$, 1) = "\" THEN
                Z$ = Y1$ + MID$(Z$, 2)
            ELSE
                Z$ = Y1$ + Z$
            END IF
        END IF
    END IF
END IF

' truncate to length
IF LEN(Z$) > Z% THEN
    Z$ = LEFT$(Z$, Z% - 3) + "..."
END IF
END SUB

REM returns ambiguated filename.
SUB GetShortFilename (Short.Filename$)
DIM TempASCIIZ2 AS STRING * 260
R = GetShortPathName(TempASCIIZ, TempASCIIZ2, MAX_PATH + 1)
IF R > 0 THEN
    Short.Filename$ = TempASCIIZ2
ELSE
    Short.Filename$ = TempASCIIZ
END IF
Imbedded = INSTR(Short.Filename$, CHR$(0))
IF Imbedded THEN
    Short.Filename$ = LEFT$(Short.Filename$, Imbedded - 1)
END IF
END SUB

' inits mouse.
SUB ClearMouse
WHILE _MOUSEINPUT: WEND ' empty buffer
END SUB

' test volume media inserted.
FUNCTION MEDIAEXISTS2 (V)
' check drive exists.
IF DRIVEEXISTS(V) THEN
    MEDIAEXISTS2 = False
    EXIT FUNCTION
END IF

' check cdrom exists
IF DriveType = "[CDROM]" THEN
    MEDIAEXISTS2 = True
    EXIT FUNCTION
END IF

' get drive info.
VarX$ = CHR$(V + 64) + ":\" + CHR$(0)
Vname$ = SPACE$(MAX_PATH)
Fname$ = SPACE$(MAX_PATH)
R = GetVolumeInformationA(VarX$, Vname$, MAX_PATH, serial~&, empty1~&, empty2~&, Fname$, MAX_PATH)
IF R THEN
    MEDIAEXISTS2 = True
ELSE
    MEDIAEXISTS2 = False
END IF
END FUNCTION

' Custom line input function.
' Input:
'    XC = X-coordinate
'    YC = Y-coordinate
'    LC = Edit area length
FUNCTION LineInput$ (XC, YC, LC)
' init entry area position.
Xposition1 = 1
Xposition2 = 0

' display edit entry area.
Insert = -1
GOSUB DisplayFileLine

' main input loop.
DO
    ' get keypress.
    _LIMIT 100
    I$ = INKEYx$

    ' check key.
    IF LEN(I$) = 1 THEN

        ' enter key.
        IF I$ = CHR$(13) THEN
            LineInput$ = X$
            EXIT DO
        END IF

        ' escape key.
        IF I$ = CHR$(27) THEN
            LineInput$ = ""
            EXIT DO
        END IF

        ' backspace.
        IF I$ = CHR$(8) THEN
            IF LEN(X$) THEN
                IF Xposition2 > 0 THEN
                    IF Insert THEN ' destructive backspace
                        X$ = LEFT$(X$, Xposition2 - 1) + MID$(X$, Xposition2 + 1)
                    END IF
                    Xposition2 = Xposition2 - 1
                    IF Xposition2 - LC <= Xposition1 THEN
                        Xposition1 = Xposition1 - 1
                        IF Xposition1 < 1 THEN
                            Xposition1 = 1
                        END IF
                    END IF
                    GOSUB DisplayFileLine
                END IF
            END IF
        ELSE
            ' append/insert character into edit area.
            OK = -1
            IF ASC(I$) >= 0 AND ASC(I$) <= 31 THEN
                OK = 0 ' disallow control charcters
            END IF
            IF OK THEN
                IF Insert THEN
                    X$ = LEFT$(X$, Xposition2) + I$ + MID$(X$, Xposition2 + 1)
                    Xposition2 = Xposition2 + 1
                    IF Xposition2 > LC THEN
                        Xposition1 = Xposition1 + 1
                    END IF
                ELSE
                    IF Xposition2 + 1 > LEN(X$) THEN
                        X$ = X$ + I$ ' append
                        Xposition2 = Xposition2 + 1
                        IF Xposition2 > LC THEN
                            Xposition1 = Xposition1 + 1
                        END IF
                    ELSE ' insert
                        MID$(X$, Xposition2 + 1, 1) = I$
                        Xposition2 = Xposition2 + 1
                    END IF
                END IF
                GOSUB DisplayFileLine
            END IF
        END IF
    END IF

    ' check extended key
    IF LEN(I$) = 2 THEN
        SELECT CASE ASC(RIGHT$(I$, 1))
            CASE 82 ' Insert
                Insert = NOT Insert
                IF Insert THEN
                    LOCATE , , 1, 8, 8 ' line cursor
                ELSE
                    LOCATE , , 1, 0, 8 ' block cursor
                END IF
            CASE 83 ' Delete
                IF LEN(X$) > 0 THEN
                    X$ = LEFT$(X$, Xposition2) + MID$(X$, Xposition2 + 2)
                    GOSUB DisplayFileLine
                END IF
            CASE 71, 73, 119 ' Home/PageUp/Ctrl-Home
                IF Xposition2 > 0 THEN
                    Xposition1 = 1
                    Xposition2 = 0
                    GOSUB DisplayFileLine
                END IF
            CASE 79, 81, 117 ' End/PageDown/Ctrl-End
                IF Xposition2 < LEN(X$) THEN
                    Xposition2 = LEN(X$)
                    Xposition1 = Xposition2 - LC + 1
                    IF Xposition1 < 1 THEN
                        Xposition1 = 1
                    END IF
                    GOSUB DisplayFileLine
                END IF
            CASE 75, 115, 155 ' Left/Ctrl-Left/Alt-Left
                IF LEN(X$) THEN
                    IF Xposition2 > 0 THEN
                        Xposition2 = Xposition2 - 1
                        IF Xposition2 - LC <= Xposition1 THEN
                            Xposition1 = Xposition1 - 1
                            IF Xposition1 < 1 THEN
                                Xposition1 = 1
                            END IF
                            GOSUB DisplayFileLine
                        END IF
                    END IF
                END IF
            CASE 77, 116, 157 ' Right/Ctrl-Right/Alt-Right
                IF LEN(X$) THEN
                    IF Xposition2 < LEN(X$) THEN
                        Xposition2 = Xposition2 + 1
                        IF Xposition2 > Xposition1 + LC - 1 THEN
                            Xposition1 = Xposition1 + 1
                        END IF
                        GOSUB DisplayFileLine
                    END IF
                END IF
        END SELECT
    END IF
LOOP
LOCATE , , 1, 8, 8
_KEYCLEAR
EXIT FUNCTION

' display current entry area.
DisplayFileLine:
Locatef XC, YC, 1
Printf SPACE$(LC)

' display edit line and position cursor.
Locatef XC, YC, 1
IF LEN(X$) THEN
    Colorf White
    IF MaskInput THEN
        LV = LEN(MID$(X$, Xposition1, LC))
        Printf STRING$(LV, MaskChar$)
    ELSE
        Printf MID$(X$, Xposition1, LC)
    END IF
END IF
Locatef XC, YC + Xposition2 - Xposition1 + 1, 1
Colorf2 Plain, 0
RETURN
END FUNCTION

' formatting function.
'   Var#  input value.
'   Var3  format type:
'      1 = filelength in hexedit.
'      0 = file status area in file menu box.
'     -1 = filesize in file menu box.
FUNCTION FormatX$ (Var#, Var3)
IF Var3 = -1 THEN
    SELECT CASE FileSizeType
        CASE 0 ' Byte
            FormatX$ = FormatStringX$(Var#, 0) + " B. "
        CASE 1 ' Kilobyte
            Var# = Var# / 1024#
            FormatX$ = FormatStringX$(Var#, 1) + " KB. "
        CASE 2 ' Megabyte
            Var# = Var# / 1048576#
            FormatX$ = FormatStringX$(Var#, 2) + " MB. "
        CASE 3 ' Gigabyte
            Var# = Var# / 1073741824#
            FormatX$ = FormatStringX$(Var#, 3) + " GB. "
    END SELECT
    IF LEFT$(FormatX$, 1) = "." THEN
        FormatX$ = "0" + FormatX$
    END IF
    EXIT FUNCTION
END IF
FormatX$ = FormatStringX$(Var#, 0)
END FUNCTION

' formats a double numeric string
FUNCTION FormatStringX$ (s#, x) ' x = 0 B, 1 K, 2 M, 3 G
x$ = Nul
s$ = STR$(s#)
IF INSTR(s$, "D") THEN ' return string
    FormatStringX$ = s$
    EXIT FUNCTION
END IF
IF LEFT$(s$, 1) = "-" THEN ' store sign
    e$ = "-"
    s$ = MID$(s$, 2)
END IF
s$ = LTRIM$(s$) ' format string
IF INSTR(s$, ".") THEN
    q$ = MID$(s$, INSTR(s$, "."))
    s$ = LEFT$(s$, INSTR(s$, ".") - 1)
END IF
FOR l = LEN(s$) TO 3 STEP -3
    x$ = MID$(s$, l - 2, 3) + "," + x$
NEXT
IF l > 0 THEN
    x$ = MID$(s$, 1, l) + "," + x$
END IF
IF LEN(s$) < 3 THEN
    x$ = s$
END IF
IF RIGHT$(x$, 1) = "," THEN
    x$ = LEFT$(x$, LEN(x$) - 1)
END IF

' construct string
SELECT CASE x
    CASE 0 ' B
        x$ = e$ + x$
    CASE 1 ' K .0
        IF LEN(q$) THEN
            q$ = LEFT$(q$, 2)
        END IF
        x$ = e$ + x$ + q$
    CASE 2 ' M .00
        IF LEN(q$) THEN
            q$ = LEFT$(q$, 3)
        END IF
        x$ = e$ + x$ + q$
    CASE 3 ' G .000
        IF LEN(q$) THEN
            q$ = LEFT$(q$, 4)
        END IF
        x$ = e$ + x$ + q$
END SELECT
FormatStringX$ = x$
END FUNCTION

' get file date
SUB FileDate (Var$)
DIM ASCIIZ4 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _OFFSET

IF Var$ = Nul THEN
    EXIT SUB
END IF

IF INSTR(Var$, "?") OR INSTR(Var$, "*") THEN
    EXIT SUB
END IF

' construct file date for display.
ASCIIZ4 = Var$ + CHR$(0)
Var$ = Nul
hfind = FindFirstFileA(_OFFSET(ASCIIZ4), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN

    SELECT CASE DateTimeType
        CASE 0
            x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
        CASE 1
            x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
        CASE 2
            x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    END SELECT

    Var$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
    Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
    Var$ = Var$ + LTRIM$(STR$(SysTime.wYear))
    x = FindClose(hfind)
END IF
END SUB

' get file time
SUB FileTime (Var$)
DIM ASCIIZ4 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _OFFSET

IF Var$ = Nul THEN
    EXIT SUB
END IF

IF INSTR(Var$, "?") OR INSTR(Var$, "*") THEN
    EXIT SUB
END IF

' construct file time for display.
ASCIIZ4 = Var$ + CHR$(0)
Var$ = Nul
hfind = FindFirstFileA(_OFFSET(ASCIIZ4), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN

    SELECT CASE DateTimeType
        CASE 0
            x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
        CASE 1
            x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
        CASE 2
            x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    END SELECT

    Var$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
    Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
    Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)
    x = FindClose(hfind)
END IF
END SUB

' get file size
SUB GetFileSize (Var#)
DIM ASCIIZ4 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _OFFSET

IF Out3 = Nul THEN
    EXIT SUB
END IF

IF INSTR(Out3, "?") OR INSTR(Out3, "*") THEN
    EXIT SUB
END IF

' get file size.
ASCIIZ4 = Out3 + CHR$(0)
Var# = 0#
hfind = FindFirstFileA(_OFFSET(ASCIIZ4), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
    Var# = finddata.nFileSizeHigh * &H100000000~&& OR finddata.nFileSizeLow
    x = FindClose(hfind)
END IF
END SUB

' get dir/file attribute
SUB FileAttr (Var#)
DIM ASCIIZ4 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _OFFSET

IF Out3 = Nul THEN
    EXIT SUB
END IF

IF INSTR(Out3, "?") OR INSTR(Out3, "*") THEN
    EXIT SUB
END IF

ASCIIZ4 = Out3 + CHR$(0)
Var# = GetFileAttributes(ASCIIZ4)
IF Var# < 0 THEN
    Var# = 0
END IF
END SUB

' lists all drives
SUB ListDrives
IF Status THEN
    LineBreak = 22
ELSE
    LineBreak = 23
END IF
h = 3
l = 0
GOSUB DriveHeader
FOR c = 1 TO 26
    GOSUB DisplayDrive
    IF h = LineBreak THEN
        h = 0
        IF LCASE$(StdoutFile$) = "scrn:" THEN
            CALL MorePrompt("-more-", "y " + CHR$(13), Output.Char$)
            GOSUB DriveHeader
        END IF
    END IF
NEXT
IF LCASE$(StdoutFile$) = "scrn:" THEN
    IF q = 0 THEN
        PRINT "<none>"
    ELSE
        COLOR White, Black
        PRINT "Total drives"; l
    END IF
END IF
EXIT SUB

DisplayDrive:
c$ = CHR$(c + 64)
Out3 = c$
IF DRIVEEXISTS(c) = 0 THEN
    h = h + 1
    l = l + 1
    q = -1

    ' display drive letter
    COLOR White, Black
    z$ = c$ + ":    "
    PRINT #StdoutHandle, z$;

    ' display volume label
    COLOR Yellow, Black
    Out3 = c$
    CALL Vlabel(Out3)
    IF RTRIM$(Out3) = "" THEN
        z$ = DriveType
    ELSE
        z$ = LEFT$(Out3, 12)
    END IF
    z$ = z$ + SPACE$(13 - LEN(z$))
    PRINT #StdoutHandle, z$;

    ' display volume serial number
    COLOR Green, Black
    Out3 = c$
    CALL Vserial(Out3)
    z$ = LEFT$(Out3, 12)
    z$ = z$ + SPACE$(13 - LEN(z$))
    PRINT #StdoutHandle, z$;

    ' display volume file system type
    COLOR Red, Black
    Out3 = c$
    CALL Vtype(Out3)
    z$ = LEFT$(Out3, 8)
    z$ = z$ + SPACE$(9 - LEN(z$))
    PRINT #StdoutHandle, z$;

    ' display volume total disk space
    COLOR Cyan, Black
    Out3 = c$
    CALL TotalSpace(Out3)
    x# = INT(VAL(Out3))
    x1# = x#
    IF x# > 0# THEN
        CALL Suffix(x#, S$) ' 1,024.0 KB
        z$ = SPACE$(11 - LEN(S$)) + S$
        PRINT #StdoutHandle, z$;
    ELSE
        PRINT #StdoutHandle, "      <n/a>";
    END IF

    ' display volume free disk space
    Out3 = c$
    CALL FreeSpace(Out3)
    y# = INT(VAL(Out3))
    y1# = y#
    IF y# > 0# THEN
        CALL Suffix(y#, S$) ' 1,024.0 KB
        z$ = SPACE$(11 - LEN(S$)) + S$
        PRINT #StdoutHandle, z$;
    ELSE
        PRINT #StdoutHandle, "      <n/a>";
    END IF

    ' display volume used disk space
    IF x1# > 0# OR y1# > 0# THEN
        z# = x1# - y1#
        CALL Suffix(z#, S$) ' 1,024.0 KB
        z$ = SPACE$(11 - LEN(S$)) + S$
        PRINT #StdoutHandle, z$
    ELSE
        PRINT #StdoutHandle, "      <n/a>"
    END IF
END IF
RETURN

DriveHeader:
h = 2
IF LCASE$(StdoutFile$) = "scrn:" THEN
    COLOR White, Black
    PRINT "Drive Label        Serial       Type           Total       Free       Used"
    PRINT "--------------------------------------------------------------------------"
END IF
RETURN
END SUB

' display dirs/files
'   VarQ = 0 display directories,
'         -1 display files
SUB ListFiles (Var4$, VarQ)
DIM ASCIIZ4 AS STRING * 260
DIM finddata AS WIN32_FIND_DATAA
DIM hfind AS _OFFSET

' display header
COLOR White, Black
IF LCASE$(StdoutFile$) = "scrn:" THEN
    PRINT "Searching: " + Var4$
END IF
GOSUB TitleHeader
IF Status THEN
    LineBreak = 22
ELSE
    LineBreak = 23
END IF
d1! = 0!
f1! = 0!
c = 0
q = 0
t = 0
v = 4
ASCIIZ4 = Var4$ + CHR$(0)
hfind = FindFirstFileA(_OFFSET(ASCIIZ4), _OFFSET(finddata))
IF hfind <> INVALID_HANDLE_VALUE THEN
    z = 0
    DO
        OK = 0
        Var# = finddata.dwFileAttributes
        IF VarQ = 0 THEN
            IF (Var# AND &H10) = &H10 THEN
                OK = -1
            END IF
        END IF
        IF VarQ THEN
            IF (Var# AND &H10) = &H0 THEN
                OK = -1
            END IF
        END IF

        X$ = finddata.cFileName
        VarX = INSTR(X$, CHR$(0))
        IF VarX THEN
            X$ = LEFT$(X$, VarX - 1)
        END IF
        X$ = RTRIM$(X$)
        IF X$ = "." OR X$ = ".." THEN
            OK = 0
        END IF
        IF OK THEN
            IF t THEN
                t = 0
                GOSUB TitleHeader
            END IF
            Attr$ = SPACE$(7)
            IF VarQ = 0 THEN
                MID$(Attr$, 2, 1) = "D"
                d1! = d1! + 1!
            ELSE
                f1! = f1! + 1!
            END IF
            q = -1

            ' print shortfilename
            z$ = finddata.cAlternateFileName
            vx = INSTR(z$, CHR$(0))
            IF vx THEN
                z$ = LEFT$(z$, vx - 1)
            END IF
            IF z$ = Nul THEN
                z$ = finddata.cFileName
                vx = INSTR(z$, CHR$(0))
                IF vx THEN
                    z$ = LEFT$(z$, vx - 1)
                END IF
            END IF
            IF LEN(z$) > 12 THEN
                z$ = LEFT$(z$, 12)
            END IF
            COLOR Yellow, Black
            PRINT #StdoutHandle, UCASE$(z$);
            IF LEN(z$) <= 14 THEN
                PRINT #StdoutHandle, SPACE$(14 - LEN(z$));
            END IF

            ' print date/time
            SELECT CASE DateTimeType
                CASE 0
                    x& = FileTimeToSystemTime&(finddata.ftCreationTime, SysTime)
                CASE 1
                    x& = FileTimeToSystemTime&(finddata.ftLastAccessTime, SysTime)
                CASE 2
                    x& = FileTimeToSystemTime&(finddata.ftLastWriteTime, SysTime)
            END SELECT
            Var$ = RIGHT$("00" + LTRIM$(STR$(SysTime.wMonth)), 2) + "-"
            Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wDay)), 2) + "-"
            Var$ = Var$ + LTRIM$(STR$(SysTime.wYear)) + "  "

            Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wHour)), 2) + ":"
            Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wMinute)), 2) + ":"
            Var$ = Var$ + RIGHT$("00" + LTRIM$(STR$(SysTime.wSecond)), 2)

            COLOR Green, Black
            PRINT #StdoutHandle, Var$;

            ' attributes of directory/filename
            Var& = finddata.dwFileAttributes
            IF (Var& AND &H20) = &H20 THEN
                MID$(Attr$, 1, 1) = "A" ' archive
            END IF
            IF (Var& AND &H4) = &H4 THEN
                MID$(Attr$, 3, 1) = "S" ' system
            END IF
            IF (Var& AND &H2) = &H2 THEN
                MID$(Attr$, 4, 1) = "H" ' hidden
            END IF
            IF (Var& AND &H1) = &H1 THEN
                MID$(Attr$, 5, 1) = "R" ' read-only
            END IF
            IF (Var& AND &H800) = &H800 THEN
                MID$(Attr$, 6, 1) = "C" ' compressed
            END IF
            IF (Var& AND &H4000) = &H4000 THEN
                MID$(Attr$, 7, 1) = "E" ' encrypted
            END IF
            COLOR Red, Black
            Attr$ = " " + Attr$ + " "
            PRINT #StdoutHandle, Attr$;

            ' print filesize
            z$ = "<DIR>"
            IF VarQ THEN
                Var# = finddata.nFileSizeHigh * &H100000000~&&
                Var# = Var# OR finddata.nFileSizeLow
                VarX# = VarX# + Var# ' add bytes
                CALL Suffix(Var#, z$) ' 1,024.0 KB
            END IF
            z$ = LEFT$(z$, 10)
            z$ = SPACE$(10 - LEN(z$)) + z$ + " "
            COLOR Cyan, Black
            PRINT #StdoutHandle, z$;

            ' print longfilename
            z$ = finddata.cFileName
            VarX = INSTR(z$, CHR$(0))
            IF VarX THEN
                z$ = LEFT$(z$, VarX - 1)
            END IF
            z$ = RTRIM$(z$)
            IF LEN(z$) THEN
                IF LEN(z$) > 23 THEN
                    z$ = LEFT$(z$, 22) + "..."
                END IF
                COLOR Yellow, Black
                PRINT #StdoutHandle, z$;
            END IF
            PRINT #StdoutHandle, ""

            ' prompt
            v = v + 1
            IF v = LineBreak THEN
                v = 3
                IF c = 0 THEN
                    IF LCASE$(StdoutFile$) = "scrn:" THEN
                        COLOR White, Black
                        PRINT "-more(y/n/c/q)-";
                        X$ = Nul
                        DO
                            _LIMIT 50
                            X$ = INKEYx$
                            IF LEN(X$) THEN
                                SELECT CASE LCASE$(X$)
                                    CASE "c" ' continuous
                                        c = -1
                                        EXIT DO
                                    CASE "n", "q" ' no/quit
                                        v = 0
                                        PRINT
                                        GOTO ExitLoop
                                    CASE " ", CHR$(13), "y" ' yes/continue
                                        EXIT DO
                                END SELECT
                            END IF
                        LOOP
                        PRINT
                        t = -1
                    END IF
                END IF
            END IF
        END IF
    LOOP WHILE FindNextFileA(hfind, _OFFSET(finddata))
    x = FindClose(hfind)
END IF
IF q = 0 THEN
    IF LCASE$(StdoutFile$) = "scrn:" THEN
        COLOR Yellow, Black
        PRINT "No files found."
    END IF
END IF
ExitLoop:

' print totals
COLOR White, Black
IF LCASE$(StdoutFile$) = "scrn:" THEN
    PRINT "------------                               ----------"
    TotalLine$ = "Files " + FormatString$(CDBL(f1!))
    TotalLine$ = TotalLine$ + " Dirs " + FormatString$(CDBL(d1!))
    TotalLine$ = LEFT$(TotalLine$, 40)
    TotalLine$ = TotalLine$ + SPACE$(42 - LEN(TotalLine$))

    CALL Suffix(VarX#, z$) ' 1,024.0 KB
    TotalLine$ = TotalLine$ + SPACE$(11 - LEN(z$)) + z$
    PRINT TotalLine$
END IF
EXIT SUB

TitleHeader:
IF LCASE$(StdoutFile$) = "scrn:" THEN
    COLOR White, Black
    PRINT "Filename      Date        Time     Attr          Size Longfilename"
    PRINT "------------  ----------  -------- ------- ---------- ------------"
END IF
RETURN
END SUB

